/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to pick algorithms for calculating Life.
 */

#include "life.h"
#include "algorithm.h"
#include "listalg.h"
#include "quadalg.h"


//
// Forward declarations.
//
static BOOL ConvertFromListToQuadCell(LifeObject * obj);
static BOOL ConvertFromQuadCellToList(LifeObject * obj);

static BOOL ConvertCallback(NUMBER row, NUMBER col, int state, void * data);


/*
 * Convert the cell data for an object to use the specified algorithm.
 * If the object is too large this cannot be done (i.e., if a large
 * quadcell object is converted to a list).  Returns TRUE if the
 * conversion was able to be performed.  Warning: This can perform
 * garbage collections on the quadcells so no temporary quadcells
 * can be held by the caller.
 */
BOOL
SetObjectAlgorithm(LifeObject * obj, LifeAlgorithm * algorithm)
{
	//
	// If the algorithm isn't changing then do nothing.
	//
	if (algorithm == obj->algorithm)
		return TRUE;

	//
	// If the object is empty then we just have to set the algorithm.
	//
	if (obj->algorithm->isObjectEmpty(obj))
	{
		obj->algorithm = algorithm;

		return TRUE;
	}

	//
	// Something really has to be done to the cells of the object.
	// Save the number of live cells since it can go wrong.
	//
	NUMBER savedCount = obj->count;
	BOOL savedIsCountUnknown = obj->isCountUnknown;

	//
	// Switch on the combination of the old and new algorithms.
	//
	if ((obj->algorithm == &quadAlgorithm) &&
		(algorithm == &listAlgorithm))
	{
		if (!ConvertFromQuadCellToList(obj))
			return FALSE;
	}
	else if ((obj->algorithm == &listAlgorithm) &&
		(algorithm == &quadAlgorithm))
	{
		if (!ConvertFromListToQuadCell(obj))
			return FALSE;
	}
	else
		Error("Unknown algorithm conversion");

	//
	// Now set the algorithm and restore the count.
	//
	obj->algorithm = algorithm;
	obj->count = savedCount;
	obj->isCountUnknown = savedIsCountUnknown;

	return TRUE;
}


/*
 * Convert an object from the quadcell to the list algorithm.
 * Returns TRUE if the conversion could be done, or FALSE if the
 * quadcell is too large or if the user wanted to stop.
 * This only processes ON cells.
 */
static BOOL
ConvertFromQuadCellToList(LifeObject * obj)
{
	QuadCell * cells = qcGetReducedQuadCell(obj);

	//
	// If the object is too large then fail.
	//
	NUMBER count = qcGetCount(cells);

	if (numCmpLong(count, MAX_WRITE) > 0)
		return FALSE;

	//
	// Make sure the list data is empty.
	// The quadcell scan will add new cells to the list.
	//
	listAlgorithm.deleteAllCells(obj);

	//
	// Perform the scan over the ON cells supplying a callback
	// routine and our life object as its data value.
	//
	if (!qcScanCells(cells, STATE_ON, ConvertCallback, obj))
		return FALSE;

	//
	// Free the quadcell data.
	//
	obj->algorithm->deleteAllCells(obj);
	obj->algorithm->freeCellData(obj);

	return TRUE;
}


/*
 * Callback routine to handle the ON cells as they are found.
 * Note: The supplied data pointer holds the object being scanned.
 * Returns FALSE if user wanted to abort the conversion.
 */
static BOOL
ConvertCallback(NUMBER row, NUMBER col, int state, void * data)
{
	LifeObject * obj = data;

	listAlgorithm.addCell(obj, row, col);

	return !Life.stop;
}


/*
 * Convert an object from the list to the quadcell algorithm.
 * Warning: This calls garbage collections as it is run since lots of
 * temporary quadcells are constructed.  Because of this, we must store
 * the quadcell in the object as we go so that it will be registered
 * for the garbage collections.  This always succeeds.
 */
static BOOL
ConvertFromListToQuadCell(LifeObject * obj)
{
	const LifeRow * rp;
	const LifeCell * cp;

	qcSetQuadCell(obj, qcEmpty());

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if (Life.stop)
				return FALSE;

			qcSetQuadCell(obj, qcSet(obj->quadCells, rp->row, cp->col));

			CheckGarbageCollect();
		}
	}

	//
	// Free the list data.
	//
	obj->algorithm->deleteAllCells(obj);
	obj->algorithm->freeCellData(obj);

	return TRUE;
}


/*
 * Return the name of the given algorithm.
 * Returns 0 if the algorithm is not known.
 */
const char *
GetAlgorithmName(LifeAlgorithm * algorithm)
{
	if (algorithm == &listAlgorithm)
		return LIST_NAME;

	if (algorithm == &quadAlgorithm)
		return QUAD_NAME;

	return 0;
}


/*
 * Return the algorithm structure given its name.
 * Abbreviations are allowed.
 * Returns 0 if the name is unknown.
 */
LifeAlgorithm *
GetAlgorithmByName(const char * name)
{
	if (IsAbbreviation(name, LIST_NAME))
		return &listAlgorithm;

	if (IsAbbreviation(name, QUAD_NAME))
		return &quadAlgorithm;

	return 0;
}

