/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines which implement a particular algorithm for manipulating
 * Life objects.
 */

#ifndef	ALGORITHM_H
#define	ALGORITHM_H

#include <stdio.h>

#include "life.h"


/*
 * The structure holding function pointers for the routines to manipulate
 * the cell data for Life objects.  This allows different data formats to
 * be used for holding the cells of the object.  (i.e., list based or
 * quadcell based.)
 */

struct LifeAlgorithm
{
	//
	// Object manipulations.
	//
	void (*allocCellData)(LifeObject * obj);
	void (*freeCellData)(LifeObject * obj);
	int (*getCellState)(const LifeObject * obj, NUMBER row, NUMBER col);
	void (*setCellState)(LifeObject * obj, NUMBER row, NUMBER col, int state);
	BOOL (*isCellSet)(const LifeObject * obj, NUMBER row, NUMBER col);
	BOOL (*addCell)(LifeObject * obj, NUMBER row, NUMBER col);
	BOOL (*deleteCell)(LifeObject * obj, NUMBER row, NUMBER col);
	void (*deleteAllCells)(LifeObject * obj);
	void (*shiftObject)(LifeObject * obj, NUMBER leftOrRight, NUMBER upOrDown);
	void (*saveUsedCells)(LifeObject * obj);
	void (*registerQuadCells)(LifeObject * obj);
	void (*copyObject)(const LifeObject * srcObj, LifeObject * destObj);
	void (*moveObject)(LifeObject * srcObj, LifeObject * destObj);
	void (*addObject)(const LifeObject * srcObj, LifeObject * destObj,
		NUMBER rowDisp, NUMBER colDisp);

	BOOL (*minMax)(const LifeObject * obj,
		NUMBER * retMinRow, NUMBER * retMaxRow,
		NUMBER * retMinCol, NUMBER * retMaxCol);

	NUMBER (*countRegion)(const LifeObject * obj,
		NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

	ViewInfo (*calculateView)(const LifeObject * obj, int viewId, int scale,
		NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

	NUMBER (*calculateCellCount)(const LifeObject * obj);
	BOOL (*isObjectEmpty)(const LifeObject * obj);
	BOOL (*searchObject)(LifeObject * obj, Count count);

	BOOL (*nextGeneration)(LifeObject * obj);
	Count (*runGenerations)(LifeObject * obj, Count genCount);

	//
	// Marking and selecting operations.
	//
	BOOL (*selectObject)(LifeObject * obj, NUMBER row, NUMBER col, Value gap);
	BOOL (*selectAll)(LifeObject * obj);
	BOOL (*markSelection)(LifeObject * obj);
	BOOL (*selectMarkedCells)(LifeObject * obj);
	void (*invertMarks)(LifeObject * obj);
	void (*clearMarks)(LifeObject * obj);
	void (*clearSearch)(LifeObject * obj);
	void (*clearTracks)(LifeObject * obj);
	void (*clearSelection)(LifeObject * obj);
	BOOL (*selectPath)(LifeObject *);
	void (*countPathCells)(const LifeObject *, NUMBER *, NUMBER *, NUMBER *);
	NUMBER (*findMarkedGap)(const LifeObject *);
	BOOL (*selectCell)(LifeObject * obj, NUMBER row, NUMBER col);

	BOOL (*findClosestSelectedCell)(const LifeObject *,
		NUMBER, NUMBER, NUMBER *, NUMBER *);

	BOOL (*selectRegion)(LifeObject * obj,
		NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

	BOOL (*markMinMax)(const LifeObject * obj,
		NUMBER * minRow, NUMBER * maxRow, NUMBER * minCol, NUMBER * maxCol);

	NUMBER (*countMarks)(const LifeObject * obj);

	//
	// Object manipulations using selections.
	//
	void (*moveSelection)(LifeObject * srcObj, LifeObject * destObj);

	void (*copyMarkedObject)(const LifeObject * srcObj, LifeObject * destObj);

	void (*rotateSelection)(LifeObject * obj);
	void (*flipColumnSelection)(LifeObject * obj);
	void (*flipRowSelection)(LifeObject * obj);

	//
	// Reading and writing of objects to files.
	//
	BOOL (*scanObject)(LifeObject * obj, int states,
		BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
		void * data);

	NUMBER (*getObjectMaxRowUptoLimit)(LifeObject * obj, NUMBER limitRow);
	NUMBER (*getObjectMinRowDowntoLimit)(LifeObject * obj, NUMBER limitRow);

	QuadCell * (*extractQuadCell)(LifeObject * obj);
};

typedef	struct LifeAlgorithm LifeAlgorithmStruct;

#endif
