/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "life.h"
#include "listalg.h"
#include "quadalg.h"


static void * SafeAlloc(int);


/*
 * Allocate a new object structure.
 * We quit if the allocation fails.
 */
LifeObject *
AllocateObject(void)
{
	LifeObject * obj = Life.freeObjects;

	if (obj)
		Life.freeObjects = obj->nextObject;
	else
	{
		obj = Life.newObjects;

		if (obj >= Life.endObjects)
		{
			obj = SafeAlloc(sizeof(LifeObject) * ALLOC_OBJ_SIZE);

			Life.newObjects = obj;
			Life.endObjects = obj + ALLOC_OBJ_SIZE;
		}

		Life.newObjects++;
	}

	memset(obj, 0, sizeof(LifeObject));

	obj->nextObject = 0;
	obj->algorithm = Life.defaultAlgorithm;
	obj->rule = Life.defaultRule;
	obj->isReserved = Life.isCreatingReserved;
	obj->changeCount = 0;
	obj->count = numZero();
	obj->gen = 0;
	obj->genLeft = 0;
	obj->scaleListCount = 0;
	obj->scaleListIndex = 0;
	obj->mouseRow = numZero();
	obj->mouseCol = numZero();
	obj->dragMinRow = numZero();
	obj->dragMaxRow = numZero();
	obj->dragMinCol = numZero();
	obj->dragMaxCol = numZero();
	obj->overlayObject = 0;
	obj->isOverlayActive = FALSE;
	obj->isJumpingEnabled = FALSE;
	obj->isCountUnknown = FALSE;
	obj->isReducedValid = FALSE;
	obj->isTracksUsed = Life.defaultTracks;
	obj->isTracksVisible = TRUE;
	obj->frequency = Life.defaultFrequency;
	obj->isJumpingEnabled = Life.defaultIsJumpingEnabled;
	obj->freqLeft = obj->frequency;
	obj->stopIsEnabled = FALSE;
	obj->stopGap = 0;
	obj->stopOnPath = FALSE;
	obj->stopInsidePath = FALSE;
	obj->stopOutsidePath = FALSE;
	obj->stopCell = FALSE;
	obj->stopState = FALSE;
	obj->stopRow = numZero();
	obj->stopCol = numZero();
	obj->windowIsAutoScale = FALSE;
	obj->windowScale = Life.initScale;
	obj->stopExpr[0] = '\0';

	InitStringBuffer(&obj->name);
	InitStringBuffer(&obj->comments);

	InitAllLocationInfo(obj);
	InitializePath(obj);

	//
	// Allocate data for all algorithms.
	//
	listAlgorithm.allocCellData(obj);
	quadAlgorithm.allocCellData(obj);

	return obj;
}


/*
 * Allocate a new rule structure.
 */
LifeRule *
AllocateRule(void)
{
	LifeRule * rp = SafeAlloc(sizeof(LifeRule));

	rp->next = 0;
	rp->name[0] = '\0';

	return rp;
}


/*
 * Initialise the specified location info.
 */
void
InitLocationInfo(LocationInfo * info, int id)
{
	memset(info, 0, sizeof(LocationInfo));

	info->id = id;
	info->row = numZero();
	info->col = numZero();
}


/*
 * Initialise the where locations info.
 * This includes the current location and the origin.
 */
void
InitAllLocationInfo(LifeObject * obj)
{
	int index;

	InitLocationInfo(&obj->originInfo, LOCINFO_ID_ORIGIN);
	InitLocationInfo(&obj->currentInfo, LOCINFO_ID_CURRENT);

	for (index = 0; index < 26; index++)
		InitLocationInfo(&obj->whereInfo[index], 'a' + index);
}


/*
 * Clear the specified location info show data.
 * This does not change the id, row, or column fields.
 */
void
ClearLocationInfo(LocationInfo * info)
{
	info->isEnabled = FALSE;
	info->isIdShown = FALSE;
	info->isForwShown = FALSE;
	info->isBackShown = FALSE;
	info->isHorizShown = FALSE;
	info->isVertShown = FALSE;
	info->isDistShown = FALSE;
}


/*
 * Clear the where locations shown data, and the current location and
 * origin shown data.  This does not change the id, row, or column fields.
 */
void
ClearAllLocationInfo(LifeObject * obj)
{
	int index;

	ClearLocationInfo(&obj->originInfo);
	ClearLocationInfo(&obj->currentInfo);

	for (index = 0; index < 26; index++)
		ClearLocationInfo(&obj->whereInfo[index]);
}


/*
 * Allocate the elements involved with the object's cell data.
 */
void
lcAllocCellData(LifeObject * obj)
{
	obj->firstRow = Life.termRow;
	obj->lastRow = 0;
	obj->usedQuadCells = qcEmpty();
}


/*
 * Allocate a new row structure.
 * It contains one row element, the termCell.
 * We quit if the allocation fails.
 */
LifeRow *
lcAllocateRow(void)
{
	LifeRow * rp = Life.freeRows;

	if (rp)
		Life.freeRows = rp->nextRow;
	else
	{
		rp = Life.newRows;

		if (rp >= Life.endRows)
		{
			rp = SafeAlloc(sizeof(LifeRow) * ALLOC_ROW_SIZE);

			Life.newRows = rp;
			Life.endRows = rp + ALLOC_ROW_SIZE;
		}

		Life.newRows++;
	}

	rp->nextRow = 0;
	rp->firstCell = Life.termCell;
	rp->lastCell = 0;
	rp->count = numZero();

	return rp;
}


/*
 * Allocate a new cell structure.
 * We quit if the allocation fails.
 */
LifeCell *
lcAllocateCell(void)
{
	LifeCell * cp = Life.freeCells;

	if (cp)
	{
		Life.freeCells = cp->nextCell;
		cp->nextCell = 0;
		cp->marks = 0;

		return cp;
	}

	cp = Life.newCells;

	if (cp >= Life.endCells)
	{
		cp = SafeAlloc(sizeof(LifeCell) * ALLOC_CELL_SIZE);

		Life.newCells = cp;
		Life.endCells = cp + ALLOC_CELL_SIZE;
	}

	Life.newCells++;
	cp->nextCell = 0;
	cp->marks = 0;

	return cp;
}


/*
 * Free a list of rows along with all their cells.
 * The list is terminated either with a NULL or with the termRow element.
 * This returns the total number of cells that were freed.
 */
NUMBER
lcFreeRows(LifeRow * firstRow)
{
	LifeRow * rp;
	LifeRow * lastRow;
	NUMBER count;

	lastRow = 0;
	count = numZero();

	/*
	 * Loop over all the rows, freeing the cells within them.
	 */
	for (rp = firstRow; rp && (rp != Life.termRow); rp = rp->nextRow)
	{
		if (rp->firstCell != Life.termCell)
		{
			rp->lastCell->nextCell = Life.freeCells;
			Life.freeCells = rp->firstCell;

			numAddrAdd(&count, rp->count);
		}

		lastRow = rp;
	}

	/*
	 * Add the list of rows to the free row list.
	 */
	if (lastRow)
	{
		lastRow->nextRow = Life.freeRows;
		Life.freeRows = firstRow;
	}

	return count;
}


/*
 * Copy a list of rows and cells.
 * Both the new and old lists are terminated with termRow entries.
 */
LifeRow *
lcCopyRows(const LifeRow * oldrp)
{
	LifeRow * rp;
	LifeRow * retRp;
	LifeRow * prevRp;
	LifeCell * oldCp;
	LifeCell * cp;
	LifeCell * prevCp;

	retRp = 0;
	prevRp = 0;

	/*
	 * Loop over each row in the old list copying them to the new list.
	 */
	for (; oldrp != Life.termRow; oldrp = oldrp->nextRow)
	{
		rp = lcAllocateRow();
		rp->row = oldrp->row;

		if (prevRp)
			prevRp->nextRow = rp;
		else
			retRp = rp;

		prevRp = rp;

		/*
		 * Copy all of the cells in the row.
		 */
		prevCp = 0;

		for (oldCp = oldrp->firstCell;
			oldCp != Life.termCell; oldCp = oldCp->nextCell)
		{
			cp = lcAllocateCell();
			cp->col = oldCp->col;

			if (prevCp)
				prevCp->nextCell = cp;
			else
				rp->firstCell = cp;

			prevCp = cp;
		}

		if (prevCp)
		{
			prevCp->nextCell = Life.termCell;
			rp->lastCell = prevCp;
		}
		else
			rp->firstCell = Life.termCell;
	}

	if (prevRp == 0)
		return Life.termRow;

	prevRp->nextRow = Life.termRow;

	return retRp;
}


/*
 * Initialise a string buffer for use.
 * Calling this is optional for a static instance of a string buffer
 * since it will have an allocation size of 0, and then the data pointer
 * won't be accessed.
 */
void
InitStringBuffer(StringBuffer * buf)
{
	buf->alloc = 0;
	buf->len = 0;
	buf->data = 0;
}


/*
 * Free a string buffer which is no longer being used.
 */
void
TermStringBuffer(StringBuffer * buf)
{
	if (buf->alloc > 0)
		free(buf->data);

	buf->alloc = 0;
	buf->len = 0;
	buf->data = 0;
}


/*
 * Add a character to a string buffer.
 * Nulls are ignored.
 */
void
AddCharToStringBuffer(StringBuffer * buf, int ch)
{
	char str[2];

	str[0] = ch;
	str[1] = '\0';

	AppendToStringBuffer(buf, str);
}


/*
 * Append a string to a string buffer.
 * The buffer is resized if necessary, with a larger then needed
 * size to anticipate future calls.
 */
void
AppendToStringBuffer(StringBuffer * buf, const char * str)
{
	int newLen = buf->len + strlen(str);

	if (newLen >= buf->alloc)
	{
		int newAlloc = newLen + 256;

		char * newData = SafeAlloc(newAlloc);

		if (buf->len > 0)
			memcpy(newData, buf->data, buf->len);

		if (buf->alloc > 0)
			free(buf->data);

		buf->data = newData;
		buf->alloc = newAlloc;
	}

	strcpy(&buf->data[buf->len], str);
	buf->len = newLen;
}


/*
 * Set the text for a string buffer.
 */
void
SetStringBuffer(StringBuffer * buf, const char * str)
{
	buf->len = 0;

	AppendToStringBuffer(buf, str);
}


/*
 * Clear the data in a string buffer.
 * This doesn't free any memory.
 */
void
ClearStringBuffer(StringBuffer * buf)
{
	buf->len = 0;
}


/*
 * Set the length of the string in the string buffer.
 * This can only shrink the existing text.
 */
void
SetStringBufferLength(StringBuffer * buf, int len)
{
	if ((len < 0) || (len > buf->len))
		Error("Cannot increase length of string buffer");

	if (len == buf->len)
		return;

	buf->data[len] = '\0';
	buf->len = len;
}


/*
 * Accessor function to return the string in a string buffer.
 */
const char *
GetStringBufferText(const StringBuffer * buf)
{
	if (buf->len == 0)
		return "";

	return buf->data;
}


/*
 * Accessor function to return the length of the string in a string buffer.
 * This doesn't count the terminating null byte.
 */
int
GetStringBufferLength(const StringBuffer * buf)
{
	return buf->len;
}


/*
 * Local memory allocation routine which exits if there is no memory.
 */
static void *
SafeAlloc(int len)
{
	void *	cp;

	cp = malloc(len);

	if (cp == 0)
	{
		fprintf(stderr, "Memory allocation failed\n");

		Life.device->closeDevice(Life.device);

		exit(1);
	}

	return cp;
}

/* END CODE */
