/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "life.h"


static	int	ReadValue(NUMBER *, BOOL *);
static	NUMBER	ScanExpression(void);
static	void	DoSelect(LifeWindow *, int, BOOL);
static	void	DoMove(LifeWindow *, NUMBER, NUMBER);
static	void	DoShift(LifeWindow *, NUMBER, NUMBER, NUMBER);
static	void	DoSearch(LifeWindow *, NUMBER);
static	void	ToggleCell(LifeObject *);
static	void	BackgroundChanged(const LifeObject *);
static	void	PositionChanged(const LifeObject *);
static	void	ViewChanged(const LifeObject *);
static	void	StatusChanged(const LifeObject *);


/*
 * Read as many commands as there are available.
 * This does not block.
 */
void
DoCommands(void)
{
	while (DoOneCommand())
		;
}


/*
 * Read a character mode command if available and execute it.
 * If a complete character mode command is not yet available
 * then nothing is done (except storing the partial command)
 * and we return anyway so that we are non-blocking.
 *
 * This allows commands to be typed without stopping the
 * computation of and display of the generations of an object
 * and allows editing of partially completed commands.
 *
 * Character mode commands are read using ScanChar, which uses
 * setjmp/longjmp to abort command parsing when there is not
 * enough input. The code after the setjmp can be reentered many
 * times in order to finish any command.  Therefore it is very
 * important that no important state is changed until a command
 * is completely parsed.
 *
 * If the current window is changed by a mouse or keystroke event
 * then any partially completed character mode command is cleared
 * to prevent confusion for the user.
 *
 * However, once the beginning of a line mode command is seen
 * then ScanChar will no longer abort due to lack of input.
 * Instead ReadLine is used based on a buffer associated with
 * the current window.  Command line display and editing is
 * performed by that routine.
 *
 * Returns TRUE if a command was successfully completed.
 */
BOOL
DoOneCommand(void)
{
	LifeDevice *	dev;
	int	ch;		/* character read */
	LifeWindow * window;	/* current window */
	LifeObject * obj;		/* current object in window */
	BOOL	got1;		/* got first argument flag */
	BOOL	got2;		/* got second argument flag */
	NUMBER	defaultArg;	/* first argument defaulted to one */
	NUMBER	arg1;		/* first command argument */
	NUMBER	arg2;		/* second command argument */
	NUMBER	retRow;		/* returned row */
	NUMBER	retCol;		/* returned column */

	/*
	 * Setjmp location for reading character mode commands
	 * without blocking.  Control will return here if there
	 * is not enough input to complete a command or if the
	 * partial command is edited.
	 */
	switch (setjmp(Life.ttyJmp.buf))
	{
		case SCAN_EOF:
			/*
			 * Not enough chars were yet available for a
			 * complete command.  Return having done nothing
			 * important except for the storing of any new
			 * characters that were read.
			 */
			return FALSE;

		case SCAN_ABORT:
			return TRUE;

		case SCAN_EDIT:
		default:
			/*
			 * Command edited before completion, or normal
			 * command completion, or normal entry point.
			 */
			break;
	}

	/*
	 * At this point we are at the beginning of parsing of a
	 * command.  We can simply act like all of the required
	 * characters for a command will be available, and if
	 * they are not available then a longjmp will be made
	 * back to the setjmp location above.
	 *
	 * Note: Because the code from this point onwards can be
	 * repeatedly executed until a command is is fully read,
	 * we cannot change any important state until then.
	 */
	if (Life.stop)
		Error("Command aborted");

	CheckStartupFiles();

	dev = Life.device;
	window = Life.currentWindow;
	obj = window->currentObject;
	Life.currentObject = obj;

	arg2 = numZero();
	got2 = FALSE;
	Life.canMakeInput = TRUE;

	/*Life.device->setViewModeIsText(dev, FALSE);*/

	/*
	 * Read a command character along with any possible count values.
	 */
	ch = ReadValue(&arg1, &got1);

	if (ch == ',')
		ch = ReadValue(&arg2, &got2);

	defaultArg = numOne();

	if (got1)
		defaultArg = arg1;

	/*
	 * Switch on the command character.
	 */
	switch (ch)
	{
		case FF:
			/*
			 * Refresh screen
			 */
			dev->redraw(dev);

			break;

		case '\n':
			/*
			 * Move to next row
			 */
			SetCurrentLocation(obj,
				numAdd(obj->currentInfo.row, defaultArg),
				obj->originInfo.col);

			PositionChanged(obj);

			break;

		case '\t':
			/*
			 * move to next tab stop
			 */
			do
			{
				numAddrInc(&obj->currentInfo.col);
			}
			while (numModLong(numSub(obj->currentInfo.col,
				obj->originInfo.col), 8) != 0);

			PositionChanged(obj);

			break;

		case ESC:
			/*
			 * An escape can either begin a cursor movement
			 * sequence or the invocation of a macro.
			 */
			ch = ScanChar();

			/*
			 * Ignore double escapes.
			 */
			if (ch == ESC)
				break;

			/*
			 * Look for the standard cursor key introduction.
			 * If found, then move in the appropriate direction.
			 */
			if (ch == '[')
			{
				ch = ScanChar();

				switch (ch)
				{
					case 'A':
						DoMove(window, numNeg(defaultArg),
							numZero());

						break;

					case 'B':
						DoMove(window, defaultArg,
							numZero());

						break;

					case 'C':
						DoMove(window, numZero(),
							defaultArg);

						break;

					case 'D':
						DoMove(window, numZero(),
							numNeg(defaultArg));

						break;

					default:
						Error("Unknown escape sequence");
				}

				break;
			}

			/*
			 * This executes a macro.
			 */
			Backup();

			if (!SetMacro(arg1, arg2, ch))
				Error("Undefined macro");

			StatusChanged(obj);

			break;

		case ' ':
		case '.':
		case '?':
			/*
			 * Move to right
			 */
			AdjustCurrentLocation(obj, numZero(), defaultArg);
			PositionChanged(obj);

			break;

		case ':':
		case ';':
		case '\'':
			/*
			 * A line mode command is being started.
			 * The character mode command is complete at
			 * this point, so clear it from the scan buffer.
			 */
			ScanReset();

			/*
			 * Now do special handling for the command line.
			 * The input for the command line will not use
			 * ScanChar, and so the setjmp above will not be
			 * called until the whole command is completed.
			 */
			DoLineCommand(arg1, arg2, got1, got2);

			break;

		case '<':
			/*
			 * Begin loop or macro definition
			 */
			if (got1 || got2)
			{
				if (got2)
					SetLoop(defaultArg, arg2, 0);
				else
					SetLoop(numOne(), defaultArg, 0);

				StatusChanged(obj);

				break;
			}

			ch = ScanChar();

			if (!MacroIsValid(ch))
				Error("Bad macro character");

			SetLoop(numOne(), numOne(), ch);
			StatusChanged(obj);

			break;

		case '>':
			/*
			 * End loop
			 */
			EndLoop();

			break;

		case '+':
			/*
			 * Increment single char variable
			 */
			ch = ScanChar();

			if (ch == '$')
				ch = ScanChar();

			SetVariable1(ch, numAdd(GetVariable1(ch), defaultArg));

			break;

		case 'a':
			/*
			 * Add coordinate to path
			 */
			CheckRun();

			if (!AddPointToPath(obj, obj->currentInfo.row,
				obj->currentInfo.col))
			{
				Error("Too many coordinates in path");
			}

			break;

		case 'A':
			/*
			 * Add closure to path
			 */
			CheckRun();

			ClosePath(obj);

			break;

		case 'b':
			/*
			 * Move lower left with action
			 */
			DoMove(window, defaultArg, numNeg(defaultArg));

			break;

		case 'B':
			/*
			 * Shift to lower left
			 */
			DoShift(window, defaultArg, numNeg(defaultArg), arg2);

			break;

		case 'c':
			/*
			 * Pick cell as current location
			 */
			SetCurrentLocation(obj,
				numAdd(obj->originInfo.row, arg1),
				numAdd(obj->originInfo.col, arg2));

			PositionChanged(obj);

			break;

		case 'C':
			/*
			 * Move to closest marked cell.
			 */
			DoSelect(window, ch, TRUE);

			if (FindClosestSelectedCell(obj, obj->currentInfo.row,
				obj->currentInfo.col, &retRow, &retCol))
			{
				SetCurrentLocation(obj, retRow, retCol);
				PositionChanged(obj);
			}

			break;

		case 'd':
			/*
			 * Delete selection
			 */
			DoSelect(window, ch, FALSE);
			Backup();
			MoveSelection(obj, Life.deleteObject);
			ViewChanged(obj);

			break;

		case 'D':
			/*
			 * Set the drag area.
			 * This accepts relative widths and heights to be
			 * set from the current position.  This requires
			 * zero or two arguments.
			 */
			if (got1 && !got2)
				Error("Drag requires 0 or 2 arguments");

			if (got2 && (numIsZero(arg1) || numIsZero(arg2)))
				Error("Invalid drag value");

			/*
			 * Set the default drag area to the current cell.
			 */
			obj->dragMinRow = obj->currentInfo.row;
			obj->dragMaxRow = obj->currentInfo.row;
			obj->dragMinCol = obj->currentInfo.col;
			obj->dragMaxCol = obj->currentInfo.col;

			/*
			 * If no arguments were given then the default
			 * drag area was right.
			 */
			if (!got2)
				break;

			if (numIsPos(arg1))
				numAddrAdd(&obj->dragMaxCol, numDec(arg1));
			else
				numAddrAdd(&obj->dragMinCol, numInc(arg1));

			if (numIsPos(arg2))
				numAddrAdd(&obj->dragMaxRow, numDec(arg2));
			else
				numAddrAdd(&obj->dragMinRow, numInc(arg2));

			break;

		case 'f':
			/*
			 * Flip selection
			 */
			ch = ScanChar();

			if ((ch != 'r') && (ch != 'c'))
				Error("Bad flip axis");

			DoSelect(window, 'f', FALSE);
			Backup();

			if (ch == 'r')
				FlipRowSelection(obj);
			else
				FlipColumnSelection(obj);

			ViewChanged(obj);

			break;

		case 'g':
			/*
			 * Compute or stop generations
			 */
			if (obj->genLeft <= 0)
				Backup();

			if (got1)
			{
				if (obj->genLeft != INFINITY)
					obj->genLeft += numToLong(arg1);
			}
			else if (obj->genLeft)
			{
				obj->genLeft = 0;
				ViewChanged(obj);
			}
			else
				obj->genLeft = 1;

			obj->freqLeft = obj->frequency;
			StatusChanged(obj);

			break;

		case 'G':
			/*
			 * Compute infinite generations
			 */
			if (obj->genLeft <= 0)
				Backup();

			obj->genLeft = INFINITY;
			obj->freqLeft = obj->frequency;
			StatusChanged(obj);

			break;

		case 'h':
			/*
			 * Move left with action
			 */
			DoMove(window, numZero(), numNeg(defaultArg));

			break;

		case 'H':
			/*
			 * Shift left lots
			 */
			DoShift(window, numZero(), numNeg(defaultArg), arg2);

			break;

		case 'j':
			/*
			 * Move down with action
			 */
			DoMove(window, defaultArg, numZero());

			break;

		case 'J':
			/*
			 * Shift down lots
			 */
			DoShift(window, defaultArg, numZero(), arg2);

			break;

		case 'k':
			/*
			 * Move up with action
			 */
			DoMove(window, numNeg(defaultArg), numZero());

			break;

		case 'K':
			/*
			 * Shift up lots
			 */
			DoShift(window, numNeg(defaultArg), numZero(), arg2);

			break;

		case 'l':
			/*
			 * Move right with action
			 */
			DoMove(window, numZero(), defaultArg);

			break;

		case 'L':
			/*
			 * Shift right lots
			 */
			DoShift(window, numZero(), defaultArg, arg2);

			break;

		case 'm':
			/*
			 * Mark selected cells.
			 */
			DoSelect(window, ch, FALSE);
			Backup();
			MarkSelection(obj);
			ViewChanged(obj);

			break;

		case 'n':
			/*
			 * Move lower right with action
			 */
			DoMove(window, defaultArg, defaultArg);

			break;

		case 'N':
			/*
			 * Shift down and right
			 */
			DoShift(window, defaultArg, defaultArg, arg2);

			break;

		case 'o':
		case 'O':
		case '*':
			/*
			 * Insert new cells
			 
			 */
			CheckRun();
			Backup();

			while (!Life.stop && numIsPos(defaultArg))
			{
				defaultArg = numDec(defaultArg);

				SetCellState(obj, obj->currentInfo.row,
					obj->currentInfo.col, STATE_ON);

				numAddrInc(&obj->currentInfo.col);

				CheckGarbageCollect();
			}

			break;

		case 'p':
			/*
			 * Place deleted object
			 */
			CheckRun();
			Backup();
			AddObject(Life.deleteObject, obj, ADD_RELATIVE);
			ViewChanged(obj);

			break;

		case 'r':
			/*
			 * Rotate selection
			 */
			DoSelect(window, ch, FALSE);
			Backup();
			RotateSelection(obj);
			ViewChanged(obj);

			break;

		case 's':
			/*
			 * Set scale factor and center object
			 */
			SetWindowIsAutoScale(window, FALSE);

			if (!got1)
				arg1 = numFromLong(window->scale);

			SetScale(window, numToLong(arg1));

			StatusChanged(obj);

			break;

		case 'S':
			/*
			 * Perform auto-scaling
			 */
			if (!got1)
				arg1 = numFromLong(window->scale);

			SetWindowIsAutoScale(window, TRUE);

			SetScale(window, numToLong(arg1));

			AutoScale(window);

			StatusChanged(obj);

			break;

		case 't':
			/*
			 * Toggle current cell
			 */
			CheckRun();
			Backup();
			ToggleCell(obj);

			break;

		case 'u':
			/*
			 * Move upper right with action
			 */
			DoMove(window, numNeg(defaultArg), defaultArg);

			break;

		case 'U':
			/*
			 * Shift to upper right
			 */
			DoShift(window, numNeg(defaultArg), defaultArg, arg2);

			break;

		case 'w':
			/*
			 * Move to where remembered
			 */
			ch = ScanChar();

			if ((ch < 'a') || (ch > 'z'))
				Error("Bad where letter");

			SetCurrentLocation(obj,
				numAdd(obj->whereInfo[ch - 'a'].row, arg1),
				numAdd(obj->whereInfo[ch - 'a'].col, arg2));

			ViewChanged(obj);

			break;

		case 'W':
			/*
			 * Remember where we are
			 */
			ch = ScanChar();

			if ((ch < 'a') || (ch > 'z'))
				Error("Bad where letter");

			CopyLocation(obj, &obj->currentInfo,
				&obj->whereInfo[ch - 'a']);

			ViewChanged(obj);

			break;

		case 'x':
		case 'X':
			/*
			 * Kill current cells
			 */
			CheckRun();
			Backup();

			while (!Life.stop && numIsPos(defaultArg))
			{
				defaultArg = numDec(defaultArg);

				SetCellState(obj, obj->currentInfo.row,
					obj->currentInfo.col, STATE_OFF);

				numAddrInc(&obj->currentInfo.col);

				CheckGarbageCollect();
			}

			ViewChanged(obj);

			break;

		case 'y':
			/*
			 * Move upper left with action
			 */
			DoMove(window, numNeg(defaultArg), numNeg(defaultArg));

			break;

		case 'Y':
			/*
			 * Shift to upper left
			 */
			DoShift(window, numNeg(defaultArg), numNeg(defaultArg), arg2);

			break;

		case 'z':
			/*
			 * Clear or set generation number
			 */
			CheckRun();
			obj->gen = numToLong(arg1);
			obj->minCount = obj->count;
			obj->maxCount = obj->count;
			obj->born = numZero();
			obj->died = numZero();
			obj->growth = numZero();
			StatusChanged(obj);

			break;

		case '/':
			/*
			 * Search for next object
			 */
			DoSearch(window, defaultArg);

			break;

		case '@':
			/*
			 * Point at current location
			 */
			CopyLocation(obj, &obj->currentInfo, &obj->originInfo);

			BackgroundChanged(obj);

			break;

		case '!':
		case '#':
			/*
			 * Comment characters
			 */
			while (ScanChar() != '\n')
				;

			break;

		case NULL_CMD:
			/*
			 * Null command
			 */
			break;

		default:
			/*
			 * Unknown commands
			 */
			if (ch == Life.vEof)
			{
				EndInputLevel();

				break;
			}
			Error("Unknown command");
	}

	/*
	 * The command is completed.
	 * Clear the scan buffer to allow a new one to be read
	 * and longjmp back to the setjmp to finish up.
	 */
	ScanAbort();

	return FALSE;
}


/*
 * Read a numeric value (if any) to be used as an argument for a command.
 * Pointers to the returned value and returned flag are given.
 * Return value is the first non-argument character read.
 */
static int
ReadValue(NUMBER * valuePtr, BOOL * flagPtr)
{
	*valuePtr = numZero();
	*flagPtr = FALSE;

	BOOL isNeg = FALSE;

	int ch = ScanChar();

	/*
	 * Handle a minus sign for a negative value.
	 */
	if (ch == '-')
	{
		isNeg = TRUE;
		ch = ScanChar();
	}

	/*
	 * A dollar sign means get the value of a variable.
	 */
	if (ch == '$')
	{
		*valuePtr = GetVariable1(ScanChar());

		if (isNeg)
			*valuePtr = numNeg(*valuePtr);

		*flagPtr = TRUE;

		return ScanChar();
	}

	/*
	 * A paranthesis means read an expression.
	 */
	if (ch == '(')
	{
		*valuePtr = ScanExpression();

		if (isNeg)
			*valuePtr = numNeg(*valuePtr);

		*flagPtr = TRUE;

		return ScanChar();
	}

	/*
	 * A digit means read a numeric value.
	 */
	while (isDigit(ch))
	{
		*valuePtr = numMulLong(*valuePtr, 10);
		*valuePtr = numAddLong(*valuePtr, ch - '0');
		*flagPtr = TRUE;

		ch = ScanChar();
	}

	/*
	 * A percent sign means get a loop value.
	 */
	if (ch == '%')
	{
		int depth = 1;

		if (*flagPtr)
			depth = numToLong(*valuePtr);

		*valuePtr = GetLoopValue(depth);
		*flagPtr = TRUE;

		ch = ScanChar();
	}

	if (isNeg)
		*valuePtr = numNeg(*valuePtr);

	return ch;
}


/*
 * Routine called from above to scan and evaluate a parenthesized expression.
 * This routines knows that one parenthesis has already been read.  Stops
 * reading on the matching parenthesis.
 */
static NUMBER
ScanExpression(void)
{
	char buf[1024];	/* expression buffer */

	int nest = 1;
	char * cp = buf;
	*cp++ = '(';

	while (nest > 0)
	{
		if (cp >= &buf[sizeof(buf)-2])
			Error("expression too long");

		*cp = ScanChar();

		if (*cp == '(')
			nest++;

		if (*cp == ')')
			nest--;

		cp++;
	}

	*cp = '\0';

	return GetExpression(buf);
}


/*
 * Select a set of cells to be used for some command.  This involves reading
 * the next character and marking cells based on that character.  Repeating
 * the command character is equivalent to selecting the current object.  On a
 * successful return, the selected cells are stored in the object ready for
 * further use.  The current cells of the object are not yet modified.
 * If that is needed then the selected cells must be used to manipulate them.
 * If no cells are found, an error is generated.
 */
static void
DoSelect(LifeWindow * window, int cmd, BOOL isRunOK)
{
	LifeObject * obj = window->currentObject;
	NUMBER minRow = numMinValue();
	NUMBER maxRow = numMaxValue();
	NUMBER minCol = minRow;
	NUMBER maxCol = maxRow;
	NUMBER curRow = obj->currentInfo.row;
	NUMBER curCol = obj->currentInfo.col;
	int ch;
	BOOL isSelected;

	if (!isRunOK)
		CheckRun();

	ch = ScanChar();

	/*
	 * If the select character is a repeat of the command character,
	 * then change it to a gap of zero command.
	 */
	if (ch == cmd)
		ch = '0';

	/*
	 * Check for convenient gap characters and translate them to
	 * their digital form.
	 */
	if (ch == 'o')
		ch = '0';

	if (ch == 'e')
		ch = '1';

	//
	// Clear any prior selection.
	//
	ClearSelection(obj);

	/*
	 * Switch on the selection character.
	 */
	switch (ch)
	{
		case 'a':
			/*
			 * All of object
			 */
			if (!SelectAll(obj))
				Error("Object is empty");

			return;

		case 'b':
			/*
			 * Below and left of cursor
			 */
			minRow = curRow;
			maxCol = curCol;

			break;

		case 'c':
			/*
			 * Current cell
			 */
			if (!SelectCell(obj, curRow, curCol))
				Error("No cell at current location");

			return;

		case 'h':
			/*
			 * Left of cursor
			 */
			maxCol = curCol;

			break;

		case 'j':
			/*
			 * Below cursor
			 */
			minRow = curRow;

			break;

		case 'k':
			/*
			 * Above cursor
			 */
			maxRow = curRow;

			break;

		case 'l':
			/*
			 * Right of cursor
			 */
			minCol = curCol;

			break;

		case 'm':
			/*
			 * Marked cells
			 */
			if (!SelectMarkedCells(obj))
				Error("No object marked");

			return;

		case 'n':
			/*
			 * Below and right of cursor
			 */
			minRow = curRow;
			minCol = curCol;

			break;

		case 'p':
			/*
			 * Mark within path
			 */
			if (obj->path.count == 0)
				Error("No path defined for object");

			if (!SelectPath(obj))
				Error("No cells within defined path");

			ClearPath(obj);

			return;

		case 'r':
			/*
			 * Rectangle to some cell
			 */
			ch = ScanChar();

			if (ch == '@')
			{
				minRow = obj->originInfo.row;
				minCol = obj->originInfo.col;
			}
			else if ((ch >= 'a') && (ch <= 'z'))
			{
				minRow = obj->whereInfo[ch - 'a'].row;
				minCol = obj->whereInfo[ch - 'a'].col;
			}
			else
				Error("Bad rectangle specification");

			maxRow = curRow;
			maxCol = curCol;

			if (numIsGreater(minRow, maxRow))
			{
				maxRow = minRow;
				minRow = curRow;
			}

			if (numIsGreater(minCol, maxCol))
			{
				maxCol = minCol;
				minCol = curCol;
			}

			break;

		case 's':
			/*
			 * Cells in selected rectangle.
			 * This is the region that was last dragged
			 * using the mouse.
			 */
			minRow = obj->dragMinRow;
			maxRow = obj->dragMaxRow;
			minCol = obj->dragMinCol;
			maxCol = obj->dragMaxCol;

			break;

		case 'u':
			/*
			 * Above and right of cursor
			 */
			maxRow = curRow;
			minCol = curCol;

			break;

		case 'v':
			/*
			 * Things visible in window
			 */
			GetVisibleRegion(window, &minRow, &maxRow, &minCol, &maxCol);

			break;

		case 'y':
			/*
			 * Above and left of cursor
			 */
			maxRow = curRow;
			maxCol = curCol;

			break;

		case 'i':
			/*
			 * Cells invisible in window
			 */
			GetVisibleRegion(window, &minRow, &maxRow, &minCol, &maxCol);

			isSelected = SelectRegion(obj,
				numInc(maxRow), numMaxValue(),
				numMinValue(), numMaxValue());

			isSelected |= SelectRegion(obj,
				numMinValue(), numDec(minRow),
				numMinValue(), numMaxValue());

			isSelected |= SelectRegion(obj,
				numMinValue(), numMaxValue(),
				numInc(maxCol), numMaxValue());

			isSelected |= SelectRegion(obj,
				numMinValue(), numMaxValue(),
				numMinValue(), numDec(minCol));

			if (!isSelected)
				Error("No cells outside of visible region");

			return;

		default:
			/*
			 * Check for a digit to mean the object whose gaps
			 * between cells are not more than the specified size.
			 */
			if ((ch >= '0') && (ch <= '9'))
			{
				if (!SelectObject(obj, curRow, curCol, ch - '0'))
				{
					Error("No object at current location");
				}

				return;
			}

			/*
			 * The command is unknown.
			 */
			Error("Unknown selection command");
	}

	/*
	 * If we get here, then we want a region selected, so do that.
	 */
	if (SelectRegion(obj, minRow, maxRow, minCol, maxCol))
		return;

	/*
	 * There are no cells in the region.
	 * Complain unless this is the mark dragged area command.
	 */
	if (ch != 's')
		Error("No cells in region");
}


/*
 * Move the current position by the indicated deltas, performing the
 * current action to the configuration.  The movement is scaled by
 * the current scaling factor.
 */
static void
DoMove(LifeWindow * window, NUMBER rowDelta, NUMBER colDelta)
{
	LifeObject * obj = window->currentObject;
	NUMBER row;		/* increment for row */
	NUMBER col;		/* increment for column */

	if (window->scale > 0)
	{
		rowDelta = numMulLong(rowDelta, window->scale);
		colDelta = numMulLong(colDelta, window->scale);
	}

	/*
	 * See if we just want to move.
	 */
	if (Life.moveMode == MOVE_MODE_MOVE)
	{
		AdjustCurrentLocation(obj, rowDelta, colDelta);

		PositionChanged(obj);

		return;
	}

	CheckRun();
	Backup();

	/*
	 * Need to loop for insert or delete.
	 */
	row = numZero();
	col = numZero();

	if (numIsPos(rowDelta))
		row = numOne();

	if (numIsNeg(rowDelta))
		row = numNegOne();

	if (numIsPos(colDelta))
		col = numOne();

	if (numIsNeg(colDelta))
		col = numNegOne();

	numAddrAdd(&rowDelta, obj->currentInfo.row);
	numAddrAdd(&colDelta, obj->currentInfo.col);

	while (!Life.stop && (!numIsEqual(obj->currentInfo.row, rowDelta) ||
		!numIsEqual(obj->currentInfo.col, colDelta)))
	{
		AdjustCurrentLocation(obj, row, col);

		PositionChanged(obj);

		switch (Life.moveMode)
		{
			case MOVE_MODE_INSERT:
				SetCellState(obj, obj->currentInfo.row,
					obj->currentInfo.col, STATE_ON);

				break;

			case MOVE_MODE_DELETE:
				SetCellState(obj, obj->currentInfo.row,
					obj->currentInfo.col, STATE_OFF);

				break;

			default:
				break;
		}
	}
}


/*
 * Shift the window lots in the indicated direction and also possibly
 * shift the cursor location the same amount so that the cursor location
 * on the screen doesn't change.  "Lots" is 1/N of the screen width or
 * height, where N is the shifting factor.  Special case: if both x and
 * y are being shifted, we shift both by the same amount, which is the
 * minimum of the two shifts.
 */
static void
DoShift(LifeWindow * window, NUMBER rowScale, NUMBER colScale, NUMBER shiftFactor)
{
	LifeObject * obj = window->currentObject;
	NUMBER rowDelta;	/* number of rows to shift */
	NUMBER colDelta;	/* number of columns to shift */
	NUMBER viewRows;
	NUMBER viewCols;

	/*
	 * If the shift factor is zero then use the default shift factor.
	 */
	if (numIsZero(shiftFactor))
		shiftFactor = numFromLong(GetShiftFactor(&window->settings));

	if ((numCmpLong(shiftFactor, MIN_SHIFT) < 0) ||
		(numCmpLong(shiftFactor, MAX_SHIFT) > 0))
	{
		Error("Invalid shift factor");
	}

	/*
	 * Get the required window information.
	 */
	GetVisibleSize(window, &viewRows, &viewCols);

	/*
	 * Calculate the movement according to the specified fraction
	 * of the view area.
	 */
	rowDelta = numDivLong(numMul(viewRows, rowScale), numToLong(shiftFactor));
	colDelta = numDivLong(numMul(viewCols, colScale), numToLong(shiftFactor));

	/*
	 * If diagonal movement is required, then take the minimums of
	 * the absolute values of the row and column deltas.
	 */
	if (numIsNonZero(rowDelta) && numIsNonZero(colDelta))
	{
		int rowSign = 1;
		int colSign = 1;

		if (numIsNeg(rowDelta))
		{
			rowSign = -1;
			rowDelta = numNeg(rowDelta);
		}

		if (numIsNeg(colDelta))
		{
			colSign = -1;
			colDelta = numNeg(colDelta);
		}

		if (numIsGreater(rowDelta, colDelta))
			rowDelta = colDelta;

		if (numIsGreater(colDelta, rowDelta))
			colDelta = rowDelta;

		numAddrMulLong(&rowDelta, rowSign);
		numAddrMulLong(&colDelta, colSign);
	}

	/*
	 * Shift the view in the window.
	 */
	ShiftView(window, rowDelta, colDelta);

	/*
	 * If the view type is cursor then also move
	 * the object's cursor to follow the shift,
	 * and clear the shift flag since it isn't
	 * really shifted.
	 */
	if (GetViewType(&window->settings) == VIEW_TYPE_CURSOR)
	{
		window->isShifted = FALSE;

		AdjustCurrentLocation(obj, rowDelta, colDelta);
	}
}


/*
 * Search forwards for the next set of connected cells of the object.
 * This tries to find each connected object in turn.  The current object
 * position is updated to an ON cell of the found object, and the view
 * is positioned if possible to show the whole connected object.
 */
static void
DoSearch(LifeWindow * window, NUMBER defaultArg)
{
	LifeObject * obj = window->currentObject;

	CheckRun();

	if (!SearchObject(obj, numToLong(defaultArg), FALSE))
		Error("Empty object");

	//
	// Mark the object that was found and position the view to it.
	//
	ClearSelection(obj);

	PositionChanged(obj);
	ViewChanged(obj);

//	PositionArea(window, minRow, maxRow, minCol, maxCol);
}


/*
 * Toggle the state of the current cell.
 * This always changes the cell to ON or OFF.
 */
static void
ToggleCell(LifeObject * obj)
{
	NUMBER row = obj->currentInfo.row;
	NUMBER col = obj->currentInfo.col;

	int state = GetCellState(obj, row, col);

	if (isStateEmpty(state))
		state = STATE_ON;
	else
		state = STATE_OFF;

	SetCellState(obj, row, col, state);
	ViewChanged(obj);
}


/*
 * Perform a backup of the current object.  This is only done if reading
 * from the terminal, since it is from the point of view of the user.
 * There are a number of backup objects in a circular list.
 * The undo depth is reset since we are about to modify the current
 * object.
 */
void
Backup(void)
{
	if (!IsTerminalInput())
		return;

	if (++Life.backedUpIndex >= MAX_BACKUPS)
		Life.backedUpIndex = 0;

	SaveCurrentViewParameters();

	CopyObject(Life.currentObject,
		Life.backupObjects[Life.backedUpIndex]);

	Life.undoDepth = 0;
}


/*
 * Check to see that generations are not being computed before proceeding
 * with the current command.
 */
void
CheckRun(void)
{
	if (Life.currentObject->genLeft > 0)
		Error("Illegal while running");
}


/*
 * Set the position moved flag for the specified object.
 * This can cause the cursor to be updated in the windows.
 * If the position moves outside the window then the window
 * view will be automatically updated too.  If showing of
 * the current location is enabled then the view will also
 * be updated.
 */
static void
PositionChanged(const LifeObject * obj)
{
	CheckCurrentLocation(obj);

	AddWindowObjectUpdateFlags(obj, U_POS);
}


/*
 * Set the view changed flag for the specified object.
 * This also sets that the status has changed since that is
 * generally the case.
 */
static void
ViewChanged(const LifeObject * obj)
{
	AddWindowObjectUpdateFlags(obj, U_VIEW | U_STATUS);
}


/*
 * Set the status changed flag for the specified object.
 */
static void
StatusChanged(const LifeObject * obj)
{
	AddWindowObjectUpdateFlags(obj, U_STATUS);
}


/*
 * Set for a complete redraw since the background changed.
 */
static void
BackgroundChanged(const LifeObject * obj)
{
	AddWindowObjectUpdateFlags(obj, U_REDRAW);
}

/* END CODE */
