/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to read and write Life patterns.
 */

#include "life.h"
#include "quadalg.h"


/*
 * Maximum width of a line written in rle format.
 * Reading doesn't have any limit.
 */
#define	RLEMAX	70

/*
 * The formats of RLE encodings we handle.
 */
#define	FORMAT_NORMAL	"Normal"
#define FORMAT_HISTORY	"History"


/*
 * Structure used during the scanning of a life object which is passed
 * to the callback routine.
 */
typedef struct
{
	LifeObject * obj;
	FILE * fp;
	NUMBER minRow;
	NUMBER minCol;
	NUMBER maxRow;
	NUMBER maxCol;
	NUMBER curRow;
	NUMBER curCol;
	NUMBER runLength;
	int linePos;
}
ScanInfo;


//
// Reading routines.
//
static void ReadComments(FILE * fp, LifeObject *);
static BOOL ReadCoords(FILE *, Value *, Value *);

static void ReadHistoryPattern(FILE *, LifeObject *);
static void ReadNormalPattern(FILE *, LifeObject *);

static BOOL ParseRleHeader(const char *, Value *, Value *,
	const char **, const char **);

//
// Writing routines.
//
static void WriteRleItem(ScanInfo *, NUMBER, int);

static void WritePicture(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol,
	Value maxRows, Value maxCols);

static void WriteXlife(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

static void WriteRle(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

static void WriteQuadCell(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

static BOOL XlifeCallback(NUMBER row, NUMBER col, int state, void * data);
static BOOL RleCallback(NUMBER row, NUMBER col, int state, void * data);
static BOOL PictureCallback(NUMBER row, NUMBER col, int state, void * data);


/*
 * Load the specified files into buffers having their names.
 */
void
LoadFiles(const char ** fileNames, int fileCount)
{
	while (fileCount-- > 0)
		LoadFile(*fileNames++);
}


/*
 * Load the specified file into a buffer having its name.
 * The exact file name is used, so no library paths are used.
 * Only the last component of the file path is used.
 * The loaded file is set to be the current object.
 */
void
LoadFile(const char * fileName)
{
	LifeObject * obj;
	FILE * fp;

	const char * bufferName = strrchr(fileName, '/');

	if (bufferName)
		bufferName = bufferName + 1;
	else
		bufferName = fileName;

	obj = GetUniqueObject(bufferName);

	fp = fopen(fileName, "r");

	if (fp == 0)
		Error("Cannot open file");

	TryReadFile(fp, obj);

	fclose(fp);

	CopyObject(obj, Life.readObject);

	SetWindowObject(Life.currentWindow, obj);

	//
	// Perform an autoscale to make the object fit.
	// This is done whether or not autoscaling is set.
	//
	AutoScale(Life.currentWindow);
}


/*
 * Try to read an object as described by the remainder of the specified
 * FILE if its format is picture, RLE, XLIFE, or MC.  Blank and comment
 * lines are skipped and saved before checking the format type.
 */
BOOL
TryReadFile(FILE * fp, LifeObject * obj)
{
	int ch;

	/*
	 * Read and save all leading comments and blank lines.
	 */
	ReadComments(fp, obj);

	/*
	 * Now peek at the next character of the file to determine what
	 * type of format it is.  It can be either picture format, rle
	 * format, xlife format, or macro-cell format.  The whole file
	 * is read here.
	 */
	ch = fgetc(fp);

	if (ch != EOF)
		ungetc(ch, fp);

	switch (ch)
	{
		case 'x':
			ReadRle(fp, obj);

			return TRUE;

		case '#':
			ReadXlife(fp, obj);

			return TRUE;

		case '[':
			ReadMC(fp, obj);

			return TRUE;

		case '?':
		case '.':
		case ' ':
		case '*':
		case 'O':
		case 'o':
		case '\t':
			ReadPicture(fp, obj);

			return TRUE;

		default:
			Error("Unknown file format");
	}

	/*
	 * The file format was not processed here.
	 */
	return FALSE;
}


/*
 * Read and save all blank lines or comment lines at the start of a file.
 * Comment lines are those which start with the '#C', '#N', '#O', '#P' or '# '
 * character pairs.  This routine requires that the file be seekable.
 */
static void
ReadComments(FILE * fp, LifeObject * obj)
{
	long pos;
	int ch;
	char buf[4];

	ClearStringBuffer(&obj->comments);

	/*
	 * Keep reading lines one by one until we see one which is not a
	 * comment line and is not blank.
	 */
	while (TRUE)
	{
		/*
		 * Remember the beginning of the current line.
		 */
		pos = ftell(fp);

		if (pos < 0)
			Error("Input file is not seekable");

		/*
		 * Read the first character of the line and check it out.
		 */
		ch = fgetc(fp);

		if (ch == EOF)
			return;

		/*
		 * A blank line.
		 */
		if (ch == '\n')
			continue;

		/*
		 * Check for a comment line.
		 */
		if (ch == '#')
		{
			ch = fgetc(fp);

			if ((ch == 'C') || (ch == 'N') ||
				(ch == 'O') || (ch == 'P') ||
				isBlank(ch))
			{
				buf[0] = '#';
				buf[1] = ch;
				buf[2] = '\0';

				AppendToStringBuffer(&obj->comments, buf);

				do
				{
					ch = fgetc(fp);

					if (ch != EOF)
					{
						AddCharToStringBuffer(
							&obj->comments, ch);
					}
				}
				while ((ch != '\n') && (ch != EOF));

				continue;
			}

			break;
		}

		/*
		 * Check for a line only containing blanks.
		 */
		if (isBlank(ch) || (ch == '\r'))
		{
			do
			{
				ch = fgetc(fp);
			}
			while ((ch != '\n') && (ch != EOF) &&
				(isBlank(ch) || (ch == '\r')));

			if ((ch == '\n') || (ch == EOF))
				continue;

			break;
		}

		/*
		 * Some other character.
		 */
		break;
	}

	/*
	 * We have eaten up all of the comments and blank lines.
	 * Seek back to the beginning of the line that was just read.
	 */
	if (fseek(fp, pos, SEEK_SET) < 0)
		Error("Seek failed for input file");
}


/*
 * Read in a run-length encoded description from an opened file.
 * This can handle various rules and a couple of formats.
 * This expects to immediately read the header line, and so any leading
 * comment lines must have been already read.
 */
void
ReadRle(FILE * fp, LifeObject * obj)
{
	int len;
	Value width;
	Value height;
	const char * rule;
	const char * format;
	char buf[1024];

	width = 0;
	height = 0;
	rule = "";
	format = "";

	/*
	 * Read the header line and trim trailing blanks.
	 */
	if (fgets(buf, sizeof(buf), fp) == 0)
		Error("Empty RLE file");

	len = strlen(buf);

	if (len == 0)
		Error("Empty header line in RLE file");

	if (buf[len - 1] != '\n')
		Error("Header line too long in RLE file");

	len--;

	if ((len > 0) && (buf[len - 1] =='\r'))
		len--;

	while ((len > 0) && isBlank(buf[len - 1]))
		len--;

	buf[len] = '\0';

	/*
	 * Try to parse the header line into its components.
	 */
	if (!ParseRleHeader(buf, &width, &height, &rule, &format))
		Error("Bad header line in RLE file");

	/*
	 * If a rule string was present then set the rule for the object.
	 * (Otherwise the object will use its previously set rule.)
	 */
	if (*rule)
		obj->rule = ParseRule(rule);

	/*
	 * Adjust the current row and column of the object so that the added
	 * pattern will be centered around the object's current position.
	 * Save the current location so it can be restored afterwards.
	 */
	LocationInfo savedLocation = obj->currentInfo;

	numAddrSubLong(&obj->currentInfo.row, height / 2);
	numAddrSubLong(&obj->currentInfo.col, width / 2);

	/*
	 * Now read further lines from the file which are the actual pattern
	 * using the appropriate format.
	 */
	if (strcmp(format, FORMAT_NORMAL) == 0)
		ReadNormalPattern(fp, obj);
	else if (strcmp(format, FORMAT_HISTORY) == 0)
		ReadHistoryPattern(fp, obj);
	else
		Error("Unknown RLE format");

	obj->currentInfo = savedLocation;
}


/*
 * Parse the specified RLE header line and store its components back
 * into the supplied width, height, rule, and format parameters.
 * The rule string will be empty if it wasn't explicitly specified.
 * Warning: The rule string might point back into the supplied buffer.
 * Returns TRUE on success.
 */
static BOOL
ParseRleHeader(const char * cp,
	Value * retWidth, Value * retHeight,
	const char ** retRule, const char ** retFormat)
{
	Value width = 0;
	Value height = 0;
	const char * rule = "";
	const char * format = FORMAT_NORMAL;

	*retWidth = 0;
	*retHeight = 0;
	*retRule = "";
	*retFormat = "";

	if (*cp++ != 'x')
		return FALSE;

	while (isBlank(*cp))
		cp++;

	if (*cp++ != '=')
		return FALSE;

	while (isBlank(*cp))
		cp++;

	if (!isDigit(*cp))
		return FALSE;

	while (isDigit(*cp))
		width = width * 10 + *cp++ - '0';

	while (isBlank(*cp))
		cp++;

	if (*cp++ != ',')
		return FALSE;

	while (isBlank(*cp))
		cp++;

	if (*cp++ != 'y')
		return FALSE;

	while (isBlank(*cp))
		cp++;

	if (*cp++ != '=')
		return FALSE;

	while (isBlank(*cp))
		cp++;

	if (!isDigit(*cp))
		return FALSE;

	while (isDigit(*cp))
		height = height * 10 + *cp++ - '0';

	while (isBlank(*cp))
		cp++;

	/*
	 * The rule string is optional.
	 */
	if (*cp == ',')
	{
		cp++;

		while (isBlank(*cp))
			cp++;

		if (memcmp(cp, "rule", 4) != 0)
			return FALSE;

		cp += 4;

		while (isBlank(*cp))
			cp++;

		if (*cp++ != '=')
			return FALSE;

		while (isBlank(*cp))
			cp++;

		/*
		 * Check for the LifeHistory or Life rules.
		 */
		if (strcmp(cp, "LifeHistory") == 0)
		{
			cp = DEFAULT_RULE;
			format = FORMAT_HISTORY;
		}
		else if (strcmp(cp, "Life") == 0)
		{
			cp = DEFAULT_RULE;
		}

		rule = cp;
		cp = "";
	}

	if ((width <= 0) || (height <= 0))
		return FALSE;

	/*
	 * The header is reasonable so return the values.
	 */
	*retWidth = width;
	*retHeight = height;
	*retRule = rule;
	*retFormat = format;

	return TRUE;
}


/*
 * Read the pattern string for the simple standard RLE pattern.
 * The header line has already been read and checked.
 * Warning: This routine assumes that no temporary quadcells are in
 * use since we perform garbage collections as we read the file.
 */
static void
ReadNormalPattern(FILE * fp, LifeObject * obj)
{
	NUMBER minCol = obj->currentInfo.col;

	while (TRUE)
	{
		if (Life.stop)
			Error("Reading aborted");

		Value val = 1;
		int ch = fgetc(fp);

		if (isDigit(ch))
		{
			val = 0;

			while (isDigit(ch))
			{
				val = val * 10 + ch - '0';
				ch = fgetc(fp);
			}
		}

		if (isUpper(ch))
			ch += 'a' - 'A';

		switch (ch)
		{
			case 'b':
				numAddrAddLong(&obj->currentInfo.col, val);

				break;

			case 'o':
				while (val-- > 0)
				{
					if (Life.stop)
						Error("Reading aborted");

					AddCell(obj, obj->currentInfo.row,
						obj->currentInfo.col);

					numAddrInc(&obj->currentInfo.col);

					CheckGarbageCollect();
				}

				break;

			case '$':
				numAddrAddLong(&obj->currentInfo.row, val);
				obj->currentInfo.col = minCol;

				break;

			case ' ':
			case '\t':
			case '\r':
			case '\n':
				break;

			case '!':
				/*
				 * This is the end of the RLE description.
				 * Read and ignore data until the end of the line
				 * then return leaving the rest of the file unread.
				 */
				do
				{
					ch = fgetc(fp);
				}
				while ((ch != EOF) && (ch != '\n'));

				return;

			case EOF:
				Error("Unexpected end of file in rle file");

			default:
				/*
				 * If it isn't a letter, it is illegal.
				 */
				if (!isLower(ch))
					Error("Illegal character in rle file");

				/*
				 * Make letters be the marked group of ON
				 * cells according to the letter.
				 */
				while (val-- > 0)
				{
					if (Life.stop)
						Error("Reading aborted");

					AddCell(obj, obj->currentInfo.row,
						obj->currentInfo.col);

					numAddrInc(&obj->currentInfo.col);

					CheckGarbageCollect();
				}

				break;
		}
	}
}


/*
 * Read the pattern string for the LifeHistory format.
 * The header line has already been read and checked.
 * Note: Not every possibility for such patterns are handled fully.
 * The letters, their state numbers, and their meanings are:
 * . (0) - OFF cell
 * A (1) - ON cell
 * B (2) - historical ON cell
 * C (3) - ON marked cell
 * D (4) - OFF marked cell
 * E (5) - initial ON cell
 * F (6) - boundary OFF cell
 * Warning: This routine assumes that no temporary quadcells are in
 * use since we perform garbage collections as we read the file.
 */
static void
ReadHistoryPattern(FILE * fp, LifeObject * obj)
{
	NUMBER minCol = obj->currentInfo.col;

	while (TRUE)
	{
		if (Life.stop)
			Error("Reading aborted");

		Value val = 1;
		int ch = fgetc(fp);

		if (isDigit(ch))
		{
			val = 0;

			while (isDigit(ch))
			{
				val = val * 10 + ch - '0';
				ch = fgetc(fp);
			}
		}

		switch (ch)
		{
			case '.':
			case 'B':
			case 'D':
			case 'F':
				numAddrAddLong(&obj->currentInfo.col, val);

				break;

			case 'A':
			case 'E':
				while (val-- > 0)
				{
					if (Life.stop)
						Error("Reading aborted");

					AddCell(obj, obj->currentInfo.row,
						obj->currentInfo.col);

					numAddrInc(&obj->currentInfo.col);

					CheckGarbageCollect();
				}

				break;

			case 'C':
				while (val-- > 0)
				{
					if (Life.stop)
						Error("Reading aborted");

					AddCell(obj, obj->currentInfo.row,
						obj->currentInfo.col);

					numAddrInc(&obj->currentInfo.col);

					CheckGarbageCollect();
				}

				break;

			case '$':
				numAddrAddLong(&obj->currentInfo.row, val);
				obj->currentInfo.col = minCol;

				break;

			case ' ':
			case '\t':
			case '\r':
			case '\n':
				break;

			case '!':
				/*
				 * This is the end of the RLE description.
				 * Read and ignore data until the end of the line
				 * then return leaving the rest of the file unread.
				 */
				do
				{
					ch = fgetc(fp);
				}
				while ((ch != EOF) && (ch != '\n'));

				return;

			case EOF:
				Error("Unexpected end of file in LifeHistory file");

			default:
				/*
				 * We don't handle any other letters.
				 */
				Error("Unhandled character in LifeHistory file");

				break;
		}
	}
}


/*
 * Read in an object in xlife format from an opened file.
 * Absolute coordinates are converted to relative ones.
 * However, we don't handle indirect files.
 */
void
ReadXlife(FILE * fp, LifeObject * obj)
{
	Value	row;
	Value	col;
	BOOL	inheader;
	int	ch;

	inheader = TRUE;

	while (inheader)
	{
		if (Life.stop)
			Error("Reading aborted");

		ch = fgetc(fp);

		if (ch != '#')
			Error("Directive expected in xlife file");

		ch = fgetc(fp);

		switch (ch)
		{
			case 'C':
			case 'N':
			case 'O':
				while ((ch != EOF) && (ch != '\n'))
					ch = fgetc(fp);

				break;

			case 'I':
				Error("Indirect files not supported for xlife files");

			case 'A':
			case 'R':
				while ((ch != EOF) && (ch != '\n'))
					ch = fgetc(fp);

				while (ReadCoords(fp, &col, &row))
				{
					if (Life.stop)
						Error("Reading aborted");

					AddCell(obj,
						numAddLong(obj->currentInfo.row, row),
						numAddLong(obj->currentInfo.col, col));
				}

				return;

			case 'P':
				if (!ReadCoords(fp, &col, &row))
					Error("Unexpected EOF in xlife file");

				numAddrAddLong(&obj->currentInfo.row, row);
				numAddrAddLong(&obj->currentInfo.col, col);

				ReadPicture(fp, obj);

				return;

			default:
				Error("Unknown directive in xlife file");
		}
	}
}


/*
 * Read in a picture of a life object from an opened file.
 */
void
ReadPicture(FILE * fp, LifeObject * obj)
{
	NUMBER minCol = obj->currentInfo.col;

	while (TRUE)
	{
		if (Life.stop)
			Error("Reading aborted");

		int ch = fgetc(fp);

		switch (ch)
		{
			case '.':
			case ' ':
			case 'x':
			case 'X':
			case '?':
				numAddrInc(&obj->currentInfo.col);

				break;

			case '*':
			case 'O':
			case 'o':
				AddCell(obj, obj->currentInfo.row,
					obj->currentInfo.col);

				numAddrInc(&obj->currentInfo.col);

				break;

			case '\t':
				do
				{
					numAddrInc(&obj->currentInfo.col);
				}
				while (numModLong(numSub(
					obj->currentInfo.col, minCol), 8) != 0);

				break;

			case '\r':
				obj->currentInfo.col = minCol;
				break;

			case '\n':
				obj->currentInfo.col = minCol;
				numAddrInc(&obj->currentInfo.row);

				break;

			case EOF:
				return;

			default:
				Error("Illegal pictorial character in file");
		}
	}
}


/*
 * Read a pair of coordinates and the following end of line from a file.
 * The pair of values are returned indirectly through pointers.  Extra
 * blanks are allowed.  Returns TRUE if coordinates were read, or FALSE
 * if EOF was found.
 */
static BOOL
ReadCoords(FILE * fp, Value * num1, Value * num2)
{
	Value	val;
	BOOL	isneg;
	int	ch;
	char *	cp;
	char	buf[80];

	*num1 = 0;
	*num2 = 0;

	val = 0;

	do
	{
		ch = fgetc(fp);

		if (ch == EOF)
		{
			if (val == 0)
				return FALSE;

			goto badline;
		}

		if (val >= sizeof(buf))
			goto badline;

		buf[val++] = ch;
	}
	while (ch != '\n');

	cp = buf;

	while (isBlank(*cp))
		cp++;

	isneg = (*cp == '-');

	if (isneg)
		cp++;

	val = 0;

	if (!isDigit(*cp))
		goto badline;

	while (isDigit(*cp))
		val = val * 10 + *cp++ - '0';

	if (isneg)
		val = -val;

	*num1 = val;

	if (!isBlank(*cp))
		goto badline;

	while (isBlank(*cp))
		cp++;

	isneg = (*cp == '-');

	if (isneg)
		cp++;

	val = 0;

	if (!isDigit(*cp))
		goto badline;

	while (isDigit(*cp))
		val = val * 10 + *cp++ - '0';

	if (isneg)
		val = -val;

	*num2 = val;

	while (isBlank(*cp))
		cp++;

	if (*cp == '\r')
		cp++;

	if (*cp != '\n')
	{
badline:	Error("Invalid coordinate value in file");
	}

	return TRUE;
}


/*
 * Read in a MC (macro-cell) format file.
 */
void
ReadMC(FILE * fp, LifeObject * obj)
{
	//
	// We can only read MC format if we are using the quadcell algorithm.
	// This check is a bit ugly.
	//
	if (strcmp(GetAlgorithmName(obj->algorithm), QUAD_NAME))
		Error("MC format only readable when using quadcell algorithm");

	QuadCell * cp = qcReadStructure(fp);

	qcSetQuadCell(obj, qcOrOperation(obj->quadCells, cp));
}


/*
 * Write an object out to a file in one of several different formats,
 * which is one of picture, xlife, rle, or quadcell.  A format of 0
 * causes us to use a default format.  The maxRows and maxCols arguments
 * are only used for limiting picture sizes.  This is a generic routine
 * which handles multiple algorithms.
 */
void
WriteObject(LifeObject * obj, const char * format, const char * fileName,
	Value maxRows, Value maxCols)
{
	FILE * fp;
	NUMBER minRow;
	NUMBER maxRow;
	NUMBER minCol;
	NUMBER maxCol;

	//
	// See if the file extension indicates a writing format.
	// Use it only if the format wasn't specificly indicated.
	//
	if (format == 0)
	{
		int len = strlen(fileName);

		if ((len > 3) && (strcmp(&fileName[len - 3], ".mc") == 0))
			format = "quadcell";
		else
			format = "rle";
	}

	//
	// Validate the parameters and make sure that the object
	// has something in it.
	//
	if (!IsAbbreviation(format, "picture") &&
		!IsAbbreviation(format, "rle") &&
		!IsAbbreviation(format, "xlife") &&
		!IsAbbreviation(format, "quadcell"))
	{
		Error("Unknown file format specified");
	}

	if (*fileName == '\0')
		Error("No filename specified");

	//
	// Compute its bounds and give an error if the object is empty.
	//
	if (!MinMax(obj, &minRow, &maxRow, &minCol, &maxCol))
		Error("Object is empty");

	//
	// Get the cell count and see if it is too large.
	// Only the quadcell format can have huge counts.
	//
	NUMBER cellCount = CalculateCellCount(obj);

	if ((*format != 'q') && (numCmpLong(cellCount, MAX_WRITE) > 0))
		Error("Object is too large to write in the specified format");

	//
	// Now we can start writing the file.
	//
	fp = fopen(fileName, "w");

	if (fp == 0)
		Error("Cannot open output file");

	//
	// Write the object in the specified format.
	// Note: The format value has been verified earlier and so we
	// know here that the first letter identifies it uniquely.
	//
	switch (*format)
	{
		case 'p':
			WritePicture(fp, obj, minRow, maxRow, minCol, maxCol,
				maxRows, maxCols);

			break;

		case 'x':
			WriteXlife(fp, obj, minRow, maxRow, minCol, maxCol);

			break;

		case 'r':
			WriteRle(fp, obj, minRow, maxRow, minCol, maxCol);

			break;

		case 'q':
			WriteQuadCell(fp, obj, minRow, maxRow, minCol, maxCol);

			break;

		default:
			verifyAlways(FALSE);
	}

	//
	// Finish the file.
	//
	fflush(fp);

	if (ferror(fp))
	{
		fclose(fp);
		Error("Write failed");
	}

	if (fclose(fp))
		Error("Close failed");

	if (Life.stop)
		Error("Writing aborted");
}


/*
 * Write the current object out to the named file as a picture, with the
 * given maximum sizes for "pretty" output.  An error is generated if the
 * object exceeds the maximum sizes.  The file as written can be read in
 * as commands which will regenerate the object.
 */
static void
WritePicture(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol,
	Value maxRows, Value maxCols)
{
	const char * rule = obj->rule->name;

	//
	// Check whether the object is too large.
	//
	NUMBER rowCount = numRange(minRow, maxRow);
	NUMBER colCount = numRange(minCol, maxCol);

	if ((numCmpLong(rowCount, maxRows) > 0) ||
		(numCmpLong(colCount, maxCols) > 0))
	{
		Error("Object exceeds maximum size for picture format");
	}

	//
	// OK, write the object as commands to be executed.
	//
	fprintf(fp, "#C \"%s\" (cells %s length %s width %s generation %ld)\n",
		GetStringBufferText(&obj->name),
		numToString(CalculateCellCount(obj)),
		numToString(numRange(minRow, maxRow)),
		numToString(numRange(minCol, maxCol)), obj->gen);

	if (*rule && (strcmp(rule, DEFAULT_RULE) != 0))
		fprintf(fp, "#C rule %s\n", rule);

#if 0
	if (numIsGreater(obj->currentInfo.row, minRow))
		fprintf(fp, "%sk",
			numToString(numSub(obj->currentInfo.row, minRow)));

	if (numIsLess(obj->currentInfo.row, minRow))
		fprintf(fp, "%sj",
			numToString(numSub(minRow, obj->currentInfo.row)));

	if (numIsGreater(obj->currentInfo.col, minCol))
		fprintf(fp, "%sh",
			numToString(numSub(obj->currentInfo.col, minCol)));

	if (numIsLess(obj->currentInfo.col, minCol))
		fprintf(fp, "%sl", numToString(numSub(minCol, obj->currentInfo.col)));

	fprintf(fp, "@!\n");
#endif

	//
	// Set the needed data for the callbacks and then scan the object.
	// The callback routine will write the picture of the object.
	//
	ScanInfo info;

	info.fp = fp;
	info.minRow = minRow;
	info.minCol = minCol;
	info.maxRow = maxRow;
	info.maxCol = maxCol;
	info.curRow = minRow;
	info.curCol = minCol;

	ScanObject(obj, STATE_NONE, PictureCallback, &info);

	//
	// End the last row.
	//
	while (numIsLessOrEqual(info.curCol, info.maxCol))
	{
		fputc('.', fp);
		numAddrInc(&info.curCol);
	}

	fputc('\n', fp);
}


/*
 * Callback routine for writing picture format.
 * Returns TRUE if the scan is to be continued.
 */
static BOOL
PictureCallback(NUMBER row, NUMBER col, int state, void * data)
{
	ScanInfo * info = data;

	//
	// Write blank rows up until this new row,
	// including ending the current row.
	//
	while (numIsLess(info->curRow, row))
	{
		while (numIsLessOrEqual(info->curCol, info->maxCol))
		{
			fputc('.', info->fp);
			numAddrInc(&info->curCol);
		}

		fputc('\n', info->fp);

		numAddrInc(&info->curRow);
		info->curCol = info->minCol;
	}

	//
	// Write blank columns until this new column.
	//
	while (numIsLess(info->curCol, col))
	{
		fputc('.', info->fp);
		numAddrInc(&info->curCol);
	}

	//
	// Write this live cell.
	//
	fputc('O', info->fp);
	info->curCol = numInc(col);

	return !Life.stop;
}


/*
 * Write an object out to a file in xlife format.
 * This format is just a list of coordinates.
 * We write the coordinates relative to the current position.
 * This is a generic routine which handles multiple algorithms.
 */
static void
WriteXlife(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	//
	// Write the header.
	//
	fprintf(fp, "#C \"%s\" (cells %s width %s height %s generation %ld)\n",
		GetStringBufferText(&obj->name),
		numToString(CalculateCellCount(obj)),
		numToString(numRange(minCol, maxCol)),
		numToString(numRange(minRow, maxRow)), obj->gen);

	fputs("#R\n", fp);

	//
	// Set the needed data for the callbacks and then scan the object.
	// The callback routine will write the coordinates of the cells.
	//
	ScanInfo info;

	info.obj = obj;
	info.fp = fp;

	ScanObject(obj, STATE_NONE, XlifeCallback, &info);
}


/*
 * Callback routine for writing xlife format.
 * Returns TRUE if the scan is to be continued.
 */
static BOOL
XlifeCallback(NUMBER row, NUMBER col, int state, void * data)
{
	ScanInfo * info = data;
	LifeObject * obj = info->obj;

	NUMBER relRow = numSub(row, obj->currentInfo.row);
	NUMBER relCol = numSub(col, obj->currentInfo.col);

	fprintf(info->fp, "%s %s\n",
		numToString(relCol), numToString(relRow));

	return !Life.stop;
}


/*
 * Write an object out to a file in RLE format.
 * This format uses 'b' for a blank cell, 'o' for an unmarked live cell,
 * other letters for marked live cells, dollar signs for end of lines,
 * and an exclaimation point for the end of pattern.  Repeat counts can be
 * applied.  This is a generic routine which handles multiple algorithms.
 * TODO: Handle marks.
 */
static void
WriteRle(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	const char *rule = obj->rule->name;

	//
	// Write the header.
	//
	fprintf(fp, "x = %s, y = %s", numToString(numRange(minCol, maxCol)),
		numToString(numRange(minRow, maxRow)));

	if (*rule && (strcmp(rule, DEFAULT_RULE) != 0))
		fprintf(fp, ", rule = %s", rule);

	fputc('\n', fp);

	//
	// Set the needed data for the callbacks and then scan the object.
	// The callback routine will write most of the RLE data.
	//
	ScanInfo info;

	info.obj = obj;
	info.fp = fp;
	info.minRow = minRow;
	info.minCol = minCol;
	info.curRow = minRow;
	info.curCol = minCol;
	info.runLength = numZero();
	info.linePos = 0;

	ScanObject(obj, STATE_NONE, RleCallback, &info);

	//
	// Finish the last run if any, then add the end of pattern.
	//
	WriteRleItem(&info, info.runLength, 'o');
	WriteRleItem(&info, numOne(), '!');
	fputc('\n', fp);
}


/*
 * Callback routine for writing RLE format.
 * Returns TRUE if the scan is to be continued.
 */
static BOOL
RleCallback(NUMBER row, NUMBER col, int state, void * data)
{
	ScanInfo * info = data;

	//
	// If this ON cell continues a run then just extend the run.
	//
	if (numIsEqual(row, info->curRow) &&
		numIsPos(info->runLength) &&
		numIsEqual(col, info->curCol))
	{
		numAddrInc(&info->runLength);
		info->curCol = numInc(col);

		return !Life.stop;
	}

	//
	// This cell will have to start a new run.
	// But first, if there was a current run in progress then
	// write an element for it.
	//
	if (numIsPos(info->runLength))
	{
		WriteRleItem(info, info->runLength, 'o');
		info->runLength = numZero();
	}

	//
	// Now if the new row is larger than the previous row then
	// write an end of line element to reach that row.
	//
	if (numIsLess(info->curRow, row))
	{
		WriteRleItem(info, numSub(row, info->curRow), '$');

		info->curRow = row;
		info->curCol = info->minCol;
	}

	//
	// Then possibly write a blank element to reach the new column.
	//
	WriteRleItem(info, numSub(col, info->curCol), 'b');

	//
	// Finally remember the location of the beginning of the new run.
	//
	info->curCol = numInc(col);
	info->runLength = numOne();

	return !Life.stop;
}


/*
 * Convert the specified count and action character to an rle string,
 * and write it to the specified file.  This adds the character count to
 * the specified line position, and if it would exceed RLEMAX, then inserts
 * a newline first.  The character position for the line is updated.
 * This is a generic routine which handles multiple algorithms.
 */
static void
WriteRleItem(ScanInfo * info, NUMBER count, int action)
{
	int len;
	char buf[64];

	if (numIsNonPos(count))
		return;

	if (numIsOne(count))
	{
		buf[0] = action;
		buf[1] = '\0';
		len = 1;
	}
	else if (numIsTwo(count))
	{
		buf[0] = action;
		buf[1] = action;
		buf[2] = '\0';
		len = 2;
	}
	else
	{
		sprintf(buf, "%s%c", numToString(count), action);
		len = strlen(buf);
	}

	if (info->linePos + len > RLEMAX)
	{
		fputc('\n', info->fp);
		info->linePos = 0;
	}

	fputs(buf, info->fp);
	info->linePos += len;
}


/*
 * Write an object out to a file in quadcell format.
 * This is a generic routine which handles multiple algorithms.
 * However, only the quadcell algorithm makes this easy.
 * This saves the quadcell with whatever coordinates it may have.
 * So it might be a good idea for the user to center the object
 * before writing it.  Warning: This can perform garbage collection.
 */
static void
WriteQuadCell(FILE * fp, LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	QuadCell * cells = ExtractQuadCell(obj);

	cells = qcReduceStates(cells);
	cells = qcTrim(cells);

	const char * rule = obj->rule->name;

	if (strcmp(rule, DEFAULT_RULE) != 0)
		fprintf(fp, "# Rule %s\n", rule);

	qcWriteStructure(fp, cells);
}

