/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Generic routines to support the running of Life generations.
 */

#include "life.h"


/*
 * A Life rule can specify any combination of the neighborhood settings.
 * There are nine cells of interest, with labels arbitrarily defined as
 * wrapping clockwise around the middle cell as follows:
 *	abc
 *	hid
 *	gfe
 *
 * Each neighborhood cell is assigned a power of two which is easily
 * derived from the letter identifying the cell of the neighborhood
 * which is set.  A rule table is indexed by the sum of the values of
 * the live cells in the neighborhood.  Entries in the table are nonzero
 * if the new cell is to be live.  The rule table has 2^9 values since
 * there are nine cells of interest.
 */
//#define	RULESIZE	512
#define	HALFRULESIZE	(RULESIZE / 2)
#define	ANDCOUNT	10
#define	VALIDCELL(ch)	(((ch) >= 'a') && ((ch) <= 'i'))
#define	VALIDCOUNT(ch)	(((ch) >= '0') && ((ch) <= '9'))
#define	VALIDNEIGHBORHOOD(ch)	(((ch) >= 'a') && ((ch) <= 'z'))
#define	CELLBIT(ch)	(1 << ((ch) - 'a'))
#define	COUNTBIT(ch)	(1 << ((ch) - '0'))
#define	CELL_ABCDEFGH	0x0ff
#define	CELL_I		0x100
#define	CELL_ABCDEFGHI	0x1ff


/*
 * The list of rules.
 * These are linked together using the next field.
 */
static LifeRule * ruleList = 0;


/*
 * The rule currently being used for running generations.
 */
static const LifeRule * currentRule = 0;


/*
 * Table of non-totalistic neighborhood types.
 * These are based on Alan Helsel's web page.
 */
const struct RuleTypes
{
	int id;			/* character identifier */
	int count;		/* neighbor count */
	const char * cellList;	/* neighborbood cell list */
}
RuleTypes[] =
{
	{'c', 1, "a"},
	{'c', 2, "ac"},
	{'c', 3, "acg"},
	{'c', 4, "aceg"},
	{'c', 5, "bdefh"},
	{'c', 6, "bdefgh"},
	{'c', 7, "bcdefgh"},
	{'e', 1, "b"},
	{'e', 2, "bh"},
	{'e', 3, "bdh"},
	{'e', 4, "bdfh"},
	{'e', 5, "acefg"},
	{'e', 6, "acdefg"},
	{'e', 7, "acdefgh"},
	{'k', 2, "ad"},
	{'k', 3, "adf"},
	{'k', 4, "acfh"},
	{'k', 5, "bcegh"},
	{'k', 6, "bcefgh"},
	{'a', 2, "ab"},
	{'a', 3, "abh"},
	{'a', 4, "abcd"},
	{'a', 5, "cdefg"},
	{'a', 6, "cdefgh"},
	{'i', 2, "bf"},
	{'i', 3, "abc"},
	{'i', 4, "acdh"},
	{'i', 5, "defgh"},
	{'i', 6, "acdegh"},
	{'n', 2, "ae"},
	{'n', 3, "acd"},
	{'n', 4, "abcg"},
	{'n', 5, "befgh"},
	{'n', 6, "bcdfgh"},
	{'y', 3, "acf"},
	{'y', 4, "acdg"},
	{'y', 5, "bdegh"},
	{'q', 3, "abe"},
	{'q', 4, "abeh"},
	{'q', 5, "cdfgh"},
	{'j', 3, "abd"},
	{'j', 4, "abdf"},
	{'j', 5, "cefgh"},
	{'r', 3, "abf"},
	{'r', 4, "abdh"},
	{'r', 5, "cdegh"},
	{'t', 4, "abcf"},
	{'w', 4, "abde"},
	{'z', 4, "abef"},
	{0, 0, ""}
};


/*
 * Translations for rotating and flipping neighbors.
 * These are indexed by (neighbor - 'a'), so that the
 * rotation of 'c' is 'e'.
 */
const char * rotations = "cdefghab";
const char * flips = "cbahgfed";


/*
 * Local procedures.
 */
static const char * FindCellList(int neighborType, int cellCount);

static const LifeRule * AddRule(const char * name, char rule[RULESIZE]);

static void SetRuleValues(char ruleTable[HALFRULESIZE],
	int neighborType, int cellCount, BOOL value);

static const char * TransformNeighborhoods(const char * cellList,
	int rotationCount, BOOL isFlipped);

static void SetOneRuleValue(char ruleTable[HALFRULESIZE],
	const char * cellList, BOOL value);

static void SetValuesForCount(char ruleTable[HALFRULESIZE],
	int neighborCount, BOOL value);

static const LifeRule * SetNeighborhoodRule(const char * ruleStr);

static BOOL StopSatisfied(const LifeObject * obj);
static int bitCount(int value);


/*
 * Parse a Life rule based on the specified rule string.
 * There are three ways to describe rules: the normal totalistic syntax,
 * Hensel's non-totalistic syntax, and a explicit neighborhood syntax syntax.
 *
 * The totalistic syntax is of the form Bmmm/Snnn or Snnn/Bmmm, where mmm
 * are digits specifying neighbors required for birth, and nnn are digits
 * specifying neighbors required for survival.  The B and S are optional,
 * and the slash may also be a comma.
 *
 * The non-totalistic syntax is similar to the totalistic syntax, except
 * that a digit can be followed by a list of letters, and then another
 * set of letters following a dash.  Each letter identifies a type of
 * neighborhood, and the dash means negation.  For example, "b3/s2-a"
 * is the JustFriends rule.
 *
 * The neighborhood syntax is more complicated and is explained in
 * its own routine.
 *
 * Returns the LifeRule object defined by the rule string.
 */
const LifeRule *
ParseRule(const char * ruleStr)
{
	int	sawBorn = FALSE;
	int	sawStay = FALSE;
	const char * saveRuleStr;
	char *	ruleBase;
	char	tempRules[RULESIZE];

	/*
	 * If the rule begins with the new neighborhood syntax,
	 * go parse it that way.
	 */
	while (isBlank(*ruleStr))
		ruleStr++;

	if (*ruleStr == 'N')
		return SetNeighborhoodRule(ruleStr);

	//
	// The rule is the old born and stay alive syntax, with possible
	// non-totalistic neighborhood letters after each digit.
	// Example: "b2a3-z7/s4e-a56".
	//
	if (strlen(ruleStr) >= sizeof(Life.defaultRule->name))
		Error("Rule string is too long");

	//
	// Look for the normal rules under other names and turn them
	// into the expected rule.
	//
	if ((strcasecmp(ruleStr, "b3/s23") == 0) ||
		(strcasecmp(ruleStr, "s23/b3") == 0) ||
		(strcasecmp(ruleStr, "23/3") == 0))
	{
		ruleStr = DEFAULT_RULE;
	}

	saveRuleStr = ruleStr;

	memset(tempRules, FALSE, RULESIZE);

	//
	// Loop parsing it to get the born and live parts.
	//
	while (*ruleStr && ((sawBorn == 0) || (sawStay == 0)))
	{
		if ((*ruleStr == 'B') || (*ruleStr == 'b'))
		{
			ruleBase = &tempRules[0];
			sawBorn++;
			ruleStr++;
		}
		else if ((*ruleStr == 'S') || (*ruleStr == 's'))
		{
			ruleBase = &tempRules[HALFRULESIZE];
			sawStay++;
			ruleStr++;
		}
		else
		{
			if (sawBorn)
			{
				ruleBase = &tempRules[HALFRULESIZE];
				sawStay++;
			}
			else
			{
				ruleBase = &tempRules[0];
				sawBorn++;
			}
		}

		//
		// Now read the neighborhood count digits.
		// For each digit look for possible neighborhood letters.
		//
		memset(ruleBase, FALSE, HALFRULESIZE);

		while ((*ruleStr >= '0') && (*ruleStr <= '8'))
		{
			int neighborCount = (*ruleStr++ - '0');

			//
			// If the next character is not a minus sign or a
			// letter then the usual totalistic neighborhood
			// rule is being defined.
			//
			if ((*ruleStr != '-') && !VALIDNEIGHBORHOOD(*ruleStr))
			{
				SetValuesForCount(ruleBase, neighborCount, TRUE);

				continue;
			}

			//
			// We are parsing a non-totalistic rule.
			// If the next character is a dash then set all of
			// of the count values so we can remove the listed ones.
			// Parse the list of included neighborhoods if any.
			//
			if (*ruleStr == '-')
			{
				SetValuesForCount(ruleBase, neighborCount, TRUE);
			}
			else while (VALIDNEIGHBORHOOD(*ruleStr))
			{
				SetRuleValues(ruleBase, *ruleStr++,
					neighborCount, TRUE);
			}

			//
			// If a dash is seen then skip it.
			//
			if (*ruleStr == '-')
				ruleStr++;

			//
			// Parse the list of negated neighborhoods if any.
			//
			while (VALIDNEIGHBORHOOD(*ruleStr))
			{
				SetRuleValues(ruleBase, *ruleStr++,
					neighborCount, FALSE);
			}
		}

		//
		// Look for an discard the separator character, ignoring blanks.
		//
		while (isBlank(*ruleStr))
			ruleStr++;

		if ((*ruleStr == ',') || (*ruleStr == '/'))
			ruleStr++;

		while (isBlank(*ruleStr))
			ruleStr++;
	}

	//
	// Check for a following comma or colon character which exists
	// in some types of enhanced rle files, and ignore them.
	//
	if ((*ruleStr == ',') || (*ruleStr == ':'))
		ruleStr = "";

	//
	// Make sure that only blanks remain, and that exactly one
	// born specification and one live specification was seen.
	//
	if (*ruleStr || (sawBorn != 1) || (sawStay != 1))
		Error("Bad rule string");

	//
	// The rule is fine, so add it to the list and return it.
	//
	return AddRule(saveRuleStr, tempRules);
}


/*
 * Set the rule based on the specified neighborhood rule string.
 * If the rule string is bad, the real rule is not changed.
 * The syntax of a neighborhood rule string is:
 *	rule = "N" orset [ "," orset ] ...
 *	orset = "A" andset | "S" andset | andset
 *	andset = countset cellset
 *	countset = [count] ..
 *	count = "0" ... "9"
 *	cellset = cell ...
 *	cell = "a" ... "i"
 * The "A" and "S" flags add or subtract conditions set by following andsets.
 * Each orset defines a set of conditions of which one or more must match.
 * Each andset defines a set of conditions all of which must match.
 * Each countset defines a set of the count of live cells among the cellset.
 * Each cellchar can only appear once in an andset.
 * Each countchar can only appear once in a count.
 * Unused cells in an orset must be off.
 * An empty countset defines a set having the single count of 1.
 * As an example, the normal Life rule can be represented as:
 *	Ni2abcdefgh,3abcdefgh
 */
const LifeRule *
SetNeighborhoodRule(const char * ruleStr)
{
	BOOL	adding;
	BOOL	applies;
	int	count;
	int	counts;
	int	andIndex;
	int	andCount;
	int	ruleIndex;
	int	cells;
	int	allCells;
	int	countTable[ANDCOUNT];
	int	cellTable[ANDCOUNT];
	char	newRule[RULESIZE];
	static char oldRule[RULESIZE];

	if (*ruleStr++ != 'N')
		Error("Rule string is not in neighborhood syntax");

	/*
	 * See what the next character is.
	 * If it is 'A' for add, or 'S' for subtract, then start with
	 * the existing rule, otherwise start with an empty rule.
	 */
	adding = TRUE;

	if ((*ruleStr == 'A') || (*ruleStr == 'S'))
		memcpy(newRule, oldRule, RULESIZE);
	else
		memset(newRule, FALSE, RULESIZE);

	/*
	 * Here is the main loop of the rule parsing routine.
	 */
	while (*ruleStr)
	{
		/*
		 * See if this changes the adding or subtracting state.
		 */
		if ((*ruleStr == 'A') || (*ruleStr == 'S'))
			adding = (*ruleStr++ == 'A');

		/*
		 * Reset all of the bits.
		 */
		andCount = 0;
		allCells = 0;
		memset(cellTable, 0, sizeof(cellTable));
		memset(countTable, 0, sizeof(countTable));

		/*
		 * Collect sets of cells that must have the specified
		 * number of live cells.
		 */
		while (VALIDCOUNT(*ruleStr) || VALIDCELL(*ruleStr))
		{
			if (andCount >= ANDCOUNT - 1)
				Error("Too many AND sets in rule");

			/*
			 * Collect the set of counts, which are made of a
			 * sequence of the digits 0 to 9, individually
			 * interpreted.
			 */
			counts = 0;

			while (VALIDCOUNT(*ruleStr))
				counts |= COUNTBIT(*ruleStr++);

			/*
			 * If no count was seen, then a count of 1 is assumed.
			 */
			if (counts == 0)
				counts = COUNTBIT('1');

			/*
			 * Now collect the set of neighborhood cells.
			 */
			cells = 0;

			while (VALIDCELL(*ruleStr))
				cells |= CELLBIT(*ruleStr++);

			/*
			 * Check the use of the same cell in multiple
			 * AND sets.
			 */
			if (allCells & cells)
				Error("Cell used in multiple AND sets");

			allCells |= cells;

			/*
			 * Store this rule.
			 */
			countTable[andCount] = counts;
			cellTable[andCount] = cells;
			andCount++;
		}

		/*
		 * Now we expect a comma or the end of the rule string.
		 */
		if (*ruleStr && (*ruleStr++ != ','))
			Error("Bad neighborhood rule string");

		/*
		 * If not all of the cells in the neighborhood were set,
		 * then make the remaining ones be forced to off.
		 */
		countTable[andCount] = COUNTBIT('0');
		cellTable[andCount] = ~allCells & CELL_ABCDEFGHI;
		andCount++;

		/*
		 * Now set the values of the rule table according to
		 * whether the defined sets of rules apply.  The index
		 * into the rule table defines the set of neighbors which
		 * are on, and thus we just use the index as a mask.
		 */
		for (ruleIndex = 0; ruleIndex < RULESIZE; ruleIndex++)
		{
			applies = TRUE;

			/*
			 * Check to see if all of the andsets are true.
			 */
			for (andIndex = 0;
				applies && (andIndex < andCount);
				andIndex++)
			{
				count = bitCount(
					ruleIndex & cellTable[andIndex]);

				if ((countTable[andIndex] & (1 << count)) == 0)
					applies = FALSE;
			}

			/*
			 * If every andset was correct, then set this index
			 * as either true or false.
			 */
			if (applies)
				newRule[ruleIndex] = adding;
		}
	}

	/*
	 * OK, add the new rule for use and return it.
	 * Use a blank rule name since it is complicated.
	 */
	memcpy(oldRule, newRule, RULESIZE);

	return AddRule("", newRule);
}


/*
 * Add a rule for use, or return the existing one if already added.
 * The rule is defined by the array of neighborhood flags.
 * The name is for identification only.
 * Returns the new or found rule structure.
 */
static const LifeRule *
AddRule(const char * name, char rule[RULESIZE])
{
	LifeRule * rp = ruleList;

	//
	// Look for the rule already existing.
	//
	while (rp)
	{
		if (memcmp(rp->rule, rule, RULESIZE) == 0)
			return rp;

		rp = rp->next;
	}

	//
	// This is a new rule.
	//
	rp = AllocateRule();

	strcpy(rp->name, name);
	memcpy(rp->rule, rule, RULESIZE);

	rp->next = ruleList;
	ruleList = rp;

	return rp;
}


/*
 * Set or clear the rule values for the specified number of neighbors.
 */
static void
SetValuesForCount(char ruleTable[HALFRULESIZE], int neighborCount, BOOL value)
{
	int ruleIndex;

	for (ruleIndex = 0; ruleIndex < HALFRULESIZE; ruleIndex++)
	{
		if (bitCount(ruleIndex) == neighborCount)
		{
			ruleTable[ruleIndex] = value;
		}
	}
}


/*
 * Set or clear the rule values for the specified combination of cell count
 * and neighborhood types.  This applies all of the rotations and flips.
 * The rule table is not cleared first, so this just adds them in.
 */
static void
SetRuleValues(char ruleTable[HALFRULESIZE], int neighborType,
	int cellCount, BOOL value)
{
	const char * cellList = FindCellList(neighborType, cellCount);

	//
	// Set the values for the cell list as given.
	//
	SetOneRuleValue(ruleTable, cellList, value);

	//
	// Then perform rotations and flips and set those values too.
	//
	SetOneRuleValue(ruleTable, TransformNeighborhoods(cellList, 0, TRUE), value);
	SetOneRuleValue(ruleTable, TransformNeighborhoods(cellList, 1, FALSE), value);
	SetOneRuleValue(ruleTable, TransformNeighborhoods(cellList, 1, TRUE), value);
	SetOneRuleValue(ruleTable, TransformNeighborhoods(cellList, 2, FALSE), value);
	SetOneRuleValue(ruleTable, TransformNeighborhoods(cellList, 2, TRUE), value);
	SetOneRuleValue(ruleTable, TransformNeighborhoods(cellList, 3, FALSE), value);
	SetOneRuleValue(ruleTable, TransformNeighborhoods(cellList, 3, TRUE), value);
}


/*
 * Set or clear the rule value for the specified cell list.
 */
static void
SetOneRuleValue(char ruleTable[HALFRULESIZE], const char * cellList, BOOL value)
{
	int ruleIndex = 0;

	while (*cellList)
	{
		ruleIndex |= CELLBIT(*cellList);

		cellList++;
	}

	ruleTable[ruleIndex] = value;
}


/*
 * Return a neighborhood cell list which has been rotated or flipped.
 * The returned string is overwritten on each call.
 */
static const char *
TransformNeighborhoods(const char * cellList, int rotationCount, BOOL isFlipped)
{
	static char newCellList[12];
	int index = 0;
	int i;

	while (cellList[index])
	{
		int value = cellList[index];

		if (isFlipped)
			value = flips[value - 'a'];

		for (i = 0; i < rotationCount; i++)
			value = rotations[value - 'a'];

		newCellList[index] = value;
		index++;
	}

	newCellList[index] = '\0';

	return newCellList;
}


/*
 * Return the neighborhood cell list for the specified neighborhood type
 * and cell count.  Generates an error if the combination is invalid.
 */
static const char *
FindCellList(int neighborType, int cellCount)
{
	const struct RuleTypes * rule;

	rule = RuleTypes;

	while (rule->id)
	{
		if ((rule->id == neighborType) && (rule->count == cellCount))
			return rule->cellList;

		rule++;
	}

	Error("Invalid rule neighborhood");

	return 0;
}


/*
 * Count the number of set bits in a integer.
 */
static int
bitCount(int value)
{
	int	count;
	int	lowBit;

	count = 0;

	while (value)
	{
		lowBit = value & (~value + 1);
		value &= ~lowBit;
		count++;
	}

	return count;
}


/*
 * Return whether any object wants to run generations.
 */
BOOL
IsRunningGenerations(void)
{
	const LifeObject * obj;

	for (obj = Life.objects; obj; obj = obj->nextObject)
	{
		if (obj->isReserved && (obj != Life.mainObject))
			continue;

		if (obj->genLeft > 0)
			return TRUE;
	}

	return FALSE;
}


/*
 * Compute one or more generatins of all of the objects which have
 * generations left to run.  This does nothing for reserved
 * objects except for main.  Returns TRUE if at least one
 * object has been run.
 */
BOOL
DoGenerationsForAll(void)
{
	LifeObject * obj;
	BOOL isRun = FALSE;

	for (obj = Life.objects; obj; obj = obj->nextObject)
	{
		if (obj->isReserved && (obj != Life.mainObject))
			continue;

		if (DoGenerations(obj))
			isRun = TRUE;
	}

	return isRun;
}


/*
 * Possibly compute one or more generations of the current configuration.
 * Returns TRUE if a generation was actually run and something changed.
 * This depends on stop conditions, the frequency, the number of
 * generations left to run, and whether jumping is enabled.
 */
BOOL
DoGenerations(LifeObject * obj)
{
	obj->born = numZero();
	obj->died = numZero();
	obj->growth = numZero();

	if (obj->genLeft <= 0)
		return FALSE;

	/*
	 * If we have a stop condition, check it, and if the stop condition
	 * is matched, don't calculate any more.  Note that this can only
	 * stop execution occasionally if jumping is enabled.
	 */
	if (StopSatisfied(obj))
	{
		obj->genLeft = 0;
		Message("Stop condition satisfied");

		return FALSE;
	}

	//
	// Set the currently used rule to the one for this object.
	//
	currentRule = obj->rule;

	//
	// If tracks are enabled then add the set of current cells to
	// the tracked cells.
	//
	if (obj->isTracksUsed)
		SaveUsedCells(obj);

	//
	// Determine the number of generations we can jump ahead.
	// If the frequency is set too high then we will have to do
	// multiple jumps.  The normal case is one generation.
	//
	Count genCount = MAX_JUMPING;

	if (genCount > obj->genLeft)
		genCount = obj->genLeft;

	if (genCount > obj->freqLeft)
		genCount = obj->freqLeft;

	if (!obj->isJumpingEnabled || (genCount <= 0))
		genCount = 1;

	//
	// Now try to actually calculate one or more generations.
	// The calculation will update the generation number and
	// the count, born, died, and growth counts.
	//
	genCount = RunGenerations(obj, genCount);

	//
	// If no generations were run then clear the generations left.
	//
	if (genCount == 0)
	{
		obj->genLeft = 0;

		return FALSE;
	}

	//
	// At least one generation was run.
	// Reduce the number of generations left and the frequency
	// left counter by the number of generations run.
	//
	if (obj->genLeft != INFINITY)
		obj->genLeft -= genCount;

	obj->freqLeft -= genCount;

	/*
	 * If the object has no more generations left or
	 * if the update frequency has been reached then
	 * mark the views of the object as needing updating
	 * and reset the frequency counter.
	 */
	if ((obj->genLeft <= 0) || (obj->freqLeft <= 0))
	{
		obj->freqLeft = obj->frequency;
		obj->changeCount++;
		AddWindowObjectUpdateFlags(obj, U_ALL);
	}

	return TRUE;
}


/*
 * Return the new cell state for the center of a 3 by 3 square of
 * cells depending on the old state of the nine cells.  These cells
 * are described in the comments at the top.  For example, the
 * middle cell of the square is cell 'i'.
 */
BOOL
IsNewCellAlive(BOOL cellA, BOOL cellB, BOOL cellC, BOOL cellD,
	BOOL cellE, BOOL cellF, BOOL cellG, BOOL cellH, BOOL cellI)
{
	int sum = 0;

	if (cellA) sum += CELLBIT('a');
	if (cellB) sum += CELLBIT('b');
	if (cellC) sum += CELLBIT('c');
	if (cellD) sum += CELLBIT('d');
	if (cellE) sum += CELLBIT('e');
	if (cellF) sum += CELLBIT('f');
	if (cellG) sum += CELLBIT('g');
	if (cellH) sum += CELLBIT('h');
	if (cellI) sum += CELLBIT('i');

	return (currentRule->rule[sum] != 0);
}


/*
 * Return whether or not the stop condition is satisfied by the
 * current object.
 */
static BOOL
StopSatisfied(const LifeObject * obj)
{
	/*
	 * If stopping is not enabled then return FALSE.
	 */
	if (!obj->stopIsEnabled)
		return FALSE;

	/*
	 * Check for an explicit cell changing state.
	 */
	if (obj->stopCell &&
		(IsCellSet(obj, obj->stopRow, obj->stopCol) == obj->stopState))
	{
		return TRUE;
	}

	/*
	 * Check for any cell being inside, outside, or on the boundary of
	 * the current path.  For efficiency only call the checking routine
	 * with non-null pointers for the numbers we actually need.
	 */
	if ((obj->stopOnPath || obj->stopInsidePath || obj->stopOutsidePath)
		&& (obj->path.count > 0))
	{
		NUMBER inCount = numZero();
		NUMBER onCount = numZero();
		NUMBER outCount = numZero();
		NUMBER * inPtr = (obj->stopInsidePath ? &inCount : 0);
		NUMBER * onPtr = (obj->stopOnPath ? &onCount : 0);
		NUMBER * outPtr = (obj->stopOutsidePath ? &outCount : 0);

		CountPathCells(obj, inPtr, onPtr, outPtr);

		if (numIsPos(onCount) || numIsPos(inCount) || numIsPos(outCount))
			return TRUE;
	}

	/*
	 * Check for the minimum gap size to be reached.
	 * This is the minimum distance between any marked call and
	 * any unmarked cell.
	 */
	if (obj->stopGap != 0)
	{
		Count gap;

		gap = numToLong(FindMarkedGap(obj));

		if ((gap != 0) && (gap <= obj->stopGap))
			return TRUE;
	}

	/*
	 * Check for a generic expression to stop for.
	 */
	if ((obj->stopExpr[0] != '\0') &&
		numIsNonZero(GetExpression(obj->stopExpr)))
	{
		return TRUE;
	}

	/*
	 * There is no reason to stop.
	 */
	return FALSE;
}

/* END CODE */
