/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdarg.h>

#include "life.h"


/*
 * The maximum input depth.
 */
#define	MAX_INPUT_DEPTH	20


/*
 * Types of input.
 */
typedef enum InputType
{ 
	INPUT_TERMINAL,		/* input is terminal */
	INPUT_FILE,		/* input is file */
	INPUT_LOOP,		/* input is command loop */
	INPUT_MACRO		/* input is command macro */
} InputType;


/*
 * The following structure holds all data necessary for processing
 * characters from some source.
 */
typedef	struct Input Input;

struct Input
{
	int	(*getChar)(Input *);	/* read next character */
	void (*term)(Input *);		/* terminate reading */
	InputType type;			/* type of input */

	/* 
	 * Data for file reading only.
	 */
	FILE *	file;		/* file handle */
	LifeObject * obj;	/* object to restore on reentry */
	NUMBER	row;		/* row to restore */
	NUMBER	col;		/* column to restore */
	NUMBER	origRow;	/* origin row to restore */
	NUMBER	origCol;	/* origin column to restore */

	/*
	 * Data for loop or macro reading only.
	 */
	char *	begPtr;		/* beginning of command data */
	char *	endPtr;		/* end of command data */
	char *	curPtr;		/* current character */
	NUMBER	curVal;		/* current iteration value */
	NUMBER	endVal;		/* ending iteration value */
	BOOL	first;		/* processing new chars */
	char	macro;		/* macro being defined */
};


/*
 * Structure for command macros.
 */
typedef	struct
{
	char *	begPtr;		/* beginning of data (NULL if none) */
	char *	endPtr;		/* end of data */
} Macro;



/*
 * Local procedures.
 */
static	int	TerminalChar(Input *);
static	void	TerminalTerm(Input *);
static	int	FileChar(Input *);
static	void	FileTerm(Input *);
static	int	LoopChar(Input *);
static	void	LoopTerm(Input *);
static	int	MacroChar(Input *);
static	void	MacroTerm(Input *);
static	FILE *	OpenLibraryFile(const char *, const char *);
static	BOOL	MatchFile(const char *, const char *);
static	int	GetLifeSuffixLength(const char * name);
static	BOOL	SpaceWait(BOOL isMoreComing);
static	int	SortFunc(const void *, const void *);
static	Macro *	FindMacro(int);
static	Input *	NextInputLevel(void);
static	void	UpdateStatus(LifeObject * obj);



/*
 * Local variables.
 */
static	int	helpRow;		/* current row of help text */
static	BOOL	helpBusy;		/* true if doing help display */
static	BOOL	helpAbort;		/* true if help is aborted */
static	BOOL	helpPaged;		/* true if help did multiple pages */
static	Macro	macros[26 * 2 + 9];	/* macros a-z, A-Z, and 1-9 */
static	char	stringBuf[STRING_SIZE];	/* characters for string value */


/*
 * The current input source.
 */
static	Input *	currentInput;

/*
 * The input source stack.
 */
static	Input	inputStack[MAX_INPUT_DEPTH];


/*
 * Initialize to read from the terminal as the top level input source.
 * This top level input level can not be closed.
 */
void
SetTopInputLevel(void)
{
	Input *	ip;

	ip = inputStack;

	ip->getChar = TerminalChar;
	ip->term = TerminalTerm;
	ip->type = INPUT_TERMINAL;

	currentInput = ip;
}


/*
 * Read the next character from the appropriate input source.
 * EOF is never returned by this routine (longjmps are done instead).
 * This routine is usually called indirectly by ScanChar.
 */
int
ReadChar(void)
{
	Input *	ip;
	int	ch;

	while (TRUE)
	{
		ip = currentInput;
		ch = ip->getChar(ip);

		if (ch != EOF)
			break;

		ip->term(ip);
	}

	return ch;
}


/*
 * Determine if the next input character to be read is from the terminal.
 * Returns TRUE if so.
 */
BOOL
IsTerminalInput(void)
{
	const Input *	ip;

	ip = currentInput;

	while ((ip->type == INPUT_LOOP) && (ip->first))
		ip--;

	return ip->type == INPUT_TERMINAL;
}


/*
 * Return the FILE that the current input level is reading from.
 * NULL is returned if the current input level is not a file.
 */
FILE *
GetInputFILE(void)
{
	const Input * ip;

	ip = currentInput;

	if (ip->type != INPUT_FILE)
		return 0;

	return ip->file;
}


/*
 * Set up to read commands from the terminal.  The lowest level of input
 * never quits, and is unusual in that it doesn't usually block waiting for
 * input.  Returns FALSE if unsuccessful.
 */
BOOL
SetTerminal(void)
{
	Input *	ip;

	ip = NextInputLevel();

	if (ip == 0)
		return FALSE;

	ip->getChar = TerminalChar;
	ip->term = TerminalTerm;
	ip->type = INPUT_TERMINAL;

	UpdateStatus(Life.currentObject);
	currentInput = ip;

	return TRUE;
}


/*
 * Read next character from the terminal if it is ready.  If nothing is
 * going on we will wait for it anyway, to prevent excessive runtimes.
 * If something is going on, we will longjmp away if input is not ready.
 */
static int
TerminalChar(Input * ip)
{
	LifeDevice * dev = Life.device;
	int ch;
	static BOOL	isWait;

	/*
	 * See if we must want for a character.
	 * This is true if either we are forced to wait,
	 * or if we are not running generations and no update
	 * is needed.
	 */
	isWait = (Life.doWait ||
		(!IsRunningGenerations() && !IsWindowUpdateNeeded()));

	if (isWait)
	{
		if (setjmp(Life.intJmp.buf))
		{
			Life.intJmpOk = FALSE;

			return '\0';
		}

		Life.intJmpOk = TRUE;
	}

	ch = dev->readChar(dev, isWait);

	Life.intJmpOk = FALSE;

	if ((ch == EOF) && !isWait)
		ScanEof();

	/*
	 * If a real character was not received then return that.
	 */
	if (ch == NULL_CMD)
		return NULL_CMD;

	/*
	 * A real character was read.
	 * Clear any message in the status window.
	 */
	dev->clearStatusMessage(dev);

	/*
	 * Clear the text in the view sub-subwindow if we are
	 * not in a command line.
	 */
	if (!Life.doWait)
			dev->setViewModeIsText(dev, FALSE);

	/*
	 * Return the character.
	 */
	return ch;
}


/*
 * Terminate reading from the terminal.  If we are reading from the lowest
 * level of terminal input, this is a no-op, since that level can never return.
 */
static void
TerminalTerm(Input * ip)
{
	if (ip > inputStack)
		ip--;

	currentInput = ip;
	UpdateStatus(Life.currentObject);
}


/*
 * Set up to read commands from a given file name.  When the file is complete,
 * some parameters (like the current cursor position) will be restored to their
 * original value.  Returns FALSE if cannot set up to read the file.
 */
BOOL
SetFile(const char * name)
{
	Input *	ip;
	FILE *	fp;
	LifeObject * obj;

	ip = NextInputLevel();

	if (ip == 0)
		return FALSE;

	fp = OpenLibraryFile(name, "r");

	if (fp == 0)
		return FALSE;

	obj = Life.currentObject;

	ip->file = fp;
	ip->getChar = FileChar;
	ip->term = FileTerm;
	ip->type = INPUT_FILE;
	ip->obj = obj;
	ip->row = obj->currentInfo.row;
	ip->col = obj->currentInfo.col;
	ip->origRow = obj->originInfo.row;
	ip->origCol = obj->originInfo.col;

	obj->originInfo.row = obj->currentInfo.row;
	obj->originInfo.col = obj->currentInfo.col;

	UpdateStatus(obj);

	currentInput = ip;

	/*
	 * Try to read the file as an picture, MC, RLE or XLIFE format.
	 * If it wasn't one of these, we will later on read the
	 * contents of the file as our own command format.
	 */
	if (TryReadFile(fp, obj))
		ip->term(ip);

	return TRUE;
}


/*
 * Open a life library file, trying various transformed names if necessary.
 * The order of the transformations which are tried is:
 *	1.  Name exactly as given.
 *	2.  Name with LIFE_EXT_1 appended to it.
 *	3.  Name with LIFE_EXT_2 appended to it.
 *	4.  Name with LIFE_EXT_3 appended to it.
 *	5.  Name in any of the user's library directories.
 *	6.  Name with LIFE_EXT_1 appended to it in the user's libraries.
 *	7.  Name with LIFE_EXT_2 appended to it in the user's libraries.
 *	8.  Name with LIFE_EXT_3 appended to it in the user's libraries.
 *	9.  Name in the system's library directory.
 *	10. Name with LIFE_EXT_1 appended to it in the system library.
 *	11. Name with LIFE_EXT_2 appended to it in the system library.
 *	12. Name with LIFE_EXT_3 appended to it in the system library.
 * Returns the file handle if the open is successful, or NULL if all the
 * open attempts failed.
 */
static FILE *
OpenLibraryFile(const char * name, const char * mode)
{
	const char * path;
	FILE *	fp;

	path = FindLibraryFile(name);

	if (path == 0)
		return 0;

	fp = fopen(path, mode);

	if (fp)
		return fp;

	return 0;
}


/*
 * Find a life library file, trying various transformed names if necessary.
 * The order of the transformations which are tried is:
 *	1.  Name exactly as given.
 *	2.  Name with LIFE_EXT_1 appended to it.
 *	3.  Name with LIFE_EXT_2 appended to it.
 *	4.  Name with LIFE_EXT_3 appended to it.
 *	5.  Name in any of the user's library directories.
 *	6.  Name with LIFE_EXT_1 appended to it in the user's libraries.
 *	7.  Name with LIFE_EXT_2 appended to it in the user's libraries.
 *	8.  Name with LIFE_EXT_3 appended to it in the user's libraries.
 *	9.  Name in the system's library directory.
 *	10. Name with LIFE_EXT_1 appended to it in the system library.
 *	11. Name with LIFE_EXT_2 appended to it in the system library.
 *	12. Name with LIFE_EXT_3 appended to it in the system library.
 *
 * Returns the path to the found file, or 0 if the file is not present.
 * The path is stored in a static buffer which is overwritten on each call.
 */
const char *
FindLibraryFile(const char * name)
{
	const char * path;
	int	i;
	static char	buf[MAX_PATH];

	/*
	 * Start by looking in the specified file name.
	 */
	path = FindFile(name);

	if (path)
		return path;

	/*
	 * If the name was absolute, then we fail now.
	 */
	if (*name == '/')
		return 0;

	/*
	 * The name is relative so we can look elsewhere.
	 * Search all of the user's specified libraries in order.
	 */
	for (i = 0; i < Life.userLibCount; i++)
	{
		strcpy(buf, Life.userLibs[i]);
		strcat(buf, "/");
		strcat(buf, name);

		path = FindFile(buf);

		if (path)
			return path;
	}

	/*
	 * Finally look in our system default library.
	 */
	strcpy(buf, LIFE_LIB);
	strcat(buf, "/");
	strcat(buf, name);

	return FindFile(buf);
}


/*
 * Find a file to be read using our set of Life extensions.
 * Returns the path to the file, or 0 if no such file exists.
 * The path is stored in a static buffer which is overridden on each call.
 */
const char *
FindFile(const char * name)
{
	static char	buf[MAX_PATH];

	/*
	 * First check the file exactly as given.
	 */
	strcpy(buf, name);

	if (access(buf, F_OK) == 0)
		return buf;

	/*
	 * Append the first suffix and check that.
	 */
	strcpy(buf, name);
	strcat(buf, LIFE_EXT_1);

	if (access(buf, F_OK) == 0)
		return buf;

	/*
	 * Append the second suffix and check that.
	 */
	strcpy(buf, name);
	strcat(buf, LIFE_EXT_2);

	if (access(buf, F_OK) == 0)
		return buf;

	/*
	 * Append the third suffix and check that.
	 */
	strcpy(buf, name);
	strcat(buf, LIFE_EXT_3);

	if (access(buf, F_OK) == 0)
		return buf;

	/*
	 * Append the fourth suffix and check that.
	 */
	strcpy(buf, name);
	strcat(buf, LIFE_EXT_4);

	if (access(buf, F_OK) == 0)
		return buf;

	/*
	 * The file could not accessed.
	 */
	return 0;
}


/*
 * Here to read next character from a file.  Called by ReadChar when
 * input source is a file.  Returns EOF on end of file.
 */
static int
FileChar(Input * ip)
{
	return fgetc(ip->file);
}


/*
 * Here on end of file or error to close the input file and restore some
 * of the previous state such as the cursor location.
 */
static void
FileTerm(Input * ip)
{
	LifeObject *obj;

	obj = Life.currentObject;

	(void) fclose(ip->file);

	obj = ip->obj;
	obj->currentInfo.row = ip->row;
	obj->currentInfo.col = ip->col;
	obj->originInfo.row = ip->origRow;
	obj->originInfo.col = ip->origCol;

	UpdateStatus(obj);

	currentInput = (ip - 1);

	if (IsTerminalInput())
	{
		SaveCurrentViewParameters();

		CopyObject(obj, Life.readObject);
	}
}


/*
 * Set up for execution of a loop.  This remembers the initial and final
 * loop values, and sets up to remember commands as they are given.
 * This is also used for defining a macro command.  The given character
 * will be assigned the string defined by the loop.
 * Returns FALSE if failed.
 */
BOOL
SetLoop(NUMBER begVal, NUMBER endVal, int ch)
{
	Input *	ip;

	ip = NextInputLevel();

	if (ip == 0)
		return FALSE;

	ip->begPtr = malloc(LOOP_SIZE);

	if (ip->begPtr == 0)
		return FALSE;

	ip->endPtr = ip->begPtr + LOOP_SIZE;
	ip->curPtr = ip->begPtr;
	ip->first = TRUE;
	ip->getChar = LoopChar;
	ip->term = LoopTerm;
	ip->type = INPUT_LOOP;
	ip->curVal = begVal;
	ip->endVal = endVal;
	ip->macro = ch;

	currentInput = ip;

	UpdateStatus(Life.currentObject);

	return TRUE;
}


/*
 * End the range of the currently defined loop.  At this point, all of
 * the characters of the loop have been read in and saved, and we can
 * just proceed to iterate over them.  The next read will find out that
 * the first iteration of the loop is over.
 */
void
EndLoop(void)
{
	Input *	ip;

	ip = currentInput;

	if (ip->type != INPUT_LOOP)
		Error("Loop not being defined");

	ip->endPtr = ip->curPtr - 1;	/* end before loop term cmd */
	ip->first = FALSE;
}


/*
 * Read one character from a loop buffer.  If at the end of the buffer, the
 * pointer is reset so that the buffer is reread.  When enough iterations
 * have been processed, EOF is returned.  A special case exists the first time
 * through the loop at the first nesting level, in that we don't yet have
 * the characters necessary for the loop, and so we have to read them by
 * ourself.
 */
static int
LoopChar(Input * ip)
{
	int	ch;

	/*
	 * If we are on the first interation of the loop,
	 * then we are collecting input characters.
	 */
	if (ip->first)
	{
		if (ip->curPtr >= ip->endPtr)
			Error("Loop too long");

		/*
		 * Get a character from the previous level.
		 */
		ch = ip[-1].getChar(ip - 1);

		if (ch == EOF)
			Error("End of file in loop");

		*ip->curPtr++ = ch;

		return ch;
	}

	/*
	 * See if we are done with one iteraction.
	 * If so, then we begin another one or return end of file.
	 */
	if (ip->curPtr >= ip->endPtr)
	{
		if (numIsEqual(ip->curVal, ip->endVal))
			return EOF;

		numAddrAddLong(&ip->curVal,
			(numIsLess(ip->curVal, ip->endVal) ? 1 : -1));

		ip->curPtr = ip->begPtr;
	}

	return (*ip->curPtr++ & 0xff);
}


/*
 * Terminate reading from a loop buffer.  If this was the definition of
 * a macro character, remember it for later.
 */
static void
LoopTerm(Input * ip)
{
	Macro *	mp;

	mp = FindMacro(ip->macro);

	if (mp)
	{
		if (mp->begPtr)
			free(mp->begPtr);

		mp->begPtr = ip->begPtr;
		mp->endPtr = ip->endPtr;
	}
	else
		free(ip->begPtr);	/* or free buffer */

	UpdateStatus(Life.currentObject);

	currentInput = (ip - 1);
}


/*
 * Find the macro structure for a macro character.
 * These are all upper and lower case letters and the digits 1 to 9.
 * Returns NULL if the macro character is invalid.
 */
Macro *
FindMacro(int ch)
{
	if ((ch >= 'a') && (ch <= 'z'))
		return &macros[ch - 'a'];

	if ((ch >= 'A') && (ch <= 'Z'))
		return &macros[ch - 'A' + 26];

	if ((ch >= '1') && (ch <= '9'))
		return &macros[ch - '1' + 26*2];

	return 0;
}


/*
 * Routine to return whether a macro character is valid.
 */
BOOL
MacroIsValid(int ch)
{
	return (FindMacro(ch) != 0);
}


/*
 * Set up to read a defined macro command.  Returns FALSE if failed.
 */
BOOL
SetMacro(NUMBER arg1, NUMBER arg2, int ch)
{
	Input *	ip;		/* current input */
	Macro *	mp;		/* macro command */

	mp = FindMacro(ch);

	if ((mp == 0) || (mp->begPtr == 0))
		return FALSE;

	ip = NextInputLevel();

	if (ip == 0)
		return FALSE;

	ip->getChar = MacroChar;
	ip->term = MacroTerm;
	ip->begPtr = mp->begPtr;
	ip->curPtr = mp->begPtr;
	ip->endPtr = mp->endPtr;
	ip->type = INPUT_MACRO;
	ip->macro = ch;
	ip->curVal = arg1;
	ip->endVal = arg2;

	currentInput = ip;

	UpdateStatus(Life.currentObject);

	return TRUE;
}


/*
 * Read next character from macro definition.
 * Returns EOF when the macro is finished.
 */
static int
MacroChar(Input * ip)
{
	if (ip->curPtr >= ip->endPtr)
		return EOF;

	return (*ip->curPtr++ & 0xff);
}


/*
 * Terminate reading from a macro definition.
 */
static void
MacroTerm(Input * ip)
{
	currentInput = (ip - 1);

	UpdateStatus(Life.currentObject);
}


/*
 * Allocate the next input level from the input stack if possible.
 * This does not make the input level active.
 * Returns NULL if there are no more input levels available.
 */
static Input *
NextInputLevel(void)
{
	Input *	ip = currentInput + 1;

	if (ip >= &inputStack[MAX_INPUT_DEPTH])
		return 0;

	return ip;
}


/*
 * Return whether or not we are at the top input level (the terminal).
 */
BOOL
IsTopInputLevel(void)
{
	return (currentInput == inputStack);
}


/*
 * End the current input level if possible.
 * This does nothing if we are at the top level input level.
 */
void
EndInputLevel(void)
{
	Input *	ip = currentInput;

	if (ip != inputStack)
		ip->term(ip);
}


/*
 * End all of the input levels except for the top level one.
 * This is called to clean up on errors.
 */
void
EndAllInputLevels(void)
{
	while (!IsTopInputLevel())
		EndInputLevel();
}


/*
 * Return a string describing the status of the current input level.
 * Returns NULL if the current input is the top level terminal.
 * The returned input string points to a static buffer which is
 * overwritten on each call.
 */
const char *
GetInputDescription(void)
{
	const Input * ip;
	char * cp;
	static char buf[256];

	/*
	 * Get the current input level.
	 */
	ip = currentInput;

	/*
	 * If we are at the top then return nothing.
	 */
	if (ip == inputStack)
		return 0;

	/*
	 * There is something to describe.
	 */
	cp = buf;

	sprintf(cp, "cmd-nest:%ld ", ((long) (ip - inputStack)));
	cp += strlen(cp);

	/*
	 * Switch on the input type.
	 */
	switch (ip->type)
	{
		case INPUT_TERMINAL:
			strcpy(cp, "tty-wait");

			break;

		case INPUT_FILE:
			strcpy(cp, "cmd-file");

			break;

		case INPUT_LOOP:
			if (ip->macro)
			{
				sprintf(cp, "macro-define-%c", ip->macro);

				break;
			}

			sprintf(cp, "loop%s (curVal:%s end:%s)",
				ip->first ? "-define" : "",
				numToString(ip->curVal),
				numToString(currentInput->endVal));

			break;

		case INPUT_MACRO:
			sprintf(cp, "macro-%c", ip->macro);

			break;

		default:
			strcpy(cp, "unknown");

			break;
	}

	/*
	 * Return the description.
	 */
	return buf;
}


/*
 * Return the current loop value for the specified looping depth
 * back from the current input level, where a value of 1 indicates
 * the most recent loop input level.  Generates an error if the
 * specified level does not exist.
 */
NUMBER
GetLoopValue(int depth)
{
	const Input * ip;

	for (ip = currentInput; ip != inputStack; ip--)
	{
		if (ip->type == INPUT_LOOP)
		{
			if (--depth == 0)
				return ip->curVal;
		}
	}

	Error("Bad loop depth");

	return numZero();
}


/*
 * Read a confirmation value from the user, which is "yes" or "no".
 * The supplied prompt is shown to the user to be answered.
 * An empty response returns the indicated default value.
 */
BOOL
Confirm(const char * prompt, BOOL defaultAnswer)
{
	const char *	answer;

	while (TRUE)
	{
		answer = ReadString(prompt);

		if (*answer == '\0')
			break;

		if (IsAbbreviation(answer, "yes"))
			return TRUE;

		if (IsAbbreviation(answer, "no"))
			return FALSE;
	}

	return defaultAnswer;
}


/*
 * Read a line from the user terminated by a newline (which is removed).
 * Editing of the input line is fully handled.  The prompt string and the
 * input line are only visible if the current input is from the terminal.
 * Returns a pointer to the null-terminated string.
 */
const char *
ReadString(const char * prompt)
{
	/*
	 * No more scan interference.
	 */
	ScanReset();

	/*
	 * Default the prompt to be empty if given null.
	 */
	if (prompt == 0)
		prompt = "";

	/*
	 * If we are not reading from the terminal then
	 * disable prompting totally.
	 */
	if (!IsTerminalInput())
		prompt = 0;

	/*
	 * We must want for tty chars.
	 */
	Life.doWait = TRUE;

	ReadLine(prompt, stringBuf, STRING_SIZE, 0, FALSE);

	Life.doWait = FALSE;

	if (prompt)
		UpdateStatus(Life.currentObject);

	if (Life.stop)
		Error("Command aborted");

	return stringBuf;
}


/*
 * Read a line from the user terminated by a newline (which is removed).
 * or by a tab character indicating that a partial line was read.
 * Editing of the input line is fully handled.  The prompt string and the
 * input line are only visible if the current input is from the terminal.
 * The specified initial data is inserted as part of the string and can
 * also be edited.  Returns a pointer to the null-terminated string.
 */
char *
ReadPartialString(const char * prompt, char * initBuf, int initCount)
{
	if ((initCount < 0) || (initCount > STRING_SIZE))
		Error("Initial string too large");

	if (initCount > 0)
		memcpy(stringBuf, initBuf, initCount);

	/*
	 * No more scan interference.
	 */
	ScanReset();

	/*
	 * Set prompt if given null.
	 */
	if (prompt == 0)
		prompt = "";

	/*
	 * No window stuff if not tty.
	 */
	if (!IsTerminalInput())
		prompt = 0;

	/*
	 * Must wait for tty chars.
	 */
	Life.doWait = TRUE;

	ReadLine(prompt, stringBuf, STRING_SIZE, initCount, TRUE);

	Life.doWait = FALSE;

	if (prompt)
		UpdateStatus(Life.currentObject);

	if (Life.stop)
		Error("Command aborted");

	return stringBuf;
}


/*
 * Read some input while possibly showing it on the status line.
 * If the prompt string is NULL, then editing is performed without
 * any display activity (useful when reading commands from scripts).
 * Otherwise, the prompt is shown in the window along with any input.
 * A trailing null character is inserted in the buffer.  The buffer may
 * contain initial data which will initially appear in the window, and
 * which can be edited by the user.  InitCount specifies the number of
 * initial characters in the buffer, so that a value of zero indicates
 * that there is no such data.  isPartial indicates whether a partial line
 * is allowed.  Partial lines are ended with a tab character, which is
 * the last character in the buffer.  Editing of the input is handled.
 * An escape character acts like the command was interrupted.
 * If the buffer fills up, the user is warned with beeps and further input
 * is ignored.  Returns the number of bytes of data read.
 */
int
ReadLine(const char * prompt, char * buf, int count,
	int initCount, BOOL isPartial)
{
	LifeDevice * dev = Life.device;
	int	ch;				/* character which was read */
	char *	bp;			/* current buffer pointer location */
	char *	endBp;		/* end of buffer */
	BOOL isRedrawNeeded;	/* need to redisplay input */

	dev->setStatusMode(dev, STATUS_MODE_INPUT);
	Life.canMakeInput = FALSE;
	isRedrawNeeded = TRUE;
	bp = buf;
	endBp = bp + count - 1;

	if (initCount > 0)
	{
		if (initCount >= count)
			initCount = count - 1;

		bp = &buf[initCount];
	}

	while (TRUE)
	{
		if (prompt)
		{
			if (isRedrawNeeded)
			{
				*bp = '\0';
				dev->setStatusMode(dev, STATUS_MODE_INPUT);
				dev->addStatusText(dev, prompt);
				dev->addStatusText(dev, buf);
				isRedrawNeeded = FALSE;
			}

			UpdateAllWindows();
		}

		ch = ReadChar();

		//
		// If this is the escape character then act like
		// we were interrypted.
		//
		if (ch == '\033')
			Life.stop = TRUE;

		/*
		 * If we had an interrupt, then return nothing.
		 */
		if (Life.stop)
		{
			buf[0] = '\0';
			dev->setStatusMode(dev, STATUS_MODE_NORMAL);
			Life.canMakeInput = TRUE;

			return 0;
		}

		/*
		 * Ignore nulls.
		 */
		if (ch == '\0')
			continue;

		/*
		 * If the literal quoting character was given,
		 * then read the next character without interpretation.
		 */
		if (ch == Life.vLiteralNext)
		{
			ch = ReadChar();

			if (Life.stop)
			{
				buf[0] = '\0';
				dev->setStatusMode(dev, STATUS_MODE_NORMAL);
				Life.canMakeInput = TRUE;

				return 0;
			}

			goto store;
		}

		/*
		 * If we allow partial lines ending with tab,
		 * then check for that and return it.
		 */
		if (isPartial && (ch == '\t'))
		{
			*bp++ = ch;
			*bp = '\0';
			dev->setStatusMode(dev, STATUS_MODE_NORMAL);
			Life.canMakeInput = TRUE;

			return bp - buf;
		}

		/*
		 * On an end of line, terminate the buffer and return
		 * the line length.
		 */
		if (ch == '\n')
		{
			*bp = '\0';
			dev->setStatusMode(dev, STATUS_MODE_NORMAL);
			Life.canMakeInput = TRUE;

			return bp - buf;
		}

		/*
		 * On either type of erase charcter, remove one character.
		 */
		if ((ch == Life.vErase1) || (ch == Life.vErase2))
		{
			if (bp <= buf)
				continue;

			bp--;
			isRedrawNeeded = TRUE;

			continue;
		}

		/*
		 * On a line erase character, clear the whole line.
		 */
		if (ch == Life.vKill)
		{
			bp = buf;
			isRedrawNeeded = TRUE;

			continue;
		}

		/*
		 * On a word erase character, delete the last blank
		 * delimited word.
		 */
		if (ch == Life.vWordErase)
		{
			if (bp <= buf)
				continue;

			while ((bp > buf) && ((bp[-1] == '\n')
				|| isBlank(bp[-1])))
			{
				bp--;
			}

			while ((bp > buf) && (bp[-1] != '\n')
				&& !isBlank(bp[-1]))
			{
				bp--;
			}

			isRedrawNeeded = TRUE;

			continue;
		}

store:
		/*
		 * Here to store a character.
		 * If the buffer is full, then beep at the user.
		 */
		if (bp >= endBp)
		{
			Beep();

			continue;
		}

		*bp = ch;
		bp[1] = '\0';

		if (prompt)
		{
			dev->addStatusText(dev, bp);
			UpdateAllWindows();
		}

		bp++;
	}
}


/*
 * Routine called to wait until a space, newline, or escape character
 * is typed.  It is assumed that the screen contains a help display
 * that we can append our message to.  If the isMoreComing flag is true
 * and the user types escape instead of a space, then we return true.
 */
static BOOL
SpaceWait(BOOL isMoreComing)
{
	LifeDevice * dev = Life.device;
	int	ch;

	/*
	 * Throw out any stored characters.
	 */
	ScanReset();

	if (isMoreComing)
		dev->addViewText(dev, "\n[Type <space> to continue or <esc> to return]");
	else
		dev->addViewText(dev, "\n[Type <space> to return]");

	Life.currentWindow->updateFlags |= U_VIEW;

	UpdateAllWindows();

	/*
	 * Set the flag indicating that we must wait for chars.
	 */
	Life.doWait = TRUE;
	Life.canMakeInput = FALSE;

	do {
		if (Life.stop)
		{
			ch = ESC;

			break;
		}

		ch = ReadChar();
	}
	while ((ch != ' ') && (ch != ESC) && (ch != '\n'));

	Life.canMakeInput = TRUE;
	Life.doWait = FALSE;

	return (isMoreComing && (ch == ESC));
}


/*
 * Write all defined macros to a the specifiled FILE handle so that
 * they can be read back in later.
 */
void
WriteMacros(FILE * fp)
{
	const Macro * mp;
	const char * cp;
	int		ch;

	fputs("! macros\n", fp);

	for (ch = 0; ch < 256; ch++)
	{
		mp = FindMacro(ch);

		if ((mp == 0) || (mp->begPtr == 0) ||
			(mp->begPtr >= mp->endPtr))
		{
			continue;
		}

		fprintf(fp, "<%c", ch);

		for (cp = mp->begPtr; cp < mp->endPtr; cp++)
			fputc(*cp, fp);

		fputs(">!\n", fp);
	}
}


/*
 * Collect all life files under a directory which contain
 * the specified pattern.  The names are appended to the given list
 * of strings.  If desired, the list can have uninitialized pointers
 * (indicated by max counts of zero).  For speed, the assumption is
 * made that files with one of our suffixes are valid life files, and
 * so are not directories.  The skip length is how many characters of
 * the given directory name are not to be displayed in the list of files.
 * The depth is how deep to allow the search to go.
 */
void
GetFiles(List * list, const char * dirName, const char * pattern,
	int skipLen, int depth)
{
	DIR * dirp;
	const struct dirent * dp;
	const char * name;
	char * nameAppend;
	int suffixLen;
	char fullName[MAX_PATH];

	dirp = opendir(dirName);

	if (dirp == 0)
		return;

	depth--;
	strcpy(fullName, dirName);
	strcat(fullName, "/");
	nameAppend = fullName + strlen(fullName);

	while ((dp = readdir(dirp)) != 0)
	{
		name = dp->d_name;
		strcpy(nameAppend, name);

		suffixLen = GetLifeSuffixLength(name);

		if (suffixLen == 0)
		{
			if (depth && strcmp(name, ".") && strcmp(name, ".."))
				GetFiles(list, fullName, pattern, skipLen, depth);

			continue;
		}

		nameAppend[strlen(name) - suffixLen] = '\0';

		if (!(*pattern) || MatchFile(fullName + skipLen, pattern))
			AddListItem(list, fullName + skipLen);
	}

	closedir(dirp);
}


/*
 * Collect all files having our life extension and directories
 * inside the specified directory which have a full path name starting
 * with the specified prefix string. The prefix string may have
 * extra directory components which will be used. The names are
 * appended to the given list of strings.  Files are not recursively
 * examined inside directories.  Directories have a trailing slash
 * appended to indicate their status, and life files have their suffix
 * removed.  For speed, the assumption is made that files having one
 * of our extensions are valid life files.  This routine is intended
 * to support filename completion.
 */
void
GetCompletionFiles(List * list, const char * dirName, const char * prefix)
{
	DIR * dirp;
	const struct dirent * dp;
	const char * prefixTail;
	const char * prefixSlash;
	const char * entryName;
	char *	nameAppend;
	int		entryLen;
	int		dirNameLen;
	int		prefixDirLen;
	int		prefixTailLen;
	int		suffixLen;
	char	fullName[MAX_PATH];
	char	fullDirName[MAX_PATH];
	char	tempEntry[MAX_PATH];
	struct stat statBuf;

	dirNameLen = strlen(dirName);

	/*
	 * Given the specified directory name, append the directory
	 * components from the prefix to it, if any.
	 */
	prefixSlash = strrchr(prefix, '/');

	if (prefixSlash)
	{
		prefixDirLen = prefixSlash - prefix;
		prefixTail = prefixSlash + 1;

		/*
		 * If the prefix is an absolute path, then only use that.
		 * Otherwise append the directory part of it to the
		 * given directory.
		 */
		if (*prefix == '/')
		{
			memcpy(fullDirName, prefix, prefixDirLen);
			fullDirName[prefixDirLen] = '\0';

			if (prefixDirLen == 0)
				strcpy(fullDirName, "/");
		}
		else
		{
			memcpy(fullDirName, dirName, dirNameLen);
			fullDirName[dirNameLen] = '/';

			memcpy(&fullDirName[dirNameLen + 1],
				prefix, prefixDirLen);

			fullDirName[dirNameLen + prefixDirLen + 1] = '\0';
		}
	}
	else
	{
		prefixTail = prefix;
		prefixDirLen = 0;
		strcpy(fullDirName, dirName);
	}

	/*
	 * Open the directory name we have gotten to after appending the
	 * directory from the prefix into the starting directory.
	 */
	dirp = opendir(fullDirName);

	if (dirp == 0)
		return;

	prefixTailLen = strlen(prefixTail);

	strcpy(fullName, fullDirName);
	strcat(fullName, "/");
	nameAppend = fullName + strlen(fullName);

	/*
	 * Read all of the entries in the directory,
	 * collecting those which match the prefix and which
	 * end in a Life suffix or which are directories.
	 */
	while ((dp = readdir(dirp)) != 0)
	{
		entryName = dp->d_name;
		entryLen = strlen(entryName);
		strcpy(nameAppend, entryName);

		/*
		 * If the entry name does not begin with the indicated
		 * prefix then skip it.
		 */
		if ((prefixTailLen > 0) &&
			(memcmp(entryName, prefixTail, prefixTailLen) != 0))
		{
			continue;
		}

		/*
		 * If the file has the right Life suffix then add the
		 * name to the list, without its suffix.
		 */
		suffixLen = GetLifeSuffixLength(entryName);

		if (suffixLen > 0)
		{
			strcpy(tempEntry, entryName);
			tempEntry[entryLen - suffixLen] = '\0';

			AddListItem(list, tempEntry);

			continue;
		}

		/*
		 * Now look for directories, ignoring the usual ones.
		 * If the file is not a directory, then ignore it.
		 */
		if ((strcmp(entryName, ".") == 0) ||
			(strcmp(entryName, "..") == 0))
		{
			continue;
		}

		if ((stat(fullName, &statBuf) < 0) || !S_ISDIR(statBuf.st_mode))
			continue;

		/*
		 * We found a directory, add it with a trailing slash.
		 */
		strcpy(tempEntry, entryName);
		strcat(tempEntry, "/");
		AddListItem(list, tempEntry);
	}

	closedir(dirp);
}


/*
 * Check the specified path to see if it ends in one of our life suffixes.
 * If so, returns the length of the matched suffix (including the dot).
 * If not, returns zero.
 */
static int
GetLifeSuffixLength(const char * name)
{
	int length;

	length = strlen(name);

	if ((length >= sizeof(LIFE_EXT_1)) &&
		(strcmp(name + length - sizeof(LIFE_EXT_1) + 1, LIFE_EXT_1) == 0))
	{
		return sizeof(LIFE_EXT_1) - 1;
	}

	if ((length >= sizeof(LIFE_EXT_2)) &&
		(strcmp(name + length - sizeof(LIFE_EXT_2) + 1, LIFE_EXT_2) == 0))
	{
		return sizeof(LIFE_EXT_2) - 1;
	}

	if ((length >= sizeof(LIFE_EXT_3)) &&
		(strcmp(name + length - sizeof(LIFE_EXT_3) + 1, LIFE_EXT_3) == 0))
	{
		return sizeof(LIFE_EXT_3) - 1;
	}

	if ((length >= sizeof(LIFE_EXT_4)) &&
		(strcmp(name + length - sizeof(LIFE_EXT_4) + 1, LIFE_EXT_4) == 0))
	{
		return sizeof(LIFE_EXT_4) - 1;
	}

	return 0;
}


/*
 * Add a string to the specified list.
 */
void
AddListItem(List * list, const char * name)
{
	int	len;
	int	size;
	char *	mem;

	len = strlen(name);

	if (list->argc >= list->maxArgc)
	{
		size = (list->maxArgc + LIST_ARG_SIZE) * sizeof(char *);

		if (list->maxArgc)
			mem = realloc((char *) list->argv, size);
		else
			mem = malloc(size);

		if (mem == 0)
			return;

		list->argv = (char **) mem;
		list->maxArgc += LIST_ARG_SIZE;
	}

	if (list->used + len >= list->maxUsed)
	{
		size = list->maxUsed + len + LIST_BUF_SIZE;

		if (list->maxUsed)
			mem = realloc(list->buf, size);
		else
			mem = malloc(size);

		if (mem == 0)
			return;

		for (size = 0; size < list->argc; size++)
		{
			list->argv[size] = mem +
				(list->argv[size] - list->buf);
		}

		list->buf = mem;
		list->maxUsed += len + LIST_BUF_SIZE;
	}

	list->argv[list->argc] = &list->buf[list->used];
	strcpy(&list->buf[list->used], name);
	list->used += (len + 1);
	list->argc++;
}


/*
 * Display a list of strings in the help window.
 * If many strings are displayed, the user is asked to
 * page through them all.  The isWaitRequired flag is TRUE
 * if the user must type something to end the display.
 */
void
DisplayList(const List * list, BOOL isWaitRequired)
{
	int	colWidth;
	int	col;
	int	len;
	int padding;
	int	argc;
	char **	argv;
	char *	cp;

	if (list->argc <= 0)
		return;

	colWidth = MAX_FILE - sizeof(LIFE_EXT_2) + 2;
	col = 0;

	argv = list->argv;
	argc = list->argc;

	while (argc-- > 0)
	{
		cp = *argv++;

		len = strlen(cp) + 2;

		padding = colWidth - (col % colWidth);

		if (padding == colWidth)
			padding = 0;

		if ((col > 0) && (col + padding + len >= Life.currentWindow->textCols))
		{
			ShowHelp("\n");
			col = 0;
			padding = 0;
		}

		while (padding-- > 0)
		{
			ShowHelp(" ");
			col++;
		}

		ShowHelp(cp);
		ShowHelp("  ");

		col += len;
	}

	if (col)
		ShowHelp("\n");

	EndHelp(isWaitRequired);
}


/*
 * Initialize a list structure for use.
 */
void
InitList(List * list)
{
	list->argv = 0;
	list->buf = 0;
	list->argc = 0;
	list->maxArgc = 0;
	list->used = 0;
	list->maxUsed = 0;
}


/*
 * Free any data contained in a list.
 */
void
FreeList(List * list)
{
	if (list->argv)
		free((char *) list->argv);

	if (list->buf)
		free(list->buf);

	InitList(list);
}


/*
 * Sort and remove duplicate entries from a list.
 */
void
SortList(List * list)
{
	int	di;
	int	si;

	if (list->argc < 2)
		return;

	/*
	 * Sort the strings.
	 */
	qsort((void *) list->argv, list->argc, sizeof(char *), SortFunc);

	/*
	 * Walk down the argument list and shuffle up identical strings.
	 * The store the count of unique strings.
	 */
	di = 0;

	for (si = 1; si < list->argc; si++)
	{
		if (strcmp(list->argv[si], list->argv[di]) != 0)
			list->argv[++di] = list->argv[si];
	}

	list->argc = di + 1;
}


/*
 * Determine the common prefix length for the entries in a list.
 * This number of characters will be in common for all of the entries.
 * If there are no common characters, then zero is returned.
 */
int
GetListPrefixLength(const List * list)
{
	const char * first;
	int		len;
	int		newLen;
	int		i;

	if (list->argc <= 0)
		return 0;

	first = list->argv[0];
	len = strlen(first);

	for (i = 1; (i < list->argc) && (len > 0); i++)
	{
		newLen = 0;

		while ((newLen < len) &&
			(first[newLen] == list->argv[i][newLen]))
		{
			newLen++;
		}

		len = newLen;
	}

	return len;
}


/*
 * Sort function for strings.
 * This is a alphabetic-numeric sort, so numbers appear in a natural order.
 */
static int
SortFunc(const void * cp1, const void * cp2)
{
	const char * s1;
	const char * s2;
	int	c1;
	int	c2;
	long n1;
	long n2;

	s1 = *((const char **) cp1);
	s2 = *((const char **) cp2);

	while (TRUE)
	{
		c1 = *s1++;
		c2 = *s2++;

		if ((c1 == '\0') && (c2 == '\0'))
			return 0;

		if (!isDigit(c1) || !isDigit(c2))
		{
			if (c1 == c2)
				continue;

			return c1 - c2;
		}

		n1 = c1 - '0';
		n2 = c2 - '0';

		while (isDigit(*s1))
			n1 = n1 * 10 + *s1++ - '0';

		while (isDigit(*s2))
			n2 = n2 * 10 + *s2++ - '0';

		if (n1 == n2)
			continue;

		return n1 - n2;
	}
}


/*
 * Format some text in the manner of printf and add the text to the
 * help screen.  Warning: The caller must make sure that the buffer is
 * not overflowed (1024 bytes).
 */
BOOL
FormatHelp(const char * format, ...)
{
	va_list	ap;
	char buf[1024];

	va_start(ap, format);
	vsprintf(buf, format, ap);
	va_end(ap);

	return ShowHelp(buf);
}


/*
 * Add some text to the help screen.  The text may contain newlines,
 * but line wrapping is not done.  If the text goes to the bottom of
 * the screen, the output is stopped and the user is requested to type
 * a space to proceed, or escape to abort.  If the user aborts, then
 * all further help output is suppressed until the EndHelp call is made,
 * and this routine returns nonzero.  If this routine is called, then
 * EndHelp MUST also be called before non-help output occurs.
 */
BOOL
ShowHelp(const char * str)
{
	LifeDevice * dev = Life.device;
	const char * cp;
	int	len;
	char buf[256];

	if (helpBusy && helpAbort)
		return TRUE;

	if (!helpBusy)
	{
		dev->setViewModeIsText(dev, TRUE);
		helpRow = 0;
		helpAbort = FALSE;
		helpPaged = FALSE;
		helpBusy = TRUE;
	}

	while (*str)
	{
		if (helpRow > Life.currentWindow->textRows - 3)
		{
			helpPaged = TRUE;

			if (SpaceWait(TRUE))
			{
				dev->setViewModeIsText(dev, FALSE);
				
				helpAbort = TRUE;

				return TRUE;
			}

			dev->setViewModeIsText(dev, TRUE);
			helpRow = 0;
		}

		cp = strchr(str, '\n');

		if ((cp == 0) || (cp[1] == '\0'))
		{
			dev->addViewText(dev, str);

			if (cp)
				helpRow++;

			return FALSE;
		}

		/*
		 * Copy the line and null terminate it in order to display it.
		 */
		cp++;

		while (str != cp)
		{
			len = (cp - str);

			if (len >= sizeof(buf))
				len = sizeof(buf) - 1;

			memcpy(buf, str, len);
			buf[len] = '\0';
			dev->addViewText(dev, buf);
			str += len;
		}

		helpRow++;
	}

	return FALSE;
}


/*
 * End the help display, by possibly waiting for the user to type a space.
 * The space is needed only if we are asked to wait, or if the display
 * went to multiple pages.  The view sub-window will continue to show the
 * last text displayed until another character is typed.
 */
void
EndHelp(BOOL isWaitRequired)
{
	if (!helpBusy)
		return;

	/*
	 * If we need to wait for a space, then do that.
	 */
	if (!helpAbort && (isWaitRequired || helpPaged))
	{
		SpaceWait(FALSE);
	}

	helpBusy = FALSE;
	helpAbort = FALSE;
	helpPaged = FALSE;
	helpRow = 0;
}


/*
 * See if the specified file name is matched by the specified pattern.
 * For our purposes, matching means the pattern is contain somewhere
 * within the name.
 */
static BOOL
MatchFile(const char * name, const char * pattern)
{
	if (*pattern == '\0')
		return TRUE;

	return (strstr(name, pattern) != 0);
}


static void
UpdateStatus(LifeObject * obj)
{
	AddWindowObjectUpdateFlags(obj, U_STATUS);
}

/* END CODE */
