/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Program to play Conway's game of LIFE on an infinite board.
 * This file contains global definitions for the program.
 */

#ifndef	LIFE_H
#define	LIFE_H


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>

#include "quad.h"
#include "number.h"
#include "verify.h"


/*
 * Special characters.
 */
#define	ESC	'\033'
#define	FF	'\014'
#define	DEL	'\177'


/*
 * Special character value for a command which does nothing.
 * This is sometimes used when events must return something
 * in order to have the command loop start again from the top.
 * This value must be nonzero.
 */
#define	NULL_CMD	0xff


/*
 * Character tests.
 */
#define	isBlank(ch)	(((ch) == ' ') || ((ch) == '\t'))
#define	isDigit(ch)	(((ch) >= '0') && ((ch) <= '9'))
#define	isLower(ch)	(((ch) >= 'a') && ((ch) <= 'z'))
#define	isUpper(ch)	(((ch) >= 'A') && ((ch) <= 'Z'))
#define	isLetter(ch)	(isLower(ch) || isUpper(ch))


//
// LocationInfo id values and tests.
// These are the lower case letters for the where locations,
// '@' for the origin, or '.' for the current location.
//
#define	LOCINFO_ID_CURRENT	'.'
#define	LOCINFO_ID_ORIGIN	'@'

#define	IS_LOCINFO_ID(ch)	((((ch) >= 'a') && ((ch) <= 'z')) || \
	((ch) == LOCINFO_ID_CURRENT) || ((ch) == LOCINFO_ID_ORIGIN))


/*
 * The directory containing the general life library.
 * This can be overridden while compiling.
 */
#ifndef	LIFE_LIB
#define	LIFE_LIB	"/usr/local/games/lib/life"
#endif


/*
 * Configuration values.
 */
#define	MAX_WRITE	10000000	/* maximum non-quadcell cells written */
#define	MAX_PATH_CELLS	1000000		/* maximum cells scanned for paths */
#define	LIFE_OPTS	"LIFEOPTS"	/* environment name for options */
#define	LIFE_EXT_1	".l"		/* first extension for life files */
#define	LIFE_EXT_2	".lif"		/* second extension for life files */
#define	LIFE_EXT_3	".rle"		/* third extension for life files */
#define	LIFE_EXT_4	".mc"		/* fourth extension for life files */
#define	MAX_PATH	1024		/* maximum size of path names */
#define	MAX_FILE	18		/* maximum size of file name for list */
#define	MAX_BACKUPS	10		/* maximum number of backups */
#define	MAX_PREFIX	128		/* maximum size of pathname prefix */
#define	MAX_LIST	5		/* maximum depth for listing files */
#define	MAX_LIBS	10		/* maximum number of user libraries */
#define	MAX_ALIASES	100		/* maximum number of command aliases */
#define	ALLOC_OBJ_SIZE	10		/* how many new objects to allocate */
#define	ALLOC_ROW_SIZE	100		/* how many new rows to allocate */
#define	ALLOC_CELL_SIZE	1000		/* how many new cells to allocate */
#define	LIST_ARG_SIZE	200		/* incremental arg size for lists */
#define	LIST_BUF_SIZE	2000		/* incremental buffer size for lists */
#define	STRING_SIZE	(1024*4)	/* temp string buffer size */
#define	ARGS_SIZE	100		/* maximum number of command line args */
#define	LOOP_SIZE	5000		/* characters in command loops */
#define	COMMAND_SIZE	512		/* maximum command length */
#define	SCALE_LIST_SIZE	10		/* maximum number of scales in a list */
#define	MAX_ARGS	10		/* maximum command line arguments */
#define	MAX_SCALE	100000000	/* maximum scale factor */
#define	MAX_DELAY	(1000*60)	/* maximum delay time (millisec) */
#define	MAX_JUMPING	(1024*1024*1024) /* maximum jumping for generations */
#define	MAX_GAP		100		/* maximum size of a gap */
#define	MAX_SETTINGS_NAME	32	/* maximum size of settings name */
#define	MIN_SCALE	-100		/* minimum scale factor */
#define	DEFAULT_SHIFT	5		/* default shifting factor */
#define	DEFAULT_GRID_SIZE	10	/* default grid size */
#define	DEFAULT_GRID_CHAR_1	'.'	/* default first grid character */
#define	DEFAULT_GRID_CHAR_2	','	/* default second grid character */
#define	MIN_SHIFT	1		/* minimum shifting factor */
#define	MAX_SHIFT	100		/* maximum shifting factor */
#define	WRITE_ROWS	100		/* default maximum rows for writing */
#define	WRITECOLS	79		/* default maximum cols for writing */
#define	DEFAULT_RULE	"B3/S23"	/* default Life rule */
#define	DEFAULT_FONT	"fixed"		/* default font name for graphics */
#define	MAX_WINDOWS	10		/* maximum number of opened windows */
#define	MAX_SETTINGS	10		/* maximum number of settings */
#define	MAX_COLORS	100		/* maximum number of colors */
#define	TOP_LEVEL_SIZE_PERCENTAGE 90	/* default size percentage for top level */
#define	SUB_WINDOW_SIZE_PERCENTAGE 25	/* default size percentage for sub window */


/*
 * Behavior when adding objects together.
 */
typedef enum AddObjectMode
{
	ADD_RELATIVE,			/* do object additions relatively */
	ADD_ABSOLUTE			/* do object additions absolutely */
} AddObjectMode;


/*
 * Behavior of movement.
 */
typedef enum MoveMode
{
	MOVE_MODE_MOVE,			/* move only */
	MOVE_MODE_INSERT,		/* insert cells while moving */
	MOVE_MODE_DELETE		/* delete cells while moving */
} MoveMode;


/*
 * Mark bit values.
 */
#define	MARK_NONE	((Mark) 0x00)	/* no marks */
#define	MARK_SELECT	((Mark) 0x01)	/* marked only for current selection */
#define	MARK_SEARCH	((Mark) 0x02)	/* marked only by searching */
#define	MARK_USER	((Mark) 0x04)	/* marked explicitly by the user */
#define	MARK_USED	((Mark) 0x08)	/* cell is known to be in usedquadcells */


/*
 * Status display mode for the status sub-window.
 */
typedef enum StatusMode
{
	STATUS_MODE_NORMAL,		/* object information */
	STATUS_MODE_INPUT,		/* command line input */
	STATUS_MODE_ERROR,		/* error message text */
	STATUS_MODE_MESSAGE		/* requested message text */
} StatusMode;


/*
 * Scanning longjmp values.
 */
typedef enum ScanJumpValue
{
	SCAN_NONE,		/* setjmp initial value (must be zero) */
	SCAN_ABORT,		/* setjmp value for aborted command */
	SCAN_EDIT,		/* setjmp value for edited command */
	SCAN_EOF		/* setjmp value for no command data */
} ScanJumpValue;


/*
 * Flags for updating of display.
 * These are used in a call to setUpdateFlags.
 */
#define	U_NONE		((Flags) 0x00)	/* no update */
#define	U_POS		((Flags) 0x01)	/* update cursor position */
#define	U_STATUS	((Flags) 0x02)	/* update status line */
#define	U_VIEW		((Flags) 0x04)	/* update view of cells */
#define	U_OUTLINE	((Flags) 0x08)	/* update outline rectangles */
#define	U_CLEAR		((Flags) 0x10)	/* clear screen before drawing */
#define	U_CONFIG	((Flags) 0x20)	/* window configuration changed */
#define	U_ALL		((Flags) (U_POS | U_OUTLINE | U_STATUS | U_VIEW))	/* most */
#define	U_REDRAW	((Flags) (U_ALL | U_CLEAR))	/* refresh too */


/*
 * Viewing types.
 * This is the location in the object which the view window
 * tries to show in the view, and which is placed in the
 * center of the window when recentering is required.
 * Note: the view can be of other window's data instead of
 * our own window.
 */
typedef enum ViewType
{
	VIEW_TYPE_CURSOR,		/* object's cursor */
	VIEW_TYPE_ORIGIN,		/* object's origin */
	VIEW_TYPE_MOUSE,		/* current mouse position */
	VIEW_TYPE_UNDER,		/* position underneath subwindow */
	VIEW_TYPE_SELF,			/* current window position */
	VIEW_TYPE_WHERE_A = 'a',	/* object's 'a' where location */
	VIEW_TYPE_WHERE_Z = 'z'		/* object's 'z' where location */
} ViewType;


/*
 * Positioning types.
 * This defines what is done to the window to make the
 * viewing position visible when necessary.
 */
typedef enum PositionType
{
	POSITION_TYPE_SHIFT,		/* shift view to make position visible */
	POSITION_TYPE_CENTER,		/* center view around position */
	POSITION_TYPE_SCALE		/* scale view to make position visible */
} PositionType;


/*
 * Shape of cells to draw.
 * Squares are the default.
 */
typedef enum CellShape
{
	SHAPE_SQUARE,			/* draw squares for cells */
	SHAPE_CIRCLE,			/* draw circles for cells */
	SHAPE_DIAMOND			/* draw diamonds for cells */
} CellShape;


/*
 * Type of grid to draw.
 * Lines are the default.
 */
typedef enum GridType
{
	GRID_TYPE_LINES,		/* draw lines on edges */
	GRID_TYPE_CHECKERBOARD,		/* draw checkboard */
	GRID_TYPE_DOTS			/* draw dots in corners */
} GridType;


/*
 * Status window output locations.
 * Depending on the value, the status window will be hidden.
 */
typedef enum StatusLocation
{
	STATUS_LOCATION_SELF,	/* output status text to our own window */
	STATUS_LOCATION_PARENT,	/* output status text to our parent window */
	STATUS_LOCATION_NONE	/* don't output status text at all */
} StatusLocation;


/*
 * Macro to detect reserved names.
 */
#define	BAD_NAME(s)	((s[0] == '.') && ((s[1] == '\0') || (s[1] == '.')))


/*
 * Basic typedefs.
 */
typedef	long long BigCount;	/* very large count value */
typedef	long Coord;		/* coordinate of a cell */
typedef	long Count;		/* counts of things */
typedef	long Value;		/* value of an expression */
typedef	int Scale;		/* scale factor */
typedef	int PixelCoord;		/* device pixel coordinate */
typedef	int PixelCount;		/* counts of device pixels */
typedef	unsigned int Flags;	/* flag values */
typedef	unsigned long Mark;	/* mark value */


/*
 * The maximum value for coordinates, counts, and values.
 * This assumes that the these values are all longs.
 */
#define	INFINITY	((long) (((unsigned long) -1L) >> 1))


/*
 * Our boolean type is in all upper case to not be confused
 * with system include files.  Also we check to see if the
 * values are defined.
 */
#ifndef BOOL
#define BOOL int
#endif

#ifndef	TRUE
#define	TRUE		((BOOL) 1)
#endif

#ifndef	FALSE
#define	FALSE		((BOOL) 0)
#endif


/*
 * A color index value.
 * These are used to index into a table of color names and
 * also to index into a device's table of pixel values.
 * The first two color indexes are always present and are
 * black and white.  Higher color indexes depend on the
 * order of allocation of colors.  A value of -1 indicates
 * an invalid color index.
 */
typedef	int	ColorIndex;

#define	COLOR_INDEX_NONE	-1
#define	COLOR_INDEX_BLACK	0
#define	COLOR_INDEX_WHITE	1


/*
 * Algorithm names.
 * We assume that these are unique on their first character.
 */
#define	LIST_NAME	"list"
#define	QUAD_NAME	"quadcell"


/*
 * Forward declarations.
 */
typedef	struct LifeObject LifeObject;
typedef struct LifeRule LifeRule;
typedef struct LifeWindow LifeWindow;
typedef	struct LifeDevice LifeDevice;
typedef	struct LifeCell LifeCell;
typedef	struct LifeRow LifeRow;
typedef	struct LifeAlgorithm LifeAlgorithm;


/*
 * Structure to hold row and column pairs and an extra boolean value.
 */
typedef struct
{
	NUMBER	row;
	NUMBER	col;
	BOOL	flag;
} Location;


//
// Structure to hold viewing information about the where locations
// and the origin.
//
typedef struct
{
	int	id;		// character id of the location (letters or '@')
	NUMBER	row;		// current row of the location
	NUMBER	col;		// current column of the location
	BOOL	isEnabled;	// whether showing anything is enabled
	BOOL	isIdShown;	// is the character id shown
	BOOL	isForwShown;	// is the forwards diagonal line shown
	BOOL	isBackShown;	// is the backwards diagonal line shown
	BOOL	isHorizShown;	// is the horizontal line shown
	BOOL	isVertShown;	// is the vertical line shown
	BOOL	isDistShown;	// is distances to cursor shown
} LocationInfo;


/*
 * Structure to hold the coordinates of a rectangular area.
 * The coordinates include the sides of the rectangle.
 */
typedef	struct
{
	NUMBER	minRow;
	NUMBER	maxRow;
	NUMBER	minCol;
	NUMBER	maxCol;
} Area;


/*
 * Structure holding information about a path.
 * Paths are used to manipulate selected regious of an object.
 * A path can consist of a single loop, or else multiple disjoint loops.
 * An extra point is assumed at the end of the last loop to close it.
 */
typedef	struct
{
	int	count;		/* number of points in path */
	int	maxCount;	/* maximum points storable */
	int	oldCount;	/* number of points before clearing */
	int	begin;		/* beginning offset of latest path */
	NUMBER	minRow;		/* bounds of the whole path */
	NUMBER	maxRow;
	NUMBER	minCol;
	NUMBER	maxCol;
	Location * locs;	/* table of coordinates */
} Path;


/*
 * A simple string buffer.
 * All the strings are null-terminated.
 * The length doesn't include the null byte.
 * If the allocated size is zero then the data pointer is unspecified.
 * This will be true for a static instance of a string buffer.
 */
typedef struct
{
	int alloc;		// Allocated space of buffer
	int len;		// current string length (without null)
	char * data;		// current string
} StringBuffer;


/*
 * Rule information.
 */
#define	RULESIZE	512

struct LifeRule
{
	struct LifeRule * next;	/* next rule */
	char rule[RULESIZE];	/* neighborhood values for cells */
	char name[256];		/* name of rule */
};


/*
 * Structure for each cell.
 * These are linked together in rows.
 */
struct LifeCell
{
	LifeCell * nextCell;		/* link to next cell */
	NUMBER	col;			/* column number */
	Mark	marks;			/* marking values for cell */
};


/*
 * Structure for each row of cells.
 */
struct LifeRow
{
	LifeRow * nextRow;		/* link to next row */
	LifeCell * firstCell;		/* link to first cell */
	LifeCell * lastCell;		/* link to last real cell */
	NUMBER	row;			/* row number */
	NUMBER	count;			/* number of cells in this row */
};


/*
 * Structure for each object.
 * Objects exists whether or not they are displayed in a window.
 * They can be displayed in multiple windows at the same time.
 * The current position exists independently of a window, but
 * windows have an idea of their last known current position.
 */
struct LifeObject
{
	LifeObject * nextObject;	/* next object */
	LifeAlgorithm * algorithm;	/* algorithm used for object */
	const LifeRule * rule;		/* current rule for evolution */

	//
	// Data to support the list algorithm.
	//
	LifeRow * firstRow;		/* first row of live cells */
	LifeRow * lastRow;		/* last row of live cells */
	QuadCell * usedQuadCells;	/* cells ever used (for tracks) */

	//
	// Data to support the quadcell algorithm.
	//
	QuadCell * quadCells;		/* current cells of the object */
	QuadCell * reducedQuadCells;	/* reduced states of the current cells */
	QuadCell * selectedQuadCells;	/* cells resulting from a selection */
	QuadCell * searchQuadCells;	/* cells used for searching */

	//
	// Other data.
	//
	LifeObject * overlayObject;	/* object which is overlayed on this one */
	BOOL	isOverlayActive;	/* whether overlay is active */
	BigCount changeCount;		/* count of changes to object */
	NUMBER	count;			/* number of live cells */
	NUMBER	minCount;		/* minimum number of life cells */
	NUMBER	maxCount;		/* maximum number of life cells */
	NUMBER	growth;			/* change in number of live cells */
	NUMBER	born;			/* number of cells born */
	NUMBER	died;			/* number of cells died */
	NUMBER	mouseRow;		/* row of last mouse activity */
	NUMBER	mouseCol;		/* column of last mouse activity */
	Count	frequency;		/* frequency of output for object */
	Count	freqLeft;		/* current count left before updating */
	Count	genLeft;		/* generations left before stopping */
	Count	gen;			/* current generation */
	Count	scaleListCount;		/* number of scales in scale list */
	Count	scaleListIndex;		/* current index into scale list */
	BOOL	isReducedValid;		/* whether the reduced cells are valid */
	BOOL	isReserved;		/* whether this is a reserved object */
	BOOL	isLocationsEnabled;	/* Whether showing locations is enabled */
	BOOL	isJumpingEnabled;	/* whether generations can jump ahead */
	BOOL	isCountUnknown;		/* whether the count of ON cells is unknown */
	BOOL	isTracksUsed;		/* track used cells for object */
	BOOL	isTracksVisible;	/* whether tracks are visible in view */
	BOOL	stopIsEnabled;		/* stop conditions are enabled */
	BOOL	stopOnPath;		/* stop if any cell is on path */
	BOOL	stopInsidePath;		/* stop if any cell is inside path */
	BOOL	stopOutsidePath;	/* stop if any cell is outside path */
	BOOL	stopCell;		/* stop if specified cell changes */
	BOOL	stopState;		/* state of cell to stop for */
	Count	stopGap;		/* gap size to stop on */
	NUMBER	stopRow;		/* row where to stop computing */
	NUMBER	stopCol;		/* column where to stop computing */
	NUMBER	dragMinRow;		/* dragged area start row */
	NUMBER	dragMinCol;		/* dragged area start column */
	NUMBER	dragMaxRow;		/* dragged area end row */
	NUMBER	dragMaxCol;		/* dragged area end column */

	BOOL	windowIsAutoScale;	/* saved autoscale flag of window view */
	Scale	windowScale;		/* saved scale factor of window view */
	NUMBER	windowCurRow;		/* saved current row of window view */
	NUMBER	windowCurCol;		/* saved current column of window view */

	StringBuffer name;		/* name of object */
	StringBuffer comments;		/* comments for object */

	Scale	scaleList[SCALE_LIST_SIZE];	/* list of scales for view */
	char	stopExpr[1024];		/* stop if expression is true */
	Path	path;			/* path defined within object */

	LocationInfo originInfo;	/* information about the origin */
	LocationInfo currentInfo;	/* information about current location */
	LocationInfo whereInfo[26];	/* information about where locations */
};


/*
 * Flags indicating which settings values for a window are
 * begin shared with other windows.  Settings which are not
 * shared are private to the window.
 */
typedef	unsigned long SettingsFlags;

#define	SETTINGS_SHIFT_FACTOR		((SettingsFlags) 0x000001)
#define	SETTINGS_DEFAULT_SCALE		((SettingsFlags) 0x000002)
#define	SETTINGS_CELL_SHAPE		((SettingsFlags) 0x000004)
#define	SETTINGS_IS_GRID_ENABLED	((SettingsFlags) 0x000008)
#define	SETTINGS_GRID_CHAR_1		((SettingsFlags) 0x000010)
#define	SETTINGS_GRID_CHAR_2		((SettingsFlags) 0x000020)
#define	SETTINGS_GRID_SIZE		((SettingsFlags) 0x000040)
#define	SETTINGS_BORDER_WIDTH		((SettingsFlags) 0x000080)
#define	SETTINGS_IS_AUTO_SCALE		((SettingsFlags) 0x000100)
#define	SETTINGS_MARGIN_PERCENTAGE	((SettingsFlags) 0x000200)
#define	SETTINGS_BACKGROUND_INDEX	((SettingsFlags) 0x000400)
#define	SETTINGS_VIEW_TYPE		((SettingsFlags) 0x000800)
#define	SETTINGS_POSITION_TYPE		((SettingsFlags) 0x001000)
#define	SETTINGS_IS_VISIBLE		((SettingsFlags) 0x002000)
#define	SETTINGS_IS_SHOW_OUTLINES	((SettingsFlags) 0x004000)
#define	SETTINGS_IS_SHOW_MY_OUTLINE	((SettingsFlags) 0x008000)
#define	SETTINGS_STATUS_LOCATION	((SettingsFlags) 0x010000)
#define	SETTINGS_CELL_COLOR_INDEX	((SettingsFlags) 0x020000)
#define	SETTINGS_TRACKS_COLOR_INDEX	((SettingsFlags) 0x040000)
#define	SETTINGS_MARK_COLOR_INDEX	((SettingsFlags) 0x080000)
#define	SETTINGS_OLD_COLOR_INDEX	((SettingsFlags) 0x100000)
#define	SETTINGS_TEXT_COLOR_INDEX	((SettingsFlags) 0x200000)
#define	SETTINGS_GRID_TYPE		((SettingsFlags) 0x400000)
#define	SETTINGS_IS_CELL_SHAPE_SMALL	((SettingsFlags) 0x800000)

/*
 * Special settings combination values for no
 * settings and all settings.
 */
#define	SETTINGS_MASK_NONE	((SettingsFlags) 0L)
#define	SETTINGS_MASK_ALL	((SettingsFlags) -1L)


/*
 * Settings which require an update of the status line
 * of windows in which these are changed.
 */
#define	SETTINGS_MASK_STATUS \
	(SETTINGS_IS_AUTO_SCALE | SETTINGS_VIEW_TYPE)

/*
 * Settings which require a redraw of windows in
 * which these are changed.
 */
#define	SETTINGS_MASK_REDRAW \
	(SETTINGS_CELL_SHAPE | SETTINGS_IS_GRID_ENABLED | \
	SETTINGS_GRID_CHAR_1 | SETTINGS_GRID_CHAR_2 | \
	SETTINGS_GRID_TYPE | SETTINGS_IS_CELL_SHAPE_SMALL | \
	SETTINGS_GRID_SIZE | SETTINGS_BACKGROUND_INDEX | \
	SETTINGS_VIEW_TYPE | SETTINGS_POSITION_TYPE | \
	SETTINGS_IS_SHOW_OUTLINES | SETTINGS_CELL_COLOR_INDEX | \
	SETTINGS_TRACKS_COLOR_INDEX | SETTINGS_MARK_COLOR_INDEX | \
	SETTINGS_OLD_COLOR_INDEX | SETTINGS_TEXT_COLOR_INDEX)

/*
 * Settings which require a reconfiguration of
 * windows in which these are changed.
 */
#define	SETTINGS_MASK_CONFIGURE	\
	(SETTINGS_BORDER_WIDTH | SETTINGS_STATUS_LOCATION)

/*
 * Settings which are grid-related.
 */
#define	SETTINGS_MASK_GRID \
	(SETTINGS_IS_GRID_ENABLED | SETTINGS_GRID_SIZE | \
	SETTINGS_GRID_CHAR_1 | SETTINGS_GRID_CHAR_2 | \
	SETTINGS_GRID_TYPE)

/*
 * Settings which are color-related.
 */
#define	SETTINGS_MASK_COLOR \
	(SETTINGS_BACKGROUND_INDEX | SETTINGS_CELL_COLOR_INDEX | \
	SETTINGS_TRACKS_COLOR_INDEX | SETTINGS_MARK_COLOR_INDEX | \
	SETTINGS_OLD_COLOR_INDEX | SETTINGS_TEXT_COLOR_INDEX)


/*
 * Display settings for a window.
 * These are settable per window, but there are also global
 * ones for initialising of new windows.
 * The higher level program code should only access these
 * values by calling the accessor functions so that the
 * proper private or shared instances of the values are used.
 */
typedef struct Settings Settings;

struct Settings
{
	char name[MAX_SETTINGS_NAME];		/* name of setting */
	SettingsFlags	changedFlags;		/* settings which have been changed */
	SettingsFlags	sharedFlags;		/* flags for which settings are shared */
	Settings *	sharedSettings;		/* settings for the shared settings values */
	Value		shiftFactor;		/* default shifting factor */
	Scale		defaultScale;		/* default scale value */
	CellShape	cellShape;		/* shape for drawing cells */
	GridType	gridType;		/* type of grid to draw */
	BOOL		isCellShapeSmall;	/* whether cell shapes are small */
	BOOL		isGridEnabled;		/* whether the grid is enabled */
	char		gridChar1;		/* first character to use for grid */
	char		gridChar2;		/* second character to use for grid */
	Count		gridSize;		/* size for showing larger grid */
	PixelCount	borderWidth;		/* width of border around edges */
	BOOL		isAutoScale;		/* doing autoscaling */
	BOOL		isVisible;		/* whether window is set as being visible */
	BOOL		isShowOutlines;		/* whether other window's outlines are shown */
	BOOL		isShowMyOutline;	/* whether other windows can show our outline */
	int		marginPercentage;	/* radius percentage margin near edges */
	ColorIndex	backgroundIndex;	/* background color index */
	ColorIndex	cellColorIndex;		/* normal cell color index */
	ColorIndex	tracksColorIndex;	/* tracks color index */
	ColorIndex	markColorIndex;		/* marked cells color index */
	ColorIndex	oldColorIndex;		/* old cells color index */
	ColorIndex	textColorIndex;		/* view text color index */
	ViewType	viewType;		/* type of view seen in window */
	PositionType	positionType;		/* type of positioning for view */
	StatusLocation	statusLocation;		/* location for showing status text */
};


/*
 * Structure defining one output window.
 * Such a window can display one life object at a time,
 * and has a combined status and command line at the top.
 * All windows are owned by the same global LifeDevice structure.
 * The windows know what row and column numbers are visible,
 * and have an idea of the position of interest within the object.
 * But the object always has the real cursor position.
 */
struct LifeWindow
{
	LifeDevice * device;		/* device window belongs to */
	LifeObject * currentObject;	/* currently displayed object */
	LifeObject * prevObject;	/* previously displayed object */
	LifeWindow * parentWindow;	/* our parent window or NULL */
	Count id;			/* window id number */
	PixelCount	width;		/* total window width in device coordinates */
	PixelCount	height;		/* total window height in device coordinates */
	PixelCoord	positionX;	/* X position of subwindow in device coordinates */
	PixelCoord	positionY;	/* Y position of subwindow in device coordinates */
	PixelCount	rows;		/* number of rows of individual cells viewable */
	PixelCount	cols;		/* number of columns of individual cells viewable */
	NUMBER	minRow;			/* minimum object row seen in window */
	NUMBER	maxRow;			/* maximum object row seen in window */
	NUMBER	minCol;			/* minimum object column seen in window */
	NUMBER	maxCol;			/* maximum object column seen in window */
	NUMBER	curRow;			/* object row of center of view */
	NUMBER	curCol;			/* object column of center of view */
	NUMBER	viewRows;		/* how many object rows can be seen at scale */
	NUMBER	viewCols;		/* how many object columns can be seen at scale */
	NUMBER	seeCount;		/* number of cells visible in view */
	Count	textRows;		/* current rows of text available */
	Count	textCols;		/* current columns of text available */
	Scale	scale;			/* scaling factor for view */
	BigCount lastChangeCount;	/* last object change count known */
	Flags	updateFlags;		/* update flags for displaying */
	BOOL	isMapped;		/* whether window is currently mapped (actually visible) */
	BOOL	isShifted;		/* whether the view is shifted from desired position */
	BOOL	isTextView;		/* whether text is being shown in view window */
	StatusMode	statusMode;	/* current usage of status sub-window */
	int	statusUsed;		/* number of characters in status buffer */
	char	statusBuffer[256];	/* buffer holding status text */
	Settings	settings;	/* settings for this window */
	int	mOutlineCount;		/* number of outlines in the table */
	Area	mOutlineTable[MAX_WINDOWS];	/* areas for window outlines to be drawn */
	void *	privateData;		/* private data for window */
};


/*
 * Structure to interface to various input/output devices.
 * This supports both normal terminals and graphics output.
 * A device can manage multiple windows at the same time.
 */
struct LifeDevice
{
	BOOL	(*openDevice)(LifeDevice *);	/* open device */
	void	(*closeDevice)(LifeDevice *);	/* close device */
	LifeWindow * (*openWindow)(LifeDevice *, LifeWindow * parentWindow);	/* open new window */
	BOOL	(*closeWindow)(LifeDevice *);	/* close current window */
	void	(*setIsVisible)(LifeDevice *, LifeWindow * window, BOOL);	/* set sub-windows visible or not */
	void	(*setStatusMode)(LifeDevice *, StatusMode);	/* set mode for status sub-window */
	void	(*addStatusText)(LifeDevice *, const char *);	/* add text to status sub-window */
	void	(*clearStatusMessage)(LifeDevice *);		/* remove any message in the status sub-window */
	void	(*setViewModeIsText)(LifeDevice *, BOOL);	/* set view window to text mode */
	void	(*addViewText)(LifeDevice *, const char *);	/* add text to view sub-window */
	void	(*update)(LifeDevice *);		/* update all windows in the display */
	void	(*redraw)(LifeDevice *);		/* redraw all windows in the display */
	void	(*delay)(LifeDevice *, int milliSeconds);		/* delay for a while or until event */
	BOOL	(*isInputReady)(LifeDevice *);	/* check whether input is ready */
	int		(*readChar)(LifeDevice *, BOOL);	/* read character with optional wait */
	BOOL	(*assign)(LifeDevice *, Value, int);	/* assign button to macro */
	void	(*writeAssign)(LifeDevice *, FILE *);	/* write assignments */	
	void	(*beep)(LifeDevice *);			/* make a beep sound for errors */
	void	(*defaultSettings)(LifeDevice *, BOOL isTopLevelWindow, Settings * settings);
	ColorIndex	(*defineColor)(LifeDevice *, const char * name);	/* define a color to be used */

	LifeWindow * windowTable[MAX_WINDOWS];	/* table of opened windows */
	Count	windowCount;			/* number of opened windows */
	Scale	minScale;			/* minimum legal scale value */
	Scale	maxScale;			/* maximum legal scale value */
};


/*
 * Structure to hold a list of strings allocated from a common buffer.
 * This is used for collecting file names for listing.
 */
typedef	struct
{
	char **	argv;		/* pointer to list of pointers */
	char *	buf;		/* pointer to string storage */
	int	argc;		/* number of pointers */
	int	maxArgc;	/* maximum number of pointers */
	int	used;		/* chars used in string storage */
	int	maxUsed;	/* maximum chars in string storage */
} List;


/*
 * A structure containing a jump buffer.
 * This is used because a jmp_buf is defined as an array,
 * and there are problems passing around its address.
 * Passing around the address of the structure works fine.
 */
typedef	struct
{
	jmp_buf	buf;
} JumpBuffer;


/*
 * Extern definitions for non-initialized data.
 * These are defined externally in all modules, except in main.
 */
#ifdef	DEFINE_GLOBALS
#define	EXTERN
#else
#define	EXTERN	extern
#endif

//#include "algorithm.h"
//#include "listalg.h"
//#include "quadalg.h"


/*
 * Structure containing the globals for the whole Life program.
 */
typedef	struct LifeGlobals
{
	LifeAlgorithm * defaultAlgorithm; /* default algorithm used for cells */
	LifeDevice * device;		/* device being used */
	LifeWindow * currentWindow;	/* current window */
	LifeObject * currentObject;	/* current object of current window */
	LifeObject * objects;		/* list of active objects */
	LifeObject * mainObject;	/* the main object */
	LifeObject * deleteObject;	/* object last deleted */
	LifeObject * readObject;	/* object just read */
	LifeObject * tempObject;	/* temporary object */
	LifeObject * freeObjects;	/* list of free objects */
	LifeObject * newObjects;	/* top of new object allocation */
	LifeObject * endObjects;	/* end of new objects */
	LifeObject * backupObjects[MAX_BACKUPS];	/* backup objects */

	const LifeRule * defaultRule;	/* the default Life rule */

	LifeRow *	freeRows;	/* list of free row structures */
	LifeRow *	newRows;	/* top of new row allocation */
	LifeRow *	endRows;	/* end of new rows */
	LifeRow *	termRow;	/* terminus row */
	LifeRow		initRow;	/* row to initialize list */

	LifeCell *	freeCells;	/* list of free cell structures */
	LifeCell *	newCells;	/* top of new cell allocation */
	LifeCell *	endCells;	/* end of new cells */
	LifeCell *	termCell;	/* terminus cell */
	LifeCell	initCell;	/* cell to initialize list */

	int		colorCount;	/* number of colors in the table */
	const char *	colorTable[MAX_COLORS];		/* table of color names */
	Scale	initScale;		/* default scale factor set on command line */
	Settings * topLevelWindowSettings;	/* default settings for top level */
	Settings * subWindowSettings;	/* default sub-window settings */
	Settings	settingsTable[MAX_SETTINGS];	/* table of settings */
	int		settingsCount;		/* number of settings in table */
	BOOL	isCreatingReserved;	/* creating reserved object names */
	BOOL	doWait;			/* must wait for input */
	BOOL	canMakeInput;		/* allowed to generate input chars */
	BOOL	stop;			/* user wants to stop current action */
	BOOL	intJmpOk;		/* ok to use interrupt jump buffer */
	Count	defaultFrequency;	/* default frequency for new objects */
	BOOL	defaultIsJumpingEnabled; /* default jumping for new objects */
	BOOL	defaultTracks;		/* default tracks flag */
	int	debugOption;		/* debugging option value */
	int		topWindowSizePercentage;	/* size for new top level window */
	int		subWindowSizePercentage;	/* size for new sub-window */
	SettingsFlags	defaultSharedSettings;	/* default shared settings */
	int	backedUpIndex;		/* index of last backed up object */
	int	undoDepth;		/* current undo depth */
	MoveMode	moveMode;	/* mode of movement */
	int	delayTime;		/* milliseconds of delay between updates */
	const char *	fontName;	/* font name for graphics */
	const char *	userLibs[MAX_LIBS];	/* user's life libraries if any */
	int	userLibCount;		/* number of user libraries specified */
	JumpBuffer	ttyJmp;		/* jump buffer for terminal input */
	JumpBuffer	intJmp;		/* jump buffer for interrupts */

	/*
	 * Editing characters
	 */
	char	vKill;		/* line kill character */
	char	vErase1;	/* first erase character */
	char	vErase2;	/* second erase character */
	char	vWordErase;	/* word erase character */
	char	vEol;		/* end of line character */
	char	vEof;		/* EOF character */
	char	vLiteralNext;	/* literal next character */
} LifeGlobals;


/*
 * The globals for the Life program.
 */
EXTERN	LifeGlobals	Life;


/*
 * Global procedures.
 * These routines are generic and don't depend on the particular
 * LifeAlgorithm implementation which is active.
 */
extern	LifeObject *	AllocateObject(void);
extern	LifeObject *	FindObject(const char *);
extern	LifeObject *	GetObject(const char *);
extern	LifeObject *	GetUniqueObject(const char * str);
extern	LifeRule *	AllocateRule(void);
extern	void		CheckStartupFiles(void);
extern	const char *	ReadString(const char *);
extern	BOOL		Confirm(const char *, BOOL);
extern	char *		ReadPartialString(const char *, char *, int);
extern	void		ExpandFileName(char *, int);
extern	void		ExpandBufferName(char *, int);
extern	void		ListVariables(void);
extern	Value		GetSmallExpression(const char * str);
extern	NUMBER		GetExpression(const char *);
extern	NUMBER		GetVariable(const char *);
extern	NUMBER		GetVariable1(int);
extern	void		SetVariable(const char *, NUMBER);
extern	void		SetVariable1(int, NUMBER);
extern	void		ZeroObject(LifeObject *);
extern	void		DestroyObject(LifeObject *);
extern	void		MoveObject(LifeObject *, LifeObject *);
extern	void		CenterObject(LifeObject *, NUMBER, NUMBER);
extern	void		ShiftObject(LifeObject *, NUMBER, NUMBER);
extern	void 		InitLocationInfo(LocationInfo * info, int id);
extern	void 		InitAllLocationInfo(LifeObject * obj);
extern	void		ClearLocationInfo(LocationInfo * info);
extern	void 		ClearAllLocationInfo(LifeObject * obj);

extern	void		CopyMarkedObject(const LifeObject * srcObj,
	LifeObject * destObj);

extern	void		MoveSelection(LifeObject * srcObj, LifeObject * destObj);

extern	void		AddObject(const LifeObject *, LifeObject *, AddObjectMode);
extern	void		CopyObject(const LifeObject *, LifeObject *);
extern	void		BackupObject(const LifeObject *, const char *);
extern	void		ListObjects(BOOL);
extern	void		ShowObjectInfo(const LifeObject *);
extern	void		ShowQuadCellStatistics(void);

extern	BOOL		AddCell(LifeObject * obj, NUMBER row, NUMBER col);
extern	BOOL		DeleteCell(LifeObject * obj, NUMBER row, NUMBER col);
extern	BOOL		IsCellSet(const LifeObject * obj, NUMBER row, NUMBER col);

extern	int		GetCellState(const LifeObject * obj,
	NUMBER row, NUMBER col);

extern	void		SetCellState(LifeObject * obj,
	NUMBER row, NUMBER col, int state);

extern	BOOL		IsObjectEmpty(const LifeObject * obj);
extern	NUMBER		CountMarks(const LifeObject * obj);

extern	BOOL		MinMax(const LifeObject * obj,
	NUMBER * retMinRow, NUMBER * retMaxRow,
	NUMBER * retMinCol, NUMBER * retMaxCol);

extern	NUMBER		CountRegion(const LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

extern	BOOL		SelectRegion(LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

extern	BOOL		SelectObject(LifeObject * obj,
	NUMBER row, NUMBER col, Value gap);

extern	BOOL		FindClosestSelectedCell(const LifeObject * obj,
	NUMBER baseRow, NUMBER baseCol, NUMBER * retRow, NUMBER * retCol);

extern	void		RotateSelection(LifeObject * obj);
extern	void		FlipRowSelection(LifeObject * obj);
extern	void		FlipColumnSelection(LifeObject * obj);
extern	BOOL		MarkSelection(LifeObject * obj);
extern	void		ClearSelection(LifeObject * obj);
extern	void		ClearSearch(LifeObject * obj);
extern	void		ClearTracks(LifeObject * obj);
extern	void		ClearMarks(LifeObject * obj);
extern	void		InvertMarks(LifeObject * obj);
extern	BOOL		SelectAll(LifeObject * obj);
extern	BOOL		SelectPath(LifeObject * obj);

extern	BOOL		SelectMarkedCells(LifeObject * obj);
extern	BOOL		SelectCell(LifeObject * obj, NUMBER row, NUMBER col);

extern	void		ClearTracks(LifeObject * obj);

extern	BOOL		MarkMinMax(const LifeObject * obj,
	NUMBER * minRow, NUMBER * maxRow, NUMBER * minCol, NUMBER * maxCol);

extern	void		SaveUsedCells(LifeObject * obj);

extern	ViewInfo	CalculateView(
	const LifeObject * obj, int viewId, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

extern	BOOL		ScanObject(LifeObject * obj, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data);

extern	NUMBER		GetObjectMinRowDowntoLimit(LifeObject * obj,
	NUMBER limitRow);

extern	QuadCell *	ExtractQuadCell(LifeObject * obj);

extern	Count		RunGenerations(LifeObject * obj, Count genCount);

extern	void		CountPathCells(const LifeObject * obj,
	NUMBER * insideCount, NUMBER * boundaryCount, NUMBER * outsideCount);

extern	NUMBER		FindMarkedGap(const LifeObject * obj);

extern	void		IncrementChangeCount(LifeObject * obj);
extern	void		AdjustCellCount(LifeObject * obj, NUMBER adjust);
extern	void		SetCellCount(LifeObject * obj, NUMBER count);
extern	NUMBER		CalculateCellCount(LifeObject * obj);
extern	BOOL		SearchObject(LifeObject *, Count, BOOL);
extern	BOOL		MacroIsValid(int);
extern	void		Error(const char *);
extern	void		Message(const char *);
extern	const char *	GenerateStatus(const LifeWindow * window);

extern	const char *	FindLibraryFile(const char * name);
extern	const char *	FindFile(const char * name);

extern	int		ReadChar(void);
extern	BOOL		IsAbbreviation(const char *, const char *);
extern	int		ReadLine(const char *, char *, int, int, BOOL);
extern	BOOL		ShowHelp(const char *);
extern	BOOL		FormatHelp(const char * fmt, ...);
extern	void		EndHelp(BOOL isWaitRequired);
extern	void		LoadFiles(const char ** fileList, int fileCount);
extern	void		LoadFile(const char * fileName);
extern	BOOL		TryReadFile(FILE *, LifeObject *);
extern	void		ReadRle(FILE *, LifeObject *);
extern	void		ReadXlife(FILE *, LifeObject *);
extern	void		ReadPicture(FILE *, LifeObject *);
extern	void		ReadMC(FILE * fp, LifeObject * obj);

extern	void		WriteObject(LifeObject *, const char *,
				const char *, Value, Value);

extern	void		WriteMacros(FILE *);
extern	BOOL		DoGenerations(LifeObject * obj);
extern	BOOL		DoGenerationsForAll(void);
extern	BOOL		IsRunningGenerations(void);

extern	BOOL		IsNewCellAlive(BOOL cellA, BOOL cellB,
	BOOL cellC, BOOL cellD, BOOL cellE, BOOL cellF, BOOL cellG,
	BOOL cellH, BOOL cellI);

extern	void		DoLineCommand(NUMBER, NUMBER, BOOL, BOOL);
extern	void		DoCommands(void);
extern	BOOL		DoOneCommand(void);
extern	void		Backup(void);
extern	void		CheckRun(void);
extern	void		GetFiles(List *, const char *, const char *,
				int, int);
extern	void		GetCompletionFiles(List *, const char *,
				const char *);
extern	void		DisplayList(const List *, BOOL);
extern	void		InitList(List *);
extern	void		AddListItem(List *, const char *);
extern	void		SortList(List *);
extern	int		GetListPrefixLength(const List *);
extern	void		FreeList(List *);
extern	void		GarbageCollect(void);
extern	void		CheckGarbageCollect(void);

extern	void		CheckSameAlgorithm(const LifeObject * srcObj,
	const LifeObject * destObj);

extern	void		PrepareTempObject(const LifeObject * srcObj);


//
// String buffer routines.
//
extern	void		InitStringBuffer(StringBuffer * buf);
extern	void		TermStringBuffer(StringBuffer * buf);
extern	void		SetStringBuffer(StringBuffer * buf, const char * str);
extern	void		ClearStringBuffer(StringBuffer * buf);
extern	void		AddCharToStringBuffer(StringBuffer * buf, int ch);
extern	void		AppendToStringBuffer(StringBuffer * buf, const char * str);
extern	void		SetStringBufferLength(StringBuffer * buf, int len);
extern	const char *	GetStringBufferText(const StringBuffer * buf);
extern	int		GetStringBufferLength(const StringBuffer * buf);


//
// Location setting and checking routines.
//
extern	void		CheckAllLocationStatus(const LifeObject * obj);
extern	void		CheckLocation(const LifeObject *, const LocationInfo *);
extern	void		SetLocation(LifeObject *, LocationInfo *, NUMBER, NUMBER);
extern	void		AdjustLocation(LifeObject *, LocationInfo *, NUMBER, NUMBER);
extern	void		AdjustCurrentLocation(LifeObject *, NUMBER, NUMBER);
extern	void		SetLocationRow(LifeObject *, LocationInfo *, NUMBER);
extern	void		SetLocationCol(LifeObject *, LocationInfo *, NUMBER);
extern	void		CopyLocation(LifeObject *, const LocationInfo *, LocationInfo *);
extern	void		CheckCurrentLocation(const LifeObject *);
extern	void		SetCurrentLocation(LifeObject *, NUMBER, NUMBER);
extern	void		SetCurrentRow(LifeObject *, NUMBER);
extern	void		SetCurrentCol(LifeObject *, NUMBER);


//
// Algorithm setting routines.
//
extern	BOOL	SetObjectAlgorithm(LifeObject * obj, LifeAlgorithm * algorithm);
extern	const char *	GetAlgorithmName(LifeAlgorithm * algorithm);
extern	LifeAlgorithm *	GetAlgorithmByName(const char * name);


/*
 * Settings accessor functions.
 */
extern	Value		GetShiftFactor(const Settings * settings);
extern	Scale		GetDefaultScale(const Settings * settings);
extern	GridType	GetGridType(const Settings * settings);
extern	CellShape	GetCellShape(const Settings * settings);
extern	BOOL		GetIsCellShapeSmall(const Settings * settings);
extern	BOOL		GetIsAutoScale(const Settings * settings);
extern	BOOL		GetIsVisible(const Settings * settings);
extern	BOOL		GetIsShowOutlines(const Settings * settings);
extern	BOOL		GetIsShowMyOutline(const Settings * settings);
extern	BOOL		GetIsGridEnabled(const Settings * settings);
extern	int		GetGridChar1(const Settings * settings);
extern	int		GetGridChar2(const Settings * settings);
extern	Count		GetGridSize(const Settings * settings);
extern	PixelCount	GetBorderWidth(const Settings * settings);
extern	int		GetMarginPercentage(const Settings * settings);
extern	ColorIndex	GetBackgroundIndex(const Settings * settings);
extern	ColorIndex	GetCellColorIndex(const Settings * settings);
extern	ColorIndex	GetTracksColorIndex(const Settings * settings);
extern	ColorIndex	GetMarkColorIndex(const Settings * settings);
extern	ColorIndex	GetOldColorIndex(const Settings * settings);
extern	ColorIndex	GetTextColorIndex(const Settings * settings);
extern	ViewType	GetViewType(const Settings * settings);
extern	PositionType	GetPositionType(const Settings * settings);
extern	StatusLocation	GetStatusLocation(const Settings * settings);

extern	void		SetShiftFactor(Settings * settings, Value);
extern	void		SetDefaultScale(Settings * settings, Scale);
extern	void		SetCellShape(Settings * settings, CellShape);
extern	void		SetGridType(Settings * settings, GridType);
extern	void		SetIsCellShapeSmall(Settings * settings, BOOL);
extern	void		SetIsGridEnabled(Settings * settings, BOOL);
extern	void		SetIsAutoScale(Settings * settings, BOOL);
extern	void		SetIsVisible(Settings * settings, BOOL);
extern	void		SetIsShowOutlines(Settings * settings, BOOL);
extern	void		SetIsShowMyOutline(Settings * settings, BOOL);
extern	void		SetGridChar1(Settings * settings, int);
extern	void		SetGridChar2(Settings * settings, int);
extern	void		SetGridSize(Settings * settings, Count);
extern	void		SetBorderWidth(Settings * settings, PixelCount);
extern	void		SetMarginPercentage(Settings * settings, int);
extern	void		SetBackgroundIndex(Settings * settings, ColorIndex);
extern	void		SetCellColorIndex(Settings * settings, ColorIndex);
extern	void		SetTracksColorIndex(Settings * settings, ColorIndex);
extern	void		SetMarkColorIndex(Settings * settings, ColorIndex);
extern	void		SetOldColorIndex(Settings * settings, ColorIndex);
extern	void		SetTextColorIndex(Settings * settings, ColorIndex);
extern	void		SetViewType(Settings * settings, ViewType);
extern	void		SetPositionType(Settings * settings, PositionType);
extern	void		SetStatusLocation(Settings * settings, StatusLocation);

extern	SettingsFlags	ParseSettingsFlags(int count, char ** nameTable);
extern	void		ClearSettingsChangedFlags(void);


/*
 * Window related functions.
 */
extern	void		ApplyWindowSettingsChangedFlags(LifeWindow * window);
extern	void		UpdateAllWindows(void);
extern	void		ShowWindowSettings(const LifeWindow * window);
extern	SettingsFlags	CompareSettings(const Settings * old, const Settings * new);
extern	void		AddSharedWindowSettings(LifeWindow * window, SettingsFlags);
extern	void		SubtractSharedWindowSettings(LifeWindow * window, SettingsFlags);
extern	BOOL		LoadWindowSettings(LifeWindow * window, const char * name);
extern	BOOL		MergeWindowSettings(LifeWindow * window, const char * name);
extern	BOOL		StoreWindowSettings(const LifeWindow * window, const char * name);
extern	BOOL		UseWindowSettings(LifeWindow * window, const char * name);
extern	void		DefaultWindowSettings(LifeWindow * window);
extern	void		InitializeWindowSettings(LifeWindow * window);
extern	LifeWindow *	FindOtherTopLevelWindow(const LifeWindow * window);
extern	BOOL		IsOnlyTopLevelWindow(const LifeWindow * window);
extern	Settings *	AllocSettings(const char * name, BOOL isTopLevelWindow);
extern	Settings *	FindSettings(const char * name);
extern	void		CopySettings(const Settings * old, Settings * new);
extern	void		CopySomeSettings(const Settings *, Settings *, SettingsFlags);
extern	void		CopySettings(const Settings * old, Settings * new);
extern	void		RawCopySettings(const Settings * old, Settings * new);
extern	BOOL		IsWindowUpdateNeeded(void);
extern	void		SetWindowStatusLocation(LifeWindow * window, StatusLocation location);
extern	void		SetWindowIsAutoScale(LifeWindow * window, BOOL isAutoScale);
extern	void		SetWindowIsGridEnabled(LifeWindow * window, BOOL isGridEnabled);
extern	void		SetWindowGridType(LifeWindow * window, GridType gridType);
extern	void		SetWindowGridSize(LifeWindow * window, Count gridSize);
extern	void		SetWindowGridChar1(LifeWindow * window, int gridChar);
extern	void		SetWindowGridChar2(LifeWindow * window, int gridChar);
extern	void		SetWindowCellShape(LifeWindow * window, CellShape cellShape);
extern	void		SetWindowIsCellShapeSmall(LifeWindow * window, BOOL);
extern	BOOL		SetWindowBackgroundColor(LifeWindow * window, const char * colorName);
extern	BOOL		SetWindowCellColor(LifeWindow * window, const char * colorName);
extern	BOOL		SetWindowTracksColor(LifeWindow * window, const char * colorName);
extern	BOOL		SetWindowMarkColor(LifeWindow * window, const char * colorName);
extern	BOOL		SetWindowOldColor(LifeWindow * window, const char * colorName);
extern	BOOL		SetWindowTextColor(LifeWindow * window, const char * colorName);
extern	void		SetWindowBorderWidth(LifeWindow * window, PixelCount borderWidth);
extern	void		AddWindowObjectUpdateFlags(const LifeObject * obj, Flags flags);
extern	void		WindowClearObject(const LifeObject * object);
extern	int		GetObjectWindowCount(const LifeObject * obj);
extern	void		AddWindowUpdateFlags(LifeWindow * window, Flags);
extern	void		AddOtherWindowOutlineFlags(const LifeWindow * window);
extern	LifeObject *	GetCurrentObject(const LifeWindow * window);
extern	LifeObject *	GetPreviousObject(const LifeWindow * window);
extern	void		SetWindowObject(LifeWindow * window, LifeObject * object);
extern	void		GetVisibleSize(const LifeWindow *, NUMBER * rows, NUMBER * cols);
extern	Scale		GetScale(const LifeWindow * window);
extern	void		GetViewTypePosition(const LifeWindow *, NUMBER * row, NUMBER * col);
extern	void		SetScale(LifeWindow *, Scale);
extern	void		CenterView(LifeWindow * window);
extern	void		ShiftView(LifeWindow * window, NUMBER rowShift, NUMBER colShift);
extern	Scale		AutoScale(LifeWindow *);
extern	void		SetWindowViewType(LifeWindow *, ViewType viewType);
extern	void		SetWindowPositionType(LifeWindow *, PositionType positionType);
extern	void		SetWindowIsShowOutlines(LifeWindow * window, BOOL isShowOutlines);
extern	void		SetWindowIsShowMyOutline(LifeWindow * window, BOOL isShowOutline);
extern	void		PositionView(LifeWindow *);
extern	void		PositionCell(LifeWindow *, NUMBER row, NUMBER col);
extern	void		PositionArea(LifeWindow *, NUMBER, NUMBER, NUMBER, NUMBER);
extern	void		BuildOutlineTable(LifeWindow * window);
extern	PixelCoord	ScreenCellToPixel(const LifeWindow *, NUMBER coord);
extern	NUMBER		PixelToScreenCell(const LifeWindow *, PixelCoord pixel);

extern	BOOL		IsOverlappingSubwindow(const LifeWindow * window,
				PixelCoord x, PixelCoord y,
				PixelCount width, PixelCount height);

extern	void		GetVisibleRegion(const LifeWindow *,
				NUMBER * minRow, NUMBER * maxRow,
				NUMBER * minCol, NUMBER * maxCol);

extern	ColorIndex	FindColorIndex(const char * colorName);
extern	ColorIndex	DefineColor(const char * colorName);
extern	ColorIndex	AddColorForDevice(const char * colorName);
extern	const char *	GetColorName(ColorIndex index);

extern	void		SaveCurrentViewParameters(void);
extern	void		RestoreCurrentViewParameters(void);


/*
 * Other routines.
 */
extern	const LifeRule * ParseRule(const char *);
extern	void		ViewStatus(LifeWindow *);
extern	int		GridMod(const LifeWindow *, NUMBER);
extern	void		ScanInit(int (*)(void), JumpBuffer *);
extern	int		ScanChar(void);
extern	void		ScanAbort(void);
extern	void		ScanEof(void);
extern	void		ScanReset(void);
extern	void		Beep(void);

extern	void		SetTopInputLevel(void);
extern	void		EndAllInputLevels(void);
extern	void		EndInputLevel(void);
extern	BOOL		IsTopInputLevel(void);
extern	BOOL		IsTerminalInput(void);
extern	BOOL		SetTerminal(void);
extern	BOOL		SetFile(const char *);
extern	BOOL		SetLoop(NUMBER, NUMBER, int);
extern	void		EndLoop(void);
extern	BOOL		SetMacro(NUMBER, NUMBER, int);
extern	const char *	GetInputDescription(void);
extern	FILE *		GetInputFILE(void);
extern	NUMBER		GetLoopValue(int);

//
// Path routines.
//
extern	void		InitializePath(LifeObject *);
extern	void		FreePath(LifeObject *);
extern	void		CopyPath(const LifeObject *, LifeObject *);
extern	void		ClearPath(LifeObject *);
extern	BOOL		RestorePath(LifeObject *);
extern	BOOL		AddPointToPath(LifeObject *, NUMBER, NUMBER);
extern	void		ClosePath(LifeObject *);
extern	void		ShortenPath(LifeObject *, Count);
extern	void		ShiftPath(LifeObject *, NUMBER, NUMBER);
extern	BOOL		PointIsOnPath(const Path *, NUMBER, NUMBER);
extern	int		PointPathSignum(const Path *, NUMBER, NUMBER);

extern	LifeDevice *	GetTtyDevice(void);

#if	defined(X11)
extern	LifeDevice *	GetX11Device(void);
#endif

#endif

/* END CODE */
