/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines which implement a particular algorithm for manipulating
 * Life objects.
 */

#include "life.h"
#include "algorithm.h"
#include "listalg.h"


/*
 * Functions to implement the life algorithm for list-based objects.
 */
LifeAlgorithm listAlgorithm =
{
	lcAllocCellData,
	lcFreeCellData,
	lcGetCellState,
	lcSetCellState,
	lcIsCellSet,
	lcAddCell,
	lcDeleteCell,
	lcDeleteAllCells,
	lcShiftObject,
	lcSaveUsedCells,
	lcRegisterQuadCells,
	lcCopyObject,
	lcMoveObject,
	lcAddObject,
	lcMinMax,
	lcCountRegion,
	lcCalculateView,
	lcCalculateCellCount,
	lcIsObjectEmpty,
	lcSearchObject,
	lcNextGeneration,
	lcRunGenerations,
	lcSelectObject,
	lcSelectAll,
	lcMarkSelection,
	lcSelectMarkedCells,
	lcInvertMarks,
	lcClearMarks,
	lcClearSearch,
	lcClearTracks,
	lcClearSelection,
	lcSelectPath,
	lcCountPathCells,
	lcFindMarkedGap,
	lcSelectCell,
	lcFindClosestSelectedCell,
	lcSelectRegion,
	lcMarkMinMax,
	lcCountMarks,
	lcMoveSelection,
	lcCopyMarkedObject,
	lcRotateSelection,
	lcFlipColumnSelection,
	lcFlipRowSelection,
	lcScanObject,
	lcGetObjectMaxRowUptoLimit,
	lcGetObjectMinRowDowntoLimit,
	lcExtractQuadCell
};


/*
 * Return whether or not the object has no ON cells.
 * This makes no use of the object's cell count value.
 */
BOOL
lcIsObjectEmpty(const LifeObject * obj)
{
	return (obj->firstRow == Life.termRow);
}


/*
 * Scan the current cells of an object, calling the indicated function
 * for each cell whose state matches one of the specified ones,
 * supplying a specified piece of data that can be used for whatever
 * purpose the called routine wants.  If the routine returns TRUE then
 * the scan is continued, otherwise it is stopped.  This routine returns
 * TRUE if the scan was completed, or FALSE if the called function
 * indicates that the scan is to be aborted.  The object cannot be changed
 * during the scan, neither can a garbage collection be performed.
 */
BOOL
lcScanObject(LifeObject * obj, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data)
{
	const LifeRow * rp;

//TODO: Use states value
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		const LifeCell * cp;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if (!func(rp->row, cp->col, STATE_ON, data))
				return FALSE;
		}
	}

	return TRUE;
}


/*
 * Return the highest row number in the object which is less than
 * or equal to the specified row.  If no such row exists then the
 * minimum negative number value is returned.
 */
NUMBER
lcGetObjectMaxRowUptoLimit(LifeObject * obj, NUMBER limitRow)
{
	const LifeRow * rp;
	NUMBER result = numMinValue();

	if (obj->firstRow == Life.termRow)
		return numMinValue();

	if (numIsLess(limitRow, obj->firstRow->row))
		return numMinValue();

	if (numIsGreaterOrEqual(limitRow, obj->lastRow->row))
		return obj->lastRow->row;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		if (numIsGreater(rp->row, limitRow))
			break;

		result = rp->row;
	}

	return result;
}


/*
 * Return the lowest row number in the object which is greater than
 * or equal to the specified row.  If no such row exists then the
 * maximum positive number value is returned.
 */
NUMBER
lcGetObjectMinRowDowntoLimit(LifeObject * obj, NUMBER limitRow)
{
	const LifeRow * rp;

	if (obj->firstRow == Life.termRow)
		return numMaxValue();

	if (numIsLessOrEqual(limitRow, obj->firstRow->row))
		return obj->firstRow->row;

	if (numIsGreater(limitRow, obj->lastRow->row))
		return numMaxValue();

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		if (numIsGreaterOrEqual(rp->row, limitRow))
			return rp->row;
	}

	//
	// We really shouldn't get here.
	//
	return numMaxValue();
}


/*
 * Register the quad cells for an object as being used.
 * This is in preparation of a garbage collection.
 * Note: We do this even if the list-based algorithm is in use
 * since quadcells are always used for the tracked cells.
 */
void
lcRegisterQuadCells(LifeObject * obj)
{
	qcRegister(obj->usedQuadCells);
}


/*
 * Find the minimum and maximum row and column numbers for an object.
 * Returns FALSE if the object has no cells.
 */
BOOL
lcMinMax(const LifeObject * obj,
	NUMBER * retMinRow, NUMBER * retMaxRow,
	NUMBER * retMinCol, NUMBER * retMaxCol)
{
	NUMBER minRow = numMaxValue();
	NUMBER maxRow = numMinValue();
	NUMBER minCol = numMaxValue();
	NUMBER maxCol = numMinValue();
	BOOL isFound = FALSE;
	const LifeRow * rp;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		if (rp->firstCell == Life.termCell)
			continue;

		minRow = numMin(minRow, rp->row);
		maxRow = rp->row;

		minCol = numMin(minCol, rp->firstCell->col);
		maxCol = numMax(maxCol, rp->lastCell->col);

		isFound = TRUE;
	}

	/*
	 * Fail if there are no cells in the object.
	 */
	if (!isFound)
		return FALSE;

	*retMinRow = minRow;
	*retMaxRow = maxRow;
	*retMinCol = minCol;
	*retMaxCol = maxCol;

	return TRUE;
}


/*
 * Calculate a view of the specified rectangular region of the specified object
 * and save the result in a returned memory array whose size is the indicated
 * number of rows times the indicated number of columns, scaled down by the
 * specified factor.  The cells in the array contain values for square areas
 * of the object having sides of 'scale', and are 0 for all cells OFF and 1
 * for one or more ON cells.  This also uses the auxillary data array to
 * store whether or not any cells are marked.  There is a limit on the size
 * of the rectangular region that can be used.  The returned array should not
 * be freed by the caller and is valid until the next call to this routine.
 */
ViewInfo
lcCalculateView(const LifeObject * obj, int viewId, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	const LifeRow *	rp;
	ViewInfo info;

	//
	// Initialise and clear the view structure including its
	// auxillary data.  We can then set cells nonzero as needed.
	//
	InitView(&info, viewId, TRUE, scale,
		minRow, maxRow, minCol, maxCol);

	ClearViewData(&info);
	FillAuxViewData(&info, 0);

	/*
	 * Now scan the object setting the appropriate view info cell
	 * nonzero for every ON cell in the region.
	 */
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = rp->row;
		const LifeCell * cp;
		int rowIndex;
		int state;

		if (numIsLess(row, minRow))
			continue;

		if (numIsGreater(row, maxRow))
			break;

		rowIndex = numToLong(numSub(row, minRow)) / scale;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER col = cp->col;
			int colIndex;
			int auxData;

			if (numIsLess(col, minCol))
				continue;

			if (numIsGreater(col, maxCol))
				break;

			//
			// This cell is on.
			// Count it and set the cell as being interesting.
			//
			numAddrInc(&info.onCount);

			colIndex = numToLong(numSub(col, minCol)) / scale;

			SetViewCell(&info, rowIndex, colIndex, TRUE);

			//
			// Set the auxillary data.
			//
			auxData = GetAuxViewData(&info, rowIndex, colIndex);

			state = ((cp->marks & MARK_USER) ? STATE_MARK : STATE_ON);

			if (scale == 1)
				auxData = state;
			else
				auxData |= state;

			SetAuxViewData(&info, rowIndex, colIndex, auxData);
		}
	}

	return info;
}


/*
 * Search forwards for the nth next object, restarting at the top if necessary.
 * This uses the MARK_SEARCH marks to remember which objects have been seen.
 * Returns FALSE if the search was aborted.
 */
BOOL
lcSearchObject(LifeObject * obj, Count count)
{
	const LifeRow * rp;
	const LifeCell * cp;

	NUMBER row = obj->currentInfo.row;
	NUMBER col = obj->currentInfo.col;

	for (rp = obj->firstRow; numIsGreater(row, rp->row); rp = rp->nextRow)
		;

	for (cp = rp->firstCell; numIsGreater(col, cp->col); cp = cp->nextCell)
		;

	if (numIsEqual(row, rp->row) && numIsEqual(col, cp->col) &&
		((cp->marks & MARK_SEARCH) == 0))
	{
	    	count++;
	}

	while (TRUE)
	{
		if (Life.stop)
			return FALSE;

		if (cp == Life.termCell)
		{
			rp = rp->nextRow;

			if (rp == Life.termRow)
			{
				lcClearSearch(obj);

				rp = obj->firstRow;
			}

			cp = rp->firstCell;

			continue;
		}

		if ((cp->marks & MARK_SEARCH) == 0)
		{
			lcSelectObject(obj, rp->row, cp->col, 1);
//WRONG
			if (--count <= 0)
				break;
		}

		cp = cp->nextCell;
	}

	obj->currentInfo.row = rp->row;
	obj->currentInfo.col = cp->col;

	return TRUE;
}


/*
 * Count the cells in a specified rectangular region.
 * Returns the number of cells in the region.
 */
NUMBER
lcCountRegion(const LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	const LifeRow *	rp;
	const LifeCell * cp;

	NUMBER count = numZero();

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		if (numIsGreater(minRow, rp->row))
			continue;

		if (numIsLess(maxRow, rp->row))
			break;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if (numIsGreater(minCol, cp->col))
				continue;

			if (numIsLess(maxCol, cp->col))
				break;

			numAddrInc(&count);
		}
	}

	return count;
}


/*
 * Calculate from scratch the number of ON cells in an object.
 * This does not reference the count or isUnknownCount elements.
 * Note: This routine probably is not ever called since the list
 * algorithm maintains correct counts.
 */
NUMBER
lcCalculateCellCount(const LifeObject * obj)
{
	const LifeRow *	rp;
	const LifeCell * cp;
	NUMBER count = numZero();

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
			numAddrInc(&count);
	}

	return count;
}


/*
 * Copy the cells from one object to another.  The source object is unchanged.
 * The current contents of the destination object are lost.  The source
 * and destination object must have the same algorithm.
 */
void
lcCopyObject(const LifeObject * srcObj, LifeObject * destObj)
{
	const LifeRow * rp;
	const LifeCell * cp;

	CheckSameAlgorithm(srcObj, destObj);

	lcDeleteAllCells(destObj);

	for (rp = srcObj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			lcAddCell(destObj, rp->row, cp->col);
		}
	}
}


/*
 * Add one object to another.  The source object is unchanged.
 * The destination object will get all cells from both objects.
 * The object is displaced as specified.  The source and destination
 * object must have the same algorithm.  All the added cells will be
 * in the MARK state.
 */
void
lcAddObject(const LifeObject * srcObj, LifeObject * destObj,
	NUMBER rowDisp, NUMBER colDisp)
{
	const LifeRow * rp;
	const LifeCell * cp;

	CheckSameAlgorithm(srcObj, destObj);

	for (rp = srcObj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER newRow = numAdd(rowDisp, rp->row);

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER newCol = numAdd(colDisp, cp->col);

			lcSetCellState(destObj, newRow, newCol, STATE_MARK);
		}
	}
}


/*
 * Move the cells from one object to another.  The source cells are
 * deleted, and the previous cells of the destination object are lost.
 * The source and destination objects must have the same algorithm.
 */
void
lcMoveObject(LifeObject * srcObj, LifeObject * destObj)
{
	CheckSameAlgorithm(srcObj, destObj);

	lcDeleteAllCells(destObj);

	destObj->firstRow = srcObj->firstRow;
	destObj->lastRow = srcObj->lastRow;
	destObj->usedQuadCells = srcObj->usedQuadCells;

	lcFreeCellData(srcObj);
}


/*
 * Free the cell data allocated to an object.
 */
void
lcFreeCellData(LifeObject * obj)
{
	obj->firstRow = Life.termRow;
	obj->lastRow = 0;
	obj->usedQuadCells = qcEmpty();
}


/*
 * Add all of the current cells of an object to its set of used cells.
 * This is used to display the cells which have ever been used by an
 * object during its evolution.  This uses the quadcell storage mechanism
 * since that can handle very large objects which have some regularity
 * in a relatively small space.  (This is bound to happen using the
 * tracks feature when a spaceship or puffer runs away.)
 * Warning: This can perform garbage collection.
 */
void
lcSaveUsedCells(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;

	//
	// Do nothing if tracking is not enabled.
	//
	if (!obj->isTracksUsed)
		return;

	/*
	 * Add all of the used cells to the quad cell.
	 */
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = rp->row;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER col = cp->col;

			if (Life.stop)
				Error("Saving tracked cells aborted");

			//
			// If we already know the cell is marked in the
			// usedQuadCells then skip on.  Otherwise add the
			// flag to avoid looking at the quadcells again.
			//
			if (cp->marks & MARK_USED)
				continue;

			cp->marks |= MARK_USED;

			//
			// First check whether the cell is already set
			// since that is common and faster.  TODO:
			// call a routine which can return multiple adjacent
			// ON cells in one call.
			//
			if (qcIsSet(obj->usedQuadCells, row, col))
				continue;

			//
			// The cell was not set, so set it.
			//
			obj->usedQuadCells = qcSet(obj->usedQuadCells, row, col);

			//
			// See if a garbage collection can be done.
			//
			CheckGarbageCollect();
		}
	}
}


/*
 * Delete all the cells of an object.
 * This also clears all marks and tracked cells.
 * Extraneous values (such as the cell count) are not reset.
 */
void
lcDeleteAllCells(LifeObject * obj)
{
	lcFreeRows(obj->firstRow);
	lcFreeCellData(obj);
}


/*
 * Shift the object left or right and up or down as specified.
 * This must also shift the marks to correspond.
 */
void
lcShiftObject(LifeObject * obj, NUMBER leftOrRight, NUMBER upOrDown)
{
	LifeRow * rp;
	LifeCell * cp;

	if (lcIsObjectEmpty(obj) ||
		(numIsZero(leftOrRight) && numIsZero(upOrDown)))
	{
		return;
	}

	/*
	 * Adjust the coordinates of the cells.
	 */
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		numAddrAdd(&rp->row, upOrDown);

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			numAddrAdd(&cp->col, leftOrRight);
		}
	}
}


/*
 * Initialise the data structures needed for the list algorithm.
 */
void
lcInit(void)
{
	Life.termCell = lcAllocateCell();
	Life.termCell->nextCell = Life.termCell;
	Life.termCell->col = numMaxValue();

	Life.termRow = lcAllocateRow();
	Life.termRow->nextRow = Life.termRow;
	Life.termRow->firstCell = Life.termCell;
	Life.termRow->row = numMaxValue();
}


