/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Definitions for routines which implement a list-based data structure
 * for manipulating Life objects.
 */

#ifndef	LISTALG_H
#define	LISTALG_H

#include "life.h"
//#include "algorithm.h"


/*
 * The structure holding function pointers to the list-based routines.
 */
extern LifeAlgorithm listAlgorithm;


/*
 * Global routines for manipulating list-based cell structures.
 * These are referenced by the function pointers in the LifeAlgorithm
 * structure.  Because of this, their usage is generic since a different
 * implementation can be used for the same functionality.
 */
extern	void lcAllocCellData(LifeObject * obj);
extern	void lcFreeCellData(LifeObject * obj);
extern	void lcRegisterQuadCells(LifeObject * obj);
extern	void lcSetCellState(LifeObject * obj, NUMBER row, NUMBER col, int state);
extern	int lcGetCellState(const LifeObject * obj, NUMBER row, NUMBER col);
extern	BOOL lcIsCellSet(const LifeObject * obj, NUMBER row, NUMBER col);
extern	BOOL lcAddCell(LifeObject *, NUMBER, NUMBER);
extern	BOOL lcDeleteCell(LifeObject *, NUMBER, NUMBER);
extern	void lcDeleteAllCells(LifeObject *);
extern	void lcShiftObject(LifeObject * obj, NUMBER, NUMBER);
extern	void lcSaveUsedCells(LifeObject *);
extern	void lcMoveObject(LifeObject * srcObj, LifeObject * destObj);
extern	void lcCopyObject(const LifeObject * srcObj, LifeObject * destObj);

extern	void lcAddObject(const LifeObject * srcObj, LifeObject * destObj,
	NUMBER rowDisp, NUMBER colDisp);

extern	NUMBER lcCalculateCellCount(const LifeObject * obj);
extern	BOOL lcIsObjectEmpty(const LifeObject * obj);
extern	BOOL lcSearchObject(LifeObject * obj, Count count);

extern	ViewInfo lcCalculateView(const LifeObject * obj, int viewId, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

extern	BOOL lcMinMax(const LifeObject *, NUMBER *, NUMBER *, NUMBER *, NUMBER *);	
extern	BOOL lcSelectObject(LifeObject *, NUMBER, NUMBER, Value);
extern	BOOL lcSelectPath(LifeObject *);
extern	BOOL lcSelectAll(LifeObject *);
extern	BOOL lcMarkSelection(LifeObject *);
extern	BOOL lcSelectMarkedCells(LifeObject *);
extern	void lcClearSelection(LifeObject *);
extern	void lcClearSearch(LifeObject *);
extern	void lcClearTracks(LifeObject *);
extern	void lcClearMarks(LifeObject *);
extern	void lcInvertMarks(LifeObject *);
extern	NUMBER lcCountMarks(const LifeObject *);

extern	BOOL lcMarkMinMax(const LifeObject *,
	NUMBER *, NUMBER *, NUMBER *, NUMBER *);

extern	BOOL lcFindClosestSelectedCell(const LifeObject *,
	NUMBER, NUMBER, NUMBER *, NUMBER *);

extern	NUMBER lcFindMarkedGap(const LifeObject *);

extern	void lcRotateSelection(LifeObject *);
extern	void lcFlipRowSelection(LifeObject *);
extern	void lcFlipColumnSelection(LifeObject *);

extern	void lcMoveSelection(LifeObject *, LifeObject *);
extern	void lcCopyMarkedObject(const LifeObject *, LifeObject *);

extern	BOOL lcSelectCell(LifeObject *, NUMBER, NUMBER);
extern	BOOL lcSelectRegion(LifeObject *, NUMBER, NUMBER, NUMBER, NUMBER);
extern	NUMBER lcCountRegion(const LifeObject *, NUMBER, NUMBER, NUMBER, NUMBER);
extern	void lcCountPathCells(const LifeObject *, NUMBER *, NUMBER *, NUMBER *);

extern	BOOL lcNextGeneration(LifeObject * obj);
extern	Count lcRunGenerations(LifeObject * obj, Count genCount);

extern	BOOL lcScanObject(LifeObject * obj, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data);

extern	NUMBER lcGetObjectMaxRowUptoLimit(LifeObject * obj, NUMBER limitRow);
extern	NUMBER lcGetObjectMinRowDowntoLimit(LifeObject * obj, NUMBER limitRow);

extern QuadCell * lcExtractQuadCell(LifeObject * obj);


//
// Supporting routines.
//
extern	LifeRow * lcAllocateRow(void);
extern	LifeRow * lcFindRow(const LifeObject *, NUMBER);
extern	LifeRow * lcGetRow(LifeObject *, NUMBER);
extern	LifeRow * lcCopyRows(const LifeRow *);
extern	NUMBER lcFreeRows(LifeRow *);
extern	LifeCell * lcAllocateCell(void);
extern	LifeCell * lcFindCell(const LifeObject *, NUMBER, NUMBER);
extern	void lcInit(void);

#endif

