/*
 * Copyright (c) 2015 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Row and column operations on objects for the list algorithm.
 */

#include "life.h"
#include "listalg.h"


/*
 * Return a row structure for a given row number, returning NULL if not there.
 */
LifeRow *
lcFindRow(const LifeObject * obj, NUMBER row)
{
	LifeRow * rp;

	/*
	 * Check for a row at the end of the object.
	 */
	if (obj->lastRow && numIsGreaterOrEqual(row, obj->lastRow->row))
	{
		if (numIsEqual(row, obj->lastRow->row))
			return obj->lastRow;

		return 0;
	}

	/*
	 * Search for the row from the beginning.
	 */
	for (rp = obj->firstRow; numIsGreater(row, rp->row); rp = rp->nextRow)
		;

	if (numIsNotEqual(rp->row, row))
		return 0;

	return rp;
}


/*
 * Return a row structure for a given row number, creating it if needed.
 */
LifeRow *
lcGetRow(LifeObject * obj, NUMBER row)
{
	LifeRow * rp;		/* current row */
	LifeRow * nrp;		/* next row */
	LifeRow * prp;		/* previous row */

	prp = 0;
	rp = obj->firstRow;

	/*
	 * Handle the case that the row is at the front.
	 */
	if (numIsLess(row, rp->row))
	{
		nrp = lcAllocateRow();

		nrp->row = row;
		nrp->nextRow = obj->firstRow;
		obj->firstRow = nrp;

		if (nrp->nextRow == Life.termRow)
			obj->lastRow = nrp;

		return nrp;
	}

	if (numIsGreaterOrEqual(row, obj->lastRow->row))
		rp = obj->lastRow;

	while (numIsGreater(row, rp->row))
	{
		prp = rp;
		rp = rp->nextRow;
	}

	if (numIsEqual(row, rp->row))
		return rp;

	nrp = lcAllocateRow();

	nrp->row = row;
	nrp->nextRow = rp;
	prp->nextRow = nrp;

	if (nrp->nextRow == Life.termRow)
		obj->lastRow = nrp;

	return nrp;
}


/*
 * Find a cell given its coordinates, returning NULL if not found.
 */
LifeCell *
lcFindCell(const LifeObject * obj, NUMBER row, NUMBER col)
{
	const LifeRow * rp;
	LifeCell * cp;

	rp = lcFindRow(obj, row);

	if (rp == 0)
		return 0;

	for (cp = rp->firstCell; numIsGreater(col, cp->col); cp = cp->nextCell)
		;

	if (numIsNotEqual(col, cp->col))
		return 0;

	return cp;
}


/*
 * Return the state of a cell at the given row and column.
 * TODO: Check DEAD and NEW states.
 */
int
lcGetCellState(const LifeObject * obj, NUMBER row, NUMBER col)
{
	LifeCell * cp = lcFindCell(obj, row, col);

	if (cp == 0)
		return STATE_OFF;

	if (cp->marks & MARK_USER)
		return STATE_MARK;

	return STATE_ON;
}


/*
 * Set the pure state of a cell at the given row and column.
 * This is one of OFF, ON, MARK, DEAD, or NEW.
 * TODO: Implement DEAD and NEW.
 */
void
lcSetCellState(LifeObject * obj, NUMBER row, NUMBER col, int state)
{
	if ((state == STATE_OFF) || (state == STATE_DEAD))
	{
		lcDeleteCell(obj, row, col);

		return;
	}

	LifeCell * cp = lcFindCell(obj, row, col);

	//
	// If the cell wasn't present then add it.
	//
	if (cp == 0)
	{
		lcAddCell(obj, row, col);

		if ((state == STATE_ON) || (state == STATE_NEW))
			return;

		cp = lcFindCell(obj, row, col);
		// This could be better with an add which returns the pointer
	}

	//
	// Set or clear the mark flag.
	//
	if (state == STATE_MARK)
		cp->marks |= MARK_USER;
	else
		cp->marks &= ~MARK_USER;
}


/*
 * Return whether the cell at a given row and column is set.
 */
BOOL
lcIsCellSet(const LifeObject * obj, NUMBER row, NUMBER col)
{
	return (lcFindCell(obj, row, col) != 0);
}


/*
 * Create a cell at a given row and column.  Returns FALSE if the cell
 * already existed.
 */
BOOL
lcAddCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	LifeCell * cp;		/* current cell */
	LifeCell * ncp;		/* next cell */
	LifeCell * pcp;		/* previous cell */
	LifeRow * rp;

	rp = lcGetRow(obj, row);
	cp = rp->firstCell;
	pcp = 0;

	if ((cp != Life.termCell) &&
		numIsGreaterOrEqual(col, rp->lastCell->col))
	{
		/*
		 * Inserting at end.
		 */
		pcp = rp->lastCell;

		if (numIsEqual(col, pcp->col))
			return FALSE;

		ncp = lcAllocateCell();

		ncp->col = col;
		ncp->nextCell = Life.termCell;
		pcp->nextCell = ncp;
		rp->lastCell = ncp;

		numAddrInc(&rp->count);

		return TRUE;
	}

	if (numIsLess(col, cp->col))
	{
		/*
		 * Inserting at the front.
		 */
		ncp = lcAllocateCell();

		ncp->col = col;
		ncp->nextCell = cp;
		rp->firstCell = ncp;

		if (cp == Life.termCell)
			rp->lastCell = ncp;

		numAddrInc(&rp->count);

		return TRUE;
	}

	while (numIsGreater(col, cp->col))
	{
		pcp = cp;
		cp = pcp->nextCell;
	}

	if (numIsEqual(col, cp->col))
		return FALSE;

	ncp = lcAllocateCell();

	ncp->col = col;
	ncp->marks = MARK_NONE;
	ncp->nextCell = cp;
	pcp->nextCell = ncp;

	if (cp == Life.termCell)
		rp->lastCell = ncp;

	numAddrInc(&rp->count);

	return TRUE;
}


/*
 * Delete a cell at a given coordinate.
 * Returns FALSE if it did not exist.
 */
BOOL
lcDeleteCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	LifeRow * rp;
	LifeCell * pcp;		/* previous cell */
	LifeCell * cp;		/* current cell */

	rp = lcFindRow(obj, row);

	if (rp == 0)
		return FALSE;

	pcp = 0;
	cp = rp->firstCell;

	while (numIsGreater(col, cp->col))
	{
		pcp = cp;
		cp = cp->nextCell;
	}

	if (numIsNotEqual(col, cp->col))
		return FALSE;

	if (pcp)
		pcp->nextCell = cp->nextCell;
	else
		rp->firstCell = cp->nextCell;

	if (cp->nextCell == Life.termCell)
		rp->lastCell = pcp;

	cp->nextCell = Life.freeCells;
	Life.freeCells = cp;

	numAddrDec(&rp->count);

	return TRUE;
}


/*
 * Return the quadcell representation of an object.
 * This is a bit painful.  We basically have to act like we are
 * reading in an object and setting the cells one by one.
 * Warning: This can perform garbage collection.
 */
QuadCell *
lcExtractQuadCell(LifeObject * obj)
{
	const LifeRow * rp;
	const LifeCell * cp;
	QuadCell * cells = qcEmpty();

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			cells = qcSet(cells, rp->row, cp->col);

			qcAddSafeQuadCell(cells);
			CheckGarbageCollect();
			qcClearSafeQuadCells(1);

			if (Life.stop)
				return cells;
		}
	}

	return cells;
}


/*
 * Show statistics about the use of quad cells.
 */
void
ShowQuadCellStatistics(void)
{
	QuadStats stats = qcGetStats();
	int size;

	ShowHelp("Quad Cell Statistics\n");
	ShowHelp("--------------------\n\n");

	FormatHelp("Find Empties:        %ld\n", stats.findEmpties);
	FormatHelp("Find Fulls:          %ld\n", stats.findFulls);
	FormatHelp("Find Misses:         %ld\n", stats.findMisses);
	FormatHelp("Find Hits:           %ld\n", stats.findHits);
	FormatHelp("Find Length:         %ld\n", stats.findLength);
	FormatHelp("QuadPair Misses:     %ld\n", stats.quadPairMisses);
	FormatHelp("QuadPair Hits:       %ld\n", stats.quadPairHits);
	FormatHelp("Garbage Collections: %ld\n", stats.garbageCollections);
	FormatHelp("Cells Freed:         %ld\n", stats.cellsFreed);
	FormatHelp("Cells Allocated:     %ld\n", stats.cellsAllocated);
	FormatHelp("Cells Available:     %ld\n", stats.cellsAvailable);
	FormatHelp("Cells Next Index:    %ld\n", stats.cellsNextIndex);
	FormatHelp("Hash Table Growths:  %ld\n", stats.hashTableGrowths);
	FormatHelp("Hash Table Size:     %ld\n", stats.hashTableSize);
	FormatHelp("Hash Table Used:     %ld\n", stats.hashTableUsed);

	for (size = 0; size < QUAD_STAT_SIZE; size++)
	{
		if (stats.sizes[size])
		{
			FormatHelp("Size %d:              %ld\n",
				size, stats.sizes[size]);
		}
	}

	for (size = 0; size < QUAD_STAT_SIZE; size++)
	{
		if (stats.regionOffCounts[size])
		{
			FormatHelp("Region Off %d:        %ld\n",
				size, stats.regionOffCounts[size]);
		}
	}

	for (size = 0; size < QUAD_STAT_SIZE; size++)
	{
		if (stats.regionOnCounts[size])
		{
			FormatHelp("Region On  %d:        %ld\n",
				size, stats.regionOnCounts[size]);
		}
	}

	EndHelp(TRUE);
}

/* END CODE */
