/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to support the running of Life generations for the
 * list algorithm.
 */

#include "life.h"
#include "listalg.h"


//
// Local routines.
//
static LifeRow * lcComputeRow(LifeObject *, LifeRow *, LifeRow *, LifeRow *);


/*
 * Try to compute a number of generations for the specified object.
 * This also updates the number of cells which are born and died
 * overall and the cell count.  We also see if the stop calculating
 * flag has been set so that running a large number of generations
 * can be stopped.  Returns the number of generations computed.
 */
Count
lcRunGenerations(LifeObject * obj, Count genCount)
{
	//
	// Save the current cells in the tracking quadcell.
	//
	if (obj->isTracksUsed)
		lcSaveUsedCells(obj);

	//
	// Handle single generations specially.
	//
	if (genCount <= 0)
	{
		obj->growth = numZero();
		obj->born = numZero();
		obj->died = numZero();

		return 0;
	}

	NUMBER oldCount = obj->count;

	if (genCount == 1)
	{
		if (lcNextGeneration(obj))
		{
			obj->growth = numSub(obj->count, oldCount);

			return 1;
		}

		return 0;
	}

	//
	// If running multiple generations, we don't want any
	// born and died counts, but only the growth count.
	//
	Count runCount = 0;

	while ((genCount-- > 0) && !Life.stop)
	{
		if (!lcNextGeneration(obj))
			break;

		runCount++;
	}

	obj->growth = numSub(obj->count, oldCount);
	obj->born = numZero();
	obj->died = numZero();

	return runCount;
}


/*
 * Try to compute one full generation of the current configuration.
 * This is done by calling lcComputeRow with each triple of rows which
 * contain any live cells, and remembering the result.  When all rows
 * are finished, we change the object.  This also updates the number of
 * cells which are born and died and the cell count.  Returns TRUE if a
 * generation was run, or FALSE if the object was empty.
 */
BOOL
lcNextGeneration(LifeObject * obj)
{
	LifeRow * curRp;	/* current row */
	LifeRow * prevRp;	/* previous row */
	LifeRow * nextRp;	/* next row */
	LifeRow * newRp;	/* current row of new list */
	LifeRow * saveRp;	/* saved row pointer */
	NUMBER countAdjustment = numZero();

	obj->born = numZero();
	obj->died = numZero();
	obj->growth = numZero();

	if (lcIsObjectEmpty(obj))
		return FALSE;

	//
	// We are not empty so we can do the calculation.
	//
	obj->gen++;

	newRp = &Life.initRow;
	prevRp = Life.termRow;
	curRp = Life.termRow;
	nextRp = obj->firstRow;
	saveRp = nextRp->nextRow;
	NUMBER row = numDec(nextRp->row);

	/*
	 * Loop over each triple of rows.
	 */
	while (numIsLess(row, numDec(numMaxValue())))
	{
		prevRp = lcComputeRow(obj, prevRp, curRp, nextRp);

		/*
		 * See if there is something in the current row.
		 */
		if (prevRp)
		{
			newRp->nextRow = prevRp;
			newRp = prevRp;
			newRp->row = row;
			numAddrAdd(&countAdjustment, prevRp->count);
		}

		numAddrInc(&row);

		prevRp = curRp;
		curRp = nextRp;
		nextRp = saveRp;

		if (numIsEqual(nextRp->row, numInc(row)))
		{
			saveRp = nextRp->nextRow;

			continue;
		}

		nextRp = Life.termRow;

		if ((prevRp != Life.termRow) || (curRp != Life.termRow))
			continue;

		nextRp = saveRp;
		saveRp = nextRp->nextRow;

		row = numDec(nextRp->row);
	}

	numAddrSub(&countAdjustment, lcFreeRows(obj->firstRow));

	AdjustCellCount(obj, countAdjustment);

	obj->firstRow = Life.termRow;
	obj->lastRow = 0;

	obj->growth = numSub(obj->born, obj->died);

	/*
	 * Handle the case that the object died off.
	 */
	if (newRp == &Life.initRow)
		return TRUE;

	obj->firstRow = Life.initRow.nextRow;
	newRp->nextRow = Life.termRow;
	obj->lastRow = newRp;

	return TRUE;
}


/*
 * Compute the result of three adjacent rows, and return a row structure
 * containing the new middle row, or NULL if no live cells are produced.
 * Any new cells are set to have all of the marks contained in any of the
 * neighboring cells.
 */
static LifeRow *
lcComputeRow(LifeObject * obj,
	LifeRow * prevrow, LifeRow * curRow, LifeRow * nextrow)
{
	LifeCell * curCp;	/* head of current row of cells */
	LifeCell * prevCp;	/* head of previous row of cells */
	LifeCell * nextCp;	/* head of next row of cells */
	LifeCell * tempCp;	/* temporary cell */
	LifeCell * newCp;	/* new row of cells */
	LifeRow * rp;		/* new row pointer */
	NUMBER col;		/* current column being examined */
	NUMBER count;		/* live cells */

	prevCp = prevrow->firstCell;
	curCp = curRow->firstCell;
	nextCp = nextrow->firstCell;
	newCp = &Life.initCell;
	count = numZero();
	col = numInc(numMinValue());

	/*
	 * Loop over the cells of all three rows.
	 * Each cell's fate depends on its neighbors according to a
	 * 3 by 3 square of cells, as in:
	 *	abc
	 *	hid
	 *	gfe
	 * Here cell I is the middle cell whose fate is being calculated.
	 */
	while (TRUE)
	{
		BOOL cellA = FALSE;
		BOOL cellB = FALSE;
		BOOL cellC = FALSE;
		BOOL cellD = FALSE;
		BOOL cellE = FALSE;
		BOOL cellF = FALSE;
		BOOL cellG = FALSE;
		BOOL cellH = FALSE;
		BOOL cellI = FALSE;
		BOOL wasLive = FALSE;

		/*
		 * Find next column where a cell exists on any row.
		 */
		NUMBER i = numDec(col);

		while (numIsLess(prevCp->col, i))
			prevCp = prevCp->nextCell;

		while (numIsLess(curCp->col, i))
			curCp = curCp->nextCell;

		while (numIsLess(nextCp->col, i))
			nextCp = nextCp->nextCell;

		i = curCp->col;

		if (numIsLess(prevCp->col, i))
			i = prevCp->col;

		if (numIsLess(nextCp->col, i))
			i = nextCp->col;

		if (numIsEqual(i, numMaxValue()))
			break;

		numAddrDec(&i);

		if (numIsLess(col, i))
			col = i;

		NUMBER colPlus1 = numInc(col);
		NUMBER colMinus1 = numDec(col);

		Mark marks = 0;

		/*
		 * Check the values for the live cells in the previous row.
		 */
		tempCp = prevCp;

		if (numIsEqual(tempCp->col, colMinus1))
		{
			cellA = TRUE;
			marks |= tempCp->marks;
			tempCp = tempCp->nextCell;
		}

		if (numIsEqual(tempCp->col, col))
		{
			cellB = TRUE;
			marks |= tempCp->marks;
			tempCp = tempCp->nextCell;
		}

		if (numIsEqual(tempCp->col, colPlus1))
		{
			cellC = TRUE;
			marks |= tempCp->marks;
		}

		/*
		 * Check the values for the live cells in the current row.
		 * Also remember whether the current cell is live.
		 */
		tempCp = curCp;

		if (numIsEqual(tempCp->col, colMinus1))
		{
			cellH = TRUE;
			marks |= tempCp->marks;
			tempCp = tempCp->nextCell;
		}

		if (numIsEqual(tempCp->col, col))
		{
			wasLive = TRUE;
			cellI = TRUE;
			marks |= tempCp->marks;
			tempCp = tempCp->nextCell;
		}

		if (numIsEqual(tempCp->col, colPlus1))
		{
			cellD = TRUE;
			marks |= tempCp->marks;
		}

		/*
		 * Check the values for the live cells in the next row.
		 */
		tempCp = nextCp;

		if (numIsEqual(tempCp->col, colMinus1))
		{
			cellG = TRUE;
			marks |= tempCp->marks;
			tempCp = tempCp->nextCell;
		}

		if (numIsEqual(tempCp->col, col))
		{
			cellF = TRUE;
			marks |= tempCp->marks;
			tempCp = tempCp->nextCell;
		}

		if (numIsEqual(tempCp->col, colPlus1))
		{
			cellE = TRUE;
			marks |= tempCp->marks;
		}

		/*
		 * Determine if the cell is alive or dead.
		 */
		if (IsNewCellAlive(cellA, cellB, cellC, cellD,
			cellE, cellF, cellG, cellH, cellI))
		{
			numAddrAddLong(&obj->born, !wasLive);
			tempCp = lcAllocateCell();
			tempCp->col = col;
			tempCp->marks = marks;
			newCp->nextCell = tempCp;
			newCp = tempCp;
			numAddrInc(&count);
		}
		else
			numAddrAddLong(&obj->died, wasLive);

		numAddrInc(&col);
	}

	if (newCp == &Life.initCell)
		return 0;

	newCp->nextCell = Life.termCell;

	rp = lcAllocateRow();
	rp->firstCell = Life.initCell.nextCell;
	rp->lastCell = newCp;
	rp->count = count;

	return rp;
}



