/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Marking routines for the list algorithm.
 */

#include "life.h"
#include "listalg.h"


/*
 * Values used for marking of cells.
 * This is the amount of space used in a stack frame, and the amount
 * of space to reserve to guarantee that the results of checking one
 * cell for the specified gap size will fit in the array.
 */
#define	MARK_LOOP_ALLOC		4096
#define	MARK_LOOP_RESERVED(gap)	(((gap) * 2 + 3) * ((gap) * 2 + 3))


static void lcSelectLoop(LifeObject *, NUMBER, NUMBER, Value);


/*
 * Local procedures.
 */
static NUMBER lcFindUnmarkedDistance(const LifeObject *, NUMBER,
	NUMBER, NUMBER);


/*
 * Select the object at a given location as specified.  Returns FALSE if
 * no object exists there.  An object for this purpose is considered the
 * largest set of cells containing the specified cell whose members are
 * separated by a distance no larger then the specified gap.  For example,
 * a gap of zero means the cells which are king-wise connected to the
 * specified cell.
 */
BOOL
lcSelectObject(LifeObject * obj, NUMBER row, NUMBER col, Value gap)
{
	if ((gap < 0) || (gap > 9))
		Error("Marking gap is illegal");
		
	LifeCell * cp = lcFindCell(obj, row, col);

	if (cp == 0)
		return FALSE;

	cp->marks |= MARK_SELECT;

	lcSelectLoop(obj, row, col, gap);

	return TRUE;
}


/*
 * Recursive subroutine called from SelectObject.
 * This stores cells as they are found in our array, and then uses
 * the found cells to add more cells to the array if they were not
 * yet marked.  We cell ourself recursively if our array gets full.
 */
static void
lcSelectLoop(LifeObject * obj, NUMBER row, NUMBER col, Value gap)
{
	int	storeOffset;
	int	examineOffset;
	int	entries;
	int	reserved;
	NUMBER	rowTable[MARK_LOOP_ALLOC];
	NUMBER	colTable[MARK_LOOP_ALLOC];

	/*
	 * Initialize the table with one entry containing the
	 * specified coordinates.
	 */
	rowTable[0] = row;
	colTable[0] = col;

	reserved = MARK_LOOP_RESERVED(gap);
	examineOffset = 0;
	storeOffset = 1;
	entries = 1;

	/*
	 * Keep examining cells until there are none left, adding new
	 * locations that have been found to the end of the table to
	 * be examined later.  If the table fills up, then call ourself
	 * recursively.  The offsets in the table wrap around as they
	 * are used.
	 */
	while (!Life.stop && (entries > 0))
	{
		/*
		 * Get the next location to be examined from the table.
		 */
		row = rowTable[examineOffset];
		col = colTable[examineOffset];

		if (++examineOffset >= MARK_LOOP_ALLOC)
			examineOffset = 0;

		entries--;

		/*
		 * If there is not room left in the table for the maximum
		 * number of cells that could be marked around this cell,
		 * then call ourself recursively to process this location.
		 */
		if (entries + reserved > MARK_LOOP_ALLOC)
		{
			lcSelectLoop(obj, row, col, gap);

			continue;
		}

		/*
		 * The table is guaranteed to have room to store any new
		 * cells around this location within the gap interval.
		 * Set up the boundaries of the area to be checked.
		 */
		NUMBER minRow = numSubLong(row, gap + 1);
		NUMBER maxRow = numAddLong(row, gap + 1);
		NUMBER minCol = numSubLong(col, gap + 1);
		NUMBER maxCol = numAddLong(col, gap + 1);

		/*
		 * Look at every cell in the area if it exists,
		 * adding their coordinates to the table if they
		 * aren't yet marked.
		 */
		for (row = minRow; numIsLessOrEqual(row, maxRow);
			numAddrInc(&row))
		{
			for (col = minCol; numIsLessOrEqual(col, maxCol);
				numAddrInc(&col))
			{
				/*
				 * If the cell does not exist or is
				 * already marked then skip it.
				 */
				LifeCell * cp = lcFindCell(obj, row, col);

				if ((cp == 0) || (cp->marks & MARK_SELECT))
					continue;

				/*
				 * Mark this new cell and add it's location
				 * to the table to be examined later.
				 */
				cp->marks |= MARK_SELECT;

				rowTable[storeOffset] = row;
				colTable[storeOffset] = col;

				if (++storeOffset >= MARK_LOOP_ALLOC)
					storeOffset = 0;

				entries++;
			}
		}
	}
}


/*
 * Select all the cells in an object.
 * Returns TRUE if there were some cells to select.
 */
BOOL
lcSelectAll(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
			cp->marks |= MARK_SELECT;
	}

	return TRUE;
}


/*
 * Mark the cells in the object from the previously made selection.
 * Returns FALSE if there was no selection.
 */
BOOL
lcMarkSelection(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;
	BOOL isFound = FALSE;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if ((cp->marks & MARK_SELECT) == 0)
				continue;

			cp->marks |= MARK_USER;
			isFound = TRUE;
		}
	}

	return isFound;
}


/*
 * Select the marked cells in the object.
 * Returns FALSE if there were no marked cells.
 */
BOOL
lcSelectMarkedCells(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;
	BOOL isFound = FALSE;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if ((cp->marks & MARK_USER) == 0)
				continue;

			cp->marks |= MARK_SELECT;
			isFound = TRUE;
		}
	}

	return isFound;
}


/*
 * Invert the marks for a whole object.
 */
void
lcInvertMarks(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
			cp->marks ^= MARK_USER;
	}
}


/*
 * Clear marks for a whole object.
 */
void
lcClearMarks(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
			cp->marks &= ~MARK_USER;
	}
}


/*
 * Clear the search state.
 */
void
lcClearSearch(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
			cp->marks &= ~MARK_SEARCH;
	}
}


/*
 * Clear the tracks for an object.
 * This replaces DEAD and NEW cells by OFF and ON cells.
 */
void
lcClearTracks(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;

	//
	// Remove the USED flag if they were set.
	//
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
			cp->marks &= ~MARK_USED;
	}

	//
	// Clear out the used quadcell.
	//
	obj->usedQuadCells = qcEmpty();
}


/*
 * Clear the selection for a whole object.
 */
void
lcClearSelection(LifeObject * obj)
{
	const LifeRow * rp;
	LifeCell * cp;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
			cp->marks &= ~MARK_SELECT;
	}
}


/*
 * Select the cell at the specified location.
 * Returns TRUE if there was a cell at the specified location.
 */
BOOL
lcSelectCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	LifeCell * cp = lcFindCell(obj, row, col);

	if (cp == 0)
		return FALSE;

	cp->marks |= MARK_SELECT;

	return TRUE;
}


/*
 * Select the cells in a specified rectangular region as desired.
 * Returns TRUE if the region contained at least one cell to be selected.
 */
BOOL
lcSelectRegion(LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	const LifeRow * rp;
	LifeCell * cp;
	BOOL isMarked = FALSE;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = rp->row;

		if (numIsLess(row, minRow))
			continue;

		if (numIsGreater(row, maxRow))
			break;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER col = cp->col;

			if (numIsLess(col, minCol))
				continue;

			if (numIsGreater(col, maxCol))
				break;

			cp->marks |= MARK_SELECT;
			isMarked = TRUE;
		}
	}

	return isMarked;
}


/*
 * Find the range of all marked cells for an object.
 * Returns FALSE if no cells were marked.
 */
BOOL
lcMarkMinMax(const LifeObject * obj,
	NUMBER * minRow, NUMBER * maxRow, NUMBER * minCol, NUMBER * maxCol)
{
	const LifeRow * rp;
	const LifeCell * cp;

	NUMBER minr = numMaxValue();
	NUMBER maxr = numMinValue();
	NUMBER minc = numMaxValue();
	NUMBER maxc = numMinValue();
	BOOL isFound = FALSE;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = rp->row;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if ((cp->marks & MARK_USER) == 0)
				continue;

			minr = numMin(minr, row);
			maxr = numMax(maxr, row);

			NUMBER col = cp->col;

			minc = numMin(minc, col);
			maxc = numMax(maxc, col);

			isFound = TRUE;
		}
	}

	if (!isFound)
		return FALSE;

	*minRow = minr;
	*maxRow = maxr;
	*minCol = minc;
	*maxCol = maxc;

	return TRUE;
}


/*
 * Count the number of marked cells in an object.
 */
NUMBER
lcCountMarks(const LifeObject * obj)
{
	const LifeRow * rp;
	const LifeCell * cp;
	NUMBER count = numZero();

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if (cp->marks & MARK_USER)
				numAddrInc(&count);
		}
	}

	return count;
}


/*
 * Move the previously selected cells to another object.  If the destination
 * object is NULL the cells are just deleted.  The destination object has
 * already been cleared.  The source and destination objects must have
 * the same algorithm.
 */
void
lcMoveSelection(LifeObject * srcObj, LifeObject * destObj)
{
	LifeRow * rp;
	LifeCell * cp;		/* current cell pointer */
	LifeCell * pcp;		/* previous cell pointer */
	LifeCell * ncp;		/* next cell pointer */
	NUMBER count = numZero();

	CheckSameAlgorithm(srcObj, destObj);

	for (rp = srcObj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		pcp = 0;
		cp = rp->firstCell;

		while (cp != Life.termCell)
		{
			if ((cp->marks & MARK_SELECT) == 0)
			{
				pcp = cp;
				cp = cp->nextCell;

				continue;
			}

			if (destObj)
			{
				lcAddCell(destObj, rp->row, cp->col);
			}

			ncp = cp->nextCell;

			if (pcp == 0)
				rp->firstCell = ncp;
			else
				pcp->nextCell = ncp;

			if (ncp == Life.termCell)
				rp->lastCell = pcp;

			cp->nextCell = Life.freeCells;
			Life.freeCells = cp;
			cp = ncp;
			numAddrDec(&rp->count);
			numAddrInc(&count);
			IncrementChangeCount(srcObj);
		}
	}

	AdjustCellCount(srcObj, numNeg(count));
}


/*
 * Copy marked cells to another object. The previous cells of the destination
 * object have already been cleared.  The source object is unaffected.
 * The source and destination objects must have the same algorithm.
 */
void
lcCopyMarkedObject(const LifeObject * srcObj, LifeObject * destObj)
{
	const LifeRow * rp;
	const LifeCell * cp;

	CheckSameAlgorithm(srcObj, destObj);

	for (rp = srcObj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if (cp->marks & MARK_USER)
			{
				lcAddCell(destObj, rp->row, cp->col);
			}
		}
	}
}


/*
 * Rotate the previously selected cells in the given object around the
 * current cursor location.  This deletes the selected cells, and reinserts
 * them after their position has been rotated by 90 degrees clockwise.
 * The cells are inserted in the MARK state.
 */
void
lcRotateSelection(LifeObject * obj)
{
	const LifeRow * rp;
	const LifeCell * cp;

	PrepareTempObject(obj);

	lcMoveSelection(obj, Life.tempObject);

	for (rp = Life.tempObject->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = numSub(rp->row, obj->currentInfo.row);

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER col = numSub(cp->col, obj->currentInfo.col);

			lcSetCellState(obj, numAdd(obj->currentInfo.row, col),
				numSub(obj->currentInfo.col, row), STATE_MARK);
		}
	}
}


/*
 * Flip the selected cells in the given object around the column of the current
 * cursor location.  This deletes the selected cells, and reinserts them after
 * their position has been flipped around the vertical axis.  The inserted
 * cells are in the MARK state.
 */
void
lcFlipColumnSelection(LifeObject * obj)
{
	const LifeRow * rp;
	const LifeCell * cp;

	PrepareTempObject(obj);

	lcMoveSelection(obj, Life.tempObject);

	for (rp = Life.tempObject->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = rp->row;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER col = numSub(cp->col, obj->currentInfo.col);

			lcSetCellState(obj, row,
				numSub(obj->currentInfo.col, col), STATE_MARK);
		}
	}
}


/*
 * Flip the previously selected cells in the given object around the row of
 * the current cursor location.  This deletes the selected cells, and reinserts
 * them after their position has been flipped around the horizontal axis.
 * The inserted cells have the MARK state.
 */
void
lcFlipRowSelection(LifeObject * obj)
{
	const LifeRow * rp;
	const LifeCell * cp;

	PrepareTempObject(obj);

	lcMoveSelection(obj, Life.tempObject);

	for (rp = Life.tempObject->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = numSub(numDouble(obj->currentInfo.row), rp->row);

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			lcSetCellState(obj, row, cp->col, STATE_MARK);
		}
	}
}


/*
 * Find the minimum distance between any marked cell to any unmarked cell.
 * The distance is computed as the taxicab distance (maximum of x and y).
 * Returns 0 if there are no marked cells or no unmarked cells.
 */
NUMBER
lcFindMarkedGap(const LifeObject * obj)
{
	const LifeRow *	rp;
	const LifeCell * cp;
	BOOL haveMark = FALSE;
	BOOL haveUnmark = FALSE;

	/*
	 * First make one pass through all the cells of the object
	 * looking for one marked cell and one unmarked cell.
	 */
	NUMBER markRow = numZero();
	NUMBER markCol = numZero();
	NUMBER unmarkRow = numZero();
	NUMBER unmarkCol = numZero();

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if (cp->marks & MARK_USER)
			{
				haveMark = TRUE;
				markRow = rp->row;
				markCol = cp->col;
			}
			else
			{
				haveUnmark = TRUE;
				unmarkRow = rp->row;
				unmarkCol = cp->col;
			}

			/*
			 * If we have found one of each type of cell
			 * then we can stop the loop.
			 */
			if (haveMark && haveUnmark)
				break;
		}

		if (haveMark && haveUnmark)
			break;
	}

	/*
	 * If we didn't find one of each type of cell, then return zero.
	 */
	if (!haveMark || !haveUnmark)
		return numZero();

	/*
	 * Initialise the gap to the distance between these two cells.
	 * Further examination might be able to shrink this gap.
	 */
	NUMBER rowGap = numGap(markRow, unmarkRow);
	NUMBER colGap = numGap(markCol, unmarkCol);

	NUMBER gap = numMax(rowGap, colGap);

	/*
	 * Now search through all of the marked cells in the object
	 * trying to shrink the gap to its minimum value.
	 */
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if ((cp->marks & MARK_USER) == 0)
				continue;

			/*
			 * We found a marked cell.
			 * Scan for the minimum distance from this cell to
			 * any unmarked cell trying to reduce the gap.
			 */
			gap = lcFindUnmarkedDistance(obj, gap, rp->row, cp->col);

			/*
			 * If the gap is is only 1, then we have the best
			 * result.
			 */
			if (numCmpLong(gap, 1) <= 0)
				return gap;
		}
	}

	/*
	 * Return the minimum gap value.
	 */
	return gap;
}


/*
 * Find the minimum distance from the specified position to an unmarked
 * cell within the specified gap distance.  Returns the new gap distance
 * if a cell is found with a smaller gap, or the incoming gap distance if
 * no cell was found with a smaller gap.
 */
NUMBER
lcFindUnmarkedDistance(const LifeObject * obj,
	NUMBER gap, NUMBER baseRow, NUMBER baseCol)
{
	const LifeRow * rp;
	const LifeCell * cp;

	/*
	 * Set an initial bounding box for the search based on the gap.
	 * We aren't interested in cells outside of this area since
	 * they can't shrink the gap value.
	 */
	NUMBER minRow = numInc(numSub(baseRow, gap));
	NUMBER maxRow = numDec(numAdd(baseRow, gap));
	NUMBER minCol = numInc(numSub(baseCol, gap));
	NUMBER maxCol = numDec(numAdd(baseCol, gap));

	/*
	 * Loop over all rows of the object.
	 */
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		/*
		 * If the current row is not closer than the current
		 * gap value, then ignore it.
		 */
		NUMBER row = rp->row;

		if (numIsLess(row, minRow))
			continue;

		if (numIsGreater(row, maxRow))
			break;

		/*
		 * Loop over all of the cells of the row.
		 */
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			/*
			 * If the current column is not closer than the
			 * current gap value, then ignore it.
			 */
			NUMBER col = cp->col;

			if (numIsLess(col, minCol))
				continue;

			if (numIsGreater(col, maxCol))
				break;

			/*
			 * Ignore marked cells.
			 */
			if (cp->marks & MARK_USER)
				continue;

			/*
			 * Calculate the gap from the base cell to this one.
			 */
			NUMBER rowGap = numGap(row, baseRow);
			NUMBER colGap = numGap(col, baseCol);

			NUMBER newGap = numMax(rowGap, colGap);

			/*
			 * If the gap is smaller than the old one,
			 * then remember it and squeeze down the area.
			 */
			if (numIsLess(newGap, gap))
			{
				gap = newGap;
				minRow = numInc(numSub(baseRow, gap));
				maxRow = numDec(numAdd(baseRow, gap));
				minCol = numInc(numSub(baseCol, gap));
				maxCol = numDec(numAdd(baseCol, gap));
			}
		}
	}

	/*
	 * Return the new gap value.
	 */
	return gap;
}


/*
 * Find the closest selected cell to the specified coordinate and return
 * its coordinates if found.  The measurement is the taxicab distance.
 * Returns TRUE if there is such a cell.
 */
BOOL
lcFindClosestSelectedCell(const LifeObject * obj,
	NUMBER baseRow, NUMBER baseCol, NUMBER * retRow, NUMBER * retCol)
{
	const LifeRow *	rp;
	const LifeCell * cp;
	NUMBER newMaxGap;
	NUMBER newMinGap;

	NUMBER bestRow = numZero();
	NUMBER bestCol = numZero();
	NUMBER bestMaxGap = numZero();
	NUMBER bestMinGap = numZero();
	BOOL isBestFound = FALSE;

	/*
	 * Loop over all rows of the object.
	 */
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = rp->row;

		/*
		 * Loop over all of the cells of the row.
		 */
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER col = cp->col;

			/*
			 * Ignore cells which are not marked.
			 */
			if ((cp->marks & MARK_SELECT) == 0)
				continue;

			/*
			 * Calculate the gap from the base cell to this one.
			 * Sort it based on the maximum and minimum gaps.
			 */
			NUMBER rowGap = numGap(row, baseRow);
			NUMBER colGap = numGap(col, baseCol);

			if (numIsGreaterOrEqual(rowGap, colGap))
			{
				newMaxGap = rowGap;
				newMinGap = colGap;
			}
			else
			{
				newMaxGap = colGap;
				newMinGap = rowGap;
			}

			/*
			 * If we don't have a cell yet or this gap is smaller
			 * then remember this cell as the desired one.
			 */
			if (!isBestFound ||
				numIsLess(newMaxGap, bestMaxGap) ||
				((numIsEqual(newMaxGap, bestMaxGap) &&
					numIsLess(newMinGap, bestMinGap))))
			{
				bestMaxGap = newMaxGap;
				bestMinGap = newMinGap;
				bestRow = row;
				bestCol = col;
				isBestFound = TRUE;
			}
		}
	}

	/*
	 * Return the coordinate of the found cell and whether there was one.
	 */
	*retRow = bestRow;
	*retCol = bestCol;

	return isBestFound;
}

/* END CODE */
