/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * List-specific routines for managing paths.
 */

#include "life.h"
#include "listalg.h"


/*
 * Select all cells which are contained within the path for the object.
 * Returns FALSE if there are no cells within the path.
 */
BOOL
lcSelectPath(LifeObject * obj)
{
	Path * path = &obj->path;
	LifeRow * rp;
	LifeCell * cp;

	if (path->count <= 0)
		return FALSE;

	BOOL isMarked = FALSE;

	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		NUMBER row = rp->row;

		if (numIsLess(row, path->minRow))
			continue;

		if (numIsGreater(row, path->maxRow))
			return isMarked;

		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			NUMBER col = cp->col;

			if (numIsLess(col, path->minCol))
				continue;

			if (numIsGreater(col, path->maxCol))
				break;

			if (PointPathSignum(path, rp->row, cp->col) <= 0)
			{
				cp->marks |= MARK_SELECT;
				isMarked = TRUE;
			}
		}
	}

	return isMarked;
}


/*
 * Return the number of cells which are truly inside, exactly on the
 * boundary of, or truly outside of the current path of the specified object.
 * If there is no current path defined, then all returned values are zero.
 * Each cell of the object contributes to exactly one of these counts.
 * The values are returned through the indicated pointers.  If a pointer
 * is NULL, that value is not returned.  Multiple pointer values can specify
 * the same return variable, which will sum the indicated counts into that
 * variable.
 */
void
lcCountPathCells(const LifeObject * obj, NUMBER * insideCount,
	NUMBER * boundaryCount, NUMBER * outsideCount)
{
	const Path * path;
	const LifeRow *	rp;
	const LifeCell * cp;
	NUMBER dummyCount;
	int sigNum;

	/*
	 * Default the pointer values to a dummy count value if necessary.
	 */
	if (insideCount == 0)
		insideCount = &dummyCount;

	if (boundaryCount == 0)
		boundaryCount = &dummyCount;

	if (outsideCount == 0)
		outsideCount = &dummyCount;

	/*
	 * Clear the counter values.
	 */
	*insideCount = numZero();
	*boundaryCount = numZero();
	*outsideCount = numZero();

	/*
	 * If there is no path, then return now.
	 */
	path = &obj->path;

	if (path->count <= 0)
		return;

	/*
	 * There is a path, examine each row of the object and increment
	 * exactly one of the count values for each live cell.
	 */
	for (rp = obj->firstRow; rp != Life.termRow; rp = rp->nextRow)
	{
		for (cp = rp->firstCell; cp != Life.termCell; cp = cp->nextCell)
		{
			if (numIsLess(rp->row, path->minRow) ||
				numIsGreater(rp->row, path->maxRow) ||
				numIsLess(cp->col, path->minCol) ||
				numIsGreater(cp->col, path->maxCol))
			{
				numAddrInc(outsideCount);

				continue;
			}

			sigNum = PointPathSignum(path, rp->row, cp->col);

			if (sigNum < 0)
				numAddrInc(insideCount);
			else if (sigNum > 0)
				numAddrInc(outsideCount);
			else
				numAddrInc(boundaryCount);
		}
	}
}


