/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Program to play Conway's Game of LIFE on an infinite board.
 */

#define	VERSION	"10.10"
#define	AUTHOR	"David I. Bell"
#define	EMAIL	"dbell@tip.net.au"


#define	DEFINE_GLOBALS

#include <time.h>
#include <unistd.h>
#include <string.h>

#include "verify.h"
#include "life.h"
#include "listalg.h"
#include "quadalg.h"


/*
 * Maximum number of init files.
 */
#define	MAXINITS	10
#define	MAXLOADS	1000

static	BOOL		isInited;		/* initialization is complete */
static	BOOL		isCleanupNeeded;	/* need cleanup after init file read */
static	int		initfileCount;		/* number of init files to do */
static	int		initfileDone;		/* number of init files done */
static	int		loadFileCount;		/* number of files to load */
static	const char *	initFiles[MAXINITS];	/* user's init files */
static	const char *	loadFiles[MAXLOADS];	/* files to be loaded on startup */


/*
 * Local procedures.
 */
static	void	CreateReservedObjects(void);
static	void	CheckLifeOpts(void);
static	void	ParseOptions(int, const char **);
static	void	InterruptHandler(int);


int
main(int argc, const char ** argv)
{
	LifeDevice * dev;
	LifeWindow * window;

	srandom(time(0));

	numInit();
	qcInit();
	lcInit();

	Life.defaultAlgorithm = &quadAlgorithm;
	Life.moveMode = MOVE_MODE_MOVE;
	Life.fontName = DEFAULT_FONT;
	Life.settingsCount = 0;
	Life.defaultFrequency = 1;
	Life.defaultIsJumpingEnabled = FALSE;
	Life.defaultSharedSettings = SETTINGS_MASK_NONE;
	Life.topWindowSizePercentage = TOP_LEVEL_SIZE_PERCENTAGE;
	Life.subWindowSizePercentage = SUB_WINDOW_SIZE_PERCENTAGE;

#ifdef	X11
	Life.device = GetX11Device();
#else
	Life.device = GetTtyDevice();
#endif

	Life.defaultRule = ParseRule(DEFAULT_RULE);

	SetTopInputLevel();

	CheckLifeOpts();
	ParseOptions(argc - 1, argv + 1);

	/*
	 * Allocate the black and white pixel index values.
	 */
	if ((AddColorForDevice("black") != COLOR_INDEX_BLACK) ||
		(AddColorForDevice("white") != COLOR_INDEX_WHITE))
	{
		fprintf(stderr, "Failed to allocate initial colors\n");

		exit(1);
	}

	/*
	 * Try to open the device.
	 */
	dev = Life.device;

	if (!dev->openDevice(dev))
		exit(1);

	/*
	 * Allocate the default settings for the top level window
	 * and the subwindow window types.
	 */
	Life.topLevelWindowSettings = AllocSettings(".top", TRUE);
	Life.subWindowSettings = AllocSettings(".sub", FALSE);

	if ((Life.topLevelWindowSettings == 0) ||
		(Life.subWindowSettings == 0))
	{
		fprintf(stderr, "Cannot create initial settings\n");

		exit(1);
	}

	/*
	 * If an initial scale factor was set on the command line
	 * then make that the default top level scale value.
	 */
	if (Life.initScale)
		Life.topLevelWindowSettings->defaultScale = Life.initScale;

	/*
	 * Now create all of the reserved objects so that they
	 * will use the default scale factor.
	 */
	CreateReservedObjects();

	/*
	 * Try to open the initial window.
	 */
	window = dev->openWindow(dev, 0);

	if (window == 0)
	{
		fprintf(stderr, "Cannot create initial window\n");

		exit(1);
	}

	/*
	 * Make the initial window be the current one.
	 */
	Life.currentWindow = window;

	signal(SIGINT, InterruptHandler);
	ScanInit(ReadChar, &Life.ttyJmp);

	//
	// Register our garbage collection checker routine so that the
	// internal quadcell routines can use it.
	//
	qcSetCallback(CheckGarbageCollect);

	/*
	 * The main loop of the Life program.
	 * Look for commands, run generations for all objects which
	 * want to run them, update the view, do garbage collection,
	 * and possibly delay for a while.
	 */
	isInited = TRUE;

	while (TRUE)
	{
		DoCommands();
		DoGenerationsForAll();
		UpdateAllWindows();
		CheckGarbageCollect();
		dev->delay(dev, Life.delayTime);
	}
}


/*
 * Create all of the reserved objects.
 */
static void
CreateReservedObjects(void)
{
	int	index;
	char backupName[32];

	/*
	 * Set the flag which allows creation of special objects.
	 */
	Life.isCreatingReserved = TRUE;

	Life.deleteObject = GetObject("..delete");
	Life.tempObject = GetObject("..temp");
	Life.readObject = GetObject("..read");
	Life.mainObject = GetObject("main");
	Life.currentObject = Life.mainObject;

	/*
	 * Create all of the backup objects.
	 */
	for (index = 0; index < MAX_BACKUPS; index++)
	{
		sprintf(backupName, "..backup-%d", index + 1);
		Life.backupObjects[index] = GetObject(backupName);
	}

	/*
	 * No more special objects.
	 */
	Life.isCreatingReserved = FALSE;
}


/*
 * Parse the optional LIFE_OPTS environment variable to define defaults.
 */
static void
CheckLifeOpts(void)
{
	const char * env;
	char * cp;
	int argc;
	const char ** argv;
	static const char * argTable[MAX_ARGS];

	env = getenv(LIFE_OPTS);

	if (env == 0)
		return;

	cp = strdup(env);

	if (cp == 0)
	{
		fprintf(stderr, "No memory\n");
		exit(1);
	}

	argc = 0;
	argv = argTable;

	while (*cp)
	{
		while (isBlank(*cp))
			cp++;

		if (*cp == '\0')
			break;

		if (++argc > MAX_ARGS)
		{ 
			fprintf(stderr, "Too many arguments in LIFE_OPTS\n");
			exit(1);
		}

		*argv++ = cp;

		while (*cp && !isBlank(*cp))
			cp++;

		if (*cp)
			*cp++ = '\0';
	}

	ParseOptions(argc, argTable);
}


/*
 * Handle command line options.  Note: The first argument argv[0] is NOT
 * the usual program name, so that must have been removed if necessary.
 * We assume that the argument strings passed are permanent.
 */
static void
ParseOptions(int argc, const char ** argv)
{
	const char * str;
	const char * cp;
	Value arg;
	BOOL isNegative;

	while (argc-- > 0)
	{
		str = *argv++;

		/*
		 * If no option letter is used, then this means
		 * read from the specified file.
		 */
		if (*str != '-')
		{
			if (loadFileCount >= MAXLOADS)
			{
				fprintf(stderr, "Too many files to load\n");
				exit(1);
			}

			loadFiles[loadFileCount++] = str;

			continue;
		}

		/*
		 * This is an option argument.  Parse each character of
		 * it, so that multiple options can be given at once.
		 */
		str++;

		while (*str) switch (*str++)
		{
			case 'a':
				/*
				 * Set algorithm to use by default.
				 */
				if (argc <= 0)
				{
					fprintf(stderr, "Missing algorithm\n");
					exit(1);
				}

				Life.defaultAlgorithm = GetAlgorithmByName(*argv);

				if (Life.defaultAlgorithm == 0)
				{
					fprintf(stderr, "Unknown algorithm\n");
					exit(1);
				}

				argv++;
				argc--;

				break;

			case 't':
				/*
				 * Use terminal
				 */
				Life.device = GetTtyDevice();

				break;

			case 'g':
				/*
				 * Use graphics
				 */
#if defined(X11)
				Life.device = GetX11Device();
#else
				fprintf(stderr, "Graphics is not available\n");
				exit(1);
#endif

				break;

			case 'f':
				/*
				 * Set font name for graphics.
				 * We can't prohibit a dash on the next argument
				 * since some font names begin with a dash.
				 */
				if (argc <= 0)
				{
					fprintf(stderr, "Missing font name\n");
					exit(1);
				}

				Life.fontName = *argv++;
				argc--;

				break;

			case 'l':
				/*
				 * Set library path
				 */
				if ((argc <= 0) || (**argv == '-'))
				{
					fprintf(stderr, "Missing library path\n");
					exit(1);
				}

				if (Life.userLibCount >= MAX_LIBS)
				{
					fprintf(stderr, "Too many library directories\n");
					exit(1);
				}
	
				Life.userLibs[Life.userLibCount++] = *argv++;
				argc--;

				break;

			case 's':
				/*
				 * Set default scale
				 */
				if (argc <= 0)
				{
					fprintf(stderr, "Missing scale\n");
					exit(1);
				}

				cp = *argv++;
				argc--;
				isNegative = FALSE;

				if (*cp == '-')
				{
					isNegative = TRUE;
					cp++;
				}

				arg = 0;

				while (isDigit(*cp))
					arg = arg * 10 + *cp++ - '0';

				if (*cp || (arg == 0))
				{
					fprintf(stderr, "Bad scale factor\n");
					exit(1);
				}

				if (isNegative)
					arg = -arg;

				Life.initScale = arg;

				break;

			case 'v':
				printf("Version: %s\n", VERSION);
				printf("Author:  %s\n", AUTHOR);
				printf("Email:   %s\n", EMAIL);
				exit(0);

			case 'i':
				/*
				 * Read initialization file
				 */
				if (argc-- <= 0)
				{
					fprintf(stderr, "Missing file name\n");
					exit(1);
				}

				if (initfileCount >= MAXINITS)
				{
					fprintf(stderr, "Too many init files\n");
					exit(1);
				}

				initFiles[initfileCount++] = *argv++;

				break;

			default:
				fprintf(stderr, "Unknown option -%c\n", str[-1]);
				exit(1);
		}
	}
}


/*
 * Here to check on the startup files.
 * This is called on every command, so it must eventually do nothing.
 * If the top level input level is reached, and there is another startup
 * file to process, then we set up to take input from it.
 */
void
CheckStartupFiles(void)
{
	int temp;

	/*
	 * If not at the top level input level, then don't start
	 * reading another file.
	 */
	if (!IsTopInputLevel())
		return;

	/*
	 * See if we have to clean up after an init file is finished.
	 * This just makes sure that we are on the main object and that
	 * it is empty.
	 */
	if (isCleanupNeeded)
	{
		isCleanupNeeded = FALSE;

		ZeroObject(Life.mainObject);
		ZeroObject(Life.deleteObject);

		SetWindowObject(Life.currentWindow, Life.mainObject);
	}

	/*
	 * Then look for another initialization file that hasn't been done.
	 */
	if (initfileDone < initfileCount)
	{
		if (!SetFile(initFiles[initfileDone++]))
			Error("Cannot read initialization file");

		isCleanupNeeded = TRUE;

		return;
	}

	/*
	 * Finally load the files specified on the command line into buffers.
	 * The first file name in the list becomes the current object.
	 * Only try to do this once.
	 */
	if (loadFileCount == 0)
		return;

	temp = loadFileCount;
	loadFileCount = 0;

	LoadFiles(loadFiles, temp);
}


/*
 * Here on an interrupt character.  Remember to stop what we are doing soon.
 * We cannot usually just longjmp away since things may be in an inconsistent
 * state.  The longjmp is only allowed if we are sitting in a terminal read.
 */
static void
InterruptHandler(int arg)
{
	signal(SIGINT, InterruptHandler);

	Life.currentObject->genLeft = 0;
	Life.stop = TRUE;
	AddWindowObjectUpdateFlags(0, U_REDRAW);

	/*
	 * Do a longjmp only if it has been set up.
	 */
	if (Life.intJmpOk)
	{
		Life.intJmpOk = FALSE;
		longjmp(Life.intJmp.buf, 1);
	}
}


/*
 * Here on an error.  Close all but the top input level, cancel the
 * current command, beep, and set up to display the indicated message.
 * The message will remain until the next command is typed by the user.
 */
void
Error(const char * str)
{
	LifeDevice * dev = Life.device;

	if (!isInited)
	{
		fprintf(stderr, "%s\n", str);
		exit(1);
	}

	EndAllInputLevels();

	Life.stop = FALSE;
	Life.doWait = FALSE;

	dev->setStatusMode(dev, STATUS_MODE_ERROR);
	dev->addStatusText(dev, str);
	dev->setViewModeIsText(dev, FALSE);

	Beep();

	ScanAbort();
}


/*
 * Show a message in the status window when convenient.
 * Only the last such message will be visible.
 * The message will remain until the next command is typed by the user.
 */
void
Message(const char * str)
{
	LifeDevice * dev = Life.device;

	if (!isInited)
	{
		printf("%s\n", str);

		return;
	}

	dev->setStatusMode(dev, STATUS_MODE_MESSAGE);
	dev->addStatusText(dev, str);
}


/*
 * Routine called when a verification failed.
 * These are program bugs and are treated as fatal errors.
 */
void
VerifyFailed(const char * fileName, int lineNumber)
{
	fprintf(stderr, "Verification failed, file %s, line %d\n",
		fileName, lineNumber);

	exit(1);
}

/* END CODE */
