/*
 * Copyright (c) 2015 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Implementation of numeric routines.
 */
#include <stdio.h>
#include <stdlib.h>

#include "number.h"
#include "verify.h"


//
// The number of NUMBER structures to allocate at one time.
// This should be a power of two.
//
#define	NUMBER_ALLOC_SIZE	(1024 * 16)
#define	MAX_NUMBER_CHUNKS	100000


typedef	struct
{
	NUMBER table[NUMBER_ALLOC_SIZE];
} NumberChunk;


/*
 * The table of all the numbers which have ever been allocated.
 * These are accessed using their NUMINDEX values using the divide and
 * modulo operations by the NUMBER_ALLOC_SIZE value.
 * The numNextIndex value indicates the first index in the table
 * which is not yet used.  We can release almost all of the numbers
 * at once.
 */
NUMINDEX numNextIndex;

NumberChunk * numberTable[MAX_NUMBER_CHUNKS];


/*
 * Return the pointer to a NUMBER given its NUMINDEX value.
 */
inline static NUMBER *
indexToNumber(NUMINDEX numIndex)
{
	NumberChunk * chunk;

	verify((numIndex >= 0) && (numIndex < numNextIndex));

	chunk = numberTable[numIndex / NUMBER_ALLOC_SIZE];

	return &chunk->table[numIndex % NUMBER_ALLOC_SIZE];
}


/*
 * Local routines.
 */
static void numFatal(const char * str);


/*
 * Initialise the number routines.
 */
void
numInit(void)
{
	int index;

	//
	// Clear the number table to nulls.
	//
	for (index = 0; index < MAX_NUMBER_CHUNKS; index++)
		numberTable[index] = 0;

	//
	// Allocate the special numbers.
	// The order of these is important since their indexes
	// are predefined.
	//
	numNextIndex = 0;

	numAllocIndex(numFromLong(0));
	numAllocIndex(numFromLong(1));
	numAllocIndex(numFromLong(-1));

	//
	// Now make sure we are at the correct index for further numbers.
	//
	verify(numNextIndex == INDEX_FIRST_ALLOC_NUMBER);
}


/*
 * Allocate a new index for the specified number and store it into
 * our table.  This does not prevent storing of duplicate values
 * except for a few small numbers which always return the same index.
 */
NUMINDEX
numAllocIndex(NUMBER num)
{
	//
	// Check for special numbers, but only after they are allocated.
	//
	if (numIsZero(num) && (numNextIndex >= INDEX_FIRST_ALLOC_NUMBER))
		return INDEX_ZERO_NUMBER;

	if (numIsOne(num) && (numNextIndex >= INDEX_FIRST_ALLOC_NUMBER))
		return INDEX_ONE_NUMBER;

	if (numIsNegOne(num) && (numNextIndex >= INDEX_FIRST_ALLOC_NUMBER))
		return INDEX_NEG_ONE_NUMBER;

	//
	// Allocate a new index and store the number into it.
	//
	NUMINDEX numIndex = numAllocOneIndex();

	numStoreToIndex(numIndex, num);

	return numIndex;
}


/*
 * Allocate a number of consecutive indexes which can later be used to
 * store numbers into.  The initial numbers stored for the allocations
 * have the value which is the most negative number.  The indexes are
 * guaranteed to be new on each call (until all the numbers are freed),
 * and there is no handling of special values.  The returned index is
 * the one for the first allocated number.  The indexes for the remaining
 * allocated numbers have indexes which are in sequential order from
 * the first one.  There is a limit of 10 on the allowed count.
 */
NUMINDEX
numAllocMultipleIndexes(int count)
{
	verifyAlways((count > 0) && (count < 10));

	NUMINDEX numIndex = numAllocOneIndex();

	while (--count > 0)
		numAllocOneIndex();

	return numIndex;
}


/*
 * Allocate a new index for a number which can be stored into later.
 * The initial value of the allocated number is the most negative number.
 */
NUMINDEX
numAllocOneIndex(void)
{
	/*
	 * If we are at a chunk boundary then possibly allocate
	 * another chunk and store it into the number table.
	 * (After the numberNextIndex has been reset, all chunks
	 * previously allocated will still be present.)
	 */
	if ((numNextIndex % NUMBER_ALLOC_SIZE) == 0)
	{
		int tableIndex = numNextIndex / NUMBER_ALLOC_SIZE;

		verify(tableIndex < MAX_NUMBER_CHUNKS);

		NumberChunk * chunk = numberTable[tableIndex];

		if (chunk == 0)
			chunk = malloc(sizeof(NumberChunk));

		if (chunk == 0)
			numFatal("Memory allocation failed");

		numberTable[tableIndex] = chunk;
	}

	/*
	 * Get the address in the table for the next number index and
	 * store a placeholder number into the table.  Note: We increment
	 * the numNextIndex value first so that the indexToNumber call
	 * will not fail its verification.
	 */
	NUMINDEX numIndex = numNextIndex++;

	numStoreToIndex(numIndex, numMinValue());

	return numIndex;
}


/*
 * Return the number corresponding to the specified index.
 */
NUMBER
numFromIndex(NUMINDEX index)
{
	const NUMBER * numPtr = indexToNumber(index);

	return *numPtr;
}


/*
 * Store a number into the table of allocated numbers at
 * the specified index.
 */
void
numStoreToIndex(NUMINDEX index, NUMBER num)
{
	NUMBER * numPtr = indexToNumber(index);

	*numPtr = num;
}


/*
 * Free all of the indexes for the numbers except for the
 * pre-allocated ones.  The chunks that were previously allocated
 * will be used again.
 */
void
numFreeIndexes(void)
{
	numNextIndex = INDEX_FIRST_ALLOC_NUMBER;
}


/*
 * Return the integer part of the log base 2 of the number.
 * This is an indication of the position of the highest bit.
 * This is approximately the inverse of the numTwoPower function.
 * For example, numTwoLog(1) is 0 and numTwoLog(18) is 4.
 * The argument must be a positive value.
 */
int
numTwoLog(NUMBER num)
{
	if (numIsNonPos(num))
		numFatal("numTwoLog for non-positive number");

	int result = 0;

	while (num.value > 1)
	{
		num.value = (num.value >> 1);
		result++;
	}

	return result;
}


/*
 * Return the lowest set bit number in the specified number.
 * For example, the lowest bit number in 80 (2^6 + 2^4) is 4.
 * Returns 0 if the number is 0.
 */
int
numLowBitNumber(NUMBER num)
{
	long value = num.value;
	int bitNumber = 0;

	while (value && ((value & 0x1) == 0))
	{
		bitNumber++;
		value >>= 1;
	}

	return bitNumber;
}


/*
 * Return the result of one number raised to the power of another.
 * This can result in huge numbers, so we have a reasonable limit.
 * If we reach the limit then zero is returned, which can be checked
 * for after the call.  (The only way zero is legitimate is if the
 * input number is zero or if the power is negative.)
 */
NUMBER
numPower(NUMBER num1, NUMBER num2)
{
	if ((num1.value == 0) || (num1.value == 1))
		return num1;

	if (num1.value == -1)
	{
		if ((num2.value & 0x01) == 0)
			num1.value = 1;

		return num1;
	}

	if (num2.value < 0)
		return numZero();

	if ((num1.value < 0) && ((num2.value & 0x01) == 0))
		num1.value = -num1.value;

	//
	// Calculate the power while checking for overflow.
	// We do this by checking for a negative result or one which
	// doesn't divide back correctly.
	//
	long oldResult = 1;
	long result = 1;

	while (num2.value-- > 0)
	{
		result = result * num1.value;

		if ((result < 0) || (result / num1.value != oldResult))
			return numZero();

		oldResult = result;
	}

	num1.value = result;

	return num1;
}


/*
 * Return the string representation of a numeric value.
 * The returned string cannot be freed by the caller.
 * The string remains valid until ten more calls to this routine after
 * which it is overwritten.  This allows up to ten numbers to be used
 * within one printf statement, for example.
 */
const char *
numToString(NUMBER num)
{
	static int index;
	static char buffer[10][MAX_NUM_DIGITS+2];

	if (++index >= 10)
		index = 0;

	sprintf(&buffer[index][0], "%ld", num.value);

	return &buffer[index][0];
}


/*
 * Handler for a numeric error.
 */
static void
numFatal(const char * str)
{
	fprintf(stderr, "Numeric error: %s\n", str);
	exit(1);
}


