/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Include file for number definition and external routines.
 */
#ifndef	NUMBER_H
#define	NUMBER_H

/*
 * Booleans
 */
#ifndef BOOL
#define BOOL int
#endif

#ifndef	FALSE
#define	FALSE	((BOOL) 0)
#endif

#ifndef	TRUE
#define	TRUE	((BOOL) 1)
#endif


/*
 * The number of digits in the longest number value (including its sign):
 * 	-9,223,372,036,854,775,808
 */
#define	MAX_NUM_DIGITS 20


/*
 * The numeric type.
 * This is a simple small value now but it can be changed to be a
 * multi-precision value without affecting the callers.
 */
typedef struct
{
	long value;
} NUMBER;


/*
 * Number index values.
 * This allows a shorter representation of a numeric value instead of an
 * actual pointer, and so saves memory (4 bytes instead of 8 on a
 * 64-bit machine).
 */
typedef int NUMINDEX;


#define	INDEX_NULL_NUMBER -1
#define	INDEX_ZERO_NUMBER 0
#define	INDEX_ONE_NUMBER 1
#define	INDEX_NEG_ONE_NUMBER 2
#define	INDEX_FIRST_ALLOC_NUMBER 3

#define	isNullNumberIndex(index) ((index) <= INDEX_NULL_NUMBER)
#define	isNumberIndex(index) ((index) > INDEX_NULL_NUMBER)


/*
 * Cheap numeric routines using longs.
 * These look like procedures so that one day they could be easily
 * replaced with real multi-precision algorithms.
 */
inline static NUMBER
numAdd(NUMBER num1, NUMBER num2)
{
	num1.value += num2.value;

	return num1;
}


inline static NUMBER
numSub(NUMBER num1, NUMBER num2)
{
	num1.value -= num2.value;

	return num1;
}


inline static NUMBER
numMul(NUMBER num1, NUMBER num2)
{
	num1.value *= num2.value;

	return num1;
}


inline static NUMBER
numDiv(NUMBER num1, NUMBER num2)
{
	num1.value /= num2.value;

	return num1;
}


inline static NUMBER
numMod(NUMBER num1, NUMBER num2)
{
	num1.value %= num2.value;

	return num1;
}


inline static NUMBER
numOr(NUMBER num1, NUMBER num2)
{
	num1.value |= num2.value;

	return num1;
}


inline static NUMBER
numAnd(NUMBER num1, NUMBER num2)
{
	num1.value &= num2.value;

	return num1;
}


inline static NUMBER
numAddLong(NUMBER num1, long num2)
{
	num1.value += num2;

	return num1;
}


inline static NUMBER
numSubLong(NUMBER num1, long num2)
{
	num1.value -= num2;

	return num1;
}


inline static NUMBER
numMulLong(NUMBER num1, long num2)
{
	num1.value *= num2;

	return num1;
}


inline static NUMBER
numDivLong(NUMBER num1, long num2)
{
	num1.value /= num2;

	return num1;
}


inline static long
numModLong(NUMBER num1, long num2)
{
	return num1.value % num2;
}


inline static NUMBER
numOrLong(NUMBER num1, long num2)
{
	num1.value |= num2;

	return num1;
}


inline static NUMBER
numAndLong(NUMBER num1, long num2)
{
	num1.value &= num2;

	return num1;
}


inline static void
numAddrMul(NUMBER * num1, NUMBER num2)
{
	num1->value *= num2.value;
}


inline static void
numAddrMulLong(NUMBER * num1, long num2)
{
	num1->value *= num2;
}


inline static void
numAddrAdd(NUMBER * num1, NUMBER num2)
{
	num1->value += num2.value;
}


inline static void
numAddrSub(NUMBER * num1, NUMBER num2)
{
	num1->value -= num2.value;
}


inline static void
numAddrAddLong(NUMBER * num1, long num2)
{
	num1->value += num2;
}


inline static void
numAddrSubLong(NUMBER * num1, long num2)
{
	num1->value -= num2;
}


inline static void
numAddrInc(NUMBER * num)
{
	num->value++;
}


inline static void
numAddrDec(NUMBER * num)
{
	num->value--;
}


inline static BOOL
numIsEqual(NUMBER num1, NUMBER num2)
{
	return num1.value == num2.value;
}


inline static BOOL
numIsNotEqual(NUMBER num1, NUMBER num2)
{
	return num1.value != num2.value;
}


inline static BOOL
numIsLess(NUMBER num1, NUMBER num2)
{
	return num1.value < num2.value;
}


inline static BOOL
numIsGreater(NUMBER num1, NUMBER num2)
{
	return num1.value > num2.value;
}


inline static BOOL
numIsLessOrEqual(NUMBER num1, NUMBER num2)
{
	return num1.value <= num2.value;
}


inline static BOOL
numIsGreaterOrEqual(NUMBER num1, NUMBER num2)
{
	return num1.value >= num2.value;
}


inline static int
numCmp(NUMBER num1, NUMBER num2)
{
	if (num1.value < num2.value)
		return -1;

	if (num1.value > num2.value)
		return 1;

	return 0;
}


inline static int
numCmpLong(NUMBER num1, long num2)
{
	if (num1.value < num2)
		return -1;

	if (num1.value > num2)
		return 1;

	return 0;
}


inline static NUMBER
numMin(NUMBER num1, NUMBER num2)
{
	if (num1.value <= num2.value)
		return num1;

	return num2;
}


inline static NUMBER
numMax(NUMBER num1, NUMBER num2)
{
	if (num1.value >= num2.value)
		return num1;

	return num2;
}


inline static NUMBER
numAverage(NUMBER num1, NUMBER num2)
{
	num1.value = (num1.value + num2.value) / 2;

	return num1;
}


inline static NUMBER
numGap(NUMBER num1, NUMBER num2)
{
	if (num1.value >= num2.value)
		num1.value -= num2.value;
	else
		num1.value = num2.value - num1.value;

	return num1;
}


inline static NUMBER
numRange(NUMBER numFrom, NUMBER numTo)
{
	numFrom.value = numTo.value - numFrom.value + 1;

	return numFrom;
}


inline static int
numSign(NUMBER num)
{
	if (num.value < 0)
		return -1;

	if (num.value > 0)
		return 1;

	return 0;
}


inline static NUMBER
numTwoPower(int power)
{
	NUMBER result;

	result.value = 1L << power;

	return result;
}


inline static NUMBER
numNeg(NUMBER num)
{
	num.value = -num.value;

	return num;
}


inline static NUMBER
numInc(NUMBER num)
{
	num.value++;

	return num;
}


inline static NUMBER
numDec(NUMBER num)
{
	num.value--;

	return num;
}


inline static NUMBER
numAbs(NUMBER num)
{
	if (num.value < 0)
		num.value = -num.value;

	return num;
}


inline static NUMBER
numHalf(NUMBER num)
{
	num.value = num.value / 2;

	return num;
}


inline static NUMBER
numDouble(NUMBER num)
{
	num.value = num.value * 2;

	return num;
}


inline static BOOL
numIsLong(NUMBER num)
{
	return TRUE;
}


inline static BOOL
numIsZero(NUMBER num)
{
	return (num.value == 0);
}


inline static BOOL
numIsNonZero(NUMBER num)
{
	return (num.value != 0);
}


inline static BOOL
numIsNeg(NUMBER num)
{
	return (num.value < 0);
}


inline static BOOL
numIsPos(NUMBER num)
{
	return (num.value > 0);
}


inline static BOOL
numIsNonNeg(NUMBER num)
{
	return (num.value >= 0);
}


inline static BOOL
numIsNonPos(NUMBER num)
{
	return (num.value <= 0);
}


inline static BOOL
numIsOne(NUMBER num)
{
	return (num.value == 1);
}


inline static BOOL
numIsTwo(NUMBER num)
{
	return (num.value == 2);
}


inline static BOOL
numIsNegOne(NUMBER num)
{
	return (num.value == -1);
}


inline static BOOL
numIsEven(NUMBER num)
{
	return ((num.value & 0x1) == 0);
}


inline static BOOL
numIsOdd(NUMBER num)
{
	return ((num.value & 0x1) != 0);
}


inline static NUMBER
numFromLong(long val)
{
	NUMBER result;

	result.value = val;

	return result;
}


inline static NUMBER
numZero(void)
{
	NUMBER result;

	result.value = 0;

	return result;
}


inline static NUMBER
numOne(void)
{
	NUMBER result;

	result.value = 1;

	return result;
}


inline static NUMBER
numTwo(void)
{
	NUMBER result;

	result.value = 2;

	return result;
}


inline static NUMBER
numNegOne(void)
{
	NUMBER result;

	result.value = -1;

	return result;
}


inline static NUMBER
numMaxValue(void)
{
	NUMBER result;

	result.value = ((unsigned long)(-1L)) >> 1;

	return result;
}


inline static NUMBER
numMinValue(void)
{
	NUMBER result;

	result.value = (((unsigned long)(-1L)) >> 1) + 1;

	return result;
}


inline static BOOL
numIsMaxValue(NUMBER num)
{
	return (num.value == (((unsigned long)(-1L)) >> 1));
}


inline static BOOL
numIsMinValue(NUMBER num)
{
	return (num.value == (((unsigned long)(-1L)) >> 1) + 1);
}


inline static long
numToLong(NUMBER num)
{
	return num.value;
}


inline static void
numAddrDiv(NUMBER * num1, NUMBER num2)
{
	num1->value /= num2.value;
}


inline static void
numAddrAnd(NUMBER * num1, NUMBER num2)
{
	num1->value &= num2.value;
}


inline static void
numAddrOr(NUMBER * num1, NUMBER num2)
{
	num1->value |= num2.value;
}


inline static void
numAddrMod(NUMBER * num1, NUMBER num2)
{
	num1->value %= num2.value;
}


inline static void
numAddrDivLong(NUMBER * num1, long num2)
{
	num1->value /= num2;
}


inline static void
numAddrModLong(NUMBER * num1, long num2)
{
	num1->value %= num2;
}


inline static void
numAddrAndLong(NUMBER * num1, long num2)
{
	num1->value &= num2;
}


inline static void
numAddrOrLong(NUMBER * num1, long num2)
{
	num1->value |= num2;
}


extern void numInit(void);

extern int numTwoLog(NUMBER num);
extern int numLowBitNumber(NUMBER num);
extern NUMBER numPower(NUMBER num1, NUMBER num2);
extern NUMINDEX numAllocIndex(NUMBER num);
extern NUMINDEX numAllocOneIndex(void);
extern NUMINDEX numAllocMultipleIndexes(int count);
extern void numStoreToIndex(NUMINDEX index, NUMBER num);
extern NUMBER numFromIndex(NUMINDEX index);
extern void numFreeIndexes(void);

extern const char * numToString(NUMBER num);

#endif

