/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to perform functions on objects.
 * These are all generic and call algorithm-specific routines as needed.
 */

#include "life.h"
#include "algorithm.h"
#include "listalg.h"
#include "quadalg.h"


/*
 * Find the given named object.  Returns NULL if nonexistant.  The special
 * name of "." means the current object.  The special name of ".." means
 * the previous object.
 */
LifeObject *
FindObject(const char * str)
{
	LifeObject * obj;

	/*
	 * Check for the special names "." or "..".
	 */
	if (str[0] == '.')
	{
		if (str[1] == '\0')
			return Life.currentObject;

		if ((str[1] == '.') && (str[2] == '\0'))
			return Life.currentWindow->prevObject;
	}

	for (obj = Life.objects; obj; obj = obj->nextObject)
	{
		if (strcmp(GetStringBufferText(&obj->name), str) == 0)
			return obj;
	}

	return 0;
}


/*
 * Create the given named object, or return it if it already exists.
 * The special names of "." and ".." mean the current or previous object.
 */
LifeObject *
GetObject(const char * str)
{
	LifeObject * obj;

	obj = FindObject(str);

	if (obj)
		return obj;

	if (*str == '\0')
		Error("Empty object name");

	if (!Life.isCreatingReserved && BAD_NAME(str))
		Error("Cannot create reserved name");

	if (strchr(str, ' ') || strchr(str, '\t'))
		Error("Object name contains blanks");

	obj = AllocateObject();
	obj->nextObject = Life.objects;
	Life.objects = obj;

	SetStringBuffer(&obj->name, str);

	return obj;
}


/*
 * Get a new object having a unique name as close as possible to
 * the specified one.  Reserved names cannot be used.
 */
LifeObject *
GetUniqueObject(const char * str)
{
	if ((*str == '\0') || (*str == '.') ||
		strchr(str, ' ') || strchr(str, '\t'))
	{
		Error("Illegal base name");
	}

	/*
	 * See if the specified name is already unique.
	 * If so, use it as is.
	 */
	if (FindObject(str) == 0)
		return GetObject(str);

	/*
	 * Keep trying to generate a new object name composed of
	 * the specified name and an incrementing counter value.
	 */
	long counter = 1;
	StringBuffer newName;

	InitStringBuffer(&newName);

	do
	{
		char extra[24];

		sprintf(extra, "-%ld", ++counter);

		SetStringBuffer(&newName, str);
		AppendToStringBuffer(&newName, extra);
	}
	while (FindObject(GetStringBufferText(&newName)) != 0);

	LifeObject * obj = GetObject(GetStringBufferText(&newName));

	TermStringBuffer(&newName);

	return obj;
}


/*
 * Remove all the cells and marks of an object and clear some related values.
 */
void
ZeroObject(LifeObject * obj)
{
	ClearPath(obj);

	obj->algorithm->deleteAllCells(obj);

	obj->algorithm = Life.defaultAlgorithm;
	obj->isCountUnknown = FALSE;
	obj->gen = 0;
	obj->genLeft = 0;
	obj->count = numZero();
	obj->minCount = numZero();
	obj->maxCount = numZero();
	obj->dragMinRow = numZero();
	obj->dragMaxRow = numZero();
	obj->dragMinCol = numZero();
	obj->dragMaxCol = numZero();
	obj->mouseRow = numZero();
	obj->mouseCol = numZero();
	obj->currentInfo.row = numZero();
	obj->currentInfo.col = numZero();
	obj->originInfo.row = numZero();
	obj->originInfo.col = numZero();
	// ? Where locations
	IncrementChangeCount(obj);
}


/*
 * Destroy the existence of an object.  If it is the current object,
 * switch the current object back to the previous object.
 */
void
DestroyObject(LifeObject * obj)
{
	LifeObject * pobj;

	if (obj == 0)
		return;

	if (obj->isReserved)
		Error("Cannot destroy reserved object");

	WindowClearObject(obj);
	Life.currentObject = Life.currentWindow->currentObject;

	ZeroObject(obj);

	FreePath(obj);

	//
	// Free data for all algorithms.
	//
	listAlgorithm.freeCellData(obj);
	quadAlgorithm.freeCellData(obj);

	/*
	 * Handle the case that the object is first in the list.
	 */
	if (Life.objects == obj)
	{
		Life.objects = obj->nextObject;
		obj->nextObject = Life.freeObjects;
		Life.freeObjects = obj;

		return;
	}

	for (pobj = Life.objects; pobj->nextObject != obj; pobj = pobj->nextObject)
		;

	pobj->nextObject = obj->nextObject;
	obj->nextObject = Life.freeObjects;
	Life.freeObjects = obj;
}


/*
 * Move one object to another.  The source object is zeroed, and the
 * previous contents of the destination object are lost.  The algorithm
 * of the destination object is changed to match the source object.
 */
void
MoveObject(LifeObject * srcObj, LifeObject * destObj)
{
	if (srcObj == destObj)
		Error("Moving object to itself");

	ZeroObject(destObj);

	destObj->algorithm = srcObj->algorithm;

	srcObj->algorithm->moveObject(srcObj, destObj);

	destObj->rule = srcObj->rule;
	destObj->isCountUnknown = srcObj->isCountUnknown;
	destObj->count = srcObj->count;
	destObj->minCount = srcObj->minCount;
	destObj->maxCount = srcObj->maxCount;
	destObj->frequency = srcObj->frequency;
	destObj->isJumpingEnabled = srcObj->isJumpingEnabled;
	destObj->mouseRow = srcObj->mouseRow;
	destObj->mouseCol = srcObj->mouseCol;
	destObj->dragMinRow = srcObj->dragMinRow;
	destObj->dragMaxRow = srcObj->dragMaxRow;
	destObj->dragMinCol = srcObj->dragMinCol;
	destObj->dragMaxCol = srcObj->dragMaxCol;
	destObj->isTracksUsed = srcObj->isTracksUsed;
	destObj->isTracksVisible = srcObj->isTracksVisible;
	destObj->isLocationsEnabled = srcObj->isLocationsEnabled;
	destObj->stopIsEnabled = srcObj->stopIsEnabled;
	destObj->stopGap = srcObj->stopGap;
	destObj->stopOnPath = srcObj->stopOnPath;
	destObj->stopInsidePath = srcObj->stopInsidePath;
	destObj->stopOutsidePath = srcObj->stopOutsidePath;
	destObj->stopCell = srcObj->stopCell;
	destObj->stopState = srcObj->stopState;
	destObj->stopRow = srcObj->stopRow;
	destObj->stopCol = srcObj->stopCol;
	destObj->windowIsAutoScale = srcObj->windowIsAutoScale;
	destObj->windowScale = srcObj->windowScale;
	destObj->windowCurRow = srcObj->windowCurRow;
	destObj->windowCurCol = srcObj->windowCurCol;

	destObj->currentInfo = srcObj->currentInfo;
	destObj->originInfo = srcObj->originInfo;

	memcpy(destObj->whereInfo, srcObj->whereInfo, sizeof(destObj->whereInfo));

	memcpy(destObj->scaleList, srcObj->scaleList, sizeof(destObj->scaleList));
	destObj->scaleListCount = srcObj->scaleListCount;
	destObj->scaleListIndex = srcObj->scaleListIndex;

	SetStringBuffer(&destObj->comments, GetStringBufferText(&srcObj->comments));

	srcObj->count = numZero();

	strcpy(destObj->stopExpr, srcObj->stopExpr);

	CopyPath(srcObj, destObj);
	ClearPath(srcObj);

	IncrementChangeCount(srcObj);
	IncrementChangeCount(destObj);
}


/*
 * Add one object to another.  The source object is unchanged.
 * The destination object will get all cells from both objects.
 * All the added cells will be in the MARK state.
 * If addMode is ADD_RELATIVE, the object is displaced as specified
 * by the two object's cursor positions. Otherwise, the addition is
 * performed with absolute coordinates.  The source and destination
 * object must have the same algorithm.
 */
void
AddObject(const LifeObject * srcObj, LifeObject * destObj, AddObjectMode addMode)
{
	if (srcObj == destObj)
		Error("Adding object to itself");

	CheckSameAlgorithm(srcObj, destObj);

	NUMBER rowDisp = numZero();
	NUMBER colDisp = numZero();

	if (addMode == ADD_RELATIVE)
	{
		rowDisp = numSub(destObj->currentInfo.row, srcObj->currentInfo.row);
		colDisp = numSub(destObj->currentInfo.col, srcObj->currentInfo.col);
	}

	srcObj->algorithm->addObject(srcObj, destObj, rowDisp, colDisp);

	IncrementChangeCount(destObj);
}


/*
 * Copy one object to another.  The source object is unchanged.
 * The current contents of the destination object are lost.
 * The algorithm of the destination object is changed to match the
 * source object.
 */
void
CopyObject(const LifeObject * srcObj, LifeObject * destObj)
{
	if (srcObj == destObj)
		Error("Copying object to itself");

	//
	// If the algorithms are different then delete the cells of
	// the destination object and make it use the same algorithm
	// as the source object.
	//
	if (srcObj->algorithm != destObj->algorithm)
	{
		destObj->algorithm->deleteAllCells(destObj);
		destObj->algorithm = srcObj->algorithm;
	}

	//
	// Now copy the cells and marks from the source to the destination.
	//
	srcObj->algorithm->copyObject(srcObj, destObj);

	//
	// Copy the other fields too.
	//
	destObj->rule = srcObj->rule;
	destObj->isCountUnknown = srcObj->isCountUnknown;
	destObj->count = srcObj->count;
	destObj->minCount = srcObj->minCount;
	destObj->maxCount = srcObj->maxCount;
	destObj->gen = srcObj->gen;
	destObj->frequency = srcObj->frequency;
	destObj->isJumpingEnabled = srcObj->isJumpingEnabled;
	destObj->isLocationsEnabled = srcObj->isLocationsEnabled;
	destObj->mouseRow = srcObj->mouseRow;
	destObj->mouseCol = srcObj->mouseCol;
	destObj->dragMinRow = srcObj->dragMinRow;
	destObj->dragMaxRow = srcObj->dragMaxRow;
	destObj->dragMinCol = srcObj->dragMinCol;
	destObj->dragMaxCol = srcObj->dragMaxCol;
	destObj->isTracksUsed = srcObj->isTracksUsed;
	destObj->isTracksVisible = srcObj->isTracksVisible;
	destObj->stopIsEnabled = srcObj->stopIsEnabled;
	destObj->stopGap = srcObj->stopGap;
	destObj->stopOnPath = srcObj->stopOnPath;
	destObj->stopInsidePath = srcObj->stopInsidePath;
	destObj->stopOutsidePath = srcObj->stopOutsidePath;
	destObj->stopCell = srcObj->stopCell;
	destObj->stopState = srcObj->stopState;
	destObj->stopRow = srcObj->stopRow;
	destObj->stopCol = srcObj->stopCol;
	destObj->windowIsAutoScale = srcObj->windowIsAutoScale;
	destObj->windowScale = srcObj->windowScale;
	destObj->windowCurRow = srcObj->windowCurRow;
	destObj->windowCurCol = srcObj->windowCurCol;
	destObj->usedQuadCells = srcObj->usedQuadCells;

	strcpy(destObj->stopExpr, srcObj->stopExpr);

	CopyPath(srcObj, destObj);

	destObj->originInfo = srcObj->originInfo;
	destObj->currentInfo = srcObj->currentInfo;

	memcpy(destObj->whereInfo, srcObj->whereInfo, sizeof(destObj->whereInfo));

	memcpy(destObj->scaleList, srcObj->scaleList, sizeof(destObj->scaleList));
	destObj->scaleListCount = srcObj->scaleListCount;
	destObj->scaleListIndex = srcObj->scaleListIndex;

	SetStringBuffer(&destObj->comments, GetStringBufferText(&srcObj->comments));

	IncrementChangeCount(destObj);
}


/*
 * Center the cells of an object around the specified location.
 * This shifts the cells left or right and up or down so that the center
 * of the object is at the location.  This also modifies the marks, origin,
 * where locations, tracked cells, and path of the object to match.
 */
void
CenterObject(LifeObject * obj, NUMBER centerRow, NUMBER centerCol)
{
	NUMBER minRow;
	NUMBER maxRow;
	NUMBER minCol;
	NUMBER maxCol;

	if (!obj->algorithm->minMax(obj, &minRow, &maxRow, &minCol, &maxCol))
		Error("Object is empty");

	NUMBER averageRow = numAverage(minRow, maxRow);
	NUMBER averageCol = numAverage(minCol, maxCol);

	NUMBER leftOrRightShift = numSub(centerCol, averageCol);
	NUMBER upOrDownShift = numSub(centerRow, averageRow);

	ShiftObject(obj, leftOrRightShift, upOrDownShift);
}


/*
 * Shift the cells of the object left or right and up or down as specified.
 * This also modifies the marks, origin, where locations, tracked cells,
 * and path of the object to match.
 */
void
ShiftObject(LifeObject * obj, NUMBER leftOrRight, NUMBER upOrDown)
{
	int index;

	if (obj->algorithm->isObjectEmpty(obj))
		Error("Object is empty");

	if (numIsZero(leftOrRight) && numIsZero(upOrDown))
		return;

	//
	// Adjust the cells and marks.
	//
	obj->algorithm->shiftObject(obj, leftOrRight, upOrDown);

	//
	// Adjust the tracked cells here since this isn't done by the
	// algorithm specific code and is used by all algorithms.
	//
	obj->usedQuadCells = qcShift(obj->usedQuadCells, leftOrRight, upOrDown);

	//
	// Adjust the other object coordinates.
	//
	numAddrAdd(&obj->mouseRow, upOrDown);
	numAddrAdd(&obj->mouseCol, leftOrRight);
	numAddrAdd(&obj->windowCurRow, upOrDown);
	numAddrAdd(&obj->windowCurCol, leftOrRight);

	AdjustLocation(obj, &obj->originInfo, upOrDown, leftOrRight);
	AdjustLocation(obj, &obj->currentInfo, upOrDown, leftOrRight);

	for (index = 0; index < 26; index++)
	{
		AdjustLocation(obj, &obj->whereInfo[index],
			upOrDown, leftOrRight);
	}

	ShiftPath(obj, leftOrRight, upOrDown);

	RestoreCurrentViewParameters();	// WRONG: Only one window

	IncrementChangeCount(obj);
}


/*
 * Copy marked cells to another object. The previous cells of the destination
 * object are deleted, and the algorithm is changed to match the source
 * object.  The source object is unaffected.
 */
void
CopyMarkedObject(const LifeObject * srcObj, LifeObject * destObj)
{
	int i;

	if (srcObj == destObj)
		Error("Copying object to itself");

	ZeroObject(destObj);

	destObj->algorithm = srcObj->algorithm;
	destObj->rule = srcObj->rule;

	destObj->currentInfo = srcObj->currentInfo;
	destObj->originInfo = srcObj->originInfo;

	for (i = 0; i < 26; i++)
		destObj->whereInfo[i] = srcObj->whereInfo[i];

	srcObj->algorithm->copyMarkedObject(srcObj, destObj);

	IncrementChangeCount(destObj);
}


/*
 * Move the selected cells to another object.  If the destination object is NULL
 * the cells are just deleted.  The previous cells of the destination object
 * are deleted, and its algorithm is changed to match the source object.
 */
void
MoveSelection(LifeObject * srcObj, LifeObject * destObj)
{
	int i;

	if (srcObj == destObj)
		Error("Moving object to itself");

	if (destObj)
	{
		ZeroObject(destObj);

		destObj->algorithm = srcObj->algorithm;
		destObj->rule = srcObj->rule;

		destObj->currentInfo = srcObj->currentInfo;
		destObj->originInfo = srcObj->originInfo;

		for (i = 0; i < 26; i++)
			destObj->whereInfo[i] = srcObj->whereInfo[i];
	}

	srcObj->algorithm->moveSelection(srcObj, destObj);

	IncrementChangeCount(srcObj);
}


/*
 * Backup an object to a newly created object having a name which is the
 * concatentaion of the specified base name and the smallest positive
 * integer which results in a new object name.  If an empty base name
 * is given then the previous base name is used.
 */
void
BackupObject(const LifeObject * obj, const char * cp)
{
	static StringBuffer baseName;

	/*
	 * If a new base name is given then save it.
	 * Otherwise if there hasn't been one set default it to "b".
	 */
	if (*cp)
	{
		if (BAD_NAME(cp) || strchr(cp, ' ') || strchr(cp, '\t'))
			Error("Illegal base name");

		SetStringBuffer(&baseName, cp);
	}
	else
	{
		if (GetStringBufferLength(&baseName) == 0)
			SetStringBuffer(&baseName, "b");
	}

	/*
	 * Keep trying to generate a new object name composed of
	 * the base name and an incrementing counter value.
	 */
	long counter = 0;
	int len = GetStringBufferLength(&baseName);

	do
	{
		char extra[24];

		sprintf(extra, "-%ld", ++counter);

		SetStringBufferLength(&baseName, len);
		AppendToStringBuffer(&baseName, extra);
	}
	while (FindObject(GetStringBufferText(&baseName)) != 0);

	/*
	 * OK, copy the object to this new object.
	 */
	CopyObject(obj, GetObject(GetStringBufferText(&baseName)));

	/*
	 * Also copy the object to a constant name for easy reference.
	 */
	SetStringBufferLength(&baseName, len);
	AppendToStringBuffer(&baseName, "-last");

	CopyObject(obj, GetObject(GetStringBufferText(&baseName)));

	SetStringBufferLength(&baseName, len);
}


/*
 * Generic routine to return the pure state of a cell.
 * This is one of OFF, DEAD, ON, NEW, or MARK.
 */
int
GetCellState(const LifeObject * obj, NUMBER row, NUMBER col)
{
	return obj->algorithm->getCellState(obj, row, col);
}


/*
 * Genetic routine to set the pure state of a cell.
 * This is one of OFF, DEAD, ON, NEW, or MARK.
 * The specified cell can be made live or made dead as desired.
 */
void
SetCellState(LifeObject * obj, NUMBER row, NUMBER col, int state)
{
	verifyAlways(isStatePure(state));

	obj->algorithm->setCellState(obj, row, col, state);

	//
	// We don't know what kind of change was made.
	//
	obj->isCountUnknown = TRUE;
	IncrementChangeCount(obj);
}


/*
 * Generic routine to try to add a cell to an object in the ON state.
 * This does not change the state of a cell in the NEW or MARK state.
 * Returns TRUE if the cell was actually added, or FALSE if the cell
 * was already live.
 */
BOOL
AddCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	if (!obj->algorithm->addCell(obj, row, col))
		return FALSE;

	AdjustCellCount(obj, numOne());
	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Generic routine to delete a cell from an object.
 * Returns TRUE if the cell was actually deleted, or FALSE if the
 * cell was already dead.
 */
BOOL
DeleteCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	if (!obj->algorithm->deleteCell(obj, row, col))
		return FALSE;

	AdjustCellCount(obj, numNegOne());
	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Generic routine to rotate the previously selected cells in the given
 * object around the current cursor location.  This deletes the selected
 * cells, and reinserts them after their position has been rotated by 90
 * degrees clockwise.  The cells are inserted with the MARK state.
 */
void
RotateSelection(LifeObject * obj)
{
	obj->algorithm->rotateSelection(obj);

	IncrementChangeCount(obj);
}


/*
 * Flip the previously selected cells in the given object around the row
 * of the current cursor location.  This deletes the selected cells,
 * and reinserts them after their position has been flipped around the
 * horizontal axis.  The added cells are in the MARK state.
 */
void
FlipRowSelection(LifeObject * obj)
{
	obj->algorithm->flipRowSelection(obj);

	IncrementChangeCount(obj);
}


/*
 * Flip the previously selected cells in the given object around the
 * column of the current cursor location.  This deletes the selected cells,
 * and reinserts them after their position has been flipped around the
 * vertical axis.
 */
void
FlipColumnSelection(LifeObject * obj)
{
	obj->algorithm->flipColumnSelection(obj);

	IncrementChangeCount(obj);
}


/*
 * Find the closest selected cell to the specified coordinate and return
 * its coordinates if found.  The measurement is the taxicab distance.
 * Returns TRUE if there is such a cell.
 */
BOOL
FindClosestSelectedCell(const LifeObject * obj,
	NUMBER baseRow, NUMBER baseCol, NUMBER * retRow, NUMBER * retCol)
{
	return obj->algorithm->findClosestSelectedCell(obj,
		baseRow, baseCol, retRow, retCol);
}


/*
 * Select the cells in a specified rectangular region as desired.
 * Returns TRUE if the region contained at least one cell to be selected.
 */
BOOL
SelectRegion(LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	if (!obj->algorithm->selectRegion(obj, minRow, maxRow, minCol, maxCol))
	{
		return FALSE;
	}

	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Select the object at a given location as specified.  Returns FALSE if
 * no object exists there.  An object for this purpose is considered the
 * largest set of cells containing the specified cell whose members are
 * separated by a distance no larger then the specified gap.  For example,
 * a gap of zero means the cells which are king-wise connected to the
 * specified cell.
 */
BOOL
SelectObject(LifeObject * obj, NUMBER row, NUMBER col, Value gap)
{
	if (!obj->algorithm->selectObject(obj, row, col, gap))
		return FALSE;

	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Select the cell at the specified location.
 * Returns TRUE if there was a cell at the specified location.
 */
BOOL
SelectCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	if (!obj->algorithm->selectCell(obj, row, col))
		return FALSE;

	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Clear the tracks for the object.
 * This removes all of the DEAD and NEW states of the cells.
 */
void
ClearTracks(LifeObject * obj)
{
	obj->algorithm->clearTracks(obj);

	IncrementChangeCount(obj);
}


/*
 * Find the range of all marked cells for an object.
 * Returns FALSE if no cells were marked.
 */
BOOL
MarkMinMax(const LifeObject * obj,
	NUMBER * minRow, NUMBER * maxRow, NUMBER * minCol, NUMBER * maxCol)
{
	return obj->algorithm->markMinMax(obj, minRow, maxRow, minCol, maxCol);
}


/*
 * Select all cells which are contained within the path for the object.
 * Returns FALSE if there are no cells within the path.
 */
BOOL
SelectPath(LifeObject * obj)
{
	if (!obj->algorithm->selectPath(obj))
		return FALSE;

	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Select all of the cells of an object.
 * Returns TRUE if there were any cells to select.
 */
BOOL
SelectAll(LifeObject * obj)
{
	if (!obj->algorithm->selectAll(obj))
		return FALSE;

	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Mark the cells of an object for the previously made selection.
 * Returns FALSE if there were no selected cells.
 */
BOOL
MarkSelection(LifeObject * obj)
{
	if (!obj->algorithm->markSelection(obj))
		return FALSE;

	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Select the marked cells in an object.
 * Returns FALSE if there were no marked cells.
 */
BOOL
SelectMarkedCells(LifeObject * obj)
{
	if (!obj->algorithm->selectMarkedCells(obj))
		return FALSE;

//	IncrementChangeCount(obj);

	return TRUE;
}


/*
 * Clear marks for a whole object.
 */
void
ClearMarks(LifeObject * obj)
{
	obj->algorithm->clearMarks(obj);

	IncrementChangeCount(obj);
}


/*
 * Clear the search state for an object.
 */
void
ClearSearch(LifeObject * obj)
{
	obj->algorithm->clearSearch(obj);

	IncrementChangeCount(obj);
}


/*
 * Clear the current selection for a whole object.
 */
void
ClearSelection(LifeObject * obj)
{
	obj->algorithm->clearSelection(obj);

	IncrementChangeCount(obj);
}


/*
 * Invert the marks for a whole object.
 */
void
InvertMarks(LifeObject * obj)
{
	obj->algorithm->invertMarks(obj);

	IncrementChangeCount(obj);
}


/*
 * Count the number of marked cells in an object.
 */
NUMBER
CountMarks(const LifeObject * obj)
{
	return obj->algorithm->countMarks(obj);
}


/*
 * Find the minimum and maximum row and column numbers for an object.
 * This includes dead cells if tracking is enabled.
 * Returns FALSE if the object has no cells.
 */
BOOL
MinMax(const LifeObject * obj,
	NUMBER * retMinRow, NUMBER * retMaxRow,
	NUMBER * retMinCol, NUMBER * retMaxCol)
{
	return obj->algorithm->minMax(obj,
		retMinRow, retMaxRow, retMinCol, retMaxCol);
}


/*
 * Count the cells in a specified rectangular region.
 * Returns the number of cells in the region.
 */
NUMBER
CountRegion(const LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	return obj->algorithm->countRegion(obj, minRow, maxRow, minCol, maxCol);
}


/*
 * Generic routine to return whether or not the object has no ON cells.
 * This makes no use of the object's cell count value.
 */
BOOL
IsObjectEmpty(const LifeObject * obj)
{
	return obj->algorithm->isObjectEmpty(obj);
}


/*
 * Generic routine to determine whether a cell is ON in an object.
 */
BOOL
IsCellSet(const LifeObject * obj, NUMBER row, NUMBER col)
{
	return obj->algorithm->isCellSet(obj, row, col);
}


/*
 * Calculate a view of the specified rectangular region of the specified object
 * and save the result in a returned memory array whose size is the indicated
 * number of rows times the indicated number of columns, scaled down by the
 * specified factor.  The cells in the array contain values for square areas
 * of the object having sides of 'scale', and are 0 for all empty cells and 1
 * for one or more nonempty cells.  This also uses the auxillary data array to
 * store the state of the cells and how many cells are ON for each area.
 * There is a limit on the size of the rectangular region that can be used.
 * The returned array should not be freed by the caller and is valid until
 * the next call to this routine.
 */
ViewInfo
CalculateView(const LifeObject * obj, int viewId, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	return obj->algorithm->calculateView(obj, viewId, scale,
		minRow, maxRow, minCol, maxCol);
}


/*
 * Return the lowest row number in the object which is greater than
 * or equal to the specified row.  If no such row exists then
 * the maximum positive number value is returned.
 */
NUMBER
GetObjectMinRowDowntoLimit(LifeObject * obj, NUMBER limitRow)
{
	return obj->algorithm->getObjectMinRowDowntoLimit(obj, limitRow);
}


/*
 * Scan the current cells of an object, calling the indicated function
 * for each cell whose state is one of those specified, supplying a
 * specified piece of data that can be used for whatever purpose the
 * called routine wants.  The special states value of STATE_NONE means
 * treat NEW and MARK cells as ON, and return the ON cells.  If the
 * routine returns TRUE then the scan is continued, otherwise it is stopped.
 * This routine returns TRUE if the scan was completed, or FALSE if the
 * called function indicates that the scan is to be aborted.  The object
 * cannot be changed during the scan, neither can a garbage collection be
 * performed.  Warning: This should not be called for a very large object
 * since it can take a long time.
 */
BOOL
ScanObject(LifeObject * obj, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data)
{
	return obj->algorithm->scanObject(obj, states, func, data);
}


/*
 * Return the quadcell representation of an object.
 * Warning: This can perform garbage collection.
 */
QuadCell *
ExtractQuadCell(LifeObject * obj)
{
	return obj->algorithm->extractQuadCell(obj);
}


/*
 * Add all of the current cells of an object to its set of used cells.
 * This is used to display the cells which have ever been used by an
 * object during its evolution.  This uses the quadcell storage mechanism
 * since that can handle very large objects which have some regularity
 * in a relatively small space.  (This is bound to happen using the
 * tracks feature when a spaceship or puffer runs away.)
 * Warning: This can perform garbage collection.
 */
void
SaveUsedCells(LifeObject * obj)
{
	if (obj->isTracksUsed)
		obj->algorithm->saveUsedCells(obj);
}


/*
 * Try to compute a number of generations for the specified object.
 * This also updates the number of cells which are born and died overall
 * and the cell count.  Returns the number of generations actually run.
 * Warning: This can perform garbage collections.
 */
Count
RunGenerations(LifeObject * obj, Count genCount)
{
	return obj->algorithm->runGenerations(obj, genCount);
}


/*
 * Return the number of cells which are truly inside, exactly on the
 * boundary of, or truly outside of the current path of the specified object.
 * If there is no current path defined, then all returned values are zero.
 * Each cell of the object contributes to exactly one of these counts.
 * The values are returned through the indicated pointers.  If a pointer
 * is NULL, that value is not returned.  Multiple pointer values can specify
 * the same return variable, which will sum the indicated counts into that
 * variable.
 */
void
CountPathCells(const LifeObject * obj,
	NUMBER * insideCount, NUMBER * boundaryCount, NUMBER * outsideCount)
{
	obj->algorithm->countPathCells(obj,
		insideCount, boundaryCount, outsideCount);
}


/*
 * Find the minimum distance between any marked cell to any unmarked cell.
 * The distance is computed as the taxicab distance (maximum of x and y).
 * Returns 0 if there are no marked cells or no unmarked cells.
 * This is an expensive operation so that there is a limit on the size
 * of the gap.  Warning: This can perform garbage collection.
 */
NUMBER
FindMarkedGap(const LifeObject * obj)
{
	return obj->algorithm->findMarkedGap(obj);
}


/*
 * Increment the change counter and zero the born, died, and
 * growth counts for an object.
 */
void
IncrementChangeCount(LifeObject * obj)
{
	obj->changeCount++;
	obj->born = numZero();
	obj->died = numZero();
	obj->growth = numZero();
}


/*
 * Try to adjust the cell count for the specified object, either upwards
 * or downwards.  This also updates the minimum and maximum count values.
 * This does nothing if the count value for the object is unknown.
 */
void
AdjustCellCount(LifeObject * obj, NUMBER adjust)
{
	if (obj->isCountUnknown)
		return;

	SetCellCount(obj, numAdd(obj->count, adjust));
}


/*
 * Set the cell count for the specified object.
 * This also updates the minimum and maximum count values and
 * clears the isCountUnknown flag.  This should not be called
 * unless we know that the specified count is correct.
 */
void
SetCellCount(LifeObject * obj, NUMBER count)
{
	obj->isCountUnknown = FALSE;
	obj->count = count;

	obj->minCount = numMin(obj->minCount, count);
	obj->maxCount = numMax(obj->maxCount, count);
}


/*
 * Calculate the number of ON cells in the object if necessary,
 * and then return the value.  If the count is known to be correct
 * then no work is needed.  This also clears the isCountUnknown flag.
 */
NUMBER
CalculateCellCount(LifeObject * obj)
{
	if (obj->isCountUnknown)
	{
		SetCellCount(obj, obj->algorithm->calculateCellCount(obj));
	}

	return obj->count;
}


/*
 * Show information about the specified object.
 */
void
ShowObjectInfo(const LifeObject * obj)
{
	ShowHelp("Object \"");
	ShowHelp(GetStringBufferText(&obj->name));
	ShowHelp("\"\n\n");
	ShowHelp(GetStringBufferText(&obj->comments));

	EndHelp(TRUE);
}


/*
 * Show the list of objects.  If all is nonzero, all objects will be
 * shown.  Otherwise, only objects not starting with a period are shown.
 */
void
ListObjects(BOOL isAll)
{
	LifeObject * obj;
	NUMBER minRow;
	NUMBER maxRow;
	NUMBER minCol;
	NUMBER maxCol;

	ShowHelp("cells        height     width      gen       win  object\n");
	ShowHelp("-----        ------     -----      ---       ---  ------\n");

	for (obj = Life.objects; obj; obj = obj->nextObject)
	{
		const char * name = GetStringBufferText(&obj->name);

		if (!isAll && (*name == '.'))
			continue;

		int ch = ' ';

		if (obj == Life.currentWindow->prevObject)
			ch = '+';

		if (obj == Life.currentObject)
			ch = '*';

		if (!obj->algorithm->minMax(obj, &minRow, &maxRow, &minCol, &maxCol))
		{
			minRow = numZero();
			maxRow = numZero();
			minCol = numZero();
			maxCol = numZero();
		}

		int windowCount = GetObjectWindowCount(obj);

		FormatHelp("%-12s %-10s %-10s %-9ld %-4d %c %s\n",
			numToString(CalculateCellCount(obj)),
			numToString(numRange(minRow, maxRow)),
			numToString(numRange(minCol, maxCol)),
			obj->gen, windowCount, ch, name);
	}

	ShowHelp("\n\nIn object column, '*' = current object, '+' = previous object\n");

	if (!isAll)
		ShowHelp("Use -a to show objects beginning with '.'\n");

	EndHelp(TRUE);
}


/*
 * Check whether this is a good time for garbage collection of quad cells.
 * If so then a garbage collection is performed.  This can only be called
 * when no unsaved temporary quad cells are being used (e.g., ones being
 * held only by local variables).  We do garbage collection when the memory
 * usage of the quad cell data grows.  Note: This routine is registered as
 * a callback routine for the internal quadcell code to use when it needs to.
 */
void
CheckGarbageCollect(void)
{
	int used = qcGetPercentUsed();

	if (used > 10000)
		GarbageCollect();

	used = qcGetPercentUsed();

	if (used > 1000)
		qcGrowHashTable();
}


/*
 * Register all of the quad cells for the objects and then call the
 * garbage collection routine for the quad cells.  This can only be
 * called when no temporary quad cells are being used (e.g., ones
 * being used in a calculation).
 */
void
GarbageCollect(void)
{
	LifeObject * obj;

	for (obj = Life.objects; obj; obj = obj->nextObject)
	{
		//
		// Register all of the quadcells for the object.
		// NOTE: We do not use the algorithm pointer in the
		// object structure for the call.
		//
		quadAlgorithm.registerQuadCells(obj);
	}

	qcGarbageCollect();
}


/*
 * Search forwards for the nth next object, restarting at the top if necessary.
 * If all is nonzero, or if wrap around occurs, the search will be over all
 * objects.  Otherwise, objects found in previous searches will be skipped.
 * Returns FALSE if the object was empty.
 */
BOOL
SearchObject(LifeObject * obj, Count count, BOOL isAll)
{
	if (isAll)
		obj->algorithm->clearSearch(obj);

	if (obj->algorithm->isObjectEmpty(obj))
		return FALSE;

	return obj->algorithm->searchObject(obj, count);
}


/*
 * Check that the source and destination objects have the same algorithm.
 * If not then an error is generated.
 */
void
CheckSameAlgorithm(const LifeObject * srcObj, const LifeObject * destObj)
{
	if (srcObj->algorithm != destObj->algorithm)
		Error("Source and destination objects have different algorithms");
}


/*
 * Prepare the temporary object for use by clearing it and setting its
 * algorithm to match the one in the specified object.  This also checks
 * that the specified source object is not the temp object.
 */
void
PrepareTempObject(const LifeObject * srcObj)
{
	if (srcObj == Life.tempObject)
		Error("Using temp object");

	ZeroObject(Life.tempObject);

	Life.tempObject->algorithm = srcObj->algorithm;
}


/*
 * Check all locations to see if any of them affect the status line.
 * If so, then the window will be marked as needing an update if required.
 * This is called when changing the current position.
 */
void
CheckAllLocationStatus(const LifeObject * obj)
{
	int flags = 0;
	int index;

	if (!obj->isLocationsEnabled)
		return;

	//
	// Check the origin and current locations.
	//
	if (obj->originInfo.isEnabled && obj->originInfo.isDistShown)
		flags |= U_STATUS;

	if (obj->currentInfo.isEnabled && obj->currentInfo.isDistShown)
		flags |= U_STATUS;

	//
	// Check the where locations.
	//
	for (index = 0; index < 26; index++)
	{
		const LocationInfo * info = &obj->whereInfo[index];

		if (info->isEnabled && info->isDistShown)
			flags |= U_STATUS;
	}

	if (flags)
		AddWindowObjectUpdateFlags(obj, flags);
}


/*
 * Check the specified location to see if it is being shown.
 * If so, then the window will be marked as needing an update if required.
 * This is called in cases where the row and column is changed
 * outside of the SetLocation function.
 */
void
CheckLocation(const LifeObject * obj, const LocationInfo * info)
{
	int flags = 0;

	//
	// If showing is disabled then do nothing.
	//
	if (!obj->isLocationsEnabled || !info->isEnabled)
		return;

	//
	// Check those items which need full redraws.
	//
	if (info->isIdShown || info->isForwShown || info->isBackShown ||
		info->isHorizShown || info->isVertShown)
	{
		flags |= U_REDRAW;
	}

	//
	// Check the item which just changes the status line.
	//
	if (info->isDistShown)
		flags |= U_STATUS;

	//
	// If something is to be shown then set to update the view.
	//
	if (flags)
		AddWindowObjectUpdateFlags(obj, flags);
}


/*
 * Check the current location to see if it is being shown.
 * If so, then the window will be marked as needing an update if required.
 * This is called in cases where the current row and column is changed
 * outside of the SetLocation function.
 */
void
CheckCurrentLocation(const LifeObject * obj)
{
	CheckLocation(obj, &obj->currentInfo);
}


/*
 * Set the row and column values for a location in an object.
 * If the location is being shown, then the window will be marked as
 * needing an update if required.
 */
void
SetLocation(LifeObject * obj, LocationInfo * info, NUMBER row, NUMBER col)
{
	if (numIsEqual(info->row, row) && numIsEqual(info->col, col))
		return;

	info->row = row;
	info->col = col;

	//
	// Check what this location needs to have updated.
	//
	CheckLocation(obj, info);

	//
	// If there is an overlay and the current location is being
	// set then set to redraw the window.
	//
	if ((info == &obj->currentInfo) && obj->isOverlayActive)
		AddWindowObjectUpdateFlags(obj, U_REDRAW);

	//
	// If this is the current location then check to see any
	// of the other locations want the status line updated.
	//
	if (info == &obj->currentInfo)
		CheckAllLocationStatus(obj);
}


/*
 * Convenience routine to adjust the row and column of a location.
 * If the location is being shown, then the window will be marked as
 * needing an update if required.
 */
void
AdjustLocation(LifeObject * obj, LocationInfo * info,
	NUMBER rowDelta, NUMBER colDelta)
{
	SetLocation(obj, info,
		numAdd(info->row, rowDelta), numAdd(info->col, colDelta));
}


/*
 * Convenience routine to copy one location to another in an object.
 * If the location is being shown, then the window will be marked as
 * needing an update if required.
 */
void
CopyLocation(LifeObject * obj, const LocationInfo * src, LocationInfo * dest)
{
	SetLocation(obj, dest, src->row, src->col);
}


/*
 * Convenience routine to set just the row for a location in an object.
 * If the location is being shown, then the window will be marked as
 * needed an update if required.
 */
void
SetLocationRow(LifeObject * obj, LocationInfo * info, NUMBER row)
{
	SetLocation(obj, info, row, info->col);
}


/*
 * Convenience routine to set just the column for a location in an object.
 * If the location is being shown, then the window will be marked as
 * needed an update if required.
 */
void
SetLocationCol(LifeObject * obj, LocationInfo * info, NUMBER col)
{
	SetLocation(obj, info, info->row, col);
}


/*
 * Convenience routine to set the current location of an object.
 * If the location is being shown, then the window will be marked as
 * needing an update if required.
 */
void
SetCurrentLocation(LifeObject * obj, NUMBER row, NUMBER col)
{
	SetLocation(obj, &obj->currentInfo, row, col);
}


/*
 * Convenience to adjust the current location by the specified values.
 * If the location is being shown, then the window will be marked as
 * needing an update if required.
 */
void
AdjustCurrentLocation(LifeObject * obj, NUMBER rowDelta, NUMBER colDelta)
{
	SetLocation(obj, &obj->currentInfo,
		numAdd(obj->currentInfo.row, rowDelta),
		numAdd(obj->currentInfo.col, colDelta));
}


/*
 * Convenience routine to set just the row for the current location
 * of an object.  If the location is being shown, then the window will
 * be marked as needing an update if required.
 */
void
SetCurrentRow(LifeObject * obj, NUMBER row)
{
	SetLocation(obj, &obj->currentInfo, row, obj->currentInfo.col);
}


/*
 * Convenience routine to set just the column for the current location
 * of an object.  If the location is being shown, then the window will
 * be marked as needing an update if required.
 */
void
SetCurrentCol(LifeObject * obj, NUMBER col)
{
	SetLocation(obj, &obj->currentInfo, obj->currentInfo.row, col);
}

/* END CODE */
