/*
 * Copyright (c) 2015 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Generic routines for managing paths.
 */

#include "life.h"

#define	PATHALLOC	20


static BOOL AllocatePath(Path *, int);
static void FindBeginLoop(Path *);
static BOOL Intersects(NUMBER y1, NUMBER x1, NUMBER y2, NUMBER x2, NUMBER ix);


/*
 * Initialize the path for an object.
 * This creates an empty path.
 */
void
InitializePath(LifeObject * obj)
{
	obj->path.count = 0;
	obj->path.maxCount = 0;
	obj->path.oldCount = 0;
	obj->path.begin = 0;
	obj->path.locs = 0;
}


/*
 * Free the path for an object.
 * This is used when an object is being deleted.
 */
void
FreePath(LifeObject * obj)
{
	if (obj->path.maxCount > 0)
		free(obj->path.locs);

	obj->path.count = 0;
	obj->path.maxCount = 0;
	obj->path.oldCount = 0;
	obj->path.begin = 0;
	obj->path.locs = 0;
}


/*
 * Copy a path from one object to another.
 */
void
CopyPath(const LifeObject * srcobj, LifeObject * destobj)
{
	const Path *	srcPath;
	Path *		destPath;

	if (srcobj == destobj)
		return;

	srcPath = &srcobj->path;
	destPath = &destobj->path;

	if ((destPath->count > 0) || (srcPath->count > 0))
		destobj->changeCount++;

	/*
	 * Grow the destination path if necessary to hold the source path.
	 */
	if (srcPath->count > destPath->maxCount)
	{
		if (!AllocatePath(destPath, srcPath->count))
			Error("No memory to copy path");
	}

	/*
	 * Now copy the path coordinates.
	 */
	destPath->count = srcPath->count;
	destPath->begin = srcPath->begin;
	destPath->oldCount = srcPath->oldCount;

	if (srcPath->count > 0)
	{
		memcpy((char *) destPath->locs, (char *) srcPath->locs,
			(srcPath->count + 1) * sizeof(Location));
	}
}


/*
 * Clear the current path from an object.
 * This resets it to the empty state.
 * The number of points the path had is saved for later restoring.
 */
void
ClearPath(LifeObject * obj)
{
	Path *	path;

	path = &obj->path;

	if (path->count > 0)
	{
		AddWindowObjectUpdateFlags(obj, U_REDRAW);
		obj->changeCount++;
	}

	path->oldCount = path->count;
	path->count = 0;
	path->begin = 0;
}


/*
 * Add a new coordinate to the path for an object.
 * Returns FALSE if the coordinate cannot be added.
 * This appends to the latest path loop, or starts a new one.
 * This clears the old path count so that no restoration of the old
 * path is possible anymore.
 */
BOOL
AddPointToPath(LifeObject * obj, NUMBER row, NUMBER col)
{
	Path *	path;
	int	count;

	path = &obj->path;
	count = path->count;

	path->oldCount = 0;

	if ((count > 0) && numIsEqual(path->locs[count - 1].row, row) &&
		numIsEqual(path->locs[count - 1].col, col))
	{
		return TRUE;
	}

	/*
	 * Make sure there is enough room for this new point.
	 */
	if (!AllocatePath(path, count + 1))
		return FALSE;

	/*
	 * Set the bounds if this is the first point in the path.
	 */
	if (count == 0)
	{
		path->minRow = row;
		path->maxRow = row;
		path->minCol = col;
		path->maxCol = col;
	}

	/*
	 * Increase the bounds to include this point.
	 */
	path->minRow = numMin(path->minRow, row);
	path->maxRow = numMax(path->maxRow, row);
	path->minCol = numMin(path->minCol, col);
	path->maxCol = numMax(path->maxCol, col);

	/*
	 * Append this new point to the end of the path.
	 */
	path->locs[count].row = row;
	path->locs[count].col = col;
	path->locs[count].flag = FALSE;
	path->count++;

	/*
	 * Duplicate the beginning coordinate of this loop at the end to
	 * implicitly close it.  The user can explicitly close it later if
	 * desired in order to start another disjoint piece of the path.
	 */
	path->locs[path->count] = path->locs[path->begin];

	AddWindowObjectUpdateFlags(obj, U_VIEW);
	obj->changeCount++;

	return TRUE;
}


/*
 * Close the current path so as to let another disjoint piece be added.
 * This connects the current end of the path back to the starting point.
 * This is always possible because the append to point routine leaves
 * one extra point available for this purpose.
 */
void
ClosePath(LifeObject * obj)
{
	Path *	path;
	int	count;

	path = &obj->path;
	count = path->count;

	/*
	 * If there is no path or if the path is already closed then
	 * do nothing.
	 */
	if ((count <= 0) || path->locs[count - 1].flag)
		return;

	/*
	 * Copy the beginning point of the latest loop to the end to close it.
	 * Remember the beginning of the next piece of the loop to come.
	 */
	path->locs[count] = path->locs[path->begin];
	path->locs[count].flag = TRUE;
	path->count++;

	path->begin = path->count;

	AddWindowObjectUpdateFlags(obj, U_VIEW);
	obj->changeCount++;
}


/*
 * Shorten the current path by the specified number of segments.
 */
void
ShortenPath(LifeObject * obj, Count count)
{
	Path * path = &obj->path;

	if (count <= 0)
		return;

	if (count >= path->count)
	{
		ClearPath(obj);

		return;
	}

	path->count -= count;
	path->oldCount = 0;

	FindBeginLoop(path);

	AddWindowObjectUpdateFlags(obj, U_REDRAW);
	obj->changeCount++;
}


/*
 * Restore the path that was just cleared.
 * This cannot be done once a new path has been started.
 * Returns TRUE if there was an old path that was restored.
 */
BOOL
RestorePath(LifeObject * obj)
{
	Path * path = &obj->path;

	if (path->oldCount == 0)
		return FALSE;

	path->count = path->oldCount;
	path->oldCount = 0;

	FindBeginLoop(path);

	obj->changeCount++;

	return TRUE;
}


/*
 * Recalculate the beginning of the last loop in a path.
 */
static void
FindBeginLoop(Path * path)
{
	int i;

	path->begin = 0;

	for (i = 0; i < path->count; i++)
	{
		if (path->locs[i].flag)
			path->begin = i + 1;
	}

	/*
	 * Duplicate the beginning of the loop at the end to implicitly
	 * close the path.
	 */
	path->locs[path->count] = path->locs[path->begin];
}


/*
 * Shift the coordinates in a path left or right and up and down.
 */
void
ShiftPath(LifeObject * obj, NUMBER leftOrRight, NUMBER upOrDown)
{
	Path * path = &obj->path;
	int count = path->count;
	int i;

	if ((count <= 0) || (numIsZero(leftOrRight) && numIsZero(upOrDown)))
		return;

	numAddrAdd(&path->minRow, upOrDown);
	numAddrAdd(&path->maxRow, upOrDown);
	numAddrAdd(&path->minCol, leftOrRight);
	numAddrAdd(&path->maxCol, leftOrRight);

	for (i = 0; i < count; i++)
	{
		numAddrAdd(&path->locs[i].row, upOrDown);
		numAddrAdd(&path->locs[i].col, leftOrRight);
	}
}


/*
 * Return the sigNum function to indicate whether the specified point is
 * truly inside of the current path, exactly on the boundary of the path,
 * or truly outside of the current path.  These cases return -1, 0, or 1
 * respectively.  An empty path always returns the value 1 for "outside".
 */
int
PointPathSignum(const Path * path, NUMBER row, NUMBER col)
{
	int i;
	int count;

	if (path->count <= 0)
		return 1;

	if (numIsLess(row, path->minRow) || numIsGreater(row, path->maxRow))
		return 1;

	if (numIsLess(col, path->minCol) || numIsGreater(col, path->maxCol))
		return 1;

	if (PointIsOnPath(path, row, col))
		return 0;

	/*
	 * Count the number of intersections of the line segment from the
	 * left edge of the path's boundaries to the coordinate of interest.
	 * The point is inside the path if the total number of intersections
	 * is odd.  For the purpose of this call we translate all the points
	 * by the leftmost edge's coordinate.
	 */
	count = 0;

	for (i = 0; i < path->count; i++)
	{
		if (path->locs[i].flag)
			continue;

		if (Intersects(
			numInc(numMulLong(numSub(path->locs[i].row, row), 15)),
			numSub(path->locs[i].col, path->minCol),
			numInc(numMulLong(numSub(path->locs[i + 1].row, row), 15)),
			numSub(path->locs[i + 1].col, path->minCol),
			numSub(col, path->minCol)))
		{
			count++;
		}
	}

	return (((count & 0x01) != 0) ? -1 : 1);
}


/*
 * Test to see whether a point is exactly on the path defined for an object.
 * Returns TRUE if the point is exactly on any of the path's line segments.
 */
BOOL
PointIsOnPath(const Path * path, NUMBER row, NUMBER col)
{
	int i;

	if (numIsLess(row, path->minRow) || numIsGreater(row, path->maxRow))
		return FALSE;

	if (numIsLess(col, path->minCol) || numIsGreater(col, path->maxCol))
		return FALSE;

	for (i = 0; i < path->count; i++)
	{
		if (path->locs[i].flag)
			continue;

		/*
		 * Get the coordinates of the endpoints of the next segment,
		 * and of the point to be tested.
		 */
		NUMBER x1 = path->locs[i].col;
		NUMBER y1 = path->locs[i].row;
		NUMBER x2 = path->locs[i+1].col;
		NUMBER y2 = path->locs[i+1].row;

		NUMBER tx = col;
		NUMBER ty = row;

		/*
		 * Check if the test point is outside of the rectangular
		 * region whose opposite corners are defined by the
		 * endpoints of the segment.  If so, then the point can't
		 * be on the segment.
		 */
		if (numIsLess(tx, x1) && numIsLess(tx, x2))
			continue;

		if (numIsGreater(tx, x1) && numIsGreater(tx, x2))
			continue;

		if (numIsLess(ty, y1) && numIsLess(ty, y2))
			continue;

		if (numIsGreater(ty, y1) && numIsGreater(ty, y2))
			continue;

		/*
		 * Now check for horizontal or vertical segments.
		 * We easily know the answer for either of these cases.
		 */
		if (numIsEqual(x1, x2))
		{
			if (numIsEqual(tx, x1))
				return TRUE;

			continue;
		}

		if (numIsEqual(y1, y2))
		{
			if (numIsEqual(ty, y1))
				return TRUE;

			continue;
		}

		/*
		 * The segment is sloping and the test point is near it.
		 * Shift the first point down to the origin to make the
		 * calculations easier, and also to help prevent overflows
		 * of the multiplication.
		 */
		numAddrSub(&tx, x1);
		numAddrSub(&x2, x1);
		numAddrSub(&ty, y1);
		numAddrSub(&y2, y1);
//		x1 = numZero();
//		y1 = numZero();

		/*
		 * Now we can compare slopes easily to get the answer:
		 *	y2/x2 = ty/tx
		 */
		if (numIsEqual(numMul(tx, y2), numMul(ty, x2)))
			return TRUE;
	}

	return FALSE;
}


/*
 * Determine if the horizontal line segment whose endpoints are at (0,0) and
 * (ix, 0) intersects the line segment with endpoints (x1, y1) and (x2, y2).
 * Returns TRUE if so.  This routine knows that the input coordinates are
 * somewhat constrained (e.g., x1 and x2 are non-negative).
 */
static BOOL
Intersects(NUMBER y1, NUMBER x1, NUMBER y2, NUMBER x2, NUMBER ix)
{
	NUMBER tmp;

	if (numIsNeg(y1) && numIsNeg(y2))
		return FALSE;

	if (numIsPos(y1) && numIsPos(y2))
		return FALSE;

	if (numIsGreater(x1, ix) && numIsGreater(x2, ix))
		return FALSE;

	if (numIsLessOrEqual(x1, ix) && numIsLessOrEqual(x2, ix))
		return TRUE;

	if (numIsLess(y1, y2))
	{
		tmp = x1; x1 = x2; x2 = tmp;
		tmp = y1; y1 = y2; y2 = tmp;
	}

	return numIsLessOrEqual(numMul(numSub(ix, x1), numSub(y2, y1)),
		numMul(y1, numSub(x1, x2)));
}


/*
 * Allocate or grow a path array to hold the specified number of elements.
 * This allocates a few extra elements for extra growth later, and also
 * reserves one more point for the implicit closure of the path.
 * Returns TRUE if successful.
 */
static BOOL
AllocatePath(Path * path, int count)
{
	Location * newLocs;
	int newLen;

	if (count <= path->maxCount)
		return TRUE;

	newLen = (count + PATHALLOC + 1) * sizeof(Location);

	if (path->maxCount > 0)
		newLocs = realloc(path->locs, newLen);
	else
		newLocs = malloc(newLen);

	if (newLocs == 0)
		return FALSE;

	path->locs = newLocs;
	path->maxCount = count + PATHALLOC;

	return TRUE;
}

/* END CODE */
