/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Implementation of the quadcell algorithm.
 * These allow the small representation of very large Life objects if
 * they have some regularity.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "quad.h"
#include "quadpriv.h"


/*
 * Hash function for a quadcell.
 * This produces a random-like number based on the quadrant indexes.
 */
typedef	unsigned int HASH;

#define	MAKE_HASH(ul, ur, ll, lr) \
	((HASH) ((ul) * 51515159 + \
	(ur) * 100000007 + \
	(ll) * 12345701 + \
	(lr) * 77777803))

/*
 * The hash table.
 * This is grown as necessary to keep the length of the chains small.
 */
static int hashTableSize;
static int hashTableUsed;

static QUADINDEX * hashTable;


/*
 * A list of the free quadcells found during a garbage collection.
 * These should not include the empty or full cells.
 * The cells are linked using their next fields.
 * The list is terminated by the INDEX_NULL value.
 */
static QUADINDEX freeList;


/*
 * Special cells containing all OFF, ON, DEAD, NEW or MARK pure states.
 * These have pre-defined special indexes and are not cached.
 * The arrays are indexed by the size of the quadcell.
 */
static QuadCell * onCell;
static QuadCell * offCell;
static QuadCell * deadCell;
static QuadCell * newCell;
static QuadCell * markCell;

static QuadCell * offCellTable[MAXSIZE];
static QuadCell * deadCellTable[MAXSIZE];

static QuadCell * onCellTable[MAXSIZE];
static QuadCell * newCellTable[MAXSIZE];
static QuadCell * markCellTable[MAXSIZE];


/*
 * Table of all pure quadcells for all the possible states.
 * This is indexed by their pure index values, which are from
 * INDEX_FIRST_OFF to INDEX_LAST_MARK.
 */
static QuadCell * pureCellTable[INDEX_LAST_MARK - INDEX_FIRST_OFF + 1];


/*
 * The table of all possible size 2 cells.
 * These are indexed by the combination of the pure states of the four
 * subquadrants making up the quadcell.  Each of the four quadrants of these
 * cells point to the markCell, onCell, newCall, offCell, or deadCell as
 * appropriate.  If the four quadcells are the same, then this will be a
 * duplicate of one of the entries in another table.  For example, the first
 * element represents four OFF cells, and is the same value as offCellTable[1].
 */
static QuadCell * size2Cells[STATE_COUNT * STATE_COUNT * STATE_COUNT * STATE_COUNT];


/*
 * The table of all the quadcells which have ever been allocated.
 * These are accessed using their QUADINDEX values using the divide and
 * modulo operations by the QUAD_ALLOC_SIZE value.
 * The quadNextIndex value indicates the first index in the table
 * which is not yet used.  quadcells are never freed, they are
 * instead put into the free list for later reuse.
 */
QUADINDEX quadNextIndex;

QuadChunk * quadCellTable[MAX_QUAD_CHUNKS];


/*
 * Temporary quadcells which are safe from garbage collection.
 */
#define	SAFE_QUADCELL_SIZE 50

static int safeQuadCellCount;
static QuadCell * safeQuadCells[SAFE_QUADCELL_SIZE];


/*
 * The statistics.
 */
QuadStats quadStats;


/*
 * The callback routine if any for the quadcell routines to call when
 * they think a garbage collection would be a good idea.  It is up to the
 * callback routine to decide whether garbage collection is to be done,
 * and if so, to actually do it.
 */
static void (*callback)(void);


/*
 * Forward definitions.
 */
static void qcInitOffCells(void);
static void qcInitDeadCells(void);
static void qcInitOnCells(void);
static void qcInitNewCells(void);
static void qcInitMarkCells(void);
static void qcInitSize2Cells(void);

static void qcInitOneSize2Cell(int ulState, int urState, int llState, int lrState);
static int qcGetSize2Index(int ulState, int urState, int llState,int lrState);
static int qcGetStateNumberFromState(int state);
static int qcGetPureStateIndex(int state);

static QuadCell * qcAlloc(void);
static QuadCell * qcRemoveHistoryStatesFunc(QuadCell * cp);
static QuadCell * qcReduceStatesFunc(QuadCell * cp);
static QuadCell * qcRaiseStatesFunc(QuadCell * cp);
static QuadCell * qcSetFunc(QuadCell * cp, NUMBER row, NUMBER col);
static QuadCell * qcClearFunc(QuadCell * cp, NUMBER row, NUMBER col);
static QuadCell * qcChangeStateFunc(QuadCell * cp, int oldState, int newState);
static QuadCell * qcExtractStatesFunc(QuadCell * cp, int states);
static QuadCell * qcSetStateFunc(QuadCell * cp, NUMBER row, NUMBER col, int state);

static void qcRegisterSafeQuadCells(void);


/*
 * Initialize the tables of OFF, ON, DEAD, NEW, MARK and SIZE2 quadcells.
 */
void
qcInit(void)
{
	/*
	 * Initialise the hash table and the free list.
	 */
	qcGrowHashTable();

	freeList = INDEX_NULL;
	quadNextIndex = 0;

	safeQuadCellCount = 0;
	callback = 0;

	/*
	 * Initialise the OFF, ON, DEAD, NEW and MARK cells for each size,
	 * and the quadcells in the size 2 cell table.
	 * The order of these calls is important.
	 */
	qcInitOffCells();
	qcInitDeadCells();
	qcInitOnCells();
	qcInitNewCells();
	qcInitMarkCells();
	qcInitSize2Cells();

	/*
	 * Now any future allocated quadcells will be cached.
	 */
	verifyAlways(quadNextIndex == INDEX_FIRST_CACHED);

	//
	// Initialise the quadpair data.
	//
	qcInitQuadPairs();
}


/*
 * Allocate the OFF quadcells for each size.
 * These have quadrants which are also OFF cells.
 * These are not inserted into the hash table since they are found specially.
 * The indexes for these cells are in a special pre-defined range.
 */
static void
qcInitOffCells(void)
{
	int size;

	/*
	 * First allocate and store the size 1 OFF cell.
	 * Its quadrants make no sense.
	 */
	QuadCell * cp = qcAlloc();

	verify(cp->index == INDEX_FIRST_OFF);

	cp->size = SIZE1;
	cp->state = STATE_OFF;
	cp->ul = INDEX_NULL;
	cp->ur = INDEX_NULL;
	cp->ll = INDEX_NULL;
	cp->lr = INDEX_NULL;

	offCell = cp;
	offCellTable[0] = cp;
	pureCellTable[INDEX_FIRST_OFF] = cp;

	/*
	 * Now allocate the remaining OFF cells having each one
	 * reference the previous size.
	 */
	for (size = 1; size < MAXSIZE; size++)
	{
		QUADINDEX prevIndex = INDEX_FIRST_OFF + size - 1;

		cp = qcAlloc();

		cp->size = size;
		cp->state = STATE_OFF;
		cp->ul = prevIndex;
		cp->ur = prevIndex;
		cp->ll = prevIndex;
		cp->lr = prevIndex;

		offCellTable[size] = cp;
		pureCellTable[INDEX_FIRST_OFF + size] = cp;
	}
}


/*
 * Allocate the DEAD quadcells for each size.
 * These have quadrants which are also DEAD cells.
 * These are not inserted into the hash table since they are found specially.
 * The indexes for these cells are in a special pre-defined range.
 */
static void
qcInitDeadCells(void)
{
	int size;

	/*
	 * First allocate and store the size 1 DEAD cell.
	 * Its quadrants make no sense.
	 */
	QuadCell * cp = qcAlloc();

	verify(cp->index == INDEX_FIRST_DEAD);

	cp->size = SIZE1;
	cp->state = STATE_DEAD;
	cp->ul = INDEX_NULL;
	cp->ur = INDEX_NULL;
	cp->ll = INDEX_NULL;
	cp->lr = INDEX_NULL;

	deadCell = cp;
	deadCellTable[0] = cp;
	pureCellTable[INDEX_FIRST_DEAD] = cp;

	/*
	 * Now allocate the remaining DEAD cells having each one
	 * reference the previous size.
	 */
	for (size = 1; size < MAXSIZE; size++)
	{
		QUADINDEX prevIndex = INDEX_FIRST_DEAD + size - 1;

		cp = qcAlloc();

		cp->size = size;
		cp->state = STATE_DEAD;
		cp->ul = prevIndex;
		cp->ur = prevIndex;
		cp->ll = prevIndex;
		cp->lr = prevIndex;

		deadCellTable[size] = cp;
		pureCellTable[INDEX_FIRST_DEAD + size] = cp;
	}
}


/*
 * Allocate the ON quadcells for each size.
 * These have quadrants which are also ON cells.
 * These are not inserted into the hash table since they are found specially.
 * The indexes for these cells are in a special pre-defined range.
 */
static void
qcInitOnCells(void)
{
	int size;

	/*
	 * First allocate and save the size 1 ON cell.
	 * Its quadrants make no sense.
	 */
	QuadCell * cp = qcAlloc();

	verifyAlways(cp->index == INDEX_FIRST_ON);

	cp->size = SIZE1;
	cp->state = STATE_ON;
	cp->ul = INDEX_NULL;
	cp->ur = INDEX_NULL;
	cp->ll = INDEX_NULL;
	cp->lr = INDEX_NULL;

	onCell = cp;
	onCellTable[0] = cp;
	pureCellTable[INDEX_FIRST_ON] = cp;

	/*
	 * Now allocate the remaining ON cells having each one
	 * reference the previous size.
	 */
	for (size = 1; size < MAXSIZE; size++)
	{
		QUADINDEX prevIndex = INDEX_FIRST_ON + size - 1;

		cp = qcAlloc();

		cp->size = size;
		cp->state = STATE_ON;
		cp->ul = prevIndex;
		cp->ur = prevIndex;
		cp->ll = prevIndex;
		cp->lr = prevIndex;

		onCellTable[size] = cp;
		pureCellTable[INDEX_FIRST_ON + size] = cp;
	}
}


/*
 * Allocate the NEW quadcells for each size.
 * These have quadrants which are also NEW cells.
 * These are not inserted into the hash table since they are found specially.
 * The indexes for these cells are in a special pre-defined range.
 */
static void
qcInitNewCells(void)
{
	int size;

	/*
	 * First allocate and save the size 1 NEW cell.
	 * Its quadrants make no sense.
	 */
	QuadCell * cp = qcAlloc();

	verify(cp->index == INDEX_FIRST_NEW);

	cp->size = SIZE1;
	cp->state = STATE_NEW;
	cp->ul = INDEX_NULL;
	cp->ur = INDEX_NULL;
	cp->ll = INDEX_NULL;
	cp->lr = INDEX_NULL;

	newCell = cp;
	newCellTable[0] = cp;
	pureCellTable[INDEX_FIRST_NEW] = cp;

	/*
	 * Now allocate the remaining NEW cells having each one
	 * reference the previous size.
	 */
	for (size = 1; size < MAXSIZE; size++)
	{
		QUADINDEX prevIndex = INDEX_FIRST_NEW + size - 1;

		cp = qcAlloc();

		cp->size = size;
		cp->state = STATE_NEW;
		cp->ul = prevIndex;
		cp->ur = prevIndex;
		cp->ll = prevIndex;
		cp->lr = prevIndex;

		newCellTable[size] = cp;
		pureCellTable[INDEX_FIRST_NEW + size] = cp;
	}
}


/*
 * Allocate the MARK quadcells for each size.
 * These have quadrants which are also MARK cells.
 * These are not inserted into the hash table since they are found specially.
 * The indexes for these cells are in a special pre-defined range.
 */
static void
qcInitMarkCells(void)
{
	int size;

	/*
	 * First allocate and save the size 1 MARK cell.
	 * Its quadrants make no sense.
	 */
	QuadCell * cp = qcAlloc();

	verify(cp->index == INDEX_FIRST_MARK);

	cp->size = SIZE1;
	cp->state = STATE_MARK;
	cp->ul = INDEX_NULL;
	cp->ur = INDEX_NULL;
	cp->ll = INDEX_NULL;
	cp->lr = INDEX_NULL;

	markCell = cp;
	markCellTable[0] = cp;
	pureCellTable[INDEX_FIRST_MARK] = cp;

	/*
	 * Now allocate the remaining MARK cells having each one
	 * reference the previous size.
	 */
	for (size = 1; size < MAXSIZE; size++)
	{
		QUADINDEX prevIndex = INDEX_FIRST_MARK + size - 1;

		cp = qcAlloc();

		cp->size = size;
		cp->state = STATE_MARK;
		cp->ul = prevIndex;
		cp->ur = prevIndex;
		cp->ll = prevIndex;
		cp->lr = prevIndex;

		markCellTable[size] = cp;
		pureCellTable[INDEX_FIRST_MARK + size] = cp;
	}
}


/*
 * Allocate the size 2 quadcells, and store them into a special small
 * table for direct indexing.  The indexing is based on the pure states
 * of the four central cells of a size 2 quadcell.  If all of those
 * cells have the same state, then so will this one and this entry will
 * be a duplication of one in another table.  For example, size2Cells[0]
 * is the same as offCellTable[1].  None of these cells are ever garbage
 * collected.
 */
static void
qcInitSize2Cells(void)
{
	int ulState, urState, llState, lrState;

	for (ulState = STATE_LOWBIT;
		ulState <= STATE_HIGHBIT; ulState <<= 1)
	{
		for (urState = STATE_LOWBIT;
			urState <= STATE_HIGHBIT; urState <<= 1)
		{
			for (llState = STATE_LOWBIT;
				llState <= STATE_HIGHBIT; llState <<= 1)
			{
				for (lrState = STATE_LOWBIT;
					lrState <= STATE_HIGHBIT; lrState <<= 1)
				{
					qcInitOneSize2Cell(ulState, urState,
						llState, lrState);
				}
			}
		}
	}
}


/*
 * Initialise one element of the size2Cells table for the specified
 * set of states for its four subquadrants.
 */
static void
qcInitOneSize2Cell(int ulState, int urState, int llState, int lrState)
{
	int index = qcGetSize2Index(ulState, urState, llState, lrState);

	//
	// If this is going to be a pure state then just copy the
	// already allocated quadcell from the appropriate pure table.
	//
	if ((ulState == urState) && (urState == llState) && (llState == lrState))
	{
		QuadCell ** table;

		switch (ulState)
		{
			case STATE_OFF: table = offCellTable; break;
			case STATE_ON: table = onCellTable; break;
			case STATE_DEAD: table = deadCellTable; break;
			case STATE_NEW: table = newCellTable; break;
			case STATE_MARK: table = markCellTable; break;

			default:
				verifyAlways(FALSE);
				table = offCellTable;
		}

		size2Cells[index] = table[1];

		return;
	}

	//
	// This is not a pure state so we will have to allocate it.
	//
	QuadCell * cp = qcAlloc();

	cp->size = SIZE2;
	cp->state = (ulState | urState | llState | lrState);
	cp->ul = qcGetPureStateIndex(ulState);
	cp->ur = qcGetPureStateIndex(urState);
	cp->ll = qcGetPureStateIndex(llState);
	cp->lr = qcGetPureStateIndex(lrState);

	size2Cells[index] = cp;
}


/*
 * Get the state value for a cell given its quadcell index.
 * This is only callable for pure quadcells (OFF, ON, DEAD, NEW or MARK).
 */
int
qcGetStateFromIndex(int quadIndex)
{
	if (isOff(quadIndex))
		return STATE_OFF;

	if (isDead(quadIndex))
		return STATE_DEAD;

	if (isOn(quadIndex))
		return STATE_ON;

	if (isNew(quadIndex))
		return STATE_NEW;

	if (isMark(quadIndex))
		return STATE_MARK;

	verifyAlways(FALSE);

	return STATE_OFF;
}


/*
 * Return the base index for the table of quadcells for the
 * specified pure state.
 */
static int
qcGetPureStateIndex(int state)
{
	switch (state)
	{
		case STATE_OFF: return INDEX_FIRST_OFF;
		case STATE_ON: return INDEX_FIRST_ON;
		case STATE_DEAD: return INDEX_FIRST_DEAD;
		case STATE_NEW: return INDEX_FIRST_NEW;
		case STATE_MARK: return INDEX_FIRST_MARK;

		default:
			verifyAlways(FALSE);
	}

	return 0;
}


/*
 * Calculate the index into the size2Cells table based on the specified
 * pure states of its four subquadrant cells.
 */
static int
qcGetSize2Index(int ulState, int urState, int llState, int lrState)
{
	int num1 = qcGetStateNumberFromState(ulState);
	int num2 = qcGetStateNumberFromState(urState);
	int num3 = qcGetStateNumberFromState(llState);
	int num4 = qcGetStateNumberFromState(lrState);

	return (num1 + (num2 * STATE_COUNT) +
		(num3 * STATE_COUNT * STATE_COUNT) +
		(num4 * STATE_COUNT * STATE_COUNT * STATE_COUNT));
}


/*
 * Get the state number from the pure state value.
 * This ranges from 0 to STATE_COUNT - 1.
 */
static int
qcGetStateNumberFromState(int state)
{
	switch (state)
	{
		case STATE_OFF: return 0;
		case STATE_DEAD: return 1;
		case STATE_ON: return 2;
		case STATE_NEW: return 3;
		case STATE_MARK: return 4;

		default:
			verifyAlways(FALSE);
	}

	return 0;
}


/*
 * Search the hash table for a quadcell with the specified quadrants.
 * Pure OFF, ON, DEAD, FULL, MARK and the and SIZE2 quadcells are handled
 * specially since they aren't cached.  If the quadcell is not found then
 * a new one is allocated.
 */
QuadCell *
qcFind(QUADINDEX ulIndex, QUADINDEX urIndex,
	QUADINDEX llIndex, QUADINDEX lrIndex)
{
	//
	// Check for all of the four quadrants having the same pure index.
	// In that case just return the next size pure cell of the same type.
	//
	if ((ulIndex == urIndex) && (urIndex == llIndex) &&
		(llIndex == lrIndex) && isPure(ulIndex))
	{
		int size = sizeFromPureIndex(ulIndex) + 1;

		verifyAlways(size < MAXSIZE);

		return pureCellTable[ulIndex + 1];
	}

	//
	// This won't be a pure quadcell, but if it is composed of SIZE1
	// pure quadrants then it will be a SIZE2 quadcell.  Get it by
	// indexing into the size2Cells table using the pure states of
	// its quadrants.
	//
	if (isPure(ulIndex) && (sizeFromPureIndex(ulIndex) == SIZE1))
	{
		verifyAlways(isPure(urIndex) &&
			isPure(llIndex) && isPure(lrIndex));

		QuadCell * ul = pureCellTable[ulIndex];
		QuadCell * ur = pureCellTable[urIndex];
		QuadCell * ll = pureCellTable[llIndex];
		QuadCell * lr = pureCellTable[lrIndex];

		//TODO: Use indexes instead of states
		int index = qcGetSize2Index(ul->state, ur->state,
			ll->state, lr->state);

		return size2Cells[index];
	}

	//
	// This is a typical larger quadcell which of mixed types and is
	// therefore cachable.  Calculate the hash value for the quadcell
	// to be found.
	//
	HASH hash = MAKE_HASH(ulIndex, urIndex, llIndex, lrIndex);

	int hashIndex = hash % hashTableSize;

	//
	// Do a linear search through the hash chain looking for the
	// quadcell whose quadrants are the specified ones.
	//
	QUADINDEX quadIndex = hashTable[hashIndex];

	while (quadIndex != INDEX_NULL)
	{
		QuadCell * cp = indexToQuadCell(quadIndex);

		quadStats.findLength++;

		/*
		 * If this is the desired quadcell, return it.
		 */
		if ((cp->ul == ulIndex) && (cp->ur == urIndex) &&
			(cp->ll == llIndex) && (cp->lr == lrIndex))
		{
			quadStats.findHits++;

			return cp;
		}

		/*
		 * Advance to the next quadcell in the hash chain.
		 */
		quadIndex = cp->next;
	}

	//
	// The quadcell was not found in the hash chain and so will
	// need to be allocated and added.  Its state will be a merging
	// of the subquadrant states and its size will be one more than
	// the subquadrant sizes.
	//
	quadStats.findMisses++;

	QuadCell * ul = indexToQuadCell(ulIndex);
	QuadCell * ur = indexToQuadCell(urIndex);
	QuadCell * ll = indexToQuadCell(llIndex);
	QuadCell * lr = indexToQuadCell(lrIndex);

	int state = (ul->state | ur->state | ll->state | lr->state);
	int size = ul->size;

	verifyAlways(size >= SIZE2);
	verifyAlways((size == ur->size) && (size == ll->size) && (size == lr->size));

	//
	// The found quadcell will be the next size up.
	// Make sure that it isn't too big.
	//
	size++;

	verifyAlways(size < MAXSIZE);

	//
	// Allocate a new quadcell and initialise it.
	//
	QuadCell * cp = qcAlloc();

	cp->size = size;
	cp->state = state;
	cp->ul = ulIndex;
	cp->ur = urIndex;
	cp->ll = llIndex;
	cp->lr = lrIndex;

	//
	// Verify that the cell is cachable and add it to the hash table.
	// (Non-cachable cells should have been found above.)
	//
	verifyAlways(isCached(cp->index));

	cp->next = hashTable[hashIndex];
	hashTable[hashIndex] = cp->index;
	hashTableUsed++;

	if (size < QUAD_STAT_SIZE)
		quadStats.sizes[size]++;

	return cp;
}


/*
 * Allocate and return a new quadcell.
 * This allocates a few at a time to minimize malloc calls.
 * If there are any cells on the free list then these are used first.
 * Newly allocated quadcells are added to the quadcell table.
 * The index of the quadcell is set in the returned quadcell.
 * If the quadcell is newly allocated then its index will be the value
 * of quadNextIndex (which is then incremented).  The returned quadcell
 * has no valid quadrants or state, and is not put into the hash table.
 */
static QuadCell *
qcAlloc(void)
{
	QuadCell * cp;
	QUADINDEX quadIndex;

	/*
	 * First see if there is anything on the free list.
	 */
	if (freeList != INDEX_NULL)
	{
		cp = indexToQuadCell(freeList);
		cp->flag = 0;
		cp->state = STATE_NONE;
		cp->numIndex = INDEX_NULL_NUMBER;
		freeList = cp->next;

		cp->next = INDEX_NULL;
		cp->result = INDEX_NULL;

		quadStats.cellsAvailable--;
		quadStats.cellsAllocated++;

		return cp;
	}

	/*
	 * If no new quadcells are available then allocate another
	 * chunk and store it into the quadcell table.
	 */
	if ((quadNextIndex % QUAD_ALLOC_SIZE) == 0)
	{
		int tableIndex = quadNextIndex / QUAD_ALLOC_SIZE;

		verify(tableIndex < MAX_QUAD_CHUNKS);

		QuadChunk * chunk = qcAllocMem("quadcells", sizeof(QuadChunk));

		quadCellTable[tableIndex] = chunk;

		quadStats.cellsAvailable += QUAD_ALLOC_SIZE;
	}

	/*
	 * Get the new quadcell's address and set its index.
	 * Note: We increment the quadNextIndex value first so that the
	 * indexToQuadCell call will not fail its verification.
	 */
	quadIndex = quadNextIndex++;

	cp = indexToQuadCell(quadIndex);
	cp->index = quadIndex;
	cp->next = INDEX_NULL;
	cp->result = INDEX_NULL;
	cp->flag = 0;
	cp->state = STATE_NONE;
	cp->numIndex = INDEX_NULL_NUMBER;

	quadStats.cellsAllocated++;

	return cp;
}


/*
 * Return an approximate percentage value (0 to 100 or more) indicating
 * how full the hash table is.  This is used to determine when a garbage
 * collection is needed or when the hash table needs to be grown.
 */
int
qcGetPercentUsed(void)
{
	return (int) ((100L * hashTableUsed) / hashTableSize);
}


/*
 * Public routine to allocate a new empty quadcell of the smallest
 * useful size.  This is a quadcell with the OFF state.
 */
QuadCell *
qcEmpty(void)
{
	return offCellTable[SIZE_INITIAL];
}


/*
 * Return a quadcell of the specified size which contains no cells,
 * and is in the OFF state.  Size is the power of two (e.g. size 3 is
 * an 8 by 8 region).
 */
QuadCell *
qcGetEmpty(int size)
{
	verifyAlways((size >= SIZE1) && (size < MAXSIZE));

	return offCellTable[size];
}


/*
 * Return a quadcell of the specified size which contains all ON cells.
 * Size is the power of two (e.g. size 3 is an 8 by 8 region).
 */
QuadCell *
qcGetFull(int size)
{
	verifyAlways((size >= SIZE1) && (size < MAXSIZE));

	return onCellTable[size];
}


/*
 * Return a pure cell of the specified size having the specified state.
 * The state must be pure (OFF, ON, DEAD, NEW, or MARK).
 */
QuadCell *
qcGetCellFromState(int size, int state)
{
	verifyAlways((size >= SIZE1) && (size < MAXSIZE));

	switch (state)
	{
		case STATE_OFF: return offCellTable[size];
		case STATE_ON: return onCellTable[size];
		case STATE_NEW: return newCellTable[size];
		case STATE_DEAD: return deadCellTable[size];
		case STATE_MARK: return markCellTable[size];

		default:
			break;
	}

	verifyAlways(FALSE);

	return offCellTable[size];
}


/*
 * Return the SIZE1 cell which is in the specified pure state.
 */
QuadCell *
qcGetSize1CellFromState(int state)
{
	switch (state)
	{
		case STATE_OFF: return offCell;
		case STATE_ON: return onCell;
		case STATE_NEW: return newCell;
		case STATE_DEAD: return deadCell;
		case STATE_MARK: return markCell;

		default:
			break;
	}

	verifyAlways(FALSE);

	return offCell;
}


/*
 * Grow the specified quadcell if necessary so that it is at
 * least as large as the specified size.  The old quadcell is
 * centered in the new one.
 */
QuadCell *
qcGrowToSize(QuadCell * cp, int size)
{
	verifyAlways((size >= SIZE2) && (size < MAXSIZE));

	while (size > cp->size)
		cp = qcGrow(cp);

	return cp;
}


/*
 * Grow the specified quadcell if necessary so that the specified
 * row and column is contained within its boundaries.  The old
 * quadcell is centered in the new one.  Example: a size 3 quadcell
 * has row and column values ranging from -4 to 3 inclusive.
 */
QuadCell *
qcGrowToContain(QuadCell * cp, NUMBER row, NUMBER col)
{
	while (!qcIsCellValid(cp, row, col))
	{
		cp = qcGrow(cp);
	}

	return cp;
}


/*
 * Grow the specified quadcell by one size. The old quadcell is
 * centered in the new one.
 */
QuadCell *
qcGrow(QuadCell * cp)
{
	QuadCell * empty = qcGetEmpty(cp->size - 1);
	QUADINDEX index = empty->index;

	QuadCell * ul = qcFind(index, index, index, cp->ul);
	QuadCell * ur = qcFind(index, index, cp->ur, index);
	QuadCell * ll = qcFind(index, cp->ll, index, index);
	QuadCell * lr = qcFind(cp->lr, index, index, index);

	return qcFind(ul->index, ur->index, ll->index, lr->index);
}


/*
 * Trim the size of a cell structure to remove extraneous empty levels.
 * This only removes extra OFF cells at the edges, while preserving cells
 * which are in the DEAD state.  Returns the resulting cell structure.
 */
QuadCell *
qcTrim(QuadCell * cp)
{
	/*
	 * Iterate while the quadcell is large enough and it has
	 * empty cells on its edges.
	 */
	while (cp->size > 3)
	{
		QuadCell * ul = indexToQuadCell(cp->ul);
		QuadCell * ur = indexToQuadCell(cp->ur);
		QuadCell * ll = indexToQuadCell(cp->ll);
		QuadCell * lr = indexToQuadCell(cp->lr);

		/*
		 * See if any of the edges of the cell are non-OFF.
		 * If so, then we can't trim it any more.
		 * These checks are only for the pure OFF quadcells.
		 */
		if (isNotOff(ul->ul) || isNotOff(ul->ur) || isNotOff(ul->ll) ||
			isNotOff(ur->ul) || isNotOff(ur->ur) || isNotOff(ur->lr) ||
			isNotOff(ll->ul) || isNotOff(ll->ll) || isNotOff(ll->lr) ||
			isNotOff(lr->ur) || isNotOff(lr->ll) || isNotOff(lr->lr))
		{
			return cp;
		}

		/*
		 * The entire edge of the quadcell is OFF.
		 * So we can extract the central square from the cell.
		 */
		cp = qcFind(ul->lr, ur->ll, ll->ur, lr->ul);
	}

	return cp;
}


/*
 * Trim the size of a cell structure to become at least as small as the
 * specified size.  Warning: This discards all cells that lie outside
 * of the trimmed area.  Returns the resulting cell structure.
 */
QuadCell *
qcTrimToSize(QuadCell * cp, int size)
{
	verify(size > SIZE1);

	/*
	 * Iterate while we still need to trim it.
	 */
	while (cp->size > size)
	{
		QuadCell * ul = indexToQuadCell(cp->ul);
		QuadCell * ur = indexToQuadCell(cp->ur);
		QuadCell * ll = indexToQuadCell(cp->ll);
		QuadCell * lr = indexToQuadCell(cp->lr);

		/*
		 * Extract the central square from the cell.
		 */
		cp = qcFind(ul->lr, ur->ll, ll->ur, lr->ul);
	}

	return cp;
}


/*
 * Return an empty quadcell which is large enough to contain the
 * specified coordinates.  The quadcell is in the OFF state.
 */
QuadCell *
qcEmptyToContain(NUMBER row, NUMBER col)
{
	int size = SIZE_INITIAL;

	while (TRUE)
	{
		NUMBER max = numTwoPower(size - 1);
		NUMBER min = numNeg(max);

		if (numIsLess(row, max) && numIsLess(col, max) &&
			numIsGreaterOrEqual(row, min) &&
			numIsGreaterOrEqual(col, min))
		{
			return qcGetEmpty(size);
		}

		size++;
	}
}


/*
 * Set an individual cell at the specified location within the specified
 * quadcell to be ON, and return the new quadcell containing the old
 * pattern plus the new cell (which might be the same quadcell).
 * This treats NEW and MARK cells as also being on.
 * The quadcell is grown as needed to contain the specified location.
 * Locations are relative to the center of the quadcell.
 */
QuadCell *
qcSet(QuadCell * cp, NUMBER row, NUMBER col)
{
	verify(cp->size >= SIZE1);

	cp = qcGrowToContain(cp, row, col);

	return qcSetFunc(cp, row, col);
}


/*
 * Set an individual cell at the specified location within the specified
 * quadcell to be ON, and return the new quadcell containing the
 * old pattern plus the new cell (which might be the same quadcell).
 * This treats NEW and MARK cells as also being ON.
 * The location must lie within the bounds of the quadcell.
 * Locations are relative to the center of the quadcell.
 * This calls itself recursively as needed.
 */
static QuadCell *
qcSetFunc(QuadCell * cp, NUMBER row, NUMBER col)
{
	QuadCell * ul;
	QuadCell * ur;
	QuadCell * ll;
	QuadCell * lr;
	NUMBER adjust;

	//
	// If this quadcell is full then just return it unchanged
	// since any location within it is already ON or NEW.
	//
	if (isCellFull(cp))
		return cp;

	//
	// If we are at the smallest cell size then return the ON cell.
	//
	if (cp->size == SIZE1)
		return onCell;

	//
	// We must call ourself recursively for the appropriate quadrant
	// with an adjusted row and column number.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	ul = indexToQuadCell(cp->ul);
	ur = indexToQuadCell(cp->ur);
	ll = indexToQuadCell(cp->ll);
	lr = indexToQuadCell(cp->lr);

	switch ((numSign(row) >= 0) * 2 + (numSign(col) >= 0))
	{
		case 0:
			numAddrAdd(&row, adjust);
			numAddrAdd(&col, adjust);

			ll = qcSetFunc(ll, row, col);

			break;

		case 1:
			numAddrAdd(&row, adjust);
			numAddrSub(&col, adjust);

			lr = qcSetFunc(lr, row, col);

			break;

		case 2:
			numAddrSub(&row, adjust);
			numAddrAdd(&col, adjust);

			ul = qcSetFunc(ul, row, col);

			break;

		case 3:
			numAddrSub(&row, adjust);
			numAddrSub(&col, adjust);

			ur = qcSetFunc(ur, row, col);

			break;

		default:
			verifyAlways(FALSE);
	}

	//
	// If our quadrants are unchanged then we can return ourself.
	// This can happen when the desired cell was already ON.
	//
	if ((cp->ul == ul->index) && (cp->ur == ur->index) &&
		(cp->ll == ll->index) && (cp->lr == lr->index))
	{
		return cp;
	}

	//
	// There was a change, so we have to return a new quadcell
	// containing the change.
	//
	return qcFind(ul->index, ur->index, ll->index, lr->index);
}


/*
 * Set an individual cell at the specified location within the specified
 * quadcell to be OFF, and return the new quadcell containing the
 * old pattern minus the new cell (which might be the same quadcell).
 * Locations are relative to the center of the quadcell.
 */
QuadCell *
qcClear(QuadCell * cp, NUMBER row, NUMBER col)
{
	verify(cp->size >= SIZE1);

	//
	// If the coordinates are not within the quadcell then the
	// cell is already OFF.
	//
	if (!qcIsCellValid(cp, row, col))
		return cp;

	return qcClearFunc(cp, row, col);
}


/*
 * Set an individual cell at the specified location within the specified
 * quadcell to be OFF, and return the new quadcell containing the
 * old pattern plus the new cell (which might be the same quadcell).
 * This treats DEAD cells as also being OFF.
 * The location must lie within the bounds of the quadcell.
 * Locations are relative to the center of the quadcell.
 * This calls itself recursively as needed.
 */
static QuadCell *
qcClearFunc(QuadCell * cp, NUMBER row, NUMBER col)
{
	QuadCell * ul;
	QuadCell * ur;
	QuadCell * ll;
	QuadCell * lr;
	NUMBER adjust;

	//
	// If this quadcell is empty then just return it unchanged
	// since any location within it is already OFF or DEAD.
	//
	if (isCellEmpty(cp))
		return cp;

	//
	// If we are at the smallest cell size then return the OFF cell.
	//
	if (cp->size == SIZE1)
		return offCell;

	//
	// We must call ourself recursively for the appropriate quadrant
	// with an adjusted row and column number.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	ul = indexToQuadCell(cp->ul);
	ur = indexToQuadCell(cp->ur);
	ll = indexToQuadCell(cp->ll);
	lr = indexToQuadCell(cp->lr);

	switch ((numSign(row) >= 0) * 2 + (numSign(col) >= 0))
	{
		case 0:
			numAddrAdd(&row, adjust);
			numAddrAdd(&col, adjust);

			ll = qcClearFunc(ll, row, col);

			break;

		case 1:
			numAddrAdd(&row, adjust);
			numAddrSub(&col, adjust);

			lr = qcClearFunc(lr, row, col);

			break;

		case 2:
			numAddrSub(&row, adjust);
			numAddrAdd(&col, adjust);

			ul = qcClearFunc(ul, row, col);

			break;

		case 3:
			numAddrSub(&row, adjust);
			numAddrSub(&col, adjust);

			ur = qcClearFunc(ur, row, col);

			break;

		default:
			verifyAlways(FALSE);
	}

	//
	// If our quadrants are unchanged then we can return ourself.
	// This can happen when the desired cell was already OFF.
	//
	if ((cp->ul == ul->index) && (cp->ur == ur->index) &&
		(cp->ll == ll->index) && (cp->lr == lr->index))
	{
		return cp;
	}

	//
	// There was a change, so we have to return a new quadcell
	// containing the change.
	//
	return qcFind(ul->index, ur->index, ll->index, lr->index);
}


/*
 * Set an individual cell at the specified location within the specified
 * quadcell to have the specified pure state, and return the new quadcell
 * containing the old pattern with the new cell (which might be the same quadcell).
 * This can be used to create a new on cell, or to delete an existing cell.
 * The quadcell is grown as needed to contain the specified location.
 * Locations are relative to the center of the quadcell.
 */
QuadCell *
qcSetState(QuadCell * cp, NUMBER row, NUMBER col, int state)
{
	verifyAlways(isStatePure(state));

	//
	// If we are setting the cell OFF and the coordinates are not
	// within the quadcell then the cell is already OFF.
	//
	if ((state == STATE_OFF) && !qcIsCellValid(cp, row, col))
		return cp;

	cp = qcGrowToContain(cp, row, col);

	return qcSetStateFunc(cp, row, col, state);
}


/*
 * Set an individual cell at the specified location within the specified
 * quadcell to the specified pure state, and return the new quadcell containing
 * the old pattern plus the new cell (which might be the same quadcell).
 * The location must lie within the bounds of the quadcell.
 * Locations are relative to the center of the quadcell.
 * This calls itself recursively as needed.
 */
static QuadCell *
qcSetStateFunc(QuadCell * cp, NUMBER row, NUMBER col, int state)
{
	QuadCell * ul;
	QuadCell * ur;
	QuadCell * ll;
	QuadCell * lr;
	NUMBER adjust;

	//
	// If this quadcell has the desired pure state then there
	// is nothing to change.
	//
	if (cp->state == state)
		return cp;

	//
	// If we are at the smallest cell size then return the
	// appropriate single cell.
	//
	if (cp->size == SIZE1)
		return qcGetSize1CellFromState(state);

	//
	// We must call ourself recursively for the appropriate quadrant
	// with an adjusted row and column number.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	ul = indexToQuadCell(cp->ul);
	ur = indexToQuadCell(cp->ur);
	ll = indexToQuadCell(cp->ll);
	lr = indexToQuadCell(cp->lr);

	switch ((numSign(row) >= 0) * 2 + (numSign(col) >= 0))
	{
		case 0:
			numAddrAdd(&row, adjust);
			numAddrAdd(&col, adjust);

			ll = qcSetStateFunc(ll, row, col, state);

			break;

		case 1:
			numAddrAdd(&row, adjust);
			numAddrSub(&col, adjust);

			lr = qcSetStateFunc(lr, row, col, state);

			break;

		case 2:
			numAddrSub(&row, adjust);
			numAddrAdd(&col, adjust);

			ul = qcSetStateFunc(ul, row, col, state);

			break;

		case 3:
			numAddrSub(&row, adjust);
			numAddrSub(&col, adjust);

			ur = qcSetStateFunc(ur, row, col, state);

			break;

		default:
			verifyAlways(FALSE);
	}

	//
	// If our quadrants are unchanged then we can return ourself.
	// This can happen when the desired cell already had the
	// desired state.
	//
	if ((cp->ul == ul->index) && (cp->ur == ur->index) &&
		(cp->ll == ll->index) && (cp->lr == lr->index))
	{
		return cp;
	}

	//
	// There was a change, so we have to return a new quadcell
	// containing the change.
	//
	return qcFind(ul->index, ur->index, ll->index, lr->index);
}


/*
 * Remove all the history states in a quadcell.  This converts DEAD cells
 * to OFF, NEW cells to ON, and leaves MARK cells as they were.
 * Returns the same quadcell if nothing changes.
 */
QuadCell *
qcRemoveHistoryStates(QuadCell * cp)
{
	//
	// If the quadcell has no history states then return it now.
	//
	if (!isCellPartHistory(cp))
		return cp;

	qcSetResultType(RESULT_REMOVEHISTORY);

	return qcRemoveHistoryStatesFunc(cp);
}


/*
 * Recursive routine to remove all the history states in a quadcell.
 * This converts DEAD cells to OFF, NEW cells to ON, and leaves MARK cells
 * as they were.  Returns the same quadcell if nothing changes.
 */
static QuadCell *
qcRemoveHistoryStatesFunc(QuadCell * cp)
{
	//
	// If the quadcell has no history states then return it now.
	//
	if (!isCellPartHistory(cp))
		return cp;

	//
	// If we are pure then return a pure ON, OFF, or MARK quadcell.
	// This handles all size1 cells.
	//
	if (isCellPure(cp))
	{
		if (isCellMark(cp))
			return cp;

		if (isCellEmpty(cp))
			return qcGetEmpty(cp->size);

		if (isCellFull(cp))
			return qcGetFull(cp->size);

		verifyAlways(FALSE);
	}

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// No result was saved so we must calculate it now.
	// Call ourself for each of our quadrants.
	//
	QuadCell * ul = qcRemoveHistoryStatesFunc(indexToQuadCell(cp->ul));
	QuadCell * ur = qcRemoveHistoryStatesFunc(indexToQuadCell(cp->ur));
	QuadCell * ll = qcRemoveHistoryStatesFunc(indexToQuadCell(cp->ll));
	QuadCell * lr = qcRemoveHistoryStatesFunc(indexToQuadCell(cp->lr));

	//
	// Build the result and save it for reuse.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Reduce the number of states in a quadcell to just the OFF and ON
 * states, eliminating all NEW, DEAD, and MARK states.  Returns the same
 * quadcell if nothing changes.
 */
QuadCell *
qcReduceStates(QuadCell * cp)
{
	//
	// If the quadcell only contains OFF and ON states then
	// just return it now.
	//
	if (isCellOnlyOnOff(cp))
		return cp;

	qcSetResultType(RESULT_REDUCE);

	return qcReduceStatesFunc(cp);
}


/*
 * Recursive routine to reduce the number of states in a quadcell to
 * just the OFF and ON states, eliminating all NEW, DEAD and MARK states.
 * Returns the same quadcell if nothing changes.
 */
static QuadCell *
qcReduceStatesFunc(QuadCell * cp)
{
	//
	// If the quadcell only contains OFF and ON states then
	// just return it now.
	//
	if (isCellOnlyOnOff(cp))
		return cp;

	//
	// If we are pure then return a pure ON or OFF quadcell.
	//
	if (isCellPure(cp))
	{
		if (isCellEmpty(cp))
			return qcGetEmpty(cp->size);

		if (isCellFull(cp))
			return qcGetFull(cp->size);

		verifyAlways(FALSE);
	}

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// No result was saved so we must calculate it now.
	// Call ourself for each of our quadrants.
	//
	QuadCell * ul = qcReduceStatesFunc(indexToQuadCell(cp->ul));
	QuadCell * ur = qcReduceStatesFunc(indexToQuadCell(cp->ur));
	QuadCell * ll = qcReduceStatesFunc(indexToQuadCell(cp->ll));
	QuadCell * lr = qcReduceStatesFunc(indexToQuadCell(cp->lr));

	//
	// Build the result and save it for reuse.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Raise the DEAD state and lower the NEW state to the ON state.
 * This is useful to make a mask out of the non-OFF cells.
 * Returns the same quadcell if nothing changes.
 */
QuadCell *
qcRaiseStates(QuadCell * cp)
{
	//
	// If the quadcell only contains OFF and ON states then
	// just return it now.
	//
	if (isCellOnlyOnOff(cp))
		return cp;

	qcSetResultType(RESULT_RAISE);

	return qcRaiseStatesFunc(cp);
}


/*
 * Recursive routine to raise all non-OFF cells of a quadcell to ON,
 * eliminating all NEW and DEAD states.  Returns the same quadcell if
 * nothing changes.
 */
static QuadCell *
qcRaiseStatesFunc(QuadCell * cp)
{
	//
	// If the quadcell only contains OFF and ON states then
	// just return it now.
	//
	if (isCellOnlyOnOff(cp))
		return cp;

	//
	// If the cell doesn't contain any OFF cells then
	// return a full cell.
	//
	if (!isCellPartOff(cp))
		return qcGetFull(cp->size);

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// No result was saved so we must calculate it now.
	// Call ourself for each of our quadrants.
	//
	QuadCell * ul = qcRaiseStatesFunc(indexToQuadCell(cp->ul));
	QuadCell * ur = qcRaiseStatesFunc(indexToQuadCell(cp->ur));
	QuadCell * ll = qcRaiseStatesFunc(indexToQuadCell(cp->ll));
	QuadCell * lr = qcRaiseStatesFunc(indexToQuadCell(cp->lr));

	//
	// Build the result and save it for reuse.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Extract the cells from a quadcell whose state is one of ones specified.
 * The states are the OR of the individual pure states.  The OFF state
 * cannot be specified.  Returns the original quadcell if everything was
 * extracted, or the empty cell if none of the states were present.
 */
QuadCell *
qcExtractStates(QuadCell * cp, int states)
{
	verifyAlways(!isStatePartOff(states))

	//
	// Check for cases in which nothing will change.
	//
	if ((cp->state & states) == 0)
		return qcEmpty();

	if ((cp->state & ~states) == 0)
		return cp;

	qcSetResultType(RESULT_NONE);

	return qcExtractStatesFunc(cp, states);
}


/*
 * Recursive routine to extract the cells from a quadcell whose state is
 * one of the ones specified.  The states are the OR of the individual pure
 * states.  The OFF state cannot be specified.  Returns the original
 * quadcell if everything was extracted, or the same size empty cell if
 * none of the states were present.
 */
static QuadCell *
qcExtractStatesFunc(QuadCell * cp, int states)
{
	//
	// Check for cases in which nothing will change.
	//
	if ((cp->state & states) == 0)
		return qcGetEmpty(cp->size);

	if ((cp->state & ~states) == 0)
		return cp;

	//
	// The above checks should have caught all single cells.
	//
	verifyAlways(cp->size > SIZE1);

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// No result was saved so we must calculate it now.
	// Call ourself for each of our quadrants.
	//
	QuadCell * ul = qcExtractStatesFunc(indexToQuadCell(cp->ul), states);
	QuadCell * ur = qcExtractStatesFunc(indexToQuadCell(cp->ur), states);
	QuadCell * ll = qcExtractStatesFunc(indexToQuadCell(cp->ll), states);
	QuadCell * lr = qcExtractStatesFunc(indexToQuadCell(cp->lr), states);

	//
	// Build the result and save it for reuse.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;

}


/*
 * Change all the cells in the quadcell having an old state to a new state.
 * Both of the states must be pure (OFF, DEAD, ON, NEW, or MARK).
 * If the new state is OFF then this effectively results in removing the
 * cells having the old state.  Returns a new quadcell with the result.
 * If nothing is changed then the original quadcell is returned.
 */
QuadCell *
qcChangeState(QuadCell * cp, int oldState, int newState)
{
	verifyAlways(isStatePure(oldState) && isStatePure(newState));

	//
	// Check for cases in which nothing will change.
	//
	if ((oldState == newState) || ((cp->state & oldState) == 0))
		return cp;

	qcSetResultType(RESULT_NONE);

	return qcChangeStateFunc(cp, oldState, newState);
}


/*
 * Recursive routine to change all the cells in the quadcell having an old
 * state to a new state.  Both of the states must be pure.  If the new
 * state is OFF then this effectively results in removing the cells having
 * the old state.  Returns a new cell of the same size with the result.
 * If nothing is changed then the original quadcell is returned.
 */
static QuadCell *
qcChangeStateFunc(QuadCell * cp, int oldState, int newState)
{
	//
	// If the quadcell has none of the old state then return itself.
	//
	if ((cp->state & oldState) == 0)
		return cp;

	//
	// If the quadcell is purely of the old state then return a
	// pure quadcell of the new state.
	//
	if (cp->state == oldState)
		return qcGetCellFromState(cp->size, newState);

	//
	// The above checks should have caught all single cells.
	//
	verifyAlways(cp->size > SIZE1);

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// No result was saved so we must calculate it now.
	// Call ourself for each of our quadrants.
	//
	QuadCell * ul = qcChangeStateFunc(
		indexToQuadCell(cp->ul), oldState, newState);

	QuadCell * ur = qcChangeStateFunc(
		indexToQuadCell(cp->ur), oldState, newState);

	QuadCell * ll = qcChangeStateFunc(
		indexToQuadCell(cp->ll), oldState, newState);

	QuadCell * lr = qcChangeStateFunc(
		indexToQuadCell(cp->lr), oldState, newState);

	//
	// Build the result and save it for reuse.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Register an external routine that is called when any of the internal
 * quadcell routines think that a gargage collection might be useful.
 * The callback routine can decide whether or not a garbage collection
 * is needed.  If so, it must first register its own quadcells using
 * qcRegister to protect them from the garbage collection, and then call
 * qcGarbageCollect to do the work.
 */
void
qcSetCallback(void (*routine)(void))
{
	callback = routine;
}


/*
 * Routine called by the internal quadcell code at times when a garbage
 * collection might be useful.  This is not guaranteed to do anything,
 * in particular if no callback routine had been set.  The callback
 * routine can perform garbage collection or not as it decides.
 * Before calling this, any temporary quadcells must be protected by
 * calling qcAddSafeQuadCell on them.
 */
void
qcCheckGarbageCollect(void)
{
	if (callback)
		callback();
}


/*
 * Add a temporary quadcell to the list of cells which are registered
 * when a garbage collection is done so that they aren't freed.  There
 * are a limited number of quadcells which can be saved in this way.
 * The qcClearSafeQuadCells routine must be called to release them.
 */
void
qcAddSafeQuadCell(QuadCell * cp)
{
	verifyAlways(safeQuadCellCount < SAFE_QUADCELL_SIZE);

	safeQuadCells[safeQuadCellCount++] = cp;
}


/*
 * Clear the specified number of the most recent temporary safe quadcells
 * when they are no longer needed to be preserved for garbage collection.
 * The special value of -1 clears them all.
 */
void
qcClearSafeQuadCells(int count)
{
	if (count == -1)
	{
		safeQuadCellCount = 0;

		return;
	}

	verifyAlways((count > 0) && (count <= safeQuadCellCount));

	safeQuadCellCount -= count;
}


/*
 * Register the safe quadcells so that they aren't garbage collected.
 */
static void
qcRegisterSafeQuadCells(void)
{
	int i;

	for (i = 0; i < safeQuadCellCount; i++)
		qcRegister(safeQuadCells[i]);

}


/*
 * Routine to register a quadcell as currently being used so that it
 * and its quadrants and results are not eligible for garbage collection.
 * This can be called multiple times for the same quadcell without harm.
 * Registration should only be done just prior to calling qcGarbageCollect
 * since it uses the flag element of the quadcells as a marker, and that
 * flag element can have other uses.  Registration is ignored for some
 * quadcells since they aren't garbage collected anyway.
 */
void
qcRegister(QuadCell * cp)
{
	if (cp->flag || (cp->size <= SIZE2) || isCellPure(cp))
		return;

	cp->flag = 1;

	qcRegister(indexToQuadCell(cp->ul));
	qcRegister(indexToQuadCell(cp->ur));
	qcRegister(indexToQuadCell(cp->ll));
	qcRegister(indexToQuadCell(cp->lr));

	if (isNotNull(cp->result))
		qcRegister(indexToQuadCell(cp->result));
}


/*
 * Perform a garbage collection for any non-registered quadcells so
 * that their memory can be reused for new quadcells.  All quadcells
 * held by the outside callers MUST be registered before calling this
 * routine otherwise those values will become indeterminate.  After calling
 * this routine all quadcells will be automatically unregistered.
 * This should be called occasionally when convenient, but not too often
 * because it does a lot of work.  All pure cells and all SIZE2 cells
 * are immune from garbage collection.
 */
void
qcGarbageCollect(void)
{
	int hashIndex;

	quadStats.garbageCollections++;

	//
	// Register any safe quadcells so that they aren't collected.
	//
	qcRegisterSafeQuadCells();

	//
	// Loop over all of the hash chains.
	//
	for (hashIndex = 0; hashIndex < hashTableSize; hashIndex++)
	{
		QuadCell * prevCp = 0;
		QUADINDEX index = hashTable[hashIndex];

		//
		// Walk down the hash chain checking all the cells.
		//
		while (index != INDEX_NULL)
		{
			QuadCell * cp = indexToQuadCell(index);
			QUADINDEX nextIndex = cp->next;

			if (cp->flag || (cp->size <= SIZE2) || isCellPure(cp))
			{
				//
				// This cell is registered as being used or
				// is special so clear its flag and leave it
				// in the hash chain.
				//
				cp->flag = 0;

				prevCp = cp;
			}
			else
			{
				//
				// This cell can be garbage collected.
				// Take it out of its hash chain and add it
				// to the front of the free list.
				//
				if (prevCp)
					prevCp->next = nextIndex;
				else
					hashTable[hashIndex] = nextIndex;

				hashTableUsed--;

				cp->next = freeList;
				freeList = index;

				quadStats.cellsFreed++;
				quadStats.cellsAvailable++;
			}

			index = nextIndex;
		}
	}

	//
	// Free up the quadpairs also while we are at it.
	//
	qcFreeQuadPairs();
}


/*
 * Calculate a new hash table size which is about 1.5 times
 * larger than the old one, and which is odd.  Also make sure
 * that it isn't divisible by a small prime.  This calculation
 * does something useful even if the old size was zero.
 */
int
qcNextTableSize(int oldSize)
{
	if (oldSize == 0)
		oldSize = 100000;

	int newSize = (((oldSize / 2) * 3) | 0x0f);

	while (((newSize % 3) == 0) || ((newSize % 5) == 0) ||
		((newSize % 7) == 0) || ((newSize % 11) == 0) ||
		((newSize % 13) == 0) || ((newSize % 17) == 0) ||
		((newSize % 19) == 0) || ((newSize % 23) == 0))
	{
		newSize += 2;
	}

	return newSize;
}


/*
 * Grow the hash table so that the average length of the chains is shorter.
 * The table is grown to be odd and be about 1.5 times its current size.
 * Note: This will change the hash indexes for the cells in the table and
 * reverse the order of the cells in the chains (which shouldn't matter much).
 * This can be called even if the hash table isn't yet allocated.
 */
void
qcGrowHashTable(void)
{
	QUADINDEX * newHashTable;
	int newUsed = 0;
	int tableIndex;

	quadStats.hashTableGrowths++;

	//
	// Calculate a new table size which is a bit larger then the old one.
	//
	int newSize = qcNextTableSize(hashTableSize);

	//
	// Allocate the new hash table and initialise its elements to
	// the null index to make each hash chain empty.
	//
	newHashTable = qcAllocMem("hash chain", sizeof(QUADINDEX) * newSize);

	for (tableIndex = 0; tableIndex < newSize; tableIndex++)
		newHashTable[tableIndex] = INDEX_NULL;

	/*
	 * Walk through the old hash chains and add their quadcells
	 * back into the new ones.
	 */
	for (tableIndex = 0; tableIndex < hashTableSize; tableIndex++)
	{
		QUADINDEX quadIndex = hashTable[tableIndex];

		//
		// Loop over the cells in the chain.
		//
		while (quadIndex != INDEX_NULL)
		{
			QuadCell * cp = indexToQuadCell(quadIndex);
			QUADINDEX nextIndex = cp->next;

			HASH hash = MAKE_HASH(cp->ul, cp->ur, cp->ll, cp->lr);
			int hashIndex = hash % newSize;

			cp->next = newHashTable[hashIndex];
			newHashTable[hashIndex] = quadIndex;

			newUsed++;
			quadIndex = nextIndex;
		}
	}

	/*
	 * Verify that the number of cells in the hash table is unchanged.
	 * Then free the old hash table if necessary and store the new one.
	 */
	verifyAlways(newUsed == hashTableUsed);

	if (hashTableSize > 0)
		free(hashTable);

	hashTable = newHashTable;
	hashTableSize = newSize;
}


/*
 * Return a copy of the quadcell statistics.
 */
QuadStats
qcGetStats(void)
{
	quadStats.hashTableSize = hashTableSize;
	quadStats.hashTableUsed = hashTableUsed;
	quadStats.cellsNextIndex = quadNextIndex;

	return quadStats;
}


/*
 * Clear the quadcell statistics.
 */
void
qcClearStats(void)
{
	QUADINDEX index = freeList;

	memset(&quadStats, 0, sizeof(quadStats));

	//
	// Recalculate the number of available cells.
	//
	if ((quadNextIndex % QUAD_ALLOC_SIZE) != 0)
	{
		quadStats.cellsAvailable =
			QUAD_ALLOC_SIZE - (quadNextIndex % QUAD_ALLOC_SIZE);
	}

	while (index != INDEX_NULL)
	{
		const QuadCell * cp = indexToQuadCell(index);
		index = cp->next;

		quadStats.cellsAvailable++;
	}
}


/*
 * Public routines to get information about a quadcell.
 * The internal routines just check directly.
 */
BOOL
qcIsIndexValid(QUADINDEX index)
{
	return ((index > INDEX_NULL) && (index < quadNextIndex));
}


BOOL
qcIsEmpty(const QuadCell * cp)
{
	return isCellEmpty(cp);
}


BOOL
qcIsFull(const QuadCell * cp)
{
	return isCellFull(cp);
}


BOOL
qcIsPure(const QuadCell *cp)
{
	return isCellPure(cp);
}


BOOL
qcIsOff(const QuadCell *cp)
{
	return isCellOff(cp);
}


BOOL
qcIsOn(const QuadCell *cp)
{
	return isCellOn(cp);
}


BOOL
qcIsDead(const QuadCell *cp)
{
	return isCellDead(cp);
}


BOOL
qcIsNew(const QuadCell *cp)
{
	return isCellNew(cp);
}


BOOL
qcIsStateUsed(const QuadCell * cp, int state)
{
	verifyAlways(isStatePure(state));

	return ((cp->state & state) != 0);
}


BOOL
qcIsReduced(const QuadCell * cp)
{
	return ((cp->state & ~(STATE_OFF|STATE_ON)) == 0);
}


int
qcGetSize(const QuadCell * cp)
{
	return cp->size;
}


NUMBER
qcGetMaxCoordinate(const QuadCell * cp)
{
	if (cp->size == SIZE1)
		return numZero();

	return numDec(numTwoPower(cp->size - 1));
}


NUMBER
qcGetMinCoordinate(const QuadCell * cp)
{
	if (cp->size == SIZE1)
		return numZero();

	return numNeg(numTwoPower(cp->size - 1));
}


/*
 * Return whether or not the specified coordinate lies within the
 * specified quadcell.  The coordinate can be either a row or a
 * column number.  For example, a size 3 quadcell has valid
 * coordinates from -4 to 3 inclusive.
 */
BOOL
qcIsCoordinateValid(const QuadCell * cp, NUMBER coord)
{
	if (cp->size == SIZE1)
		return numIsZero(coord);

	NUMBER limit = numTwoPower(cp->size - 1);

	if (numIsGreaterOrEqual(coord, limit))
		return FALSE;

	limit = numNeg(limit);

	return (numIsGreaterOrEqual(coord, limit));
}


/*
 * Return whether or not the cell at the specified row and column
 * lies within the specified quadcell.
 */
BOOL
qcIsCellValid(const QuadCell * cp, NUMBER row, NUMBER col)
{
	if (cp->size == SIZE1)
		return (numIsZero(row) && numIsZero(col));

	NUMBER limit = numTwoPower(cp->size - 1);

	//
	// If the row or column is out of bounds then the cell is OFF.
	//
	if (numIsGreaterOrEqual(row, limit) ||
		numIsGreaterOrEqual(col, limit))
	{
		return FALSE;
	}

	limit = numNeg(limit);

	if (numIsLess(row, limit) || numIsLess(col, limit))
		return FALSE;

	return TRUE;
}


/*
 * Public routine to return a quadcell given its index value.
 * This lets callers walk down through a quadcell structure if they want.
 */
QuadCell *
qcIndexToQuadCell(QUADINDEX index)
{
	return indexToQuadCell(index);
}


/*
 * Return the appropriate SIZE2 quadcell given the states of its
 * four quadrants.  The states must be one of the pure ones (OFF,
 * ON, DEAD, or NEW).
 */
QuadCell *
qcGetSize2Cell(int ulState, int urState, int llState, int lrState)
{
	int index = qcGetSize2Index(ulState, urState, llState, lrState);

	return size2Cells[index];
}


/*
 * Private routine to allocate memory.
 * This just exits on failure.
 */
void *
qcAllocMem(const char * what, int size)
{
	void * data = malloc(size);

	if (data == 0)
	{
		fprintf(stderr, "Cannot allocate memory for %s (size %d)\n",
			what, size);

		exit(1);
	}

	return data;
}

