/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Include file for public quadcell definition and external routines.
 * These are provided for the callers of the quadcell routines.
 * The definitions in the quadpriv include file are for internal use.
 *
 * Quadcells are not manipulated in the manner of most objects.  Each one
 * should be considered immutable (except for some housekeeping elements).
 * Instead an operation such as the setting or clearing of a cell within
 * them results in the creation of a NEW quadcell representing that change.
 *
 * Public routines which take or return coordinates do so with respect to
 * the center of the quad cell.  Coordinates are like in Cartesian geometry,
 * with positive values going right or up, and negative coordinates going
 * left or down.  The maximum and minimum coordinates within a quad cell
 * depend on its size, which specifies a power of two, as follows:
 *	min = -(2 ^ (size - 1))
 *	max = 2 ^ (size - 1) - 1
 * For example, a size 3 quadcell has coordinates from -4 to 3, inclusive.
 * Its bottom left corner is row -4, column -4.  Its top right corner is
 * row 3, column 3.
 *
 * Note: quadcells are allocated on demand and stored within a cache.
 * There is no routine to free a single quadcell when it is no longer needed.
 * Instead, the quadcells are garbage collected by the caller when necessary.
 * The caller must first mark all quadcells they are using using qcRegister
 * and then call qcGarbageCollect to reclaim all old quadcells for reuse.
 */
#ifndef	QUADLIFE_H
#define	QUADLIFE_H

#include <stdio.h>

#include "number.h"
#include "viewinfo.h"


/*
 * Booleans in case they aren't defined already.
 */
#ifndef BOOL
#define BOOL int
#endif

#ifndef	FALSE
#define	FALSE	((BOOL) 0)
#endif

#ifndef	TRUE
#define	TRUE	((BOOL) 1)
#endif


/*
 * State values for quadcells which are bit masks.
 * The SIZE1 quadcells have a pure state of OFF, DEAD, ON, NEW or MARK.
 * Larger quadcells can also have these pure states if all their subquadrants
 * have the same pure state.  Otherwise they have the OR of the states of
 * their subquadrants.  This means you can easily tell what types of cells a
 * quadcell contains.
 */
#define	STATE_NONE	0x00
#define	STATE_OFF	0x01
#define	STATE_DEAD	0x02
#define	STATE_ON	0x04
#define	STATE_NEW	0x08
#define	STATE_MARK	0x10

#define	STATE_LOWBIT	STATE_OFF
#define	STATE_HIGHBIT	STATE_MARK
#define	STATE_COUNT	5

//
// Combined states.
//
#define	STATE_ALL_EMPTY	(STATE_OFF | STATE_DEAD)
#define	STATE_ALL_ON	(STATE_ON | STATE_NEW | STATE_MARK)
#define	STATE_ALL	(STATED_OFF | STATE_ON | STATE_DEAD | STATE_NEW | STATE_MARK)
#define	STATE_ALL_NOT_OFF	(STATE_ON | STATE_DEAD | STATE_NEW | STATE_MARK)


//
// Macros to check the states for useful combinations.
// Pure means that every cell is composed of the same kind of state.
//
#define	isStatePure(state)	(((state) == STATE_OFF) || \
	((state) == STATE_DEAD) || ((state) == STATE_ON) || \
	((state) == STATE_NEW) || ((state) == STATE_MARK))

// Empty means that every cell is off in some manner, without any on cells.

#define	isStateEmpty(state)	(((state) == STATE_OFF) || \
	((state) == STATE_DEAD) || ((state) == (STATE_OFF|STATE_DEAD)))

#define	isStateNotEmpty(state)	(!isStateEmpty(state))

// Full means that every cell is on in some manner, without any off cells.

#define	isStateFull(state)	((((state) & STATE_ALL_ON) != 0) && \
	(((state) & ~STATE_ALL_ON) == 0))


#define	isStateOff(state)	((state) == STATE_OFF)
#define	isStateNotOff(state)	((state) != STATE_OFF)
#define	isStateOn(state)	((state) == STATE_ON)
#define	isStateDead(state)	((state) == STATE_DEAD)
#define	isStateNew(state)	((state) == STATE_NEW)
#define	isStateMark(state)	((state) == STATE_MARK)

#define	isStatePartOff(state)	(((state) & STATE_OFF) != 0)
#define	isStatePartOn(state)	(((state) & STATE_ON) != 0)
#define	isStatePartDead(state)	(((state) & STATE_DEAD) != 0)
#define	isStatePartNew(state)	(((state) & STATE_NEW) != 0)
#define	isStatePartMark(state)	(((state) & STATE_MARK) != 0)

#define	isStateOnlyOnOff(state)	(((state) & (~(STATE_ON|STATE_OFF))) == 0)
#define	isStatePartHistory(state) (((state) & (STATE_DEAD|STATE_NEW)) != 0)


//
// Short cuts to check whether a cell is empty or full.
// These handle non-pure combinations of the pure states.
//
#define	isCellEmpty(cp)		isStateEmpty((cp)->state)
#define	isCellNotEmpty(cp)	isStateNotEmpty((cp)->state)
#define	isCellFull(cp)		isStateFull((cp)->state)

#define	isCellPartOff(cp)	isStatePartOff((cp)->state)
#define	isCellPartOn(cp)	isStatePartOn((cp)->state)
#define	isCellPartDead(cp)	isStatePartDead((cp)->state)
#define	isCellPartNew(cp)	isStatePartNew((cp)->state)
#define	isCellPartMark(cp)	isStatePartMark((cp)->state)
#define	isCellOnlyOnOff(cp)	isStateOnlyOnOff((cp)->state)
#define	isCellPartHistory(cp)	isStatePartHistory((cp)->state)


//
// Short cuts for checking pure states.
//
#define	isCellPure(cp)		isStatePure((cp)->state)
#define	isCellOff(cp)		isStateOff((cp)->state)
#define	isCellNotOff(cp)	isStateNotOff((cp)->state)
#define	isCellOn(cp)		isStateOn((cp)->state)
#define	isCellDead(cp)		isStateDead((cp)->state)
#define	isCellNew(cp)		isStateNew((cp)->state)
#define	isCellMark(cp)		isStateMark((cp)->state)


/*
 * An index value for quadcells.
 * These are used because pointers take 8 bytes instead of the 4 bytes
 * that an index takes.  The quadcell can be easily found from its index.
 */
typedef	int QUADINDEX;


/*
 * A quadcell.
 * These contain the indexes of the four quadrants which make up the cell.
 * Each quadrant covers 1/4 of its parent's area, and has the size which
 * is one less than the parent.  quadcells are linked in hash chains based
 * on their quadrant index values.  Indexes into a table are used instead of
 * pointers to save memory on 64 bit machines (4 bytes each instead of 8).
 *
 * Quadcells have a state which partially or fully indicates what is contained
 * within its quadrants.  The state can be pure (OFF, ON, DEAD, NEW or MARK),
 * or else can be a mixed state containing the OR of some pure states.
 * Any pure state must match the state of all of its subquadrants, and so
 * it represents a large square of cells all in that state.  Mixed states
 * contain more than one type of cell in their subquadrants.
 *
 * Note: The number indexes are only set temporarily during any particular
 * calculation (e.g., qcGetMaxRow) since they are used to stop the recursion,
 * and must always be cleared when the operation finishes so that the next
 * operation which needs them can be performed correctly.
 */
typedef struct
{
	short size;		/* size of quadcell (which power of 2) */
	char flag;		/* flag used for bookkeeping */
	char state;		/* the state of the quadcell */
	QUADINDEX index;	/* index of this quadcell */
	QUADINDEX next;		/* next cell in the hash or free chain */
	QUADINDEX ul;		/* sub-cell in upper left quadrant */
	QUADINDEX ur;		/* sub-cell in upper right quadrant */
	QUADINDEX ll;		/* sub-cell in lower left quadrant */
	QUADINDEX lr;		/* sub-cell in lower right quadrant */
	QUADINDEX result;	/* result cell from a calculation */
	NUMINDEX numIndex;	/* index of number used during calculations */
} QuadCell;


/*
 * Information returned about a found quadcell from a qcGetRegion call.
 * The indicated square region of cells all have the same pure state
 * (ON, OFF, DEAD, NEW or MARK), and contains the cell that was specified.
 */
typedef struct
{
	int state;
	NUMBER minRow;
	NUMBER maxRow;
	NUMBER minCol;
	NUMBER maxCol;
} QuadRegion;


/*
 * Statistics
 */
#define	QUAD_STAT_SIZE	200

typedef	struct
{
	long	findEmpties;
	long	findFulls;
	long	findMisses;
	long	findHits;
	long	findLength;
	long	resultHits;
	long	resultMisses;
	long	garbageCollections;
	long	cellsFreed;
	long	cellsAllocated;
	long	cellsAvailable;
	long	cellsNextIndex;
	long	hashTableGrowths;
	long	hashTableSize;
	long	hashTableUsed;
	long	quadPairMisses;
	long	quadPairHits;
	long	regionOffCounts[QUAD_STAT_SIZE + 1];
	long	regionOnCounts[QUAD_STAT_SIZE + 1];
	long	sizes[QUAD_STAT_SIZE + 1];
} QuadStats;


/*
 * The maximum distance of a single spread operation.
 */
#define	QUAD_MAX_SPREAD 20


/*
 * Public routines.
 */
extern void qcInit(void);

extern QuadCell * qcEmpty(void);
extern QuadCell * qcIndexToQuadCell(QUADINDEX index);
extern QuadCell * qcFind(QUADINDEX ul, QUADINDEX ur, QUADINDEX ll, QUADINDEX lr);

extern BOOL qcIsIndexValid(QUADINDEX index);
extern BOOL qcIsEmpty(const QuadCell * cp);
extern BOOL qcIsFull(const QuadCell * cp);

extern BOOL qcIsPure(const QuadCell * cp);
extern BOOL qcIsOff(const QuadCell * cp);
extern BOOL qcIsOn(const QuadCell * cp);
extern BOOL qcIsDead(const QuadCell * cp);
extern BOOL qcIsNew(const QuadCell * cp);
extern int qcGetStateFromIndex(int quadIndex);

extern BOOL qcIsReduced(const QuadCell * cp);
extern BOOL qcIsStateUsed(const QuadCell * cp, int state);

extern int qcGetSize(const QuadCell * cp);
extern QuadCell * qcGetSize1CellFromState(int state);

extern NUMBER qcGetMaxCoordinate(const QuadCell * cp);
extern NUMBER qcGetMinCoordinate(const QuadCell * cp);

extern BOOL qcIsCoordinateValid(const QuadCell * cp, NUMBER coord);
extern BOOL qcIsCellValid(const QuadCell * cp, NUMBER row, NUMBER col);

extern QuadCell * qcFullDiamond(NUMBER radius);

extern QuadCell * qcGrowToSize(QuadCell * cp, int size);
extern QuadCell * qcGrowToContain(QuadCell * cp, NUMBER row, NUMBER col);
extern QuadCell * qcEmptyToContain(NUMBER row, NUMBER col);
extern QuadCell * qcTrim(QuadCell * cp);
extern QuadCell * qcTrimToSize(QuadCell * cp, int size);

extern QuadCell * qcCalculateGeneration(QuadCell * cp,
	NUMBER genCount, BOOL isEnabled);

extern QuadCell * qcSet(QuadCell * cp, NUMBER row, NUMBER col);
extern QuadCell * qcClear(QuadCell * cp, NUMBER row, NUMBER col);
extern QuadCell * qcClearToRight(QuadCell * cp, NUMBER col);
extern QuadCell * qcClearToLeft(QuadCell * cp, NUMBER col);
extern QuadCell * qcClearToTop(QuadCell * cp, NUMBER row);
extern QuadCell * qcClearToBottom(QuadCell * cp, NUMBER row);

extern QuadCell * qcReduceStates(QuadCell * cp);
extern QuadCell * qcRemoveHistoryStates(QuadCell * cp);
extern QuadCell * qcRaiseStates(QuadCell * cp);

extern QuadCell * qcExtractStates(QuadCell * cp, int states);
extern QuadCell * qcChangeState(QuadCell * cp, int oldState, int newState);

extern QuadCell * qcGetFullRegion(NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol);

extern QuadCell * qcExtractRegion(QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol);

extern QuadCell * qcExtractRow(QuadCell * cp, NUMBER row);
extern QuadCell * qcDeleteRow(QuadCell * cp, NUMBER row);

extern QuadCell * qcClearRegion(QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol);

extern QuadCell * qcFillRegion(QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol);

extern BOOL qcIsEqual(QuadCell * cp1, QuadCell * cp2);
extern BOOL qcIsOverlapping(QuadCell * cp1, QuadCell * cp2);

extern QuadCell * qcInvertOperation(QuadCell * cp);
extern QuadCell * qcOrOperation(QuadCell * cp1, QuadCell * cp2);
extern QuadCell * qcAndOperation(QuadCell * cp1, QuadCell * cp2);
extern QuadCell * qcXorOperation(QuadCell * cp1, QuadCell * cp2);
extern QuadCell * qcAndNotOperation(QuadCell * cp1, QuadCell * cp2);
extern QuadCell * qcCenter(QuadCell * cp);

extern QuadCell * qcShift(QuadCell * cp,
	NUMBER leftOrRightShift, NUMBER upOrDownShift);

extern QuadCell * qcLeftOrRightShift(QuadCell * cp, NUMBER shift);
extern QuadCell * qcUpOrDownShift(QuadCell * cp, NUMBER shift);

extern QuadCell * qcLeftShift(QuadCell * cp, NUMBER shift);
extern QuadCell * qcRightShift(QuadCell * cp, NUMBER shift);
extern QuadCell * qcUpShift(QuadCell * cp, NUMBER shift);
extern QuadCell * qcDownShift(QuadCell * cp, NUMBER shift);

extern QuadCell * qcRotateRight(QuadCell * cp, NUMBER row, NUMBER col,
	BOOL isOddSymmetry);

extern QuadCell * qcRotateLeft(QuadCell * cp, NUMBER row, NUMBER col,
	BOOL isOddSymmetry);

extern QuadCell * qcVerticalFlip(QuadCell * cp, NUMBER col, BOOL isOddSymmetry);
extern QuadCell * qcHorizontalFlip(QuadCell * cp, NUMBER row, BOOL isOddSymmetry);

extern QuadCell * qcSpread(QuadCell * cp, int distance);

extern QuadCell * qcExtractObject(QuadCell * cp,
	NUMBER row, NUMBER col, int distance);

extern QuadRegion qcGetRegion(const QuadCell * cp, NUMBER row, NUMBER col);

extern BOOL qcIsSet(const QuadCell * cp, NUMBER row, NUMBER col);
extern int qcGetState(const QuadCell * cp, NUMBER row, NUMBER col);
extern QuadCell * qcSetState(QuadCell * cp, NUMBER row, NUMBER col, int state);

extern int qcGetRegionStates(const QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol, int goalState);

extern BOOL qcIsRegionSet(const QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol);

extern BOOL qcIsRegionOff(const QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol);

extern NUMBER qcCalculateDistance(QuadCell * cp);
extern BOOL qcFindClosestCell(QuadCell * cp, NUMBER * retRow, NUMBER * retCol);
extern BOOL qcFindFirstCell(const QuadCell * cp, NUMBER * retRow, NUMBER * retCol);

extern BOOL qcVisitCells(QuadCell * cp, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data);

extern BOOL qcScanCells(QuadCell * cp, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data);

extern BOOL qcScanColumns(QuadCell * cp, NUMBER row, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data);

extern NUMBER qcGetCount(QuadCell * cp);
extern NUMBER qcGetStateCount(QuadCell * cp, int state);
extern NUMBER qcGetMaxRow(QuadCell * cp);
extern NUMBER qcGetMinRow(QuadCell * cp);
extern NUMBER qcGetMaxCol(QuadCell * cp);
extern NUMBER qcGetMinCol(QuadCell * cp);
extern NUMBER qcGetMaxRowUptoLimit(QuadCell * cp, NUMBER limitRow);
extern NUMBER qcGetMinRowDowntoLimit(QuadCell * cp, NUMBER limitRow);

extern void qcDump(QuadCell * cp);

extern QuadStats qcGetStats(void);
extern void qcClearStats(void);

extern ViewInfo qcCalculateView(const QuadCell * cp, int viewId, BOOL isAuxused,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

extern ViewInfo qcCalculateScaledView(const QuadCell * cp, int viewId,
	BOOL isAuxUsed, int scale, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol);

extern void qcWriteStructure(FILE * fp, QuadCell * cp);
extern QuadCell * qcReadStructure(FILE * fp);


//
// Garbage collection and memory usage routines.
//
extern void qcSetCallback(void (*routine)(void));
extern void qcCheckGarbageCollect(void);
extern void qcRegister(QuadCell * cp);
extern void qcAddSafeQuadCell(QuadCell * cp);
extern void qcClearSafeQuadCells(int count);
extern void qcGarbageCollect(void);
extern void qcGrowHashTable(void);
extern int qcGetPercentUsed(void);
extern long qcGetUsage(QuadCell * cp);

#endif

