/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines which implement a particular algorithm for manipulating
 * Life objects.
 */

#include "life.h"
#include "quadpriv.h"
#include "algorithm.h"
#include "quadalg.h"


typedef struct
{
	NUMBER row;
	NUMBER col;
}
SearchInfo;


/*
 * Functions to implement the life algorithm for quadcell based objects.
 */
LifeAlgorithm quadAlgorithm =
{
	qcAllocCellData,
	qcFreeCellData,
	qcGetCellState,
	qcSetCellState,
	qcIsCellSet,
	qcAddCell,
	qcDeleteCell,
	qcDeleteAllCells,
	qcShiftObject,
	qcSaveUsedCells,
	qcRegisterQuadCells,
	qcCopyObject,
	qcMoveObject,
	qcAddObject,
	qcMinMax,
	qcCountRegion,
	qcCalculateView2,
	qcCalculateCellCount,
	qcIsObjectEmpty,
	qcSearchObject,
	qcNextGeneration,
	qcRunGenerations,
	qcSelectObject,
	qcSelectAll,
	qcMarkSelection,
	qcSelectMarkedCells,
	qcInvertMarks,
	qcClearMarks,
	qcClearSearch,
	qcClearTracks,
	qcClearSelection,
	qcSelectPath,
	qcCountPathCells,
	qcFindMarkedGap,
	qcSelectCell,
	qcFindClosestSelectedCell,
	qcSelectRegion,
	qcMarkMinMax,
	qcCountMarks,
	qcMoveSelection,
	qcCopyMarkedObject,
	qcRotateSelection,
	qcFlipColumnSelection,
	qcFlipRowSelection,
	qcScanObject,
	qcGetObjectMaxRowUptoLimit,
	qcGetObjectMinRowDowntoLimit,
	qcExtractQuadCell
};


/*
 * The current rule being calculated.
 */
static const LifeRule * currentRule = 0;


//
// Forward definitions.
//
static void qcRemoveSelectedCells(LifeObject * obj);
static BOOL SearchCallback(NUMBER row, NUMBER col, int state, void * data);


/*
 * Allocate the data for the quadcell algorithm.
 * This just sets the quadcell elements of the object to an empty quadcell
 * and doesn't actually allocate any memory.  This can be called whenever
 * the cells of the object are cleared since nothing ever needs freeing.
 */
void
qcAllocCellData(LifeObject * obj)
{
	QuadCell * empty = qcEmpty();

	obj->quadCells = empty;
	obj->reducedQuadCells = empty;
	obj->selectedQuadCells = empty;
	obj->searchQuadCells = empty;
	obj->usedQuadCells = empty;
	obj->isReducedValid = FALSE;
	obj->isCountUnknown = TRUE;
}


void
qcFreeCellData(LifeObject * obj)
{
	qcAllocCellData(obj);
}


/*
 * Register the quadcells used by the object in preparation of a
 * garbage collection.
 */
void
qcRegisterQuadCells(LifeObject * obj)
{
	qcRegister(obj->quadCells);
	qcRegister(obj->reducedQuadCells);
	qcRegister(obj->selectedQuadCells);
	qcRegister(obj->searchQuadCells);
	qcRegister(obj->usedQuadCells);
}


/*
 * Return the current state of the cell at the given row and column.
 * This is one of OFF, DEAD, ON, NEW, or MARK.
 */
int
qcGetCellState(const LifeObject * obj, NUMBER row, NUMBER col)
{
	return qcGetState(obj->quadCells, row, col);
}


/*
 * Set the pure state of the cell at the given row and column.
 * This is one of OFF, DEAD, ON, NEW, or MARK.
 * This can turn a cell live or make it dead.
 */
void
qcSetCellState(LifeObject * obj, NUMBER row, NUMBER col, int state)
{
	qcSetQuadCell(obj, qcSetState(obj->quadCells, row, col, state));
}


/*
 * Return whether the cell at a given row and column is set.
 * NEW and MARK states are treated as ON.
 */
BOOL
qcIsCellSet(const LifeObject * obj, NUMBER row, NUMBER col)
{
	return qcIsSet(obj->quadCells, row, col);
}


/*
 * Create an ON cell at a given row and column.  Returns FALSE if the cell
 * already existed.  NEW and MARK states are treated as ON.
 */
BOOL
qcAddCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	//
	// Try to set the specified cell ON.
	// If the quadcell didn't change then the cell was already set.
	//
	QuadCell * newQuadCell = qcSet(obj->quadCells, row, col);

	if (qcIsEqual(newQuadCell, obj->quadCells))
		return FALSE;

	qcSetQuadCell(obj, newQuadCell);

	return TRUE;
}


/*
 * Delete a cell at a given coordinate.
 * Returns FALSE if it did not exist.
 */
BOOL
qcDeleteCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	QuadCell * newQuadCell = qcClear(obj->quadCells, row, col);

	if (qcIsEqual(newQuadCell, obj->quadCells))
		return FALSE;

	qcSetQuadCell(obj, newQuadCell);

	return TRUE;
}


/*
 * Delete all the cells of an object.
 * This also clears all marks and tracked cells.
 */
void
qcDeleteAllCells(LifeObject * obj)
{
	qcAllocCellData(obj);
}


/*
 * Shift the object left or right and up or down as specified.
 * This also shifts the marks to correspond.
 */
void
qcShiftObject(LifeObject * obj, NUMBER leftOrRight, NUMBER upOrDown)
{
	if (qcIsEmpty(obj->quadCells) ||
		(numIsZero(leftOrRight) && numIsZero(upOrDown)))
	{
		return;
	}

	qcSetQuadCell(obj, qcShift(obj->quadCells, leftOrRight, upOrDown));

	obj->selectedQuadCells = qcShift(obj->selectedQuadCells,
		leftOrRight, upOrDown);

	obj->searchQuadCells = qcShift(obj->searchQuadCells,
		leftOrRight, upOrDown);
}


/*
 * Add all of the current cells of an object to its set of used cells.
 * This is used to display the cells which have ever been used by an
 * object during its evolution.  We do nothing for the quadcell
 * algorithm since tracking data is kept within the quadcells.
 */
void
qcSaveUsedCells(LifeObject * obj)
{
}


/*
 * Calculate the view for the specified object.
 */
ViewInfo
qcCalculateView2(const LifeObject * obj, int viewId, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	ViewInfo info;

	if (scale == 1)
	{
		info = qcCalculateView(obj->quadCells, viewId, TRUE,
			minRow, maxRow, minCol, maxCol);
	}
	else
	{
		info = qcCalculateScaledView(obj->quadCells, viewId,
			TRUE, scale, minRow, maxRow, minCol, maxCol);
	}

	return info;
}


/*
 * Calculate from scratch the number of live cells in an object.
 * These are the ON, NEW, or MARK cells.
 * This does not reference the count or isUnknownCount elements.
 * The called function does what we want so we don't have to get the
 * reduced states of the object.
 */
NUMBER
qcCalculateCellCount(const LifeObject * obj)
{
	return qcGetCount(obj->quadCells);
}


/*
 * Return whether or not the object has no ON cells.
 * This makes no use of the object's cell count value.
 */
BOOL
qcIsObjectEmpty(const LifeObject * obj)
{
	return qcIsEmpty(obj->quadCells);
}


/*
 * Find the minimum and maximum row and column numbers for an object.
 * This includes dead cells if tracking is enabled.
 * Returns FALSE if the object has no cells.
 */
BOOL
qcMinMax(const LifeObject * obj,
	NUMBER * retMinRow, NUMBER * retMaxRow,
	NUMBER * retMinCol, NUMBER * retMaxCol)
{
	QuadCell * cp = obj->quadCells;

	if (qcIsEmpty(cp))
		return FALSE;

	*retMinRow = qcGetMinRow(cp);
	*retMaxRow = qcGetMaxRow(cp);
	*retMinCol = qcGetMinCol(cp);
	*retMaxCol = qcGetMaxCol(cp);

	return TRUE;
}


/*
 * Search forwards for the nth next object, restarting at the top if necessary.
 * This only manipulates OFF and ON cells.  Returns FALSE if the search was aborted.
 */
BOOL
qcSearchObject(LifeObject * obj, Count count)
{
	//
	// Get the current search quadcell for the object and make
	// sure it is still valid with respect to the object's cells.
	//
	QuadCell * searchCell = obj->searchQuadCells;

	searchCell = qcAndOperation(searchCell, qcGetReducedQuadCell(obj));

	//
	// If the search quadcell is empty then initialise it with
	// the whole object again for a fresh search.
	//
	if (isCellEmpty(searchCell))
		searchCell = qcGetReducedQuadCell(obj);

	if (isCellEmpty(searchCell))
		return FALSE;

	obj->searchQuadCells = searchCell;

	//
	// Now perform the start of a scan of the search cells to find
	// the first ON cell.  Note: The callback routine returns FALSE
	// to stop the scan after the first call to it.
	//
	SearchInfo info;

	info.row = numZero();
	info.col = numZero();

	qcScanCells(searchCell, STATE_ON, SearchCallback, &info);

	//
	// Select the connected object at this location allowing for
	// one cell gaps.
	//
	obj->selectedQuadCells = qcEmpty();

	if (!qcSelectObject(obj, info.row, info.col, 1))
		return FALSE;

	//
	// Remove the selected cells from the search quadcell
	// so that something else will be found next call.
	//
	obj->searchQuadCells = qcAndNotOperation(
		obj->searchQuadCells, obj->selectedQuadCells);

	//
	// Set the current position.
	//
	obj->currentInfo.row = info.row;
	obj->currentInfo.col = info.col;

	return TRUE;
}


/*
 * Callback from the search scan to save the first cell found.
 * This returns FALSE to stop the scan after the first cell.
 */
static BOOL
SearchCallback(NUMBER row, NUMBER col, int state, void * data)
{
	SearchInfo * info = data;

	info->row = row;
	info->col = col;

	return FALSE;
}


/*
 * Select the object at a given location as specified.  Returns FALSE if
 * no object exists there.  An object for this purpose is considered the
 * largest set of cells containing the specified cell whose members are
 * separated by a distance no larger then the specified gap.  For example,
 * a gap of zero means the cells which are king-wise connected to the
 * specified cell.  Note: This is a VERY expensive operation for quadcells
 * and so there is a limit on the size of the operation.
 * Warning: this routine can perform garbage collection.
 */
BOOL
qcSelectObject(LifeObject * obj, NUMBER row, NUMBER col, Value gap)
{
	QuadCell * cp = qcGetReducedQuadCell(obj);

	cp = qcExtractObject(cp, row, col, gap + 1);

	if (qcIsEmpty(cp))
		return FALSE;

	obj->selectedQuadCells = qcOrOperation(obj->selectedQuadCells, cp);

	//
	// See if we failed to mark it all.
	// If so then act as if the user told us to stop.
	//
	if (cp->flag)
	{
		cp->flag = 0;

		Life.stop = TRUE;
	}

	return TRUE;
}


/*
 * Select all of the cells of an object.
 * Returns TRUE if there were any cells to select.
 */
BOOL
qcSelectAll(LifeObject * obj)
{
	QuadCell * cp = qcGetReducedQuadCell(obj);

	if (qcIsEmpty(cp))
		return FALSE;

	obj->selectedQuadCells = cp;

	return TRUE;
}


/*
 * Mark the cells of the object which are in the previously made selection.
 * Returns FALSE if there was no selection.
 */
BOOL
qcMarkSelection(LifeObject * obj)
{
	QuadCell * cp = obj->selectedQuadCells;

	if (qcIsEmpty(cp))
		return FALSE;

	//
	// OR the set of marked cells into the object.
	// We know that the MARK state has the highest priority
	// so we don't have to remove the previous cells first.
	//
	cp = qcChangeState(cp, STATE_ON, STATE_MARK);

	qcSetQuadCell(obj, qcOrOperation(obj->quadCells, cp));

	return TRUE;
}


/*
 * Select the marked cells from an object.
 * The cells are added to the selection.
 * Returns FALSE if there were no marked cells.
 */
BOOL
qcSelectMarkedCells(LifeObject * obj)
{
	QuadCell * cp = qcExtractStates(obj->quadCells, STATE_MARK);

	if (qcIsEmpty(cp))
		return FALSE;

	//
	// OR the new selected cells into the selection with the ON state.
	//
	cp = qcChangeState(cp, STATE_MARK, STATE_ON);

	obj->selectedQuadCells = qcOrOperation(obj->selectedQuadCells, cp);

	return TRUE;
}


/*
 * Clear marks for a whole object.
 */
void
qcClearMarks(LifeObject * obj)
{
	qcSetQuadCell(obj, qcChangeState(obj->quadCells, STATE_MARK, STATE_ON));
}


/*
 * Clear the search state for an object.
 */
void
qcClearSearch(LifeObject * obj)
{
	obj->searchQuadCells = qcEmpty();
}


/*
 * Clear the tracks for an object.
 * This replaces all DEAD and NEW cells by OFF and ON.
 * The marked cells are left alone.
 */
void
qcClearTracks(LifeObject * obj)
{
	qcSetQuadCell(obj, qcRemoveHistoryStates(obj->quadCells));
}


/*
 * Clear the current selection for a whole object.
 */
void
qcClearSelection(LifeObject * obj)
{
	obj->selectedQuadCells = qcEmpty();
}


/*
 * Invert the marks for a whole object.
 * NEW and ON cells are changed to MARK, and MARK cells are changed to ON.
 * and DEAD cells remain as they were. 
 */
void
qcInvertMarks(LifeObject * obj)
{
	QuadCell * markCells = qcExtractStates(obj->quadCells, STATE_MARK);
	QuadCell * normalCells = qcExtractStates(obj->quadCells, STATE_ON|STATE_NEW);
	QuadCell * deadCells = qcExtractStates(obj->quadCells, STATE_DEAD);

	markCells = qcChangeState(markCells, STATE_MARK, STATE_ON);
	normalCells = qcChangeState(normalCells, STATE_ON, STATE_MARK);
	normalCells = qcChangeState(normalCells, STATE_NEW, STATE_MARK);

	QuadCell * newCells = qcOrOperation(markCells, normalCells);
	newCells = qcOrOperation(newCells, deadCells);

	qcSetQuadCell(obj, newCells);
}


/*
 * Count the number of marked cells in an object.
 */
NUMBER
qcCountMarks(const LifeObject * obj)
{
	return qcGetStateCount(obj->quadCells, STATE_MARK);
}


/*
 * Find the range of all marked cells for an object.
 * Returns FALSE if no cells were marked.
 */
BOOL
qcMarkMinMax(const LifeObject * obj,
	NUMBER * minRow, NUMBER * maxRow, NUMBER * minCol, NUMBER * maxCol)
{
	QuadCell * cells = qcExtractStates(obj->quadCells, STATE_MARK);

	*minRow = numZero();
	*maxRow = numZero();
	*minCol = numZero();
	*maxCol = numZero();

	if (qcIsEmpty(cells))
		return FALSE;

	*minRow = qcGetMinRow(cells);
	*maxRow = qcGetMaxRow(cells);
	*minCol = qcGetMinCol(cells);
	*maxCol = qcGetMaxCol(cells);

	return FALSE;
}


/*
 * Find the closest selected cell to the specified coordinate and return
 * its coordinates if found.  The measurement is the taxicab distance.
 * Returns TRUE if there is such a cell.
 */
BOOL
qcFindClosestSelectedCell(const LifeObject * obj,
	NUMBER baseRow, NUMBER baseCol, NUMBER * retRow, NUMBER * retCol)
{
	NUMBER row;
	NUMBER col;

	QuadCell * cp = obj->selectedQuadCells;

	if (qcIsEmpty(cp))
		return FALSE;

	//
	// Shift the quadcell to bring the base location to the origin.
	//
	cp = qcLeftOrRightShift(cp, numNeg(baseCol));
	cp = qcUpOrDownShift(cp, numNeg(baseRow));

	//
	// Now find the closest cell to the origin.
	//
	if (!qcFindClosestCell(cp, &row, &col))
		return FALSE;

	//
	// Adjust the resulting row and column numbers by the base
	// coordinates and return them.
	//
	*retRow = numAdd(row, baseRow);
	*retCol = numAdd(col, baseCol);

	return TRUE;
}


/*
 * Find the minimum distance between any marked cell to any unmarked cell.
 * The distance is computed as the taxicab distance (maximum of x and y).
 * Returns 0 if there are no marked cells or no unmarked cells.
 * This is an expensive operation so that there is a limit on the size
 * of the gap.  Warning: This can perform garbage collection.
 */
NUMBER
qcFindMarkedGap(const LifeObject * obj)
{
	QuadCell * cells = obj->quadCells;

	if (isCellEmpty(cells))
		return numZero();

	//
	// Get the set of marked cells.
	//
	QuadCell * marks = qcExtractStates(cells, STATE_MARK);

	if (isCellEmpty(marks))
		return numZero();

	//
	// Now get the set of unmarked cells.
	//
	cells = qcAndNotOperation(cells, marks);

	if (isCellEmpty(cells))
		return numZero();

	//
	// Now iterate by spreading the marked cells out one cell
	// at a time until they overlap the object.
	//
	int gap = 1;

	while ((gap < MAX_GAP) && !Life.stop)
	{
		marks = qcSpread(marks, 1);

		if (qcIsOverlapping(cells, marks))
			break;

		gap++;

		//
		// Protect our quadcells from garbage collection
		// and then try to do one.
		//
		qcAddSafeQuadCell(cells);
		qcAddSafeQuadCell(marks);
		qcCheckGarbageCollect();
		qcClearSafeQuadCells(2);
	}

	return numFromLong(gap);
}


/*
 * Rotate the previously selected cells in the given object around the
 * current cursor location.  This deletes the selected cells, and
 * reinserts them after their position has been rotated by 90 degrees
 * clockwise.  The selection is inserted using MARK cells.
 */
void
qcRotateSelection(LifeObject * obj)
{
	QuadCell * cp = obj->selectedQuadCells;

	if (qcIsEmpty(cp))
		return;

	qcRemoveSelectedCells(obj);

	//
	// Rotate the selected cells around the current location.
	//
	cp = qcRotateLeft(cp, obj->currentInfo.row,
		obj->currentInfo.col, TRUE);

	//
	// Insert the result back into the object in the MARK state.
	//
	cp = qcChangeState(cp, STATE_ON, STATE_MARK);

	qcSetQuadCell(obj, qcOrOperation(obj->quadCells, cp));
}


/*
 * Flip the previously selected cells in the given object around the row
 * of the current cursor location.  This deletes the selected cells,
 * and reinserts them after their position has been flipped around the
 * horizontal axis.  The cells are inserted in the MARK state.
 */
void
qcFlipRowSelection(LifeObject * obj)
{
	QuadCell * cp = obj->selectedQuadCells;

	if (qcIsEmpty(cp))
		return;

	qcRemoveSelectedCells(obj);

	//
	// Flip the selected cells around the current row.
	//
	cp = qcHorizontalFlip(cp, obj->currentInfo.row, TRUE);

	//
	// Insert the result back into the object in the MARK state.
	//
	cp = qcChangeState(cp, STATE_ON, STATE_MARK);

	qcSetQuadCell(obj, qcOrOperation(obj->quadCells, cp));
}


/*
 * Flip the selected cells in the given object around the column of the current
 * cursor location.  This deletes the selected cells, and reinserts them after
 * their position has been flipped around the vertical axis.  The inserted
 * cells have the MARK state.
 */
void
qcFlipColumnSelection(LifeObject * obj)
{
	QuadCell * cp = obj->selectedQuadCells;

	if (qcIsEmpty(cp))
		return;

	qcRemoveSelectedCells(obj);

	//
	// Flip the selected cells around the current column.
	//
	cp = qcVerticalFlip(cp, obj->currentInfo.col, TRUE);

	//
	// Insert the result back into the object in the MARK state.
	//
	cp = qcChangeState(cp, STATE_ON, STATE_MARK);

	qcSetQuadCell(obj, qcOrOperation(obj->quadCells, cp));
}


/*
 * Move the previously selected cells to another object.  If the destination
 * object is NULL the cells are just deleted.  The destination object has
 * already been cleared.  The source and destination objects must have the
 * same algorithm.
 */
void
qcMoveSelection(LifeObject * srcObj, LifeObject * destObj)
{
	CheckSameAlgorithm(srcObj, destObj);

	QuadCell * cp = srcObj->selectedQuadCells;

	if (qcIsEmpty(cp))
		return;

	qcRemoveSelectedCells(srcObj);

	//
	// Put the moved cells into the destination object if given.
	//
	if (destObj)
		qcSetQuadCell(destObj, cp);
}


/*
 * Copy marked cells to another object. The previous cells of the destination
 * object have already been cleared.  The source object is unaffected.
 * The source and destination object must have the same algorithm.
 */
void
qcCopyMarkedObject(const LifeObject * srcObj, LifeObject * destObj)
{
	CheckSameAlgorithm(srcObj, destObj);

	QuadCell * cp = qcExtractStates(srcObj->quadCells, STATE_MARK);

	//
	// Put the cells in the destination object.
	//
	qcSetQuadCell(destObj, cp);
}


/*
 * Select the cell at the specified location.
 * Returns TRUE if there was a cell at the specified location.
 * The selected cell is added to the current selection.
 */
BOOL
qcSelectCell(LifeObject * obj, NUMBER row, NUMBER col)
{
	if (!qcIsSet(obj->quadCells, row, col))
		return FALSE;

	obj->selectedQuadCells = qcSet(obj->selectedQuadCells, row, col);

	return TRUE;
}


/*
 * Select the cells in a specified rectangular region as desired.
 * The selected cells are added to the set already selected.
 * Returns TRUE if the region contained at least one cell to be selected.
 */
BOOL
qcSelectRegion(LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	QuadCell * cp = qcGetReducedQuadCell(obj);

	cp = qcExtractRegion(cp, minRow, maxRow, minCol, maxCol);

	if (qcIsEmpty(cp))
		return FALSE;

	obj->selectedQuadCells = qcOrOperation(obj->selectedQuadCells, cp);

	return TRUE;
}


/*
 * Count the cells in a specified rectangular region.
 * Returns the number of live cells in the region.
 */
NUMBER
qcCountRegion(const LifeObject * obj,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
// TODO: Fix const problem
	QuadCell * cp = qcGetReducedQuadCell((LifeObject *) obj);

	cp = qcExtractRegion(cp, minRow, maxRow, minCol, maxCol);

	return qcGetCount(cp);
}


/*
 * Try to compute one full generation of the current configuration.
 * Returns TRUE if a generation was run, or FALSE if the object was empty.
 * Note: This routine is not called.
 */
BOOL
qcNextGeneration(LifeObject * obj)
{
	return qcRunGenerations(obj, 1);
}


/*
 * Try to compute a number of generations for the specified object.
 * This also updates the number of cells which are born and died overall
 * and the cell count.  Returns the number of generations actually run.
 * Warning: This can perform garbage collections.
 */
Count
qcRunGenerations(LifeObject * obj, Count genCount)
{
	obj->born = numZero();
	obj->died = numZero();
	obj->growth = numZero();

	QuadCell * cp = obj->quadCells;

	if ((qcIsEmpty(cp) || genCount <= 0))
		return 0;

	//
	// We are committed to the calculation now.
	// See if the rule is being changed from previous calls,
	// and if so clear any saved results since they are invalid.
	//
	if (obj->rule != currentRule)
	{
		qcClearResults();
		currentRule = obj->rule;
	}

	//
	// Save the current count of cells if we are calculating
	// multiple generations.  We don't need the count when doing
	// only one generation because then we calculate the born and
	// died values.
	//
	NUMBER oldCount = numZero();

	if (genCount > 1)
		oldCount = CalculateCellCount(obj);

	//
	// Calculate the result of running generations.
	//
	QuadCell * result = qcCalculateGeneration(cp,
		numFromLong(genCount), obj->isTracksUsed);

	qcSetQuadCell(obj, result);
	obj->gen += genCount;

	//
	// If we ran just one generation then calculate the number of
	// cells born and died.  Otherwise calculate the difference in
	// the population between the generations.
	//
	if (genCount == 1)
	{
		obj->born = qcGetCount(qcAndNotOperation(result, cp));
		obj->died = qcGetCount(qcAndNotOperation(cp, result));
		obj->growth = numSub(obj->born, obj->died);
	}
	else
	{
		obj->growth = numSub(CalculateCellCount(obj), oldCount);
	}

	//
	// Try to do a garbage collection.
	//
	qcCheckGarbageCollect();

	return genCount;
}


/*
 * Move the cells from one object to another.  The source cells are
 * deleted, and the previous cells of the destination object are lost.
 * The source and destination objects must have the same algorithm.
 */
void
qcMoveObject(LifeObject * srcObj, LifeObject * destObj)
{
	CheckSameAlgorithm(srcObj, destObj);

	qcCopyObject(srcObj, destObj);
	qcAllocCellData(srcObj);
}


/*
 * Copy the cells from one object to another.  The source object is unchanged.
 * The current contents of the destination object are lost.  The source
 * and destination object must have the same algorithm.
 */
void
qcCopyObject(const LifeObject * srcObj, LifeObject * destObj)
{
	CheckSameAlgorithm(srcObj, destObj);

	qcFreeCellData(destObj);

	qcSetQuadCell(destObj, srcObj->quadCells);
}


/*
 * Add one object to another.  The source object is unchanged.
 * The destination object will get all cells from both objects.
 * Only the ON cells are added.  The object is displaced as specified.
 * The source and destination object must have the same algorithm.
 * All the added cells will be in the MARK state.
 */
void
qcAddObject(const LifeObject * srcObj, LifeObject * destObj,
	NUMBER rowDisp, NUMBER colDisp)
{
	CheckSameAlgorithm(srcObj, destObj);

	QuadCell * cells = srcObj->quadCells;

	if (qcIsEmpty(cells))
		return;

	cells = qcReduceStates(cells);

	cells = qcUpOrDownShift(cells, rowDisp);
	cells = qcLeftOrRightShift(cells, colDisp);
	cells = qcChangeState(cells, STATE_ON, STATE_MARK);

	qcSetQuadCell(destObj, qcOrOperation(cells, destObj->quadCells));

	AddWindowObjectUpdateFlags(destObj, U_ALL);
}


/*
 * Scan the current cells of an object, calling the indicated function
 * for each cell whose state is one of those specified, supplying a
 * specified piece of data that can be used for whatever purpose the called
 * routine wants.  The special states value of STATE_NONE means treat NEW
 * and MARK cells as being ON, and scan the ON cells.  If the routine
 * returns TRUE then the scan is continued, otherwise it is stopped.
 * This routine returns TRUE if the scan was completed, or FALSE if the
 * called function indicates that the scan is to be aborted.  The object
 * cannot be changed during the scan, neither can a garbage collection
 * be performed.  Warning: This should not be called for a very large
 * object since it can take a long time.
 */
BOOL
qcScanObject(LifeObject * obj, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data)
{
	//
	// Use either the real quadcells or the reduced one depending on
	// the state mask.  For convenience, the special value of STATE_NONE
	// includes NEW and MARK cells also, treating them as ON.
	//
	QuadCell * cp = obj->quadCells;

	if (states == STATE_NONE)
	{
		states = STATE_ON;
		cp = qcGetReducedQuadCell(obj);
	}

	return qcScanCells(cp, states, func, data);
}


/*
 * Return the highest row number in the object which is less than
 * or equal to the specified row.  If no such row exists then
 * the minimum negative number value is returned.
 */
NUMBER
qcGetObjectMaxRowUptoLimit(LifeObject * obj, NUMBER limitRow)
{
	QuadCell * cp = qcGetReducedQuadCell(obj);

	return qcGetMaxRowUptoLimit(cp, limitRow);
}


/*
 * Return the lowest row number in the object which is greater than
 * or equal to the specified row.  If no such row exists then
 * the maximum positive number value is returned.
 */
NUMBER
qcGetObjectMinRowDowntoLimit(LifeObject * obj, NUMBER limitRow)
{
	QuadCell * cp = qcGetReducedQuadCell(obj);

	return qcGetMinRowDowntoLimit(cp, limitRow);
}


/*
 * Remove the previously selected cells from the current cells of an object.
 * If tracking is turned on then the removed cells will changed to the
 * DEAD state.
 */
static void
qcRemoveSelectedCells(LifeObject * obj)
{
	QuadCell * cp = obj->selectedQuadCells;

	if (qcIsEmpty(cp))
		return;

	//
	// Delete the selected cells from the current object.
	//
	qcSetQuadCell(obj, qcAndNotOperation(obj->quadCells, cp));

	//
	// If tracks are enabled then replace the deleted cells
	// by DEAD cells.
	//
	if (obj->isTracksUsed)
	{
		QuadCell * cpDead = qcChangeState(cp, STATE_ON, STATE_DEAD);
		qcSetQuadCell(obj, qcOrOperation(obj->quadCells, cpDead));
	}
}


/*
 * Return the quadcell representation of an object.
 * This is easy since we are already in the format.
 */
QuadCell *
qcExtractQuadCell(LifeObject * obj)
{
	return obj->quadCells;
}


/*
 * Accessor function to return the current quadcells.
 * This is just for completeness.
 */
QuadCell *
qcGetQuadCell(LifeObject * obj)
{
	return obj->quadCells;
}


/*
 * Accessor function to set the current quadcells for an object.
 * If the current quadcells are changed, then the reduced quadcells are
 * marked invalid and will have to be recomputed when needed, and the
 * counter value is set unknown.
 */
void
qcSetQuadCell(LifeObject * obj, QuadCell * cp)
{
	if (cp != obj->quadCells)
	{
		obj->quadCells = cp;
		obj->reducedQuadCells = qcEmpty();
		obj->isReducedValid = FALSE;
		obj->isCountUnknown = TRUE;
	}
}


/*
 * Accessor function to return the reduced quadcells for the current cells.
 * This only contains OFF and ON states.  This calculates the quadcell if
 * it isn't already known.
 */
QuadCell *
qcGetReducedQuadCell(LifeObject * obj)
{
	if (isCellOnlyOnOff(obj->quadCells))
		return obj->quadCells;

	if (!obj->isReducedValid)
	{
		obj->reducedQuadCells = qcReduceStates(obj->quadCells);
		obj->isReducedValid = TRUE;
	}

	return obj->reducedQuadCells;
}

