/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Definitions for routines which implement a quadcell data structure
 * for manipulating Life objects.
 */

#ifndef	QUADALG_H
#define	QUADALG_H

#include "life.h"
#include "algorithm.h"


/*
 * The structure holding function pointers to the quadcell based routines.
 */
extern LifeAlgorithm quadAlgorithm;


/*
 * Global routines for manipulating quadcell based cell structures.
 * These are referenced by the function pointers in the LifeAlgorithm
 * structure.  Because of this, their usage is generic since a different
 * implementation can be used for the same functionality.
 */
extern	void qcAllocCellData(LifeObject * obj);
extern	void qcFreeCellData(LifeObject * obj);
extern	void qcRegisterQuadCells(LifeObject * obj);
extern	int qcGetCellState(const LifeObject * obj, NUMBER row, NUMBER col);
extern	void qcSetCellState(LifeObject * obj, NUMBER row, NUMBER col, int state);
extern	BOOL qcIsCellSet(const LifeObject * obj, NUMBER row, NUMBER col);
extern	BOOL qcAddCell(LifeObject *, NUMBER, NUMBER);
extern	BOOL qcDeleteCell(LifeObject *, NUMBER, NUMBER);
extern	void qcDeleteAllCells(LifeObject *);
extern	void qcShiftObject(LifeObject *, NUMBER, NUMBER);
extern	void qcSaveUsedCells(LifeObject *);
extern	void qcMoveObject(LifeObject * srcObj, LifeObject * destObj);
extern	void qcCopyObject(const LifeObject * srcObj, LifeObject * destObj);

extern	void qcAddObject(const LifeObject * srcObj, LifeObject * destObj,
	NUMBER rowDisp, NUMBER colDisp);

extern	NUMBER qcCalculateCellCount(const LifeObject * obj);
extern	BOOL qcIsObjectEmpty(const LifeObject * obj);
extern	BOOL qcSearchObject(LifeObject * obj, Count count);

extern	ViewInfo qcCalculateView2(const LifeObject * obj, int viewId, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

extern	BOOL qcMinMax(const LifeObject *, NUMBER *, NUMBER *, NUMBER *, NUMBER *);	
extern	BOOL qcSelectObject(LifeObject *, NUMBER, NUMBER, Value);
extern	BOOL qcSelectPath(LifeObject *);
extern	BOOL qcSelectAll(LifeObject *);
extern	BOOL qcMarkSelection(LifeObject *);
extern	BOOL qcSelectMarkedCells(LifeObject *);
extern	void qcClearSelection(LifeObject *);
extern	void qcClearSearch(LifeObject *);
extern	void qcClearTracks(LifeObject *);
extern	void qcClearMarks(LifeObject *);
extern	void qcInvertMarks(LifeObject *);
extern	NUMBER qcCountMarks(const LifeObject *);

extern	BOOL qcMarkMinMax(const LifeObject *,
	NUMBER *, NUMBER *, NUMBER *, NUMBER *);

extern	BOOL qcFindClosestSelectedCell(const LifeObject *,
	NUMBER, NUMBER, NUMBER *, NUMBER *);

extern	NUMBER qcFindMarkedGap(const LifeObject *);

extern	void qcRotateSelection(LifeObject *);
extern	void qcFlipRowSelection(LifeObject *);
extern	void qcFlipColumnSelection(LifeObject *);

extern	void qcMoveSelection(LifeObject *, LifeObject *);
extern	void qcCopyMarkedObject(const LifeObject *, LifeObject *);

extern	BOOL qcSelectCell(LifeObject *, NUMBER, NUMBER);
extern	BOOL qcSelectRegion(LifeObject *, NUMBER, NUMBER, NUMBER, NUMBER);
extern	NUMBER qcCountRegion(const LifeObject *, NUMBER, NUMBER, NUMBER, NUMBER);
extern	void qcCountPathCells(const LifeObject *, NUMBER *, NUMBER *, NUMBER *);

extern	BOOL qcNextGeneration(LifeObject * obj);
extern	Count qcRunGenerations(LifeObject * obj, Count genCount);

extern	BOOL qcScanObject(LifeObject * obj, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data);

extern	NUMBER qcGetObjectMaxRowUptoLimit(LifeObject * obj, NUMBER limitRow);
extern	NUMBER qcGetObjectMinRowDowntoLimit(LifeObject * obj, NUMBER limitRow);

extern QuadCell * qcExtractQuadCell(LifeObject * obj);

extern QuadCell * qcGetQuadCell(LifeObject * obj);
extern QuadCell * qcGetReducedQuadCell(LifeObject * obj);
extern void qcSetQuadCell(LifeObject * obj, QuadCell * cp);

#endif

