/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Functions that are calculated from the quadcell tree.
 */
#include "quad.h"
#include "quadpriv.h"


//
// Forward declarations.
//
static NUMBER qcGetCountFunc(QuadCell * cp);
static NUMBER qcGetStateCountFunc(QuadCell * cp, int state);
static NUMBER qcGetMaxRowFunc(QuadCell * cp);
static NUMBER qcGetMinRowFunc(QuadCell * cp);
static NUMBER qcGetMaxColFunc(QuadCell * cp);
static NUMBER qcGetMinColFunc(QuadCell * cp);

static NUMBER qcGetMaxRowUptoLimitFunc(QuadCell * cp, NUMBER limitRow);
static NUMBER qcGetMinRowDowntoLimitFunc(QuadCell * cp, NUMBER limitRow);


/*
 * Get the maximum row number which has at least one ON cell.
 * This routine needs to temporarily store numbers in the quadcells
 * so that values aren't repeatedly calculated during the recursion.
 * This can not be called for an empty quadcell.
 */
NUMBER
qcGetMaxRow(QuadCell * cp)
{
	verifyAlways(isCellNotEmpty(cp));

	//
	// Call the recursive routine to get the answer.
	//
	NUMBER maxRow = qcGetMaxRowFunc(cp);

	//
	// Adjust the coordinate to the center of our quadcell.
	//
	numAddrSub(&maxRow, numTwoPower(cp->size - 1));

	//
	// Now clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return maxRow;
}


/*
 * Get the maximum row number which has at least one ON cell and which is
 * smaller than or equal to the specified row.  If no such row exists then
 * the minimum negative number value is returned.
 * This routine needs to temporarily store pairs of numbers in the quadcells
 * so that values aren't repeatedly calculated during the recursion.
 * Each pair of numbers stores one value when the quadcell is cut by the
 * limit row, and another value when it is not cut.
 */
NUMBER
qcGetMaxRowUptoLimit(QuadCell * cp, NUMBER limitRow)
{
	//
	// If the quadcell is empty or if the limiting row is less
	// than the minimum coordinate then return the most negative number.
	// We do this since the normal coordinates can be both
	// positive and negative.
	//
	NUMBER halfSize = numTwoPower(cp->size - 1);
	NUMBER minCoord = qcGetMinCoordinate(cp);

	if (isCellEmpty(cp) || numIsLess(limitRow, minCoord))
		return numMinValue();

	//
	// Lower the limit row down to the maximum coordinate, and then
	// adjust it to be based from the bottom of the quadcell.
	//
	limitRow = numMin(limitRow, qcGetMaxCoordinate(cp));
	numAddrAdd(&limitRow, halfSize);

	//
	// Call the recursive routine to get the answer.
	// This will return a negative number is there is no cell found.
	//
	NUMBER maxRow = qcGetMaxRowUptoLimitFunc(cp, limitRow);

	//
	// Adjust the coordinate to the center of our quadcell or make
	// it less than the minimum coordinate according to whether
	// the row was found.
	//
	if (numIsNeg(maxRow))
		maxRow = numMinValue();
	else
		numAddrSub(&maxRow, halfSize);

	//
	// Clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return maxRow;
}


/*
 * Get the minimum row number which has at least one ON cell and which is
 * greater than or equal to the specified row.  If no such row exists then
 * the maximum positive number value is returned.
 * This routine needs to temporarily store pairs of numbers in the quadcells
 * so that values aren't repeatedly calculated during the recursion.
 * Each pair of numbers stores one value when the quadcell is cut by the
 * limit row, and another value when it is not cut.
 */
NUMBER
qcGetMinRowDowntoLimit(QuadCell * cp, NUMBER limitRow)
{
	//
	// If the quadcell is empty or if the limiting row is greater
	// than the maximum coordinate then return the most positive number.
	// We do this since the normal coordinates can be both
	// positive and negative.
	//
	NUMBER halfSize = numTwoPower(cp->size - 1);
	NUMBER maxCoord = qcGetMaxCoordinate(cp);

	if (isCellEmpty(cp) || numIsGreater(limitRow, maxCoord))
		return numMaxValue();

	//
	// Raise the limit row up to the minimum coordinate, and then
	// adjust it to be based from the top of the quadcell.
	//
	limitRow = numMax(limitRow, qcGetMinCoordinate(cp));
	limitRow = numSub(numDec(halfSize), limitRow);

	//
	// Call the recursive routine to get the answer.
	// This will return a negative number is there is no cell found.
	//
	NUMBER minRow = qcGetMinRowDowntoLimitFunc(cp, limitRow);

	//
	// Adjust the coordinate to the center of our quadcell or make
	// it less than the minimum coordinate according to whether
	// the row was found.
	//
	if (numIsNeg(minRow))
		minRow = numMaxValue();
	else
		minRow = numSub(numDec(halfSize), minRow);

	//
	// Clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return minRow;
}


/*
 * Get the minimum row number which has at least one ON cell.
 * This routine needs to temporarily store numbers in the quadcells
 * so that values aren't repeatedly calculated during the recursion.
 * This can not be called for an empty quadcell.
 */
NUMBER
qcGetMinRow(QuadCell * cp)
{
	verifyAlways(isCellNotEmpty(cp));

	//
	// Call the recursive routine to get the answer.
	//
	NUMBER minRow = qcGetMinRowFunc(cp);

	//
	// Adjust the coordinate to the center of our quadcell.
	//
	numAddrSub(&minRow, numTwoPower(cp->size - 1));

	//
	// Now clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return minRow;
}


/*
 * Get the maximum column number which has at least one ON cell.
 * This routine needs to temporarily store numbers in the quadcells
 * so that values aren't repeatedly calculated during the recursion.
 * This can not be called for an empty quadcell.
 */
NUMBER
qcGetMaxCol(QuadCell * cp)
{
	verifyAlways(isCellNotEmpty(cp));

	//
	// Call the recursive routine to get the answer.
	//
	NUMBER maxCol = qcGetMaxColFunc(cp);

	//
	// Adjust the coordinate to the center of our quadcell.
	//
	numAddrSub(&maxCol, numTwoPower(cp->size - 1));

	//
	// Now clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return maxCol;
}


/*
 * Get the minimum column number which has at least one ON cell.
 * This routine needs to temporarily store numbers in the quadcells
 * so that values aren't repeatedly calculated during the recursion.
 * This can not be called for an empty quadcell.
 */
NUMBER
qcGetMinCol(QuadCell * cp)
{
	verifyAlways(isCellNotEmpty(cp));

	//
	// Call the recursive routine to get the answer.
	//
	NUMBER minCol = qcGetMinColFunc(cp);

	//
	// Adjust the coordinate to the center of our quadcell.
	//
	numAddrSub(&minCol, numTwoPower(cp->size - 1));

	//
	// Now clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return minCol;
}


/*
 * Return the count of the number of live cells within the quadcell.
 * These are the ones having ON, NEW, or MARK states.
 * This value can get quite large, and so can exceed the size of a NUMBER.
 */
NUMBER
qcGetCount(QuadCell * cp)
{
	//
	// Call the recursive routine to get the answer.
	//
	NUMBER count = qcGetCountFunc(cp);

	//
	// Now clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return count;
}


/*
 * Return the count of the number of cells within the quadcell having
 * the specified state.  This value can get quite large, and so can
 * exceed the size of a NUMBER.
 */
NUMBER
qcGetStateCount(QuadCell * cp, int state)
{
	//
	// Call the recursive routine to get the answer.
	//
	NUMBER count = qcGetStateCountFunc(cp, state);

	//
	// Now clear out the numbers for their next use.
	//
	qcClearNumbers(cp);
	numFreeIndexes();

	return count;
}


/*
 * Recursive routine to get the maximum row number which has at
 * least one ON cell.  This should not be called for an empty quadcell.
 * NOTE: This doesn't return a coordinate based on the center of the quadcell.
 * Instead, it returns a coordinate relative to the bottom edge (so that the
 * bottom row in the quadcell has the coordinate zero).
 */
static NUMBER
qcGetMaxRowFunc(QuadCell * cp)
{
	NUMBER number1 = numZero();
	NUMBER number2 = numZero();

	//
	// If this is the single cell then the maximum row is zero.
	//
	if (cp->size == SIZE1)
		return numZero();

	//
	// If the cell is full then the maximum row is the topmost one.
	// Note: Coordinates in this routine are relative to the bottom edge.
	//
	if (isCellFull(cp))
		return numDec(numTwoPower(cp->size));

	//
	// If the maximum value is already known for this quadcell
	// then just return it.
	//
	if (isNumberIndex(cp->numIndex))
		return numFromIndex(cp->numIndex);

	//
	// Check the upper two quadrants first.
	// We will adjust their coordinates by the quadrant size.
	//
	NUMBER adjust = numTwoPower(cp->size - 1);

	BOOL isFound1 = FALSE;
	BOOL isFound2 = FALSE;

	QuadCell * ul = indexToQuadCell(cp->ul);

	if (isCellNotEmpty(ul))
	{
		number1 = qcGetMaxRowFunc(ul);
		numAddrAdd(&number1, adjust);
		isFound1 = TRUE;
	}

	QuadCell * ur = indexToQuadCell(cp->ur);

	if (isCellNotEmpty(ur))
	{
		number2 = qcGetMaxRowFunc(ur);
		numAddrAdd(&number2, adjust);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		cp->numIndex = numAllocIndex(max);

		return max;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// The upper two quadrants were empty so now we have to check the
	// bottom two quadrants.
	//
	QuadCell * ll = indexToQuadCell(cp->ll);

	if (isCellNotEmpty(ll))
	{
		number1 = qcGetMaxRowFunc(ll);
		isFound1 = TRUE;
	}

	QuadCell * lr = indexToQuadCell(cp->lr);

	if (isCellNotEmpty(lr))
	{
		number2 = qcGetMaxRowFunc(lr);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		cp->numIndex = numAllocIndex(max);

		return max;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// Fail since all of the quadrants were empty.
	//
	verifyAlways(FALSE);

	return numZero();
}


/*
 * Recursive routine to get the maximum row number which has at least
 * one ON cell and which is smaller than or equal to the specified row.
 * If no such row exists then a negative number is returned.
 *
 * NOTE: This doesn't accept or return a coordinate based on the center of
 * the quadcell.  Instead, it uses coordinates relative to the bottom edge
 * (so that the bottom row in the quadcell has the coordinate zero).
 *
 * Also note that the same quadcell can be seen in two circumstances.
 * Either it is not cut by the row limit, or it is cut.  If not cut,
 * then the quadcell's highest row is the answer.  But if it is cut,
 * then the highest row less than or equal to the row limit is the answer,
 * and the cut always occurs at the same row for the same quadcell.
 * So we only need to store either of two numbers for the quadcell in
 * order to avoid multiple recursions.
 */
NUMBER
qcGetMaxRowUptoLimitFunc(QuadCell * cp, NUMBER limitRow)
{
	//
	// If the limit row is negative or the cell is empty then
	// return a negative value.
	//
	if (numIsNeg(limitRow) || isCellEmpty(cp))
		return numNegOne();

	//
	// If this is the single cell then it is ON and so the
	// maximum row is zero.
	//
	if (cp->size == SIZE1)
		return numZero();

	//
	// If the cell is full then the result is the minimum of the
	// cell height and the limit row.  Note: Coordinates in this
	// routine are relative to the bottom edge.
	//
	NUMBER maxCoord = numDec(numTwoPower(cp->size));

	if (isCellFull(cp))
		return numMin(maxCoord, limitRow);

	//
	// The quadcell is neither empty nor full.
	// We need to do recursion on this quadcell for the answer.
	// Allocate a pair of numbers for this quadcell if we haven't yet.
	// The values of the pair are initialised to the most negative
	// number to indicate that no answers are stored yet.
	//
	if (isNullNumberIndex(cp->numIndex))
		cp->numIndex = numAllocMultipleIndexes(2);

	//
	// Now calculate whether or not the limit row cuts the quadcell.
	// Note that the limit row being at the top of the quadcell
	// doesn't count as a cut.
	//
	BOOL isCut = numIsLess(limitRow, maxCoord);

	//
	// See if we have a saved value for this quadcell in our pair
	// of allocated numbers.  For uncut quadcells we use the first
	// number of the pair, and for cut quadcells we use the second
	// number of the pair.  A result was actually saved if the value
	// is not the most negative number.  An uncut cell can never
	// have a negative saved value.
	//
	NUMINDEX numIndex = cp->numIndex + (isCut ? 1 : 0);

	NUMBER savedValue = numFromIndex(numIndex);

	if (!numIsMinValue(savedValue))
	{
		if (!isCut)
			verifyAlways(numSign(savedValue) >= 0);

		return savedValue;
	}

	//
	// We don't have a result saved for this cut quadcell so we
	// have to recurse on our quadrants to calculate the answer.
	// We will adjust their coordinates by the quadrant size.
	// Try to check the upper two quadrants first.  But don't check the
	// quadrants if this cell was cut and the limit row is below them.
	//
	NUMBER adjust = numTwoPower(cp->size - 1);
	NUMBER newLimitRow = numSub(limitRow, adjust);

	BOOL isLimitInUpperHalf = (!isCut || (numSign(newLimitRow) >= 0));

	BOOL isFound1 = FALSE;
	BOOL isFound2 = FALSE;

	NUMBER number1 = numZero();
	NUMBER number2 = numZero();

	QuadCell * ul = indexToQuadCell(cp->ul);

	if (isCellNotEmpty(ul) && isLimitInUpperHalf)
	{
		number1 = qcGetMaxRowUptoLimitFunc(ul, newLimitRow);

		if (numSign(number1) >= 0)
			numAddrAdd(&number1, adjust);

		isFound1 = (numSign(number1) >= 0);
	}

	QuadCell * ur = indexToQuadCell(cp->ur);

	if (isCellNotEmpty(ur) && isLimitInUpperHalf)
	{
		number2 = qcGetMaxRowUptoLimitFunc(ur, newLimitRow);

		if (numSign(number2) >= 0)
			numAddrAdd(&number2, adjust);

		isFound2 = (numSign(number2) >= 0);
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		numStoreToIndex(numIndex, max);

		return max;
	}

	if (isFound1)
	{
		numStoreToIndex(numIndex, number1);

		return number1;
	}

	if (isFound2)
	{
		numStoreToIndex(numIndex, number2);

		return number2;
	}

	//
	// The upper two quadrants were empty or above the limit row
	// so now we have to check the bottom two quadrants.
	//
	QuadCell * ll = indexToQuadCell(cp->ll);

	if (isCellNotEmpty(ll))
	{
		number1 = qcGetMaxRowUptoLimitFunc(ll, limitRow);
		isFound1 = (numSign(number1) >= 0);
	}

	QuadCell * lr = indexToQuadCell(cp->lr);

	if (isCellNotEmpty(lr))
	{
		number2 = qcGetMaxRowUptoLimitFunc(lr, limitRow);
		isFound2 = (numSign(number2) >= 0);
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		numStoreToIndex(numIndex, max);

		return max;
	}

	if (isFound1)
	{
		numStoreToIndex(numIndex, number1);

		return number1;
	}

	if (isFound2)
	{
		numStoreToIndex(numIndex, number2);

		return number2;
	}

	//
	// There were no cells found below or on the limit row.
	// If the cell was not cut this is an error since the cell
	// was known to be nonempty.
	//
	if (!isCut)
		verifyAlways(FALSE);

	//
	// A cut cell can fail to find an ON cell below the limit.
	// Save and return a negative row number to indicate this.
	//
	numStoreToIndex(numIndex, numNegOne());

	return numNegOne();
}


/*
 * Recursive routine to get the minimum row number which has at least
 * one ON cell and which is greater than or equal to the specified row.
 * If no such row exists then a negative number is returned.
 *
 * NOTE: This doesn't accept or return a coordinate based on the center of
 * the quadcell.  Instead, it uses coordinates relative to the top edge
 * (so that the top row in the quadcell has the coordinate zero).
 *
 * Also note that the same quadcell can be seen in two circumstances.
 * Either it is not cut by the row limit, or it is cut.  If not cut,
 * then the quadcell's lowest row is the answer.  But if it is cut,
 * then the lowest row greater than or equal to the row limit is the answer,
 * and the cut always occurs at the same row for the same quadcell.
 * So we only need to store either of two numbers for the quadcell in
 * order to avoid multiple recursions.
 */
NUMBER
qcGetMinRowDowntoLimitFunc(QuadCell * cp, NUMBER limitRow)
{
	//
	// If the limit row is negative or the cell is empty then
	// return a negative value.
	//
	if (numIsNeg(limitRow) || isCellEmpty(cp))
		return numNegOne();

	//
	// If this is the single cell then it is ON and so the
	// minimum row is zero.
	//
	if (cp->size == SIZE1)
		return numZero();

	//
	// If the cell is full then the result is the minimum of the
	// cell height and the limit row.  Note: Coordinates in this
	// routine are relative to the top edge.
	//
	NUMBER maxCoord = numDec(numTwoPower(cp->size));

	if (isCellFull(cp))
		return numMin(maxCoord, limitRow);

	//
	// The quadcell is neither empty nor full.
	// We need to do recursion on this quadcell for the answer.
	// Allocate a pair of numbers for this quadcell if we haven't yet.
	// The values of the pair are initialised to the most negative
	// number to indicate that no answers are stored yet.
	//
	if (isNullNumberIndex(cp->numIndex))
		cp->numIndex = numAllocMultipleIndexes(2);

	//
	// Now calculate whether or not the limit row cuts the quadcell.
	// Note that the limit row being at the bottom of the quadcell
	// doesn't count as a cut.
	//
	BOOL isCut = numIsLess(limitRow, maxCoord);

	//
	// See if we have a saved value for this quadcell in our pair
	// of allocated numbers.  For uncut quadcells we use the first
	// number of the pair, and for cut quadcells we use the second
	// number of the pair.  A result was actually saved if the value
	// is not the most negative number.  An uncut cell can never
	// have a negative saved value.
	//
	NUMINDEX numIndex = cp->numIndex + (isCut ? 1 : 0);

	NUMBER savedValue = numFromIndex(numIndex);

	if (!numIsMinValue(savedValue))
	{
		if (!isCut)
			verifyAlways(numSign(savedValue) >= 0);

		return savedValue;
	}

	//
	// We don't have a result saved for this cut quadcell so we
	// have to recurse on our quadrants to calculate the answer.
	// We will adjust their coordinates by the quadrant size.
	// Try to check the lower two quadrants first.  But don't check the
	// quadrants if this cell was cut and the limit row is below them.
	//
	NUMBER adjust = numTwoPower(cp->size - 1);
	NUMBER newLimitRow = numSub(limitRow, adjust);

	BOOL isLimitInLowerHalf = (!isCut || (numSign(newLimitRow) >= 0));

	BOOL isFound1 = FALSE;
	BOOL isFound2 = FALSE;

	NUMBER number1 = numZero();
	NUMBER number2 = numZero();

	QuadCell * ll = indexToQuadCell(cp->ll);

	if (isCellNotEmpty(ll) && isLimitInLowerHalf)
	{
		number1 = qcGetMinRowDowntoLimitFunc(ll, newLimitRow);

		if (numSign(number1) >= 0)
			numAddrAdd(&number1, adjust);

		isFound1 = (numSign(number1) >= 0);
	}

	QuadCell * lr = indexToQuadCell(cp->lr);

	if (isCellNotEmpty(lr) && isLimitInLowerHalf)
	{
		number2 = qcGetMinRowDowntoLimitFunc(lr, newLimitRow);

		if (numSign(number2) >= 0)
			numAddrAdd(&number2, adjust);

		isFound2 = (numSign(number2) >= 0);
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		numStoreToIndex(numIndex, max);

		return max;
	}

	if (isFound1)
	{
		numStoreToIndex(numIndex, number1);

		return number1;
	}

	if (isFound2)
	{
		numStoreToIndex(numIndex, number2);

		return number2;
	}

	//
	// The lower two quadrants were empty or below the limit row
	// so now we have to check the upper two quadrants.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);

	if (isCellNotEmpty(ul))
	{
		number1 = qcGetMinRowDowntoLimitFunc(ul, limitRow);
		isFound1 = (numSign(number1) >= 0);
	}

	QuadCell * ur = indexToQuadCell(cp->ur);

	if (isCellNotEmpty(ur))
	{
		number2 = qcGetMinRowDowntoLimitFunc(ur, limitRow);
		isFound2 = (numSign(number2) >= 0);
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		numStoreToIndex(numIndex, max);

		return max;
	}

	if (isFound1)
	{
		numStoreToIndex(numIndex, number1);

		return number1;
	}

	if (isFound2)
	{
		numStoreToIndex(numIndex, number2);

		return number2;
	}

	//
	// There were no cells found above or on the limit row.
	// If the cell was not cut this is an error since the cell
	// was known to be nonempty.
	//
	if (!isCut)
		verifyAlways(FALSE);

	//
	// A cut cell can fail to find an ON cell above the limit.
	// Save and return a negative row number to indicate this.
	//
	numStoreToIndex(numIndex, numNegOne());

	return numNegOne();
}


/*
 * Recursive routine to get the minimum (bottom) row number which has at
 * least one ON cell.  This should not be called for an empty quadcell.
 * NOTE: This doesn't return a coordinate based on the center of the quadcell.
 * Instead, it returns a coordinate relative to the bottom edge (so that the
 * bottom row in the quadcell has the coordinate zero).
 */
static NUMBER
qcGetMinRowFunc(QuadCell * cp)
{
	NUMBER number1 = numZero();
	NUMBER number2 = numZero();

	//
	// If this is the single cell or it is full then the minimum
	// row is zero (since coordinates in this routine are relative
	// to the bottom edge).
	//
	if ((cp->size == SIZE1) || isCellFull(cp))
		return numZero();

	//
	// If the minimum value is already known for this quadcell
	// then just return it.
	//
	if (isNumberIndex(cp->numIndex))
		return numFromIndex(cp->numIndex);

	//
	// Check the lower two quadrants first.
	//
	BOOL isFound1 = FALSE;
	BOOL isFound2 = FALSE;

	QuadCell * ll = indexToQuadCell(cp->ll);

	if (isCellNotEmpty(ll))
	{
		number1 = qcGetMinRowFunc(ll);
		isFound1 = TRUE;
	}

	QuadCell * lr = indexToQuadCell(cp->lr);

	if (isCellNotEmpty(lr))
	{
		number2 = qcGetMinRowFunc(lr);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER min = numMin(number1, number2);

		cp->numIndex = numAllocIndex(min);

		return min;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// The bottom two quadrants were empty so now we have to check the
	// top two quadrants.  We will adjust their coordinates by the
	// quadrant size.
	//
	NUMBER adjust = numTwoPower(cp->size - 1);

	QuadCell * ul = indexToQuadCell(cp->ul);

	if (isCellNotEmpty(ul))
	{
		number1 = qcGetMinRowFunc(ul);
		numAddrAdd(&number1, adjust);
		isFound1 = TRUE;
	}

	QuadCell * ur = indexToQuadCell(cp->ur);

	if (isCellNotEmpty(ur))
	{
		number2 = qcGetMinRowFunc(ur);
		numAddrAdd(&number2, adjust);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER min = numMin(number1, number2);

		cp->numIndex = numAllocIndex(min);

		return min;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// Fail since all of the quadrants were empty.
	//
	verifyAlways(FALSE);

	return numZero();
}


/*
 * Recursive routine to get the maximum column number which has at
 * least one ON cell.  This should not be called for an empty quadcell.
 * NOTE: This doesn't return a coordinate based on the center of the quadcell.
 * Instead, it returns a coordinate relative to the left edge (so that the
 * first column in the quadcell has the coordinate zero).
 */
static NUMBER
qcGetMaxColFunc(QuadCell * cp)
{
	NUMBER number1 = numZero();
	NUMBER number2 = numZero();

	//
	// If this is the single cell then the maximum column is zero.
	//
	if (cp->size == SIZE1)
		return numZero();

	//
	// If the cell is full then the maximum column is the rightmost one.
	// Note: Coordinates in this routine are relative to the left edge.
	//
	if (isCellFull(cp))
		return numDec(numTwoPower(cp->size));

	//
	// If the maximum value is already known for this quadcell
	// then just return it.
	//
	if (isNumberIndex(cp->numIndex))
		return numFromIndex(cp->numIndex);

	//
	// Check the right two quadrants first.
	// We will adjust their coordinates by the quadrant size.
	//
	NUMBER adjust = numTwoPower(cp->size - 1);

	BOOL isFound1 = FALSE;
	BOOL isFound2 = FALSE;

	QuadCell * ur = indexToQuadCell(cp->ur);

	if (isCellNotEmpty(ur))
	{
		number1 = qcGetMaxColFunc(ur);
		numAddrAdd(&number1, adjust);
		isFound1 = TRUE;
	}

	QuadCell * lr = indexToQuadCell(cp->lr);

	if (isCellNotEmpty(lr))
	{
		number2 = qcGetMaxColFunc(lr);
		numAddrAdd(&number2, adjust);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		cp->numIndex = numAllocIndex(max);

		return max;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// The right two quadrants were empty so now we have to check the
	// left two quadrants.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);

	if (isCellNotEmpty(ul))
	{
		number1 = qcGetMaxColFunc(ul);
		isFound1 = TRUE;
	}

	QuadCell * ll = indexToQuadCell(cp->ll);

	if (isCellNotEmpty(ll))
	{
		number2 = qcGetMaxColFunc(ll);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER max = numMax(number1, number2);

		cp->numIndex = numAllocIndex(max);

		return max;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// Fail since all of the quadrants were empty.
	//
	verifyAlways(FALSE);

	return numZero();
}


/*
 * Recursive routine to get the minimum column number which has at
 * least one ON cell.  This should not be called for an empty quadcell.
 * NOTE: This doesn't return a coordinate based on the center of the quadcell.
 * Instead, it returns a coordinate relative to the left edge (so that the
 * first column in the quadcell has the coordinate zero).
 */
static NUMBER
qcGetMinColFunc(QuadCell * cp)
{
	NUMBER number1 = numZero();
	NUMBER number2 = numZero();

	//
	// If this is the single cell or it is full then the minimum
	// column is zero (since coordinates in this routine are relative
	// to the left edge).
	//
	if ((cp->size == SIZE1) || isCellFull(cp))
		return numZero();

	//
	// If the minimum value is already known for this quadcell
	// then just return it.
	//
	if (isNumberIndex(cp->numIndex))
		return numFromIndex(cp->numIndex);

	//
	// Check the left two quadrants first.
	//
	BOOL isFound1 = FALSE;
	BOOL isFound2 = FALSE;

	QuadCell * ul = indexToQuadCell(cp->ul);

	if (isCellNotEmpty(ul))
	{
		number1 = qcGetMinColFunc(ul);
		isFound1 = TRUE;
	}

	QuadCell * ll = indexToQuadCell(cp->ll);

	if (isCellNotEmpty(ll))
	{
		number2 = qcGetMinColFunc(ll);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER min = numMin(number1, number2);

		cp->numIndex = numAllocIndex(min);

		return min;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// The left two quadrants were empty so now we have to check the
	// right two quadrants.  We will adjust their coordinates by the
	// quadrant size.
	//
	NUMBER adjust = numTwoPower(cp->size - 1);

	QuadCell * ur = indexToQuadCell(cp->ur);

	if (isCellNotEmpty(ur))
	{
		number1 = qcGetMinColFunc(ur);
		numAddrAdd(&number1, adjust);
		isFound1 = TRUE;
	}

	QuadCell * lr = indexToQuadCell(cp->lr);

	if (isCellNotEmpty(lr))
	{
		number2 = qcGetMinColFunc(lr);
		numAddrAdd(&number2, adjust);
		isFound2 = TRUE;
	}

	if (isFound1 && isFound2)
	{
		NUMBER min = numMin(number1, number2);

		cp->numIndex = numAllocIndex(min);

		return min;
	}

	if (isFound1)
	{
		cp->numIndex = numAllocIndex(number1);

		return number1;
	}

	if (isFound2)
	{
		cp->numIndex = numAllocIndex(number2);

		return number2;
	}

	//
	// Fail since all of the quadrants were empty.
	//
	verifyAlways(FALSE);

	return numZero();
}


/*
 * Recursive routine to get the count of live cells within this quadcell.
 * These are the ones having a state of ON, NEW, or MARK.
 * This count could get very large and so might overflow a NUMBER.
 */
static NUMBER
qcGetCountFunc(QuadCell * cp)
{
	//
	// If the quadcell is made up only of OFF or DEAD cells then
	// the count is zero.
	//
	if (isCellEmpty(cp))
		return numZero();

	//
	// If the quadcell is made up only of ON, NEW, or MARK cells
	// then the count is the square of the side, which is the
	// two-power of twice the size.
	//
	if (isCellFull(cp))
		return numTwoPower(cp->size * 2);

	//
	// SIZE1 cells should have been handled above.
	//
	verifyAlways(cp->size >= SIZE2);

	//
	// The quadcell is made of a mixture of on and off cells.
	// If the count value is already known for this quadcell
	// then just return it.
	//
	if (isNumberIndex(cp->numIndex))
		return numFromIndex(cp->numIndex);

	//
	// We have to calculate the count by recursing on our
	// quadrants and adding their counts together.
	//
	NUMBER count = qcGetCountFunc(indexToQuadCell(cp->ul));
	numAddrAdd(&count, qcGetCountFunc(indexToQuadCell(cp->ur)));
	numAddrAdd(&count, qcGetCountFunc(indexToQuadCell(cp->ll)));
	numAddrAdd(&count, qcGetCountFunc(indexToQuadCell(cp->lr)));

	//
	// Save the count value so it can be used again later.
	//
	cp->numIndex = numAllocIndex(count);

	return count;
}


/*
 * Recursive routine to get the count of cells within this quadcell
 * having the specified pure state.  This count could get very large
 * and so might overflow a NUMBER.
 */
static NUMBER
qcGetStateCountFunc(QuadCell * cp, int state)
{
	verifyAlways(isStatePure(state));

	//
	// Check for pure quadcells which have obvious counts.
	// Note: matching cells have a count which is the square of
	// their side, which is the two-power of twice their size.
	//
	if (isCellPure(cp))
	{
		if (cp->state != state)
			return numZero();

		return numTwoPower(cp->size * 2);
	}

	verifyAlways(cp->size >= SIZE2);

	//
	// If the quadcell doesn't contain any subquadrants with the
	// desired state then we can quit now.
	//
	if ((state & cp->state) == 0)
		return numZero();

	//
	// If the count value is already known for this quadcell
	// then just return it.
	//
	if (isNumberIndex(cp->numIndex))
		return numFromIndex(cp->numIndex);

	//
	// We have to calculate the count by recursing on our
	// quadrants and adding their counts together.
	//
	NUMBER count = qcGetStateCountFunc(indexToQuadCell(cp->ul), state);
	numAddrAdd(&count, qcGetStateCountFunc(indexToQuadCell(cp->ur), state));
	numAddrAdd(&count, qcGetStateCountFunc(indexToQuadCell(cp->ll), state));
	numAddrAdd(&count, qcGetStateCountFunc(indexToQuadCell(cp->lr),state));

	//
	// Save the count value so it can be used again later.
	//
	cp->numIndex = numAllocIndex(count);

	return count;
}


/*
 * Clear all the numbers from the specified tree of quadcells after a
 * calculation has set them.  This MUST be called after each calculation
 * is complete.  Warning: This assumes that an upper level quadcell
 * cannot have a null number index while any of its sub-level quadrants
 * do have a number index.
 */
void
qcClearNumbers(QuadCell * cp)
{
	if (isNullNumberIndex(cp->numIndex))
		return;

	cp->numIndex = INDEX_NULL_NUMBER;

	if (cp->size == SIZE1)
		return;

	qcClearNumbers(indexToQuadCell(cp->ul));
	qcClearNumbers(indexToQuadCell(cp->ur));
	qcClearNumbers(indexToQuadCell(cp->ll));
	qcClearNumbers(indexToQuadCell(cp->lr));
}

