/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Bare quadcell routines for calculating generations of an object.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "life.h"
#include "quad.h"
#include "quadpriv.h"


static int qcCalculateNewState(int cellA, int cellB, int cellC, int cellD,
	int cellE, int cellF, int cellG, int cellH, int cellI);


/*
 * Whether history is enabled when calculating generations.
 * This adds DEAD and NEW cells to base ON and OFF cells in the result.
 */
static BOOL isHistoryEnabled;


/*
 * Routines to find the quadrant sized cells formed by combining
 * parts of the original quadrants of a quadcell.
 */
static QuadCell *
TOPMERGE(QuadCell * cp)
{
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);

	return qcFind(ul->ur, ur->ul, ul->lr, ur->ll);
}


static QuadCell *
LEFTMERGE(QuadCell * cp)
{
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ll = indexToQuadCell(cp->ll);

	return qcFind(ul->ll, ul->lr, ll->ul, ll->ur);
}


static QuadCell *
RIGHTMERGE(QuadCell * cp)
{
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * lr = indexToQuadCell(cp->lr);

	return qcFind(ur->ll, ur->lr, lr->ul, lr->ur);
}


static QuadCell *
BOTTOMMERGE(QuadCell * cp)
{
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	return qcFind(ll->ur, lr->ul, ll->lr, lr->ll);
}


static QuadCell *
CENTERMERGE(QuadCell * cp)
{
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	return qcFind(ul->lr, ur->ll, ll->ur, lr->ul);
}


/*
 * Forward definitions.
 */
static QuadCell * qcSize4Result(QuadCell * cp);


/*
 * Calculate a new quadcell structure by advancing the specified
 * quadcell structure the specified number of generations.
 * A power of two generations is fastest to compute.
 * The isHistEnabled flag determines whether the DEAD or NEW cells
 * are to be produced.
 */
QuadCell *
qcCalculateGeneration(QuadCell * cp, NUMBER genCount, BOOL isHistEnabled)
{
	verifyAlways(numSign(genCount) >= 0);

	if (isCellEmpty(cp) || numIsZero(genCount))
		return cp;

	ResultType resultType = (isHistEnabled ? RESULT_GENHIST : RESULT_GEN);

	isHistoryEnabled = isHistEnabled;

	cp = qcTrim(cp);

	while (numIsPos(genCount))
	{
		/*
		 * Advance generations from the low to the high
		 * powers of two contained within the generation number.
		 */
		int bitNumber = numLowBitNumber(genCount);

		NUMBER twoPower = numTwoPower(bitNumber);
		genCount = numSub(genCount, twoPower);

		cp = qcGrowToSize(cp, cp->size + bitNumber + 2);

		qcSetResultTypeWithParam(resultType, twoPower);

		cp = qcBottomResult(cp, bitNumber);

		cp = qcTrim(cp);
	}

	return cp;
}


/*
 * Return the result of calculating life for this cell.
 * This is done by grouping together various quadrants, and combining
 * results from half-sized "scoops" to get the result.
 * Returns a cell structure whose size is 1/2 of the specified cell
 * structure and whose cells are moved forward in time by 1/4 generations.
 */
QuadCell *
qcResult(QuadCell * cp)
{
	QuadCell *	s1;		/* first 4 scoops */
	QuadCell *	s2;
	QuadCell *	s3;
	QuadCell *	s4;
	QuadCell *	s5;		/* next 5 scoops */
	QuadCell *	s6;
	QuadCell *	s7;
	QuadCell *	s8;
	QuadCell *	s9;
	QuadCell *	s10;		/* bottom 4 scoops */
	QuadCell *	s11;
	QuadCell *	s12;
	QuadCell *	s13;
	QuadCell *	t1;		/* temporaries */
	QuadCell *	t2;
	QuadCell *	t3;
	QuadCell *	t4;
	QuadCell *	t5;

	if (isCellOff(cp))
		return qcGetEmpty(cp->size - 1);

	/*
	 * If we already know the result then return it.
	 */
	if (isNotNull(cp->result))
	{
//		quadStatsresultHits++;

		return indexToQuadCell(cp->result);
	}

	/*
	 * For the lowest size (4 by 4), do the result by brute force.
	 */
	verifyAlways(cp->size >= SIZE4);

	if (cp->size == SIZE4)
		return qcSize4Result(cp);

	/*
	 * Scoop the four quadrants.
	 * All of these are of size 1/4.
	 */
	s1 = qcResult(indexToQuadCell(cp->ul));
	s2 = qcResult(indexToQuadCell(cp->ur));
	s3 = qcResult(indexToQuadCell(cp->ll));
	s4 = qcResult(indexToQuadCell(cp->lr));

	/*
	 * Rearrange the quadrants to get 5 more half sized quadcells.
	 * All of these are of size 1/2.
	 */
	t1 = TOPMERGE(cp);
	t2 = LEFTMERGE(cp);
	t3 = CENTERMERGE(cp);
	t4 = RIGHTMERGE(cp);
	t5 = BOTTOMMERGE(cp);

	/*
	 * Scoop the overlapping quadrants to excavate to a uniform level.
	 * All of these are of size 1/4.
	 */
	s5 = qcResult(t1);
	s6 = qcResult(t2);
	s7 = qcResult(t3);
	s8 = qcResult(t4);
	s9 = qcResult(t5);

	/*
	 * Rearrange the 9 quadrants into four half-sized cells.
	 * All of these are of size 1/2.
	 */
	t1 = qcFind(s1->index, s5->index, s6->index, s7->index);
	t2 = qcFind(s5->index, s2->index, s7->index, s8->index);
	t3 = qcFind(s6->index, s7->index, s3->index, s9->index);
	t4 = qcFind(s7->index, s8->index, s9->index, s4->index);

	/*
	 * Scoop each of those to get to the bottom level.
	 * All of these are of size 1/4.
	 */
	s10 = qcResult(t1);
	s11 = qcResult(t2);
	s12 = qcResult(t3);
	s13 = qcResult(t4);

	/*
	 * Combine the 4 quadrants of the bottom level.
	 * This results in a size of 1/2.
	 */
	QuadCell * result = qcFind(s10->index, s11->index, s12->index, s13->index);

	//
	// Save the result for reuse and return it.
	//
	cp->result = result->index;

//	quadStatsresultMisses++;

	return result;
}


/*
 * Return the result of calculating life for a size 4 quadcell.
 * This is done by actually calculating the next generation of the
 * central 2 by 2 square of the 4 by 4 cell to get a 2 by 2 answer.
 * This calculation is not time critical since it is only done a
 * maximum of 2^16 times per generation due to caching.
 * The cells in square groups of nine cells are defined as follows:
 *	abc
 *	hid
 *	gfe
 */
static QuadCell *
qcSize4Result(QuadCell * cp)
{
	int cellStates[4][4];

	verifyAlways(cp->size == SIZE4);

	/*
	 * If we already know the result then return it.
	 */
	if (isNotNull(cp->result))
	{
		return indexToQuadCell(cp->result);
	}

	/*
	 * Set up the matrix of actual states for the 4 by 4 area.
	 */
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	cellStates[0][0] = qcGetStateFromIndex(ul->ul);
	cellStates[0][1] = qcGetStateFromIndex(ul->ur);
	cellStates[0][2] = qcGetStateFromIndex(ur->ul);
	cellStates[0][3] = qcGetStateFromIndex(ur->ur);
	cellStates[1][0] = qcGetStateFromIndex(ul->ll);
	cellStates[1][1] = qcGetStateFromIndex(ul->lr);
	cellStates[1][2] = qcGetStateFromIndex(ur->ll);
	cellStates[1][3] = qcGetStateFromIndex(ur->lr);
	cellStates[2][0] = qcGetStateFromIndex(ll->ul);
	cellStates[2][1] = qcGetStateFromIndex(ll->ur);
	cellStates[2][2] = qcGetStateFromIndex(lr->ul);
	cellStates[2][3] = qcGetStateFromIndex(lr->ur);
	cellStates[3][0] = qcGetStateFromIndex(ll->ll);
	cellStates[3][1] = qcGetStateFromIndex(ll->lr);
	cellStates[3][2] = qcGetStateFromIndex(lr->ll);
	cellStates[3][3] = qcGetStateFromIndex(lr->lr);

	/*
	 * Get result for upper left central cell.
	 */
	int state1 = qcCalculateNewState(
		cellStates[0][0], cellStates[0][1], cellStates[0][2],
		cellStates[1][2], cellStates[2][2], cellStates[2][1],
		cellStates[2][0], cellStates[1][0], cellStates[1][1]);

	/*
	 * Get result for upper right central cell.
	 */
	int state2 = qcCalculateNewState(
		cellStates[0][1], cellStates[0][2], cellStates[0][3],
		cellStates[1][3], cellStates[2][3], cellStates[2][2],
		cellStates[2][1], cellStates[1][1], cellStates[1][2]);

	/*
	 * Get result for lower left central cell.
	 */
	int state3 = qcCalculateNewState(
		cellStates[1][0], cellStates[1][1], cellStates[1][2],
		cellStates[2][2], cellStates[3][2], cellStates[3][1],
		cellStates[3][0], cellStates[2][0], cellStates[2][1]);

	/*
	 * Get result for lower right central cell.
	 */
	int state4 = qcCalculateNewState(
		cellStates[1][1], cellStates[1][2], cellStates[1][3],
		cellStates[2][3], cellStates[3][3], cellStates[3][2],
		cellStates[3][1], cellStates[2][1], cellStates[2][2]);

	/*
	 * Use the calculated states to save and return the
	 * correct size 2 quadcell.
	 */
	QuadCell * result = qcGetSize2Cell(state1, state2, state3, state4);

	cp->result = result->index;

	return result;
}


/*
 * Calculate the new state for a cell given the nine states around it.
 * The cells in square groups of nine cells are defined as follows:
 *	abc
 *	hid
 *	gfe
 * Therefore, the old state for the cell is cellI.
 */
static int
qcCalculateNewState(int cellA, int cellB, int cellC, int cellD,
	int cellE, int cellF, int cellG, int cellH, int cellI)
{
	BOOL isLiveA = isStateFull(cellA);
	BOOL isLiveB = isStateFull(cellB);
	BOOL isLiveC = isStateFull(cellC);
	BOOL isLiveD = isStateFull(cellD);
	BOOL isLiveE = isStateFull(cellE);
	BOOL isLiveF = isStateFull(cellF);
	BOOL isLiveG = isStateFull(cellG);
	BOOL isLiveH = isStateFull(cellH);
	BOOL isLiveI = isStateFull(cellI);

	//
	// Calculate whether the new state is alive depending on
	// how the neighbors are alive.
	//
	BOOL isNewAlive = IsNewCellAlive(isLiveA, isLiveB, isLiveC,
		isLiveD, isLiveE, isLiveF, isLiveG, isLiveH, isLiveI);

	//
	// If the new cell is dead then return OFF or DEAD depending on
	// whether history is enabled.  However, preserve existing DEAD
	// states always as a convenience.
	//
	if (!isNewAlive)
	{
		if (isStateDead(cellI) ||
			(isHistoryEnabled && isStateNotOff(cellI)))
		{
			return STATE_DEAD;
		}

		return STATE_OFF;
	}

	//
	// The new cell is alive.
	// See if any of the neighbors are marked.
	// If so, then we are marked too since marks expand as
	// generations are run.
	//
	if (isStateMark(cellA) || isStateMark(cellB) ||
		isStateMark(cellC) || isStateMark(cellD) ||
		isStateMark(cellE) || isStateMark(cellF) ||
		isStateMark(cellG) || isStateMark(cellH) ||
		isStateMark(cellI))
	{
		return STATE_MARK;
	}

	//
	// If history is enabled and the cell was OFF then it is NEW,
	// otherwise it's just plain ON.
	//
	if (isHistoryEnabled && isStateOff(cellI))
		return STATE_NEW;

	return STATE_ON;
}


/*
 * Return the result of calculating life for the specified quadcell by
 * the specified power of two.  (e.g., a power of 0 advances 1 generation,
 * while a power of 3 advances 8 generations.)
 * This is done by grouping together various quadrants, and combining
 * results from half-sized "scoops" to get the result.
 * This is similar to the result of qcResult, except that the result calls
 * are delayed until the appropriate levels are processed.
 * The size of the cell structure must be large enough for the specified
 * power of two.
 */
QuadCell *
qcBottomResult(QuadCell * cp, int twoPower)
{
	QuadCell *	s1;		/* first 4 scoops */
	QuadCell *	s2;
	QuadCell *	s3;
	QuadCell *	s4;
	QuadCell *	s5;		/* next 5 scoops */
	QuadCell *	s6;
	QuadCell *	s7;
	QuadCell *	s8;
	QuadCell *	s9;
	QuadCell *	s10;		/* bottom 4 scoops */
	QuadCell *	s11;
	QuadCell *	s12;
	QuadCell *	s13;
	QuadCell *	t1;		/* temporaries */
	QuadCell *	t2;
	QuadCell *	t3;
	QuadCell *	t4;
	QuadCell *	t5;

	verifyAlways(twoPower >= 0);

	if (isCellOff(cp))
		return qcGetEmpty(cp->size - 1);

	/*
	 * If we already know the result then return it.
	 */
	if (isNotNull(cp->result))
	{
//		quadStatsresultHits++;

		return indexToQuadCell(cp->result);
	}

	/*
	 * If we are down to the level where we are calculating
	 * actual generations then do that.
	 */
	if (cp->size < twoPower + 3)
		return qcResult(cp);

	/*
	 * Make sure we aren't recursing too low.
	 */
	verifyAlways(cp->size > SIZE4);

	/*
	 * Collect the four quadrants.
	 * All of these are of size 1/4.
	 */
	s1 = CENTERMERGE(indexToQuadCell(cp->ul));
	s2 = CENTERMERGE(indexToQuadCell(cp->ur));
	s3 = CENTERMERGE(indexToQuadCell(cp->ll));
	s4 = CENTERMERGE(indexToQuadCell(cp->lr));

	/*
	 * Rearrange the quadrants to get 5 more half sized quadcells.
	 * All of these are of size 1/2.
	 */
	t1 = TOPMERGE(cp);
	t2 = LEFTMERGE(cp);
	t3 = CENTERMERGE(cp);
	t4 = RIGHTMERGE(cp);
	t5 = BOTTOMMERGE(cp);

	/*
	 * Collect the overlapping quadrants to excavate to a uniform level.
	 * All of these are of size 1/4.
	 */
	s5 = CENTERMERGE(t1);
	s6 = CENTERMERGE(t2);
	s7 = CENTERMERGE(t3);
	s8 = CENTERMERGE(t4);
	s9 = CENTERMERGE(t5);

	/*
	 * Rearrange the 9 quadrants into four half-sized cells.
	 * All of these are of size 1/2.
	 */
	t1 = qcFind(s1->index, s5->index, s6->index, s7->index);
	t2 = qcFind(s5->index, s2->index, s7->index, s8->index);
	t3 = qcFind(s6->index, s7->index, s3->index, s9->index);
	t4 = qcFind(s7->index, s8->index, s9->index, s4->index);

	/*
	 * Collect each of those to get to the bottom level.
	 * All of these are of size 1/4.
	 */
	s10 = qcBottomResult(t1, twoPower);
	s11 = qcBottomResult(t2, twoPower);
	s12 = qcBottomResult(t3, twoPower);
	s13 = qcBottomResult(t4, twoPower);

	/*
	 * Combine the 4 quadrants of the bottom level.
	 * This results in a size of 1/2.
	 */
	QuadCell * result = qcFind(s10->index, s11->index, s12->index, s13->index);

//	quadStatsresultMisses++;

	//
	// Save the result for reuse and return it.
	//
	cp->result = result->index;

	return result;
}

