/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Functions that return information about the quad tree.
 */
#include <stdio.h>
#include <memory.h>

#include "quad.h"
#include "quadpriv.h"
#include "life.h"

#define	REGION_left	1
#define	REGION_right	2
#define	REGION_top	3
#define	REGION_bottom	4

#define	MIX(a,b)	((a) * 5 + (b))


/*
 * Structure holding the status of a scan for the cells in a row.
 * These are the initial parameters of the call and a flag indicating
 * whether the scan was fully completed.
 */
typedef struct
{
	NUMBER row;
	int states;
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data);
	void * data;
	NUMBER count;
	BOOL isCompleted;
}
ScanInfo;


//
// Local routines.
//
static long qcGetUsageFunc(QuadCell * cp);
static void qcDumpFunc(QuadCell * cp, int indent);
static void qcStoreViewRegion(ViewInfo * info, const QuadRegion * region);

static void qcScanColumnsFunc(QuadCell * cp, NUMBER lowerRow, NUMBER leftCol,
	ScanInfo * scanInfo);

static BOOL qcVisitCellsFunc(QuadCell * cp, NUMBER row, NUMBER col, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data);

static int qcGetRegionStatesFunc(const QuadCell * cp,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol,
	int goalState);

static void qcNewCalculateScaledViewFunc(const QuadCell * cp,
	NUMBER minRow, NUMBER minCol, ViewInfo * info);


/*
 * Check whether two quadcells have the same set of ON cells.
 * Quadcells which could be trimmed compare as if they have been trimmed.
 * Quadcells which are shifted with respect to each other are not equal.
 * Note: This test relies on the fact that qcFind always returns the same
 * quadcell for the same inputs (as long as the quadcells remain in the
 * cache), and so the index element is a reliable way to check for equality.
 */
BOOL
qcIsEqual(QuadCell * cp1, QuadCell * cp2)
{
	//
	// First make the easy checks.
	// We can do these before the harder work of trimming.
	//
	if (cp1->index == cp2->index)
		return TRUE;
//TODO: Not quite right
	if (isCellEmpty(cp1) && isCellEmpty(cp2))
		return TRUE;

	if (isCellEmpty(cp1) || isCellEmpty(cp2))
		return FALSE;

	if (isCellFull(cp1) && isCellFull(cp2) && (cp1->size == cp2->size))
		return TRUE;

	if (isCellFull(cp1) || isCellFull(cp2))
		return FALSE;

	//
	// Now the quadcells might be aliases to each other if
	// either of them had been padded with blank quadrants.
	// Trim them both if possible.
	//
	cp1 = qcTrim(cp1);
	cp2 = qcTrim(cp2);

	//
	// Now they will be equal if and only if their indexes are the same.
	//
	return (cp1->index == cp2->index);
}


/*
 * Routine to determine whether the specified cell at the specified row
 * and column is ON or NEW.  The center of the quadcell has zero coordinates.
 * This could be a recursive routine but we can easily do it in a loop.
 */
BOOL
qcIsSet(const QuadCell * cp, NUMBER row, NUMBER col)
{
	//
	// If the cell is not within the quadcell then it is OFF.
	//
	if (!qcIsCellValid(cp, row, col))
		return FALSE;

	//
	// The cell is within the quadcell so we can walk down the
	// quadrants until we can determine the answer.
	//
	while (TRUE)
	{
		QUADINDEX index;
		NUMBER adjust;

		//
		// We know the answer if the quadcell is empty or full.
		//
		if (isCellEmpty(cp))
			return FALSE;

		if (isCellFull(cp))
			return TRUE;

		//
		// The cell must at least be two by two now since a
		// single cell must either be empty or full.
		//
		verifyAlways(cp->size > SIZE1);

		//
		// Select the appropriate quadrant to examine next
		// and adjust the coordinates within it.
		//
		if (cp->size == SIZE2)
			adjust = numZero();
		else
			adjust = numTwoPower(cp->size - 2);

		switch ((numSign(row) >= 0) * 2 + (numSign(col) >= 0))
		{
			case 0:
				numAddrAdd(&row, adjust);
				numAddrAdd(&col, adjust);
				index = cp->ll;
				break;

			case 1:
				numAddrAdd(&row, adjust);
				numAddrSub(&col, adjust);
				index = cp->lr;
				break;

			case 2:
				numAddrSub(&row, adjust);
				numAddrAdd(&col, adjust);
				index = cp->ul;
				break;

			case 3:
				numAddrSub(&row, adjust);
				numAddrSub(&col, adjust);
				index = cp->ur;
				break;

			default:
				verifyAlways(FALSE);
		}

		cp = indexToQuadCell(index);
	}
}


/*
 * Routine to return the pure state of a cell at the specified row and column.
 * The center of the quadcell has zero coordinates.  This could be a
 * recursive routine but we can easily do it in a loop.
 */
int
qcGetState(const QuadCell * cp, NUMBER row, NUMBER col)
{
	//
	// If the cell is not within the quadcell then it is OFF.
	//
	if (!qcIsCellValid(cp, row, col))
		return STATE_OFF;

	//
	// The cell is within the quadcell so we can walk down the
	// quadrants until we can determine the answer.
	//
	while (TRUE)
	{
		QUADINDEX index;
		NUMBER adjust;

		//
		// If the state of the quadcell is pure then that is
		// the cell's state.  This must be the case for a
		// SIZE1 quadcell.
		//
		if (isCellPure(cp))
			return cp->state;

		//
		// The cell must at least be two by two now since a
		// single cell must be pure.
		//
		verifyAlways(cp->size > SIZE1);

		//
		// Select the appropriate quadrant to examine next
		// and adjust the coordinates within it.
		//
		if (cp->size == SIZE2)
			adjust = numZero();
		else
			adjust = numTwoPower(cp->size - 2);

		switch ((numSign(row) >= 0) * 2 + (numSign(col) >= 0))
		{
			case 0:
				numAddrAdd(&row, adjust);
				numAddrAdd(&col, adjust);
				index = cp->ll;
				break;

			case 1:
				numAddrAdd(&row, adjust);
				numAddrSub(&col, adjust);
				index = cp->lr;
				break;

			case 2:
				numAddrSub(&row, adjust);
				numAddrAdd(&col, adjust);
				index = cp->ul;
				break;

			case 3:
				numAddrSub(&row, adjust);
				numAddrSub(&col, adjust);
				index = cp->ur;
				break;

			default:
				verifyAlways(FALSE);
		}

		cp = indexToQuadCell(index);
	}
}


/*
 * Routine to determine whether there is at least one ON or NEW cell within
 * the specified region.  This doesn't count how many ON cells are present
 * in the region, and so can return immediately once the first one is found.
 * This uses either recursive calls or a loop depending on the location of
 * the region with respect to our quadrants.
 */
BOOL
qcIsRegionSet(const QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol)
{
	//
	// If the region doesn't make sense then there is nothing in it.
	//
	if (numIsGreater(minRow, maxRow) || numIsGreater(minCol, maxCol))
		return FALSE;

	//
	// If the object is empty then nothing is in the region.
	//
loop:
	if (isCellEmpty(cp))
		return FALSE;

	//
	// Check for a region of only one cell.
	// We can do this slightly faster.
	//
	if (numIsEqual(minRow, maxRow) && numIsEqual(minCol, maxCol))
		return qcIsSet(cp, minRow, minCol);

	//
	// Calculate the minimum and maximum coordinates of the rows and
	// columns of this quadcell for easy comparions.  For example,
	// a size 3 quadcell has coordinates from -4 to 3 inclusive.
	//
	NUMBER halfSize = numTwoPower(cp->size - 1);
	NUMBER maxObjNum = numDec(halfSize);
	NUMBER minObjNum = numNeg(halfSize);

	//
	// Check whether the object overlaps the region at all.
	// If not then then nothing in the region is set.
	//
	if (numIsGreater(minRow, maxObjNum) ||
		numIsGreater(minCol, maxObjNum) ||
		numIsLess(maxRow, minObjNum) ||
		numIsLess(maxCol, minObjNum))
	{
		return FALSE;
	}

	//
	// The region does overlap the quadcell.
	// If the cell is full then there is an ON cell in the region.
	//
	if (isCellFull(cp))
		return TRUE;

	//
	// Check whether the object is totally contained within the region.
	// If so then something within the region is set (since we already
	// know the quadcell is non-empty).
	//
	if (numIsLessOrEqual(minRow, minObjNum) &&
		numIsLessOrEqual(minCol, minObjNum) &&
		numIsGreaterOrEqual(maxRow, maxObjNum) &&
		numIsGreaterOrEqual(maxCol, maxObjNum))
	{
		return TRUE;
	}

	//
	// The region and the quadcell partially overlap.  Make sure we
	// aren't at a single cell since a single cell must be either
	// entirely inside or outside of the region.
	//
	verifyAlways(cp->size > SIZE1);

	//
	// Trim the region to the boundaries of the quadcell.
	//
	minRow = numMax(minRow, minObjNum);
	maxRow = numMin(maxRow, maxObjNum);
	minCol = numMax(minCol, minObjNum);
	maxCol = numMin(maxCol, maxObjNum);

	//
	// Get the value needed to adjust the coordinates of the
	// region for use in examining the quadrants.
	//
	NUMBER adjust;

	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	//
	// If the whole region is within one of our quadrants then we
	// can just loop by checking the region against that single one
	// after adjusting the region coordinates appropriately.
	// Zero or positive signs are in the upper or right quadrants.
	// Negative signs are in the lower or left quadrants.
	//
	int upperSign = numSign(minRow);
	int lowerSign = numSign(maxRow);
	int rightSign = numSign(minCol);
	int leftSign = numSign(maxCol);

	//
	// Check for the region totally in the upper right quadrant.
	//
	if ((upperSign >= 0) && (rightSign >= 0))
	{
		cp = indexToQuadCell(cp->ur);
		numAddrSub(&minRow, adjust);
		numAddrSub(&maxRow, adjust);
		numAddrSub(&minCol, adjust);
		numAddrSub(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the lower left quadrant.
	//
	if ((lowerSign < 0) && (leftSign < 0))
	{
		cp = indexToQuadCell(cp->ll);
		numAddrAdd(&minRow, adjust);
		numAddrAdd(&maxRow, adjust);
		numAddrAdd(&minCol, adjust);
		numAddrAdd(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the upper left quadrant.
	//
	if ((upperSign >= 0) && (leftSign < 0))
	{
		cp = indexToQuadCell(cp->ul);
		numAddrSub(&minRow, adjust);
		numAddrSub(&maxRow, adjust);
		numAddrAdd(&minCol, adjust);
		numAddrAdd(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the lower right quadrant.
	//
	if ((lowerSign < 0) && (rightSign >= 0))
	{
		cp = indexToQuadCell(cp->lr);
		numAddrAdd(&minRow, adjust);
		numAddrAdd(&maxRow, adjust);
		numAddrSub(&minCol, adjust);
		numAddrSub(&maxCol, adjust);

		goto loop;
	}

	//
	// The region overlaps two or more of our quadrants.
	// Call ourself recursively for all of our quadrants with
	// the appropriate adjustments of the region coordinates.
	// (Those quadrants which don't overlap the region will
	// return FALSE without too much work.)
	//
	const QuadCell * ul = indexToQuadCell(cp->ul);
	const QuadCell * ur = indexToQuadCell(cp->ur);
	const QuadCell * ll = indexToQuadCell(cp->ll);
	const QuadCell * lr = indexToQuadCell(cp->lr);

	NUMBER addMinRow = numAdd(minRow, adjust);
	NUMBER addMaxRow = numAdd(maxRow, adjust);
	NUMBER addMinCol = numAdd(minCol, adjust);
	NUMBER addMaxCol = numAdd(maxCol, adjust);

	NUMBER subMinRow = numSub(minRow, adjust);
	NUMBER subMaxRow = numSub(maxRow, adjust);
	NUMBER subMinCol = numSub(minCol, adjust);
	NUMBER subMaxCol = numSub(maxCol, adjust);

	if (qcIsRegionSet(ul, subMinRow, subMaxRow, addMinCol, addMaxCol))
		return TRUE;

	if (qcIsRegionSet(ur, subMinRow, subMaxRow, subMinCol, subMaxCol))
		return TRUE;

	if (qcIsRegionSet(ll, addMinRow, addMaxRow, addMinCol, addMaxCol))
		return TRUE;

	if (qcIsRegionSet(lr, addMinRow, addMaxRow, subMinCol, subMaxCol))
		return TRUE;

	//
	// There were no ON cells within the region in any of the
	// quadrants.
	//
	return FALSE;
}


/*
 * Routine to determine whether all of the cells within the specified
 * region are OFF.  This uses either recursive calls or a loop depending
 * on the location of the region with respect to our quadrants.
 * As soon as we find a non-OFF cell in the region then we can return.
 */
BOOL
qcIsRegionOff(const QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol)
{
	//
	// If the region doesn't make sense then it is off.
	//
	if (numIsGreater(minRow, maxRow) || numIsGreater(minCol, maxCol))
		return TRUE;

loop:
	//
	// If the whole quadcell is OFF then so is the region.
	//
	if (isCellOff(cp))
		return TRUE;

	//
	// Check for a region of only one cell.
	// We can do this slightly faster.
	//
//	if (numIsEqual(minRow, maxRow) && numIsEqual(minCol, maxCol))
//		return qcIsSet(cp, minRow, minCol);

	//
	// Calculate the minimum and maximum coordinates of the rows and
	// columns of this quadcell for easy comparions.  For example,
	// a size 3 quadcell has coordinates from -4 to 3 inclusive.
	//
	NUMBER halfSize = numTwoPower(cp->size - 1);
	NUMBER maxObjNum = numDec(halfSize);
	NUMBER minObjNum = numNeg(halfSize);

	//
	// Check whether the quadcell overlaps the region at all.
	// If not then then nothing in the region is set.
	//
	if (numIsGreater(minRow, maxObjNum) ||
		numIsGreater(minCol, maxObjNum) ||
		numIsLess(maxRow, minObjNum) ||
		numIsLess(maxCol, minObjNum))
	{
		return TRUE;
	}

	//
	// The region does overlap the quadcell.
	// If the cell is pure then it is not OFF since we checked
	// for OFF cells above.
	//
	if (isCellPure(cp))
		return FALSE;

	//
	// Check whether the quadcell is totally contained within the region.
	// If so then something within the region is set (since we already
	// know the quadcell is non-OFF).
	//
	if (numIsLessOrEqual(minRow, minObjNum) &&
		numIsLessOrEqual(minCol, minObjNum) &&
		numIsGreaterOrEqual(maxRow, maxObjNum) &&
		numIsGreaterOrEqual(maxCol, maxObjNum))
	{
		return FALSE;
	}

	//
	// The region and the quadcell partially overlap.  Make sure we
	// aren't at a single cell since a single cell must be either
	// entirely inside or outside of the region.
	//
	verifyAlways(cp->size > SIZE1);

	//
	// Trim the region to the boundaries of the quadcell.
	//
	minRow = numMax(minRow, minObjNum);
	maxRow = numMin(maxRow, maxObjNum);
	minCol = numMax(minCol, minObjNum);
	maxCol = numMin(maxCol, maxObjNum);

	//
	// Get the value needed to adjust the coordinates of the
	// region for use in examining the quadrants.
	//
	NUMBER adjust;

	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	//
	// If the whole region is within one of our quadrants then we
	// can just loop by checking the region against that single one
	// after adjusting the region coordinates appropriately.
	// Zero or positive signs are in the upper or right quadrants.
	// Negative signs are in the lower or left quadrants.
	//
	int upperSign = numSign(minRow);
	int lowerSign = numSign(maxRow);
	int rightSign = numSign(minCol);
	int leftSign = numSign(maxCol);

	//
	// Check for the region totally in the upper right quadrant.
	//
	if ((upperSign >= 0) && (rightSign >= 0))
	{
		cp = indexToQuadCell(cp->ur);
		numAddrSub(&minRow, adjust);
		numAddrSub(&maxRow, adjust);
		numAddrSub(&minCol, adjust);
		numAddrSub(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the lower left quadrant.
	//
	if ((lowerSign < 0) && (leftSign < 0))
	{
		cp = indexToQuadCell(cp->ll);
		numAddrAdd(&minRow, adjust);
		numAddrAdd(&maxRow, adjust);
		numAddrAdd(&minCol, adjust);
		numAddrAdd(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the upper left quadrant.
	//
	if ((upperSign >= 0) && (leftSign < 0))
	{
		cp = indexToQuadCell(cp->ul);
		numAddrSub(&minRow, adjust);
		numAddrSub(&maxRow, adjust);
		numAddrAdd(&minCol, adjust);
		numAddrAdd(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the lower right quadrant.
	//
	if ((lowerSign < 0) && (rightSign >= 0))
	{
		cp = indexToQuadCell(cp->lr);
		numAddrAdd(&minRow, adjust);
		numAddrAdd(&maxRow, adjust);
		numAddrSub(&minCol, adjust);
		numAddrSub(&maxCol, adjust);

		goto loop;
	}

	//
	// The region overlaps two or more of our quadrants.
	// Call ourself recursively for all of our quadrants with
	// the appropriate adjustments of the region coordinates.
	// (Those quadrants which don't overlap the region will
	// return TRUE without too much work.)
	//
	const QuadCell * ul = indexToQuadCell(cp->ul);
	const QuadCell * ur = indexToQuadCell(cp->ur);
	const QuadCell * ll = indexToQuadCell(cp->ll);
	const QuadCell * lr = indexToQuadCell(cp->lr);

	NUMBER addMinRow = numAdd(minRow, adjust);
	NUMBER addMaxRow = numAdd(maxRow, adjust);
	NUMBER addMinCol = numAdd(minCol, adjust);
	NUMBER addMaxCol = numAdd(maxCol, adjust);

	NUMBER subMinRow = numSub(minRow, adjust);
	NUMBER subMaxRow = numSub(maxRow, adjust);
	NUMBER subMinCol = numSub(minCol, adjust);
	NUMBER subMaxCol = numSub(maxCol, adjust);

	if (!qcIsRegionOff(ul, subMinRow, subMaxRow, addMinCol, addMaxCol))
		return FALSE;

	if (!qcIsRegionOff(ur, subMinRow, subMaxRow, subMinCol, subMaxCol))
		return FALSE;

	if (!qcIsRegionOff(ll, addMinRow, addMaxRow, addMinCol, addMaxCol))
		return FALSE;

	if (!qcIsRegionOff(lr, addMinRow, addMaxRow, subMinCol, subMaxCol))
		return FALSE;

	//
	// There were no non-OFF cells within the region in any of the
	// quadrants.
	//
	return TRUE;
}


/*
 * Routine to return a bitmask representing which states of the pure
 * cells are present in the region (OFF, ON, DEAD, NEW, and MARK), and
 * are specified in goalState.  This doesn't count how many cells of each
 * state are present, and so can return immediately once all of the
 * states in the goal are found or not.
 */
int
qcGetRegionStates(const QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol, int goalState)
{
	//
	// If the region doesn't make sense then there is nothing in it.
	//
	if (numIsGreater(minRow, maxRow) || numIsGreater(minCol, maxCol))
		return STATE_NONE;

	//
	// Call the recursive routine to find the states, starting
	// with the goal of the states actually used in the quadcell.
	//
	goalState &= cp->state;

	return qcGetRegionStatesFunc(cp, minRow, maxRow,
		minCol, maxCol, goalState);
}


/*
 * Recursive routine to return a bitmask representing which types of
 * pure cells are present in the region (OFF, ON, DEAD, NEW and MARK).
 * This only looks at the quadcell tree until all the states in the
 * specified goalStates bit mask are either found or not found.
 * This uses either recursive calls or a loop depending on the
 * location of the region with respect to our quadrants.
 */
static int
qcGetRegionStatesFunc(const QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol, int goalState)
{
	int state = STATE_NONE;

loop:
	//
	// If this cell doesn't have any of the states remaining
	// in the goal then return the states we did find.
	// This also checks whether there is any goal left at all.
	//
	if ((cp->state & goalState) == 0)
		return state;

	//
	// Check for a region of only one cell.
	// We can do this slightly faster.
	//
	if (numIsEqual(minRow, maxRow) && numIsEqual(minCol, maxCol))
		return qcGetState(cp, minRow, minCol);

	//
	// Calculate the minimum and maximum coordinates of the rows and
	// columns of this quadcell for easy comparions.  For example,
	// a size 3 quadcell has coordinates from -4 to 3 inclusive.
	//
	NUMBER halfSize = numTwoPower(cp->size - 1);
	NUMBER maxObjNum = numDec(halfSize);
	NUMBER minObjNum = numNeg(halfSize);

	//
	// Check whether the object overlaps the region at all.
	// If not then just return the states we found.
	//
	if (numIsGreater(minRow, maxObjNum) ||
		numIsGreater(minCol, maxObjNum) ||
		numIsLess(maxRow, minObjNum) ||
		numIsLess(maxCol, minObjNum))
	{
		return state;
	}

	//
	// The region does overlap the quadcell.
	// If the quadcell is pure then the region certainly contains
	// that state, and there is no point looking at its subquadrants.
	// This also handles the SIZE1 quadcells.
	//
	if (isCellPure(cp))
	{
		state |= cp->state;

		return state;
	}

	verifyAlways(cp->size >= SIZE1);

	//
	// Trim the region to the boundaries of the quadcell.
	//
	minRow = numMax(minRow, minObjNum);
	maxRow = numMin(maxRow, maxObjNum);
	minCol = numMax(minCol, minObjNum);
	maxCol = numMin(maxCol, maxObjNum);

	//
	// Get the value needed to adjust the coordinates of the
	// region for use in examining the quadrants.
	//
	NUMBER adjust;

	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	//
	// If the whole region is within one of our quadrants then we
	// can just loop by checking the region against that single one
	// after adjusting the region coordinates appropriately.
	// Zero or positive signs are in the upper or right quadrants.
	// Negative signs are in the lower or left quadrants.
	//
	int upperSign = numSign(minRow);
	int lowerSign = numSign(maxRow);
	int rightSign = numSign(minCol);
	int leftSign = numSign(maxCol);

	//
	// Check for the region totally in the upper right quadrant.
	//
	if ((upperSign >= 0) && (rightSign >= 0))
	{
		cp = indexToQuadCell(cp->ur);
		numAddrSub(&minRow, adjust);
		numAddrSub(&maxRow, adjust);
		numAddrSub(&minCol, adjust);
		numAddrSub(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the lower left quadrant.
	//
	if ((lowerSign < 0) && (leftSign < 0))
	{
		cp = indexToQuadCell(cp->ll);
		numAddrAdd(&minRow, adjust);
		numAddrAdd(&maxRow, adjust);
		numAddrAdd(&minCol, adjust);
		numAddrAdd(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the upper left quadrant.
	//
	if ((upperSign >= 0) && (leftSign < 0))
	{
		cp = indexToQuadCell(cp->ul);
		numAddrSub(&minRow, adjust);
		numAddrSub(&maxRow, adjust);
		numAddrAdd(&minCol, adjust);
		numAddrAdd(&maxCol, adjust);

		goto loop;
	}

	//
	// Check for the region totally in the lower right quadrant.
	//
	if ((lowerSign < 0) && (rightSign >= 0))
	{
		cp = indexToQuadCell(cp->lr);
		numAddrAdd(&minRow, adjust);
		numAddrAdd(&maxRow, adjust);
		numAddrSub(&minCol, adjust);
		numAddrSub(&maxCol, adjust);

		goto loop;
	}

	//
	// The region overlaps two or more of our quadrants.
	// Call ourself recursively for all of our quadrants with
	// the appropriate adjustments of the region coordinates.
	// (Those quadrants which don't overlap the region or which
	// don't have the states left that we want will return
	// without too much work.)
	//
	const QuadCell * ul = indexToQuadCell(cp->ul);
	const QuadCell * ur = indexToQuadCell(cp->ur);
	const QuadCell * ll = indexToQuadCell(cp->ll);
	const QuadCell * lr = indexToQuadCell(cp->lr);

	NUMBER addMinRow = numAdd(minRow, adjust);
	NUMBER addMaxRow = numAdd(maxRow, adjust);
	NUMBER addMinCol = numAdd(minCol, adjust);
	NUMBER addMaxCol = numAdd(maxCol, adjust);

	NUMBER subMinRow = numSub(minRow, adjust);
	NUMBER subMaxRow = numSub(maxRow, adjust);
	NUMBER subMinCol = numSub(minCol, adjust);
	NUMBER subMaxCol = numSub(maxCol, adjust);

	//
	// Examine the subquadrants to see if they have the states
	// that we want.  If we have found them all then quit early.
	//
	state |= qcGetRegionStatesFunc(ul, subMinRow, subMaxRow,
		addMinCol, addMaxCol, goalState);

	goalState &= ~state;

	if (goalState == 0)
		return state;

	state |= qcGetRegionStatesFunc(ur, subMinRow, subMaxRow,
		subMinCol, subMaxCol, goalState);

	goalState &= ~state;

	if (goalState == 0)
		return state;

	state |= qcGetRegionStatesFunc(ll, addMinRow, addMaxRow,
		addMinCol, addMaxCol, goalState);

	goalState &= ~state;

	if (goalState == 0)
		return state;

	state |= qcGetRegionStatesFunc(lr, addMinRow, addMaxRow,
			subMinCol, subMaxCol, goalState);

	return state;
}


/*
 * Get information about the largest square region of cells around the
 * specified location whose cells all have the same pure setting (ON, OFF,
 * DEAD, NEW or MARK).  In the worst case this will return just one cell's
 * information but hopefully in many cases a larger quadcell region can
 * be returned.  This could call itself recursively for the smaller
 * quadcells, but we can easily do the required logic in a loop.  The
 * specified location must lie within the object.
 */
QuadRegion
qcGetRegion(const QuadCell * cp, NUMBER row, NUMBER col)
{
	QuadRegion region;
	NUMBER halfSize;
	NUMBER rowCenter;
	NUMBER colCenter;
	int rowRegion;
	int colRegion;

	/*
	 * Save the bounding coordinates of the starting quadcell and verify
	 * that the specified row and column are within those bounds.
	 */
	halfSize = numTwoPower(cp->size - 1);

	region.maxRow = numDec(halfSize);
	region.maxCol = numDec(halfSize);

	verifyAlways(numIsLess(row, halfSize) && numIsLess(col, halfSize));

	halfSize = numNeg(halfSize);

	region.minRow = halfSize;
	region.minCol = halfSize;

	verifyAlways(numIsGreaterOrEqual(row, halfSize) &&
		numIsGreaterOrEqual(col, halfSize));

	/*
	 * Now keep walking down the appropriate quadrants until we
	 * find one which is a pure state.  This must eventually
	 * succeed since any size 1 quadcell works.
	 */
loop:
	//
	// If this cell has a pure state then we can return its
	// current bounds as the region.  This has to hold at least
	// for the SIZE1 quadcell.
	//
	if (isCellPure(cp))
	{
		region.state = cp->state;

//		if (cp->size < QUAD_STAT_SIZE)
//			quadStats.regionOffCounts[cp->size]++;

//		if (cp->size < QUAD_STAT_SIZE)
//			quadStats.regionOnCounts[cp->size]++;

		return region;
	}

	verify(cp->size > SIZE1);

	/*
	 * This quadcell is not completely empty or full and so we
	 * must examine the appropriate quadrant.  First calculate
	 * the center coordinate of the quadcell by adding half of the
	 * cell's size to its minimum boundary coordinates.
	 */
	halfSize = numTwoPower(cp->size - 1);

	rowCenter = numAdd(region.minRow, halfSize);
	colCenter = numAdd(region.minCol, halfSize);

	/*
	 * Determine what region the row lies in (top or bottom half).
	 */
	if (numIsLess(row, rowCenter))
		rowRegion = REGION_bottom;
	else
		rowRegion = REGION_top;

	/*
	 * Determine what region the column lies in (left or right half).
	 */
	if (numIsLess(col, colCenter))
		colRegion = REGION_left;
	else
		colRegion = REGION_right;

	/*
	 * Select the proper quadrant to iterate on based on the row
	 * and column regions found above.  Also adjust the row and column
	 * values of the boundary depending on the quadrant picked.
	 */
	switch (MIX(rowRegion, colRegion))
	{
		case MIX(REGION_bottom, REGION_left):
			cp = indexToQuadCell(cp->ll);
			numAddrSub(&region.maxRow, halfSize);
			numAddrSub(&region.maxCol, halfSize);

			break;

		case MIX(REGION_top, REGION_left):
			cp = indexToQuadCell(cp->ul);
			numAddrAdd(&region.minRow, halfSize);
			numAddrSub(&region.maxCol, halfSize);

			break;

		case MIX(REGION_bottom, REGION_right):
			cp = indexToQuadCell(cp->lr);
			numAddrSub(&region.maxRow, halfSize);
			numAddrAdd(&region.minCol, halfSize);

			break;

		case MIX(REGION_top, REGION_right):
			cp = indexToQuadCell(cp->ur);
			numAddrAdd(&region.minRow, halfSize);
			numAddrAdd(&region.minCol, halfSize);

			break;

		default:
			verifyAlways(FALSE);
	}

	goto loop;
}


/*
 * Scan the ON, DEAD, NEW or MARK cells of the specified quadcell from the
 * min row to the max row, and within each row from the min column to the
 * max column.  The states parameter contains the OR of the state values
 * to be returned (with the OFF state ignored).  The specified function is
 * called with each pair of found row and column numbers and the state of
 * the found cell, along with a pointer to any information of interest to
 * the called function.  The called function must return TRUE to continue
 * scanning, or FALSE if the scan is to be aborted.  This routine returns
 * TRUE if the scan was completed, or FALSE if the called function wanted
 * to abort the scan.
 *
 * Warning: This routine could take a long time to run if the object is
 * very large and so it is a good idea to check that fact first.
 */
BOOL
qcScanCells(QuadCell * cp, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data)
{
	//
	// Iterate over the rows from the lowest to the highest.
	//
	NUMBER limitRow = numMinValue();

	while (TRUE)
	{
		// TODO: use a call which skips rows without our state.
		NUMBER row = qcGetMinRowDowntoLimit(cp, limitRow);

		if (numIsMaxValue(row))
			return TRUE;

		if (!qcScanColumns(cp, row, states, func, data))
			return FALSE;

		limitRow = numAdd(row, numOne());
	}
}


/*
 * Scan the cells of the specified row of the specified quadcell by calling
 * the specified function for each DEAD, ON, NEW, or MARK cell in the row,
 * according to the specified states value, which is the OR of the desired
 * states (with OFF being ignored).  The function is called with each pair
 * of found row and columm numbers and the state of the cell, along with a
 * pointer to any information of interest to the called function.
 * The called function must return TRUE to continue scanning, or FALSE if
 * the scan is to be aborted.  This function returns TRUE if the scan was
 * completed without aborting.
 *
 * Warning: This routine could take a long time to run if the object is
 * very large and so it is a good idea to check that fact first.
 */
BOOL
qcScanColumns(QuadCell * cp, NUMBER row, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data)
{
	//
	// First see if the row number is outside the quadcell or the
	// quadcell is empty.  If so then the row has no cells to return.
	// But the scan is still successful.
	//
	if (!qcIsCoordinateValid(cp, row) || isCellOff(cp))
		return TRUE;

	//
	// Initialise the information to be passed to the recursive routine.
	// The completed flag will be set to FALSE if any calls to the
	// user's function returns FALSE.
	//
	ScanInfo scanInfo;

	scanInfo.row = row;
	scanInfo.states = states & ~STATE_OFF;
	scanInfo.func = func;
	scanInfo.data = data;
	scanInfo.count = numZero();
	scanInfo.isCompleted = TRUE;

	//
	// Call the recursive function to get the answer with the
	// row and column numbers starting at the minimum coordinates.
	//
	NUMBER minCoord = qcGetMinCoordinate(cp);

	qcScanColumnsFunc(cp, minCoord, minCoord, &scanInfo);

	return scanInfo.isCompleted;
}


/*
 * Recursive function to examine a row and call the specified function for
 * each found cell.  We examine the cells in the row from left to right.
 * The desired row to examine is specified in the ScanInfo structure.
 * The incoming row and column coordinates are those of the bottom and left
 * edges of the quadcell.
 */
static void
qcScanColumnsFunc(QuadCell * cp, NUMBER lowerRow, NUMBER leftCol,
	ScanInfo * scanInfo)
{
	//
	// If this cell is OFF or doesn't contain any of the desired
	// states then return now.
	//
	if (isCellOff(cp) || ((cp->state & scanInfo->states) == 0))
		return;

	//
	// If the caller function had previously returned FALSE then
	// stop scanning for new cells.
	//
	if (!scanInfo->isCompleted)
		return;

	//
	// If this is a single cell then call the supplied function with
	// our coordinates and the supplied data.
	//
	if (cp->size == SIZE1)
	{
		//
		// When this size quadcell is reached the calculated
		// row number must match the desired row number.
		//
		verifyAlways(numIsEqual(lowerRow, scanInfo->row));

		//
		// Add to the number of found cells.
		//
		numAddrInc(&scanInfo->count);

		//
		// Call the specified function.
		// If it returns FALSE then set to stop the scan.
		//
		if (!scanInfo->func(lowerRow, leftCol, cp->state, scanInfo->data))
			scanInfo->isCompleted = FALSE;

		return;
	}

	//
	// This is a larger size cell and so has quadrants to examine.
	// Calculate the coordinates of the upper and rightmost halfs.
	//
	NUMBER halfSize = numTwoPower(cp->size - 1);

	NUMBER rightCol = numAdd(leftCol, halfSize);
	NUMBER upperRow = numAdd(lowerRow, halfSize);

	//
	// See if the destination row is in the upper or lower half of
	// this quadcell.  Whichever half is in, examine its left and
	// right quadrants in that order and call ourself recursively.
	//
	if (numIsLess(scanInfo->row, upperRow))
	{
		qcScanColumnsFunc(indexToQuadCell(cp->ll),
			lowerRow, leftCol, scanInfo);

		qcScanColumnsFunc(indexToQuadCell(cp->lr),
			lowerRow, rightCol, scanInfo);
	}
	else
	{
		qcScanColumnsFunc(indexToQuadCell(cp->ul),
			upperRow, leftCol, scanInfo);

		qcScanColumnsFunc(indexToQuadCell(cp->ur),
			upperRow, rightCol, scanInfo);
	}
}


/*
 * Visit all of the cells of the specified quadcell, calling the
 * specified function for each found cell.  The specified function is
 * called with each pair of found row and column numbers, along with a
 * pointer to any information of interest to the called function.
 * The called function must return TRUE to continue scanning, or FALSE
 * if the scan is to be aborted.  This routine returns TRUE if the scan
 * was completed, or FALSE if the called function wanted to abort the scan.
 *
 * This routine is very similar to the qcScanCells routine, except that
 * there is no order to the rows and columns that are visited, making
 * this routine much faster.
 *
 * Warning: This routine could take a long time to run if the object is
 * very large and so it is a good idea to check that fact first.
 */
BOOL
qcVisitCells(QuadCell * cp, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data)
{
	NUMBER minCoord = qcGetMinCoordinate(cp);

	states &= ~STATE_OFF;

	return qcVisitCellsFunc(cp, minCoord, minCoord, states, func, data);
}


/*
 * Recursive function to visit the ON cells of a quadcell.
 * The supplied row and column numbers are for the bottom and left edges
 * of the quadcell.  Returns TRUE if the visiting is to be continued,
 * or FALSE if the called function indicates that it wants to stop.
 */
static BOOL
qcVisitCellsFunc(QuadCell * cp, NUMBER row, NUMBER col, int states,
	BOOL (*func)(NUMBER row, NUMBER col, int state, void * data),
	void * data)
{
	//
	// Check if the quadcell is OFF or doesn't contain any cells
	// having one of the specified states.  If so, do nothing but
	// return TRUE to indicate that the visiting can continue.
	//
	if (isCellOff(cp) || ((cp->state & states) == 0))
		return TRUE;

	//
	// If this is a single cell then call the supplied function with
	// our coordinates and the supplied data.  Its return value
	// indicates whether we should stop.
	//
	if (cp->size == SIZE1)
		return func(row, col, cp->state, data);

	//
	// This is a larger size cell and so has quadrants to examine.
	// Calculate the coordinates of the upper and rightmost halfs.
	//
	NUMBER halfSize = numTwoPower(cp->size - 1);

	NUMBER rightCol = numAdd(col, halfSize);
	NUMBER upperRow = numAdd(row, halfSize);

	//
	// Recursively call ourself for our quadrants.
	// If any of the calls returns FALSE then so do we.
	//
	if (!qcVisitCellsFunc(indexToQuadCell(cp->ll),
		row, col, states, func, data))
	{
		return FALSE;
	}

	if (!qcVisitCellsFunc(indexToQuadCell(cp->lr),
		row, rightCol, states, func, data))
	{
		return FALSE;
	}

	if (!qcVisitCellsFunc(indexToQuadCell(cp->ul),
		upperRow, col, states, func, data))
	{
		return FALSE;
	}

	if (!qcVisitCellsFunc(indexToQuadCell(cp->ur),
		upperRow, rightCol, states, func, data))
	{
		return FALSE;
	}

	return TRUE;
}


static ViewInfo
qcNewCalculateScaledView(const QuadCell * cp, int viewId, BOOL isAuxUsed,
	int scale, NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);


/*
 * Calculate a view of the specified rectangular region of the specified object
 * and save the result in a returned memory array whose size is the indicated
 * number of rows times the indicated number of columns.  The cells in the
 * array contain values row by row and are 0 for OFF cells and 1 for ON cells.
 * The array can then be easily examined for ON cells by calling
 * FindNextRowIndex and FindNextRowIndex.  There is a limit on the size
 * of the rectangular region that can be used.  The returned array should
 * not be freed by the caller and is valid until the next call to this routine.
 */
ViewInfo
qcCalculateView(const QuadCell * cp, int viewId, BOOL isAuxUsed,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	if (Life.debugOption == 0)
	{
		return qcNewCalculateScaledView(cp, viewId, isAuxUsed,
			1, minRow, maxRow, minCol, maxCol);
	}

	ViewInfo info;

	//
	// Initialise the view structure.
	//
	InitView(&info, viewId, isAuxUsed, 1, minRow, maxRow, minCol, maxCol);

	NUMBER cpMinRow;
	NUMBER cpMaxRow;
	NUMBER cpMinCol;
	NUMBER cpMaxCol;
	int rowIndex;
	int colIndex;

	/*
	 * Get the boundaries of the quadcell so that we can quickly
	 * weed out those coordinates which aren't in it.
	 */
	NUMBER halfSize = numTwoPower(cp->size - 1);

	cpMaxRow = numDec(halfSize);
	cpMaxCol = numDec(halfSize);

	halfSize = numNeg(halfSize);

	cpMinRow = halfSize;
	cpMinCol = halfSize;

	/*
	 * Check for a trivial case where the whole object is totally
	 * outside of the specified area.  In that case set all of the
	 * data values to OFF.
	 */
	if (numIsLess(maxRow, cpMinRow) ||
		numIsGreater(minRow, cpMaxRow) ||
		numIsLess(maxCol, cpMinCol) ||
		numIsGreater(minCol, cpMaxCol))
	{
		memset(info.data, 0, info.dataSize);

		return info;
	}

	/*
	 * There is some overlap between the object and the area so we
	 * really have to do the scan.  Initialise the data values to -1
	 * to indicate that the cells are not yet calculated.
	 */
	FillViewData(&info, -1);

//	if (isAuxUsed)
//		FillAuxViewData(&info, STATE_OFF);

	/*
	 * Now iterate over the rows and columns of the data values
	 * to be calculated.
	 */
	for (rowIndex = 0; rowIndex < info.rowCount; rowIndex++)
	{
		for (colIndex = 0; colIndex < info.colCount; colIndex++)
		{
			int offset = rowIndex * info.colCount + colIndex;
			QuadRegion region;

			/*
			 * If we already know the result for this location
			 * then go on.  This can be the case if a large empty
			 * or full quadcell was encountered earlier.
			 */
			if (info.data[offset] >= 0)
				continue;

			//
			// Calculate the row and column number for this
			// location and then see that is outside of the
			// object's boundaries.  If so then the cell at
			// this data offset is OFF.
			//
			NUMBER row = numAddLong(minRow, rowIndex);
			NUMBER col = numAddLong(minCol, colIndex);

			if (numIsLess(row, cpMinRow) ||
				numIsGreater(row, cpMaxRow) ||
				numIsLess(col, cpMinCol) ||
				numIsGreater(col, cpMaxCol))
			{
				info.data[offset] = 0;

				continue;
			}

			//
			// The location is really within the object.
			// Make the call to get a region containing it.
			//
			region = qcGetRegion(cp, row, col);

			//
			// Make sure that the region contains the cell!
			//
			verify(numIsLessOrEqual(region.minRow, row) &&
				numIsLessOrEqual(region.minCol, col) &&
				numIsGreaterOrEqual(region.maxRow, row) &&
				numIsGreaterOrEqual(region.maxCol, col));

			//
			// Trim the region to the requested area.
			//
			region.minRow = numMax(region.minRow, minRow);
			region.minCol = numMax(region.minCol, minCol);
			region.maxRow = numMin(region.maxRow, maxRow);
			region.maxCol = numMin(region.maxCol, maxCol);

			//
			// Now fill in all of the cells in the view info
			// table which are within the found region.
			//
			qcStoreViewRegion(&info, &region);
		}
	}

	return info;
}


/*
 * Calculate a view of the specified rectangular region of the specified object
 * and save the result in a returned memory array whose size is the indicated
 * number of rows times the indicated number of columns divided by the scale.
 * The number of rows and columns must both be an exact multiple of scale.
 * Scale values determine squares having 'scale' cells on each side.
 * (For example, scale 1 is individual cells, and scale 10 is information
 * about groups of 100 cells.)  The cells in the array contain values which
 * are 0 for all OFF cells in each group, and 1 for one or more ON cells
 * in each group.  If the auxillary data is specified, then the combined
 * states found in the group is also stored.  For speed reasons the total
 * number of ON cells in the region is NOT calculated (only the count of
 * non-empty groups is obtained).  The array can then be easily examined
 * for groups of ON cells by calling FindNextRowIndex and FindNextRowIndex.
 * There is a limit on the size of the rectangular region that can be used.
 * The returned array should not be freed by the caller and is valid until
 * the next call to this routine.
 */
ViewInfo
qcCalculateScaledView(const QuadCell * cp, int viewId, BOOL isAuxUsed,
	int scale, NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	if (Life.debugOption == 0)
	{
		return qcNewCalculateScaledView(cp, viewId, isAuxUsed,
			scale, minRow, maxRow, minCol, maxCol);
	}

	int rowIndex = 0;
	NUMBER rowNum = minRow;
	ViewInfo info;

	verifyAlways(scale >= 1);

	//
	// Initialise and clear the view structure using the scale factor.
	//
	InitView(&info, viewId, isAuxUsed, scale, minRow, maxRow, minCol, maxCol);

	ClearViewData(&info);

	/*
	 * Perform the mostly obvious iteration over the rows and columns.
	 * We examine the regions one by one where each region is a
	 * square of size 'scale'.
	 */
	while (numIsLess(rowNum, maxRow))
	{
		int colIndex = 0;
		NUMBER endRowNum = numAddLong(rowNum, scale - 1);
		NUMBER colNum = minCol;

		while (numIsLess(colNum, maxCol))
		{
			//
			// Examine a strip of cells to see if all of them
			// are OFF.  If so then we can save some work.
			// (If not, then we have wasted a bit of extra work,
			// but not too much.)
			//
#define	COL_SKIP 20
			NUMBER endColNum = numAddLong(colNum,
				scale * COL_SKIP - 1);

			if (qcIsRegionOff(cp, rowNum, endRowNum,
				colNum, endColNum))
			{
				numAddrAddLong(&colNum, scale * COL_SKIP);
				colIndex += COL_SKIP;

				continue;
			}

			//
			// There was something in the strip, so now
			// we can do a more detailed analysis.
			//
			endColNum = numAddLong(colNum, scale - 1);

			int states = qcGetRegionStates(cp, rowNum, endRowNum,
				colNum, endColNum, STATE_ALL_NOT_OFF);

			if (states != STATE_NONE)
			{
				SetViewCell(&info, rowIndex, colIndex, TRUE);

				if (isAuxUsed)
					SetAuxViewData(&info, rowIndex, colIndex, states);

				if (isStatePartOn(states) || isStatePartNew(states))
					numAddrInc(&info.onCount);
			}

			numAddrAddLong(&colNum, scale);
			colIndex++;
		}

		numAddrAddLong(&rowNum, scale);
		rowIndex++;
	}

	return info;
}


//NEW:
static ViewInfo
qcNewCalculateScaledView(const QuadCell * cp, int viewId, BOOL isAuxUsed,
	int scale, NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	ViewInfo info;

	verifyAlways(scale >= 1);

	//
	// Initialise and clear the view structure using the scale factor.
	//
	InitView(&info, viewId, isAuxUsed, scale, minRow, maxRow, minCol, maxCol);

	ClearViewData(&info);

	if (isAuxUsed)
		FillAuxViewData(&info, STATE_OFF);

	//
	// Now walk down the quadcells to set the ViewInfo values.
	//
	NUMBER minCoord = qcGetMinCoordinate(cp);

	qcNewCalculateScaledViewFunc(cp, minCoord, minCoord, &info);

	return info;
}


/*
 * Recursive routine to set the appropriate values in the specified
 * ViewInfo for the specified quadcell.  The minRow and minCol values are
 * the coordinates of bottom and left edges of the quadcell.
 */
static void
qcNewCalculateScaledViewFunc(const QuadCell * cp,
	NUMBER minRow, NUMBER minCol, ViewInfo * info)
{
	//
	// First see if the cell is off or empty with no auxillary data needed.
	// If so then just return.
	//
	if (isCellOff(cp))
		return;

	if ((info->auxData == 0) && isCellEmpty(cp))
		return;

	//
	// Calculate the maximum coordinates of the quadcell.
	//
	NUMBER adjust = numDec(numTwoPower(cp->size));

	NUMBER maxRow = numAdd(minRow, adjust);
	NUMBER maxCol = numAdd(minCol, adjust);

	//
	// If the quadcell and the view area don't overlap then return.
	//
	if (numIsLess(maxRow, info->minRow) ||
		numIsGreater(minRow, info->maxRow) ||
		numIsLess(maxCol, info->minCol) ||
		numIsGreater(minCol, info->maxCol))
	{
		return;
	}

	//
	// Trim the quadcell coordinates to the view area rectangle.
	//
	NUMBER trimMinRow = numMax(minRow, info->minRow);
	NUMBER trimMaxRow = numMin(maxRow, info->maxRow);
	NUMBER trimMinCol = numMax(minCol, info->minCol);
	NUMBER trimMaxCol = numMin(maxCol, info->maxCol);

	//
	// Calculate the minimum and maximum row and column indexes
	// within the view area of the trimmed quadcell.
	//
	int minRowIndex = numToLong(
		numDivLong(numSub(trimMinRow, info->minRow), info->scale));

	int maxRowIndex = numToLong(
		numDivLong(numSub(trimMaxRow, info->minRow), info->scale));

	int minColIndex = numToLong(
		numDivLong(numSub(trimMinCol, info->minCol), info->scale));

	int maxColIndex = numToLong(
		numDivLong(numSub(trimMaxCol, info->minCol), info->scale));

	//
	// If the cell is pure than add its state to all of the view area
	// squares that it overlaps with and return.
	//
	if (isCellPure(cp))
	{
		int row, col;

		for (row = minRowIndex; row <= maxRowIndex; row++)
		{
			for (col = minColIndex; col <= maxColIndex; col++)
			{
				SetViewCell(info, row, col, TRUE);

				if (info->auxData)
					OrAuxViewData(info, row, col, cp->state);
			}
		}

		return;
	}

	//
	// If the quadcell is totally contained within a single view area
	// square then just add its state to that square and return.
	//
	if ((minRowIndex == maxRowIndex) && (minColIndex == maxColIndex) &&
		numIsGreaterOrEqual(minRow, info->minRow) &&
		numIsGreaterOrEqual(minCol, info->minCol) &&
		numIsLessOrEqual(maxRow, info->maxRow) &&
		numIsLessOrEqual(maxCol, info->maxCol))
	{
		SetViewCell(info, minRowIndex, minColIndex, TRUE);

		if (info->auxData)
			OrAuxViewData(info, minRowIndex, minColIndex, cp->state);

		return;
	}

	//
	// This quadcell is not pure and is not totally contained within
	// one view square.  Check the view squares that it overlaps with
	// and see if all those squares already have all of the states that
	// we could contribute.  If so we can save time and not recurse.
	//
	if (Life.debugOption != 1)
	if (AreViewDataBitsSet(info, minRowIndex, maxRowIndex,
		minColIndex, maxColIndex, cp->state & ~STATE_OFF))
	{
		return;
	}

	//
	// Now we have to recurse on its subquadrants.
	// Adjust their lower and leftmost coordinates as required.
	//
	verifyAlways(cp->size > SIZE1);

	adjust = numTwoPower(cp->size - 1);

	qcNewCalculateScaledViewFunc(indexToQuadCell(cp->ul),
		numAdd(minRow, adjust), minCol, info);

	qcNewCalculateScaledViewFunc(indexToQuadCell(cp->ur),
		numAdd(minRow, adjust), numAdd(minCol, adjust), info);

	qcNewCalculateScaledViewFunc(indexToQuadCell(cp->ll),
		minRow, minCol, info);

	qcNewCalculateScaledViewFunc(indexToQuadCell(cp->lr),
		minRow, numAdd(minCol, adjust),  info);
}


/*
 * Store all of the values in the view information array for those cells
 * in the specified region.  The cells must NOT have been set already.
 * If auxillary data is not present then cells with states ON or NEW
 * are marked.  If auxillary data is present, then all non-OFF cells are
 * marked, and the state of the cell is saved in auxData.
 */
static void
qcStoreViewRegion(ViewInfo * info, const QuadRegion * region)
{
	int minRowIndex = numToLong(numSub(region->minRow, info->minRow));
	int maxRowIndex = numToLong(numSub(region->maxRow, info->minRow));
	int minColIndex = numToLong(numSub(region->minCol, info->minCol));
	int maxColIndex = numToLong(numSub(region->maxCol, info->minCol));

	BOOL isNotEmpty = isStateNotEmpty(region->state);
	int cellValue = isNotEmpty;
	int rowIndex;

	if (info->auxData)
		cellValue = isStateNotOff(region->state);

	for (rowIndex = minRowIndex; rowIndex <= maxRowIndex; rowIndex++)
	{
		int colIndex;

		for (colIndex = minColIndex; colIndex <= maxColIndex; colIndex++)
		{
			int offset = rowIndex * info->colCount + colIndex;

			verify(info->data[offset] < 0);

			info->data[offset] = cellValue;

			if (info->auxData)
				info->auxData[offset] = region->state;

			if (isNotEmpty)
				numAddrInc(&info->onCount);
		}
	}
}


/*
 * Dump out the structure of a quadcell.
 * This assumes that all flag values within the quadcells start off
 * being zero, and then clears them all again when done.
 */
void
qcDump(QuadCell * cp)
{
	if (cp == 0)
	{
		printf("Null quadcell\n");

		return;
	}

	qcDumpFunc(cp, 0);
	qcClearFlags(cp);
}


/*
 * Recursive routine to dump out the structure of a quadcell.
 * This uses the flag element of the quadcells to prevent multiple
 * dumps of the same quadcell.
 */
static void
qcDumpFunc(QuadCell * cp, int indent)
{
	int i;

	if (cp->flag)
		return;

	cp->flag = 1;

	for (i = 0; i < indent; i++)
		fputs(" ", stdout);

	printf("index:%d size:%d state:%d ul:%d ur:%d ll:%d lr:%d",
		cp->index, cp->size, cp->state, cp->ul, cp->ur, cp->ll, cp->lr);

	if (isCellPure(cp))
		printf(" PURE");

	if (isCellPartDead(cp))
		printf(" DEAD");

	if (isCellPartOff(cp))
		printf(" OFF");

	if (isCellPartOn(cp))
		printf(" ON");

	if (isCellPartNew(cp))
		printf(" NEW");

	if (isCellPartMark(cp))
		printf(" MARK");

	printf("\n");

	if (cp->size <= SIZE1)
		return;

	indent++;

	qcDumpFunc(indexToQuadCell(cp->ul), indent);
	qcDumpFunc(indexToQuadCell(cp->ur), indent);
	qcDumpFunc(indexToQuadCell(cp->ll), indent);
	qcDumpFunc(indexToQuadCell(cp->lr), indent);
}


/*
 * Get the count of the number of quadcells used by the specified
 * quadcell tree.  This is an indication of the amount of memory used.
 * Note: It has NO relation to the number of ON cells within the quadcell.
 * This temporarily uses the flag element of the quadcells.
 */
long
qcGetUsage(QuadCell * cp)
{
	long count = qcGetUsageFunc(cp);

	qcClearFlags(cp);

	return count;
}


/*
 * Recursive function to count the number of quadcells used by the
 * specified quadcell tree.
 */
static long
qcGetUsageFunc(QuadCell * cp)
{
	if (cp->flag)
		return 0;

	cp->flag = 1;

	long count = 1;

	if (cp->size > SIZE1)
	{
		count += qcGetUsageFunc(indexToQuadCell(cp->ul));
		count += qcGetUsageFunc(indexToQuadCell(cp->ur));
		count += qcGetUsageFunc(indexToQuadCell(cp->ll));
		count += qcGetUsageFunc(indexToQuadCell(cp->lr));
	}

	return count;
}



/*
 * Recursive routine to clear all of the flags from a quadcell tree
 * which has had its flags set.
 */
void
qcClearFlags(QuadCell * cp)
{
	if (cp->flag == 0)
		return;

	cp->flag = 0;

	if (cp->size <= SIZE1)
		return;

	qcClearFlags(indexToQuadCell(cp->ul));
	qcClearFlags(indexToQuadCell(cp->ur));
	qcClearFlags(indexToQuadCell(cp->ll));
	qcClearFlags(indexToQuadCell(cp->lr));
}

