/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Operations which act on groups of cells in a quadcell which are
 * treated as objects.
 */
#include "life.h"
#include "quad.h"
#include "quadpriv.h"


/*
 * A table of the results from cutting quadcells into diamond shapes.
 * Totally full or totally empty quadcells aren't stored here.
 */
static QuadCell * lowCuts[MAXSIZE];
static QuadCell * highCuts[MAXSIZE];


//
// Forward declarations.
//
static QuadCell * qcQuarterDiamondFunc(QuadCell * cp, NUMBER radius);


/*
 * Extract the cells of a connected object whose cells are all within the
 * specified distance of each other.  The specified location should be any
 * one of the ON cells of the object.  The maximum allowed distance between
 * the cells is not very large since this is an expensive operation.
 * Furthermore, this operation scales according to the size of the object.
 * This only extracts cells with the ON and NEW states, ignoring DEAD and
 * OFF cells.
 * 
 * Note: There is a limit on the size of the number of iterations used.
 * If this limit is exceeded only a partial result is returned.
 * This fact is set in the flag element of the returned quadcell,
 * and MUST be cleared by the caller.
 * Warning: This routine can perform garbage collection.
 */
QuadCell *
qcExtractObject(QuadCell * cp, NUMBER row, NUMBER col, int distance)
{
	int iterations = 0;

	QuadCell * cpObject = qcEmpty();

	if (!qcIsSet(cp, row, col))
		return cpObject;

	//
	// Reduce the states of the object so that the logical
	// operations do what we want.
	//
	cp = qcReduceStates(cp);

	//
	// Start off the object with the known ON cell.
	//
	cpObject = qcSet(cpObject, row, col);

	//
	// If the distance is zero then return just that cell.
	//
	if (distance <= 0)
		return cpObject;

	//
	// Now iterate as long as a spread operation from the current
	// set of new cells picks up any more new cells.  This process
	// propagates as an edge-wave through the object's cells.  We
	// only process edge cells on each iteration instead of the full
	// set of the  object's cells since that keeps the quadcells
	// simpler (even though more operations are used).
	//
	QuadCell * cpNew = cpObject;

	while (!Life.stop && (iterations++ < 1000))
	{
		//
		// Spread out the current set of new cells to get all the
		// candidates for the next set of new cells.
		//
		cpNew = qcSpread(cpNew, distance);

		//
		// Remove the candidates which aren't in the input object.
		//
		cpNew = qcAndOperation(cp, cpNew);

		//
		// Remove those candidates which are already in the
		// found object.  The ones which are left become the next
		// set of new cells.
		//
		cpNew = qcAndNotOperation(cpNew, cpObject);

		//
		// If there are no new cells then the object is complete.
		//
		if (isCellEmpty(cpNew))
			return cpObject;

		//
		// Add the new cells to the found object.
		//
		cpObject = qcOrOperation(cpObject, cpNew);

		//
		// See if garbage collection is useful.
		// We must temporarily save our quadcells from the
		// garbage collection.
		//
		qcAddSafeQuadCell(cpObject);
		qcAddSafeQuadCell(cpNew);
		qcCheckGarbageCollect();
		qcClearSafeQuadCells(2);
	}

	//
	// We ran out of iterations.
	// Set the flag element to indicate this and return what we got.
	//
	cpObject->flag = 1;

	return cpObject;
}


/*
 * Spread all of the cells in a quadcell out by the specified distance.
 * This means that any OFF cell which is within the specified distance
 * to an existing ON cell (in any of the 8 directions) is turned ON.
 * The maximum allowed distance of the spread is not very large since
 * this is an expensive operation.
 */
QuadCell *
qcSpread(QuadCell * cp, int distance)
{
	int i;

	if (isCellEmpty(cp) || (distance <= 0))
		return cp;

	verifyAlways(distance <= QUAD_MAX_SPREAD);

	//
	// Perform left, right, up, and down shifts in sequence.
 	// We do the OR operations of each quadcell as we go to
 	// hopefully collect together solid regions.
	//
	for (i = 0; i < distance; i++)
	{
		cp = qcOrOperation(cp, qcLeftShift(cp, numOne()));
		cp = qcOrOperation(cp, qcRightShift(cp, numOne()));
		cp = qcOrOperation(cp, qcUpShift(cp, numOne()));
		cp = qcOrOperation(cp, qcDownShift(cp, numOne()));
	}

	return cp;
}


/*
 * Get a quadcell which contains a full diamond shape centered on the origin
 * with sides at 45 degree angles and sharp points on the left, right, top,
 * and bottom edges.  The radius is the distance from the origin to any edge.
 * (The result represents all cells whose taxicab distance from the origin
 * is less than or equal to the radius.)  A radius of 0 returns just one
 * ON cell at the origin.
 */
QuadCell *
qcFullDiamond(NUMBER radius)
{
	int i;

	//
	// Check for very small radius.
	//
	verifyAlways(numSign(radius) >= 0);

	if (numIsZero(radius))
		return qcGetFull(0);

	//
	// Clear the special results for saving cut quadcells.
	//
	for (i = 0; i < MAXSIZE; i++)
	{
		lowCuts[i] = 0;
		highCuts[i] = 0;
	}

	//
	// Start with an empty quadcell large enough to hold the result
	// in the upper right quadrant.
	//
	QuadCell * cp = qcEmptyToContain(radius, radius);

	//
	// Call the recursive routine to get the result for just the
	// upper right quadrant.  Note: There are no results to clear
	// afterwards since we store them in our own private table.
	//
	QuadCell * result = qcQuarterDiamondFunc(cp, radius);

	//
	// Replace the other quadrants with empty ones.
	//
	QUADINDEX emptyIndex = INDEX_FIRST_OFF + result->size - 1;

	result = qcFind(emptyIndex, result->ur, emptyIndex, emptyIndex);

	//
	// Reflect the quarter diamond and combine it to get the
	// right half of the result.
	//
	result = qcOrOperation(result,
		qcVerticalFlip(result, numZero(), TRUE));

	//
	// Reflect it again and combine it to get the full diamond.
	//
	result = qcOrOperation(result,
		qcHorizontalFlip(result, numZero(), TRUE));

	return result;
}


/*
 * Recursive routine to get the result for the upper right quarter
 * of a diamond.  We can get this answer by considering where a
 * single 45 degree line cuts through the quadrants.  For any size
 * quadrant, the line either lies below the quadrant, lies above the
 * quadrant, or cuts through the quadrant.  If the quadrant is cut
 * there are only two different places that it can be cut, so we only
 * have to store two different results for those quadrants.
 */
static QuadCell *
qcQuarterDiamondFunc(QuadCell * cp, NUMBER radius)
{
	//
	// Handle 2 by 2 cells specially.
	// This is easier then calculating adjustment values.
	//
	if (cp->size == SIZE2)
	{
		QUADINDEX ulIndex = ((numCmpLong(radius, -1) >= 0) ?
			 INDEX_FIRST_ON : INDEX_FIRST_OFF);

		QUADINDEX urIndex = ((numCmpLong(radius, 0) >= 0) ?
			 INDEX_FIRST_ON : INDEX_FIRST_OFF);

		QUADINDEX llIndex = ((numCmpLong(radius, -2) >= 0) ?
			 INDEX_FIRST_ON : INDEX_FIRST_OFF);

		QUADINDEX lrIndex = ((numCmpLong(radius, -1) >= 0) ?
			 INDEX_FIRST_ON : INDEX_FIRST_OFF);

		return qcFind(ulIndex, urIndex, llIndex, lrIndex);
	}

	//
	// We shouldn't see individual cells here.
	//
	verifyAlways(cp->size > SIZE2);

	//
	// See what the taxicab distance to the upper right cell is
	// with respect to the diagonal line which lies at the
	// specified radius to the right.
	//
	NUMBER maxCoord = qcGetMaxCoordinate(cp);
	NUMBER minCoord = qcGetMinCoordinate(cp);

	NUMBER maxDistance = numAdd(maxCoord, maxCoord);
	NUMBER minDistance = numAdd(minCoord, minCoord);

	//
	// Determine whether the whole quadcell lies on one side of the
	// diagonal line determined by the radius.  If so then the quadcell
	// is completely empty or full.
	//
	if (numIsGreater(minDistance, radius))
		return qcGetEmpty(cp->size);

	if (numIsLessOrEqual(maxDistance, radius))
		return qcGetFull(cp->size);

	//
	// This quadcell is cut by the line.
	// There are only two possible cuts, either in the upper right
	// corner or the lower left corner.  Both cuts are exactly at
	// the same place for this size quadcell, and so can be computed
	// once and saved.
	//
	BOOL isUpperRightCut = (numSign(radius) >= 0);

	//
	// See if there is a result saved for the cut.
	//
	if (isUpperRightCut && highCuts[cp->size])
		return highCuts[cp->size];

	if (!isUpperRightCut && lowCuts[cp->size])
		return lowCuts[cp->size];

	//
	// We have no result saved, so we have to calculate it.
	// Get the results for the quadrants while adjusting the radius.
	//
	NUMBER adjust = numTwoPower(cp->size - 1);

	NUMBER radius1 = numAdd(radius, adjust);
	NUMBER radius2 = numSub(radius, adjust);

	QuadCell * ul = qcQuarterDiamondFunc(indexToQuadCell(cp->ul), radius);
	QuadCell * ur = qcQuarterDiamondFunc(indexToQuadCell(cp->ur), radius2);
	QuadCell * ll = qcQuarterDiamondFunc(indexToQuadCell(cp->ll), radius1);
	QuadCell * lr = qcQuarterDiamondFunc(indexToQuadCell(cp->lr), radius);

	//
	// Find the result and save it appropriately.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	if (isUpperRightCut)
		highCuts[cp->size] = result;
	else
		lowCuts[cp->size] = result;

	return result;
}


/*
 * Return the taxicab distance from the origin to the closest ON or NEW cell.
 * This basically tests full diamonds of different sizes against the
 * quadcell looking for overlaps.  The sizes are chosen based on a
 * increasing binary search.  The quadcell must not be empty.
 */
NUMBER
qcCalculateDistance(QuadCell * cp)
{
	QuadCell * diamondCell;
	NUMBER lowRadius;
	NUMBER highRadius;

	verifyAlways(isCellNotEmpty(cp));

	//
	// Check to see if the origin contains an ON cell.
	//
	if (isCellFull(cp))
		return numZero();

	if (qcIsSet(cp, numZero(), numZero()))
		return numZero();

	//
	// Walk upwards by larger and larger full diamond shapes
	// looking for the first one which overlaps the quadcell.
	// This is an exponential process so it shouldn't take long.
	//
	lowRadius = numZero();
	highRadius = numFromLong(1);

	while (TRUE)
	{
		diamondCell = qcFullDiamond(highRadius);

		if (qcIsOverlapping(cp, diamondCell))
			break;

		lowRadius = highRadius;
		numAddrAdd(&highRadius, highRadius);
	}

	//
	// We found the lowest power of two radius which intersects.
	// Now use a binary-like search on the low and high radius
	// values to find the smallest diamond which still intersects.
	// At all times the high radius value creates a diamond which
	// intersects, whereas the low radius creates a diamond which
	// doesn't intersect.
	//
	while (TRUE)
	{
		//
		// Take the average between the low and high values.
		// If the average is the same as the low radius, then
		// the high and low radius are within one of each other,
		// and so we have our answer.
		//
		NUMBER averageRadius = numAverage(lowRadius, highRadius);

		if (numIsEqual(averageRadius, lowRadius))
			break;

		//
		// Create and test a diamond having the average radius.
		// If it intersects then replace the high radius with it,
		// otherwise replace the low radius with it.
		//
		diamondCell = qcFullDiamond(averageRadius);

		if (qcIsOverlapping(cp, diamondCell))
			highRadius = averageRadius;
		else
			lowRadius = averageRadius;
	}

	return highRadius;
}


/*
 * Return the coordinates of one of the closest cells to the origin.
 * If there is a tie we try first for cells on the backward diagonal.
 * Returns FALSE if the quadcell is empty.
 */
BOOL
qcFindClosestCell(QuadCell * cp, NUMBER * retRow, NUMBER * retCol)
{
	*retRow = numZero();
	*retCol = numZero();

	if (isCellEmpty(cp))
		return FALSE;

	//
	// Find out how far away the closest ON cell is.
	// If the answer is zero then we are done.
	//
	NUMBER radius = qcCalculateDistance(cp);

	if (numIsZero(radius))
		return TRUE;

	//
	// Create the diamond for that radius and then AND it with the
	// given quadcell.  This preserves only the cells at that radius.
	// There should be at least one ON cell found since the distance
	// calculation found it.
	//
	QuadCell * diamondCell = qcFullDiamond(radius);
	QuadCell * cells = qcAndOperation(cp, diamondCell);

	verifyAlways(isCellNotEmpty(cells));

	return qcFindFirstCell(cells, retRow, retCol);
}


/*
 * Find the first ON cell in the quadcell structure, searching from
 * the upper left row by row.  Returns TRUE if a cell was found with
 * its coordinates returned.
 */
BOOL
qcFindFirstCell(const QuadCell * cells, NUMBER * retRow, NUMBER * retCol)
{
	*retRow = numZero();
	*retCol = numZero();

	if (isCellEmpty(cells))
		return FALSE;

	//
	// Now walk down the quadcell structure to find one of the
	// ON cells. Try the backward diagonal first.  Calculate the
	// coordinates of the found cell starting from the left and
	// bottom edges of the quadcell, and adjust them as needed.
	//
	NUMBER row = qcGetMinCoordinate(cells);
	NUMBER col = row;

	while (cells->size >= SIZE2)
	{
		QUADINDEX ulIndex = cells->ul;
		QUADINDEX urIndex = cells->ur;
		QUADINDEX llIndex = cells->ll;
		QUADINDEX lrIndex = cells->lr;

		NUMBER adjust = numTwoPower(cells->size - 1);

		cells = indexToQuadCell(ulIndex);

		if (isCellNotEmpty(cells))
		{
			numAddrAdd(&row, adjust);

			continue;
		}

		cells = indexToQuadCell(urIndex);

		if (isCellNotEmpty(cells))
		{
			numAddrAdd(&row, adjust);
			numAddrAdd(&col, adjust);

			continue;
		}

		cells = indexToQuadCell(llIndex);

		if (isCellNotEmpty(cells))
			continue;

		cells = indexToQuadCell(lrIndex);

		if (isCellNotEmpty(cells))
		{
			numAddrAdd(&col, adjust);

			continue;
		}

		//
		// One of the quadrants should have been non-empty.
		//
		verifyAlways(FALSE);
	}

	//
	// Return the coordinates.
	//
	*retRow = row;
	*retCol = col;

	return TRUE;
}

