/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Operations that are calculated from the quadcell tree.
 */
#include "quad.h"
#include "quadpriv.h"

#include <stdio.h>


/*
 * The result type and its parameter to be calculated by an operation.
 */
static ResultType currentResultType = RESULT_NONE;
static NUMBER currentResultParam;


//
// Local routines.
//
static BOOL qcIsOverlappingFunc(const QuadCell * cp1, const QuadCell * cp2);
static QuadCell * qcInvertOperationFunc(QuadCell * cp);
static QuadCell * qcOrOperationFunc(QuadCell * cp1, QuadCell * cp2);
static QuadCell * qcAndOperationFunc(QuadCell * cp1, QuadCell * cp2);
static QuadCell * qcXorOperationFunc(QuadCell * cp1, QuadCell * cp2);
static QuadCell * qcAndNotOperationFunc(QuadCell * cp1, QuadCell * cp2);

static QuadCell * qcClearToRightFunc(QuadCell * cp, NUMBER col);
static QuadCell * qcClearToLeftFunc(QuadCell * cp, NUMBER col);
static QuadCell * qcClearToTopFunc(QuadCell * cp, NUMBER row);
static QuadCell * qcClearToBottomFunc(QuadCell * cp, NUMBER row);
static QuadCell * qcExtractRowFunc(QuadCell * cp, NUMBER row);
static QuadCell * qcDeleteRowFunc(QuadCell *cp, NUMBER row);


/*
 * Perform the inversion of all the current cells in a quadcell.
 * This turns all of its ON cells into OFF cells, and vice-versa.
 * This treats DEAD and NEW cells as OFF and ON cells.
 * The size of the quadcell is not changed.
 */
QuadCell *
qcInvertOperation(QuadCell * cp)
{
	//
	// Check the easy cases.
	//
	if (isCellEmpty(cp))
		return qcGetFull(cp->size);

	if (isCellFull(cp))
		return qcGetEmpty(cp->size);

	qcSetResultType(RESULT_NONE);

	//
	// Call our recursive routine.
	//
	return qcInvertOperationFunc(cp);
}


/*
 * Recursive routine to invert the cells in a quadcell.
 * This turns all of its ON cells into OFF cells, and vice-versa.
 * The size of the quadcell is not changed.
 */
static QuadCell *
qcInvertOperationFunc(QuadCell * cp)
{
	//
	// Check for the empty and full cases.
	// This also handles the size 1 quadcell.
	//
	if (isCellEmpty(cp))
		return qcGetFull(cp->size);

	if (isCellFull(cp))
		return qcGetEmpty(cp->size);

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// No result was saved so we must calculate it now.
	// Call ourself for each of our quadrants.
	//
	QuadCell * ul = qcInvertOperationFunc(indexToQuadCell(cp->ul));
	QuadCell * ur = qcInvertOperationFunc(indexToQuadCell(cp->ur));
	QuadCell * ll = qcInvertOperationFunc(indexToQuadCell(cp->ll));
	QuadCell * lr = qcInvertOperationFunc(indexToQuadCell(cp->lr));

	//
	// Build the result and save it for reuse.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Perform the OR operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * OR function of the ON cells in both of the two quadcells.
 * This handles any DEAD, NEW and MARK states in reasonable ways too.
 * This temporarily uses quadpair structures to avoid repeated recursions.
 */
QuadCell *
qcOrOperation(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for really trivial results.
	//
	if (cp1->index == cp2->index)
		return cp1;

	if (isCellOff(cp1))
		return cp2;

	if (isCellOff(cp2))
		return cp1;

	if ((cp1->size >= cp2->size) && isCellMark(cp1))
		return cp1;

	if ((cp2->size >= cp1->size) && isCellMark(cp2))
		return cp2;

	//
	// Grow the size of the smaller quadcell so that the quadrants
	// of the two quadcells can be matched easily.
	//
	if (cp1->size > cp2->size)
		cp2 = qcGrowToSize(cp2, cp1->size);

	if (cp2->size > cp1->size)
		cp1 = qcGrowToSize(cp1, cp2->size);

	//
	// Prepare the quadpair hash table for this operation.
	//
	qcSetQuadPairOperation(OP_OR, numZero());

	//
	// Now call our recursive routine for the quadrants.
	//
	return qcOrOperationFunc(cp1, cp2);
}


/*
 * Perform the AND operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * AND function of the ON cells in both of the two quadcells.
 * This handles any DEAD and NEW states in reasonable ways too.
 * This temporarily uses quadpair structures to avoid repeated recursions.
 */
QuadCell *
qcAndOperation(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results.
	//
	if (cp1->index == cp2->index)
		return cp1;

	if (isCellOff(cp1))
		return cp1;

	if (isCellOff(cp2))
		return cp2;

	if ((cp1->size >= cp2->size) && isCellNew(cp1))
		return cp2;

	if ((cp2->size >= cp1->size) && isCellNew(cp2))
		return cp1;

	//
	// Trim the size of the larger quadcell so that the quadrants
	// of the two quadcells can be matched easily.
	//
	if (cp1->size > cp2->size)
		cp1 = qcTrimToSize(cp1, cp2->size);

	if (cp2->size > cp1->size)
		cp2 = qcTrimToSize(cp2, cp1->size);

	//
	// Prepare the quadpair hash table for this operation.
	//
	qcSetQuadPairOperation(OP_AND, numZero());

	//
	// Now call our recursive routine for the quadrants.
	//
	return qcAndOperationFunc(cp1, cp2);
}


/*
 * Perform the XOR operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * XOR function of the ON cells in both of the two quadcells.
 * This temporarily uses quadpair structures to avoid repeated recursions.
 */
QuadCell *
qcXorOperation(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results.
	//
	if (cp1->index == cp2->index)
		return qcEmpty();

	if (isCellEmpty(cp1))
		return cp2;

	if (isCellEmpty(cp2))
		return cp1;

	if ((cp1->size == cp2->size) && isCellFull(cp1) && isCellFull(cp2))
		return qcEmpty();

	//
	// Grow the size of the smaller quadcell so that the quadrants
	// of the two quadcells can be matched easily.
	//
	if (cp1->size > cp2->size)
		cp2 = qcGrowToSize(cp2, cp1->size);

	if (cp2->size > cp1->size)
		cp1 = qcGrowToSize(cp1, cp2->size);

	//
	// Prepare the quadpair hash table for this operation.
	//
	qcSetQuadPairOperation(OP_XOR, numZero());

	//
	// Now call our recursive routine for the quadrants.
	//
	return qcXorOperationFunc(cp1, cp2);
}


/*
 * Perform the ANDNOT operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * ANDNOT function of the ON or NEW cells in both of the two quadcells.
 * The cells in the DEAD state are also affected similarly.
 * Note: This operation is not symmetric on the two quadcells.
 * This temporarily uses quadpair structures to avoid repeated recursions.
 */
QuadCell *
qcAndNotOperation(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results.
	//
	if (cp1->index == cp2->index)
		return qcEmpty();

	if (isCellOff(cp1) || isCellOff(cp2))
		return cp1;

	if ((cp2->size >= cp1->size) && isCellFull(cp2))
		return qcEmpty();

	//
	// Grow or trim the sizes the quadcells so that the quadrants
	// of the two quadcells can be matched easily.
	//
	if (cp1->size < cp2->size)
		cp2 = qcTrimToSize(cp2, cp1->size);

	if (cp1->size > cp2->size)
		cp2 = qcGrowToSize(cp2, cp1->size);

	//
	// Prepare the quadpair hash table for this operation.
	//
	qcSetQuadPairOperation(OP_ANDNOT, numZero());

	//
	// Now call our recursive routine for the quadrants.
	//
	return qcAndNotOperationFunc(cp1, cp2);
}


/*
 * Return whether or not two quadcells have any ON cells in common.
 * This is similar to an AND operation, but is much quicker since no
 * quadcell is calculated and a TRUE answer can be short-cutted.
 */
BOOL
qcIsOverlapping(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results.
	//
	if (isCellEmpty(cp1) || isCellEmpty(cp2))
		return FALSE;

	if (cp1->index == cp2->index)
		return TRUE;

	if ((cp1->size >= cp2->size) && isCellFull(cp1))
		return TRUE;

	if ((cp2->size >= cp1->size) && isCellFull(cp2))
		return TRUE;

	//
	// Trim the size of the larger quadcell so that the quadrants
	// of the two quadcells can be matched easily.
	//
	if (cp1->size > cp2->size)
		cp1 = qcTrimToSize(cp1, cp2->size);

	if (cp2->size > cp1->size)
		cp2 = qcTrimToSize(cp2, cp1->size);

	//
	// Prepare the quadpair hash table for this operation.
	//
	qcSetQuadPairOperation(OP_OVERLAPPING, numZero());

	//
	// Now perform the recursive call to get the answer.
	//
	return qcIsOverlappingFunc(cp1, cp2);
}


/*
 * Recursive routine to do the OR operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * OR function of the ON cells in both of the two quadcells.
 * This handles any DEAD, NEW and MARK states in reasonable ways too.
 * The two quadcells must have the same size.
 */
static QuadCell *
qcOrOperationFunc(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results for pure cells in the order that
	// we want to promote.  These checks should also catch all the
	// SIZE1 cells since they must have pure states.
	//
	if (cp1->index == cp2->index)
		return cp1;

	verifyAlways(cp1->size == cp2->size);

	//
	// First check for the overridding pure states which
	// work for all sizes.
	//
	if (isCellOff(cp1))
		return cp2;

	if (isCellOff(cp2))
		return cp1;

	if (isCellMark(cp1))
		return cp1;

	if (isCellMark(cp2))
		return cp2;

	if (isCellOn(cp1) && !isCellPartMark(cp2))
		return cp1;

	if (isCellOn(cp2) && !isCellPartMark(cp1))
		return cp2;

	//
	// Check for the remaining states for SIZE1 quadcells.
	// We can't shortcut higher sizes.
	//
	if (cp1->size == SIZE1)
	{
		if (isCellNew(cp1))
			return cp1;

		if (isCellNew(cp2))
			return cp2;

		if (isCellOn(cp1))
			return cp1;

		if (isCellOn(cp2))
			return cp2;

		if (isCellDead(cp1))
			return cp1;

		if (isCellDead(cp2))
			return cp2;

		verifyAlways(FALSE);
	}

	//
	// Now we need to do recursion on the subquadrants.
	// Look up a previous result if any for these quadcells.
	// We can use a symmmetrical lookup for this operation.
	//
	QuadPair * pair = qcFindQuadPair(cp1->index, cp2->index, TRUE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// The result was not saved yet so we need to calculate it now.
	// Recurse on our quadrants.
	//
	QuadCell * ul = qcOrOperationFunc(
		indexToQuadCell(cp1->ul), indexToQuadCell(cp2->ul));

	QuadCell * ur = qcOrOperationFunc(
		indexToQuadCell(cp1->ur), indexToQuadCell(cp2->ur));

	QuadCell * ll = qcOrOperationFunc(
		indexToQuadCell(cp1->ll), indexToQuadCell(cp2->ll));

	QuadCell * lr = qcOrOperationFunc(
		indexToQuadCell(cp1->lr), indexToQuadCell(cp2->lr));

	//
	// Save and return the result quadcell.
	//
	QuadCell * cp = qcFind(ul->index, ur->index, ll->index, lr->index);

	pair->result = cp->index;

	return cp;
}


/*
 * Recursive routine to do the AND operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * AND function of the ON cells in both of the two quadcells.
 * This handles any DEAD and NEW states in reasonable ways too.
 * The two quadcells must have the same size.
 */
static QuadCell *
qcAndOperationFunc(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results for pure cells in the order that
	// we want to promote.  These checks should also catch all the
	// SIZE1 cells since they must have pure states.
	//
	if (cp1->index == cp2->index)
		return cp1;

	verifyAlways(cp1->size == cp2->size);

	//
	// First check for the overridding pure states which
	// work for all sizes.
	//
	if (isCellOff(cp1))
		return cp1;

	if (isCellOff(cp2))
		return cp2;

	if (isCellNew(cp1))
		return cp2;

	if (isCellNew(cp2))
		return cp1;

	//
	// Check for the remaining states for SIZE1 quadcells.
	// We can't shortcut higher sizes.
	//
	if (cp1->size == SIZE1)
	{
		if (isCellOn(cp1))
			return cp2;

		if (isCellOn(cp2))
			return cp1;

		if (isCellDead(cp1))
			return cp2;

		if (isCellDead(cp2))
			return cp1;

		verifyAlways(FALSE);
	}

	//
	// Now we need to do recursion on the subquadrants.
	// Look up a previous result if any for these quadcells.
	// We can use a symmmetrical lookup for this operation.
	//
	QuadPair * pair = qcFindQuadPair(cp1->index, cp2->index, TRUE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// The result was not saved yet so we need to calculate it now.
	// Recurse on our quadrants.
	//
	QuadCell * ul = qcAndOperationFunc(
		indexToQuadCell(cp1->ul), indexToQuadCell(cp2->ul));

	QuadCell * ur = qcAndOperationFunc(
		indexToQuadCell(cp1->ur), indexToQuadCell(cp2->ur));

	QuadCell * ll = qcAndOperationFunc(
		indexToQuadCell(cp1->ll), indexToQuadCell(cp2->ll));

	QuadCell * lr = qcAndOperationFunc(
		indexToQuadCell(cp1->lr), indexToQuadCell(cp2->lr));

	//
	// Save and return the result quadcell.
	//
	QuadCell * cp = qcFind(ul->index, ur->index, ll->index, lr->index);

	pair->result = cp->index;

	return cp;
}


/*
 * Recursive routine to do the XOR operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * XOR function of the ON cells in both of the two quadcells.
 * The two quadcells must have the same size.
 */
static QuadCell *
qcXorOperationFunc(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results.
	// These checks should also catch the SIZE1 cells since they
	// must be empty or full.  When returning empty cells make them
	// the same size as the input cells.
	//
	if (cp1->index == cp2->index)
		return qcGetEmpty(cp1->size);

	if (isCellEmpty(cp1))
		return cp2;

	if (isCellEmpty(cp2))
		return cp1;

	if (isCellFull(cp1) && isCellFull(cp2))
		return qcGetEmpty(cp1->size);

	verifyAlways((cp1->size > SIZE1) && (cp1->size == cp2->size));

	//
	// Look up a previous result if any for these quadcells.
	// We can use a symmmetrical lookup for this operation.
	//
	QuadPair * pair = qcFindQuadPair(cp1->index, cp2->index, TRUE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// The result was not saved yet so we need to calculate it now.
	// Recurse on our quadrants.
	//
	QuadCell * ul = qcXorOperationFunc(
		indexToQuadCell(cp1->ul), indexToQuadCell(cp2->ul));

	QuadCell * ur = qcXorOperationFunc(
		indexToQuadCell(cp1->ur), indexToQuadCell(cp2->ur));

	QuadCell * ll = qcXorOperationFunc(
		indexToQuadCell(cp1->ll), indexToQuadCell(cp2->ll));

	QuadCell * lr = qcXorOperationFunc(
		indexToQuadCell(cp1->lr), indexToQuadCell(cp2->lr));

	//
	// Save and return the result quadcell.
	//
	QuadCell * cp = qcFind(ul->index, ur->index, ll->index, lr->index);

	pair->result = cp->index;

	return cp;
}


/*
 * Recursive routine to do the ANDNOT operation for two quadcell structures.
 * This results in a possibly new quadcell whose ON cells are the
 * ANDNOT function of the ON or NEW cells in both of the two quadcells.
 * The cells in the DEAD state are also affected similarly.
 * Note: This operation is not symmetric on the two quadcells.
 * The two quadcells must have the same size.
 */
static QuadCell *
qcAndNotOperationFunc(QuadCell * cp1, QuadCell * cp2)
{
	//
	// Check for trivial results.
	// These checks should also catch the SIZE1 cells since they
	// must be empty or full.  When returning empty cells make them
	// the same size as the input cells.
	//
	if (cp1->index == cp2->index)
		return qcGetEmpty(cp1->size);

	if (isCellOff(cp1) || isCellOff(cp2))
		return cp1;

	if (isCellFull(cp2))
		return qcGetEmpty(cp1->size);

	if (isCellDead(cp1) && isCellDead(cp2))
		return qcGetEmpty(cp1->size);

	if (isCellFull(cp1) && isCellDead(cp2))
		return cp1;

	verifyAlways((cp1->size > SIZE1) && (cp1->size == cp2->size));

	//
	// Look up a previous result if any for these quadcells.
	// This is NOT a symmetrical operation and so we must preserve
	// the order of the indexes in the lookup.
	//
	QuadPair * pair = qcFindQuadPair(cp1->index, cp2->index, FALSE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// The result was not saved yet so we need to calculate it now.
	// Recurse on our quadrants.
	//
	QuadCell * ul = qcAndNotOperationFunc(
		indexToQuadCell(cp1->ul), indexToQuadCell(cp2->ul));

	QuadCell * ur = qcAndNotOperationFunc(
		indexToQuadCell(cp1->ur), indexToQuadCell(cp2->ur));

	QuadCell * ll = qcAndNotOperationFunc(
		indexToQuadCell(cp1->ll), indexToQuadCell(cp2->ll));

	QuadCell * lr = qcAndNotOperationFunc(
		indexToQuadCell(cp1->lr), indexToQuadCell(cp2->lr));

	//
	// Save and return the result quadcell.
	//
	QuadCell * cp = qcFind(ul->index, ur->index, ll->index, lr->index);

	pair->result = cp->index;

	return cp;
}


/*
 * Recursive routine to check for overlaps for two quadcell structures.
 * The two quadcells must have the same size.  Returns TRUE if the two
 * quadcells have at least one ON cell in common.
 */
static BOOL
qcIsOverlappingFunc(const QuadCell * cp1, const QuadCell * cp2)
{
	//
	// Check for trivial results.
	// These checks should also catch the SIZE1 cells since they
	// must be empty or full.
	//
	if (isCellEmpty(cp1) || isCellEmpty(cp2))
		return FALSE;

	if (isCellFull(cp1) || isCellFull(cp2))
		return TRUE;

	if (cp1->index == cp2->index)
		return TRUE;

	verifyAlways((cp1->size > SIZE1) && (cp1->size == cp2->size));

	//
	// Look up a previous result if any for these quadcells.
	// We can use a symmmetrical lookup for this operation.
	// Any non-null result indicates that we know that these two
	// quadcells do not overlap.
	//
	QuadPair * pair = qcFindQuadPair(cp1->index, cp2->index, TRUE);

	if (isNotNull(pair->result))
		return FALSE;

	//
	// The result was not saved yet so we need to calculate it now.
	// Recurse on our quadrants one by one and see if they overlap.
	// If so then we can return right away.
	//
	if (qcIsOverlappingFunc(
		indexToQuadCell(cp1->ul), indexToQuadCell(cp2->ul)))
	{
		return TRUE;
	}

	if (qcIsOverlappingFunc(
		indexToQuadCell(cp1->ur), indexToQuadCell(cp2->ur)))
	{
		return TRUE;
	}

	if (qcIsOverlappingFunc(
		indexToQuadCell(cp1->ll), indexToQuadCell(cp2->ll)))
	{
		return TRUE;
	}

	if (qcIsOverlappingFunc(
		indexToQuadCell(cp1->lr), indexToQuadCell(cp2->lr)))
	{
		return TRUE;
	}

	//
	// These two quadrants do not overlap.
	// Save a non-null result so the answer can be reused.
	//
	pair->result = INDEX_FIRST_OFF;

	return FALSE;
}


/*
 * Return a quadcell which contains all ON cells within the specified region
 * (including its boundaries) and all OFF cells outside of it.
 */
QuadCell *
qcGetFullRegion(NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	//
	// Check for a nonsense region.
	//
	if (numIsGreater(minRow, maxRow) || numIsGreater(minCol, maxCol))
		return qcEmpty();

	//
	// Check for a region of just one cell.
	//
	if (numIsEqual(minRow, maxRow) && numIsEqual(minCol, maxCol))
		return qcSet(qcEmpty(), minRow, minCol);

	//
	// The region is larger than one cell.
	// Get the size of a full quadcell which contains the region
	// and then get that full cell.
	//
	NUMBER minCoord = numMin(minRow, minCol);
	NUMBER maxCoord = numMax(maxRow, maxCol);

	NUMBER max;
	NUMBER min;
	int size = SIZE1;

	do
	{
		size++;

		verifyAlways(size < MAXSIZE - 1);

		max = numTwoPower(size);
		min = numNeg(max);
	}
	while (numIsGreaterOrEqual(maxCoord, max) || numIsLess(minCoord, min));

	QuadCell * cp = qcGetFull(size + 1);

	//
	// Now trim off its edges to leave just the region.
	// We have to adjust the coordinates by one since we want to
	// preserve the boundaries.
	//
	cp = qcClearToRight(cp, numInc(maxCol));
	cp = qcClearToLeft(cp, numDec(minCol));
	cp = qcClearToTop(cp, numInc(maxRow));
	cp = qcClearToBottom(cp, numDec(minRow));

	return cp;
}


/*
 * Clear the specified region of a quadcell.  All cells inside the
 * region (including on its boundary) are set OFF.  The states of the
 * remaining cells are preserved.  The result is returned as a new quadcell.
 */
QuadCell *
qcClearRegion(QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol)
{
	//
	// Check for an OFF quadcell or an invalid region.
	//
	if (isCellOff(cp))
		return cp;

	if (numIsGreater(minRow, maxRow) || numIsGreater(minCol, maxCol))
		return cp;

	//
	// Calculate the bounds of the quadcell and then check for the
	// region not overlapping the object.
	//
	NUMBER num = numTwoPower(cp->size - 1);
	NUMBER max = numDec(num);
	NUMBER min = numNeg(num);

	if (numIsGreater(minRow, max) || numIsGreater(minCol, max))
		return cp;

	if (numIsLess(maxRow, min) || numIsLess(maxCol, min))
		return cp;

	//
	// The region does overlap the quadcell.
	// Reduce the region if possible to the size of the quadcell.
	//
	minRow = numMax(minRow, min);
	minCol = numMax(minCol, min);
	maxRow = numMin(maxRow, max);
	maxCol = numMin(maxCol, max);

	//
	// Check for a small region of just one cell.
	//
	if (numIsEqual(minRow, maxRow) && numIsEqual(minCol, maxCol))
		return qcClear(cp, minRow, minCol);

	//
	// Get a quadcell of the size to be cleared full of ON cells.
	//
	QuadCell * full = qcGetFullRegion(minRow, maxRow, minCol, maxCol);

	//
	// Finally ANDNOT the quadcell with the full quadcell.
	//
	return qcAndNotOperation(cp, full);
}


/*
 * Fill the specified region of a quadcell (including its boundary)
 * with ON cells.  The result is returned as a new quadcell.
 */
QuadCell *
qcFillRegion(QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol)
{
	//
	// Check for an invalid region.
	//
	if (numIsGreater(minRow, maxRow) || numIsGreater(minCol, maxCol))
		return cp;

	//
	// Check for a small region of just one cell.
	//
	if (numIsEqual(minRow, maxRow) && numIsEqual(minCol, maxCol))
		return qcSet(cp, minRow, minCol);

	//
	// Get a quadcell of the size to be filled full of ON cells.
	//
	QuadCell * full = qcGetFullRegion(minRow, maxRow, minCol, maxCol);

	//
	// Finally OR the quadcell with the full quadcell.
	//
	return qcOrOperation(cp, full);
}


/*
 * Extract the non-OFF cells contained with the specified region of a quadcell.
 * This means that all the cells outside of the region are cleared.
 * The states of the cells that are not cleared are preserved.
 * The result is returned as a new quadcell.
 */
QuadCell *
qcExtractRegion(QuadCell * cp, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol)
{
	//
	// Check for an OFF quadcell or an invalid region.
	//
	if (isCellOff(cp))
		return qcEmpty();

	if (numIsGreater(minRow, maxRow) || numIsGreater(minCol, maxCol))
		return qcEmpty();

	//
	// Calculate the bounds of the quadcell and then check for the
	// region not overlapping the object.
	//
	NUMBER num = numTwoPower(cp->size - 1);
	NUMBER max = numDec(num);
	NUMBER min = numNeg(num);

	if (numIsGreater(minRow, max) || numIsGreater(minCol, max))
		return qcEmpty();

	if (numIsLess(maxRow, min) || numIsLess(maxCol, min))
		return qcEmpty();

	//
	// The region does overlap the quadcell.
	// Reduce the region if possible to the size of the quadcell.
	//
	minRow = numMax(minRow, min);
	minCol = numMax(minCol, min);
	maxRow = numMin(maxRow, max);
	maxCol = numMin(maxCol, max);

	//
	// Check for a small region of just one cell.
	//
//	if (numIsEqual(minRow, maxRow) && numIsEqual(minCol, maxCol))
//	{
//		if (qcIsSet(cp, minRow, minCol))
//			return qcSet(qcEmpty(), minRow, minCol);
//
//		return qcEmpty();
//	}

	//
	// The overlap is more substantial and so lots of work is needed.
	// Calculate it by slicing away the cells outside of the four
	// sides of the region.  We have to adjust the coordinates by one
	// since we want to preserve the boundaries.
	//
	cp = qcClearToRight(cp, numInc(maxCol));
	cp = qcClearToLeft(cp, numDec(minCol));
	cp = qcClearToTop(cp, numInc(maxRow));
	cp = qcClearToBottom(cp, numDec(minRow));

	return cp;
}


/*
 * Clear all of the cells which are either in the specified column or
 * else to the right of it.  The states of the other cells are preserved.
 * Returns the new quadcell tree with the result.
 * The coordinate is relative to the center of the quadcell.
 * This temporarily uses the result elements to store information for
 * those quadcells which are cut by the column (which is necessary to
 * short-cut repeated calculations for very large objects).
 */
QuadCell *
qcClearToRight(QuadCell * cp, NUMBER col)
{
	qcSetResultType(RESULT_NONE);

	return qcClearToRightFunc(cp, col);
}


/*
 * Clear all of the cells which are either in the specified column or
 * else to the left of it.  The states of the other cells are preserved.
 * Returns the new quadcell tree with the result.
 * The coordinate is relative to the center of the quadcell.
 * This temporarily uses the result elements to store information for
 * those quadcells which are cut by the column (which is necessary to
 * short-cut repeated calculations for very large objects).
 */
QuadCell *
qcClearToLeft(QuadCell * cp, NUMBER col)
{
	qcSetResultType(RESULT_NONE);

	return qcClearToLeftFunc(cp, col);
}


/*
 * Clear all of the cells which are either in the specified row or
 * else above it.  The states of the other cells are preserved.
 * Returns the new quadcell tree with the result.
 * The coordinate is relative to the center of the quadcell.
 * This temporarily uses the result elements to store information for
 * those quadcells which are cut by the row (which is necessary to
 * short-cut repeated calculations for very large objects).
 */
QuadCell *
qcClearToTop(QuadCell * cp, NUMBER row)
{
	qcSetResultType(RESULT_NONE);

	return qcClearToTopFunc(cp, row);
}


/*
 * Clear all of the cells which are either in the specified row or
 * else below it.  The states of the other cells are preserved.
 * Returns the new quadcell tree with the result.
 * The coordinate is relative to the center of the quadcell.
 * This temporarily uses the result elements to store information for
 * those quadcells which are cut by the row (which is necessary to
 * short-cut repeated calculations for very large objects).
 */
QuadCell *
qcClearToBottom(QuadCell * cp, NUMBER row)
{
	qcSetResultType(RESULT_NONE);

	return qcClearToBottomFunc(cp, row);
}


/*
 * Recursive routine to clear the cells at the specified column or to
 * its right.  The states of the other cells are preserved.
 * We check whether we have a result already saved to eliminate
 * unnecessary recursions.  NOTE: We can only calculate a result
 * for cells which are actually cut.  Uncut quadcells will simply return
 * themselves.  Using the saved results depends on the fact that all
 * quadcells of a particular size which are cut are done so on exactly
 * the same column, and so the saved results are always correct.
 */
static QuadCell *
qcClearToRightFunc(QuadCell * cp, NUMBER col)
{
	QuadCell * newCp;
	NUMBER adjust;

	//
	// If the cell is OFF then nothing can be deleted so we can
	// return ourself.
	//
	if (isCellOff(cp))
		return cp;

	//
	// If we have a result saved then just return that.
	// (The column argument can be ignored since the result must have
	// had the same column argument when it was saved before.)
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// We have to actually do the work.
	// If this is the single cell then we can either return ourself
	// or the OFF cell depending on whether the column is positive.
	//
	if (cp->size == SIZE1)
		return (numIsPos(col) ? cp : qcGetEmpty(SIZE1));

	//
	// If the column is to the right of the whole quadcell then
	// nothing is deleted and we can return ourself.  Note: We
	// can't save ourself as a result since that would compete
	// with the results from similar cells which ARE cut.
	//
	NUMBER num = numTwoPower(cp->size - 1);

	if (numIsGreaterOrEqual(col, num))
		return cp;

	//
	// If the column is to the left of the whole quadcell then
	// everything is deleted and we can return an empty quadcell.
	// Return the one of the expected size.  Once again we can't
	// save the empty cell as a result since that would compete
	// with the results from similar cells which ARE cut.
	//
	num = numNeg(num);

	if (numIsLessOrEqual(col, num))
		return qcGetEmpty(cp->size);

	//
	// This quadcell is cut by the column.  If the column number is
	// zero we can simply replace the right two quadrants with empty
	// ones.  Save the result for future use.
	//
	if (numIsZero(col))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * newCp = qcFind(cp->ul, emptyIndex, cp->ll, emptyIndex);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// This quadcell is cut other than through its middle.
	// Get the adjustment value for the column number when recursing
	// on the quadrants.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	//
	// See whether the left or the right quadrants are the ones cut.
	// The uncut quadrants will either be unchanged or deleted.
	// The cut quadrants are the ones we will need to recurse on.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	if (numIsPos(col))
	{
		//
		// The rightmost quadrants are cut.
		// If they are both OFF then nothing will be deleted
		// and so we can return ourself.  Save ourself as the
		// result to avoid work for similar cells later.
		//
		if (isCellOff(ur) && isCellOff(lr))
		{
			cp->result = cp->index;

			return cp;
		}

		//
		// Recursively calculate the result for the cut quadrants
		// using an adjusted column number.
		//
		numAddrSub(&col, adjust);

		ur = qcClearToRightFunc(ur, col);
		lr = qcClearToRightFunc(lr, col);

		//
		// Finally combine the changed and unchanged quadrants
		// into a new quadcell and save it as our result.
		//
		newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// The leftmost quadrants are cut and so both of the right
	// quadrants will become empty.
	//
	ur = qcGetEmpty(cp->size - 1);
	lr = qcGetEmpty(cp->size - 1);

	//
	// Recursively calculate the result for the cut quadrants
	// using an adjusted column number.
	//
	numAddrAdd(&col, adjust);

	ul = qcClearToRightFunc(ul, col);
	ll = qcClearToRightFunc(ll, col);

	//
	// Finally combine the changed and unchanged quadrants into a
	// new quadcell and save it as our result.
	//
	newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = newCp->index;

	return newCp;
}


/*
 * Recursive routine to clear the cells at the specified column or to
 * its left.  The states of the other cells are preserved.
 * We check whether we have a result already saved to eliminate
 * unnecessary recursions.  NOTE: We can only calculate a result
 * for cells which are actually cut.  Uncut quadcells will simply return
 * themselves.  Using the saved results depends on the fact that all
 * quadcells of a particular size which are cut are done so on exactly
 * the same column, and so the saved results are always correct.
 */
static QuadCell *
qcClearToLeftFunc(QuadCell * cp, NUMBER col)
{
	QuadCell * newCp;
	NUMBER adjust;

	//
	// If the cell is OFF then nothing can be deleted so we can
	// return ourself.
	//
	if (isCellOff(cp))
		return cp;

	//
	// If we have a result saved then just return that.
	// (The column argument can be ignored since the result must have
	// had the same column argument when it was saved before.)
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// We have to actually do the work.
	// If this is the single cell then we can either return ourself
	// or the OFF cell depending on whether the column is non-positive.
	//
	if (cp->size == SIZE1)
		return (numIsNeg(col) ? cp : qcGetEmpty(SIZE1));

	//
	// Get the bounds on the quadcell.
	//
	NUMBER num = numTwoPower(cp->size - 1);
	NUMBER max = numDec(num);
	NUMBER min = numNeg(num);

	//
	// If the column is to the left of the whole quadcell then
	// nothing is deleted and we can return ourself.  Note: We
	// can't save ourself as a result since that would compete
	// with the results from similar cells which ARE cut.
	//
	if (numIsLess(col, min))
		return cp;

	//
	// If the column is to the right of the whole quadcell then
	// everything is deleted and we can return an empty quadcell.
	// Return the one of the expected size.  Note: We can't
	// save the empty cell as a result since that would compete
	// with the results from similar cells which ARE cut.
	//
	if (numIsGreaterOrEqual(col, max))
		return qcGetEmpty(cp->size);

	//
	// This quadcell is cut by the column.  If the column number is
	// minus one we can simply replace the left two quadrants with
	// empty ones.  Save the result for future use.
	//
	if (numIsNegOne(col))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * newCp = qcFind(emptyIndex, cp->ur, emptyIndex, cp->lr);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// This quadcell is cut other than through its middle.
	// Get the adjustment value for the column number when recursing
	// on the quadrants.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	//
	// See whether the left or the right quadrants are the ones cut.
	// The uncut quadrants will either be unchanged or deleted.
	// The cut quadrants are the ones we will need to recurse on.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	if (numIsNeg(col))
	{
		//
		// The leftmost quadrants are cut.
		// If they are both OFF then nothing will be deleted
		// and so we can return ourself.  Save ourself as the
		// result to avoid work for similar cells later.
		//
		if (isCellOff(ul) && isCellOff(ll))
		{
			cp->result = cp->index;

			return cp;
		}

		//
		// Recursively calculate the result for the cut quadrants
		// using an adjusted column number.
		//
		numAddrAdd(&col, adjust);

		ul = qcClearToLeftFunc(ul, col);
		ll = qcClearToLeftFunc(ll, col);

		//
		// Finally combine the changed and unchanged quadrants
		// into a new quadcell and save it as our result.
		//
		newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// The rightmost quadrants are cut and so both of the left
	// quadrants will become empty.
	//
	ul = qcGetEmpty(cp->size - 1);
	ll = qcGetEmpty(cp->size - 1);

	//
	// Recursively calculate the result for the cut quadrants
	// using an adjusted column number.
	//
	numAddrSub(&col, adjust);

	ur = qcClearToLeftFunc(ur, col);
	lr = qcClearToLeftFunc(lr, col);

	//
	// Finally combine the changed and unchanged quadrants into a
	// new quadcell and save it as our result.
	//
	newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = newCp->index;

	return newCp;
}


/*
 * Recursive routine to clear the cells at the specified row or above it.
 * The states of the other cells are preserved.  We check whether we
 * have a result already saved to eliminate unnecessary recursions.
 * NOTE: We can only calculate a result for cells which are actually cut.
 * Uncut quadcells will simply return themselves.  Using the saved
 * results depends on the fact that all quadcells of a particular size
 * which are cut are done so on exactly the same row, and so the saved
 * results are always correct.
 */
static QuadCell *
qcClearToTopFunc(QuadCell * cp, NUMBER row)
{
	QuadCell * newCp;
	NUMBER adjust;

	//
	// If the cell is OFF then nothing can be deleted so we can
	// return ourself.
	//
	if (isCellOff(cp))
		return cp;

	//
	// If we have a result saved then just return that.
	// (The row argument can be ignored since the result must have
	// had the same row argument when it was saved before.)
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// We have to actually do the work.
	// If this is the single cell then we can either return ourself
	// or the OFF cell depending on whether the row is positive.
	//
	if (cp->size == SIZE1)
		return (numIsPos(row) ? cp : qcGetEmpty(SIZE1));

	//
	// If the row is above the whole quadcell then nothing is
	// deleted and we can return ourself.  Note: We can't save
	// ourself as a result since that would compete with the
	// results from similar cells which ARE cut.
	//
	NUMBER num = numTwoPower(cp->size - 1);

	if (numIsGreaterOrEqual(row, num))
		return cp;

	//
	// If the row is below the whole quadcell then everything is
	// deleted and we can return an empty quadcell.  Return the
	// one of the expected size.  Once again we can't save the
	// empty cell as a result since that would compete with the
	// results from similar cells which ARE cut.
	//
	num = numNeg(num);

	if (numIsLessOrEqual(row, num))
		return qcGetEmpty(cp->size);

	//
	// This quadcell is cut by the row.  If the row number is
	// zero we can simply replace the top two quadrants with empty
	// ones.  Save the result for future use.
	//
	if (numIsZero(row))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * newCp = qcFind(emptyIndex, emptyIndex, cp->ll, cp->lr);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// This quadcell is cut other than through its middle.
	// Get the adjustment value for the row number when recursing
	// on the quadrants.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	//
	// See whether the top or the bottom quadrants are the ones cut.
	// The uncut quadrants will either be unchanged or deleted.
	// The cut quadrants are the ones we will need to recurse on.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	if (numIsPos(row))
	{
		//
		// The upper quadrants are cut.
		// If they are both OFF then nothing will be deleted
		// and so we can return ourself.  Save ourself as the
		// result to avoid work for similar cells later.
		//
		if (isCellOff(ul) && isCellOff(ur))
		{
			cp->result = cp->index;

			return cp;
		}

		//
		// Recursively calculate the result for the cut quadrants
		// using an adjusted row number.
		//
		numAddrSub(&row, adjust);

		ul = qcClearToTopFunc(ul, row);
		ur = qcClearToTopFunc(ur, row);

		//
		// Finally combine the changed and unchanged quadrants
		// into a new quadcell and save it as our result.
		//
		newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// The bottom quadrants are cut and so both of the upper
	// quadrants will become empty.
	//
	ul = qcGetEmpty(cp->size - 1);
	ur = qcGetEmpty(cp->size - 1);

	//
	// Recursively calculate the result for the cut quadrants
	// using an adjusted row number.
	//
	numAddrAdd(&row, adjust);

	ll = qcClearToTopFunc(ll, row);
	lr = qcClearToTopFunc(lr, row);

	//
	// Finally combine the changed and unchanged quadrants into a
	// new quadcell and save it as our result.
	//
	newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = newCp->index;

	return newCp;
}


/*
 * Recursive routine to clear the cells at the specified row or below it.
 * The states of the other cells are preserved.  We check whether we
 * have a result already saved to eliminate unnecessary recursions.
 * NOTE: We can only calculate a result for cells which are actually cut.
 * Uncut quadcells will simply return themselves.  Using the saved
 * results depends on the fact that all quadcells of a particular size
 * which are cut are done so on exactly the same row, and so the saved
 * results are always correct.
 */
static QuadCell *
qcClearToBottomFunc(QuadCell * cp, NUMBER row)
{
	QuadCell * newCp;
	NUMBER adjust;

	//
	// If the cell is OFF then nothing can be deleted so we can
	// return ourself.
	//
	if (isCellOff(cp))
		return cp;

	//
	// If we have a result saved then just return that.
	// (The row argument can be ignored since the result must have
	// had the same row argument when it was saved before.)
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// We have to actually do the work.
	// If this is the single cell then we can either return ourself
	// or the OFF cell depending on whether the row is non-positive.
	//
	if (cp->size == SIZE1)
		return (numIsNeg(row) ? cp : qcGetEmpty(SIZE1));

	//
	// Get the bounds on the quadcell.
	//
	NUMBER num = numTwoPower(cp->size - 1);
	NUMBER max = numDec(num);
	NUMBER min = numNeg(num);

	//
	// If the row is below the whole quadcell then nothing is
	// deleted and we can return ourself.  Note: We can't save
	// ourself as a result since that would compete with the
	// results from similar cells which ARE cut.
	//
	if (numIsLess(row, min))
		return cp;

	//
	// If the row is above the whole quadcell then everything is
	// deleted and we can return an empty quadcell.  Return the
	// one of the expected size.  Note: We can't save the empty
	// cell as a result since that would compete with the results
	// from similar cells which ARE cut.
	//
	if (numIsGreaterOrEqual(row, max))
		return qcGetEmpty(cp->size);

	//
	// This quadcell is cut by the row.  If the row number is
	// minus one we can simply replace the bottom two quadrants with
	// empty ones.  Save the result for future use.
	//
	if (numIsNegOne(row))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * newCp = qcFind(cp->ul, cp->ur, emptyIndex, emptyIndex);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// This quadcell is cut other than through its middle.
	// Get the adjustment value for the row number when recursing
	// on the quadrants.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	//
	// See whether the upper or the lower quadrants are the ones cut.
	// The uncut quadrants will either be unchanged or deleted.
	// The cut quadrants are the ones we will need to recurse on.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	if (numIsNeg(row))
	{
		//
		// The bottommost quadrants are cut.
		// If they are both OFF then nothing will be deleted
		// and so we can return ourself.  Save ourself as the
		// result to avoid work for similar cells later.
		//
		if (isCellOff(ll) && isCellOff(lr))
		{
			cp->result = cp->index;

			return cp;
		}

		//
		// Recursively calculate the result for the cut quadrants
		// using an adjusted column number.
		//
		numAddrAdd(&row, adjust);

		ll = qcClearToBottomFunc(ll, row);
		lr = qcClearToBottomFunc(lr, row);

		//
		// Finally combine the changed and unchanged quadrants
		// into a new quadcell and save it as our result.
		//
		newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

		cp->result = newCp->index;

		return newCp;
	}

	//
	// The topmost quadrants are cut and so both of the bottom
	// quadrants will become empty.
	//
	ll = qcGetEmpty(cp->size - 1);
	lr = qcGetEmpty(cp->size - 1);

	//
	// Recursively calculate the result for the cut quadrants
	// using an adjusted column number.
	//
	numAddrSub(&row, adjust);

	ul = qcClearToBottomFunc(ul, row);
	ur = qcClearToBottomFunc(ur, row);

	//
	// Finally combine the changed and unchanged quadrants into a
	// new quadcell and save it as our result.
	//
	newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = newCp->index;

	return newCp;
}


/*
 * Extract the specified row from a quadcell.
 * All of the other rows will be OFF.
 */
QuadCell *
qcExtractRow(QuadCell * cp, NUMBER row)
{
	if (qcIsOff(cp) || !qcIsCoordinateValid(cp, row))
		return qcEmpty();

	qcSetResultType(RESULT_NONE);

	return qcExtractRowFunc(cp, row);
}


/*
 * Recursive routine to extract the specified row from a quadcell.
 */
static QuadCell *
qcExtractRowFunc(QuadCell * cp, NUMBER row)
{
	NUMBER adjust;

	//
	// If we are OFF or a single cell then return ourself.
	//
	if (qcIsOff(cp) || (cp->size == SIZE1))
		return cp;

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// We have to calculate the result.
	// Assume the subquadrants are OFF since half of them will be.
	//
	QuadCell * ul = qcGetEmpty(cp->size - 1);
	QuadCell * ur = ul;
	QuadCell * ll = ul;
	QuadCell * lr = ul;

	//
	// Determine whether the row is in the upper or lower half
	// of this quadcell.  Adjust the row number for the recursion
	// and call ourself for the two quadcells containing the row.
	// The other two subquadrants will be OFF.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	if (numIsNonNeg(row))
	{
		numAddrSub(&row, adjust);

		ul = qcExtractRowFunc(indexToQuadCell(cp->ul), row);
		ur = qcExtractRowFunc(indexToQuadCell(cp->ur), row);
	}
	else
	{
		numAddrAdd(&row, adjust);

		ll = qcExtractRowFunc(indexToQuadCell(cp->ll), row);
		lr = qcExtractRowFunc(indexToQuadCell(cp->lr), row);
	}

	//
	// Finally combine quadrants into a new quadcell and save
	// it as our result.
	//
	QuadCell * newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = newCp->index;

	return newCp;
}


/*
 * Delete the specified row from a quadcell.
 */
QuadCell *
qcDeleteRow(QuadCell * cp, NUMBER row)
{
	if (qcIsOff(cp) || !qcIsCoordinateValid(cp, row))
		return cp;

	qcSetResultType(RESULT_NONE);

	return qcDeleteRowFunc(cp, row);
}


/*
 * Recursive routine to delete the specified row from a quadcell.
 */
static QuadCell *
qcDeleteRowFunc(QuadCell * cp, NUMBER row)
{
	QuadCell * ul;
	QuadCell * ur;
	QuadCell * ll;
	QuadCell * lr;
	NUMBER adjust;

	//
	// If we are OFF or a single cell then return an empty cell.
	//
	if (qcIsOff(cp))
		return cp;

	if (cp->size == SIZE1)
		return qcGetEmpty(SIZE1);

	//
	// See if we have a saved result.  If so return it.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// We have to calculate the result.
	// Determine whether the row is in the upper or lower half
	// of this quadcell.  Adjust the row number for the recursion
	// and call ourself for the two quadcells containing the row.
	//
	if (cp->size == SIZE2)
		adjust = numZero();
	else
		adjust = numTwoPower(cp->size - 2);

	if (numIsNonNeg(row))
	{
		numAddrSub(&row, adjust);

		ul = qcDeleteRowFunc(indexToQuadCell(cp->ul), row);
		ur = qcDeleteRowFunc(indexToQuadCell(cp->ur), row);
		ll = indexToQuadCell(cp->ll);
		lr = indexToQuadCell(cp->lr);
	}
	else
	{
		numAddrAdd(&row, adjust);

		ul = indexToQuadCell(cp->ul);
		ur = indexToQuadCell(cp->ur);
		ll = qcDeleteRowFunc(indexToQuadCell(cp->ll), row);
		lr = qcDeleteRowFunc(indexToQuadCell(cp->lr), row);
	}

	//
	// Finally combine quadrants into a new quadcell and save
	// it as our result.
	//
	QuadCell * newCp = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = newCp->index;

	return newCp;
}


/*
 * Set the result type without a parameter for an operation that is
 * about to start.  If the result type is being changed then all
 * results are cleared.  The RESULT_NONE type always clears the results.
 */
void
qcSetResultType(ResultType resultType)
{
	qcSetResultTypeWithParam(resultType, numZero());
}


/*
 * Set the result type and an associated parameter for an operation that
 * is about to start.  If the result type or the parameter is being
 * changed then all results are cleared.  The RESULT_NONE type always
 * clears the results.
 */
void
qcSetResultTypeWithParam(ResultType resultType, NUMBER resultParam)
{
	if ((resultType != currentResultType) ||
		(resultType == RESULT_NONE) ||
		numIsNotEqual(resultParam, currentResultParam))
	{
		qcClearResults();

		currentResultType = resultType;
		currentResultParam = resultParam;
	}
}


/*
 * Clear the results from all quadcells either before or after a specific
 * type of operation.  This is used as necessary so that a different type
 * of operation will not get invalid answers.
 */
void
qcClearResults(void)
{
	QUADINDEX index;

	for (index = 0; index < quadNextIndex; index++)
	{
		QuadCell * cp = indexToQuadCell(index);
		cp->result = INDEX_NULL;
	}

	currentResultType = RESULT_NONE;
}


/*
 * Clear the results from almost all quadcells.  This preserves results
 * from the SIZE2 cells, which is useful when running generations.
 */
void
qcClearMostResults(void)
{
	QUADINDEX index;

	for (index = 0; index < quadNextIndex; index++)
	{
		QuadCell * cp = indexToQuadCell(index);

		if (cp->size > SIZE2)
			cp->result = INDEX_NULL;
	}

	currentResultType = RESULT_NONE;
}

