/*
 * Copyright (c) 2015 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to implement quadpairs, which are just a cache used to map
 * pairs of input quadcells to result quadcells.  This cache is used to
 * avoid many recursive calls when calculating operations on two quadcell
 * trees (such as OR and AND).  If the quadcells have a reasonable amount
 * of regularity, then this usage turns an exponential algorithm into an
 * almost linear one, allowing the operations to be performed on very
 * large objects.
 *
 * QuadPairs are only used temporarily during a calculation and cannot be
 * stored by the caller.  However, if multiple operations are performed
 * of the same type then the results from the previous operations might
 * still be useful.  Quadpairs are not individually freed.  Instead, all
 * of the quadpairs are freed at once when necessary.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "quad.h"
#include "quadpriv.h"


//
// The number of quadPairs to allocate at one time.
// This should be a power of two.
//
#define	PAIR_ALLOC_SIZE	(1024 * 16)
#define	MAX_PAIR_CHUNKS	100000

typedef	struct
{
	QuadPair table[PAIR_ALLOC_SIZE];
} PairChunk;


/*
 * Hash function for a quadpair.
 * This produces a random-like number based on the quadrant indexes.
 */
typedef	unsigned int HASH;

#define	MAKE_PAIR_HASH(quad1, quad2) \
	((HASH) ((quad1) * 51515159 + (quad2) * 100000007))

/*
 * The index value indicating no quadpair.
 */
#define	PAIR_INDEX_NULL -1

#define	isNullPairIndex(index) ((index) == PAIR_INDEX_NULL)
#define	isNotNullPairIndex(index) ((index) != PAIR_INDEX_NULL)


/*
 * The hash table.
 * This is grown as necessary to keep the length of the chains small.
 */
static int hashTableSize;
static int hashTableUsed;

static PAIRINDEX * hashTable;


/*
 * The table of all the quadpairs which have ever been allocated.
 * These are accessed using their PAIRINDEX values using the divide and
 * modulo operations by the PAIR_ALLOC_SIZE value.  The quadPairNextIndex
 * value indicates the first index in the table which is not yet used.
 */
static PAIRINDEX quadPairNextIndex;

static PairChunk * quadPairTable[MAX_PAIR_CHUNKS];


/*
 * Return the pointer to a QuadPair given its PAIRINDEX value.
 */
inline static QuadPair *
indexToQuadPair(PAIRINDEX pairIndex)
{
	PairChunk * chunk;

	verify((pairIndex >= 0) && (pairIndex < quadPairNextIndex));

	chunk = quadPairTable[pairIndex / PAIR_ALLOC_SIZE];

	return &chunk->table[pairIndex % PAIR_ALLOC_SIZE];
}


//
// Forward declarations.
//
static QuadPair * qcAllocQuadPair(QUADINDEX quad1, QUADINDEX quad2);
static void qcGrowQuadPairTable(void);


/*
 * Search the cache for the quadpair for the specified two quadcells.
 * The isSymmetrical flag is set if the order of the quadcells is unimportant.
 * If the requested quadpair is found in the cache, then that is returned.
 * Otherwise a new quadpair is allocated and added to the cache.
 * If this is a new quadpair then the result element is the null index,
 * otherwise the result is what was stored by a previous caller.
 */
QuadPair *
qcFindQuadPair(QUADINDEX quad1, QUADINDEX quad2, BOOL isSymmetrical)
{
	//
	// If the indexes are symmetrical then sort them.
	// This hopefully will reduce the number of quadpairs used.
	//
	if (isSymmetrical && (quad1 > quad2))
	{
		QUADINDEX tempIndex = quad1;
		quad1 = quad2;
		quad2 = tempIndex;
	}

	//
	// Calculate a hash for the two indexes and then the index into
	// the hash table.
	//
	HASH hash = MAKE_PAIR_HASH(quad1, quad2);
	int hashIndex = hash % hashTableSize;

	//
	// Search the hash table chain for the quadpair having the indexes.
	// If the quadpair is found then return it.
	//
	PAIRINDEX pairIndex = hashTable[hashIndex]; 

	while (isNotNullPairIndex(pairIndex))
	{
		QuadPair * qp = indexToQuadPair(pairIndex);

		if ((qp->quad1 == quad1) && (qp->quad2 == quad2))
		{
			quadStats.quadPairHits++;

			return qp;
		}

		pairIndex = qp->next;
	}

	//
	// The quad pair wasn't found, so allocate a new one.
	//
	quadStats.quadPairMisses++;

	return qcAllocQuadPair(quad1, quad2);
}


/*
 * Initialise the quadpair data structures.
 */
void
qcInitQuadPairs(void)
{
	int index;

	//
	// Clear the chunk pointers.
	//
	for (index = 0; index < MAX_PAIR_CHUNKS; index++)
		quadPairTable[0] = 0;

	//
	// Allocate the initial hash table.
	//
	qcGrowQuadPairTable();
}


/*
 * Free all of the quadpairs so that we can begin a new operation.
 * This doesn't actually free any memory, it simply lets the previously
 * allocated memory be reused.
 */
void
qcFreeQuadPairs(void)
{
	int index;

	//
	// Reset the allocation index and the amount of the hash table used.
	//
	quadPairNextIndex = 0;
	hashTableUsed = 0;

	//
	// Clear all of the hash chains.
	//
	for (index = 0; index < hashTableSize; index++)
		hashTable[index] = PAIR_INDEX_NULL;
}


/*
 * Allocate and return a new quadpair for the specified two quadcells.
 * This allocates a few at a time to minimize malloc calls.
 * Newly allocated quadpairs are added to the quadpair and hash tables.
 * The allocated index of the quadpair is only known within this routine
 * since it has no use other than to manage the hash chain.
 */
static QuadPair *
qcAllocQuadPair(QUADINDEX quad1, QUADINDEX quad2)
{
	QuadPair * qp;
	PAIRINDEX pairIndex;

	/*
	 * If no new quadcells are available then possibly allocate
	 * another chunk and store it into the quadcell table.
	 * Chunks might not need to be allocated if the quadpairs were
	 * previously freed.
	 */
	if ((quadPairNextIndex % QUAD_ALLOC_SIZE) == 0)
	{
		int tableIndex = quadPairNextIndex / QUAD_ALLOC_SIZE;

		verify(tableIndex < MAX_QUAD_CHUNKS);

		PairChunk * chunk = quadPairTable[tableIndex];

		if (chunk == 0)
		{
			chunk = qcAllocMem("quadpairs", sizeof(PairChunk));

			quadPairTable[tableIndex] = chunk;
		}
	}

	/*
	 * Get the new quadpair's address and set its values.
	 * Note: We increment the quadPairNextIndex value first so that
	 * the indexToQuadPair call will not fail its verification.
	 */
	pairIndex = quadPairNextIndex++;

	qp = indexToQuadPair(pairIndex);
	qp->quad1 = quad1;
	qp->quad2 = quad2;
	qp->result = PAIR_INDEX_NULL;

	//
	// Calculate which hash chain will contain the new quad pair
	// and add the quadpair into it.
	//
	HASH hash = MAKE_PAIR_HASH(quad1, quad2);
	int hashIndex = hash % hashTableSize;

	qp->next = hashTable[hashIndex];
	hashTable[hashIndex] = pairIndex;

	hashTableUsed++;

	//
	// If the table is getting full then reallocate it.
	//
	if (hashTableUsed * 4 > hashTableSize * 3)
		qcGrowQuadPairTable();

	return qp;
}


/*
 * Grow the quadpair hash table so that the average length of the
 * chains is shorter.  The table is grown to be odd and be about 1.5
 * times its current size.  This can be called even if the hash table
 * isn't yet allocated.
 */
static void
qcGrowQuadPairTable(void)
{
	PAIRINDEX * newHashTable;
	int newUsed = 0;
	int tableIndex;

	//
	// Calculate a new table size which is a bit larger then the old one.
	//
	int newSize = qcNextTableSize(hashTableSize);

	//
	// Allocate the new hash table and initialise its elements to
	// make each hash chain empty.
	//
	newHashTable = qcAllocMem("hash chain", sizeof(PAIRINDEX) * newSize);

	for (tableIndex = 0; tableIndex < newSize; tableIndex++)
		newHashTable[tableIndex] = PAIR_INDEX_NULL;

	/*
	 * Walk through the old hash chains and add their quadpairs
	 * back into the new ones.
	 */
	for (tableIndex = 0; tableIndex < hashTableSize; tableIndex++)
	{
		PAIRINDEX pairIndex = hashTable[tableIndex];

		//
		// Loop over the cells in the chain.
		//
		while (isNotNullPairIndex(pairIndex))
		{
			QuadPair * qp = indexToQuadPair(pairIndex);

			PAIRINDEX nextIndex = qp->next;

			HASH hash = MAKE_PAIR_HASH(qp->quad1, qp->quad2);
			int hashIndex = hash % newSize;

			qp->next = newHashTable[hashIndex];
			newHashTable[hashIndex] = pairIndex;

			newUsed++;

			pairIndex = nextIndex;
		}
	}

	/*
	 * Verify that the number of quadpairs in the hash table is unchanged.
	 * Then free the old hash table if necessary and store the new one.
	 */
	verifyAlways(newUsed == hashTableUsed);

	if (hashTableSize > 0)
		free(hashTable);

	hashTable = newHashTable;
	hashTableSize = newSize;
}


/*
 * Remember the operation and a parameter we are about to calculate,
 * and if it is changing then free the quadpair cache since old
 * results are wrong.
 */
void
qcSetQuadPairOperation(PairOperation operation, NUMBER parameter)
{
	static PairOperation oldOperation = OP_NONE;
	static NUMBER oldParameter;

	if ((operation != oldOperation) || !numIsEqual(parameter, oldParameter))
		qcFreeQuadPairs();

	oldOperation = operation;
	oldParameter = parameter;
}

