/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines for manipulating paths for quadcells.
 */

#include "life.h"
#include "quadpriv.h"
#include "quadalg.h"


/*
 * Callback information provided during the scan of ON cells.
 */
typedef struct
{
	const Path * path;
	QuadCell * cells;
	QuadCell * marks;
	NUMBER insideCount;
	NUMBER outsideCount;
	NUMBER boundaryCount;
	BOOL isInsideDone;
	BOOL isOutsideDone;
	BOOL isBoundaryDone;
}
ScanInfo;


//
// Local routines.
//
static BOOL MarkCallback(NUMBER row, NUMBER col, int state, void * data);
static BOOL CountCallback(NUMBER row, NUMBER col, int state, void * data);


/*
 * Select all cells which are contained within the path for the object.
 * Returns FALSE if there are no cells within the path.
 * This treats NEW and MARK cells are being ON.
 * An error is generated if there are too many cells to check.
 * Warning: This routine performs garbage collections since many
 * cells can be added.
 */
BOOL
qcSelectPath(LifeObject * obj)
{
	const Path * path = &obj->path;
	QuadCell * cells = qcGetReducedQuadCell(obj);

	if ((path->count <= 0) || qcIsEmpty(cells))
		return FALSE;

	//
	// Extract the cells in the bounding region of the path.
	// In this way a really huge object can still use paths if
	// the bounding box area isn't too large.
	//
	cells = qcExtractRegion(cells, path->minRow, path->maxRow,
		path->minCol, path->maxCol);

	if (qcIsEmpty(cells))
		return FALSE;

	//
	// If the extracted region has too many cells then fail.
	//
	NUMBER count = qcGetCount(cells);

	if (numCmpLong(count, MAX_PATH_CELLS) > 0)
		Error("Too many cells to be marked");

	//
	// Visit all the cells in the region.
	// The callback routine will add those cells which are within
	// the path into a set of marks.
	//
	ScanInfo info;

	info.path = path;
	info.cells = cells;
	info.marks = qcEmpty();

	if (!qcVisitCells(cells, STATE_ON, MarkCallback, &info))
		return FALSE;

	//
	// Return FALSE if nothing was marked.
	//
	if (qcIsEmpty(info.marks))
		return FALSE;

	//
	// Add the marked cells to the selection.
	//
	obj->selectedQuadCells =
		qcOrOperation(obj->selectedQuadCells, info.marks);

	return TRUE;
}


/*
 * Callback to mark a cell if it is within the path.
 * Returns TRUE if the scan is to be continued.
 */
static BOOL
MarkCallback(NUMBER row, NUMBER col, int state, void * data)
{
	ScanInfo * info = data;

	if (PointPathSignum(info->path, row, col) <= 0)
	{
		info->marks = qcSet(info->marks, row, col);

		//
		// Protect our temporary quadcells from garbage
		// collection, and then check whether we should do one.
		// Remove the protection afterwards.
		//
		qcAddSafeQuadCell(info->cells);
		qcAddSafeQuadCell(info->marks);
		qcCheckGarbageCollect();
		qcClearSafeQuadCells(2);
	}

	return !Life.stop;
}


/*
 * Return the number of cells which are truly inside, exactly on the
 * boundary of, or truly outside of the current path of the specified object.
 * If there is no current path defined, then all returned values are zero.
 * Each cell of the object contributes to exactly one of these counts.
 * The values are returned through the indicated pointers.  If a pointer
 * is NULL, that value is not returned.  Multiple pointer values can specify
 * the same return variable, which will sum the indicated counts into that
 * variable.
 *
 * Note: This is a very expensive routine, and so we cheat by only
 * returning nonzero counts instead of the full counts.  (It is only
 * called by a place which isn't interested in the full counts.)
 */
void
qcCountPathCells(const LifeObject * obj,
	NUMBER * insideCount, NUMBER * boundaryCount, NUMBER * outsideCount)
{
	const Path * path = &obj->path;
	//TODO: Fix const problem.
	QuadCell * cells = qcGetReducedQuadCell((LifeObject *) obj);
	NUMBER dummyCount;
	BOOL isInsideDone = FALSE;
	BOOL isOutsideDone = FALSE;
	BOOL isBoundaryDone = FALSE;

	/*
	 * Default the pointer values to a dummy count value if necessary
	 * and remember which counts we still want.
	 */
	if (insideCount == 0)
	{
		isInsideDone = TRUE;
		insideCount = &dummyCount;
	}

	if (boundaryCount == 0)
	{
		isBoundaryDone = TRUE;
		boundaryCount = &dummyCount;
	}

	if (outsideCount == 0)
	{
		isOutsideDone = TRUE;
		outsideCount = &dummyCount;
	}

	*insideCount = numZero();
	*boundaryCount = numZero();
	*outsideCount = numZero();

	//
	// If there are no cells then the counts are all zero.
	//
	if (qcIsEmpty(cells))
		return;

	//
	// If there is no path then all the cells are outside.
	//
	if (path->count <= 0)
	{
		*outsideCount = numOne();

		return;
	}

	//
	// Extract the cells in the bounding region of the path.
	// In this way a really huge object can still use paths if
	// the bounding box area isn't too large.  If there are no
	// cells in the boundary box then all cells are outside.
	//
	cells = qcExtractRegion(cells, path->minRow, path->maxRow,
		path->minCol, path->maxCol);

	if (qcIsEmpty(cells))
	{
		*outsideCount = numOne();

		return;
	}

	//
	// There are outside cells if the region didn't get them all.
	//
	if (!qcIsEqual(cells, obj->reducedQuadCells))
	{
		isOutsideDone = TRUE;
		*outsideCount = numOne();
	}

	if (isInsideDone && isOutsideDone && isBoundaryDone)
		return;

	//
	// If the extracted region has too many cells then fail.
	//
	NUMBER count = qcGetCount(cells);

	if (numCmpLong(count, MAX_PATH_CELLS) > 0)
		Error("Too many cells to be counted");

	//
	// Now visit the cells within the extracted region.
	//
	ScanInfo info;

	info.path = path;
	info.insideCount = numZero();
	info.outsideCount = *outsideCount;
	info.boundaryCount = numZero();
	info.isInsideDone = isInsideDone;
	info.isOutsideDone = isOutsideDone;
	info.isBoundaryDone = isBoundaryDone;

	qcVisitCells(cells, STATE_ON, CountCallback, &info);

	//
	// Copy back the results.
	//
	*insideCount = info.insideCount;
	*outsideCount = info.outsideCount;
	*boundaryCount = info.boundaryCount;
}


/*
 * Callback to count a cell as being inside, outside, or on a path.
 * Returns TRUE if the scan is to be continued.
 */
static BOOL
CountCallback(NUMBER row, NUMBER col, int state, void * data)
{
	ScanInfo * info = data;

	//
	// Determine the location of the point in the path and increment
	// the corresponding counter.
	//
	int sigNum = PointPathSignum(info->path, row, col);

	if (sigNum < 0)
	{
		info->isInsideDone = TRUE;
		info->insideCount = numOne();
	}
	else if (sigNum > 0)
	{
		info->isOutsideDone = TRUE;
		info->outsideCount = numOne();
	}
	else
	{
		info->isBoundaryDone = TRUE;
		info->boundaryCount = numOne();
	}

	//
	// If we have seen all the cases then stop the scan.
	//
	if (info->isInsideDone && info->isOutsideDone && info->isBoundaryDone)
		return FALSE;

	//
	// Continue unless the user wants to stop.
	//
	return !Life.stop;
}

