/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Private include file for quadcell definition and external routines.
 * Callers into the quadcell routines should not include this.
 */
#ifndef	QUADPRIV_H
#define	QUADPRIV_H

#include "quad.h"
#include "verify.h"


/*
 * Important quadcell sizes for the algorithms.
 */
#define	SIZE1	0		/* offCell, onCell , deadCell, newCell or markCell */
#define	SIZE2	1		/* a 2 by 2 square of cells */
#define	SIZE4	2		/* a 4 by 4 square which can have a RESULT */
#define	SIZE8	3		/* a 8 by 8 square */
#define	SIZE_INITIAL	1	/* initial size of a new object - 2 by 2 */
#define	MAXSIZE	100		/* the maximum size we can handle */


/*
 * Special index values for the quadcells.
 * The relative order of these values are important.
 * The 625 - 5 in the FIRST_CACHED value comes because there are five
 * quadcells in the 625 elements of the size2Cells table which have pure
 * states and so are duplicates of the already allocated quadcells in the
 * other tables.
 */
#define INDEX_NULL		((QUADINDEX) -1)

#define	INDEX_FIRST_OFF		((QUADINDEX) 0)
#define	INDEX_LAST_OFF		((QUADINDEX) (1 * MAXSIZE - 1))

#define	INDEX_FIRST_DEAD	((QUADINDEX) (1 * MAXSIZE))
#define	INDEX_LAST_DEAD		((QUADINDEX) (2 * MAXSIZE - 1))

#define	INDEX_FIRST_ON		((QUADINDEX) (2 * MAXSIZE))
#define	INDEX_LAST_ON		((QUADINDEX) (3 * MAXSIZE - 1))

#define	INDEX_FIRST_NEW		((QUADINDEX) (3 * MAXSIZE))
#define	INDEX_LAST_NEW		((QUADINDEX) (4 * MAXSIZE - 1))

#define	INDEX_FIRST_MARK	((QUADINDEX) (4 * MAXSIZE))
#define	INDEX_LAST_MARK		((QUADINDEX) (5 * MAXSIZE - 1))

#define	INDEX_FIRST_CACHED	((QUADINDEX) ((5 * MAXSIZE) + 625 - 5))


/*
 * Tests for the types of the SIZE1 quadcells based on their indexes.
 * Note: These checks are only valid for pure quadcells.
 * Non-pure quadcells must check the state instead.
 */
#define	isNull(index)	((index) <= INDEX_NULL)
#define	isNotNull(index)	((index) > INDEX_NULL)

#define	isOff(index)	(((index) >= INDEX_FIRST_OFF) && ((index) <= INDEX_LAST_OFF))
#define	isDead(index)	(((index) >= INDEX_FIRST_DEAD) && ((index) <= INDEX_LAST_DEAD))
#define	isOn(index)	(((index) >= INDEX_FIRST_ON) && ((index) <= INDEX_LAST_ON))
#define	isNew(index)	(((index) >= INDEX_FIRST_NEW) && ((index) <= INDEX_LAST_NEW))
#define	isMark(index)	(((index) >= INDEX_FIRST_MARK) && ((index) <= INDEX_LAST_MARK))
#define	isPure(index)	(((index) >= INDEX_FIRST_OFF) && ((index) <= INDEX_LAST_MARK))
#define	isNotOff(index)	((index) >= INDEX_FIRST_DEAD)

#define	isCached(index)	((index) >= INDEX_FIRST_CACHED)
#define	offIndexFromSize(size)	(INDEX_FIRST_OFF + (size))
#define	sizeFromPureIndex(index)	((index) % MAXSIZE)


//
// The number of quadcells to allocate at one time.
// This should be a power of two.
//
#define	QUAD_ALLOC_SIZE	(1024 * 16)
#define	MAX_QUAD_CHUNKS	100000

typedef	struct
{
	QuadCell table[QUAD_ALLOC_SIZE];
} QuadChunk;


/*
 * An index value for quadpairs.
 * These are used because pointers take 8 bytes instead of the 4 bytes
 * that an index takes.  The quadpair can be easily found from its index.
 * We only use the indexes to link together quadpairs into hash chains.
 */
typedef	int PAIRINDEX;


/*
 * A data structure used to operate on pairs of quadcells with a result
 * for them (for example, in AND and OR operations).  For symmetrical
 * operations the quad1 and quad2 indexes are sorted so that both orders
 * of using them results in the same quadpair.  Quadpairs are only used
 * during calculations for an operation and are all freed at one time.
 */
typedef struct
{
	QUADINDEX quad1;	// first quad cell
	QUADINDEX quad2;	// second quad cell
	QUADINDEX result;	// result of operation
	PAIRINDEX next;		// next quadpair in hash chain
} QuadPair;


/*
 * The type of operation that was last used for the quadpairs.
 */
typedef enum
{
	OP_NONE, OP_OVERLAPPING, OP_AND, OP_OR, OP_XOR, OP_ANDNOT,
	OP_LEFTSHIFT, OP_RIGHTSHIFT, OP_UPSHIFT, OP_DOWNSHIFT
} PairOperation;


/*
 * The type of result that is to be calculated in an operation.
 * This allows multiple operations to correctly use results from a
 * previous operation of the same type.  Not all operations are
 * listed here.  Those that aren't use RESULT_NONE which means that
 * all quadcells need clearing.
 */
typedef enum
{
	RESULT_NONE, RESULT_GEN, RESULT_GENHIST, RESULT_AND, RESULT_OR,
	RESULT_XOR, RESULT_INVERT, RESULT_ANDNOT, RESULT_REDUCE,
	RESULT_REMOVEHISTORY, RESULT_RAISE
} ResultType;


/*
 * Global data.
 */
extern QuadChunk * quadCellTable[MAX_QUAD_CHUNKS];

extern QUADINDEX quadNextIndex;
extern QuadStats quadStats;


/*
 * Return the pointer to a QuadCell given its QUADINDEX value.
 * In almost all cases the returned quadcell contains valid values,
 * except temporarily during its allocation.
 */
inline static QuadCell *
indexToQuadCell(QUADINDEX quadIndex)
{
	QuadChunk * chunk;

	verify((quadIndex >= 0) && (quadIndex < quadNextIndex));

	chunk = quadCellTable[quadIndex / QUAD_ALLOC_SIZE];

	return &chunk->table[quadIndex % QUAD_ALLOC_SIZE];
}


/*
 * Internal routines.
 */
extern QuadCell * qcGetEmpty(int size);
extern QuadCell * qcGetFull(int size);
extern QuadCell * qcGetCellFromState(int size, int state);

extern QuadCell * qcGrow(QuadCell * cp);

extern QuadCell * qcGetSize2Cell(int state1, int state2, int state3, int state4);

extern void * qcAllocMem(const char * what, int size);

extern QuadCell * qcBottomResult(QuadCell * cp, int twoPower);
extern QuadCell * qcResult(QuadCell * cp);

extern int qcNextTableSize(int oldSize);

extern QuadPair * qcFindQuadPair(QUADINDEX quad1, QUADINDEX quad2,
	BOOL isSymmetrical);

extern void qcClearResults(void);
extern void qcClearMostResults(void);
extern void qcSetResultType(ResultType result);
extern void qcSetResultTypeWithParam(ResultType result, NUMBER param);

extern void qcClearFlags(QuadCell * cp);
extern void qcClearNumbers(QuadCell * cp);

extern void qcInitQuadPairs(void);
extern void qcFreeQuadPairs(void);

extern void qcSetQuadPairOperation(PairOperation operation, NUMBER parameter);

#endif

