/*
 * Copyright (c) 2015 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Read a file in macro-cell format.
 * Example pattern (a glider):
 * [M1] (hlife 0.96)
 * $$$$$$$.*$
 * *$***$
 * 4 0 1 0 2
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "life.h"
#include "quad.h"
#include "quadpriv.h"


/*
 * The table of allocated cells.
 * This is needed in order to lookup previously defined cells.
 */
#define	INITIAL_TABLE_SIZE	1024

static QuadCell ** table = 0;
static int tableAlloc;
static int tableUsed;

/*
 * Local routines.
 */
static QuadCell * parseBottomLevel(const char * line);
static QuadCell * parseUpperLevel(const char * line);
static BOOL getLine(FILE * fp, char * buf, int bufLen);
static void addCell(QuadCell * cp);
static void freeResources(const char * errorText);


/*
 * Read an object in MC (macro-cell) format from the specified file
 * handle and return the resulting quadcell structure.  Generates an
 * error if it is invalid.
 */
QuadCell *
qcReadStructure(FILE * fp)
{
	QuadCell * cp;
	char line[128];

	table = 0;
	tableAlloc = 0;
	tableUsed = 0;

	/*
	 * Skip lines which are empty or which start with hash marks.
	 */
	while (getLine(fp, line, sizeof(line)))
	{
		if (*line && (*line != '#'))
			break;
	}

	/*
	 * Make sure the line looks like the start of a macro-cell definition.
	 * Ignore the version value (maybe it will still be ok).
	 */
	if ((line[0] != '[') || (line[1] != 'M') || (line[3] != ']'))
	{
		freeResources("Macro-cell format introduction not found");
	}

	/*
	 * Now read the macro-cell lines.
	 * Start the table off with an empty cell.
	 * Then the actual macro-cells read will have indexes starting with 1.
	 */
	addCell(qcGetEmpty(SIZE1));

	while (getLine(fp, line, sizeof(line)))
	{
		/*
		 * If the line starts with an exclamation point then stop.
		 * This isn't exactly right, but the defined format doesn't
		 * have any termination except EOF, which is an oversight.
		 */
		if (*line == '!')
			break;

		/*
		 * Skip comment or blank lines.
		 */
		if ((*line == '\0') || (*line == '#'))
			continue;

		/*
		 * If the line has pictorial characters then it defines a
		 * "bottom" size 3 macro-cell (8 by 8).  Otherwise it is a
		 * higher sized cell which refers to previous macro-cells.
		 */
		if ((*line == '.') || (*line == '*') || (*line == '$'))
			cp = parseBottomLevel(line);
		else
			cp = parseUpperLevel(line);

		/*
		 * Add the macro-call to the table, giving it the next index
		 * which can then be referred to by later lines in the file.
		 */
		addCell(cp);
	}

	/*
	 * Get the final macro-cell from the table and return it.
	 */
	cp = table[tableUsed - 1];

	freeResources(0);

	return cp;
}


/*
 * Parse the "bottom" level data for an 8 by 8 size 3 macro-cell.
 * This is a pictorial format of the 64 cells in a square where '.'
 * is an OFF cell, '*' is an ON cell, '$' is the end of a row,
 * and cells which aren't explicitly specified are OFF.
 */
static QuadCell *
parseBottomLevel(const char * line)
{
	int row;
	int col;
	QuadCell * ul;
	QuadCell * ur;
	QuadCell * ll;
	QuadCell * lr;
	QUADINDEX size0[8][8];
	QuadCell * size1[4][4];
	QuadCell * size2[2][2];

	/*
	 * Initialise the size 0 table to the index of the OFF cell.
	 */
	for (row = 0; row < 8; row++)
	{
		for (col = 0; col < 8; col++)
			size0[row][col] = INDEX_FIRST_OFF;
	}

	/*
	 * Walk through the characters on the line setting the ON
	 * indexes in the size0 table as indicated.
	 */
	row = 0;
	col = 0;

	while (*line)
	{
		switch (*line++)
		{
			case '.':
				col++;
				break;

			case '*':
				if ((row >= 8) || (col >= 8))
					freeResources("Invalid cell in lowest level");

				size0[row][col++] = INDEX_FIRST_ON;

				break;

			case '$':
				row++;
				col = 0;

				break;

			default:
				freeResources("Invalid char in lowest level");
		}
	}

	/*
	 * Build the size 1 macro-cells from the size 0 data.
	 */
	for (row = 0; row < 4; row++)
	{
		for (col = 0; col < 4; col++)
		{
			NUMINDEX ulIndex = size0[row * 2][col * 2];
			NUMINDEX urIndex = size0[row * 2][col * 2 + 1];
			NUMINDEX llIndex = size0[row * 2 + 1][col * 2];
			NUMINDEX lrIndex = size0[row * 2 + 1][col * 2 + 1];

			size1[row][col] =
				qcFind(ulIndex, urIndex, llIndex, lrIndex);
		}
	}

	/*
	 * Build the size 2 macro-cells from the size 1 data.
	 */
	for (row = 0; row < 2; row++)
	{
		for (col = 0; col < 2; col++)
		{
			ul = size1[row * 2][col * 2];
			ur = size1[row * 2][col * 2 + 1];
			ll = size1[row * 2 + 1][col * 2];
			lr = size1[row * 2 + 1][col * 2 + 1];

			size2[row][col] = qcFind(
				ul->index, ur->index, ll->index, lr->index);
		}
	}

	/*
	 * Finally build the size 3 macro-cell.
	 */
	ul = size2[0][0];
	ur = size2[0][1];
	ll = size2[1][0];
	lr = size2[1][1];

	return qcFind(ul->index, ur->index, ll->index, lr->index);
}


/*
 * Parse a line to create an upper level macro-cell (size 4 or larger).
 * This is of the format:
 *	size ulIndex urIndex llIndex lrIndex
 * where the indexes are zero for empty quadrants or else the index
 * in our table of previously read macro-cells.
 */
static QuadCell *
parseUpperLevel(const char * line)
{
	int size;
	int halfSize;
	int ulIndex;
	int urIndex;
	int llIndex;
	int lrIndex;
	QuadCell * ul;
	QuadCell * ur;
	QuadCell * ll;
	QuadCell * lr;

	/*
	 * Parse five integers from the line.
	 * These are the size of the macro-cell and the indexes of
	 * its four quadrant macro-cells.
	 */
	int count = sscanf(line, "%d %d %d %d %d",
		&size, &ulIndex, &urIndex, &llIndex, &lrIndex);

	if (count != 5)
		freeResources("Invalid macro-cell row");

	/*
	 * Validate the values to be within range.
	 */
	if ((size < 4) || (size >= MAXSIZE) ||
		(ulIndex < 0) || (ulIndex >= tableUsed) ||
		(urIndex < 0) || (urIndex >= tableUsed) ||
		(llIndex < 0) || (llIndex >= tableUsed) ||
		(lrIndex < 0) || (lrIndex >= tableUsed))
	{
		freeResources("Invalid macro-cell row value");
	}

	/*
	 * Get the referenced quadrant macro-cells either from the
	 * indicated indexes in the table or from an empty macro cell.
	 * The indexes are zero for empty macro-cells.
	 */
	halfSize = size - 1;

	ul = ulIndex ? table[ulIndex] : qcGetEmpty(halfSize);
	ur = urIndex ? table[urIndex] : qcGetEmpty(halfSize);
	ll = llIndex ? table[llIndex] : qcGetEmpty(halfSize);
	lr = lrIndex ? table[lrIndex] : qcGetEmpty(halfSize);

	/*
	 * All the quadrants must have a size one less than our new one.
	 */
	if ((ul->size != halfSize) || (ur->size != halfSize) ||
		(ll->size != halfSize) || (lr->size != halfSize))
	{
		freeResources("Mismatched macro-cell size");
	}

	/*
	 * Build the result macro-cell from the quadrants.
	 */
	return qcFind(ul->index, ur->index, ll->index, lr->index);
}


/*
 * Read the next line from the file and remove trailing newlines,
 * carriage returns, and spaces.  Returns FALSE on end of file.
 */
static BOOL
getLine(FILE * fp, char * buf, int bufLen)
{
	if (fgets(buf, bufLen, fp) == 0)
	{
		if (ferror(fp))
			freeResources("Error reading file");

		return FALSE;
	}

	int len = strlen(buf);

	while (len > 0)
	{
		int ch = buf[len - 1];

		if ((ch != '\n') && (ch != '\r') && (ch != ' '))
			break;
		
		buf[--len] = '\0';
	}

	return TRUE;
}


/*
 * Append a new quadcell to our table, growing the table if necessary.
 */
static void
addCell(QuadCell * cp)
{
	QuadCell ** newTable;
	int newAlloc;

	/*
	 * If the table isn't full then just store the cell.
	 */
	if (tableUsed < tableAlloc)
	{
		table[tableUsed++] = cp;

		return;
	}

	/*
	 * We need to allocate or reallocate the table.
	 * Either use its initial size or double it.
	 */
	newAlloc = (tableAlloc ? (tableAlloc * 2) : INITIAL_TABLE_SIZE);

	newTable = realloc(table, newAlloc * (sizeof(QuadCell *)));

	if (newTable == 0)
		freeResources("Cannot allocate cell table");

	table = newTable;
	tableAlloc = newAlloc;

	table[tableUsed++] = cp;
}


/*
 * Clean up resources.  An error is then generated if the
 * error text is not 0.
 */
static void
freeResources(const char * errorText)
{
	if (table)
		free(table);

	table = 0;
	tableAlloc = 0;
	tableUsed = 0;

	if (errorText)
		Error(errorText);
}


