/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Shifting operations that are calculated from the quadcell tree.
 */
#include "quad.h"
#include "quadpriv.h"


//
// Local routines.
//
static QuadCell * qcRotateRightFunc(QuadCell * cp);
static QuadCell * qcRotateLeftFunc(QuadCell * cp);
static QuadCell * qcVerticalFlipFunc(QuadCell * cp);
static QuadCell * qcHorizontalFlipFunc(QuadCell * cp);

static QuadCell * qcLeftShiftFunc(QuadCell * cp, int shiftSize);
static QuadCell * qcRightShiftFunc(QuadCell * cp, int shiftSize);
static QuadCell * qcUpShiftFunc(QuadCell * cp, int shiftSize);
static QuadCell * qcDownShiftFunc(QuadCell * cp, int shiftSize);

static QuadCell * qcShiftPairLeft(QuadCell * cpLeft, QuadCell * cpRight,
	int shiftSize);

static QuadCell * qcShiftPairRight(QuadCell * cpRight, QuadCell * cpLeft,
	int shiftSize);

static QuadCell * qcShiftPairUp(QuadCell * cpUpper, QuadCell * cpLower,
	int shiftSize);

static QuadCell * qcShiftPairDown(QuadCell * cpLower, QuadCell * cpUpper,
	int shiftSize);


/*
 * Rotate the cells in a quadcell clockwise around the specified location.
 * The isOddSymmetry flag indicates that the rotation uses odd symmetry
 * instead of even symmetry.  An odd symmetry leaves the specified location
 * in place, whereas an even symmetry replaces that location with an
 * adjacent one.  This temporarily uses saved results.
 */
QuadCell *
qcRotateRight(QuadCell * cp, NUMBER row, NUMBER col, BOOL isOddSymmetry)
{
	//
	// Check for cases which do nothing.
	//
	if (isCellOff(cp) ||
		(isCellPure(cp) && numIsZero(row) && numIsZero(col)))
	{
		return cp;
	}

	//
	// Shift the quadcell to bring the specified location to the origin.
	//
	cp = qcLeftOrRightShift(cp, numNeg(col));
	cp = qcUpOrDownShift(cp, numNeg(row));

	//
	// Now do the rotation around the origin.
	//
	qcSetResultType(RESULT_NONE);

	cp = qcRotateRightFunc(cp);

	//
	// Finally shift back to the specified location.
	// But if the symmetry is odd then adjust the size of one of
	// the shifts by one cell so that the specified location
	// remains in place.
	//
	if (isOddSymmetry)
		numAddrInc(&row);

	cp = qcLeftOrRightShift(cp, col);
	cp = qcUpOrDownShift(cp, row);

	return cp;
}


/*
 * Recursive routine to rotate the cells in a quadrant clockwise
 * around the origin.
 */
static QuadCell *
qcRotateRightFunc(QuadCell * cp)
{
	//
	// An pure quadcell doesn't change by a rotate.
	// This also handles the size 1 quadcell.
	//
	if (isCellPure(cp))
		return cp;

	//
	// If a result was saved then return that.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There was no result yet so we have to calculate it.
	// Call ourself recursively for our quadrants while rotating
	// the quadrants clockwise.
	//
	QuadCell * ul = qcRotateRightFunc(indexToQuadCell(cp->ll));
	QuadCell * ur = qcRotateRightFunc(indexToQuadCell(cp->ul));
	QuadCell * ll = qcRotateRightFunc(indexToQuadCell(cp->lr));
	QuadCell * lr = qcRotateRightFunc(indexToQuadCell(cp->ur));

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Rotate the cells in a quadcell counter-clockwise around the specified
 * location.  The isOddSymmetry flag indicates that the flipping uses odd
 * symmetry instead of even symmetry.  An odd symmetry leaves the specified
 * location in place, whereas an even symmetry replaces that location with
 * an adjacent one.  This temporarily uses saved results.
 */
QuadCell *
qcRotateLeft(QuadCell * cp, NUMBER row, NUMBER col, BOOL isOddSymmetry)
{
	//
	// Check for cases which do nothing.
	//
	if (isCellOff(cp) ||
		(isCellPure(cp) && numIsZero(row) && numIsZero(col)))
	{
		return cp;
	}

	//
	// Shift the quadcell to bring the specified location to the origin.
	//
	cp = qcLeftOrRightShift(cp, numNeg(col));
	cp = qcUpOrDownShift(cp, numNeg(row));

	//
	// Now do the rotation around the origin.
	//
	qcSetResultType(RESULT_NONE);

	cp = qcRotateLeftFunc(cp);

	//
	// Finally shift back to the specified location.
	// But if the symmetry is odd then adjust the size of one of
	// the shifts by one cell so that the specified location
	// remains in place.
	//
	if (isOddSymmetry)
		numAddrInc(&col);

	cp = qcLeftOrRightShift(cp, col);
	cp = qcUpOrDownShift(cp, row);

	return cp;
}


/*
 * Recursive routine to rotate the cells in a quadrant counter-clockwise
 * around the origin.
 */
static QuadCell *
qcRotateLeftFunc(QuadCell * cp)
{
	//
	// An pure quadcell doesn't change by a rotate.
	// This also handles the size 1 quadcell.
	//
	if (isCellPure(cp))
		return cp;

	//
	// If a result was saved then return that.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There was no result yet so we have to calculate it.
	// Call ourself recursively for our quadrants while rotating
	// the quadrants counter-clockwise.
	//
	QuadCell * ul = qcRotateLeftFunc(indexToQuadCell(cp->ur));
	QuadCell * ur = qcRotateLeftFunc(indexToQuadCell(cp->lr));
	QuadCell * ll = qcRotateLeftFunc(indexToQuadCell(cp->ul));
	QuadCell * lr = qcRotateLeftFunc(indexToQuadCell(cp->ll));

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Flip the cells in a quadcell vertically around the specified column.
 * The isOddSymmetry flag indicates that the flipping uses odd symmetry
 * instead of even symmetry.  An odd symmetry leaves the specified column
 * in place, whereas an even symmetry exchanges that column with the
 * adjacent one.  This temporarily uses saved results.
 */
QuadCell *
qcVerticalFlip(QuadCell * cp, NUMBER col, BOOL isOddSymmetry)
{
	//
	// Check for cases which do nothing.
	//
	if (isCellOff(cp) || (isCellPure(cp) && numIsZero(col)))
		return cp;

	//
	// Shift the quadcell so that the specified column moves to the origin.
	//
	cp = qcLeftOrRightShift(cp, numNeg(col));

	//
	// Perform the vertical flip function around the origin.
	//
	qcSetResultType(RESULT_NONE);

	cp = qcVerticalFlipFunc(cp);

	//
	// Finally shift back to the specified column.
	// But if the symmetry is odd then adjust the size of the shift
	// by one cell so that the specified column remains in place.
	//
	if (isOddSymmetry)
		numAddrInc(&col);

	return qcLeftOrRightShift(cp, col);
}


/*
 * Recursive routine to flip the cells in a quadcell along its
 * vertical axis.  This uses even symmetry flipping.
 */
static QuadCell *
qcVerticalFlipFunc(QuadCell * cp)
{
	//
	// A pure quadcell doesn't change by a flip.
	// This also handles the size 1 quadcell.
	//
	if (isCellPure(cp))
		return cp;

	//
	// If a result was saved then return that.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There was no result yet so we have to calculate it.
	// Call ourself recursively for our quadrants while flipping
	// the results of the left and right quadrants.
	//
	QuadCell * ul = qcVerticalFlipFunc(indexToQuadCell(cp->ur));
	QuadCell * ur = qcVerticalFlipFunc(indexToQuadCell(cp->ul));
	QuadCell * ll = qcVerticalFlipFunc(indexToQuadCell(cp->lr));
	QuadCell * lr = qcVerticalFlipFunc(indexToQuadCell(cp->ll));

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Flip the cells in a quadcell horizontally around the specified row.
 * The isOddSymmetry flag indicates that the flipping uses odd symmetry
 * instead of even symmetry.  An odd symmetry leaves the specified row
 * in place, whereas an even symmetry exchanges that row with the
 * adjacent one.  This temporarily uses saved results.
 */
QuadCell *
qcHorizontalFlip(QuadCell * cp, NUMBER row, BOOL isOddSymmetry)
{
	//
	// Check for cases which do nothing.
	//
	if (isCellOff(cp) || (isCellPure(cp) && numIsZero(row)))
		return cp;

	//
	// Shift the quadcell so that the specified row moves to the origin.
	//
	cp = qcUpOrDownShift(cp, numNeg(row));

	//
	// Perform the horizontal flip function around the origin.
	//
	qcSetResultType(RESULT_NONE);

	cp = qcHorizontalFlipFunc(cp);

	//
	// Finally shift back to the specified row.
	// But if the symmetry is odd then adjust the size of the shift
	// by one cell so that the specified row remains in place.
	//
	if (isOddSymmetry)
		numAddrInc(&row);

	return qcUpOrDownShift(cp, row);
}


/*
 * Recursive routine to flip the cells in a quadcell along its
 * horizontal axis.  This uses even symmetry flipping.
 */
static QuadCell *
qcHorizontalFlipFunc(QuadCell * cp)
{
	//
	// An pure quadcell doesn't change by a flip.
	// This also handles the size 1 quadcell.
	//
	if (isCellPure(cp))
		return cp;

	//
	// If a result was saved then return that.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There was no result yet so we have to calculate it.
	// Call ourself recursively for our quadrants while flipping
	// the results of the top and bottom quadrants.
	//
	QuadCell * ul = qcHorizontalFlipFunc(indexToQuadCell(cp->ll));
	QuadCell * ur = qcHorizontalFlipFunc(indexToQuadCell(cp->lr));
	QuadCell * ll = qcHorizontalFlipFunc(indexToQuadCell(cp->ul));
	QuadCell * lr = qcHorizontalFlipFunc(indexToQuadCell(cp->ur));

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ul->index, ur->index, ll->index, lr->index);

	cp->result = result->index;

	return result;
}


/*
 * Center the cells in a quadcell so that the minimum and maximum coordinates
 * of the rows and columns are as close to each other (in opposite signs) as
 * they can be.  This also trims the result.  Warning: The coordinates of the
 * cells in the object are changed by a nonobvious amount by this operation,
 * so it is unlikely that the result can be usefully combined with other
 * operations on the original quadcell.
 */
QuadCell *
qcCenter(QuadCell * cp)
{
	if (isCellPure(cp))
		return cp;

	NUMBER minRow = qcGetMinRow(cp);
	NUMBER maxRow = qcGetMaxRow(cp);
	NUMBER minCol = qcGetMinCol(cp);
	NUMBER maxCol = qcGetMaxCol(cp);

	NUMBER averageRow = numAverage(minRow, maxRow);
	NUMBER averageCol = numAverage(minCol, maxCol);

	return qcShift(cp, numNeg(averageCol), numNeg(averageRow));
}


/*
 * Convenience routine to shift the cells in a quadcell both left or right
 * and up or down.  As a further convenience the result is trimmed.
 */
QuadCell *
qcShift(QuadCell * cp, NUMBER leftOrRightShift, NUMBER upOrDownShift)
{
	cp = qcLeftOrRightShift(cp, leftOrRightShift);
	cp = qcUpOrDownShift(cp, upOrDownShift);
	cp = qcTrim(cp);

	return cp;
}


/*
 * Convenience routine to shift the cells of a quadcell either to the
 * right or to the left according to the sign of the shift value.
 * A positive value moves cells to the right (increasing columns) while
 * a negative value moves cells to the left (decreasing columns).
 */
QuadCell *
qcLeftOrRightShift(QuadCell * cp, NUMBER shift)
{
	int sign = numSign(shift);

	if (isCellOff(cp) || (sign == 0))
		return cp;

	if (sign > 0)
		cp = qcRightShift(cp, shift);

	if (sign < 0)
		cp = qcLeftShift(cp, numNeg(shift));

	return cp;
}


/*
 * Routine to perform a left shift operation of a quadcell by the specified
 * number of cells (which must be non-negative).  This temporarily uses
 * saved results and quadpair structures to avoid repeated recursions.
 * A number of smaller shifts are needed if the shift value is not an
 * exact power of two.
 */
QuadCell *
qcLeftShift(QuadCell * cp, NUMBER shift)
{
	//
	// If the cell is OFF or the shift is zero then nothing
	// is changed.
	//
	verifyAlways(numSign(shift) >= 0);

	if (isCellOff(cp) || numIsZero(shift))
		return cp;

	//
	// Grow the quadcell enough so that any ON cells on the left
	// edge of the input quadcell will not be lost.
	//
	NUMBER minCoord = numNeg(numTwoPower(cp->size - 1));
	NUMBER leftMostCol = numSub(minCoord, shift);

	cp = qcGrowToContain(cp, numZero(), leftMostCol);

	//
	// Perform a loop doing shifts of a power of two as long as
	// the shift value is nonzero.
	//
	while (numIsNonZero(shift))
	{
		//
		// Get the highest power of two remaining in the shift.
		//
		int shiftSize = numTwoLog(shift);
		NUMBER highBitValue = numTwoPower(shiftSize);

		//
		// Prepare the quadpair hash table for this shift.
		//
		qcSetQuadPairOperation(OP_LEFTSHIFT, highBitValue);

		//
		// Perform the left shift operation.
		//
		qcSetResultType(RESULT_NONE);

		cp = qcLeftShiftFunc(cp, shiftSize);

		//
		// Subtract off the high bit from the required shift.
		// Eventually this will clear its value and end the loop.
		//
		numAddrSub(&shift, highBitValue);
	}

	//
	// Finally trim the quadcell if possible.
	//
	return qcTrim(cp);
}


/*
 * Recursive routine to perform a left shift operation by the specified
 * power of two cells.  Any ON cells shifted out of the left edge of the
 * quadcell are lost.
 */
static QuadCell *
qcLeftShiftFunc(QuadCell * cp, int shiftSize)
{
	if (isCellOff(cp))
		return cp;

	//
	// See if we have a result saved.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There isn't any result saved yet so we have to calculate it.
	// If the shift is larger then the quadcell size then the result
	// will be an empty quadcell.
	//
	if (cp->size <= shiftSize)
		return qcGetEmpty(cp->size);

	//
	// If the shift is by a quadrant size then we can get the result now.
	// The right two quadrants move over to the left two quadrants while
	// the right two quadrants are replaced by empty quadcells.
	// Save the result for a future use.
	//
	if (cp->size - 1 == shiftSize)
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * result = qcFind(cp->ur, emptyIndex, cp->lr, emptyIndex);

		cp->result = result->index;

		return result;
	}

	//
	// We have to do the shift for a larger quadcell and so need recursion.
	// Do this by examining quadpairs made of the horizontally adjacent
	// quadcells contained within the upper and lower quadrants along
	// with an empty quadcell for the OFF cells coming in on the right.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	QuadCell * empty = qcGetEmpty(cp->size - 1);

	QuadCell * ulNew = qcShiftPairLeft(ul, ur, shiftSize);
	QuadCell * urNew = qcShiftPairLeft(ur, empty, shiftSize);
	QuadCell * llNew = qcShiftPairLeft(ll, lr, shiftSize);
	QuadCell * lrNew = qcShiftPairLeft(lr, empty, shiftSize);

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ulNew->index, urNew->index,
		llNew->index, lrNew->index);

	cp->result = result->index;

	return result;
}


/*
 * Shift a pair of horizontally adjacent quadcells left by the indicated
 * power of two cells.  Returns a quadcell which replaces the left one
 * with the result.  Nothing is computed for the right quadcell (it is
 * only used as input for the right edge of the left quadcell).  Any ON
 * cells shifted out of the left edge of the left quadcell are lost.
 */
static QuadCell *
qcShiftPairLeft(QuadCell * cpLeft, QuadCell * cpRight, int shiftSize)
{
	verifyAlways(cpLeft->size == cpRight->size);

	//
	// If both quadcells are OFF then so is the result.
	//
	if (isCellOff(cpLeft) && isCellOff(cpRight))
		return cpLeft;

	//
	// If both quadcells are the same pure state then so is the result.
	//
	if (isCellPure(cpLeft) && (cpRight->state == cpLeft->state))
		return cpLeft;

	//
	// If the right quadcell doesn't have OFF right quadrants
	// then replace it with one which does since none of the cells
	// in the right quadrants are examined.  This lets us share
	// more quadpair results.
	//
	QuadCell * ur = indexToQuadCell(cpRight->ur);
	QuadCell * lr = indexToQuadCell(cpRight->lr);

	if (isCellNotOff(ur) || isCellNotOff(lr))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cpLeft->size - 1);

		cpRight = qcFind(cpRight->ul, emptyIndex, cpRight->ll, emptyIndex);
	}

	//
	// See if there is a quadpair saved for this combination.
	// If so then return it.  This is an unsymmetrical lookup.
	//
	QuadPair * pair = qcFindQuadPair(cpLeft->index, cpRight->index, FALSE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// Nothing was saved so we have to calculate it now.
	// Combine the left and right halfs of the input quadcells
	// to make a new quadcell which spans them.
	//
	QuadCell * cpSpan = qcFind(cpLeft->ur, cpRight->ul,
		cpLeft->lr, cpRight->ll);

	//
	// Now do recursive calls on two smaller quadcells.
	// These are the input left quadcell and the quadcell which
	// spans the input quadcells.
	//
	QuadCell * leftResult = qcLeftShiftFunc(cpLeft, shiftSize);
	QuadCell * spanResult = qcLeftShiftFunc(cpSpan, shiftSize);

	//
	// Build the quadpair result for our quadpair from the left
	// and right quadrants of the smaller results.
	//
	QuadCell * result = qcFind(leftResult->ul, spanResult->ul,
		leftResult->ll, spanResult->ll);

	//
	// Save the result for this quadpair and return it.
	//
	pair->result = result->index;

	return result;
}


/*
 * Routine to perform a right shift operation of a quadcell by the specified
 * number of cells (which must be non-negative).  This temporarily uses
 * saved results and quadpair structures to avoid repeated recursions.
 * A number of smaller shifts are needed if the shift value is not an
 * exact power of two.
 */
QuadCell *
qcRightShift(QuadCell * cp, NUMBER shift)
{
	//
	// If the cell is OFF or the shift is zero then nothing
	// is changed.
	//
	verifyAlways(numSign(shift) >= 0);

	if (isCellOff(cp) || numIsZero(shift))
		return cp;

	//
	// Grow the quadcell enough so that any ON cells on the right
	// edge of the input quadcell will not be lost.
	//
	NUMBER maxCoord = numDec(numTwoPower(cp->size - 1));
	NUMBER rightMostCol = numAdd(maxCoord, shift);

	cp = qcGrowToContain(cp, numZero(), rightMostCol);

	//
	// Perform a loop doing shifts of a power of two as long as
	// the shift value is nonzero.
	//
	while (numIsNonZero(shift))
	{
		//
		// Get the highest power of two remaining in the shift.
		//
		int shiftSize = numTwoLog(shift);
		NUMBER highBitValue = numTwoPower(shiftSize);

		//
		// Prepare the quadpair hash table for this shift.
		//
		qcSetQuadPairOperation(OP_RIGHTSHIFT, highBitValue);

		//
		// Perform the right shift operation.
		//
		qcSetResultType(RESULT_NONE);

		cp = qcRightShiftFunc(cp, shiftSize);

		//
		// Subtract off the high bit from the required shift.
		// Eventually this will clear its value and end the loop.
		//
		numAddrSub(&shift, highBitValue);
	}

	//
	// Finally trim the quadcell if possible.
	//
	return qcTrim(cp);
}


/*
 * Recursive routine to perform a right shift operation by the specified
 * power of two cells.  Any ON cells shifted out of the right edge of the
 * quadcell are lost.
 */
static QuadCell *
qcRightShiftFunc(QuadCell * cp, int shiftSize)
{
	if (isCellOff(cp))
		return cp;

	//
	// See if we have a result saved.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There isn't any result saved yet so we have to calculate it.
	// If the shift is larger then the quadcell size then the result
	// will be an empty quadcell.
	//
	if (cp->size <= shiftSize)
		return qcGetEmpty(cp->size);

	//
	// If the shift is by a quadrant size then we can get the result now.
	// The left two quadrants move over to the right two quadrants while
	// the left two quadrants are replaced by empty quadcells.
	// Save the result for a future use.
	//
	if (cp->size - 1 == shiftSize)
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * result = qcFind(emptyIndex, cp->ul, emptyIndex, cp->ll);

		cp->result = result->index;

		return result;
	}

	//
	// We have to do the shift for a larger quadcell and so need recursion.
	// Do this by examining quadpairs made of the horizontally adjacent
	// quadcells contained within the upper and lower quadrants along
	// with an empty quadcell for the OFF cells coming in on the left.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	QuadCell * empty = qcGetEmpty(cp->size - 1);

	QuadCell * urNew = qcShiftPairRight(ur, ul, shiftSize);
	QuadCell * ulNew = qcShiftPairRight(ul, empty, shiftSize);
	QuadCell * lrNew = qcShiftPairRight(lr, ll, shiftSize);
	QuadCell * llNew = qcShiftPairRight(ll, empty, shiftSize);

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ulNew->index, urNew->index,
		llNew->index, lrNew->index);

	cp->result = result->index;

	return result;
}


/*
 * Shift a pair of horizontally adjacent quadcells right by the indicated
 * power of two cells.  Returns a quadcell which replaces the right one
 * with the result.  Nothing is computed for the left quadcell (it is
 * only used as input for the left edge of the right quadcell).  Any ON
 * cells shifted out of the right edge of the right quadcell are lost.
 */
static QuadCell *
qcShiftPairRight(QuadCell * cpRight, QuadCell * cpLeft, int shiftSize)
{
	verifyAlways(cpLeft->size == cpRight->size);

	//
	// If both quadcells are OFF then so is the result.
	//
	if (isCellOff(cpLeft) && isCellOff(cpRight))
		return cpLeft;

	//
	// If both quadcells are the same pure state then so is the result.
	//
	if (isCellPure(cpLeft) && (cpRight->state == cpLeft->state))
		return cpLeft;

	//
	// If the left quadcell doesn't have OFF left quadrants
	// then replace it with one which does since none of the cells
	// in the left quadrants are examined.  This lets us share
	// more quadpair results.
	//
	QuadCell * ul = indexToQuadCell(cpLeft->ul);
	QuadCell * ll = indexToQuadCell(cpLeft->ll);

	if (isCellNotOff(ul) || isCellNotOff(ll))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cpLeft->size - 1);

		cpLeft = qcFind(emptyIndex, cpLeft->ur, emptyIndex, cpLeft->lr);
	}

	//
	// See if there is a quadpair saved for this combination.
	// If so then return it.  This is an unsymmetrical lookup.
	//
	QuadPair * pair = qcFindQuadPair(cpRight->index, cpLeft->index, FALSE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// Nothing was saved so we have to calculate it now.
	// Combine the left and right halfs of the input quadcells
	// to make a new quadcell which spans them.
	//
	QuadCell * cpSpan = qcFind(cpLeft->ur, cpRight->ul,
		cpLeft->lr, cpRight->ll);

	//
	// Now do recursive calls on two smaller quadcells.
	// These are the input right quadcell and the quadcell which
	// spans the input quadcells.
	//
	QuadCell * rightResult = qcRightShiftFunc(cpRight, shiftSize);
	QuadCell * spanResult = qcRightShiftFunc(cpSpan, shiftSize);

	//
	// Build the quadpair result for our quadpair from the left
	// and right quadrants of the smaller results.
	//
	QuadCell * result = qcFind(spanResult->ur, rightResult->ur,
		spanResult->lr, rightResult->lr);

	//
	// Save the result for this quadpair and return it.
	//
	pair->result = result->index;

	return result;
}


/*
 * Convenience routine to shift the cells of a quadcell either upwards
 * or downwards according to the sign of the shift value.
 * A positive value moves cells upwards (increasing rows) while a
 * negative value moves cells downwards (decreasing rows).
 */
QuadCell *
qcUpOrDownShift(QuadCell * cp, NUMBER shift)
{
	int sign = numSign(shift);

	if (isCellOff(cp) || (sign == 0))
		return cp;

	if (sign > 0)
		cp = qcUpShift(cp, shift);

	if (sign < 0)
		cp = qcDownShift(cp, numNeg(shift));

	return cp;
}


/*
 * Routine to perform an upwards shift operation of a quadcell by the
 * specified number of cells (which must be non-negative).  This
 * temporarily uses saved results and quadpair structures to avoid
 * repeated recursions.  A number of smaller shifts are needed if the
 * shift value is not an exact power of two.
 */
QuadCell *
qcUpShift(QuadCell * cp, NUMBER shift)
{
	//
	// If the cell is OFF or the shift is zero then nothing
	// is changed.
	//
	verifyAlways(numSign(shift) >= 0);

	if (isCellOff(cp) || numIsZero(shift))
		return cp;

	//
	// Grow the quadcell enough so that any ON cells on the top
	// edge of the input quadcell will not be lost.
	//
	NUMBER maxCoord = numDec(numTwoPower(cp->size - 1));
	NUMBER topMostRow = numAdd(maxCoord, shift);

	cp = qcGrowToContain(cp, topMostRow, numZero());

	//
	// Perform a loop doing shifts of a power of two as long as
	// the shift value is nonzero.
	//
	while (numIsNonZero(shift))
	{
		//
		// Get the highest power of two remaining in the shift.
		//
		int shiftSize = numTwoLog(shift);
		NUMBER highBitValue = numTwoPower(shiftSize);

		//
		// Prepare the quadpair hash table for this shift.
		//
		qcSetQuadPairOperation(OP_UPSHIFT, highBitValue);

		qcSetResultType(RESULT_NONE);

		//
		// Perform the up shift operation.
		//
		cp = qcUpShiftFunc(cp, shiftSize);

		//
		// Subtract off the high bit from the required shift.
		// Eventually this will clear its value and end the loop.
		//
		numAddrSub(&shift, highBitValue);
	}

	//
	// Finally trim the quadcell if possible.
	//
	return qcTrim(cp);
}


/*
 * Recursive routine to perform an upwards shift operation by the specified
 * power of two cells.  Any ON cells shifted out of the top edge of the
 * quadcell are lost.
 */
static QuadCell *
qcUpShiftFunc(QuadCell * cp, int shiftSize)
{
	if (isCellOff(cp))
		return cp;

	//
	// See if we have a result saved.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There isn't any result saved yet so we have to calculate it.
	// If the shift is larger then the quadcell size then the result
	// will be an empty quadcell.
	//
	if (cp->size <= shiftSize)
		return qcGetEmpty(cp->size);

	//
	// If the shift is by a quadrant size then we can get the result now.
	// The bottom two quadrants move to the top two quadrants while
	// the bottom two quadrants are replaced by empty quadcells.
	// Save the result for a future use.
	//
	if (cp->size - 1 == shiftSize)
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * result = qcFind(cp->ll, cp->lr, emptyIndex, emptyIndex);

		cp->result = result->index;

		return result;
	}

	//
	// We have to do the shift for a larger quadcell and so need recursion.
	// Do this by examining quadpairs made of the vertically adjacent
	// quadcells contained within the leftmost and rightmost quadrants
	// along with an empty quadcell for the OFF cells coming in on the bottom.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	QuadCell * empty = qcGetEmpty(cp->size - 1);

	QuadCell * ulNew = qcShiftPairUp(ul, ll, shiftSize);
	QuadCell * llNew = qcShiftPairUp(ll, empty, shiftSize);
	QuadCell * urNew = qcShiftPairUp(ur, lr, shiftSize);
	QuadCell * lrNew = qcShiftPairUp(lr, empty, shiftSize);

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ulNew->index, urNew->index,
		llNew->index, lrNew->index);

	cp->result = result->index;

	return result;
}


/*
 * Shift a pair of vertically adjacent quadcells up by the indicated
 * power of two cells.  Returns a quadcell which replaces the upper one
 * with the result.  Nothing is computed for the lower quadcell (it is
 * only used as input for the bottom edge of the upper quadcell).  Any ON
 * cells shifted out of the top edge of the top quadcell are lost.
 */
static QuadCell *
qcShiftPairUp(QuadCell * cpUpper, QuadCell * cpLower, int shiftSize)
{
	verifyAlways(cpLower->size == cpUpper->size);

	//
	// If both quadcells are OFF then so is the result.
	//
	if (isCellOff(cpLower) && isCellOff(cpUpper))
		return cpLower;

	//
	// If both quadcells are the same pure state then so is the result.
	//
	if (isCellPure(cpLower) && (cpUpper->state == cpLower->state))
		return cpLower;

	//
	// If the bottom quadcell doesn't have OFF bottommost quadrants
	// then replace it with one which does since none of the cells
	// in the lower quadrants are examined.  This lets us share
	// more quadpair results.
	//
	QuadCell * ll = indexToQuadCell(cpLower->ll);
	QuadCell * lr = indexToQuadCell(cpLower->lr);

	if (isCellNotOff(ll) || isCellNotOff(lr))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cpLower->size - 1);

		cpLower = qcFind(cpLower->ul, cpLower->ur, emptyIndex, emptyIndex);
	}

	//
	// See if there is a quadpair saved for this combination.
	// If so then return it.  This is an unsymmetrical lookup.
	//
	QuadPair * pair = qcFindQuadPair(cpUpper->index, cpLower->index, FALSE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// Nothing was saved so we have to calculate it now.
	// Combine the top and bottom halfs of the input quadcells
	// to make a new quadcell which spans them.
	//
	QuadCell * cpSpan = qcFind(cpUpper->ll, cpUpper->lr,
		cpLower->ul, cpLower->ur);

	//
	// Now do recursive calls on two smaller quadcells.
	// These are the input top quadcell and the quadcell which
	// spans the input quadcells.
	//
	QuadCell * upperResult = qcUpShiftFunc(cpUpper, shiftSize);
	QuadCell * spanResult = qcUpShiftFunc(cpSpan, shiftSize);

	//
	// Build the quadpair result for our quadpair from the top
	// and bottom quadrants of the smaller results.
	//
	QuadCell * result = qcFind(upperResult->ul, upperResult->ur,
		spanResult->ul, spanResult->ur);

	//
	// Save the result for this quadpair and return it.
	//
	pair->result = result->index;

	return result;
}


/*
 * Routine to perform a downwards shift operation of a quadcell by the
 * specified number of cells (which must be non-negative).  This
 * temporarily uses saved results and quadpair structures to avoid
 * repeated recursions.  A number of smaller shifts are needed if the
 * shift value is not an exact power of two.
 */
QuadCell *
qcDownShift(QuadCell * cp, NUMBER shift)
{
	//
	// If the cell is OFF or the shift is zero then nothing
	// is changed.
	//
	verifyAlways(numSign(shift) >= 0);

	if (isCellOff(cp) || numIsZero(shift))
		return cp;

	//
	// Grow the quadcell enough so that any ON cells on the bottom
	// edge of the input quadcell will not be lost.
	//
	NUMBER minCoord = numNeg(numTwoPower(cp->size - 1));
	NUMBER bottomMostRow = numSub(minCoord, shift);

	cp = qcGrowToContain(cp, bottomMostRow, numZero());

	//
	// Perform a loop doing shifts of a power of two as long as
	// the shift value is nonzero.
	//
	while (numIsNonZero(shift))
	{
		//
		// Get the highest power of two remaining in the shift.
		//
		int shiftSize = numTwoLog(shift);
		NUMBER highBitValue = numTwoPower(shiftSize);

		//
		// Prepare the quadpair hash table for this shift.
		//
		qcSetQuadPairOperation(OP_DOWNSHIFT, highBitValue);

		qcSetResultType(RESULT_NONE);

		//
		// Perform the down shift operation.
		//
		cp = qcDownShiftFunc(cp, shiftSize);

		//
		// Subtract off the high bit from the required shift.
		// Eventually this will clear its value and end the loop.
		//
		numAddrSub(&shift, highBitValue);
	}

	//
	// Finally trim the quadcell if possible.
	//
	return qcTrim(cp);
}


/*
 * Recursive routine to perform a downwards shift operation by the specified
 * power of two cells.  Any ON cells shifted out of the bottom edge of the
 * quadcell are lost.
 */
static QuadCell *
qcDownShiftFunc(QuadCell * cp, int shiftSize)
{
	if (isCellOff(cp))
		return cp;

	//
	// See if we have a result saved.
	//
	if (isNotNull(cp->result))
		return indexToQuadCell(cp->result);

	//
	// There isn't any result saved yet so we have to calculate it.
	// If the shift is larger then the quadcell size then the result
	// will be an empty quadcell.
	//
	if (cp->size <= shiftSize)
		return qcGetEmpty(cp->size);

	//
	// If the shift is by a quadrant size then we can get the result now.
	// The top two quadrants move to the bottom two quadrants while
	// the top two quadrants are replaced by empty quadcells.
	// Save the result for a future use.
	//
	if (cp->size - 1 == shiftSize)
	{
		QUADINDEX emptyIndex = offIndexFromSize(cp->size - 1);

		QuadCell * result = qcFind(emptyIndex, emptyIndex, cp->ul, cp->ur);

		cp->result = result->index;

		return result;
	}

	//
	// We have to do the shift for a larger quadcell and so need recursion.
	// Do this by examining quadpairs made of the vertically adjacent
	// quadcells contained within the leftmost and rightmost quadrants
	// along with an empty quadcell for the OFF cells coming in on the top.
	//
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	QuadCell * empty = qcGetEmpty(cp->size - 1);

	QuadCell * llNew = qcShiftPairDown(ll, ul, shiftSize);
	QuadCell * ulNew = qcShiftPairDown(ul, empty, shiftSize);
	QuadCell * lrNew = qcShiftPairDown(lr, ur, shiftSize);
	QuadCell * urNew = qcShiftPairDown(ur, empty, shiftSize);

	//
	// Now build and save the result.
	//
	QuadCell * result = qcFind(ulNew->index, urNew->index,
		llNew->index, lrNew->index);

	cp->result = result->index;

	return result;
}


/*
 * Shift a pair of vertically adjacent quadcells down by the indicated
 * power of two cells.  Returns a quadcell which replaces the lower one
 * with the result.  Nothing is computed for the upper quadcell (it is
 * only used as input for the upper edge of the lower quadcell).  Any ON
 * cells shifted out of the bottom edge of the bottom quadcell are lost.
 */
static QuadCell *
qcShiftPairDown(QuadCell * cpLower, QuadCell * cpUpper, int shiftSize)
{
	verifyAlways(cpLower->size == cpUpper->size);

	//
	// If both quadcells are OFF then so is the result.
	//
	if (isCellOff(cpLower) && isCellOff(cpUpper))
		return cpLower;

	//
	// If both quadcells are the same pure state then so is the result.
	//
	if (isCellPure(cpLower) && (cpUpper->state == cpLower->state))
		return cpLower;

	//
	// If the top quadcell doesn't have empty topmost quadrants
	// then replace it with one which does since none of the cells
	// in the upper quadrants are examined.  This lets us share
	// more quadpair results.
	//
	QuadCell * ul = indexToQuadCell(cpUpper->ul);
	QuadCell * ur = indexToQuadCell(cpUpper->ur);

	if (isCellNotOff(ul) || isCellNotOff(ur))
	{
		QUADINDEX emptyIndex = offIndexFromSize(cpLower->size - 1);

		cpUpper = qcFind(emptyIndex, emptyIndex, cpUpper->ll, cpUpper->lr);
	}

	//
	// See if there is a quadpair saved for this combination.
	// If so then return it.  This is an unsymmetrical lookup.
	//
	QuadPair * pair = qcFindQuadPair(cpLower->index, cpUpper->index, FALSE);

	if (isNotNull(pair->result))
		return indexToQuadCell(pair->result);

	//
	// Nothing was saved so we have to calculate it now.
	// Combine the top and bottom halfs of the input quadcells
	// to make a new quadcell which spans them.
	//
	QuadCell * cpSpan = qcFind(cpUpper->ll, cpUpper->lr,
		cpLower->ul, cpLower->ur);

	//
	// Now do recursive calls on two smaller quadcells.
	// These are the input bottom quadcell and the quadcell which
	// spans the input quadcells.
	//
	QuadCell * lowerResult = qcDownShiftFunc(cpLower, shiftSize);
	QuadCell * spanResult = qcDownShiftFunc(cpSpan, shiftSize);

	//
	// Build the quadpair result for our quadpair from the top
	// and bottom quadrants of the smaller results.
	//
	QuadCell * result = qcFind(spanResult->ll, spanResult->lr,
		lowerResult->ll, lowerResult->lr);

	//
	// Save the result for this quadpair and return it.
	//
	pair->result = result->index;

	return result;
}

