/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routine to write a file in macro-cell format.
 * Example pattern (a glider):
 * [M1] (hlife 0.96)
 * $$$$$$$.*$
 * *$***$
 * 4 0 1 0 2
 */
#include <stdio.h>
#include <stdlib.h>

#include "quad.h"
#include "quadpriv.h"


/*
 * The next id to to be assigned to quadcells as they are found.
 * Note: This id is NOT a reference to numbers allocated by the
 * numeric routines.
 */
static NUMINDEX nextIndex;


/*
 * Local routines.
 */
static void writeFunc(FILE * fp, QuadCell * cp);
static void writeBottomLevel(FILE * fp, QuadCell *cp);
static void writeUpperLevel(FILE * fp, QuadCell * cp);

static QuadCell * getQuadrant(QuadCell * cp, int row, int col);


/*
 * Write the specified quadcell structure into the specified file.
 * It will be in macro-cell format as demonstrated above.
 *
 * Warning: This routine temporarily uses the numIndex element of
 * the quadcell structures to label the quadcells with increasing ids.
 * These ids have NOTHING to do with the numeric routines.
 */
void
qcWriteStructure(FILE * fp, QuadCell * cp)
{
	cp = qcTrim(cp);
	cp = qcGrowToSize(cp, SIZE8);

	nextIndex = 0;

	fprintf(fp, "[M1] hashlife\n");

	writeFunc(fp, cp);

	qcClearNumbers(cp);
	numFreeIndexes();
}


/*
 * Recursive routine to write a quadcell structure to the file.
 * This only does something if the quadcell is non-empty and hasn't
 * been written yet (i.e., its numIndex field is zero).
 */
static void
writeFunc(FILE * fp, QuadCell * cp)
{
	/*
	 * If the cell is empty or has been written then do nothing.
	 */
	if (isCellEmpty(cp) || isNumberIndex(cp->numIndex))
		return;

	/*
	 * Write the cell depending on whether it is the bottom level or not.
	 */
	if (cp->size == SIZE8)
		writeBottomLevel(fp, cp);
	else
		writeUpperLevel(fp, cp);

	/*
	 * Assign the next index for this cell.
	 * After this, this cell won't be written again.
	 */
	cp->numIndex = ++nextIndex;
}


/*
 * Write the "bottom" level data for an 8 by 8 size 3 quadcell.
 * This is a pictorial format of the 64 cells in a square where '.'
 * is an OFF cell, '*' is an ON cell, '$' is the end of a row,
 * and cells which aren't explicitly specified are OFF.
 */
static void
writeBottomLevel(FILE * fp, QuadCell *cp)
{
	int row;
	int col;
	int lastRow;
	char data[8][8];

	/*
	 * Collect the cell data from the smaller quadcells.
	 */
	for (row = 0; row < 8; row++)
	{
		for (col = 0; col < 8; col++)
		{
			QuadCell * temp = getQuadrant(cp, row, col);

			temp = getQuadrant(temp, row, col);
			temp = getQuadrant(temp, row, col);

			data[row][col] = isCellNotEmpty(temp);
		}
	}

	/*
	 * Write out the line as much as is necessary,
	 * trimming any trailing empty columns and rows.
	 */
	lastRow = 0;

	for (row = 0; row < 8; row++)
	{
		for (col = 0; col < 8; col++)
		{
			if (data[row][col])
				lastRow = row;
		}
	}

	for (row = 0; row <= lastRow; row++)
	{
		int lastCol = -1;

		for (col = 0; col < 8; col++)
		{
			if (data[row][col])
				lastCol = col;
		}

		for (col = 0; col <= lastCol; col++)
		{
			fputc((data[row][col] ? '*' : '.'), fp);
		}

		fputc('$', fp);
	}

	fputc('\n', fp);
}


/*
 * Helper routine to select the appropriate quadrant of a small-sized
 * quadcell given row and column offsets from its upper left corner.
 * The offsets are reduced modulo the size of the cell.
 */
static QuadCell *
getQuadrant(QuadCell * cp, int row, int col)
{
	int full = (1 << cp->size);
	int half = full / 2;

	row %= full;
	col %= full;

	if ((row < half) && (col < half))
		return indexToQuadCell(cp->ul);

	if ((row < half) && (col >= half))
		return indexToQuadCell(cp->ur);

	if ((row >= half) && (col < half))
		return indexToQuadCell(cp->ll);

	return indexToQuadCell(cp->lr);
}


/*
 * Write out an upper level non-empty quadcell which has not yet
 * been written.  This is of the format:
 *	size ulIndex urIndex llIndex lrIndex
 */
static void
writeUpperLevel(FILE * fp, QuadCell * cp)
{
	QuadCell * ul = indexToQuadCell(cp->ul);
	QuadCell * ur = indexToQuadCell(cp->ur);
	QuadCell * ll = indexToQuadCell(cp->ll);
	QuadCell * lr = indexToQuadCell(cp->lr);

	/*
	 * First write out this cell's quadrants if necessary.
	 * This will make sure they have their indexes assigned.
	 */
	writeFunc(fp, ul);
	writeFunc(fp, ur);
	writeFunc(fp, ll);
	writeFunc(fp, lr);

	/*
	 * Get the indexes for the quadrants, or zero if they are empty.
	 */
	NUMINDEX ulIndex = (isCellEmpty(ul) ? 0 : ul->numIndex);
	NUMINDEX urIndex = (isCellEmpty(ur) ? 0 : ur->numIndex);
	NUMINDEX llIndex = (isCellEmpty(ll) ? 0 : ll->numIndex);
	NUMINDEX lrIndex = (isCellEmpty(lr) ? 0 : lr->numIndex);

	/*
	 * Write a line giving our size and the indexes of our quadrants.
	 */
	fprintf(fp, "%d %d %d %d %d\n", cp->size,
		ulIndex, urIndex, llIndex, lrIndex);
}


