/*
 * Copyright (c) 2011 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "life.h"


#ifdef	TERMIOS
#include <termios.h>
#endif

#ifdef	TERMIO
#include <sys/termio.h>
#endif

#ifdef	SGTTY
#include <sys/ioctl.h>
#include <sgtty.h>
#endif


static	int	(*scanRoutine)(void);	/* routine to read characters */
static	JumpBuffer * scanJumpBuf;	/* jump buffer to use in ScanChar */
static	char *	scanReadPtr;		/* current read pointer */
static	char *	scanWritePtr;		/* current write pointer */
static	char	scanBuffer[COMMAND_SIZE+1];/* storage for characters */


/*
 * Initialize for later calls to ScanChar.
 */
void
ScanInit(int (*routine)(void), JumpBuffer * jumpBuf)
{
	/*
	 * Init some static variables.
	 */
	scanRoutine = routine;
	scanJumpBuf = jumpBuf;
	scanWritePtr = scanBuffer;
	scanReadPtr = scanBuffer;

	/*
	 * Ask the system for the settings of the editing characters.
	 * All of this is within a pair of brackets so that local data
	 * structures can be used.
	 */
	{
#ifdef	TERMIOS
		struct	termios	term;

		if (tcgetattr(STDIN_FILENO, &term) == 0)
#endif

#ifdef	TERMIO
		struct	termio	term;

		if (ioctl(STDIN_FILENO, TCGETA, &term) == 0)
#endif

#if	defined(TERMIO) || defined(TERMIOS)
		{
			Life.vKill = term.c_cc[VKILL];
			Life.vErase1 = term.c_cc[VERASE];
			Life.vEof = term.c_cc[VEOF];
#ifdef	VWERASE
			Life.vWordErase = term.c_cc[VWERASE];
#endif

#ifdef	VLNEXT
			Life.vLiteralNext = term.c_cc[VLNEXT];
#endif
		}
#endif


#ifdef	SGTTY
		struct	sgttyb	sgbuf;
		struct	ltchars	ltbuf;

		if (ioctl(STDIN_FILENO, TIOCGETP, &sgbuf) == 0)
		{
			Life.vErase1 = sgbuf.sg_erase;
			Life.vKill = sgbuf.sg_kill;
		}

		if (ioctl(STDIN_FILENO, TIOCGLTC, &ltbuf) == 0)
		{
			Life.vWordErase = ltbuf.t_werasc;
			Life.vLiteralNext = ltbuf.t_lnextc;
		}
#endif
	}

	/*
	 * Undefine any values are aren't set to control characters.
	 * Things like the ancient # and @ kill characters will interfere
	 * with our own commands.
	 */
	if (Life.vKill >= ' ')
		Life.vKill = '\0';

	if (Life.vErase1 >= ' ')
		Life.vErase1 = '\0';

	if (Life.vEof >= ' ')
		Life.vEof = '\0';

	if (Life.vWordErase >= ' ')
		Life.vWordErase = '\0';

	if (Life.vLiteralNext >= ' ')
		Life.vLiteralNext = '\0';

	/*
	 * Now set some defaults if they weren't defined by the system.
	 * Set the second erase character to the other DEL or BS key if the
	 * first erase char is one of them, otherwise set it identically.
	 */
	if (!Life.vErase1)
		Life.vErase1 = '\b';

	Life.vErase2 = Life.vErase1;

	if (Life.vErase1 == '\b')
		Life.vErase2 = DEL;

	if (Life.vErase1 == DEL)
		Life.vErase2 = '\b';

	if (!Life.vKill)
		Life.vKill = 'U'-'@';

	if (!Life.vWordErase)
		Life.vWordErase = 'W'-'@';

	if (!Life.vLiteralNext)
		Life.vLiteralNext = 'V'-'@';

	if (!Life.vEof)
		Life.vEof = 'D'-'@';
}


/*
 * Read the next input character.  If it is an editing character,
 * abort the current context and longjmp back to the last setjmp.
 * NOTE: for proper results, the caller should not alter the global
 * state until the full command has been read in.  This includes such
 * things as prompting for input or saving values.  Otherwise, improper
 * results will occur if the user edits the command.
 */
int
ScanChar(void)
{
	int	ch;		/* current character */

loop:
	/*
	 * If there is a saved character then return that.
	 */
	if (scanReadPtr < scanWritePtr)
		return *scanReadPtr++;

	/*
	 * Get a new character.
	 */
	ch = scanRoutine();

	/*
	 * If EOF was returned then no character is ready now
	 * so longjmp away.
	 */
	if (ch == EOF)
		ScanEof();

	/*
	 * Handle literal input.
	 */
	if (ch == Life.vLiteralNext)
	{
		ch = scanRoutine();

		goto store;
	}

	/*
	 * Handle character erase, by either of the two erase characters.
	 */
	if ((ch == Life.vErase1) || (ch == Life.vErase2))
	{
		if (scanWritePtr <= scanBuffer)
		{
			Beep();

			goto loop;
		}

		scanWritePtr--;
		scanReadPtr = scanBuffer;

		longjmp(scanJumpBuf->buf, SCAN_EDIT);
	}

	/*
	 * Handle word erase.
	 */
	if (ch == Life.vWordErase)
	{
		if (scanWritePtr <= scanBuffer)
			goto loop;

		while ((--scanWritePtr >= scanBuffer) &&
			isBlank(*scanWritePtr))
		    		;

		scanWritePtr++;

		while ((--scanWritePtr >= scanBuffer) &&
			!isBlank(*scanWritePtr))
			    	;

		scanWritePtr++;
		scanReadPtr = scanBuffer;

		longjmp(scanJumpBuf->buf, SCAN_EDIT);
	}

	/*
	 * Handle line erase.
	 */
	if (ch == Life.vKill)
	{
		if (scanWritePtr <= scanBuffer)
			goto loop;

		scanWritePtr = scanBuffer;
		scanReadPtr = scanBuffer;

		longjmp(scanJumpBuf->buf, SCAN_EDIT);
	}

store:	
	if (scanWritePtr >= scanBuffer + COMMAND_SIZE)
	{
		Beep();

		goto loop;
	}

	*scanWritePtr++ = ch;

	return (*scanReadPtr++ & 0xff);
}


/*
 * Abort reading of the current command
 */
void
ScanAbort(void)
{
	scanReadPtr = scanBuffer;
	scanWritePtr = scanBuffer;

	longjmp(scanJumpBuf->buf, SCAN_ABORT);
}


/*
 * Indicate no more characters ready yet
 */
void
ScanEof(void)
{
	scanReadPtr = scanBuffer;

	longjmp(scanJumpBuf->buf, SCAN_EOF);
}


/*
 * Simply reset input and output pointers without longjmping
 */
void
ScanReset(void)
{
	scanReadPtr = scanBuffer;
	scanWritePtr = scanBuffer;
}


void
Beep(void)
{
	/*
	 * Let the output device do the beeping.
	 */
	Life.device->beep(Life.device);
}

/* END CODE */
