/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines having to do with window settings.
 * Each setting can be private to a window or possibly shared
 * with other windows, and can be copied around.
 */

#include "life.h"


/*
 * A structure holding the flag value of a setting,
 * the parameter name of the settings flag, and a
 * description of the parameter.
 */
typedef	struct
{
	SettingsFlags	flag;
	const char *	name;
	const char *	desc;
} SettingsInfo;


/*
 * The table of settings parameters.
 * The names are ordered such that the preferred abbreviations are earlier.
 * The table is ended with an entry containing all zeroes.
 */
static const SettingsInfo settingsTable[] =
{
	{SETTINGS_IS_AUTO_SCALE,	"autoscale",		"Auto scaling is on"},
	{SETTINGS_BACKGROUND_INDEX,	"backgroundcolor",	"Window background color"},
	{SETTINGS_BORDER_WIDTH,		"borderwidth",		"Border width of subwindows"},
	{SETTINGS_CELL_COLOR_INDEX,	"cellcolor",		"Normal cell color"},
	{SETTINGS_CELL_SHAPE,		"cellshape",		"Cell shape"},
	{SETTINGS_IS_CELL_SHAPE_SMALL,	"cellshapeissmall",	"Cell shape is small"},
	{SETTINGS_GRID_TYPE,		"gridtype",		"Grid type"},
	{SETTINGS_DEFAULT_SCALE,	"defaultscale",		"Default scale factor for new objects"},
	{SETTINGS_GRID_CHAR_1,		"gridchar1",		"First character for showing grid"},
	{SETTINGS_GRID_CHAR_2,		"gridchar2",		"Second character for showing grid"},
	{SETTINGS_IS_GRID_ENABLED,	"gridenabled",		"Showing of a grid is enabled"},
	{SETTINGS_GRID_SIZE,		"gridsize",		"Size of grid divivisions"},
	{SETTINGS_MARGIN_PERCENTAGE,	"marginpercentage",	"Window radius percentage for margins"},
	{SETTINGS_MARK_COLOR_INDEX,	"markcolor",		"Marked cell color"},
	{SETTINGS_POSITION_TYPE,	"positiontype",		"Type of positioning to keep view"},
	{SETTINGS_OLD_COLOR_INDEX,	"oldcolor",		"Old cells color"},
	{SETTINGS_SHIFT_FACTOR,		"shiftfactor",		"Shift factor for movement commands"},
	{SETTINGS_IS_SHOW_MY_OUTLINE,	"showmyoutline",	"Allow this window's outline to be seen"},
	{SETTINGS_IS_SHOW_OUTLINES,	"showoutlines",		"Show outlines of other windows"},
	{SETTINGS_STATUS_LOCATION,	"statuslocation",	"Location to show status line"},
	{SETTINGS_OLD_COLOR_INDEX,	"oldcolor",		"Old cells color"},
	{SETTINGS_TEXT_COLOR_INDEX,	"textcolor",		"View text color"},
	{SETTINGS_VIEW_TYPE,		"viewtype",		"Type of location being viewed"},
	{SETTINGS_IS_VISIBLE,		"visible",		"Window is visible"},
	{0,				0,			0}
};


/*
 * Parse a list of settings names and return the 'OR'
 * of all of their values.  Abbreviations are handled.
 * The special value of "all" indicates all settings.
 * The special value of "default" indicates a default.
 * Returns 0 if one or more of the names is not known.
 */
SettingsFlags
ParseSettingsFlags(int count, char ** nameTable)
{
	SettingsFlags flags = SETTINGS_MASK_NONE;

	/*
	 * Check for the single word of "all".
	 */
	if ((count == 1) && (strcmp(nameTable[0], "all") == 0))
		return SETTINGS_MASK_ALL;

	/*
	 * Check for the single word of default.
	 */
	if ((count == 1) && (strcmp(nameTable[0], "default") == 0))
		return Life.defaultSharedSettings;

	/*
	 * Look for settings parameter names or special names
	 * which are combinations of parameters.
	 */
	while (count-- > 0)
	{
		const char * name = *nameTable++;
		const SettingsInfo * info = settingsTable;

		/*
		 * Look for special words for combinations
		 * of related settings.
		 */
		if (IsAbbreviation(name, "grid"))
		{
			flags |= SETTINGS_MASK_GRID;

			continue;
		}

		if (IsAbbreviation(name, "color"))
		{
			flags |= SETTINGS_MASK_COLOR;

			continue;
		}

		/*
		 * Look for an actual setting name.
		 */
		while (info->flag && !IsAbbreviation(name, info->name))
			info++;

		if (info->flag == 0)
			return 0;

		flags |= info->flag;
	}

	return flags;
}


/*
 * Return the settings info structure for the indicated
 * settings flag.  Returns 0 if not found.
 */
static const SettingsInfo *
FindSettingsInfo(SettingsFlags flag)
{
	const SettingsInfo * info = settingsTable;

	while (info->flag && (info->flag != flag))
		info++;

	if (info->flag)
		return info;

	return 0;
}


/*
 * Return the text to describe a cell shape.
 */
static const char *
GetCellShapeText(CellShape shape)
{
	switch (shape)
	{
		case SHAPE_SQUARE:
			return "squares";

		case SHAPE_CIRCLE:
			return "circles";

		case SHAPE_DIAMOND:
			return "diamonds";

		default:
			return "unknown";
	}
}


/*
 * Return the text to describe a grid type.
 */
static const char *
GetGridTypeText(GridType gridType)
{
	switch (gridType)
	{
		case GRID_TYPE_DOTS:
			return "dots";

		case GRID_TYPE_LINES:
			return "lines";

		case GRID_TYPE_CHECKERBOARD:
			return "checkboard";

		default:
			return "unknown";
	}
}


/*
 * Return the text to describe a viewing type.
 * The text is only good until the next call.
 */
static const char *
GetViewTypeText(ViewType type)
{
	static char buf[2];

	switch (type)
	{
		case VIEW_TYPE_CURSOR:
			return "cursor";

		case VIEW_TYPE_ORIGIN:
			return "origin";

		case VIEW_TYPE_MOUSE:
			return "mouse";

		case VIEW_TYPE_UNDER:
			return "under";

		case VIEW_TYPE_SELF:
			return "self";

		default:
			if ((type < VIEW_TYPE_WHERE_A) ||
				(type > VIEW_TYPE_WHERE_Z))
			{
				return "unknown";
			}

			buf[0] = type;
			buf[1] = '\0';

			return buf;
	}
}


/*
 * Return the text to describe a positioning type.
 */
static const char *
GetPositionTypeText(PositionType type)
{
	switch (type)
	{
		case POSITION_TYPE_SHIFT:
			return "shift";

		case POSITION_TYPE_CENTER:
			return "center";

		case POSITION_TYPE_SCALE:
			return "scale";

		default:
			return "unknown";
	}
}


/*
 * Return the text to describe a status location.
 */
static const char *
GetStatusLocationText(StatusLocation location)
{
	switch (location)
	{
		case STATUS_LOCATION_SELF:
			return "visible";

		case STATUS_LOCATION_PARENT:
			return "parent";

		case STATUS_LOCATION_NONE:
			return "invisible";

		default:
			return "unknown";
	}
}


/*
 * Return the text to describe a color index.
 */
static const char *
GetColorIndexText(ColorIndex index)
{
	const char * text = GetColorName(index);

	if (text == 0)
		text = "unknown";

	return text;
}


/*
 * Return the text to describe a boolean value.
 */
static const char *
GetBooleanText(BOOL flag)
{
	return (flag ? "true" : "false");
}


/*
 * Return the text to describe a character value.
 * The text is only good until the next call.
 */
static const char *
GetCharText(int value)
{
	static char buf[2];

	if (value == 0)
		return " ";

	buf[0] = value;
	buf[1] = '\0';

	return buf;
}


/*
 * Return the text to describe a long integer value.
 * The text is only good until the next call.
 */
static const char *
GetLongText(long value)
{
	static char buf[32];

	sprintf(buf, "%ld", value);

	return buf;
}


/*
 * Allocate a new settings object initialized with the default values.
 * The settings object is allocated in our global array.
 * Return 0 if it cannot be allocated or the name is already used.
 */
Settings *
AllocSettings(const char * name, BOOL isTopLevelWindow)
{
	Settings * settings;
	int index;

	/*
	 * If the name is empty or too long or if the number
	 * of settings is exceeded then fail.
	 */
	if ((*name == '\0')|| (strlen(name) >= MAX_SETTINGS_NAME))
		return 0;

	if (Life.settingsCount >= MAX_SETTINGS)
		return 0;

	/*
	 * If the name is already in use then fail.
	 */
	for (index = 0; index < Life.settingsCount; index++)
	{
		if (strcmp(Life.settingsTable[index].name, name) == 0)
			return 0;
	}

	/*
	 * Point to the next settings entry.
	 */
	settings = &Life.settingsTable[Life.settingsCount++];

	/*
	 * Initialize the settings values and set the name.
	 */
	memset(settings, 0, sizeof(Settings));

	settings->isAutoScale = TRUE;

	strcpy(settings->name, name);

	Life.device->defaultSettings(Life.device, isTopLevelWindow, settings);

	/*
	 * Return the address of the settings.
	 */
	return settings;
}


/*
 * Find the settings entry in the table having the specified name.
 * Returns 0 if the specified settings does not exist.
 */
Settings *
FindSettings(const char * name)
{
	int index;

	if (*name == '\0')
		return 0;

	for (index = 0; index < Life.settingsCount; index++)
	{
		Settings * settings = &Life.settingsTable[index];

		if (strcmp(name, settings->name) == 0)
			return settings;
	}

	return 0;
}


/*
 * Compare two settings structures and return a SettingsFlags value
 * indicating what settings values are different between the two.
 * The compared settings values are the private or shared instances
 * according to the sharedFlags of the settings structures.
 */
SettingsFlags
CompareSettings(const Settings * old, const Settings * new)
{
	/*
	 * Initialise a temporary Settings structure.
	 */
	Settings tempSettings;

	memset(&tempSettings, 0, sizeof(Settings));

	tempSettings.changedFlags = SETTINGS_MASK_NONE;
	tempSettings.sharedFlags = SETTINGS_MASK_NONE;
	tempSettings.sharedSettings = &tempSettings;

	/*
	 * Copy the old settings into it without checking changes.
	 */
	RawCopySettings(old, &tempSettings);

	/*
	 * Now copy the new settings into the temporary settings
	 * while updating the changed flags.
	 */
	CopySettings(new, &tempSettings);

	/*
	 * Return the mask of changed settings.
	 */
	return tempSettings.changedFlags;
}


/*
 * Copy the settings values from one instance to another.
 * Everything but the name, changed flags, and share flags
 * are copied.  This doesn't use the shared values.
 */
void
RawCopySettings(const Settings * old, Settings * new)
{
	if (old == new)
		return;

	new->borderWidth = old->borderWidth;
	new->shiftFactor = old->shiftFactor;
	new->defaultScale = old->defaultScale;
	new->cellShape = old->cellShape;
	new->isCellShapeSmall = old->isCellShapeSmall;
	new->gridType = old->gridType;
	new->isVisible = old->isVisible;
	new->isShowOutlines = old->isShowOutlines;
	new->isShowMyOutline = old->isShowMyOutline;
	new->isGridEnabled = old->isGridEnabled;
	new->gridChar1 = old->gridChar1;
	new->gridChar2 = old->gridChar2;
	new->gridSize = old->gridSize;
	new->isAutoScale = old->isAutoScale;
	new->marginPercentage = old->marginPercentage;
	new->backgroundIndex = old->backgroundIndex;
	new->cellColorIndex = old->cellColorIndex;
	new->tracksColorIndex = old->tracksColorIndex;
	new->markColorIndex = old->markColorIndex;
	new->oldColorIndex = old->oldColorIndex;
	new->textColorIndex = old->textColorIndex;
	new->viewType = old->viewType;
	new->positionType = old->positionType;
	new->statusLocation = old->statusLocation;
}


/*
 * Clear all of the changed flags in the settings table.
 * This is called after all window updates are completed.
 */
void
ClearSettingsChangedFlags(void)
{
	int index;

	for (index = 0; index < Life.settingsCount; index++)
	{
		Life.settingsTable[index].changedFlags = SETTINGS_MASK_NONE;
	}
}


/*
 * Copy all the settings values from one instance to another.
 * Everything but the name and share flags are copied.
 * The values copied are the private or shared instances
 * as specified by the sharedFlags.  Any windows using the
 * new settings values have the appropriate update flags set.
 */
void
CopySettings(const Settings * old, Settings * new)
{
	CopySomeSettings(old, new, SETTINGS_MASK_ALL);
}


/*
 * Copy some settings values from one instance to another.
 * Everything but the name and share flags are copied
 * according to the specified settings mask value.
 * The values copied are the private or shared instances
 * as specified by the sharedFlags.  Any windows using the
 * new settings values have the appropriate update flags set.
 */
void
CopySomeSettings(const Settings * old, Settings * new, SettingsFlags copyMask)
{
	if (old == new)
		return;

	if (copyMask & SETTINGS_BORDER_WIDTH)
		SetBorderWidth(new, GetBorderWidth(old));

	if (copyMask & SETTINGS_SHIFT_FACTOR)
		SetShiftFactor(new, GetShiftFactor(old));

	if (copyMask & SETTINGS_DEFAULT_SCALE)
		SetDefaultScale(new, GetDefaultScale(old));

	if (copyMask & SETTINGS_CELL_SHAPE)
		SetCellShape(new, GetCellShape(old));

	if (copyMask & SETTINGS_IS_CELL_SHAPE_SMALL)
		SetIsCellShapeSmall(new, GetIsCellShapeSmall(old));

	if (copyMask & SETTINGS_GRID_TYPE)
		SetGridType(new, GetGridType(old));

	if (copyMask & SETTINGS_IS_VISIBLE)
		SetIsVisible(new, GetIsVisible(old));

	if (copyMask & SETTINGS_IS_SHOW_OUTLINES)
		SetIsShowOutlines(new, GetIsShowOutlines(old));

	if (copyMask & SETTINGS_IS_SHOW_MY_OUTLINE)
		SetIsShowMyOutline(new, GetIsShowMyOutline(old));

	if (copyMask & SETTINGS_IS_GRID_ENABLED)
		SetIsGridEnabled(new, GetIsGridEnabled(old));

	if (copyMask & SETTINGS_GRID_CHAR_1)
		SetGridChar1(new, GetGridChar1(old));

	if (copyMask & SETTINGS_GRID_CHAR_2)
		SetGridChar2(new, GetGridChar2(old));

	if (copyMask & SETTINGS_GRID_SIZE)
		SetGridSize(new, GetGridSize(old));

	if (copyMask & SETTINGS_IS_AUTO_SCALE)
		SetIsAutoScale(new, GetIsAutoScale(old));

	if (copyMask & SETTINGS_MARGIN_PERCENTAGE)
		SetMarginPercentage(new, GetMarginPercentage(old));

	if (copyMask & SETTINGS_BACKGROUND_INDEX)
		SetBackgroundIndex(new, GetBackgroundIndex(old));

	if (copyMask & SETTINGS_CELL_COLOR_INDEX)
		SetCellColorIndex(new, GetCellColorIndex(old));

	if (copyMask & SETTINGS_TRACKS_COLOR_INDEX)
		SetTracksColorIndex(new, GetTracksColorIndex(old));

	if (copyMask & SETTINGS_MARK_COLOR_INDEX)
		SetMarkColorIndex(new, GetMarkColorIndex(old));

	if (copyMask & SETTINGS_OLD_COLOR_INDEX)
		SetOldColorIndex(new, GetOldColorIndex(old));

	if (copyMask & SETTINGS_TEXT_COLOR_INDEX)
		SetTextColorIndex(new, GetTextColorIndex(old));

	if (copyMask & SETTINGS_VIEW_TYPE)
		SetViewType(new, GetViewType(old));

	if (copyMask & SETTINGS_POSITION_TYPE)
		SetPositionType(new, GetPositionType(old));

	if (copyMask & SETTINGS_STATUS_LOCATION)
		SetStatusLocation(new, GetStatusLocation(old));
}


/*
 * Format and show a line for showing a settings value.
 */
static void
ShowSettingLine(const LifeWindow * window,
	SettingsFlags settingFlag, const char * value)
{
	int sharedIndicator;
	const SettingsInfo * info = FindSettingsInfo(settingFlag);

	if (info == 0)
		return;

	/*
	 * Get a flag character to indicate if the setting is shared.
	 */
	sharedIndicator =
		((window->settings.sharedFlags & settingFlag) ? '*' : ' ');

	/*
	 * Format the info.
	 */
	FormatHelp("%c  %-18s  %-18s  %s\n", sharedIndicator,
		info->name, value, info->desc);
}


/*
 * Show the window settings.
 */
void
ShowWindowSettings(const LifeWindow * window)
{
	const Settings * settings = &window->settings;

	FormatHelp("Current settings for window %ld:\n\n", window->id);

	ShowHelp("   Name                Value               Description\n");
	ShowHelp("   ----------------    ----------------    -------------------\n");

	ShowSettingLine(window, SETTINGS_IS_AUTO_SCALE,
		GetBooleanText(GetIsAutoScale(settings)));

	ShowSettingLine(window, SETTINGS_BACKGROUND_INDEX,
		GetColorIndexText(GetBackgroundIndex(settings)));

	ShowSettingLine(window, SETTINGS_BORDER_WIDTH,
		GetLongText(GetBorderWidth(settings)));

	ShowSettingLine(window, SETTINGS_CELL_COLOR_INDEX,
		GetColorIndexText(GetCellColorIndex(settings)));

	ShowSettingLine(window, SETTINGS_CELL_SHAPE,
		GetCellShapeText(GetCellShape(settings)));

	ShowSettingLine(window, SETTINGS_IS_CELL_SHAPE_SMALL,
		GetBooleanText(GetIsCellShapeSmall(settings)));

	ShowSettingLine(window, SETTINGS_DEFAULT_SCALE,
		GetLongText(GetDefaultScale(settings)));

	ShowSettingLine(window, SETTINGS_GRID_TYPE,
		GetGridTypeText(GetGridType(settings)));

	ShowSettingLine(window, SETTINGS_GRID_CHAR_1,
		GetCharText(GetGridChar1(settings)));

	ShowSettingLine(window, SETTINGS_GRID_CHAR_2,
		GetCharText(GetGridChar2(settings)));

	ShowSettingLine(window, SETTINGS_IS_GRID_ENABLED,
		GetBooleanText(GetIsGridEnabled(settings)));

	ShowSettingLine(window, SETTINGS_GRID_SIZE,
		GetLongText(GetGridSize(settings)));

	ShowSettingLine(window, SETTINGS_MARGIN_PERCENTAGE,
		GetLongText(GetMarginPercentage(settings)));

	ShowSettingLine(window, SETTINGS_MARK_COLOR_INDEX,
		GetColorIndexText(GetMarkColorIndex(settings)));

	ShowSettingLine(window, SETTINGS_OLD_COLOR_INDEX,
		GetColorIndexText(GetOldColorIndex(settings)));

	ShowSettingLine(window, SETTINGS_POSITION_TYPE,
		GetPositionTypeText(GetPositionType(settings)));

	ShowSettingLine(window, SETTINGS_SHIFT_FACTOR,
		GetLongText(GetShiftFactor(settings)));

	ShowSettingLine(window, SETTINGS_IS_SHOW_MY_OUTLINE,
		GetBooleanText(GetIsShowMyOutline(settings)));

	ShowSettingLine(window, SETTINGS_IS_SHOW_OUTLINES,
		GetBooleanText(GetIsShowOutlines(settings)));

	ShowSettingLine(window, SETTINGS_STATUS_LOCATION,
		GetStatusLocationText(GetStatusLocation(settings)));

	ShowSettingLine(window, SETTINGS_TEXT_COLOR_INDEX,
		GetColorIndexText(GetTextColorIndex(settings)));

	ShowSettingLine(window, SETTINGS_TRACKS_COLOR_INDEX,
		GetColorIndexText(GetTracksColorIndex(settings)));

	ShowSettingLine(window, SETTINGS_VIEW_TYPE,
		GetViewTypeText(GetViewType(settings)));

	ShowSettingLine(window, SETTINGS_IS_VISIBLE,
		GetBooleanText(GetIsVisible(settings)));

	ShowHelp("\n\nSettings marked with an asterisk have values shared with \"");
	ShowHelp(settings->sharedSettings->name);
	ShowHelp("\".\n");

	ShowHelp("Unmarked settings have values only for this window.\n\n");

	EndHelp(TRUE);
}


/*
 * Add the specified settings flags to the settings
 * being shared with other windows.
 */
void
AddSharedWindowSettings(LifeWindow * window, SettingsFlags flags)
{
	/*
	 * Calculate the window's settings which are not now
	 * shared but which will be shared using the new flags.
	 */
	SettingsFlags changedFlags = (flags & ~window->settings.sharedFlags);

	/*
	 * Add in the changed flags to the window's changed
	 * flags and add in the specified shared flags.
	 */
	window->settings.changedFlags |= changedFlags;
	window->settings.sharedFlags |= flags;

	/*
	 * Redo everything.
	 * TODO: Only update what is needed.
	 */
	window->updateFlags |= (U_REDRAW | U_CONFIG);
}


/*
 * Substract the specified settings flags from the settings
 * being shared with other windows so that they become private.
 */
void
SubtractSharedWindowSettings(LifeWindow * window, SettingsFlags flags)
{
	/*
	 * Calculate the window's settings which are currently
	 * shared but which will not be shared using the new flags.
	 */
	SettingsFlags changedFlags = (flags & window->settings.sharedFlags);

	/*
	 * Make a copy of all of the current settings.
	 */
	Settings newSettings;

	newSettings.sharedFlags = SETTINGS_MASK_NONE;

	CopySettings(&window->settings, &newSettings);

	/*
	 * Add the changed flags to the window's changed
	 * flags and remove the specified shared flags.
	 */
	window->settings.changedFlags |= changedFlags;
	window->settings.sharedFlags &= ~flags;

	/*
	 * Now copy back the raw values that were saved to
	 * set the window's private values correctly.
	 */
	RawCopySettings(&newSettings, &window->settings);

	/*
	 * Redo everything.
	 * TODO: Only update what is needed.
	 */
	window->updateFlags |= (U_REDRAW | U_CONFIG);
}


/*
 * Accessor functions for settings.
 */
Value
GetShiftFactor(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_SHIFT_FACTOR)
		settings = settings->sharedSettings;

	return settings->shiftFactor;
}

Scale
GetDefaultScale(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_DEFAULT_SCALE)
		settings = settings->sharedSettings;

	return settings->defaultScale;
}

CellShape
GetCellShape(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_CELL_SHAPE)
		settings = settings->sharedSettings;

	return settings->cellShape;
}

GridType
GetGridType(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_GRID_TYPE)
		settings = settings->sharedSettings;

	return settings->gridType;
}

BOOL
GetIsVisible(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_IS_VISIBLE)
		settings = settings->sharedSettings;

	return settings->isVisible;
}

BOOL
GetIsShowOutlines(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_IS_SHOW_OUTLINES)
		settings = settings->sharedSettings;

	return settings->isShowOutlines;
}

BOOL
GetIsShowMyOutline(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_IS_SHOW_MY_OUTLINE)
		settings = settings->sharedSettings;

	return settings->isShowMyOutline;
}

BOOL
GetIsGridEnabled(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_IS_GRID_ENABLED)
		settings = settings->sharedSettings;

	return settings->isGridEnabled;
}

BOOL
GetIsCellShapeSmall(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_IS_CELL_SHAPE_SMALL)
		settings = settings->sharedSettings;

	return settings->isCellShapeSmall;
}

int
GetGridChar1(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_GRID_CHAR_1)
		settings = settings->sharedSettings;

	return settings->gridChar1;
}

int
GetGridChar2(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_GRID_CHAR_2)
		settings = settings->sharedSettings;

	return settings->gridChar2;
}

Count
GetGridSize(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_GRID_SIZE)
		settings = settings->sharedSettings;

	return settings->gridSize;
}

PixelCount
GetBorderWidth(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_BORDER_WIDTH)
		settings = settings->sharedSettings;

	return settings->borderWidth;
}

BOOL
GetIsAutoScale(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_IS_AUTO_SCALE)
		settings = settings->sharedSettings;

	return settings->isAutoScale;
}

int
GetMarginPercentage(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_MARGIN_PERCENTAGE)
		settings = settings->sharedSettings;

	return settings->marginPercentage;
}

ColorIndex
GetBackgroundIndex(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_BACKGROUND_INDEX)
		settings = settings->sharedSettings;

	return settings->backgroundIndex;
}

ColorIndex
GetCellColorIndex(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_CELL_COLOR_INDEX)
		settings = settings->sharedSettings;

	return settings->cellColorIndex;
}

ColorIndex
GetTracksColorIndex(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_TRACKS_COLOR_INDEX)
		settings = settings->sharedSettings;

	return settings->tracksColorIndex;
}

ColorIndex
GetMarkColorIndex(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_MARK_COLOR_INDEX)
		settings = settings->sharedSettings;

	return settings->markColorIndex;
}

ColorIndex
GetOldColorIndex(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_OLD_COLOR_INDEX)
		settings = settings->sharedSettings;

	return settings->oldColorIndex;
}

ColorIndex
GetTextColorIndex(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_TEXT_COLOR_INDEX)
		settings = settings->sharedSettings;

	return settings->textColorIndex;
}

ViewType
GetViewType(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_VIEW_TYPE)
		settings = settings->sharedSettings;

	return settings->viewType;
}

PositionType
GetPositionType(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_POSITION_TYPE)
		settings = settings->sharedSettings;

	return settings->positionType;
}

StatusLocation
GetStatusLocation(const Settings * settings)
{
	if (settings->sharedFlags & SETTINGS_STATUS_LOCATION)
		settings = settings->sharedSettings;

	return settings->statusLocation;
}


/*
 * Setting set functions.
 * If the setting is changed then the changed flag
 * is set for the setting.
 */
void
SetShiftFactor(Settings * settings, Value shiftFactor)
{
	if (settings->sharedFlags & SETTINGS_SHIFT_FACTOR)
		settings = settings->sharedSettings;

	if (settings->shiftFactor == shiftFactor)
		return;

	settings->changedFlags |= SETTINGS_SHIFT_FACTOR;
	settings->shiftFactor = shiftFactor;
}

void
SetDefaultScale(Settings * settings, Scale scale)
{
	if (settings->sharedFlags & SETTINGS_DEFAULT_SCALE)
		settings = settings->sharedSettings;

	if (settings->defaultScale == scale)
		return;

	settings->changedFlags |= SETTINGS_DEFAULT_SCALE;
	settings->defaultScale = scale;
}

void
SetCellShape(Settings * settings, CellShape cellShape)
{
	if (settings->sharedFlags & SETTINGS_CELL_SHAPE)
		settings = settings->sharedSettings;

	if (settings->cellShape == cellShape)
		return;

	settings->changedFlags |= SETTINGS_CELL_SHAPE;
	settings->cellShape = cellShape;
}

void
SetGridType(Settings * settings, GridType gridType)
{
	if (settings->sharedFlags & SETTINGS_GRID_TYPE)
		settings = settings->sharedSettings;

	if (settings->gridType == gridType)
		return;

	settings->changedFlags |= SETTINGS_GRID_TYPE;
	settings->gridType = gridType;
}

void
SetIsVisible(Settings * settings, BOOL isVisible)
{
	if (settings->sharedFlags & SETTINGS_IS_VISIBLE)
		settings = settings->sharedSettings;

	if (settings->isVisible == isVisible)
		return;

	settings->changedFlags |= SETTINGS_IS_VISIBLE;
	settings->isVisible = isVisible;
}

void
SetIsShowOutlines(Settings * settings, BOOL isEnabled)
{
	if (settings->sharedFlags & SETTINGS_IS_SHOW_OUTLINES)
		settings = settings->sharedSettings;

	if (settings->isShowOutlines == isEnabled)
		return;

	settings->changedFlags |= SETTINGS_IS_SHOW_OUTLINES;
	settings->isShowOutlines = isEnabled;
}

void
SetIsShowMyOutline(Settings * settings, BOOL isEnabled)
{
	if (settings->sharedFlags & SETTINGS_IS_SHOW_MY_OUTLINE)
		settings = settings->sharedSettings;

	if (settings->isShowMyOutline == isEnabled)
		return;

	settings->changedFlags |= SETTINGS_IS_SHOW_MY_OUTLINE;
	settings->isShowMyOutline = isEnabled;
}

void
SetIsGridEnabled(Settings * settings, BOOL isEnabled)
{
	if (settings->sharedFlags & SETTINGS_IS_GRID_ENABLED)
		settings = settings->sharedSettings;

	if (settings->isGridEnabled == isEnabled)
		return;

	settings->changedFlags |= SETTINGS_IS_GRID_ENABLED;
	settings->isGridEnabled = isEnabled;
}

void
SetIsCellShapeSmall(Settings * settings, BOOL isSmall)
{
	if (settings->sharedFlags & SETTINGS_IS_CELL_SHAPE_SMALL)
		settings = settings->sharedSettings;

	if (settings->isCellShapeSmall == isSmall)
		return;

	settings->changedFlags |= SETTINGS_IS_CELL_SHAPE_SMALL;
	settings->isCellShapeSmall = isSmall;
}

void
SetGridChar1(Settings * settings, int gridChar)
{
	if (settings->sharedFlags & SETTINGS_GRID_CHAR_1)
		settings = settings->sharedSettings;

	if (settings->gridChar1 == gridChar)
		return;

	settings->changedFlags |= SETTINGS_GRID_CHAR_1;
	settings->gridChar1 = gridChar;
}

void
SetGridChar2(Settings * settings, int gridChar)
{
	if (settings->sharedFlags & SETTINGS_GRID_CHAR_2)
		settings = settings->sharedSettings;

	if (settings->gridChar2 == gridChar)
		return;

	settings->changedFlags |= SETTINGS_GRID_CHAR_2;
	settings->gridChar2 = gridChar;
}

void
SetGridSize(Settings * settings, Count gridSize)
{
	if (settings->sharedFlags & SETTINGS_GRID_SIZE)
		settings = settings->sharedSettings;

	if (settings->gridSize == gridSize)
		return;

	settings->changedFlags |= SETTINGS_GRID_SIZE;
	settings->gridSize = gridSize;
}

void
SetBorderWidth(Settings * settings, PixelCount width)
{
	if (settings->sharedFlags & SETTINGS_BORDER_WIDTH)
		settings = settings->sharedSettings;

	if (settings->borderWidth == width)
		return;

	settings->changedFlags |= SETTINGS_BORDER_WIDTH;
	settings->borderWidth = width;
}

void
SetIsAutoScale(Settings * settings, BOOL isEnabled)
{
	if (settings->sharedFlags & SETTINGS_IS_AUTO_SCALE)
		settings = settings->sharedSettings;

	if (settings->isAutoScale == isEnabled)
		return;

	settings->changedFlags |= SETTINGS_IS_AUTO_SCALE;
	settings->isAutoScale = isEnabled;
}

void
SetMarginPercentage(Settings * settings, int percentage)
{
	if (settings->sharedFlags & SETTINGS_MARGIN_PERCENTAGE)
		settings = settings->sharedSettings;

	if (settings->marginPercentage == percentage)
		return;

	settings->changedFlags |= SETTINGS_MARGIN_PERCENTAGE;
	settings->marginPercentage = percentage;
}

void
SetBackgroundIndex(Settings * settings, ColorIndex colorIndex)
{
	if (settings->sharedFlags & SETTINGS_BACKGROUND_INDEX)
		settings = settings->sharedSettings;

	if (settings->backgroundIndex == colorIndex)
		return;

	settings->changedFlags |= SETTINGS_BACKGROUND_INDEX;
	settings->backgroundIndex = colorIndex;
}

void
SetCellColorIndex(Settings * settings, ColorIndex colorIndex)
{
	if (settings->sharedFlags & SETTINGS_CELL_COLOR_INDEX)
		settings = settings->sharedSettings;

	if (settings->cellColorIndex == colorIndex)
		return;

	settings->changedFlags |= SETTINGS_CELL_COLOR_INDEX;
	settings->cellColorIndex = colorIndex;
}

void
SetTracksColorIndex(Settings * settings, ColorIndex colorIndex)
{
	if (settings->sharedFlags & SETTINGS_TRACKS_COLOR_INDEX)
		settings = settings->sharedSettings;

	if (settings->tracksColorIndex == colorIndex)
		return;

	settings->changedFlags |= SETTINGS_TRACKS_COLOR_INDEX;
	settings->tracksColorIndex = colorIndex;
}

void
SetMarkColorIndex(Settings * settings, ColorIndex colorIndex)
{
	if (settings->sharedFlags & SETTINGS_MARK_COLOR_INDEX)
		settings = settings->sharedSettings;

	if (settings->markColorIndex == colorIndex)
		return;

	settings->changedFlags |= SETTINGS_MARK_COLOR_INDEX;
	settings->markColorIndex = colorIndex;
}

void
SetOldColorIndex(Settings * settings, ColorIndex colorIndex)
{
	if (settings->sharedFlags & SETTINGS_OLD_COLOR_INDEX)
		settings = settings->sharedSettings;

	if (settings->oldColorIndex == colorIndex)
		return;

	settings->changedFlags |= SETTINGS_OLD_COLOR_INDEX;
	settings->oldColorIndex = colorIndex;
}

void
SetTextColorIndex(Settings * settings, ColorIndex colorIndex)
{
	if (settings->sharedFlags & SETTINGS_TEXT_COLOR_INDEX)
		settings = settings->sharedSettings;

	if (settings->textColorIndex == colorIndex)
		return;

	settings->changedFlags |= SETTINGS_TEXT_COLOR_INDEX;
	settings->textColorIndex = colorIndex;
}

void
SetViewType(Settings * settings, ViewType viewType)
{
	if (settings->sharedFlags & SETTINGS_VIEW_TYPE)
		settings = settings->sharedSettings;

	if (settings->viewType == viewType)
		return;

	settings->changedFlags |= SETTINGS_VIEW_TYPE;
	settings->viewType = viewType;
}

void
SetPositionType(Settings * settings, PositionType positionType)
{
	if (settings->sharedFlags & SETTINGS_POSITION_TYPE)
		settings = settings->sharedSettings;

	if (settings->positionType == positionType)
		return;

	settings->changedFlags |= SETTINGS_POSITION_TYPE;
	settings->positionType = positionType;
}

void
SetStatusLocation(Settings * settings, StatusLocation location)
{
	if (settings->sharedFlags & SETTINGS_STATUS_LOCATION)
		settings = settings->sharedSettings;

	if (settings->statusLocation == location)
		return;

	settings->changedFlags |= SETTINGS_STATUS_LOCATION;
	settings->statusLocation = location;
}


/*
 * Store the current window settings into a settings structure
 * having the specified name.  The settings structure is created
 * if it does not exist.  Returns TRUE if successful.
 */
BOOL
StoreWindowSettings(const LifeWindow * window, const char * name)
{
	Settings * settings;

	/*
	 * Find the settings if it already exists.
	 */
	settings = FindSettings(name);

	/*
	 * If the name isn't known then allocate a new settings structure.
	 */
	if (settings == 0)
	{
		settings = AllocSettings(name, (window->parentWindow == 0));

		if (settings == 0)
			return FALSE;
	}

	/*
	 * Save the window settings and return success.
	 */
	CopySettings(&window->settings, settings);

	return TRUE;
}


/*
 * Load the window settings from the setting structure having
 * the specified name.  Returns TRUE if successful, or FALSE
 * if there is no such settings.
 */
BOOL
LoadWindowSettings(LifeWindow * window, const char * name)
{
	const Settings * settings;

	/*
	 * Find the settings structure and fail if not found.
	 */
	settings = FindSettings(name);

	if (settings == 0)
		return FALSE;

	/*
	 * Copy the settings into our window's settings.
	 */
	CopySettings(settings, &window->settings);

	return TRUE;
}


/*
 * Merge the window settings from the setting structure having
 * the specified name.  This means that only the private settings
 * of the window are copied over.  Returns TRUE if successful,
 * or FALSE if there is no such settings.
 */
BOOL
MergeWindowSettings(LifeWindow * window, const char * name)
{
	const Settings * settings;

	/*
	 * Find the settings structure and fail if not found.
	 */
	settings = FindSettings(name);

	if (settings == 0)
		return FALSE;

	/*
	 * Copy the settings wbhich are private into our
	 * window's settings, preserving our shared settings.
	 */
	CopySomeSettings(settings, &window->settings,
		~window->settings.sharedFlags);

	return TRUE;
}


/*
 * Use the specified window settings for all parameters which
 * are not set as private to the window.  The settings are
 * created if they don't yet exist.  Returns TRUE if successful
 * or FALSE on a failure.
 */
BOOL
UseWindowSettings(LifeWindow * window, const char * name)
{
	Settings * settings;
	SettingsFlags changedFlags;

	/*
	 * Find the settings structure if it exists.
	 */
	settings = FindSettings(name);

	/*
	 * If the settings don't exist then create them and
	 * copy our settings to it and share with that.
	 */
	if (settings == 0)
	{
		settings = AllocSettings(name, (window->parentWindow == 0));

		if (settings == 0)
			return FALSE;

		CopySettings(&window->settings, settings);

		window->settings.sharedSettings = settings;

		return TRUE;
	}

	/*
	 * We are sharing someone else's existing settings.
	 * Get the changed flags for the settings change.
	 */
	changedFlags = CompareSettings(&window->settings, settings);

	/*
	 * Use the shared settings and add the changed flags.
	 */
	window->settings.sharedSettings = settings;
	window->settings.changedFlags |= changedFlags;

	return TRUE;
}


/*
 * Set the window's settings to the initial ones.
 */
void
InitializeWindowSettings(LifeWindow * window)
{
	LifeDevice * dev = Life.device;
	Settings settings;

	/*
	 * Get the initial device settings.
	 */
	dev->defaultSettings(dev, window->parentWindow == 0, &settings);

	/*
	 * Copy those settings to the window settings.
	 */
	CopySettings(&settings, &window->settings);
}


/*
 * Copy the appropriate default window settings into the
 * window's settings.
 */
void
DefaultWindowSettings(LifeWindow * window)
{
	const Settings * defaultSettings;

	/*
	 * Get the appropriate settings to use for a top
	 * level window or a sub-window.
	 */
	if (window->parentWindow)
		defaultSettings = Life.topLevelWindowSettings;
	else
		defaultSettings = Life.subWindowSettings;

	/*
	 * Copy the settings into our window's settings.
	 */
	CopySettings(defaultSettings, &window->settings);
}

/* END CODE */
