/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include <signal.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <errno.h>
#include <curses.h>
#include "/usr/include/sys/termio.h"
#include "life.h"


/*
 * Device interface routines for the rest of the life program to call.
 */
static	BOOL	TtyOpenDevice(LifeDevice *);
static	void	TtyCloseDevice(LifeDevice *);
static	LifeWindow * TtyOpenWindow(LifeDevice *, LifeWindow * parent);
static	BOOL	TtyCloseWindow(LifeDevice *);
static	void	TtySetStatusMode(LifeDevice *, StatusMode);
static	void	TtyAddStatusText(LifeDevice *, const char *);
static	void	TtyClearStatusMessage(LifeDevice *);
static	void	TtySetIsVisible(LifeDevice *, LifeWindow * window, BOOL isVisible);
static	void	TtySetViewModeIsText(LifeDevice *, BOOL);
static	void	TtyAddViewText(LifeDevice *, const char *);
static	void	TtyUpdate(LifeDevice *);
static	void	TtyRedraw(LifeDevice *);
static	void	TtyDelay(LifeDevice *, int);
static	BOOL	TtyIsInputReady(LifeDevice *);
static	int	TtyReadChar(LifeDevice *, BOOL);
static	BOOL	TtyAssign(LifeDevice *, Value, int);
static	void	TtyWriteAssign(LifeDevice *, FILE *);
static	void	TtyBeep(LifeDevice *);
static	void	TtyDefaultSettings(LifeDevice *, BOOL isTopLevelWindow, Settings * settings);
static	ColorIndex TtyDefineColor(LifeDevice *, const char * name);


static LifeDevice ttyDevice =
{
	TtyOpenDevice, TtyCloseDevice, TtyOpenWindow,
	TtyCloseWindow, TtySetIsVisible, TtySetStatusMode,
	TtyAddStatusText, TtyClearStatusMessage,
	TtySetViewModeIsText, TtyAddViewText, TtyUpdate,
	TtyRedraw, TtyDelay, TtyIsInputReady,
	TtyReadChar, TtyAssign, TtyWriteAssign, TtyBeep,
	TtyDefaultSettings, TtyDefineColor
};


/*
 * Private data for a terminal window.
 */
typedef struct
{
	WINDOW *	statWin;	/* status curses window */
	WINDOW *	viewWin;	/* view curses window */
	PixelCount	viewHeight;	/* height of view window */
} TtyData;


/*
 * The type of event returned while waiting for characters.
 */
typedef enum EventType
{
	EVENT_INPUT,		/* input data can be read */
	EVENT_TIMEOUT,		/* timeout occurred */
	EVENT_SIGNAL,		/* signal interrupted the wait */
	EVENT_FAILED		/* the select failed */
} EventType;


/*
 * Local routines.
 */
static void UpdateWindow(LifeWindow * window);
static void UpdateView(LifeWindow * window);

static void ViewNormal(LifeWindow * window);

static void ViewNormalRow(LifeWindow * window, int rowIndex,
	ViewInfo * currentInfo, ViewInfo * usedInfo);

static void ViewBlankNormalRow(LifeWindow * window, int rowIndex);

static void ViewScaled(LifeWindow * window);

static void ViewScaledRow(LifeWindow * window, int rowIndex,
	ViewInfo * currentInfo);

static void ViewBlankScaledRow(LifeWindow * window, int rowIndex);

static void PositionCursor(LifeWindow * window);
static int GridChar(const LifeWindow * window, NUMBER row, NUMBER col);
static int TryReadChar(int timeoutMilliSeconds);
static void UpdateStatusLine(LifeWindow * window);
static void HandleResize(int arg);
static void CheckTerminalSize(void);
static void ConfigureWindow(LifeWindow * window, int rows, int cols);
static EventType WaitForEvent(int milliSeconds);


/*
 * Static data for the terminal display.
 */
static BOOL	isSizeChanged;	/* terminal size has changed */
static BOOL	isCharSaved;	/* whether there is a saved character */
static int	savedChar;	/* the saved character if any */
static int	lastTimeout;	/* last timeout value set for curses */


/*
 * Return the instance of the terminal device.
 */
LifeDevice *
GetTtyDevice(void)
{
	return &ttyDevice;
}


static BOOL
TtyOpenDevice(LifeDevice * dev)
{
	/*
	 * Initialise the curses settings.
	 */
	initscr();
	cbreak();
	noecho();
	nonl();
	typeahead(-1);
	intrflush(stdscr, FALSE);

	/*
	 * Set the read values to match curses initially.
	 */
	lastTimeout = -1;
	isCharSaved = FALSE;

/*	keypad(stdscr, TRUE);	*/

	dev->minScale = 1;
	dev->maxScale = MAX_SCALE;
	dev->windowCount = 0;

	/*
	 * Set up to catch terminal resizing.
	 */
	signal(SIGWINCH, HandleResize);

	/*
	 * Act as if the terminal changed size
	 */
	isSizeChanged = TRUE;
	CheckTerminalSize();

	return TRUE;
}


static void
TtyCloseDevice(LifeDevice * dev)
{
	refresh();
	endwin();
}


/*
 * Signal handler for resizing of window.
 * This only sets a flag so that we can handle the resize later.
 * (Changing the size at unpredictable times would be dangerous.)
 */
static void
HandleResize(int arg)
{
	isSizeChanged = TRUE;

	signal(SIGWINCH, HandleResize);
}


/*
 * Routine called to get the new terminal size from the kernel.
 * We inform curses of this change and let it resize its window.
 * This does nothing if the signal handler has not been called.
 */
static void
CheckTerminalSize(void)
{
	LifeDevice * dev = Life.device;
	struct winsize	size;
	int rows;
	int cols;
	int index;

	/*
	 * If there is no indication of a window size change
	 * then do nothing.
	 */
	if (!isSizeChanged)
		return;

	/*
	 * Clear the size changed flag and get the new
	 * terminal size.
	 */
	isSizeChanged = FALSE;

	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &size) < 0)
		return;

	rows = size.ws_row;
	cols = size.ws_col;

	if (rows <= 0)
		rows = 1;

	if (cols <= 0)
		cols = 1;

	/*
	 * If the values have not changed then do nothing.
	 */
	if ((rows == LINES) && (cols == COLS))
		return;

	/*
	 * Tell curses the new window size.
	 */
	if (resize_term(rows, cols) == ERR)
		return;

	/*
	 * Loop over the windows to configure their sizes.
	 * Note: This is only correct for one window.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * window = dev->windowTable[index];

		ConfigureWindow(window, rows, cols);
	}
}


/*
 * Reconfigure a window to be of the specified size.
 */
static void
ConfigureWindow(LifeWindow * window, int rows, int cols)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	Scale scale = window->scale;
	PixelCount viewRows = rows - 1;
	PixelCount viewCols = cols - 1;

	if (scale <= 0)
		scale = 1;

	/*
	 * Clear and resize the windows.
	 */
	wclear(ttyData->statWin);
	wclear(ttyData->viewWin);
	wresize(ttyData->statWin, 1, cols);
	wresize(ttyData->viewWin, viewRows, cols);

	/*
	 * Recalculate the window parameters.
	 */
	window->updateFlags |= U_REDRAW;
	window->width = cols;
	window->height = rows;
	window->rows = viewRows;
	window->cols = viewCols;
	window->viewRows = numFromLong(viewRows * scale);
	window->viewCols = numFromLong(viewCols * scale);
	window->maxRow = numDec(numAdd(window->minRow, window->viewRows));
	window->maxCol = numDec(numAdd(window->minCol, window->viewCols));
	window->textRows = viewRows;
	window->textCols = viewCols;

	ttyData->viewHeight = viewRows;
}


static LifeWindow *
TtyOpenWindow(LifeDevice * dev, LifeWindow * parent)
{
	LifeWindow * window;
	TtyData * ttyData;
	Settings * settings;
	int windowWidth;
	int windowHeight;
	int viewHeight;
	int viewWidth;

	/*
	 * If too many windows are opened then fail.
	 * We only allow one window.
	 */
	if (dev->windowCount > 0)
		return 0;

	/*
	 * If for some reason a parent was given then fail.
	 */
	if (parent)
		return 0;

	/*
	 * Allocate a new window structure.
	 */
	window = malloc(sizeof(LifeWindow));

	if (window == 0)
		return 0;

	/*
	 * Set everything to zero to begin with.
	 */
	memset(window, 0, sizeof(LifeWindow));

	/*
	 * Remember our parent window (or none) and
	 * set the status window as ourself.
	 */
	window->device = &ttyDevice;
	window->parentWindow = 0;
	window->id = dev->windowCount + 1;

	/*
	 * Set some parameters.
	 */
	window->isMapped = TRUE;

	/*
	 * Use the main object as the current and previous.
	 */
	window->currentObject = Life.mainObject;
	window->prevObject = Life.mainObject;

	/*
	 * Set up the default settings.
	 * Since we only allow one window for now, these settings
	 * are the only ones we can really use.
	 */
	settings = &window->settings;
	settings->sharedSettings = Life.topLevelWindowSettings;
	settings->sharedFlags = Life.defaultSharedSettings;
	settings->name[0] = '\0';

	RawCopySettings(Life.topLevelWindowSettings, settings);

	/*
	 * Allocate the window's private data.
	 */
	ttyData = malloc(sizeof(TtyData));

	if (ttyData == 0)
	{
		free(window);

		return 0;
	}

	window->privateData = ttyData;

	memset(ttyData, 0, sizeof(TtyData));

	windowWidth = COLS;
	windowHeight = LINES;

	/*
	 * Calculate and save the window sizes.
	 */
	viewWidth = windowWidth - 1;
	viewHeight = windowHeight - 1;

	window->width = windowWidth;
	window->height = windowHeight;

	window->rows = viewHeight;
	window->cols = viewWidth;
	window->textRows = viewHeight;
	window->textCols = viewWidth;

	ttyData->viewHeight = viewHeight;

	/*
	 * Set the scale value and thus set the viewing area.
	 */
	SetScale(window, GetDefaultScale(settings));

	/*
	 * Allocate the curses windows.
	 */
	ttyData->statWin = newwin(1, windowWidth, 0, 0);
	ttyData->viewWin = newwin(viewHeight, windowWidth, 1, 0);

	/*
	 * Add the window to the table.
	 */
	dev->windowTable[dev->windowCount] = window;
	dev->windowCount++;

	/*
	 * If there is no current window then make this be it.
	 */
	if (Life.currentWindow == 0)
		Life.currentWindow = window;

	/*
	 * Return the opened window.
	 */
	return window;
}


static BOOL
TtyCloseWindow(LifeDevice * dev)
{
	return FALSE;
}


/*
 * Fill in the initial default settings for a window
 * according to whether it is a top level window or a subwindow.
 */
static void
TtyDefaultSettings(LifeDevice * dev, BOOL isTopLevelWindow, Settings * settings)
{
	settings->sharedSettings = Life.topLevelWindowSettings;
	settings->sharedFlags = Life.defaultSharedSettings;
	settings->isVisible = TRUE;
	settings->isShowOutlines = FALSE;
	settings->isShowMyOutline = FALSE;
	settings->isGridEnabled = FALSE;
	settings->isAutoScale = FALSE;
	settings->gridSize = DEFAULT_GRID_SIZE;
	settings->gridChar1 = DEFAULT_GRID_CHAR_1;
	settings->gridChar2 = DEFAULT_GRID_CHAR_2;
	settings->shiftFactor = DEFAULT_SHIFT;
	settings->cellShape = SHAPE_CIRCLE;
	settings->backgroundIndex = COLOR_INDEX_WHITE;
	settings->cellColorIndex = COLOR_INDEX_BLACK;
	settings->markColorIndex = COLOR_INDEX_BLACK;
	settings->tracksColorIndex = COLOR_INDEX_BLACK;
	settings->oldColorIndex = COLOR_INDEX_BLACK;
	settings->textColorIndex = COLOR_INDEX_BLACK;
	settings->viewType = VIEW_TYPE_CURSOR;
	settings->positionType = POSITION_TYPE_SHIFT;
	settings->statusLocation = STATUS_LOCATION_SELF;
	settings->borderWidth = 0;
	settings->defaultScale = 1;
	settings->marginPercentage = 0;

	if (settings->sharedSettings == 0)
		settings->sharedSettings = settings;
}


/*
 * Define the specified color for use later.
 * Returns the index of the color if found or
 * COLOR_INDEX_NONE if not known.  We don't do new
 * colors so only check for the predefined ones.
 */
static ColorIndex
TtyDefineColor(LifeDevice * dev, const char * name)
{
	return FindColorIndex(name);
}


static void
TtyBeep(LifeDevice * dev)
{
	printf("\a");
	fflush(stdout);
}


static void
TtyRedraw(LifeDevice * dev)
{
	wrefresh(curscr);
	fflush(stdout);
}


/*
 * Wait up to the specified number of milliseconds for an
 * input character from stdin or a signal.  A value of 0
 * is no wait.  A value of -1 is an indefinite wait.
 * Returns the reason the wait was terminated.
 */
static EventType
WaitForEvent(int milliSeconds)
{
	int status;
	struct timeval * timeOutPtr = 0;
	struct timeval timeOut;
	fd_set readFds;

	/*
	 * Set the stdin file handle for the select.
	 */
	FD_ZERO(&readFds);
	FD_SET(STDIN_FILENO, &readFds);

	/*
 	 * Set the timeout value if nonnegative.
	 */
	if (milliSeconds >= 0)
	{
		timeOut.tv_sec = milliSeconds / 1000;
		timeOut.tv_usec = (milliSeconds % 1000) * 1000;

		timeOutPtr = &timeOut;
	}

	/*
	 * Perform the select.
	 */
	status = select(STDIN_FILENO + 1, (int *)&readFds, 0, 0, timeOutPtr);

	/*
	 * Check the result.
	 */
	if (status > 0)
		return EVENT_INPUT;

	if (status == 0)
		return EVENT_TIMEOUT;

	if (errno == EINTR)
		return EVENT_SIGNAL;

	return EVENT_FAILED;
}


/*
 * Try to read a character using curses using the given timeout value.
 * A value of -1 means an indefinite wait, a value of 0 means no wait,
 * and positive values means wait that many milliseconds.  Returns
 * the curses value ERR if the timeout occurred.  This routine also
 * checks for a previously saved character and returns that first.
 *
 * This doesn't use the curses timeout function to do the waiting
 * because a SIGWINCH signal doesn't cause the badly designed getch
 * function to return, and we can't resize the window in the signal
 * handler since that isn't safe.
 */
static int
TryReadChar(int milliSeconds)
{
	EventType eventType;

	/*
	 * If there is a saved character then return that.
	 */
	if (isCharSaved)
	{
		isCharSaved = FALSE;

		return savedChar;
	}

	/*
	 * Check the terminal size before waiting.
	 */
	CheckTerminalSize();

	/*
	 * Wait for a timeout, a character, or a signal.
	 */
	eventType = WaitForEvent(milliSeconds);

	/*
	 * Check the terminal size after the wait.
	 */
	CheckTerminalSize();

	/*
	 * If an input character is ready then
	 * ask curses to read it.
	 */
	if (eventType == EVENT_INPUT)
		return getch();

	/*
	 * Return an error result for anything else.
	 */
	return ERR;
}


/*
 * Read the next character from the terminal, waiting if requested.
 * If not waiting and no character is ready, then EOF is returned.
 * When waiting and the read is interrupted, EOF is also returned.
 */
static int
TtyReadChar(LifeDevice * dev, BOOL isWaitRequired)
{
	int timeout;
	int ch;

	/*
	 * Set the timeout to -1 to wait forever or 0 for no wait.
	 */
	timeout = (isWaitRequired ? -1 : 0);

	/*
	 * Try to read the character with the timeout value.
	 */
	ch = TryReadChar(timeout);

	/*
	 * If there was no character ready then return EOF.
	 */
	if (ch == ERR)
		return EOF;

	/*
	 * Translate returns into newlines.
	 */
	if (ch == '\r')
		ch = '\n';

	/*
	 * Return the character.
	 */
	return ch;
}


/*
 * See if input is ready from the terminal.
 * Returns TRUE if so.
 */
static BOOL
TtyIsInputReady(LifeDevice * dev)
{
	/*
	 * Do a non-blocking read of a character.
	 */
	int ch = TryReadChar(0);

	/*
	 * If the curses ERR value was returned then no input is ready.
	 */
	if (ch == ERR)
		return FALSE;

	/*
	 * Save the character for re-reading and return true.
	 */
	savedChar = ch;
	isCharSaved = TRUE;

	return TRUE;
}


/*
 * Delay until the specified number of milliseconds has passed,
 * or until a keystroke is typed.
 */
static void
TtyDelay(LifeDevice * dev, int milliSeconds)
{
	int	ch;

	/*
	 * If there is no sleep time or a character has been
	 * saved then return immediately.
	 */
	if ((milliSeconds <= 0) || isCharSaved)
		return;

	/*
	 * Wait for the timeout or until a character is ready.
	 */
	ch = TryReadChar(milliSeconds);

	/*
	 * If the curses ERR value was returned then the
	 * timeout occurred without any character being read.
	 */
	if (ch == ERR)
		return;

	/*
	 * Save the character for re-reading.
	 */
	savedChar = ch;
	isCharSaved = TRUE;
}


static BOOL
TtyAssign(LifeDevice * dev, Value button, int macro)
{
	return FALSE;
}


static void
TtyWriteAssign(LifeDevice * dev, FILE * fp)
{
}


/*
 * Update the status line for the window.
 * This can show the object information or another text
 * message depending on the window's current status mode.
 * This should only be called after the view has been
 * updated since the status information includes the
 * number of cells which are invisible.
 */
static void
UpdateStatusLine(LifeWindow * window)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	const char * str;

	/*
	 * If the window is not mapped or has its status window
	 * not set to our own window then do nothing.
	 */
	if (!window->isMapped ||
		(GetStatusLocation(&window->settings) != STATUS_LOCATION_SELF))
	{
		return;
	}

	/*
	 * If the view is displaying text then leave the status line.
	 */
	if (window->isTextView)
		return;

	/*
	 * Get the text for the status line.
	 * This depends on the status mode.
	 */
	switch (window->statusMode)
	{
		case STATUS_MODE_NORMAL:
			str = GenerateStatus(window);
			break;

		default:
			str = window->statusBuffer;
			break;
	}

	/*
	 * Put the text into the status subwindow and
	 * remember the length.
	 */
	wmove(ttyData->statWin, 0, 0);
	waddstr(ttyData->statWin, str);
	wclrtoeol(ttyData->statWin);
}


/*
 * Position the cursor to the correct location in the window.
 * This can either be in the status subwindow if input is wanted,
 * or at the cursor location in the view subwindow.
 */
static void
PositionCursor(LifeWindow * window)
{
	const LifeObject * obj = window->currentObject;
	TtyData * ttyData = (TtyData *) window->privateData;
	Scale scale = window->scale;

	/*
	 * If the input is in the status subwindow then put
	 * the cursor there.
	 */
	if (window->statusMode == STATUS_MODE_INPUT)
	{
		wmove(ttyData->statWin, 0, strlen(window->statusBuffer));
		wnoutrefresh(ttyData->statWin);

		return;
	}

	/*
	 * The cursor is to be in the view subwindow.
	 * If the window is in text mode then do nothing.
	 */
	if (window->isTextView)
		return;

	/*
	 * Get the row and column offset from the window edge.
	 */
	NUMBER row = numSub(obj->currentInfo.row, window->minRow);
	NUMBER col = numSub(obj->currentInfo.col, window->minCol);

	row = numMax(row, numZero());
	col = numMax(col, numZero());

	/*
	 * Scale down the offsets by the scale factor.
	 */
	if (scale < 1)
		scale = 1;

	numAddrDivLong(&row, scale);
	numAddrDivLong(&col, scale);

	row = numMin(row, numFromLong(ttyData->viewHeight - 1));
	col = numMin(col, numFromLong(window->width - 1));

	wmove(ttyData->viewWin, numToLong(row), numToLong(col));
	wnoutrefresh(ttyData->viewWin);
}


static void
UpdateView(LifeWindow * window)
{
	/*
	 * If text is being shown in the view window then
	 * do nothing.
	 */
	if (window->isTextView)
		return;

	/*
	 * The view is showing an object.
	 * Draw the view appropriately depending on the scale.
	 */
	if (window->scale <= 1)
		ViewNormal(window);
	else
		ViewScaled(window);
}


static void
UpdateWindow(LifeWindow * window)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	LifeObject * obj = window->currentObject;
	Flags flags = window->updateFlags;

	/*
	 * If the object modify count differs from ours
	 * and generations are not being run then set
	 * the view flag and save the new count.
	 * (While running generations we don't update
	 * unless something else prods us to do so such
	 * as the frequency counter reaching zero.)
	 */
	if ((window->lastChangeCount != obj->changeCount) &&
		(obj->genLeft == 0))
	{
		flags |= (U_VIEW | U_STATUS);
		window->lastChangeCount = obj->changeCount;
	}

	/*
	 * If nothing needs updating then return.
	 */
	if ((flags & (U_VIEW | U_CLEAR | U_STATUS | U_POS)) == 0)
	{
		window->updateFlags = 0;

		return;
	}

	if (flags & (U_VIEW | U_CLEAR))
		UpdateView(window);

	if (flags & U_STATUS)
		UpdateStatusLine(window);

	/*
	 * Make the changes visible.
	 * Update the windows in the appropriate order so that
	 * the cursor will be left in the correct subwindow.
	 */
	if (window->statusMode == STATUS_MODE_INPUT)
	{
		if (flags & (U_VIEW | U_CLEAR))
			wnoutrefresh(ttyData->viewWin);

		if (flags & U_STATUS)
			wnoutrefresh(ttyData->statWin);
	}
	else
	{
		if (flags & U_STATUS)
			wnoutrefresh(ttyData->statWin);

		if (flags & (U_VIEW | U_CLEAR))
			wnoutrefresh(ttyData->viewWin);
	}

	/*
	 * Position the cursor to the correct location.
 	 */
	PositionCursor(window);

	/*
	 * Now finally do the output to the terminal.
	 */
	doupdate();

	window->updateFlags = 0;
}


/*
 * Update all of the windows to show the latest data.
 */
static void
TtyUpdate(LifeDevice * dev)
{
	int index;

	/*
	 * If the terminal size has changed then handle that.
	 */
	CheckTerminalSize();

	/*
	 * Loop over the windows.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * window = dev->windowTable[index];

		UpdateWindow(window);
	}

	/*
	 * Force the output to be seen now.
	 */
	fflush(stdout);
}


static void
TtySetIsVisible(LifeDevice * dev, LifeWindow * window, BOOL isVisible)
{
}


static void
TtySetStatusMode(LifeDevice * dev, StatusMode mode)
{
	LifeWindow * window = Life.currentWindow;
	TtyData * ttyData = (TtyData *) window->privateData;

	window->statusMode = mode;
	window->statusUsed = 0;
	window->statusBuffer[0] = '\0';

	window->updateFlags |= U_STATUS;

	wmove(ttyData->statWin, 0, 0);
	wclrtoeol(ttyData->statWin);
	wmove(ttyData->statWin, 0, 0);
}


/*
 * Add the specified status to the status line.
 */
static void
TtyAddStatusText(LifeDevice * dev, const char * str)
{
	LifeWindow * window = Life.currentWindow;
	TtyData * ttyData = (TtyData *) window->privateData;
	int len;

	/*
	 * Get the length of the string and truncate it
	 * to fit if necessary.
	 */
	len = strlen(str);

	if (window->statusUsed + len >= sizeof(window->statusBuffer))
		len = sizeof(window->statusBuffer) - window->statusUsed - 1;

	if (len <= 0)
		return;

	/*
	 * Position to the new status location and add it.
	 */
	wmove(ttyData->statWin, 0, window->statusUsed);
	waddnstr(ttyData->statWin, str, len);
	window->updateFlags |= U_STATUS;

	/*
	 * Copy the text into the buffer and update the amount used.
	 */
	memcpy(&window->statusBuffer[window->statusUsed], str, len);

	window->statusUsed += len;

	window->statusBuffer[window->statusUsed] = '\0';
}


/*
 * Clear the message in the status window if required.
 */
static void
TtyClearStatusMessage(LifeDevice * dev)
{
	LifeWindow * window = Life.currentWindow;

	if ((window->statusMode == STATUS_MODE_ERROR) ||
		(window->statusMode == STATUS_MODE_MESSAGE))
	{
		TtySetStatusMode(dev, STATUS_MODE_NORMAL);
	}
}


static void
TtySetViewModeIsText(LifeDevice * dev, BOOL isText)
{
	LifeWindow * window = Life.currentWindow;
	TtyData * ttyData = (TtyData *) window->privateData;

	/*
	 * If the window is in the normal view and that is
	 * what we want then do nothing.
	 */
	if (!isText && !window->isTextView)
		return;

	/*
	 * Remember the mode.
	 */
	window->isTextView = isText;
	window->updateFlags |= U_VIEW;

	/*
	 * Clear the window.
	 */
	wmove(ttyData->viewWin, 0, 0);
	wclrtobot(ttyData->viewWin);
	wmove(ttyData->viewWin, 0, 0);
}


static void
TtyAddViewText(LifeDevice * dev, const char * str)
{
	LifeWindow * window = Life.currentWindow;
	TtyData * ttyData = (TtyData *) window->privateData;

	waddstr(ttyData->viewWin, str);

	window->updateFlags |= U_VIEW;
}


/*
 * Show the cells around the cursor normally (scale factor of 1).
 * This shows normal cells as 'O', marked cells as 'o', previously live
 * cells as '+', and empty cells as either ' ' or a user-defined character.
 */
static void
ViewNormal(LifeWindow * window)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	Settings * settings = &window->settings;
	const LifeObject * obj = window->currentObject;
	BOOL isGridEnabled = GetIsGridEnabled(settings);
	NUMBER minRow = window->minRow;
	NUMBER maxRow = window->maxRow;
	NUMBER minCol = window->minCol;
	NUMBER maxCol = window->maxCol;
	int rowCount = 0;
	ViewInfo currentInfo;
	ViewInfo usedInfo;

	/*
	 * Get the data for the current cells of the object for the
	 * region being viewed and remember how many ON cells are seen.
	 */
	currentInfo = CalculateView(obj, VIEW_ID_OBJECT, 1,
		minRow, maxRow, minCol, maxCol);

	window->seeCount = currentInfo.onCount;

	//
	// Get the first nonempty row of the current cells (if any).
	//
	int currentRowIndex = FindNextRowIndex(&currentInfo);

	/*
	 * Get the data for the cells ever used by the object for the
	 * region being viewed, or else an empty info structure if
	 * tracks are currently invisible or not present.
	 * Note: We call the quadcell routine directly since the
	 * used cells always use that algorithm.
	 */
	if (obj->isTracksVisible && !qcIsEmpty(obj->usedQuadCells))
	{
		usedInfo = qcCalculateView(obj->usedQuadCells,
			VIEW_ID_TRACKS, FALSE, minRow, maxRow,
			minCol, maxCol);
	}
	else
	{
		InitEmptyView(&usedInfo);
	}

	//
	// Get the first nonempty row of the used cells (if any).
	//
	int usedRowIndex = FindNextRowIndex(&usedInfo);

	wmove(ttyData->viewWin, 0, 0);

	/*
	 * Scan the cells of the object row by row.
	 * This scan examines the current cells of the object and
	 * the used cells of the object at the same time.
	 */
	while ((currentRowIndex >= 0) || (usedRowIndex >= 0))
	{
		/*
		 * Find the minimum row index of the sets of cells.
		 */
		int rowIndex = PickMinimumIndex(
			currentRowIndex, usedRowIndex, -1, -1);

		//
		// Write out blank lines until the row index is reached.
		//
		while (rowCount < rowIndex)
			ViewBlankNormalRow(window, rowCount++);

		//
		// Examine the cells of the found row to output them.
		//
		ViewNormalRow(window, rowIndex, &currentInfo, &usedInfo);

		rowCount++;

		//
		// Now advance the current and used rows to the next
		// used ones if they were on the row just examined.
		// Note: the checks still work correctly if the appropriate
		// row index was negative when no more rows exist.
		//
		if (rowIndex == currentRowIndex)
			currentRowIndex = FindNextRowIndex(&currentInfo);

		if (rowIndex == usedRowIndex)
			usedRowIndex = FindNextRowIndex(&usedInfo);
	}

	//
	// If the grid is enabled then write blank lines to the end.
	//
	while (isGridEnabled && (rowCount < currentInfo.rowCount))
		ViewBlankNormalRow(window, rowCount++);

	//
	// Clear the rest of the window.
	//
	wclrtobot(ttyData->viewWin);
}


/*
 * Write out a non-empty row to the screen.
 */
static void
ViewNormalRow(LifeWindow * window, int rowIndex,
	ViewInfo * currentInfo, ViewInfo * usedInfo)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	Settings * settings = &window->settings;
	BOOL isGridEnabled = GetIsGridEnabled(settings);
	NUMBER row = numAddLong(window->minRow, rowIndex);
	int colCount = 0;

	//
	// Find the index of the first ON cell in the row (if any) for
	// the current object and the used object if they are on the
	// specified row index.  Note: These calls work correctly even
	// if there were no more rows left to be examined.
	//
	int currentColIndex = -1;
	int usedColIndex = -1;

	if (currentInfo->rowIndex == rowIndex)
		currentColIndex = FindNextColIndex(currentInfo);

	if (usedInfo->rowIndex == rowIndex)
		usedColIndex = FindNextColIndex(usedInfo);

	/*
	 * Keep looping while there are columns of interest in any of
	 * the sets of cells.
	 */
	while ((currentColIndex >= 0) || (usedColIndex >= 0))
	{
		int ch = ' ';

		/*
		 * Find the minimum column index of the sets of cells.
		 */
		int colIndex = PickMinimumIndex(currentColIndex,
			usedColIndex, -1, -1);

		//
		// While we haven't reached the required column output
		// spaces or the grid character.
		//
		while (colCount < colIndex)
		{
			NUMBER col = numAddLong(window->minCol, colCount++);
			waddch(ttyData->viewWin, GridChar(window, row, col));
		}

		if (colIndex == usedColIndex)
			ch = '+';

		if (colIndex == currentColIndex)
		{
			ch = 'O';

			if (GetAuxViewData(currentInfo, rowIndex, colIndex) < 0)
				ch = 'o';
		}

		waddch(ttyData->viewWin, ch);

		colCount++;

		//
		// Now advance the current and used columns to the next
		// used ones on the row.  Note: the checks still work
		// correctly if the appropriate column index was negative
		// when no more columns exist.
		//
		if ((currentInfo->rowIndex == rowIndex) &&
			(colIndex == currentColIndex))
		{
			currentColIndex = FindNextColIndex(currentInfo);
		}

		if ((usedInfo->rowIndex == rowIndex) &&
			(colIndex == usedColIndex))
		{
			usedColIndex = FindNextColIndex(usedInfo);
		}
	}

	//
	// Add possible trailing grid characters to the rest of the line.
	//
	while (isGridEnabled && (colCount < currentInfo->colCount))
	{
		NUMBER col = numAddLong(window->minCol, colCount++);
		waddch(ttyData->viewWin, GridChar(window, row, col));
	}

	waddch(ttyData->viewWin, '\n');
}


/*
 * Write out a blank line to the screen.
 * This is just a newline character unless the grid is enabled,
 * in case case we must output the grid characters first.
 */
static void
ViewBlankNormalRow(LifeWindow * window, int rowIndex)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	Settings * settings = &window->settings;
	BOOL isGridEnabled = GetIsGridEnabled(settings);

	/*
	 * If the grid is enabled then fill the whole row with
	 * the grid characters
	 */
	if (isGridEnabled)
	{
		NUMBER row = numAddLong(window->minRow, rowIndex);
		NUMBER col = window->minCol;

		while (numIsLessOrEqual(col, window->maxCol))
		{
			waddch(ttyData->viewWin, GridChar(window, row, col));
			numAddrInc(&col);
		}
	}

	waddch(ttyData->viewWin, '\n');
}


/*
 * Show the view around the cursor with an arbitrary scale factor.
 * When in this mode, characters from 1 to 9 (or * if 10 or more)
 * are used to indicate how many cells are in each n by n square.
 * This routine doesn't currently show tracked cells.
 */
static void
ViewScaled(LifeWindow * window)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	Settings * settings = &window->settings;
	const LifeObject * obj = window->currentObject;
	BOOL isGridEnabled = GetIsGridEnabled(settings);
	int scale = window->scale;
	NUMBER minRow = window->minRow;
	NUMBER maxRow = window->maxRow;
	NUMBER minCol = window->minCol;
	NUMBER maxCol = window->maxCol;
	int rowCount = 0;
	ViewInfo currentInfo;

	if (scale < 1)
		scale = 1;

	/*
	 * Get the data for the current cells of the object for the
	 * region being viewed and remember how many ON cells are seen.
	 */
	currentInfo = CalculateView(obj, VIEW_ID_OBJECT, scale,
		minRow, maxRow, minCol, maxCol);

	window->seeCount = currentInfo.onCount;

	//
	// Get the first nonempty row of the current cells (if any).
	//
	int currentRowIndex = FindNextRowIndex(&currentInfo);

	wmove(ttyData->viewWin, 0, 0);

	/*
	 * Scan the cells of the object row by row.
	 */
	while (currentRowIndex >= 0)
	{
		int rowIndex = currentRowIndex;

		//
		// Write out blank lines until the row index is reached.
		//
		while (rowCount < rowIndex)
			ViewBlankScaledRow(window, rowCount++);

		//
		// Examine the cells of the found row to output them.
		//
		ViewScaledRow(window, rowIndex, &currentInfo);

		rowCount++;

		//
		// Now advance the current row to the next used one.
		// Note: the checks still work correctly if the appropriate
		// row index was negative when no more rows exist.
		//
		if (rowIndex == currentRowIndex)
			currentRowIndex = FindNextRowIndex(&currentInfo);
	}

	//
	// If the grid is enabled then write blank lines to the end.
	//
	while (isGridEnabled && (rowCount < currentInfo.rowCount))
		ViewBlankScaledRow(window, rowCount++);

	//
	// Clear the rest of the window.
	//
	wclrtobot(ttyData->viewWin);
}


/*
 * Write out a non-empty scaled row to the screen.
 */
static void
ViewScaledRow(LifeWindow * window, int rowIndex, ViewInfo * currentInfo)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	Settings * settings = &window->settings;
	BOOL isGridEnabled = GetIsGridEnabled(settings);
	int gridChar = GetGridChar1(settings);
	int colCount = 0;

	if (!isGridEnabled)
		gridChar = ' ';

	//
	// Find the index of the first ON cell in the row (if any) for
	// the current object.  Note: These calls work correctly even
	// if there were no more rows left to be examined.
	//
	int currentColIndex = -1;

	if (currentInfo->rowIndex == rowIndex)
		currentColIndex = FindNextColIndex(currentInfo);

	/*
	 * Keep looping while there are columns of interest.
	 */
	while (currentColIndex >= 0)
	{
		int ch = gridChar;
		int colIndex = currentColIndex;

		//
		// While we haven't reached the required column output
		// spaces or the grid character.
		//
		while (colCount < colIndex)
		{
			waddch(ttyData->viewWin, gridChar);
			colCount++;
		}

		//
		// Determine what character to write here depending on
		// the count of ON cells in the auxillary data element.
		//
		int auxData = GetAuxViewData(currentInfo, rowIndex, colIndex);

		if (auxData < 0)
			auxData = -auxData;

		if (auxData > 9)
			ch = '*';
		else
			ch = '0' + auxData;

		waddch(ttyData->viewWin, ch);

		colCount++;

		//
		// Now advance the current column to the next used ones
		// on the row.  Note: the checks still work
		// correctly if the appropriate column index was negative
		// when no more columns exist.
		//
		if ((currentInfo->rowIndex == rowIndex) &&
			(colIndex == currentColIndex))
		{
			currentColIndex = FindNextColIndex(currentInfo);
		}
	}

	//
	// Add possible trailing grid characters to the rest of the line.
	//
	while (isGridEnabled && (colCount < currentInfo->colCount))
	{
		waddch(ttyData->viewWin, gridChar);
		colCount++;
	}

	waddch(ttyData->viewWin, '\n');
}


/*
 * Write out a blank scaled row.
 * This doesn't do much unless the grid is set.
 */
static void
ViewBlankScaledRow(LifeWindow * window, int rowIndex)
{
	TtyData * ttyData = (TtyData *) window->privateData;
	Settings * settings = &window->settings;
	BOOL isGridEnabled = GetIsGridEnabled(settings);

	if (isGridEnabled)
	{
		int gridChar = GetGridChar1(settings);
		int colCount = 0;

		while (++colCount < window->width)
			waddch(ttyData->viewWin, gridChar);
	}

	waddch(ttyData->viewWin, '\n');
}


/*
 * Return the appropriate grid character for the specified row and column
 * number of the object.  Returns a space character if the grid is not enabled.
 */
static int
GridChar(const LifeWindow * window, NUMBER row, NUMBER col)
{
	const Settings * settings = &window->settings;
	const LifeObject * obj = window->currentObject;

	if (!GetIsGridEnabled(settings))
		return ' ';

	NUMBER gridRow = numSub(row, obj->originInfo.row);
	NUMBER gridCol = numSub(col, obj->originInfo.col);

	if (GridMod(window, gridRow) && GridMod(window, gridCol))
		return GetGridChar1(settings);

	return GetGridChar2(settings);
}

/* END CODE */
