/*
 * Copyright (c) 2017 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "life.h"


#define	MAX_FUNC_ARGS	20
#define	MAX_NAME_SIZE	16


/*
 * Built-in functions.
 */
typedef	NUMBER	(*Func)(const NUMBER *, int);


typedef	struct
{
	const char *	name;
	Func		func;
	int		minArgs;
	int		maxArgs;
} FuncInfo;


static	NUMBER	FuncGcd(const NUMBER *, int);
static	NUMBER	FuncLcm(const NUMBER *, int);
static	NUMBER	FuncMin(const NUMBER *, int);
static	NUMBER	FuncMax(const NUMBER *, int);
static	NUMBER	FuncAbs(const NUMBER *, int);
static	NUMBER	FuncSign(const NUMBER *, int);
static	NUMBER	FuncFactor(const NUMBER *, int);


/*
 * Table of built-in functions.
 */
static const FuncInfo	func_table[] =
{
	{"gcd", FuncGcd, 2, MAX_FUNC_ARGS},
	{"lcm", FuncLcm, 2, MAX_FUNC_ARGS},
	{"min", FuncMin, 2, MAX_FUNC_ARGS},
	{"max", FuncMax, 2, MAX_FUNC_ARGS},
	{"abs", FuncAbs, 1, 1},
	{"sign", FuncSign, 1, 1},
	{"factor", FuncFactor, 1, 1},
	{0, 0}
};


/*
 * Variable enum.
 */
typedef	enum Vars
{
	VarMinx, VarMiny, VarMaxx, VarMaxy, VarCells, VarCx, VarCy, VarOx,
	VarOy, VarVminx, VarVmaxx, VarVminy, VarVmaxy, VarVcells, VarVx,
	VarVy, VarSminx, VarSmaxx, VarSminy, VarSmaxy, VarMcells, VarGen,
	VarScale, VarFreq, VarBorn, VarDied, VarGrowth, VarDefscale,
	VarDeffreq, VarGenleft, VarDelay, VarMinCell, VarMaxCell, VarWidth,
	VarHeight, VarDragMinX, VarDragMinY, VarDragMaxX, VarDragMaxY,
	VarGridSize, VarEndlist
} Vars;


/*
 * Table of multi-character variables.
 * This is arranged in listing order (one item per line).
 */
typedef	struct
{
	const char * name;		/* name of variable */
	Vars type;			/* variable id */
} VarTab;


static const VarTab	vartab[] =
{
	{"width",	VarWidth},
	{"height",	VarHeight},
	{"minx",	VarMinx},
	{"maxx",	VarMaxx},
	{"miny",	VarMiny},
	{"maxy",	VarMaxy},
	{"vminx",	VarVminx},
	{"vmaxx",	VarVmaxx},
	{"vminy",	VarVminy},
	{"vmaxy",	VarVmaxy},
	{"sminx",	VarSminx},
	{"smaxx",	VarSmaxx},
	{"sminy",	VarSminy},
	{"smaxy",	VarSmaxy},
	{"cells",	VarCells},
	{"vcells",	VarVcells},
	{"mcells",	VarMcells},
	{"cx",		VarCx},
	{"cy",		VarCy},
	{"vx",		VarVx},
	{"vy",		VarVy},
	{"ox",		VarOx},
	{"oy",		VarOy},
	{"gen",		VarGen},
	{"genLeft",	VarGenleft},
	{"freq",	VarFreq},
	{"born",	VarBorn},
	{"died",	VarDied},
	{"growth",	VarGrowth},
	{"delay",	VarDelay},
	{"scale",	VarScale},
	{"dscale",	VarDefscale},
	{"dfreq",	VarDeffreq},
	{"mincell",	VarMinCell},
	{"maxcell",	VarMaxCell},
	{"dminx",	VarDragMinX},
	{"dmaxx",	VarDragMaxX},
	{"dminy",	VarDragMinY},
	{"dmaxy",	VarDragMaxY},
	{"grid",	VarGridSize},
	{0,		VarEndlist}
};


static	const char * curCp;	/* current character to parse */
static	NUMBER	lowerVars[26];	/* lower case single-char variable values */
static	NUMBER	upperVars[26];	/* upper case single-char variable values */


/*
 * Local procedures.
 */
static	NUMBER	GetVariableByType(Vars);
static	NUMBER	GetWhereVariable(const char *);
static	NUMBER	ParseExpression(void);
static	NUMBER	ParseRelation(void);
static	NUMBER	ParseSum(void);
static	NUMBER	ParseProduct(void);
static	NUMBER	ParseOr(void);
static	NUMBER	ParseAnd(void);
static	NUMBER	ParsePower(void);
static	NUMBER	ParseTerm(void);
static	NUMBER	ParseSymbol(void);
static	NUMBER	ParseFunction(void);


/*
 * Return the value of a multiple character variable name.  This can be
 * either a single character name, or else one of a fixed set of multi-
 * character names.  All variable names must start with either a letter
 * or a dollar sign followed by a letter.
 */
NUMBER
GetVariable(const char * cp)
{
	const VarTab *	vp;

	if (*cp == '$')
		cp++;

	if (cp[1] == '\0')
		return GetVariable1(*cp);

	if (!isLower(*cp) && !isUpper(*cp))
		Error("Bad variable name");

	for (vp = vartab; ; vp++)
	{
		if (vp->name == 0)
		{
			if (*cp == 'w')
				return GetWhereVariable(cp + 1);

			Error("Unknown variable");
		}

		if (strcmp(vp->name, cp) == 0)
			break;
	}

	return GetVariableByType(vp->type);
}


/*
 * Return the value of a variable given its enum value.
 */
static NUMBER
GetVariableByType(Vars type)
{
	const LifeWindow * window = Life.currentWindow;
	LifeObject * obj = Life.currentObject;
	const NUMBER * ptr = 0;
	const NUMBER * sptr = 0;
	int sign = 1;
	NUMBER value = numZero();
	NUMBER minRow;
	NUMBER maxRow;
	NUMBER minCol;
	NUMBER maxCol;

	switch (type)
	{
		case VarDragMinX:
			value = obj->dragMinCol;

			break;

		case VarDragMaxX:
			value = obj->dragMaxCol;

			break;

		case VarDragMinY:
			value = obj->dragMinRow;

			break;

		case VarDragMaxY:
			value = obj->dragMaxRow;

			break;

		case VarCx:
			value = obj->currentInfo.col;

			break;

		case VarCy:
			value = numNeg(obj->currentInfo.row);

			break;

		case VarOx:
			value = obj->originInfo.col;

			break;

		case VarOy:
			value = numNeg(obj->originInfo.row);

			break;

		case VarVx:
			value = numAverage(window->minCol, window->maxCol);

			break;

		case VarVy:
			value = numNeg(numAverage(window->minRow, window->maxRow));

			break;

		case VarVminx:
			value = window->minCol;

			break;

		case VarVmaxx:
			value = window->maxCol;

			break;

		case VarVminy:
			value = numNeg(window->maxRow);

			break;

		case VarVmaxy:
			value = numNeg(window->minRow);

			break;

		case VarVcells:
			value = CountRegion(obj,
				window->minRow, window->maxRow,
				window->minCol, window->maxCol);

			break;

		case VarCells:
			value = CalculateCellCount(obj);

			break;

		case VarMinCell:
			value = obj->minCount;

			break;

		case VarMaxCell:
			value = obj->maxCount;

			break;

		case VarGen:
			value = numFromLong(obj->gen);

			break;

		case VarGenleft:
			value = numFromLong(obj->genLeft);

			break;

		case VarDelay:
			value = numFromLong(Life.delayTime);

			break;

		case VarBorn:
			value = obj->born;

			break;

		case VarDied:
			value = obj->died;

			break;

		case VarGrowth:
			value = obj->growth;

			break;

		case VarFreq:
			value = numFromLong(obj->frequency);

			break;

		case VarScale:
			value = numFromLong(window->scale);

			break;

		case VarDeffreq:
			value = numFromLong(Life.defaultFrequency);

			break;

		case VarDefscale:
			value = numFromLong(GetDefaultScale(&window->settings));

			break;

		case VarGridSize:
			value = numFromLong(GetGridSize(&window->settings));

			break;

		case VarMinx:
			ptr = &minCol;

			break;

		case VarMaxx:
			ptr = &maxCol;

			break;

		case VarMiny:
			ptr = &maxRow;
			sign = -1;

			break;

		case VarMaxy:
			ptr = &minRow;
			sign = -1;

			break;

		case VarMcells:
			value = CountMarks(obj);

			break;

		case VarSminx:
			sptr = &minCol;

			break;

		case VarSmaxx:
			sptr = &maxCol;

			break;

		case VarSminy:
			sptr = &maxRow;
			sign = -1;

			break;

		case VarSmaxy:
			sptr = &minRow;
			sign = -1;

			break;

		case VarWidth:
			if (MinMax(obj, &minRow, &maxRow, &minCol, &maxCol))
				value = numRange(minCol, maxCol);

			break;

		case VarHeight:
			if (MinMax(obj, &minRow, &maxRow, &minCol, &maxCol))
				value = numRange(minRow, maxRow);

			break;

		default:
			return numZero();
	}

	/*
	 * Call proper minmax routines if we need to
	 */
	if (ptr && MinMax(obj, &minRow, &maxRow, &minCol, &maxCol))
		value = *ptr;

	if (sptr && MarkMinMax(obj,
		&minRow, &maxRow, &minCol, &maxCol))
	{
		value = *sptr;
	}

	if (sign < 0)
		value = numNeg(value);

	return value;
}


/*
 * Get the value of a 'where' variable.
 * The value is either the x or the y coordinate, depending on the argument.
 * The incoming string is the where variable letter followed by x or y.
 */
static NUMBER
GetWhereVariable(const char * cp)
{
	int ch = *cp++;

	if ((ch < 'a') || (ch >= 'z'))
		Error("Bad where variable name");

	if (((*cp != 'x') && (*cp != 'y')) || (cp[1] != '\0'))
		Error("Bad where coordinate selector");

	if (*cp == 'x')
		return Life.currentObject->whereInfo[ch - 'a'].col;

	return numNeg(Life.currentObject->whereInfo[ch - 'a'].row);
}


/*
 * Return the value of a single character variable name (a-z or A-Z).
 */
NUMBER
GetVariable1(int ch)
{
	if (isLower(ch))
		return lowerVars[ch - 'a'];

	if (isUpper(ch))
		return lowerVars[ch - 'A'];

	Error("Bad variable name");

	return numZero();
}


/*
 * Set the value of a variable name.  Multi-character names cannot be set.
 */
void
SetVariable(const char * cp, NUMBER value)
{
	if (*cp == '$')
		cp++;

	if (cp[1] != '\0')
		Error("Cannot set multi-character variables");

	SetVariable1(*cp, value);
}


/*
 * Set the value of a single-character variable (a-z or A-Z).
 */
void
SetVariable1(int ch, NUMBER value)
{
	if (isLower(ch))
	{
		lowerVars[ch - 'a'] = value;

		return;
	}

	if (isUpper(ch))
	{
		lowerVars[ch - 'A'] = value;

		return;
	}

	Error("Bad variable name");
}


/*
 * Display the current values of the variables.  Show all multi-character
 * variable values, and those single character variables which have a
 * nonzero value.
 */
void
ListVariables(void)
{
	LifeObject * obj = Life.currentObject;
	int count;
	const VarTab * vp;
	NUMBER * var;
	int i;

	ShowHelp("Variable names and values:\n\n");

	/*
	 * Show all fixed variables, even if their value is zero.
	 */
	for (vp = vartab; vp->name; vp++)
	{
		FormatHelp("%-8s %s\n", vp->name,
			numToString(GetVariableByType(vp->type)));
	}

	/*
	 * Show all where locations which have a nonzero coordinate.
	 */
	count = 0;

	for (i = 0; i < 26; i++)
	{
		LocationInfo * info = &obj->whereInfo[i];

		if (numIsNonZero(info->col) || numIsNonZero(info->row))
		{
			if (count++ == 0)
				ShowHelp("\n");

			FormatHelp("w%cx      %s\n", 'a' + i,
				numToString(info->col));

			FormatHelp("w%cy      %s\n", 'a' + i,
				numToString(info->row));
		}
	}

	/*
	 * Show the upper case and lower case single character variables
	 * which are nonzero.
	 */
	count = 0;

	for (var = upperVars; var < &upperVars[26]; var++)
	{
		if (numIsZero(*var))
			continue;

		if (count++ == 0)
			ShowHelp("\n");

		FormatHelp("%c        %s\n", 'A' + (var - upperVars),
			numToString(*var));
	}

	for (var = lowerVars; var < &lowerVars[26]; var++)
	{
		if (numIsZero(*var))
			continue;

		if (count++ == 0)
			ShowHelp("\n");

		FormatHelp("%c        %s\n", 'a' + (var - lowerVars),
			numToString(*var));
	}

	ShowHelp("\n\nMany names starting with 'v' refer to visible cells\n");
	ShowHelp("Many names starting with 's' refer to selected cells\n");
	ShowHelp("Also, 'where' variable coordinates may be referenced as\n");
	ShowHelp("w<ch>x or w<ch>y, where <ch> is the variable name.\n");

	EndHelp(TRUE);
}


/*
 * Evaluate an expression and return its value.  The expression can contain
 * numbers, variables, the normal arithmetic operators, and parenthesized
 * expressions.  The usual precedence rules are used.
 */
NUMBER
GetExpression(const char * str)
{
	NUMBER value;

	while isBlank(*str)
		str++;

	if (*str == '\0')
		Error("Null expression");

	curCp = str;
	value = ParseExpression();

	if (*curCp != '\0')
		Error("Bad expression");

	return value;
}


/*
 * Evaluate an expression and return its value.  The expression can contain
 * numbers, variables, the normal arithmetic operators, and parenthesized
 * expressions.  The usual precedence rules are used.  The size of the
 * result must fit within a Value.
 */
Value
GetSmallExpression(const char * str)
{
	NUMBER value = GetExpression(str);

	if (numIsLong(value))
		return numToLong(value);

	Error("Expression value is too large");

	return 0;
}


/*
 * Parse a top level expression.
 */
static NUMBER
ParseExpression(void)
{
	return ParseRelation();
}


/*
 * Parse a relation between two sums.
 * This is equality and inequality.
 */
static NUMBER
ParseRelation(void)
{
	NUMBER	left = ParseSum();

	while (isBlank(*curCp))
		curCp++;

	switch (*curCp)
	{
		case '<':
			if (curCp[1] == '=')
			{
				curCp += 2;

				return numFromLong(
					numIsLessOrEqual(left, ParseSum()));
			}

			curCp++;

			return numFromLong(numIsLess(left, ParseSum()));


		case '>':
			if (curCp[1] == '=')
			{
				curCp += 2;

				return numFromLong(
					numIsGreaterOrEqual(left, ParseSum()));
			}

			curCp++;

			return numFromLong(numIsGreater(left, ParseSum()));


		case '!':
			if (curCp[1] == '=')
			{
				curCp += 2;

				return numFromLong(!numIsEqual(left, ParseSum()));
			}

			break;

		case '=':
			if (curCp[1] == '=')
			{
				curCp += 2;

				return numFromLong(numIsEqual(left, ParseSum()));
			}

			break;

		default:
			break;
	}

	return left;
}


/*
 * Parse the sum of products.
 */
static NUMBER
ParseSum(void)
{
	NUMBER value;

	while (isBlank(*curCp))
		curCp++;

	switch (*curCp)
	{
		case '-':
			curCp++;
			value = numNeg(ParseProduct());

			break;

		case '+':
			curCp++;
			/* proceed into default case */

		default:
			value = ParseProduct();
	}

	while (TRUE)
	{
		switch (*curCp++)
		{
			case '+':		/* sum of products */
				numAddrAdd(&value, ParseProduct());

				continue;

			case '-':		/* difference of products */
				numAddrSub(&value,  ParseProduct());

				continue;

			case ' ':		/* space */
			case '\t':
				continue;

			default:		/* end of sum */
				curCp--;

				return value;
		}
	}
}


/* 
 * Parse the product of OR'd values.
 */
static NUMBER
ParseProduct(void)
{
	NUMBER value;
	NUMBER value2;

	value = ParseOr();

	while (TRUE)
	{
		switch (*curCp++)
		{
			case '*':
				numAddrMul(&value, ParseOr());

				continue;

			case '/':
				value2 = ParseOr();

				if (numIsZero(value2))
					Error("Division by zero");

				numAddrDiv(&value,  value2);

				continue;

			case '%':
				value2 = ParseOr();

				if (numIsZero(value2))
					Error("Division by zero");

				numAddrMod(&value, value2);

				continue;

			case ' ':
			case '\t':
				continue;

			default:
				curCp--;

				return value;
		}
	}
}


/*
 * Parse a list of values OR'd together.
 */
static NUMBER
ParseOr(void)
{
	NUMBER value = ParseAnd();

	while (TRUE)
	{
		while (isBlank(*curCp))
			curCp++;

		if (*curCp != '|')
			break;

		curCp++;

		numAddrOr(&value, ParseAnd());
	}

	return value;
}


/*
 * Parse a list of values AND'd together.
 */
static NUMBER
ParseAnd(void)
{
	NUMBER value = ParsePower();

	while (TRUE)
	{
		while (isBlank(*curCp))
			curCp++;

		if (*curCp != '&')
			break;

		curCp++;

		numAddrAnd(&value,  ParsePower());
	}

	return value;
}


/*
 * Parse a power expression (e.g., 2^10).
 */
static NUMBER
ParsePower(void)
{
	NUMBER value = ParseTerm();

	while (isBlank(*curCp))
		curCp++;

	if (*curCp != '^')
		return value;

	curCp++;

	NUMBER value2 = ParseTerm();
	NUMBER result = numPower(value, value2);

	if (numIsZero(result) && !numIsZero(value) && !numIsNeg(value2))
		Error("Very large power");

	return result;
}


/*
 * Parse a single term.
 */
static NUMBER
ParseTerm(void)
{
	NUMBER value;
	int ch;

	while (isBlank(*curCp))
		curCp++;

	ch = *curCp;

	/*
	 * If a digit is seen, then read a number.
	 */
	if (isDigit(ch))
	{
		value = numZero();

		do
		{
			numAddrMulLong(&value, 10);
			numAddrAddLong(&value, *curCp++ - '0');
		}
		while (isDigit(*curCp));

		return value;
	}

	/*
	 * If a parenthesis is seen, then read in a parenthesized expression.
	 */
	if (ch == '(')
	{
		curCp++;

		while (isBlank(*curCp))
			curCp++;

		if (*curCp == ')')
			Error("Null expression");

		value = ParseExpression();

		while (isBlank(*curCp))
			curCp++;

		if (*curCp != ')')
			Error("Bad expression");

		curCp++;

		return value;
	}

	if (ch == ')')
		Error("Bad expression");

	return ParseSymbol();
}


/*
 * Parse a symbol name and return its value.
 */
static NUMBER
ParseSymbol(void)
{
	const char * cp;
	int len;
	char name[MAX_NAME_SIZE];

	/*
	 * Look for a function call.
	 * This is a name followed by a parenthesis.
	 */
	while (isBlank(*curCp))
		curCp++;

	cp = curCp;

	if (!isLower(*cp) && !isUpper(*cp))
		Error("Bad expression");

	while (isLower(*cp) || isUpper(*cp) || isDigit(*cp))
		cp++;

	while (isBlank(*cp))
		cp++;

	if (*cp == '(')
		return ParseFunction();

	/*
	 * Not a function, so it must be a variable name.
	 */
	cp = curCp;

	if (*cp == '$')
		cp++;

	while (isLower(*cp) || isUpper(*cp) || isDigit(*cp))
    		cp++;

	len = cp - curCp;
	
	if (len >= sizeof(name))
		Error("Variable name too long");

	memcpy(name, curCp, len);
	name[len] = '\0';

	curCp = cp;

	return GetVariable(name);
}


/*
 * Parse a built-in function and return its value.
 */
static NUMBER
ParseFunction(void)
{
	const char * cp;
	const FuncInfo * func;
	int argCount;
	int len;
	char name[MAX_NAME_SIZE];
	NUMBER args[MAX_FUNC_ARGS];

	cp = curCp;

	while (isLower(*cp) || isUpper(*cp) || isDigit(*cp))
    		cp++;

	len = cp - curCp;

	if (len >= sizeof(name))
		Error("Function name too long");

	memcpy(name, curCp, len);
	name[len] = '\0';

	curCp = cp;

	/*
	 * Search for the function name or the end of the table.
	 */
	func = func_table;

	while ((func->name != 0) && (strcmp(func->name, name) != 0))
		func++;

	if (func->name == 0)
		Error("Unknown function");

	/*
	 * Collect the parenthesis introducing the arguments.
	 */
	while (isBlank(*cp))
		cp++;

	if (*cp++ != '(')
		Error("Missing parenthesis for function");

	curCp = cp;

	/*
	 * Now collect all of the comma separated arguments.
	 */
	argCount = 0;

	while (isBlank(*curCp))
		curCp++;

	if (*curCp != ')')
	{
		for (;;)
		{
			if (argCount >= MAX_FUNC_ARGS)
				Error("Too many function arguments");

			args[argCount++] = ParseExpression();

			while (isBlank(*curCp))
				curCp++;

			if (*curCp != ',')
				break;

			curCp++;
		}
	}

	if (*curCp++ != ')')
		Error("Missing right parenthesis for function");

	/*
	 * Verify the number of arguments, and then call the function.
	 */
	if (argCount < func->minArgs)
		Error("Too few arguments for function");

	if (argCount > func->maxArgs)
		Error("Too many arguments for function");

	return func->func(args, argCount);
}


/*
 * Built-in functions.
 */
static NUMBER
FuncGcd(const NUMBER * args, int argCount)
{
	NUMBER value;
	NUMBER newArg;
	NUMBER temp;

	value = *args++;
	argCount--;

	while (argCount-- > 0)
	{
		newArg = *args++;

		while (numIsNonZero(newArg))
		{
			temp = numMod(value, newArg);
			value = newArg;
			newArg = temp;
		}
	}

	return value;
}


static NUMBER
FuncLcm(const NUMBER * args, int argCount)
{
	NUMBER values[2];

	values[0] = *args++;
	argCount--;

	while (argCount-- > 0)
	{
		values[1] = *args++;

		numAddrMul(&values[0], numDiv(values[1], FuncGcd(values, 2)));
	}

	return values[0];
}


static NUMBER
FuncMin(const NUMBER * args, int argCount)
{
	NUMBER value;

	value = *args++;
	argCount--;

	while (argCount-- > 0)
		value = numMin(value, *args++);

	return value;
}


static NUMBER
FuncMax(const NUMBER * args, int argCount)
{
	NUMBER value;

	value = *args++;
	argCount--;

	while (argCount-- > 0)
		value = numMax(value, *args++);

	return value;
}


static NUMBER
FuncAbs(const NUMBER * args, int argCount)
{
	return numAbs(*args);
}


static NUMBER
FuncSign(const NUMBER * args, int argCount)
{
	return numFromLong(numSign(*args));
}


static NUMBER
FuncFactor(const NUMBER * args, int argCount)
{
	NUMBER value = *args;

	value = numAbs(value);

	if (numCmpLong(value, 4) < 0)
		return numOne();

	if (numIsZero(numAndLong(value, 0x01)))
		return numTwo();

	/*
	 * Try dividing by odd numbers.
	 * TODO: This is wrong and slow for large numbers.
	 */
	NUMBER test = numFromLong(3);

	while (numIsLessOrEqual(numMul(test, test), value))
	{
		if (numIsZero(numMod(value, test)))
			return test;

		numAddrAddLong(&test, 2);
	}

	return numOne();
}

/* END CODE */
