/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Generic code associated with viewing of objects.
 * These routines are called by the device functions.
 */

#include "life.h"

static const char * GetDistanceInfo(LifeObject * obj, const LocationInfo * info);


/*
 * Return the color name that was defined for a color index.
 * Returns 0 if the color index is not defined.
 */
const char *
GetColorName(ColorIndex index)
{
	if ((index < 0) || (index >= Life.colorCount))
		return 0;

	return Life.colorTable[index];
}


/*
 * Return the color index of the specified previously
 * defined color name.  Returns the found color index or
 * COLOR_INDEX_NONE if the color name had not been defined.
 */
ColorIndex
FindColorIndex(const char * colorName)
{
	ColorIndex index;

	for (index = 0; index < Life.colorCount; index++)
	{
		if (strcmp(colorName, Life.colorTable[index]) == 0)
			return index;
	}

	return COLOR_INDEX_NONE;
}


/*
 * Define a new color to be used and return its index.
 * If called for a color name already defined, the already
 * defined index is returned.  Returns the color index
 * or COLOR_INDEX_NONE if the color could not be defined.
 */
ColorIndex
DefineColor(const char * colorName)
{
	/*
	 * See if the color is already defined.
	 * If so then return that index.
	 */
	ColorIndex index = FindColorIndex(colorName);

	if (index >= 0)
		return index;

	/*
	 * A new color needs to be allocated.
	 * Fail if the color table is full.
	 */
	if (Life.colorCount >= MAX_COLORS)
		return COLOR_INDEX_NONE;

	/*
	 * Ask the device to define the new color.
	 * It will call AddColorForDevice if successful.
	 */
	return Life.device->defineColor(Life.device, colorName);
}


/*
 * Add a color to the color table and return its index.
 * This is called by the device when a color is successfully
 * defined.  This doesn't check for duplicates.
 * Returns the color index or COLOR_INDEX_NONE on an error.
 */
ColorIndex
AddColorForDevice(const char * colorName)
{
	char * copiedName;

	/*
	 * Fail if the color table is full.
	 */
	if (Life.colorCount >= MAX_COLORS)
		return COLOR_INDEX_NONE;

	/*
	 * Copy the color name since we need to keep it.
	 */
	copiedName = strdup(colorName);

	if (copiedName == 0)
		return COLOR_INDEX_NONE;

	/*
	 * Save the color name in the color table.
	 */
	Life.colorTable[Life.colorCount] = copiedName;

	/*
	 * Return the allocated index and increment it.
	 */
	return Life.colorCount++;
}


/*
 * Set the appropriate update flags for the window according
 * to the settings which have been changed since the last update.
 * The private changed settings flags are then cleared.
 */
void
ApplyWindowSettingsChangedFlags(LifeWindow * window)
{
	Settings * settings = &window->settings;

	/*
	 * Get the settings which have changed privately.
	 */
	SettingsFlags flags = settings->changedFlags;

	/*
	 * Add in the shared settings which have changed.
	 */
	flags |= (settings->sharedSettings->changedFlags & settings->sharedFlags);

	/*
	 * Test the settings to see what update flags are needed.
	 */
	if (flags & SETTINGS_MASK_STATUS)
		window->updateFlags |= U_STATUS;

	if (flags & SETTINGS_MASK_REDRAW)
		window->updateFlags |= U_REDRAW;

	if (flags & SETTINGS_MASK_CONFIGURE)
		window->updateFlags |= (U_CONFIG | U_REDRAW);

	/*
	 * Clear the private changed settings flags for the window.
	 * The shared changed flags will be cleared later.
	 */
	settings->changedFlags = SETTINGS_MASK_NONE;
}


/*
 * Set the location to actually display the status messages
 * for a window.
 */
void
SetWindowStatusLocation(LifeWindow * window, StatusLocation location)
{
	SetStatusLocation(&window->settings, location);
}


/*
 * Set the autoscale flag on or off for a window.
 */
void
SetWindowIsAutoScale(LifeWindow * window, BOOL isAutoScale)
{
	SetIsAutoScale(&window->settings, isAutoScale);
}


/*
 * Set whether the grid is enabled for a window.
 */
void
SetWindowIsGridEnabled(LifeWindow * window, BOOL isGridEnabled)
{
	SetIsGridEnabled(&window->settings, isGridEnabled);
}


/*
 * Set the grid size for a window.
 */
void
SetWindowGridSize(LifeWindow * window, Count gridSize)
{
	if (gridSize <= 0)
		gridSize = 1;

	SetGridSize(&window->settings, gridSize);
}


/*
 * Set the first grid character for a window.
 */
void
SetWindowGridChar1(LifeWindow * window, int gridChar)
{
	SetGridChar1(&window->settings, gridChar);
}


/*
 * Set the second grid character for a window.
 */
void
SetWindowGridChar2(LifeWindow * window, int gridChar)
{
	SetGridChar2(&window->settings, gridChar);
}


/*
 * Set the cell shape for a window.
 */
void
SetWindowCellShape(LifeWindow * window, CellShape cellShape)
{
	SetCellShape(&window->settings, cellShape);
}


/*
 * Set whether the cell shape is small.
 */
void
SetWindowIsCellShapeSmall(LifeWindow * window, BOOL isSmall)
{
	SetIsCellShapeSmall(&window->settings, isSmall);
}


/*
 * Set the grid type for a window.
 */
void
SetWindowGridType(LifeWindow * window, GridType gridType)
{
	SetGridType(&window->settings, gridType);
}


/*
 * Set the background color for a window.
 * Returns true if the color was known.
 */
BOOL
SetWindowBackgroundColor(LifeWindow * window, const char * colorName)
{
	ColorIndex colorIndex = DefineColor(colorName);

	if (colorIndex < 0)
		return FALSE;

	SetBackgroundIndex(&window->settings, colorIndex);

	return TRUE;
}


/*
 * Set the normal cell color for a window.
 * Returns true if the color was known.
 */
BOOL
SetWindowCellColor(LifeWindow * window, const char * colorName)
{
	ColorIndex colorIndex = DefineColor(colorName);

	if (colorIndex < 0)
		return FALSE;

	SetCellColorIndex(&window->settings, colorIndex);

	return TRUE;
}


/*
 * Set the tracks color for a window.
 * Returns true if the color was known.
 */
BOOL
SetWindowTracksColor(LifeWindow * window, const char * colorName)
{
	ColorIndex colorIndex = DefineColor(colorName);

	if (colorIndex < 0)
		return FALSE;

	SetTracksColorIndex(&window->settings, colorIndex);

	return TRUE;
}


/*
 * Set the marked cells color for a window.
 * Returns true if the color was known.
 */
BOOL
SetWindowMarkColor(LifeWindow * window, const char * colorName)
{
	ColorIndex colorIndex = DefineColor(colorName);

	if (colorIndex < 0)
		return FALSE;

	SetMarkColorIndex(&window->settings, colorIndex);

	return TRUE;
}


/*
 * Set the old cells color for a window.
 * Returns true if the color was known.
 */
BOOL
SetWindowOldColor(LifeWindow * window, const char * colorName)
{
	ColorIndex colorIndex = DefineColor(colorName);

	if (colorIndex < 0)
		return FALSE;

	SetOldColorIndex(&window->settings, colorIndex);

	return TRUE;
}


/*
 * Set the view text color for a window.
 * Returns true if the color was known.
 */
BOOL
SetWindowTextColor(LifeWindow * window, const char * colorName)
{
	ColorIndex colorIndex = DefineColor(colorName);

	if (colorIndex < 0)
		return FALSE;

	SetTextColorIndex(&window->settings, colorIndex);

	return TRUE;
}


/*
 * Set the border width for a window.
 */
void
SetWindowBorderWidth(LifeWindow * window, PixelCount borderWidth)
{
	SetBorderWidth(&window->settings, borderWidth);
}


/*
 * Given a window try to find a top level window other
 * then the given one.  If the given one is a sub-window
 * then its parent is returned.  Returns 0 if the window
 * is a top level window but there is no other top level
 * window.
 */
LifeWindow *
FindOtherTopLevelWindow(const LifeWindow * window)
{
	const LifeDevice * dev = Life.device;
	int index;

	/*
	 * If we have a parent window then return that.
	 */
	if (window->parentWindow)
		return window->parentWindow;

	/*
	 * If another window is found which is a top level window
	 * then return that.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * testWindow = dev->windowTable[index];

		if ((testWindow != window) && (testWindow->parentWindow == 0))
			return testWindow;
	}

	/*
	 * We were the only top level window so return null.
	 */
	return 0;
}


/*
 * Return whether the indicated rectangular region overlaps
 * any subwindow of the specified window.  Returns TRUE if so.
 */
BOOL
IsOverlappingSubwindow(const LifeWindow * window,
	PixelCoord x, PixelCoord y, PixelCount width, PixelCount height)
{
	const LifeDevice * dev = Life.device;
	int index;

	for (index = 0; index < dev->windowCount; index++)
	{
		const LifeWindow * testWindow = dev->windowTable[index];

		if ((testWindow == window) || (testWindow->parentWindow != window))
			continue;

		if (x + width <= testWindow->positionX)
			continue;

		if (y + height <= testWindow->positionY)
			continue;

		if (testWindow->positionX + testWindow->width <= x)
			continue;

		if (testWindow->positionY + testWindow->height <= y)
			continue;

		return TRUE;
	}

	return FALSE;
}


/*
 * Return whether any window needs to be updated.
 */
BOOL
IsWindowUpdateNeeded(void)
{
	const LifeDevice * dev = Life.device;
	int index;

	for (index = 0; index < dev->windowCount; index++)
	{
		const LifeWindow * window = dev->windowTable[index];
		const LifeObject * obj = window->currentObject;

		if (!window->isMapped)
			continue;

		if (window->updateFlags)
			return TRUE;

		if (window->lastChangeCount != obj->changeCount)
			return TRUE;
	}

	return FALSE;
}


/*
 * Add to the list of update flags for the windows showing
 * the specified object.  If the object is NULL then the
 * flags are set in all windows.
 */
void
AddWindowObjectUpdateFlags(const LifeObject * obj, Flags flags)
{
	LifeDevice * dev = Life.device;
	int index;

	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * window = dev->windowTable[index];

		if ((obj == 0) || (window->currentObject == obj))
		{
			window->updateFlags |= flags;
		}
	}
}


/*
 * Return the number of windows which are showing the object.
 */
int
GetObjectWindowCount(const LifeObject * obj)
{
	int count = 0;
	int index;

	for (index = 0; index < Life.device->windowCount; index++)
	{
		const LifeWindow * window = Life.device->windowTable[index];

		if (window->currentObject == obj)
			count++;
	}

	return count;
}


/*
 * Clear the specified object from all windows.
 * This is used when an object is being destroyed.
 * The removed objects are replaced by the main object
 * (which is not destroyable).
 */
void
WindowClearObject(const LifeObject * object)
{
	LifeDevice * dev = Life.device;
	int index;

	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * window = dev->windowTable[index];

		if (window->currentObject == object)
		{
			window->currentObject = Life.mainObject;
			window->updateFlags |= U_REDRAW;
		}

		if (window->prevObject == object)
			window->prevObject = Life.mainObject;
	}
}



/*
 * Set the update outlines flag for all other windows
 * which show the same object as the specified window.
 */
void
AddOtherWindowOutlineFlags(const LifeWindow * window)
{
	const LifeObject * obj = window->currentObject;
	int index;

	for (index = 0; index < Life.device->windowCount; index++)
	{
		LifeWindow * testWindow = Life.device->windowTable[index];

		if ((testWindow != window) && (testWindow->currentObject == obj))
			testWindow->updateFlags |= U_OUTLINE;
	}
}


/*
 * Add the specified update flags to the window.
 */
void
AddWindowUpdateFlags(LifeWindow * window, Flags flags)
{
	window->updateFlags |= flags;
}


/*
 * Return the row and column values for the desired center
 * position according to the window's view type.
 */
void
GetViewTypePosition(const LifeWindow * window, NUMBER * row, NUMBER * col)
{
	const LifeWindow * parentWindow = window->parentWindow;
	const LifeObject * obj = window->currentObject;
	ViewType viewType = GetViewType(&window->settings);

	/*
	 * If this is a where location view type then get the
	 * coordinates of the appropriate where location.
	 */
	if ((viewType >= VIEW_TYPE_WHERE_A) &&
		(viewType <= VIEW_TYPE_WHERE_Z))
	{
		const LocationInfo * info = &obj->whereInfo[viewType - VIEW_TYPE_WHERE_A];

		*row = info->row;
		*col = info->col;

		return;
	}

	/*
	 * Switch on the view type to get the appropriate position.
	 */
	switch (viewType)
	{
		case VIEW_TYPE_SELF:
			*row = window->curRow;
			*col = window->curCol;

			break;

		case VIEW_TYPE_CURSOR:
			*row = obj->currentInfo.row;
			*col = obj->currentInfo.col;

			break;

		case VIEW_TYPE_ORIGIN:
			*row = obj->originInfo.row;
			*col = obj->originInfo.col;

			break;

		case VIEW_TYPE_MOUSE:
			*row = obj->mouseRow;
			*col = obj->mouseCol;

			break;

		case VIEW_TYPE_UNDER:
			/*
			 * If we are not a subwindow or if the parent window's
			 * object is not the same as ours then do nothing.
			 */
			if ((parentWindow == 0) || (parentWindow->currentObject != obj))
			{
				*row = window->curRow;
				*col = window->curCol;

				break;
			}

			/*
			 * Get the device coordinates of the center of this
			 * subwindow with respect to the parent window.
			 */
			PixelCoord centerX = window->positionX + window->width / 2;
			PixelCoord centerY = window->positionY + window->height / 2;

			/*
			 * Return the row and column number of the coordinates
			 * within the parent window.
			 */
			*row = numAdd(parentWindow->minRow,
				PixelToScreenCell(parentWindow, centerY));

			*col = numAdd(parentWindow->minCol,
				PixelToScreenCell(parentWindow, centerX));

			break;

		default:
			*row = numZero();
			*col = numZero();

			Error("Unknown view type");
	}
}


/*
 * Center the view around the current window's view position.
 * This clears the shifted flag.
 */
void
CenterView(LifeWindow * window)
{
	Scale	scale = window->scale;
	NUMBER	newViewRows;
	NUMBER	newViewCols;
	NUMBER	newMinRow;
	NUMBER	newMaxRow;
	NUMBER	newMinCol;
	NUMBER	newMaxCol;
	NUMBER	row;
	NUMBER	col;

	/*
	 * Get the position to center around according to
	 * the current view type.
	 */
	GetViewTypePosition(window, &row, &col);

	/*
	 * Calculate the number of rows and columns visible
	 * according the window size and scale values.
	 */
	if (scale == 0)
		scale = 1;

	if (scale > 0)
	{
		newViewRows = numFromLong(window->rows * scale);
		newViewCols = numFromLong(window->cols * scale);
	}
	else
	{
		newViewRows = numFromLong(window->rows / (-scale));
		newViewCols = numFromLong(window->cols / (-scale));

		if (numSign(newViewRows) <= 0)
			newViewRows = numOne();

		if (numSign(newViewCols) <= 0)
			newViewCols = numOne();
	}

	/*
	 * Calculate the new minimum and maximum 
	 * row and column values seen in the window.
	 */
	newMinRow = numSub(row, numHalf(newViewRows));
	newMaxRow = numDec(numAdd(newMinRow, newViewRows));
	newMinCol = numSub(col, numHalf(newViewCols));
	newMaxCol = numDec(numAdd(newMinCol, newViewCols));

	/*
	 * If any of the values are changing then
	 * the view will have to be redrawn and other
	 * windows will have to know our outline has
	 * changed.
	 */
	if (!numIsEqual(window->minRow, newMinRow) ||
		!numIsEqual(window->maxRow, newMaxRow) ||
		!numIsEqual(window->minCol, newMinCol) ||
		!numIsEqual(window->maxCol, newMaxCol))
	{
		window->updateFlags |= U_REDRAW;

		AddOtherWindowOutlineFlags(window);
	}

	/*
	 * Store the new viewing information.
	 */
	window->curRow = row;
	window->curCol = col;
	window->minRow = newMinRow;
	window->maxRow = newMaxRow;
	window->minCol = newMinCol;
	window->maxCol = newMaxCol;
	window->viewRows = newViewRows;
	window->viewCols = newViewCols;
	window->seeCount = numZero();
	window->isShifted = FALSE;
}


/*
 * Shift the view by the indicated amount.
 * Positive shift values are to the right or down.
 * The shift values are in terms of object rows and columns.
 * This temporarily disables the window automatically
 * repositioning itself to make its viewing position visible.
 * The repositioning is re-enabled when the view is recentered
 * or the view type is set again.
 */
void
ShiftView(LifeWindow * window, NUMBER rowShift, NUMBER colShift)
{
	if (numIsZero(rowShift) && numIsZero(colShift))
		return;

	/*
	 * Perform the shift.
	 */
	numAddrAdd(&window->curRow, rowShift);
	numAddrAdd(&window->curCol, colShift);
	numAddrAdd(&window->minRow, rowShift);
	numAddrAdd(&window->maxRow, rowShift);
	numAddrAdd(&window->minCol, colShift);
	numAddrAdd(&window->maxCol, colShift);
	window->seeCount = numZero();

	/*
	 * A full update is needed and other windows
	 * have to update our outline.
	 */
	window->updateFlags |= U_ALL;

	AddOtherWindowOutlineFlags(window);

	/*
	 * If the view type is not ourself then remember
	 * that the view is shifted so that automatic
	 * positioning will not be done.  We don't bother
	 * setting the flag for the self view type since
	 * that is always positioned correctly.
	 */
	if (GetViewType(&window->settings) != VIEW_TYPE_SELF)
		window->isShifted = TRUE;
}


/*
 * Set the scaling factor for the object in the specified window.
 * This also centers the view around the current view position.
 * Positive scale factors mean compress the view so that each
 * displayed item represents multiple cells.  Negative scale factors
 * mean use multiple displayed items to represent individual cells.
 */
void
SetScale(LifeWindow * window, Scale scale)
{
	LifeDevice * dev = Life.device;

	/*
	 * Make sure the scale is in range.
	 */
	if (scale < dev->minScale)
		scale = dev->minScale;

	if (scale > dev->maxScale)
		scale = dev->maxScale;

	if ((scale == 0) || (scale == -1))
		scale = 1;

	/*
	 * If the scale is changing then the window will
	 * have to be redrawn and other windows will have
	 * to update our outline.
	 */
	if (window->scale != scale)
	{
		window->updateFlags |= U_REDRAW;

		AddOtherWindowOutlineFlags(window);
	}

	/*
	 * Set the scale value and then recenter the
	 * view around the current view position.
	 */
	window->scale = scale;

	CenterView(window);
}


/*
 * Perform auto-scaling of the specified window.  This implies
 * picking a scaling factor such that the whole object fits in
 * the screen.  The scale factor is never decreased.  When the
 * scale factor is large, convenient ones are picked.  Returns
 * the new scale factor.
 */
Scale
AutoScale(LifeWindow * window)
{
	LifeDevice * dev = Life.device;
	LifeObject * obj = window->currentObject;
	Scale scale = window->scale;
	NUMBER minRow;
	NUMBER maxRow;
	NUMBER minCol;
	NUMBER maxCol;

	/*
	 * Get the bounds on the object.
	 * If the object is empty do nothing.
	 */
	if (!MinMax(obj, &minRow, &maxRow, &minCol, &maxCol))
		return scale;

	/*
	 * Keep increasing the scale as required until the
	 * whole object will be visible in the window.
	 */
	while ((scale <= dev->maxScale) &&
		(numIsLess(minRow, window->minRow) ||
			numIsGreater(maxRow, window->maxRow) ||
			numIsLess(minCol, window->minCol) ||
			numIsGreater(maxCol, window->maxCol)))
	{
		scale++;

		if ((scale == -1) || (scale == 0))
			scale = 1;

		if (scale > 20)
			scale += (5 - (scale % 5));

		if (scale > 50)
			scale += (10 - (scale % 10));

		if (scale > 200)
			scale += (100 - (scale % 100));

		if (scale > 1000)
			scale += (500 - (scale % 500));

		if (scale > 10000)
			scale += (5000 - (scale % 5000));

		if (scale > 100000)
			scale += (50000 - (scale % 50000));

		/*
		 * Apply this scale value in preparation
		 * for the next check.
		 */
		SetScale(window, scale);
	}

	return scale;
}


/*
 * Set the view type for the window.
 */
void
SetWindowViewType(LifeWindow * window, ViewType viewType)
{
	Settings * settings = &window->settings;

	if (window->isShifted || (GetViewType(settings) != viewType))
	{
		window->updateFlags |= U_ALL;
	}

	SetViewType(settings, viewType);
	window->isShifted = FALSE;
}


/*
 * Set the position type for the window.
 */
void
SetWindowPositionType(LifeWindow * window, PositionType positionType)
{
	Settings * settings = &window->settings;

	if (window->isShifted ||
		(GetPositionType(settings) != positionType))
	{
		window->updateFlags |= U_ALL;
	}

	SetPositionType(settings, positionType);
	window->isShifted = FALSE;
}


/*
 * Set whether or not this window is to show the
 * outlines of other windows which allow it.
 */
void
SetWindowIsShowOutlines(LifeWindow * window, BOOL isShowOutlines)
{
	Settings * settings = &window->settings;

	if (GetIsShowOutlines(settings) == isShowOutlines)
		return;

	SetIsShowOutlines(settings, isShowOutlines);
	window->updateFlags |= U_REDRAW;
}


/*
 * Set whether or not this window's outline can be
 * shown in other windows which might want to see it.
 */
void
SetWindowIsShowMyOutline(LifeWindow * window, BOOL isShowMyOutline)
{
	LifeDevice * dev = Life.device;
	LifeObject * obj = window->currentObject;
	Settings * settings = &window->settings;
	int index;

	if (GetIsShowMyOutline(settings) == isShowMyOutline)
		return;

	SetIsShowMyOutline(settings, isShowMyOutline);

	/*
	 * Scan the windows looking for other ones showing our
	 * object and which have the show outlines feature enabled.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * testWindow = dev->windowTable[index];

		if ((testWindow == window) || (window->currentObject != obj))
			continue;

		if (GetIsShowOutlines(&testWindow->settings))
			testWindow->updateFlags |= U_REDRAW;
	}
}


/*
 * Position the view so that the current window position is visible.
 * How this is done depends on the position type.
 */
void
PositionView(LifeWindow * window)
{
	NUMBER row;
	NUMBER col;

	GetViewTypePosition(window, &row, &col);

	switch (GetPositionType(&window->settings))
	{
		case POSITION_TYPE_SHIFT:
			PositionCell(window, row, col);
			break;

		case POSITION_TYPE_SCALE:
			AutoScale(window);
			break;

		case POSITION_TYPE_CENTER:
			CenterView(window);
			break;

		default:
			Error("Unknown position type");
	}
}


/*
 * Position the view of the current object by shifting it
 * if necessary to show the specified cell's location.
 * If positioning is required then the margin is used.
 */
void
PositionCell(LifeWindow * window, NUMBER row, NUMBER col)
{
	NUMBER minRow = window->minRow;
	NUMBER maxRow = window->maxRow;
	NUMBER minCol = window->minCol;
	NUMBER maxCol = window->maxCol;
	NUMBER viewRows = window->viewRows;
	NUMBER viewCols = window->viewCols;
	BOOL isMoved = FALSE;

	/*
	 * Calculate the required margins if repositioning is needed.
	 * The margin is a percentage of the window's minimum radius
	 * giving the extra division factor of 2.
	 */
	NUMBER margin =
		(numIsLessOrEqual(window->viewCols, window->viewRows) ?
		window->viewCols : window->viewRows);

	numAddrMulLong(&margin, GetMarginPercentage(&window->settings) / 200);

	/*
	 * If the cell is outside of the visible area then
	 * shift the visible area so as to show it, with the
	 * margin around it.
	 */
	if (numIsLess(row, minRow))
	{
		minRow = numSub(row, margin);
		maxRow = numDec(numAdd(minRow, viewRows));
		isMoved = TRUE;
	}

	if (numIsGreater(row, maxRow))
	{
		maxRow = numAdd(row, margin);
		minRow = numInc(numSub(maxRow, viewRows));
		isMoved = TRUE;
	}

	if (numIsLess(col, minCol))
	{
		minCol = numSub(col, margin);
		maxCol = numDec(numAdd(minCol, viewCols));
		isMoved = TRUE;
	}

	if (numIsGreater(col, maxCol))
	{
		maxCol = numAdd(col, margin);
		minCol = numInc(numSub(maxCol, viewCols));
		isMoved = TRUE;
	}

	/*
	 * If we have moved the window then update
	 * our location and tell other windows that
	 * our outline has changed.
	 */
	if (isMoved)
	{
		window->minRow = minRow;
		window->maxRow = maxRow;
		window->minCol = minCol;
		window->maxCol = maxCol;
		window->updateFlags |= U_ALL;

		AddOtherWindowOutlineFlags(window);
	}

	/*
	 * Update the new position.
	 */
	window->curRow = row;
	window->curCol = col;
	window->isShifted = FALSE;
}


/*
 * Position the view of the current object to show both the
 * given region and the current window position.  If this
 * is impossible, just the window position will be positioned.
 * The positioning is done leaving a border on the edge.
 */
void
PositionArea(LifeWindow * window, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol)
{
	NUMBER curRow;
	NUMBER curCol;
	BOOL isMoved = FALSE;

	/*
	 * Calculate the required margins if repositioning is needed.
	 * The margin is a percentage of the window's minimum radius
	 * giving the extra division factor of 2.
	 */
	NUMBER margin =
		(numIsLessOrEqual(window->viewCols, window->viewRows) ?
		window->viewCols : window->viewRows);

	numAddrMulLong(&margin, GetMarginPercentage(&window->settings) / 200);

	GetViewTypePosition(window, &curRow, &curCol);

	/*
	 * If the new current position is outside of the
	 * specified area then expand the area to include
	 * the current position.
	 */
	minRow = numMin(minRow, curRow);
	maxRow = numMax(maxRow, curRow);
	minCol = numMin(minCol, curCol);
	maxCol = numMax(maxCol, curCol);

	/*
	 * If the rectangle is too large to fit
	 * within the window along with the required
	 * margin then reduce the rectangle as much
	 * as is needed.
	 */
	if (numIsGreaterOrEqual(numSub(maxRow, minRow),
		numSub(window->viewRows, numDouble(margin))))
	{
		minRow = numSub(curRow, margin);
		maxRow = curRow;
	}

	if (numIsGreaterOrEqual(numSub(maxCol, minCol),
		numSub(window->viewCols, numDouble(margin))))
	{
		minCol = numSub(curCol, margin);
		maxCol = curCol;
	}

	if (numIsLess(minRow, numAdd(window->minRow, margin)))
	{
		window->minRow = numSub(minRow, margin);
		window->maxRow = numDec(numAdd(window->minRow, window->viewRows));
		isMoved = TRUE;
	}

	if (numIsGreater(maxRow, numSub(window->maxRow, margin)))
	{
		window->maxRow = numAdd(maxRow, margin);
		window->minRow = numInc(numSub(window->maxRow, window->viewRows));
		isMoved = TRUE;
	}

	if (numIsLess(minCol, numAdd(window->minCol, margin)))
	{
		window->minCol = numSub(minCol, margin);
		window->maxCol = numDec(numAdd(window->minCol, window->viewCols));
		isMoved = TRUE;
	}

	if (numIsGreater(maxCol, numSub(window->maxCol, margin)))
	{
		window->maxCol = numAdd(maxCol, margin);
		window->minCol = numInc(numSub(window->maxCol, window->viewCols));
		isMoved = TRUE;
	}

	if (isMoved)
	{
		window->updateFlags |= U_ALL;

		AddOtherWindowOutlineFlags(window);
	}

	window->curRow = curRow;
	window->curCol = curCol;
	window->isShifted = FALSE;
}


/*
 * Update all of the windows which require it.
 * This does a bit of preliminary checking and then calls the
 * device update routine for the real work.
 */
void
UpdateAllWindows(void)
{
	LifeDevice * dev = Life.device;
	int index;

	/*
	 * First loop over all the windows to position and scale
	 * them as required.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * window = dev->windowTable[index];

		/*
		 * Move the view if required to show the current position.
		 * Don't do this if the user has shifted the view or if
		 * the window is showing text.
		 */
		if (!window->isShifted && !window->isTextView)
			PositionView(window);

		/*
		 * If autoscaling is enabled in the window then
		 * perform that if needed.
		 */
		if (GetIsAutoScale(&window->settings) && !window->isTextView)
			AutoScale(window);

		/*
		 * Set the window update flags according to the changed
		 * settings flags if any.
		 */
		ApplyWindowSettingsChangedFlags(window);
	}

	/*
	 * Now call the device update routine to update any
	 * windows which require it.
	 */
	dev->update(dev);

	/*
	 * Clear the modified flags for the shared settings.
	 */
	ClearSettingsChangedFlags();
}


/*
 * Build the table of area coordinates which describe the
 * views that other windows have within our own view.
 * This is used to draw rectangles showing those view areas.
 * Before calling this the old outline should be removed
 * from the drawn window since it uses this data.
 */
void
BuildOutlineTable(LifeWindow * window)
{
	int index;

	/*
	 * Clear the old outline count.
	 */
	window->mOutlineCount = 0;

	/*
	 * If the window is not showing outlines then do nothing else.
	 */
	if (!GetIsShowOutlines(&window->settings))
		return;

	/*
	 * Examine the window table to find the windows which
	 * can generate outlines for us.
	 */
	for (index = 0; index < Life.device->windowCount; index++)
	{
		LifeWindow * testWindow = Life.device->windowTable[index];
		Area * area;

		/*
		 * Don't consider our own window, one which is not visible,
		 * one which not showing the same object as us, or one
		 * which doesn't want its outline show.
		 */
		if ((testWindow == window) ||
			(testWindow->currentObject != window->currentObject) ||
			!GetIsVisible(&testWindow->settings) ||
			!GetIsShowMyOutline(&testWindow->settings))
		{
			continue;
		}

		/*
		 * This window's outline wants to be seen by us.
		 * Add an entry to the table to remember its viewing area.
		 */
		area = &window->mOutlineTable[window->mOutlineCount];

		area->minRow = testWindow->minRow;
		area->maxRow = testWindow->maxRow;
		area->minCol = testWindow->minCol;
		area->maxCol = testWindow->maxCol;

		window->mOutlineCount++;
	}
}


/*
 * Generate the status line text for the specified window.
 * Returns a pointer to the status text which is in a static
 * buffer overwritten on each call.
 */
const char *
GenerateStatus(const LifeWindow * window)
{
	static char buf[8192];
	LifeObject * obj = window->currentObject;
	const Settings * settings = &window->settings;
	ViewType viewType = GetViewType(settings);
	NUMBER cellCount = CalculateCellCount(obj);
	const char * rule;
	const char * input;
	BOOL usingStop;
	char * cp = buf;

	if (GetStatusLocation(settings) == STATUS_LOCATION_PARENT)
	{
		sprintf(cp, "Win[%ld] ", window->id);
		cp += strlen(cp);
	}

	sprintf(cp, "Gen:%ld", obj->gen);
	cp += strlen(cp);

	if ((obj->genLeft > 0) && (obj->genLeft != INFINITY))
	{
		sprintf(cp, "(%ldr)", obj->genLeft);
		cp += strlen(cp);
	}

	sprintf(cp, " cells:%s", numToString(cellCount));
	cp += strlen(cp);

//	if (numIsGreater(cellCount, window->seeCount))
//	{
//		sprintf(cp, "(%si)",
//			numToString(numSub(cellCount, window->seeCount)));
//
//		cp += strlen(cp);
//	}

	if (numIsNonZero(obj->born))
	{
		sprintf(cp, " born:%s", numToString(obj->born));
		cp += strlen(cp);
	}

	if (numIsNonZero(obj->died))
	{
		sprintf(cp, " died:%s", numToString(obj->died));
		cp += strlen(cp);
	}

	if (numIsNonZero(obj->growth) &&
		numIsZero(obj->born) && numIsZero(obj->died))
	{
		sprintf(cp, " growth:%s", numToString(obj->growth));
		cp += strlen(cp);
	}

	if (obj->frequency > 1)
	{
		sprintf(cp, " %s:%ld",
			(obj->isJumpingEnabled ? "jump" : "freq"),
			obj->frequency);

		cp += strlen(cp);
	}

	if ((window->scale != 1) || GetIsAutoScale(settings))
	{
		sprintf(cp, " %scale:%d",
			(GetIsAutoScale(settings) ? "autos" : "s") ,
			window->scale);

		cp += strlen(cp);
	}

	rule = obj->rule->name;

	if (*rule == '\0')
		rule = "complex";

	if (strcmp(rule, DEFAULT_RULE))
	{
		sprintf(cp, " rule:%s", rule);
		cp += strlen(cp);
	}

	input = GetInputDescription();

	if (input)
	{
		*cp++ = ' ';
		strcpy(cp, input);
		cp += strlen(cp);
	}

	if (obj->isTracksUsed)
	{
		if (obj->isTracksVisible)
			strcpy(cp, " tracks");
		else
			strcpy(cp, " tracks-invisible");

		cp += strlen(cp);
	}

	usingStop = FALSE;

	if (obj->stopCell)
	{
		strcpy(cp, " stop-cell");
		cp += strlen(cp);
		usingStop = TRUE;
	}

	if (obj->stopOnPath)
	{
		strcpy(cp, " stop-on-path");
		cp += strlen(cp);
		usingStop = TRUE;
	}

	if (obj->stopInsidePath)
	{
		strcpy(cp, " stop-inside");
		cp += strlen(cp);
		usingStop = TRUE;
	}

	if (obj->stopOutsidePath)
	{
		strcpy(cp, " stop-outside");
		cp += strlen(cp);
		usingStop = TRUE;
	}

	if (obj->stopGap != 0)
	{
		sprintf(cp, " stop-gap:%ld", obj->stopGap);
		usingStop = TRUE;
		cp += strlen(cp);
	}

	if (obj->stopExpr[0] != '\0')
	{
		strcpy(cp, " stop-expr");
		cp += strlen(cp);
		usingStop = TRUE;
	}

	if (usingStop && !obj->stopIsEnabled)
	{
		strcpy(cp, " (disabled)");
		cp += strlen(cp);
	}

	if (Life.moveMode == MOVE_MODE_INSERT)
	{
		strcpy(cp, " inserting");
		cp += strlen(cp);
	}

	if (Life.moveMode == MOVE_MODE_DELETE)
	{
		strcpy(cp, " deleting");
		cp += strlen(cp);
	}

	if (Life.delayTime > 0)
	{
		if (Life.delayTime < 1000)
			sprintf(cp, " delay:%d", Life.delayTime);
		else if ((Life.delayTime % 1000) == 0)
			sprintf(cp, " delay:%ds", Life.delayTime / 1000);
		else
			sprintf(cp, " delay:%d.%ds", Life.delayTime / 1000,
				(Life.delayTime / 100) % 10);

		cp += strlen(cp);
	}

	switch (viewType)
	{
		case VIEW_TYPE_CURSOR:
			break;

		case VIEW_TYPE_MOUSE:
			strcpy(cp, " view:mouse");
			break;

		case VIEW_TYPE_ORIGIN:
			strcpy(cp, " view:origin");
			break;

		case VIEW_TYPE_SELF:
			strcpy(cp, " view:self");
			break;

		case VIEW_TYPE_UNDER:
			strcpy(cp, " view:under");
			break;

		default:
			if ((viewType >= VIEW_TYPE_WHERE_A) &&
				(viewType <= VIEW_TYPE_WHERE_Z))
			{
				sprintf(cp, " view:%c",
					viewType - VIEW_TYPE_WHERE_A + 'a');
			}
	}

	cp += strlen(cp);

	switch (GetPositionType(settings))
	{
		case POSITION_TYPE_SHIFT:
			break;

		case POSITION_TYPE_CENTER:
			strcpy(cp, " center");
			break;

		case POSITION_TYPE_SCALE:
			strcpy(cp, " autoscale");
			break;
	}

	cp += strlen(cp);

	if (settings->sharedSettings->name[0] != '.')
	{
		strcat(cp, " settings:");
		strcat(cp, settings->sharedSettings->name);
		cp += strlen(cp);
	}

	if (window->isShifted)
	{
		strcat(cp, " shifted");
		cp += strlen(cp);
	}

	if (Life.undoDepth > 0)
	{
		sprintf(cp, " undo:%d", Life.undoDepth);
		cp += strlen(cp);
	}

	if (obj != Life.mainObject)
	{
		sprintf(cp, " \"%s\"", GetStringBufferText(&obj->name));
		cp += strlen(cp);
	}

	if (obj->isOverlayActive)
	{
		sprintf(cp, " \(%s)",
			GetStringBufferText(&obj->overlayObject->name));

		cp += strlen(cp);
	}

	//
	// Check whether distances are shown.
	//
	if (obj->isLocationsEnabled)
	{
		int i;

		strcat(cp, GetDistanceInfo(obj, &obj->originInfo));
		cp += strlen(cp);

		for (i = 0; i < 26; i++)
		{
			strcat(cp, GetDistanceInfo(obj, &obj->whereInfo[i]));
			cp += strlen(cp);
		}
	}

	return buf;
}


/*
 * Get the distance info for the specified where location.
 * Returns an empty string if it is not enabled.
 */
static const char *
GetDistanceInfo(LifeObject * obj, const LocationInfo * info)
{
	static char buf[8192];

	if (!info->isEnabled || !info->isDistShown)
		return "";

	NUMBER rowDiff = numSub(obj->currentInfo.row, info->row);
	NUMBER colDiff = numSub(obj->currentInfo.col, info->col);

	NUMBER forwDiagDiff = numAdd(rowDiff, colDiff);
	NUMBER backDiagDiff = numSub(rowDiff, colDiff);

	rowDiff = numAbs(rowDiff);
	colDiff = numAbs(colDiff);
	forwDiagDiff = numAbs(forwDiagDiff);
	backDiagDiff = numAbs(backDiagDiff);

	const char * forwFrac = numIsOdd(forwDiagDiff) ? ".5" : "";
	const char * backFrac = numIsOdd(backDiagDiff) ? ".5" : "";

	forwDiagDiff = numDivLong(forwDiagDiff, 2);
	backDiagDiff = numDivLong(backDiagDiff, 2);

	sprintf(buf, " %c:(r=%s c=%s f=%s%s b=%s%s)", info->id,
		numToString(rowDiff), numToString(colDiff),
		numToString(forwDiagDiff), forwFrac,
		numToString(backDiagDiff), backFrac);

	return buf;
}


/*
 * Set the current object for the current window.
 * If the object is changed, the old object is saved
 * as the previous one.
 */
void
SetWindowObject(LifeWindow * window, LifeObject * object)
{
	if (object == window->currentObject)
		return;

	window->prevObject = window->currentObject;
	window->prevObject->windowScale = window->scale;
	window->currentObject = object;

	SetWindowIsAutoScale(window, object->windowIsAutoScale);
	SetScale(window, object->windowScale);
	window->curRow = object->windowCurRow;
	window->curCol = object->windowCurCol;

	window->lastChangeCount = 0;
	window->updateFlags |= U_REDRAW;
}


/*
 * Return the current window scale value.
 */
Scale
GetScale(const LifeWindow * window)
{
	return window->scale;
}


/*
 * Return the current number of rows and columns that are
 * visible in the window.
 */
void
GetVisibleSize(const LifeWindow * window,
	NUMBER * rowCount, NUMBER * colCount)
{
	*rowCount = window->viewRows;
	*colCount = window->viewCols;
}


/*
 * Return the bounds of the region visible in the window.
 */
void
GetVisibleRegion(const LifeWindow * window,
	NUMBER * minRow, NUMBER * maxRow,
	NUMBER * minCol, NUMBER * maxCol)
{
	*minRow = window->minRow;
	*maxRow = window->maxRow;
	*minCol = window->minCol;
	*maxCol = window->maxCol;
}


/*
 * Return the current object in a window.
 */
LifeObject *
GetCurrentObject(const LifeWindow * window)
{
	return window->currentObject;
}


/*
 * Return the previous object in a window.
 */
LifeObject *
GetPreviousObject(const LifeWindow * window)
{
	return window->prevObject;
}


/*
 * Calculate the residue of the specifed coordinate value
 * modulo the current grid size of the window.  This is
 * slightly complicated because it has to work correctly
 * for negative coordinates too.
 */
int
GridMod(const LifeWindow * window, NUMBER coord)
{
	Coord gridSize = GetGridSize(&window->settings);

	if (numSign(coord) >= 0)
		return numModLong(coord, gridSize);

	coord = numFromLong(numModLong(numNeg(coord), gridSize));

	if (!numIsZero(coord))
		coord = numSub(numFromLong(gridSize), coord);

	return numToLong(coord);
}


/*
 * Convert a window cell offset into a pixel offset.
 * The returned pixel is in the center of the cell.
 */
PixelCoord
ScreenCellToPixel(const LifeWindow * window, NUMBER coord)
{
	Scale scale = window->scale;

	if ((scale == 0) || (scale == -1))
		scale = 1;

	if (scale > 0)
		return numToLong(numDivLong(coord, scale));

	scale = -scale;

	return (numToLong(numMulLong(coord, scale)) + ((scale - 1) / 2));
}


/*
 * Convert a pixel offset into a window cell offset.
 * The pixel offset ranges from 0 to the size of the window.
 * The cell offset ranges from 0 to the number of visible cells.
 */
NUMBER
PixelToScreenCell(const LifeWindow * window, PixelCoord pixel)
{
	Scale scale = window->scale;

	if ((scale == 0) || (scale == -1))
		scale = 1;

	if (scale >= 0)
		return numFromLong(pixel * scale);

	return numFromLong(pixel / (-scale));
}


/*
 * Save the current window viewing parameters into
 * the current object.
 */
void
SaveCurrentViewParameters(void)
{
	const LifeWindow * window = Life.currentWindow;
	LifeObject * obj = Life.currentObject;

	obj->windowIsAutoScale = GetIsAutoScale(&window->settings);
	obj->windowScale = GetScale(window);
	obj->windowCurRow = window->curRow;
	obj->windowCurCol = window->curCol;
}


/*
 * Restore the current window viewing parameters from
 * the current object.
 */
void
RestoreCurrentViewParameters(void)
{
	LifeWindow * window = Life.currentWindow;
	const LifeObject * obj = Life.currentObject;

	SetWindowIsAutoScale(window, obj->windowIsAutoScale);
	SetScale(window, obj->windowScale);
	window->curRow = obj->windowCurRow;
	window->curCol = obj->windowCurCol;
	AddWindowObjectUpdateFlags(obj, U_POS);
}


/* END CODE */
