/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to help in the efficient viewing of a reasonably small region
 * of a life object.  This is useful when displaying the data in a window.
 */
#include <stdlib.h>
#include <memory.h>
#include <stdio.h>

#include "verify.h"
#include "viewinfo.h"
#include "number.h"


static char * allocateData(int size, int id);
static void * allocMem(int size);


//
// The maximum number of rows or columns that can be specified.
//
#define	MAX_VIEW_SIDE 10000


/*
 * Initialise the view structure for the specified rectangular region.
 * This allocates a data array large enough to store the region data,
 * and optionally another data array which can be used to store other data.
 * There is a limit on the size of the rectangular region that can be used.
 * The scale factor is used to group rows and columns together into single
 * elements.  (i.e, scale 1 means individual rows and columns and scale 10
 * means 100 cells of a square are compressed into one element of the data.)
 * Scale factors must always be positive for view info calculations.
 * The number of rows and columns must be multiples of the scale factor.
 * The data array in the returned structure should NOT be freed by the caller
 * except by calling FreeView and is only valid only until the next call to
 * this routine with the same id (if the id is non-negative).  There are
 * only a small number of allowable ids.  Having multiple ids allows multiple
 * views being used at the same time.  The data array values are NOT
 * initialised here.
 */
void
InitView(ViewInfo * info, int id, BOOL isAuxUsed, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol)
{
	//
	// Check the parameters and make sure that the area isn't too large.
	// Calculate the number of rows and columns taking into account the
	// scale factor, and make sure that the number of rows and columns
	// is an exact multiple of the scale factor.
	//
	verifyAlways(numIsLessOrEqual(minRow, maxRow) &&
		numIsLessOrEqual(minCol, maxCol));

	NUMBER rowCountNum = numInc(numSub(maxRow, minRow));
	NUMBER colCountNum = numInc(numSub(maxCol, minCol));

	verify((scale >= 1) && (numModLong(rowCountNum, scale) == 0) &&
		(numModLong(colCountNum, scale) == 0));

	numAddrDivLong(&rowCountNum, scale);
	numAddrDivLong(&colCountNum, scale);

	verify((numCmpLong(rowCountNum, MAX_VIEW_SIDE) <= 0) &&
		(numCmpLong(colCountNum, MAX_VIEW_SIDE) <= 0));

	/*
	 * Allocate a data buffer for the specified id having the correct size.
	 * This buffer will remain valid until the next call to this routine
	 * using the same id.
	 */
	int rowCount = (int) numToLong(rowCountNum);
	int colCount = (int) numToLong(colCountNum);
	int dataSize = rowCount * colCount;
	int allocSize = isAuxUsed ? dataSize * 2 : dataSize;

	char * data = allocateData(allocSize, id);

	//
	// Store the information.
	//
	info->data = data;
	info->auxData = (isAuxUsed ? (data + dataSize) : 0);
	info->dataSize = rowCount * colCount;
	info->rowCount = rowCount;
	info->colCount = colCount;
	info->minRow = minRow;
	info->maxRow = maxRow;
	info->minCol = minCol;
	info->maxCol = maxCol;
	info->id = id;
	info->scale = scale;
	info->rowIndex = 0;
	info->dataOffset = 0;
	info->endRowOffset = 0;
	info->onCount = numZero();
	info->isValid = TRUE;
}


/*
 * Initialise the view structure to a state having almost no data.
 * This means that the find next row and find next column routines
 * always return -1.
 */
void
InitEmptyView(ViewInfo * info)
{
	info->data = allocMem(1);
	info->auxData = 0;
	info->dataSize = 0;
	info->rowCount = 1;
	info->colCount = 1;
	info->id = -1;
	info->scale = 1;
	info->isValid = TRUE;
	info->onCount = numZero();
	info->minRow = numZero();
	info->maxRow = numZero();
	info->minCol = numZero();
	info->maxCol = numZero();
	info->rowIndex = 0;
	info->dataOffset = 0;
	info->endRowOffset = 0;

	info->data[0] = 0;
}


/*
 * Free the data buffer from the specified view if it was allocated
 * privately and clean up some fields.  This frees nothing for
 * non-negative ids since their data buffer is kept for re-use.
 */
void
FreeView(ViewInfo * info)
{
	if (info->isValid && (info->id < 0) && (info->data != 0))
		free(info->data);

	info->isValid = FALSE;
	info->data = 0;
	info->auxData = 0;
	info->dataSize = 0;
	info->rowIndex = 0;
	info->dataOffset = 0;
	info->endRowOffset = 0;
}


/*
 * Reset the data offset in the view info for another scan.
 */
void
ResetView(ViewInfo * info)
{
	info->rowIndex = 0;
	info->dataOffset = 0;
	info->endRowOffset = 0;
}


/*
 * Return the index (zero based) of the next row in the data which
 * has at least one ON cell in it.  This also calculates the end of
 * the row containing that first ON cell.  Each call to this routine
 * will advance to the next row having an ON cell, whether or not the
 * FindNextColumnIndex routine has been called.  Returns -1 if
 * there are no more rows having ON cells.
 */
int
FindNextRowIndex(ViewInfo * info)
{
	//
	// Start the scan after the previously seen row if any.
	//
	info->dataOffset = info->endRowOffset;

	//
	// If there is no more data then there is no next row.
	//
	int dataLeft = info->dataSize - info->dataOffset;

	if (dataLeft <= 0)
	{
		info->rowIndex = -1;

		return -1;
	}

	//
	// Find the next ON cell (if any) in the remainder of the data.
	//
	const char * nextOn =
		memchr(info->data + info->dataOffset, 1, dataLeft);

	//
	// If there are no more ON cells then clean up and report that.
	//
	if (nextOn == 0)
	{
		info->dataOffset = info->dataSize;
		info->endRowOffset = info->dataSize;

		return -1;
	}

	//
	// Calculate the row index of the ON cell and calculate the
	// end of the row.
	//
	int foundOffset = (nextOn - info->data);
	int rowIndex = foundOffset / info->colCount;
	int colIndex = foundOffset % info->colCount;

	info->rowIndex = rowIndex;
	info->dataOffset = foundOffset;
	info->endRowOffset = foundOffset + (info->colCount - colIndex);

	return rowIndex;
}


/*
 * Return the index (zero based) of the next column of the current
 * row of the data which has at least one ON cell in it.  This uses
 * the data calculated by a previous qcFindNextRowIndex call.
 * Returns -1 if there are no more columns on the row having ON cells.
 */
int
FindNextColIndex(ViewInfo * info)
{
	//
	// If there is no more data in the row then there is no next column.
	//
	int dataLeft = info->endRowOffset - info->dataOffset;
 
	if (dataLeft <= 0)
		return -1;

	//
	// Find the next ON cell (if any) in the remainder of the
	// current row.
	//
	const char * nextOn =
		memchr(info->data + info->dataOffset, 1, dataLeft);

	//
	// If there are no more ON cells then clean up and report that.
	//
	if (nextOn == 0)
	{
		info->dataOffset = info->endRowOffset;

		return -1;
	}

	//
	// Calculate the column index and adjust the data offset
	// past the found cell.
	//
	int foundOffset = (nextOn - info->data);
	int colIndex = foundOffset % info->colCount;

	info->dataOffset = foundOffset + 1;

	return colIndex;
}


/*
 * Set the state of the cell at the specified row and column index.
 */
void
SetViewCell(ViewInfo * info, int rowIndex, int colIndex, BOOL isSet)
{
	int value = (isSet ? 1 : 0);

	verifyAlways((rowIndex >= 0) && (rowIndex < info->rowCount) &&
		(colIndex >= 0) && (colIndex < info->colCount));

	info->data[rowIndex * info->colCount + colIndex] = value;
}


/*
 * Return whether or not the cell at the indicated row and column index
 * in the specified view data is set.
 */
BOOL
IsViewCellSet(const ViewInfo * info, int rowIndex, int colIndex)
{
	if ((rowIndex < 0) || (rowIndex >= info->rowCount) ||
		(colIndex < 0) || (colIndex >= info->colCount))
	{
		return FALSE;
	}

	return (info->data[rowIndex * info->colCount + colIndex] != 0);
}


/*
 * Clear all of the view data values to zero.
 * This does not affect any auxillary data (if any).
 */
void
ClearViewData(ViewInfo * info)
{
	memset(info->data, 0, info->dataSize);
}


/*
 * Fill all of the view data elements to the specified value.
 * This does not affect any auxillary data (if any).
 */
void
FillViewData(ViewInfo * info, int value)
{
	memset(info->data, value, info->dataSize);
}


/*
 * Return the auxillary data value for the cell at the indicated row
 * and column index.  Returns zero if there is no auxillary data or
 * if the coordinates are illegal.
 */
int
GetAuxViewData(const ViewInfo * info, int rowIndex, int colIndex)
{
	if (info->auxData == 0)
		return 0;

	if ((rowIndex < 0) || (rowIndex >= info->rowCount) ||
		(colIndex < 0) || (colIndex >= info->colCount))
	{
		return 0;
	}

	return info->auxData[rowIndex * info->colCount + colIndex];
}


/*
 * Set the auxillary data value for the cell at the indicated row and
 * column index.  The data value cannot be larger than a char.
 * This can only be called for view info data which actually has auxillary data.
 */
void
SetAuxViewData(ViewInfo * info, int rowIndex, int colIndex, int value)
{
	verifyAlways(info->auxData != 0);

	if ((rowIndex < 0) || (rowIndex >= info->rowCount) ||
		(colIndex < 0) || (colIndex >= info->colCount))
	{
		return;
	}

	info->auxData[rowIndex * info->colCount + colIndex] = value;
}


/*
 * Perform an OR operation for the auxillary data value for the cell at the
 * indicated row and column index.  The data value cannot be larger than a char.
 * This can only be called for view info data which actually has auxillary data.
 */
void
OrAuxViewData(ViewInfo * info, int rowIndex, int colIndex, int value)
{
	verifyAlways(info->auxData != 0);

	if ((rowIndex < 0) || (rowIndex >= info->rowCount) ||
		(colIndex < 0) || (colIndex >= info->colCount))
	{
		return;
	}

	info->auxData[rowIndex * info->colCount + colIndex] |= value;
}


/*
 * Fill all of cells of the auxillary data array with a data value.
 * The data value cannot be larger than a char.  This can only be called
 * for view info data which actually has auxillary data.
 */
void
FillAuxViewData(ViewInfo * info, int value)
{
	verifyAlways(info->auxData != 0);

	memset(info->auxData, value, info->dataSize);
}


/*
 * Scan the main and auxillary values in the specified region and check
 * the auxillary values to see if they all contain all the specified bits.
 * If there is no auxillary data then this just checks the state.
 * If any state is FALSE then FALSE is returned.
 */
BOOL
AreViewDataBitsSet(const ViewInfo * info,
	int minRow, int maxRow, int minCol, int maxCol, int bitValue)
{
	int row, col;

	verifyAlways((minRow >= 0) && (maxRow < info->rowCount) &&
		(minCol >= 0) && (maxCol < info->colCount));

	for (row = minRow; row <= maxRow; row++)
	{
		for (col = minCol; col <= maxCol; col++)
		{
			int index = row * info->colCount + col;

			if (!info->data[index])
				return FALSE;

			if (~info->auxData[index] & bitValue)
				return FALSE;
		}
	}

	return TRUE;
}


/*
 * Pick the minimum non-negative index number from the given values.
 * Any of the values can be negative to not be part of the comparison.
 * This is essentially the min function, except that negative
 * index values are ignored.  Returns a negative value if all index
 * values were negative.
 */
int
PickMinimumIndex(int index1, int index2, int index3, int index4)
{
	int minIndex = index1;

	if ((minIndex < 0) || ((index2 >= 0) && (index2 < minIndex)))
		minIndex = index2;

	if ((minIndex < 0) || ((index3 >= 0) && (index3 < minIndex)))
		minIndex = index3;

	if ((minIndex < 0) || ((index4 >= 0) && (index4 < minIndex)))
		minIndex = index4;

	return minIndex;
}


/*
 * Allocate a temporary or private data array of at least the specified
 * size for the specified id.  If the id is non-negative then the array is
 * only valid until the next call to this routine using the same id and
 * should NOT be freed by the caller.  There are only a small number of
 * allowable ids.  If the id is negative then the data is always allocate
 * and is owned by the caller.
 */
static char *
allocateData(int size, int id)
{
	static char * data[5];
	static int dataSize[5];

	if (id < 0)
		return allocMem(size);

	verifyAlways(id < 5);

	if (size > dataSize[id])
	{
		if (dataSize[id])
			free(data[id]);

		data[id] = 0;
		dataSize[id] = 0;
	}

	if (dataSize[id] == 0)
	{
		data[id] = allocMem(size);
		dataSize[id] = size;
	}

	return data[id];
}


/*
 * Allocate some memory.
 */
static void *
allocMem(int size)
{
	void * data = malloc(size);

	if (data == 0)
	{
		fprintf(stderr, "Cannot allocate viewinfo\n");

		exit(1);
	}

	return data;
}

