/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#ifdef	X11

#include <sys/time.h>
#include <memory.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/StringDefs.h>

#include "life.h"


#define	ALL_MOD_MASKS	(Mod1Mask | Mod2Mask | Mod3Mask | Mod4Mask | Mod5Mask)
#define	ALL_KEY_MASKS	(ShiftMask | LockMask | ControlMask | ALL_MOD_MASKS)

#define	WM_RES_X	20		/* X pixels to reserve for window manager */
#define	WM_RES_Y	30		/* Y pixels to reserve for window manager */
#define	MIN_WINDOW_WIDTH	20	/* minimum window width */
#define	MIN_WINDOW_HEIGHT	40	/* minimum window height */
#define	SUB_WINDOW_GAP	5		/* gap to leave when placing subwindows */
#define	SUB_WINDOW_TRIES	1000	/* tries to find a nice sub-window location */
#define	MIN_SCALE	-100		/* minimum scale */
#define	DEFAULT_SCALE_TOP_LEVEL	-15	/* default scale for top level window */
#define	DEFAULT_SCALE_SUB_WINDOW -5	/* default scale for sub window */
#define	CURSOR_LARGE	3		/* magnification when large cursor shown */
#define	CURSOR_HUGE		8	/* magnification when huge cursor shown */
#define	GRID_MAG	3		/* magnification when grid is shown */
#define	GRID_BIG_MAG	20		/* magnification when big grid is shown */
#define	CIRCLE_MAG	5		/* magnification when circles can be shown */
#define	DIAMOND_MAG	3		/* magnification when diamonds can be shown */
#define	TEXT_X_MARGIN	8		/* margin for text on left and right edges */
#define	TEXT_Y_MARGIN	3		/* margin for text on top and bottom edges */
#define	RESIZE_SIZE	5		/* pixels for resize area in bottom right past border */
#define	MAX_DRAW	500		/* maximum things to draw in one X call */
#define	ASSIGNS		10		/* number of button assignments */
#define	BORDER_WIDTH	3		/* width of border for sub-window */
#define	DEFAULT_MARGIN_PERCENTAGE	10	/* percentage of radius for repositioning */


/*
 * Mouse motion actions which occur while the mouse
 * is moved while being held down.  Only one of these
 * actions can be happening at a time.
 */
typedef enum MotionAction
{
	MOTION_NONE,	/* no movement in progress */
	MOTION_DRAG,	/* selecting of drag area */
	MOTION_MOVE,	/* moving a subwindow */
	MOTION_RESIZE,	/* resizing a subwindow */
	MOTION_POSITION	/* just updating mouse position */
} MotionAction;


/*
 * Actions to perform for a row of cells (a set of flags, even though any
 * particular cell only has one of them set).
 */
#define	CELL_IGNORE	0x00	/* cell is not to be changed */
#define	CELL_CLEAR	0x01	/* cell is to be cleared */
#define	CELL_NORMAL	0x02	/* cell is to be set normal */
#define	CELL_OLD	0x04	/* cell is to be set old */
#define	CELL_MARK	0x08	/* cell is to be set marked */
#define	CELL_TRACK	0x10	/* cell is to be set tracked */
#define	CELL_OVERLAY	0x20	/* cell is to be set as an overlay */


/*
 * A pixel color value.
 */
typedef	unsigned long	Color;


/*
 * Device interface routines for the rest of the life program to call.
 */
static	BOOL	X11OpenDevice(LifeDevice *);
static	void	X11CloseDevice(LifeDevice *);
static	LifeWindow * X11OpenWindow(LifeDevice *, LifeWindow * parent);
static	BOOL	X11CloseWindow(LifeDevice *);
static	void	X11SetStatusMode(LifeDevice *, StatusMode);
static	void	X11AddStatusText(LifeDevice *, const char *);
static	void	X11ClearStatusMessage(LifeDevice *);
static	void	X11SetIsVisible(LifeDevice *, LifeWindow * window, BOOL isVisible);
static	void	X11SetViewModeIsText(LifeDevice *, BOOL);
static	void	X11AddViewText(LifeDevice *, const char *);
static	void	X11Update(LifeDevice *);
static	void	X11Redraw(LifeDevice *);
static	void	X11Delay(LifeDevice *, int);
static	BOOL	X11IsInputReady(LifeDevice *);
static	int	X11ReadChar(LifeDevice *, BOOL);
static	BOOL	X11Assign(LifeDevice *, Value, int);
static	void	X11WriteAssign(LifeDevice *, FILE *);
static	void	X11Beep(LifeDevice *);
static	void	X11DefaultSettings(LifeDevice *, BOOL isTopLevelWindow, Settings * settings);
static	ColorIndex	X11DefineColor(LifeDevice *, const char * name);


static LifeDevice x11Device =
{
	X11OpenDevice, X11CloseDevice, X11OpenWindow,
	X11CloseWindow, X11SetIsVisible, X11SetStatusMode,
	X11AddStatusText, X11ClearStatusMessage,
	X11SetViewModeIsText, X11AddViewText, X11Update,
	X11Redraw, X11Delay, X11IsInputReady,
	X11ReadChar, X11Assign, X11WriteAssign, X11Beep,
	X11DefaultSettings, X11DefineColor
};


/*
 * Private data values for an X window.
 */
typedef struct
{
	Window		mainWid;	/* main window id */
	Window		viewWid;	/* window id for view of cells */
	Window		statWid;	/* window id for status line */
	Pixmap		viewPixmap;	/* pixmap for drawing view */
	PixelCount	statWidth;	/* width of status window */
	PixelCount	statHeight;	/* height of status window */
	PixelCount	viewWidth;	/* width of view window */
	PixelCount	viewHeight;	/* height of view window */
	PixelCount	charHeight;	/* height of characters */
	PixelCount	charAscent;	/* ascent of characters */
	PixelCount	spaceWidth;	/* width of a space */
	PixelCount	usWidth;	/* width of underscore */
	PixelCoord	cursorX;	/* current cursor x location if drawn */
	PixelCoord	cursorY;	/* current cursor y location if drawn */
	PixelCoord	statusTextX;	/* current X offset for new status text */
	NUMBER		dragBeginRow;	/* row when button started drag */
	NUMBER		dragBeginCol;	/* column when button started drag */
	PixelCoord	dragBeginX;	/* x position of begin drag */
	PixelCoord	dragBeginY;	/* y position of begin drag */
	PixelCoord	dragEndX;	/* x position of end drag */
	PixelCoord	dragEndY;	/* y position of end drag */
	PixelCoord	moveOldRootX;	/* old root x position while moving */
	PixelCoord	moveOldRootY;	/* old root y position while moving */
	PixelCount	moveStartWidth;	/* window width when resize started */
	PixelCount	moveStartHeight;	/* window height when resize started */
	MotionAction	motionAction;	/* current mouse motion action */
	Color		statusBackground;	/* status window background color */
	ColorIndex	viewBackgroundIndex;	/* color index of view subwindow */
	BOOL		isMotionAborted;		/* TRUE if motion action is being aborted */
	BOOL		isCursorShown;	/* TRUE if cursor and paths are drawn */
	BOOL		isDragShown;	/* TRUE if drag rectangle is drawn */
	BOOL		isOutlineShown;	/* TRUE if window outlines are shown */
	StringBuffer	textBuffer;	/* allocated text buffer */
	ViewInfo	seenInfo;	/* info about currently seen cells */
} X11Data;


/*
 * Local variables for the X11 device.
 * These values are shared for all window instances.
 */
static	Display *	display;	/* display for X */
static	int		screen;		/* screen number */
static	Window	rootWid;		/* top level window id */

static	GC		statTextGC;	/* GC for status line text */
static	GC		viewTextGC;	/* GC for for text in view window */
static	GC		viewCellGC;	/* GC for normal view of cells */
static	GC		viewMarkGC;	/* GC for view of marked cells */
static	GC		viewTracksGC;	/* GC for tracked cells */
static	GC		viewOldGC;	/* GC for view of old cells */
static	GC		gridGC;		/* GC for grid */
static	GC		checkerGC;	/* GC for checkerboard */
static	GC		cursorGC;	/* GC for cursor */
static	GC		pathGC;		/* GC for paths */
static	GC		whereGC;	/* GC for where lines */
static	GC		dragGC;		/* GC for dragging */
static	GC		outlineGC;	/* GC for window outlines */
static	GC		gridLineGC;	/* GC for grid lines */
static	GC		clearGC;	/* GC for clearing pixmap */
static	GC		overlayGC;	/* GC for overlay object cells */

static	XFontStruct *	font;		/* font for text */

static	Colormap	colorMap;	/* default color map */

static	Color		black;		/* black pixel value */
static	Color		white;		/* white pixel value */
static	Color		red;		/* red pixel value */
static	Color		blue;		/* blue pixel value */
static	Color		grey;		/* grey pixel value */
static	Color		darkGrey;	/* dark grey pixel value */
static	Color		green;		/* green pixel value */
static	Color		yellow;		/* yellow pixel value */
static	Color		darkBlue;	/* dark blue pixel value */
static	Color		darkerBlue;	/* darker blue */
static	Color		darkGreen;	/* dark green pixel value */
static	Color		brown;		/* brown pixel value */

static	Color		colorTable[MAX_COLORS];		/* table of pixel values */

static	ColorIndex	redIndex;	/* index of red */
static	ColorIndex	darkBlueIndex;	/* index of dark blue */
static	ColorIndex	blueIndex;	/* index of blue */
static	ColorIndex	greyIndex;	/* index of grey */

static	ColorIndex	textColorIndex;	/* color index of view text */
static	ColorIndex	cellColorIndex;	/* color index of normal cell */
static	ColorIndex	markColorIndex;	/* color index of marked cell */
static	ColorIndex	tracksColorIndex;	/* color index of tracks */
static	ColorIndex	oldColorIndex;	/* color index of old cells */
static	ColorIndex	clearColorIndex = COLOR_INDEX_BLACK;	/* color index of clearGC */

static	LifeWindow *	lastInputWindow;	/* most recent window of event */
static	PixelCoord	lastInputX;		/* last window X position of event */
static	PixelCoord	lastInputY;		/* last window Y position of event */
static	char *		genStr;			/* current char of generated string */
static	char		genStrBuf[256];		/* generated string to return */
static	char		assigns[ASSIGNS];	/* macro assignments for buttons */
static	char *		cellActions;		/* actions to do for a row of cells */
static	int		maxViewWidth;	/* maximum view width so far */


/*
 * Local procedures.
 */
static	int	DoEvent(const XEvent *);
static	int	DoKeyDown(const XKeyEvent *);
static	void	DoMapped(const XMapEvent *, BOOL isMapped);
static	void	DoExposure(const XExposeEvent *);
static	void	DoFocusIn(const XFocusChangeEvent *);
static	void	DoButtonDown(const XButtonEvent *);
static	void	DoButtonMotion(const XButtonEvent *);
static	int	DoButtonUp(const XButtonEvent *);
static	void	DoConfigure(const XConfigureEvent *);
static	void	ViewNormal(LifeWindow * window);

static	void	ViewNormalRow(LifeWindow * window, int rowIndex,
	ViewInfo * currentInfo, ViewInfo * usedInfo, ViewInfo * overlayInfo,
	ViewInfo * seenInfo);

static	void	ViewScaledRow(LifeWindow * window, int rowIndex,
	ViewInfo * currentInfo, ViewInfo * usedInfo, ViewInfo * overlayInfo);

static ViewInfo	GetOverlayView(const LifeWindow * window, const LifeObject * obj);

static	void	ViewScaled(LifeWindow * window);
static	void	DrawCursor(LifeWindow * window);
static	void	DrawOutlines(LifeWindow * window);
static	void	DrawRow(LifeWindow * window, Coord, int action);
static	void	ClearMotionAction(LifeWindow * window);
static	void	DrawDragRectangle(LifeWindow * window);
static	void	DrawPath(LifeWindow * window);
static	void	ClearRow(LifeWindow * window, Coord row);

static	void	DrawGridDots(LifeWindow * window,
	Coord rowIndex, Coord colIndex, Count rowCount, Count colCount);

static	void	DrawGridLines(LifeWindow * window,
	Coord rowIndex, Coord colIndex, Count rowCount, Count colCount);

static	void	DrawGridCheckerBoard(LifeWindow * window,
	Coord rowIndex, Coord colIndex, Count rowCount, Count colCount);

static	int	AddGridPoints(XPoint * point, PixelCoord xp, PixelCoord yp, BOOL isLarge);

static	void	ClearArea(LifeWindow * window,
	Coord rowIndex, Coord colIndex, Count rowCount, Count colCount);

static void	ClearPixmapArea(LifeWindow * window,
	Coord x, Coord y, Count width, Count height);

static	void	DrawAllWhereLocations(LifeWindow * window);
static	void	DrawWhereLocation(LifeWindow * window, const LocationInfo * info);

static	void	DrawCharacter(LifeWindow * window, int ch,
	NUMBER row, NUMBER col);

static	void	DrawWhereLine(LifeWindow * window,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

static	LifeWindow * FindWindow(Window id);
static	BOOL	SetCurrentWindow(LifeWindow * window);
static	void	SetWindowIsMapped(LifeWindow * window, BOOL isVisible);
static	void	DrawBorder(LifeWindow * window);
static	void	UpdateWindow(LifeWindow * window);
static	void	UpdateCursor(LifeWindow * window);
static	void	UpdateStatusLine(LifeWindow * window);
static	void	ShowView(LifeWindow * window);
static	void	ClearCursor(LifeWindow * window);
static	void	ClearOutlines(LifeWindow * window);
static	BOOL	ConfigureWindow(LifeWindow * window, PixelCount width, PixelCount height);
static	void	SaveMousePosition(const LifeWindow * window, PixelCoord x, PixelCoord y);
static	void	DestroyWindow(LifeWindow * window);
static	ColorIndex	AllocateColor(const char * name, Color * retColor, Color alternate);
static	LifeWindow * GetStatusWindow(LifeWindow * window);
static	Color	FindColor(ColorIndex index);
static	void	SetTextGC(ColorIndex colorIndex);
static	void	SetCellGC(ColorIndex colorIndex);
static	void	SetMarkGC(ColorIndex colorIndex);
static	void	SetTracksGC(ColorIndex colorIndex);
static	void	SetOldGC(ColorIndex colorIndex);
static	void	CalculateWindowPosition(LifeWindow * window);
static	void	DrawViewText(LifeWindow * window, const char * str);
static	void	ShowViewText(LifeWindow * window);
static	BOOL	IsSeenInfoValid(LifeWindow * window);
static	void	CopyPixmapToWindow(LifeWindow * window);
static	void	CreatePixmap(LifeWindow * window);


/*
 * Error handler.
 */
static int
X11ErrorHandler(Display * display, XErrorEvent * event)
{
	fprintf(stderr, "X11 error\n");
	exit(1);
}


/*
 * Return the instance of the X11 device.
 */
LifeDevice *
GetX11Device(void)
{
	return &x11Device;
}


/*
 * Open the X11 graphics screen.
 * This also opens the initial window for the screen.
 * TRUE on success.
 */
static BOOL
X11OpenDevice(LifeDevice * dev)
{
	const char * displayName;

	/*
	 * Try to open the display.
	 */
	displayName = getenv( "DISPLAY" );

	if (displayName == 0)
	{
		fprintf(stderr, "DISPLAY environment variable not set\n");

		return FALSE;
	}

	display = XOpenDisplay(displayName);

	if (display == 0)
	{
		fprintf(stderr, "Cannot open display %s\n", displayName);

		return FALSE;
	}

	dev->minScale = MIN_SCALE;
	dev->maxScale = MAX_SCALE;
	dev->windowCount = 0;

	/*
	 * Get the default screen and the root window id.
	 */
	screen = DefaultScreen(display);
	rootWid = RootWindow(display, screen);

	/*
	 * Get the font to be used for all the windows.
	 */
	font = XLoadQueryFont(display, Life.fontName);

	if (font == 0)
	{
		fprintf(stderr, "Cannot load \"%s\" font\n", Life.fontName);

		return FALSE;
	}

	/*
	 * Get and save the black and white color values.
	 * Their indexes have been predefined in the color table.
	 */
	black = BlackPixel(display, screen);
	white = WhitePixel(display, screen);

	colorTable[COLOR_INDEX_BLACK] = black;
	colorTable[COLOR_INDEX_WHITE] = white;

	/*
	 * Allocate other colors, defaulting them if they are not available.
	 */
	colorMap = DefaultColormap(display, screen);

	darkBlueIndex = AllocateColor("medium blue", &darkBlue, black);
	greyIndex = AllocateColor("grey", &grey, black);
	redIndex = AllocateColor("red", &red, white);
	blueIndex = AllocateColor("blue", &blue, black);

	AllocateColor("green", &green, white);
	AllocateColor("yellow", &yellow, white);
	AllocateColor("#888", &darkGrey, black);
	AllocateColor("#10a", &darkerBlue, black);
	AllocateColor("#070", &darkGreen, black);
	AllocateColor("#770", &brown, white);

	/*
	 * Check for a few colors which we must have.
	 */
	if ((darkBlueIndex < 0) || (greyIndex < 0))
	{
		fprintf(stderr, "Failed to allocate required color\n");

		return FALSE;
	}

	/*
	 * Allocate the graphics contexts based on the root window.
	 * These will work on any of the windows we create since
	 * they will all have the same depth as the root window.
	 */
	statTextGC = XCreateGC(display, rootWid, 0, 0);
	viewTextGC = XCreateGC(display, rootWid, 0, 0);
	viewCellGC = XCreateGC(display, rootWid, 0, 0);
	viewOldGC = XCreateGC(display, rootWid, 0, 0);
	viewMarkGC = XCreateGC(display, rootWid, 0, 0);
	viewTracksGC = XCreateGC(display, rootWid, 0, 0);
	gridGC = XCreateGC(display, rootWid, 0, 0);
	checkerGC = XCreateGC(display, rootWid, 0, 0);
	cursorGC = XCreateGC(display, rootWid, 0, 0);
	pathGC =  XCreateGC(display, rootWid, 0, 0);
	whereGC =  XCreateGC(display, rootWid, 0, 0);
	dragGC = XCreateGC(display, rootWid, 0, 0);
	outlineGC = XCreateGC(display, rootWid, 0, 0);
	gridLineGC = XCreateGC(display, rootWid, 0, 0);
	clearGC = XCreateGC(display, rootWid, 0, 0);
	overlayGC = XCreateGC(display, rootWid, 0, 0);

	//
	// Disable exposure events for copying the pixmap.
	//
	XSetGraphicsExposures(display, clearGC, False);

	/*
	 * Set the initial index values of the changable colors.
	 */
	textColorIndex = COLOR_INDEX_WHITE;
	cellColorIndex = COLOR_INDEX_WHITE;
	markColorIndex = redIndex;
	tracksColorIndex = blueIndex;
	oldColorIndex = greyIndex;

	/*
	 * Set the attributes of the graphics contexts.
	 */
	XSetForeground(display, statTextGC, darkBlue);
	XSetForeground(display, viewTextGC, white);
	XSetForeground(display, viewCellGC, white);
	XSetForeground(display, viewOldGC, grey);
	XSetForeground(display, viewMarkGC, red);
	XSetForeground(display, viewTracksGC, blue);
	XSetForeground(display, gridGC, (grey != black) ? grey : white);
	XSetForeground(display, checkerGC, darkerBlue);
	XSetForeground(display, gridLineGC, black);
	XSetForeground(display, cursorGC, white ^ darkBlue);
	XSetForeground(display, dragGC, white ^ darkBlue);
	XSetForeground(display, outlineGC, darkGrey ^ darkBlue);
	XSetForeground(display, pathGC, green);
	XSetForeground(display, whereGC, darkGreen);
	XSetForeground(display, clearGC, black);
	XSetForeground(display, overlayGC, brown);

	XSetFunction(display, cursorGC, GXxor);
	XSetFunction(display, dragGC, GXxor);
	XSetFunction(display, outlineGC, GXxor);

	XSetFont(display, statTextGC, font->fid);
	XSetFont(display, viewCellGC, font->fid);
	XSetFont(display, viewTextGC, font->fid);
	XSetFont(display, pathGC, font->fid);

	genStr = genStrBuf;

	/*
	 * Allocate some cell actions.
	 * This will be reallocated if necessary later.
	 */
	cellActions = malloc(1024 + 1);

	if (cellActions == 0)
	{
		fprintf(stderr, "Cannot allocate memory\n");

		return FALSE;
	}

	maxViewWidth = 1024;

	/*
	 * Set an error handler.
	 * This is used to catch errors from closed windows.
	 */
	XSetErrorHandler(X11ErrorHandler);

	/*
	 * We are successful.
	 */
	return TRUE;
}


static void
X11CloseDevice(LifeDevice * dev)
{
	/*
	 * TODO: Close all windows.
	 */
	if (display)
		XCloseDisplay(display);
}


/*
 * Fill in the initial default settings for a window
 * according to whether it is a top level window or a subwindow.
 */
static void
X11DefaultSettings(LifeDevice * dev, BOOL isTopLevelWindow, Settings * settings)
{
	/*
	 * Do the settings which are independent of the type of window.
	 */
	settings->sharedFlags = Life.defaultSharedSettings;
	settings->isVisible = TRUE;
	settings->isShowMyOutline = TRUE;
	settings->isShowOutlines = FALSE;
	settings->isGridEnabled = FALSE;
	settings->gridSize = DEFAULT_GRID_SIZE;
	settings->gridChar1 = DEFAULT_GRID_CHAR_1;
	settings->gridChar2 = DEFAULT_GRID_CHAR_2;
	settings->cellShape = SHAPE_SQUARE;
	settings->shiftFactor = DEFAULT_SHIFT;
	settings->marginPercentage = DEFAULT_MARGIN_PERCENTAGE;
	settings->backgroundIndex = darkBlueIndex;
	settings->textColorIndex = COLOR_INDEX_WHITE;
	settings->cellColorIndex = COLOR_INDEX_WHITE;
	settings->markColorIndex = redIndex;
	settings->tracksColorIndex = blueIndex;
	settings->oldColorIndex = greyIndex;
	settings->viewType = VIEW_TYPE_CURSOR;
	settings->positionType = POSITION_TYPE_SHIFT;
	settings->statusLocation = STATUS_LOCATION_SELF;
	settings->isAutoScale = TRUE;

	/*
	 * Do the other settings according to whether the
	 * window is a top level window or a subwindow.
	 */
	if (isTopLevelWindow)
	{
		settings->sharedSettings = Life.topLevelWindowSettings;
		settings->borderWidth = 0;
		settings->defaultScale = DEFAULT_SCALE_TOP_LEVEL;
	}
	else
	{
		settings->sharedSettings = Life.subWindowSettings;
		settings->borderWidth = BORDER_WIDTH;
		settings->defaultScale = DEFAULT_SCALE_SUB_WINDOW;
	}

	/*
	 * If this is one of the first settings allocated then
	 * point the settings at itself.
	 */
	if (settings->sharedSettings == 0)
		settings->sharedSettings = settings;
}


/*
 * Define the specified color for use later.
 * Returns the index of the color if found or
 * COLOR_INDEX_NONE if not known.
 */
static ColorIndex
X11DefineColor(LifeDevice * dev, const char * name)
{
	Color value;

	return AllocateColor(name, &value, black);
}


/*
 * Make a beep sound to indicate an error.
 * This is flushed for a quick effect.
 */
static void
X11Beep(LifeDevice * dev)
{
	XBell(display, 0);
	XFlush(display);
}


/*
 * Clear the message in the status window if required.
 */
static void
X11ClearStatusMessage(LifeDevice * dev)
{
	LifeWindow * window = GetStatusWindow(Life.currentWindow);

	if (window == 0)
		return;

	if ((window->statusMode == STATUS_MODE_ERROR) ||
		(window->statusMode == STATUS_MODE_MESSAGE))
	{
		X11SetStatusMode(dev, STATUS_MODE_NORMAL);
	}
}


/*
 * Set the usage mode for the current window's status sub-window.
 * This lets input lines and messages to be displayed instead
 * of the normal object information.
 */
static void
X11SetStatusMode(LifeDevice * dev, StatusMode mode)
{
	LifeWindow * window = GetStatusWindow(Life.currentWindow);
	X11Data * x11Data;

	if (window == 0)
		return;

	x11Data = (X11Data *) window->privateData;

	/*
	 * Save the status mode and clear the status bufer.
	 */
	window->statusMode = mode;
	window->statusUsed = 0;
	window->statusBuffer[0] = '\0';

	if (mode == STATUS_MODE_NORMAL)
		window->updateFlags |= U_STATUS;

	/*
	 * Clear the window in preparation for text being
	 * added to it.
	 */
	if (window->isMapped)
		XClearWindow(display, x11Data->statWid);
}


/*
 * Add text to the status sub-window for the current window.
 * What this does exactly depends on the current status mode.
 * For command line input mode the text is shown immediately.
 * Other types of output are shown by the normal update methods. 
 */
static void
X11AddStatusText(LifeDevice * dev, const char * str)
{
	LifeWindow * window = GetStatusWindow(Life.currentWindow);
	X11Data * x11Data;
	int	oldLen;
	int newLen;
	PixelCount oldWidth;
	BOOL isInputMode;
	char * buffer;

	if (window == 0)
		return;

	x11Data = (X11Data *) window->privateData;
	buffer = window->statusBuffer;

	/*
	 * Get the old and new lengths of the status text.
	 */
	oldLen = window->statusUsed;
	newLen = strlen(str);

	/*
	 * If the total text (plus the terminating null and
	 * a possible underscore) won't fit then ignore it.
	 */
	if (oldLen + newLen + 2 > sizeof(window->statusBuffer))
		return;

	/*
	 * See if the status window is being used for input.
	 * We will do special handling if so.
	 */
	isInputMode = (window->statusMode == STATUS_MODE_INPUT);

	/*
	 * If this is input mode then remove the trailing
	 * underscore from the old text if present.
	 */
	if (isInputMode && (oldLen > 0) && (buffer[oldLen-1] == '_'))
	{
		buffer[--oldLen] = '\0';
	}

	/*
	 * Append the new text to the buffer.
	 */
	memcpy(buffer + oldLen, str, newLen + 1);

	/*
	 * If this is input mode then append an underscore
	 * to indicate that user input is wanted and add to
	 * the count.
	 */
	if (isInputMode)
	{
		buffer[oldLen + newLen] = '_';
		buffer[oldLen + newLen + 1] = '\0';
		newLen++;
	}

	/*
	 * Update the space used in the buffer.
	 */
	window->statusUsed = oldLen + newLen;

	/*
	 * Do nothing more if the window is not mapped.
	 */
	if (!window->isMapped)
		return;

	/*
	 * Get the length and width of the existing text
	 * in the status buffer.
	 */
	oldWidth = 0;

	if (oldLen > 0)
		oldWidth = XTextWidth(font, buffer, oldLen);

	/*
	 * If we are displaying an input line then clear the
	 * trailing underscore from the previous output.
	 */
	if (isInputMode)
	{
		XClearArea(display, x11Data->statWid,
			TEXT_X_MARGIN + oldWidth, 0, x11Data->usWidth,
			0, False);
	}

	/*
	 * Draw the new text at the appropriate position.
	 * The previous text should still be visible.
	 */
	XDrawString(display, x11Data->statWid, statTextGC,
		TEXT_X_MARGIN + oldWidth,
		x11Data->charAscent + TEXT_Y_MARGIN,
		buffer + oldLen, newLen);

	/*
	 * If this is input mode then flush the output to get
	 * the new characters to show quickly.
	 */
	if (isInputMode)
		XFlush(display);
}


/*
 * Redraw all of the windows to show the latest data.
 */
static void
X11Redraw(LifeDevice * dev)
{
	int index;

	/*
	 * Set the redraw and configuration flags for all windows
	 * then do a normal update.
	 */
	for (index = 0; index < dev->windowCount; index++)
		dev->windowTable[index]->updateFlags |= (U_REDRAW | U_CONFIG);

	X11Update(dev);
}


/*
 * Update all of the windows to show the latest data.
 */
static void
X11Update(LifeDevice * dev)
{
	int	index;

	/*
	 * First update all but the current window.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * window = dev->windowTable[index];

		if (window != Life.currentWindow)
			UpdateWindow(window);
	}

	/*
	 * Now update the current window.
	 * This will make its status be the last one seen.
	 */
	UpdateWindow(Life.currentWindow);

	/*
	 * Flush the display so it is up to date quickly.
	 */
	XFlush(display);
}


/*
 * Read the next character from the terminal, waiting if requested.
 * If not waiting and no character is ready, then EOF is returned.
 * Returns characters from a previously generated string first, if present.
 * Button press events are translated into generated string.
 * This also handles other types of X11 events which are seen.
 * For input events, the current window is set.
 */
static int
X11ReadChar(LifeDevice * dev, BOOL isWaitRequired)
{
	XEvent	event;
	int	ch;

	do
	{
		if (*genStr)
			return *genStr++ & 0xff;

		XSync(display, False);

		if (!isWaitRequired && (XEventsQueued(display, QueuedAfterFlush) <= 0))
			return EOF;

		XNextEvent(display, &event);

		ch = DoEvent(&event);
	}
	while (ch == EOF);

	return ch & 0xff;
}


/*
 * Check whether input is ready from the terminal.  Returns TRUE if so.
 * Input consists of either key presses or button presses, both of which
 * are left in the event queue for later reading.  All other types of
 * X11 events are processed here when seen.
 */
static BOOL
X11IsInputReady(LifeDevice * dev)
{
	XEvent	event;

	while (TRUE)
	{
		if (*genStr)
			return TRUE;

		XSync(display, False);

		if (XEventsQueued(display, QueuedAfterFlush) <= 0)
			return FALSE;

		XPeekEvent(display, &event);

		if ((event.type == KeyPress) || (event.type == ButtonPress))
			return TRUE;

		XNextEvent(display, &event);

		DoEvent(&event);
	}
}


/*
 * Delay for the specified number of milliseconds, or until a keystroke
 * is typed (actually, any X11 event).
 */
static void
X11Delay(LifeDevice * dev, int milliSec)
{
	struct	timeval	timeOut;
	fd_set		readFds;
	int		fd;

	if (milliSec <= 0)
		return;

	fd = ConnectionNumber(display);

	FD_ZERO(&readFds);
	FD_SET(fd, &readFds);

	timeOut.tv_sec = milliSec / 1000;
	timeOut.tv_usec = (milliSec % 1000) * 1000;

	(void) select(fd + 1, &readFds, 0, 0, &timeOut);
}


/*
 * Assign a macro to be automatically executed by a mouse button.
 * Returns TRUE if the button number is legal, or FALSE if it is illegal.
 * The first five numbers represent the unmodified buttons, whereas
 * the second five numbers are for the buttons used with any modifier
 * such as control or shift.
 */
static BOOL
X11Assign(LifeDevice * dev, Value button, int macro)
{
	if ((button <= 0) || (button > ASSIGNS))
		return FALSE;

	assigns[button - 1] = macro;

	return TRUE;
}


/*
 * Write button macro assignments to the specified FILE handle.
 */
static void
X11WriteAssign(LifeDevice * dev, FILE * fp)
{
	int	button;
	int	macro;

	fputs("! button assignments\n", fp);

	for (button = 0; button < ASSIGNS; button++)
	{
		macro = assigns[button];

		if (macro)
			fprintf(fp, ";assign %d %c\n", button + 1, macro);
	}
}


/*
 * Update the cursor position (and the path) to the newest
 * information for the window's current object.
 * This is easy to do because we use XOR to draw it.
 */
static void
UpdateCursor(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	LifeObject * obj = window->currentObject;

	if (!window->isMapped)
		return;

	/*
	 * Remove the cursor if it is present.
	 */
	ClearCursor(window);

	/*
	 * Calculate and save the new position of the cursor.
	 */
	x11Data->cursorY = ScreenCellToPixel(
		window, numSub(obj->currentInfo.row, window->minRow));

	x11Data->cursorX = ScreenCellToPixel(
		window, numSub(obj->currentInfo.col, window->minCol));

	/*
	 * Draw the cursor at the new position.
	 */
	DrawCursor(window);
}


/*
 * Update the status line for the window.
 * This can show the object information or another text
 * message depending on the window's current status mode.
 * This should only be called after the view has been
 * updated since the status information includes the
 * number of cells which are invisible.
 */
static void
UpdateStatusLine(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	const char * str;

	/*
	 * If the window is not mapped or has its status window
	 * set to none then do nothing.
	 */
	if (!window->isMapped ||
		(GetStatusLocation(&window->settings) == STATUS_LOCATION_NONE))
	{
		return;
	}

	/*
	 * Get the color to use for the status background.
	 * This depends on whether the window is the current one.
	 */
	long color = grey;

	if (window != Life.currentWindow)
		color = darkGrey;

	/*
	 * If the status window background color needs
	 * changing then do so.
	 */
	if (x11Data->statusBackground != color)
	{
		x11Data->statusBackground = color;

		XSetWindowBackground(display, x11Data->statWid, color);
	}

	/*
	 * Get the text for the status line.
	 * This depends on the status mode.
	 */
	switch (window->statusMode)
	{
		case STATUS_MODE_NORMAL:
			str = GenerateStatus(window);
			break;

		default:
			str = window->statusBuffer;
			break;
	}

	/*
	 * Move to the correct window for output.
	 */
	window = GetStatusWindow(window);

	if (window == 0)
		return;

	x11Data = (X11Data *) window->privateData;

	/*
	 * Clear the window and draw the new text.
	 */
	XClearWindow(display, x11Data->statWid);

	XDrawString(display, x11Data->statWid, statTextGC, TEXT_X_MARGIN,
		x11Data->charAscent + TEXT_Y_MARGIN, str, strlen(str));
}


/*
 * Update everything in the window that is required.
 */
static void
UpdateWindow(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	LifeObject * obj = window->currentObject;

	//
	// If we are in text mode then make sure the text is visible.
	//
	if (window->isTextView)
	{
		CopyPixmapToWindow(window);

		return;
	}

	/*
	 * Get the current update flags.
	 */
	Flags flags = window->updateFlags;

	/*
	 * If the view subwindow background color has changed
	 * then do that.
	 */
	if (x11Data->viewBackgroundIndex != GetBackgroundIndex(&window->settings))
	{
		x11Data->viewBackgroundIndex = GetBackgroundIndex(&window->settings);

		XSetWindowBackground(display, x11Data->viewWid,
			FindColor(x11Data->viewBackgroundIndex));

		flags |= (U_CLEAR | U_ALL);
	}

	/*
	 * If the window configuration has changed then
	 * redo it.
	 */
	if (flags & U_CONFIG)
	{
		ConfigureWindow(window, window->width, window->height);
	}

	/*
	 * If the object modify count differs from ours
	 * and generations are not being run then set
	 * the view flag and save the new count.
	 * (While running generations we don't update
	 * unless something else prods us to do so such
	 * as the frequency counter reaching zero.)
	 */
	if ((window->lastChangeCount != obj->changeCount) &&
		(obj->genLeft == 0))
	{
		flags |= U_VIEW;
		window->lastChangeCount = obj->changeCount;
	}

	/*
	 * If the view subwindow is to be cleared then do that.
	 */
	if (flags & U_CLEAR)
	{
		ClearPixmapArea(window, 0, 0,
			x11Data->viewWidth, x11Data->viewHeight);

		x11Data->isCursorShown = FALSE;
		x11Data->isDragShown = FALSE;
		x11Data->isOutlineShown = FALSE;

		if (x11Data->seenInfo.isValid > 0)
		{
			FillAuxViewData(&x11Data->seenInfo, CELL_CLEAR);
		}

		flags |= U_ALL;
	}

	/*
	 * If the view is being changed then we will also need
	 * to redraw the outlines.
	 */
	if (flags & U_VIEW)
		flags |= U_OUTLINE;

	/*
	 * If the outlines need to be updated then clear them first.
	 */
	if (flags & U_OUTLINE)
	{
		ClearOutlines(window);
	}

	/*
	 * If the view is to be updated then do that.
	 */
	if (flags & U_VIEW)
	{
		ShowView(window);
	}

	/*
	 * If the cursor position is to be updated then do that.
	 */
	if (flags & (U_POS | U_VIEW))
	{
		UpdateCursor(window);
	}

	/*
	 * If the outlines are to be updated then generate
	 * the new ones and show them.
	 */
	if (flags & U_OUTLINE)
	{
		BuildOutlineTable(window);

		DrawOutlines(window);
	}

	//
	// Now copy the pixmap to the actual window.
	//
	if (flags & (U_VIEW | U_POS))
		CopyPixmapToWindow(window);

	/*
	 * If the status line is to be updated then do that.
	 * This must be done after the view has been updated
	 * since some of the status text depends on the view.
	 */
	if (flags & (U_STATUS | U_VIEW))
	{
		UpdateStatusLine(window);
	}

	/*
	 * Clear the update flags.
	 */
	window->updateFlags = 0;
}


/*
 * Copy the pixmap for the view into the view subwindow.
 */
static void
CopyPixmapToWindow(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	XCopyArea(display, x11Data->viewPixmap, x11Data->viewWid, clearGC,
		0, 0, x11Data->viewWidth, x11Data->viewHeight, 0, 0);
}


/*
 * Set a window to be mapped or not.
 * This maps or unmaps the main window.
 */
static void
SetWindowIsMapped(LifeWindow * window, BOOL isMapped)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	if (window->isMapped == isMapped)
		return;

	window->isMapped = isMapped;

	if (isMapped)
	{
		XMapWindow(display, x11Data->mainWid);
	}
	else
	{
		XUnmapWindow(display, x11Data->mainWid);
	}
}


/*
 * Open a new window on the display with or without a parent.
 * This is set to show the main object or the object in
 * the parent window with default settings.
 * A top level window is specified by a NULL parent.
 * If we have a parent, we are constrained to always fit
 * within the parent window, and we have no status line.
 * Returns the opened window on success, or NULL.
 */
static LifeWindow *
X11OpenWindow(LifeDevice * dev, LifeWindow * parentWindow)
{
	LifeWindow * window;
	X11Data *	x11Data;
	Settings *	settings;
	XSizeHints	sizeHints;
	XWMHints	wmHints;
	Window parentWid;
	PixelCount displayWidth;
	PixelCount displayHeight;
	PixelCount windowWidth;
	PixelCount windowHeight;
	PixelCount borderWidth;
	long eventMask;

	/*
	 * If too many windows are opened then fail.
	 */
	if (dev->windowCount >= MAX_WINDOWS)
		return 0;

	/*
	 * If we want to create ourself as a child of
	 * another child window then fail.
	 */
	if (parentWindow && parentWindow->parentWindow)
		return 0;

	/*
	 * Allocate a new window structure.
	 */
	window = malloc(sizeof(LifeWindow));

	if (window == 0)
		return 0;

	/*
	 * Set everything to zero to begin with.
	 */
	memset(window, 0, sizeof(LifeWindow));

	/*
	 * Remember our parent window (or none).
	 */
	window->device = &x11Device;
	window->parentWindow = parentWindow;
	window->id = dev->windowCount + 1;
	window->isMapped = FALSE;
	window->isShifted = FALSE;

	/*
	 * Initialise our settings.
	 */
	settings = &window->settings;
	settings->sharedSettings = settings;
	settings->sharedFlags = Life.defaultSharedSettings;
	settings->name[0] = '\0';

	/*
	 * Use the parent (or current window) objects
	 * as our own current and previous objects.
	 * If there isn't any then use the main object.
	 */
	if (parentWindow)
	{
		window->currentObject = parentWindow->currentObject;
		window->prevObject = parentWindow->prevObject;
	}
	else if (Life.currentWindow)
	{
		window->currentObject = Life.currentWindow->currentObject;
		window->prevObject = Life.currentWindow->prevObject;
	}
	else
	{
		window->currentObject = Life.mainObject;
		window->prevObject = Life.mainObject;
	}

	/*
	 * Copy the default settings for this window type into our
	 * private settings.
	 */
	if (parentWindow == 0)
	{
		settings->sharedSettings = Life.topLevelWindowSettings;
		RawCopySettings(Life.topLevelWindowSettings, settings);
	}
	else
	{
		settings->sharedSettings = Life.subWindowSettings;
		RawCopySettings(Life.subWindowSettings, settings);
	}

	/*
	 * Allocate the window's private data.
	 */
	x11Data = malloc(sizeof(X11Data));

	if (x11Data == 0)
	{
		free(window);

		return 0;
	}

	window->privateData = x11Data;

	memset(x11Data, 0, sizeof(X11Data));

	/*
	 * Allocate an initial text buffer.
	 */
	InitStringBuffer(&x11Data->textBuffer);

	x11Data->isCursorShown = FALSE;
	x11Data->isDragShown = FALSE;
	x11Data->motionAction = MOTION_NONE;
	x11Data->isMotionAborted = FALSE;

	/*
	 * Get the sizes of characters.
	 * Note: This assumes that all our windows use the same font.
	 */
	x11Data->charAscent = font->ascent;
	x11Data->charHeight = x11Data->charAscent + font->descent;
	x11Data->spaceWidth = font->max_bounds.width;
	x11Data->usWidth = XTextWidth(font, "_", 1);

	/*
	 * Get the window id for the window which is either the root
	 * window, or the view sub-window of the parent window.
	 * Also calculate the window size based on the display size.
	 */
	parentWid = rootWid;
	borderWidth = 0;

	displayWidth = DisplayWidth(display, screen);
	displayHeight = DisplayHeight(display, screen);

	windowWidth = displayWidth;
	windowHeight = displayHeight;

	/*
	 * If this isn't the first window then use the percentage
	 * size value to reduce the window size as a percentage of the
	 * display size.
	 */
	if (dev->windowCount > 0)
	{
		int sizePercentage = Life.topWindowSizePercentage;

		windowWidth = (windowWidth * sizePercentage) / 100;
		windowHeight = (windowHeight * sizePercentage) / 100;
	}

	/*
	 * If this is a sub-window then set its size according to the
	 * percentage of the parent window and give it a border width
	 * for visibility and resizing.
	 */
	if (parentWindow)
	{
		X11Data * parentData = (X11Data *) parentWindow->privateData;
		int sizePercentage = Life.subWindowSizePercentage;

		parentWid = parentData->viewWid;
		
		windowWidth = (parentData->viewWidth * sizePercentage) / 100;
		windowHeight = (parentData->viewHeight * sizePercentage) / 100;
		borderWidth = GetBorderWidth(settings);
	}

	/*
	 * Shrink the window if necessary to account for window manager
	 * borders and panels.
	 */
	if (windowWidth > displayWidth - WM_RES_X)
		windowWidth = displayWidth - WM_RES_X;

	if (windowHeight > displayHeight - WM_RES_Y)
		windowHeight = displayHeight - WM_RES_Y;

	/*
	 * Enforce a minimum window size.
	 */
	if (windowWidth < MIN_WINDOW_WIDTH)
		windowWidth = MIN_WINDOW_WIDTH;

	if (windowHeight < MIN_WINDOW_HEIGHT)
		windowHeight = MIN_WINDOW_HEIGHT;

	/*
	 * Save the window size.
	 */
	window->width = windowWidth;
	window->height = windowHeight;

	/*
	 * Calculate the subwindow sizes.
	 */
	x11Data->viewWidth = windowWidth - borderWidth * 2;
	x11Data->statWidth = windowWidth - borderWidth * 2;
	x11Data->statHeight = x11Data->charHeight + TEXT_Y_MARGIN * 2;
	x11Data->viewHeight = windowHeight - x11Data->statHeight - borderWidth;
	x11Data->statusBackground = grey;
	x11Data->viewBackgroundIndex = GetBackgroundIndex(settings);

	/*
	 * Find a nice position for the window.
	 */
	CalculateWindowPosition(window);

	/*
	 * Create the main window.
	 */
	x11Data->mainWid = XCreateSimpleWindow(display, parentWid,
		window->positionX, window->positionY,
		windowWidth, windowHeight, 0, 0, darkGrey);

	/*
	 * Create the status and view sub-windows in the main window.
	 * Put the view sub-window in its default position under the
	 * status sub-window.
	 */
	x11Data->statWid = XCreateSimpleWindow(display,
		x11Data->mainWid, borderWidth, 0, x11Data->viewWidth,
		x11Data->statHeight, 0, 0, grey);

	x11Data->viewWid = XCreateSimpleWindow(display,
		x11Data->mainWid, borderWidth, x11Data->statHeight,
		x11Data->viewWidth, x11Data->viewHeight, 0, 0,
		FindColor(x11Data->viewBackgroundIndex));

	//
	// Create the initial pixmap for the view subwindow.
	//
	CreatePixmap(window);

	/*
	 * Select for the events we are interested in.
	 * The top level window is slightly different if it is a
	 * child window.
	 */
	eventMask = (ExposureMask | KeyPressMask | FocusChangeMask |
		ButtonPressMask | ButtonMotionMask | ButtonReleaseMask);

	if (parentWindow == 0)
		eventMask |= StructureNotifyMask;

	XSelectInput(display, x11Data->mainWid, eventMask);

	XSelectInput(display, x11Data->statWid, ExposureMask |
		KeyPressMask | FocusChangeMask );

	XSelectInput(display, x11Data->viewWid,
		(ExposureMask | KeyPressMask | ButtonPressMask |
			ButtonReleaseMask | ButtonMotionMask | FocusChangeMask ));

	/*
	 * Set the window manager hints.
	 */
	sizeHints.flags = PSize | PMinSize;
	sizeHints.width = windowWidth;
	sizeHints.height = windowHeight;
	sizeHints.min_width = 25;
	sizeHints.min_height = x11Data->statHeight + 25;

	XSetStandardProperties(display, x11Data->mainWid, "life", "life",
		None, (char **) "", 0, &sizeHints);

	wmHints.flags = InputHint;
	wmHints.input = True;

	XSetWMHints(display, x11Data->mainWid, &wmHints);

	/*
	 * Perform some more configuration for the window.
	 */
	if (!ConfigureWindow(window, windowWidth, windowHeight))
		return 0;

	/*
	 * Set the scale value and thus the viewing area.
	 */
	SetScale(window, GetDefaultScale(settings));

	/*
	 * Map the windows.
	 * We will get exposure events later so that we can draw
	 * the initial contents.
	 */
	XMapWindow(display, x11Data->mainWid);
	XMapWindow(display, x11Data->statWid);
	XMapWindow(display, x11Data->viewWid);

	/*
	 * Add the window to the table.
	 */
	dev->windowTable[dev->windowCount] = window;
	dev->windowCount++;

	/*
	 * If there is no current window then make this be it.
	 */
	if (Life.currentWindow == 0)
		Life.currentWindow = window;

	/*
	 * Return the opened window.
	 */
	return window;
}


/*
 * Create the pixmap for the view subwindow's data,
 * either for the initial creation of the window or for a resize.
 * Any previous pixmap must be deleted first.
 * The pixmap is cleared to the window's background color.
 */
static void
CreatePixmap(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	x11Data->viewPixmap = XCreatePixmap(display, x11Data->viewWid,
		x11Data->viewWidth, x11Data->viewHeight,
		DefaultDepth(display, DefaultScreen(display)));

	ClearPixmapArea(window, 0, 0, x11Data->viewWidth, x11Data->viewHeight);
}


/*
 * Find a good initial window position for the window.
 * This only does something useful for sub-windows.
 * This assumes that the initial window size is already set
 * and that the window is not yet in the window table.
 */
static void
CalculateWindowPosition(LifeWindow * window)
{
	LifeWindow * parentWindow = window->parentWindow;
	X11Data * parentData;
	PixelCoord minX;
	PixelCoord maxX;
	PixelCoord minY;
	PixelCoord maxY;
	PixelCoord rangeX;
	PixelCoord rangeY;
	PixelCoord tryX;
	PixelCoord tryY;
	long areaRange;
	int tryCount;

	/*
	 * Start with a position at the upper left.
	 */
	window->positionX = 0;
	window->positionY = 0;

	/*
	 * If this is a top-level window then that's good enough.
	 */
	if (parentWindow == 0)
		return;

	/*
	 * This is a subwindow.
	 * Calculate the coordinates where the top left corner
	 * of the sub-window could appear without the rest of
	 * the window being off of the parent window.
	 * Try to leave a small gap for niceness.
	 */
	parentData = (X11Data *) parentWindow->privateData;

	minX = SUB_WINDOW_GAP;
	minY = SUB_WINDOW_GAP;
	maxX = parentData->viewWidth - window->width - minX;
	maxY = parentData->viewHeight - window->height - minY;

	if (maxX < minX)
	{
		minX = 0;
		maxX = 0;
	}

	if (maxY < minY)
	{
		minY = 0;
		maxY = 0;
	}

	/*
	 * Try to calculate a random position for the window
	 * which doesn't overlap other subwindows.  First
	 * calculate the number of pixels to choose from.
	 */
	rangeX = (maxX - minX + 1);
	rangeY = (maxY - minY + 1);

	areaRange = ((long)rangeX) * rangeY;

	/*
	 * Now calculate the random positions and test them.
	 * If we don't find a nice one then the last position
	 * will be used anyway.
	 */
	for (tryCount = 0; tryCount < SUB_WINDOW_TRIES; tryCount++)
	{
		long value = random() % areaRange;

		tryX = value / rangeY;
		tryY = value % rangeY;

		/*
		 * If the intended location overlaps any
		 * existing subwindow then we don't want it.
		 */
		if (IsOverlappingSubwindow(parentWindow,
			tryX - SUB_WINDOW_GAP, tryY - SUB_WINDOW_GAP,
			window->width + 2 * SUB_WINDOW_GAP, 
			window->height + 2 * SUB_WINDOW_GAP))
		{
			continue;
		}

		/*
		 * The window doesn't overlap any subwindows.
		 * But see if the mouse position will be within it.
		 * If not, then we can use it.
		 */
		if ((lastInputWindow != parentWindow) ||
			(lastInputX < tryX) || (lastInputY < tryY) ||
			(lastInputX > tryX + window->width) ||
			(lastInputY > tryY + window->height))
		{
			break;
		}
	}

	/*
	 * Set the found position.
	 */
	window->positionX = tryX;
	window->positionY = tryY;
}


/*
 * Set values dependent on the specified new window size and
 * on whether or not the status window is being displayed.
 */
static BOOL
ConfigureWindow(LifeWindow * window, PixelCount windowWidth, PixelCount windowHeight)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	PixelCount borderWidth;
	PixelCoord viewY;
	BOOL isStatusShown;

	/*
	 * Enforce a minimum size.
	 */
	if (windowWidth < MIN_WINDOW_WIDTH)
		windowWidth = MIN_WINDOW_WIDTH;

	if (windowHeight < MIN_WINDOW_HEIGHT)
		windowHeight = MIN_WINDOW_HEIGHT;

	/*
	 * If the view is growing, then reallocate the actions table.
	 * If that fails, just ignore the reconfiguration.
	 */
	if (windowWidth > maxViewWidth)
	{
		char * newActions = realloc(cellActions, windowWidth + 1);

		if (newActions == 0)
			return FALSE;

		cellActions = newActions;

		maxViewWidth = windowWidth;
	}

	/*
	 * Determine whether or not our status text is being
	 * shown in our own window.
	 */
	isStatusShown =
		(GetStatusLocation(&window->settings) == STATUS_LOCATION_SELF);

	/*
	 * Calculate the subwindow dimensions.
	 * If we have a parent then we have a border.
	 */
	borderWidth = 0;

	if (window->parentWindow)
	{
		borderWidth = GetBorderWidth(&window->settings);

		if (windowWidth <= 2 * borderWidth)
			borderWidth = 0;
	}

	/*
	 * Update some sizes.
	 */
	window->width = windowWidth;
	window->height = windowHeight;

	x11Data->viewWidth = windowWidth - borderWidth * 2;
	x11Data->statWidth = windowWidth - borderWidth * 2;
	x11Data->statHeight = x11Data->charHeight + TEXT_Y_MARGIN * 2;

	/*
	 * Calculate the view window position and height according
	 * to whether the status window is being shown.
	 */
	if (isStatusShown)
	{
		x11Data->viewHeight = windowHeight - x11Data->statHeight - borderWidth;
		viewY = x11Data->statHeight;
	}
	else
	{
		x11Data->viewHeight = windowHeight - borderWidth * 2;
		viewY = borderWidth;
	}

	if (x11Data->viewHeight <= 0)
		x11Data->viewHeight = 1;

	/*
	 * Recalculate the number of rows and columns of text the
	 * view window can display.
	 */
	window->textRows = (x11Data->viewHeight - TEXT_Y_MARGIN * 2) / x11Data->charHeight;
	window->textCols = (x11Data->viewWidth - TEXT_X_MARGIN * 2) / x11Data->spaceWidth;

	/*
	 * Update the number of rows and columns available for individual cells.
	 */
	window->rows = x11Data->viewHeight;
	window->cols = x11Data->viewWidth;

	/*
	 * Recenter the view around the window's view position.
	 * This will recalculate all of the row numbers that are visible
	 * according to the current window size and scale value.
	 */
	CenterView(window);

	/*
	 * Map or unmap the status window according to whether it is shown.
	 */
	if (isStatusShown)
		XMapWindow(display, x11Data->statWid);
	else
		XUnmapWindow(display, x11Data->statWid);

	/*
	 * If this is a child window then resize the main window since we are
	 * responsible for it.  (Top level windows are resized for us by
	 * window managers.)
	 */
	if (window->parentWindow)
		XResizeWindow(display, x11Data->mainWid, windowWidth, windowHeight);

	/*
	 * Resize the sub-windows.
	 */
	XResizeWindow(display, x11Data->statWid, x11Data->statWidth, x11Data->statHeight);
	XResizeWindow(display, x11Data->viewWid, x11Data->viewWidth, x11Data->viewHeight);

	/*
	 * Move the view sub-window to the correct position according
	 * to whether or not the status sub-window is being shown.
	 */
	XMoveWindow(display, x11Data->viewWid, borderWidth, viewY);

	/*
	 * Indicate a full redraw is needed.
	 */
	window->updateFlags |= U_REDRAW;

	/*
	 * Redraw the border to be sure.
	 */
	DrawBorder(window);

	/*
	 * Tell other windows that our outline has changed.
	 */
	AddOtherWindowOutlineFlags(window);

	//
	// Free the old pixmap for the view window and make a new
	// pixmap of the correct size.
	//
	XFreePixmap(display, x11Data->viewPixmap);

	CreatePixmap(window);

	return TRUE;
}


/*
 * Utility routine to destroy the specified window.
 * This doesn't check whether it is legal or whether there
 * are children windows.
 */
static void
DestroyWindow(LifeWindow * window)
{
	LifeDevice * dev = window->device;
	X11Data * x11Data = (X11Data *) window->privateData;
	int lastIndex;
	int index;

	//
	// Free the pixmap.
	//
	XFreePixmap(display, x11Data->viewPixmap);

	/*
	 * Destroy the X window ids.
	 * Note: We only have to destroy the main window id
	 * since this will also destroy its child windows.
	 */
	XDestroyWindow(display, x11Data->mainWid);

	/*
	 * Remove the window from the window table by moving
	 * the last entry on top of our own.
	 */
	lastIndex = dev->windowCount - 1;
	index = lastIndex;

	while ((index >= 0) && (dev->windowTable[index] != window))
	{
		index--;
	}

	if (index >= 0)
	{
		dev->windowTable[index] = dev->windowTable[lastIndex];
	}

	dev->windowCount--;

	/*
	 * Finally free the memory.
	 */
	FreeView(&x11Data->seenInfo);

	TermStringBuffer(&x11Data->textBuffer);

	free(x11Data);
	free(window);
}


static BOOL
X11CloseWindow(LifeDevice * dev)
{
	LifeWindow * window = Life.currentWindow;
	LifeWindow * newCurrentWindow;
	int index;

	/*
	 * See if there is a new top-level window available
	 * to be made the current window.  If not then fail.
	 */
	newCurrentWindow = FindOtherTopLevelWindow(window);

	if (newCurrentWindow == 0)
		return FALSE;

	/*
	 * Set the new current window.
	 */
	Life.currentWindow = newCurrentWindow;

	newCurrentWindow->updateFlags |= (U_OUTLINE | U_STATUS);

	/*
	 * Look for any child windows of us and destroy them first.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * testWindow = dev->windowTable[index];

		if (testWindow->parentWindow == window)
			DestroyWindow(testWindow);
	}

	/*
	 * Now destroy our window.
	 */
	DestroyWindow(window);

	return TRUE;
}


/*
 * Draw the cursor and the current path at the set position.
 * If the scale factor is large enough, then the cursor is made larger.
 * The cursor is drawn using XOR, so that if the call is made
 * twice for the same position, it will be removed.
 */
static void
DrawCursor(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	Drawable drawable = x11Data->viewPixmap;
	PixelCoord x = x11Data->cursorX;
	PixelCoord y = x11Data->cursorY;

	/*
	 * Toggle the cursor is shown flag.
	 */
	x11Data->isCursorShown = !x11Data->isCursorShown;

	/*
	 * Draw the path for the object if any.
	 */
	DrawPath(window);

	/*
	 * Draw the center point for the cursor.
	 */
	XDrawPoint(display, drawable, cursorGC, x, y);

	/*
	 * If the scale factor is not large then do no more.
	 */
	if (window->scale > -CURSOR_LARGE)
		return;

	/*
	 * Draw other nearby points to create a cross.
	 */
	XDrawPoint(display, drawable, cursorGC, x - 1, y);
	XDrawPoint(display, drawable, cursorGC, x + 1, y);
	XDrawPoint(display, drawable, cursorGC, x, y - 1);
	XDrawPoint(display, drawable, cursorGC, x, y + 1);

	/*
	 * If the scale factor is not even larger then
	 * do no more.
	 */
	if (window->scale > -CURSOR_HUGE)
		return;

	/*
	 * Draw other nearby points to create a larger cross.
	 */
	XDrawPoint(display, drawable, cursorGC, x - 2, y);
	XDrawPoint(display, drawable, cursorGC, x - 3, y);
	XDrawPoint(display, drawable, cursorGC, x + 2, y);
	XDrawPoint(display, drawable, cursorGC, x + 3, y);
	XDrawPoint(display, drawable, cursorGC, x, y - 2);
	XDrawPoint(display, drawable, cursorGC, x, y - 3);
	XDrawPoint(display, drawable, cursorGC, x, y + 2);
	XDrawPoint(display, drawable, cursorGC, x, y + 3);
}


/*
 * Clear the current cursor if it is being displayed.
 */
static void
ClearCursor(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	if (x11Data->isCursorShown)
		DrawCursor(window);
}


/*
 * Clear any mouse motion action.
 */
static void
ClearMotionAction(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	if (x11Data->isDragShown)
	{
		DrawDragRectangle(window);
		CopyPixmapToWindow(window);
	}

	x11Data->motionAction = MOTION_NONE;
}


/*
 * Clear the current outline rectangles if they are being displayed.
 */
static void
ClearOutlines(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	if (x11Data->isOutlineShown)
		DrawOutlines(window);
}


/*
 * Draw the window outlines on the specified window.
 * This is drawn using XOR so calling this again will
 * remove the outlines.
 */
static void
DrawOutlines(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	Drawable drawable = x11Data->viewPixmap;
	const Area * area;
	PixelCoord minX;
	PixelCoord maxX;
	PixelCoord minY;
	PixelCoord maxY;
	int index;

	x11Data->isOutlineShown = !x11Data->isOutlineShown;

	if (!window->isMapped)
		return;

	for (index = 0; index < window->mOutlineCount; index++ )
	{
		area = &window->mOutlineTable[index];

		/*
		 * If the area is outside of the window view area
		 * then we don't have to draw it.
		 */
		if (numIsLess(area->maxRow, window->minRow) ||
			numIsGreater(area->minRow, window->maxRow) ||
			numIsLess(area->maxCol, window->minCol) ||
			numIsGreater(area->minCol, window->maxCol))
		{
			continue;
		}

		/*
		 * Translate the cell coordinates to pixel coordinates.
		 */
		minX = ScreenCellToPixel(window,
			numSub(area->minCol, window->minCol));

		maxX = ScreenCellToPixel(window,
			numSub(area->maxCol, window->minCol));

		minY = ScreenCellToPixel(window,
			numSub(area->minRow, window->minRow));

		maxY = ScreenCellToPixel(window,
			numSub(area->maxRow, window->minRow));

		/*
		 * If the scale is negative then adjust the coordinates from
		 * the centers to the edges of the cells so that the area
		 * is nicely framed.
		 */
		if (window->scale < -1)
		{
			int offset = (-window->scale) / 2;

			minX -= offset;
			maxX += offset;
			minY -= offset;
			maxY += offset;
		}

		/*
		 * Draw the lines for the edges of the rectangular
		 * area where they would be visible.
		 */
		if (minY >= 0)
		{
			XDrawLine(display, drawable, outlineGC,
				minX, minY, maxX, minY);
		}

		if (maxY < x11Data->viewHeight)
		{
			XDrawLine(display, drawable, outlineGC,
				minX, maxY, maxX, maxY);
		}

		if (minX >= 0)
		{
			XDrawLine(display, drawable, outlineGC,
				minX, minY, minX, maxY);
		}

		if (maxX < x11Data->viewWidth)
		{
			XDrawLine(display, drawable, outlineGC,
				maxX, minY, maxX, maxY);
		}
	}
}


/*
 * Draw the current drag region on the specified window.
 * This is drawn using XOR so calling this again will remove the drag
 * rectangle.  The drawing is to the pixmap, which will eventually need to
 * be copied to the window to be seen.
 */
static void
DrawDragRectangle(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	Drawable drawable = x11Data->viewPixmap;
	PixelCoord	x1;
	PixelCoord	x2;
	PixelCoord	y1;
	PixelCoord	y2;

	x11Data->isDragShown = !x11Data->isDragShown;

	/*
	 * Get the endpoints of the rectangle such that x1 is not
	 * more then x2 and y1 is not more than y2.
	 */
	x1 = ((x11Data->dragBeginX <= x11Data->dragEndX) ?
		x11Data->dragBeginX : x11Data->dragEndX);

	x2 = ((x11Data->dragBeginX <= x11Data->dragEndX) ?
		x11Data->dragEndX : x11Data->dragBeginX);

	y1 = ((x11Data->dragBeginY <= x11Data->dragEndY) ?
		x11Data->dragBeginY : x11Data->dragEndY);

	y2 = ((x11Data->dragBeginY <= x11Data->dragEndY) ?
		x11Data->dragEndY : x11Data->dragBeginY);

	/*
	 * Check for a point sized rectangle.
	 */
	if ((x1 == x2) && (y1 == y2))
	{
		XDrawPoint(display, drawable, dragGC, x1, y1);

		return;
	}

	/*
	 * Check for a zero thickness rectangle.
	 */
	if ((x1 == x2) || (y1 == y2))
	{
		XDrawLine(display, drawable, dragGC, x1, y1, x2, y2);

		return;
	}

	/*
	 * The rectangle has all four sides.
	 */
	XDrawLine(display, drawable, dragGC, x1, y1, x2 - 1, y1);
	XDrawLine(display, drawable, dragGC, x1, y2, x2 - 1, y2);
	XDrawLine(display, drawable, dragGC, x1, y1, x1, y2 - 1);
	XDrawLine(display, drawable, dragGC, x2, y1, x2, y2 - 1);
}


/*
 * Show the currently defined path on the window, if any.
 * The path is drawn using XOR so that if this is called again
 * the path will be removed.  The endpoints of each segment of
 * the path are in the centers of the cells.  The path can
 * actually consist of more than one loop of segments.
 */
static void
DrawPath(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	Drawable drawable = x11Data->viewPixmap;
	const LifeObject * obj = window->currentObject;
	const Path * path = &obj->path;

	if (!window->isMapped)
		return;

	NUMBER minRow = window->minRow;
	NUMBER minCol = window->minCol;

	int index = 0;

	while (index + 1 < path->count)
	{
		if (path->locs[index].flag)
		{
			index++;

			continue;
		}

		PixelCoord x1 = ScreenCellToPixel(
			window, numSub(path->locs[index].col, minCol));

		PixelCoord y1 = ScreenCellToPixel(
			window, numSub(path->locs[index].row, minRow));

		index++;

		PixelCoord x2 = ScreenCellToPixel(
			window, numSub(path->locs[index].col, minCol));

		PixelCoord y2 = ScreenCellToPixel(
			window, numSub(path->locs[index].row, minRow));

		XDrawLine(display, drawable, pathGC, x1, y1, x2, y2);
	}
}


/*
 * Redraw the border of the window.
 * This does nothing for top level windows.
 * This just clears the window and then paints a
 * resize area at the bottom right corner.
 */
static void
DrawBorder(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	Window wid = x11Data->mainWid;
	PixelCount borderWidth;

	if (window->parentWindow == 0)
		return;

	/*
	 * Clear the main window to make the border be the
	 * background color.
	 */
	XClearWindow(display, wid);

	/*
	 * Draw the bottom corner to mark the resize area.
	 */
	borderWidth = GetBorderWidth(&window->settings);

	XFillRectangle(display, wid, pathGC,
		window->width - RESIZE_SIZE - borderWidth - 1,
		window->height - RESIZE_SIZE - borderWidth - 1,
		window->width, window->height);
}


/*
 * Redraw the text in the specified subwindow.
 */
static void
ShowViewText(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	/*
	 * Draw all of the text into the pixmap and then to the screen.
	 */
	DrawViewText(window, GetStringBufferText(&x11Data->textBuffer));

	CopyPixmapToWindow(window);
}


/*
 * Draw the specified text into the view pixmap.
 * We must interpret certain control codes.
 * The pixmap needs to be copied to the view window to be seen.
 */
static void
DrawViewText(LifeWindow * window, const char * str)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	const char * cp = str;
	int textCol = 0;
	int textX = TEXT_X_MARGIN;
	int textY = x11Data->charAscent + TEXT_Y_MARGIN;

	/*
	 * Make sure the text is in the right color.
	 */
	SetTextGC(GetTextColorIndex(&window->settings));

	/*
	 * Scan the string looking for lines.
	 */
	for (;;)
	{
		while (*cp >= ' ')
			cp++;

		int len = cp - str;

		if (len > 0)
		{
			XDrawString(display, x11Data->viewPixmap, viewTextGC,
				textX, textY, str, len);

			textCol += len;
			textX += XTextWidth(font, str, len);
		}

		switch (*cp++)
		{
			case '\0':
				return;

			case '\n':
				textCol = 0;
				textX = TEXT_X_MARGIN;
				textY += x11Data->charHeight;

				break;

			case '\t':
				len = 8 - (textCol + 1) % 8;

				XDrawString(display, x11Data->viewPixmap,
					viewTextGC, textX, textY,
					"        ", len);

				textCol += len;
				textX += x11Data->spaceWidth * len;

				break;
		}

		str = cp;
	}
}


/*
 * Show the view around the current location in a window.
 */
static void
ShowView(LifeWindow * window)
{
	if (!window->isMapped)
		return;

	/*
	 * If we are in text mode then redraw the text.
	 */
	if (window->isTextView)
	{
		ShowViewText(window);

		return;
	}

	/*
	 * Set the colors for this window.
	 */
	SetTextGC(GetTextColorIndex(&window->settings));
	SetCellGC(GetCellColorIndex(&window->settings));
	SetMarkGC(GetMarkColorIndex(&window->settings));
	SetTracksGC(GetTracksColorIndex(&window->settings));
	SetOldGC(GetOldColorIndex(&window->settings));

	/*
	 * Now draw the view according to the scale factor.
	 */
	if (window->scale > 1)
		ViewScaled(window);
	else
		ViewNormal(window);
}


/*
 * Set whether or not text is being viewed in the
 * current view subwindow.  If this is called to set
 * the view to text then the window is cleared.
 */
static void
X11SetViewModeIsText(LifeDevice * dev, BOOL isText)
{
	LifeWindow * window = Life.currentWindow;
	X11Data * x11Data = (X11Data *) window->privateData;

	/*
	 * If the window is in the normal view and that is
	 * what we want then do nothing.
	 */
	if (!isText && !window->isTextView)
		return;

	/*
	 * The usage is changing.
	 * Save the view state and set the update flag.
	 */
	window->isTextView = isText;
	window->updateFlags |= U_REDRAW;

	/*
	 * Set up initial state for the text.
	 */
	x11Data->isCursorShown = FALSE;
	x11Data->motionAction = MOTION_NONE;
	x11Data->isMotionAborted = FALSE;

	ClearStringBuffer(&x11Data->textBuffer);

	/*
	 * Clear the window in preparation.
	 */
	ClearPixmapArea(window, 0, 0, x11Data->viewWidth, x11Data->viewHeight);
}


/*
 * Add information to the current view sub-window.
 * We must also save the text in a buffer for exposure handling.
 */
static void
X11AddViewText(LifeDevice * dev, const char * str)
{
	LifeWindow * window = Life.currentWindow;
	X11Data * x11Data = (X11Data *) window->privateData;

	//
	// Save the new text and display it.
	//
	AppendToStringBuffer(&x11Data->textBuffer, str);

	DrawViewText(window, GetStringBufferText(&x11Data->textBuffer));
}


/*
 * Set the specified sub-window visible or not, or else
 * set all the sub-windows of the window visible or not.
 */
static void
X11SetIsVisible(LifeDevice * dev, LifeWindow * window, BOOL isVisible)
{
	int index;

	/*
	 * If this is a sub-window then set just its state.
	 */
	if (window->parentWindow)
	{
		SetWindowIsMapped(window, isVisible);

		return;
	}

	/*
	 * This is a top-level window.
	 * Set the state of all its sub-windows.
	 */
	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * testWindow = dev->windowTable[index];

		if (testWindow->parentWindow == window)
			SetWindowIsMapped(testWindow, isVisible);
	}
}


/*
 * Handle the specified event.
 * Returns a character generated by the event, or else
 * EOF if the event does not produce a character.
 * Some events change the current window and so the
 * current object.
 */
static int
DoEvent(const XEvent * event)
{
	switch (event->type)
	{
		case FocusIn:
			DoFocusIn(&event->xfocus);

			if (Life.canMakeInput)
				return NULL_CMD;

			break;

		case Expose:
			DoExposure(&event->xexpose);

			if (Life.canMakeInput)
				return NULL_CMD;

			break;

		case ButtonPress:
			DoButtonDown(&event->xbutton);

			break;

		case MotionNotify:
			DoButtonMotion(&event->xbutton);

			break;

		case ButtonRelease:
			return DoButtonUp(&event->xbutton);

		case KeyPress:
			return DoKeyDown(&event->xkey);

		case MapNotify:
			DoMapped(&event->xmap, TRUE);

			break;

		case UnmapNotify:
			DoMapped(&event->xmap, FALSE);

			break;

		case ConfigureNotify:
			DoConfigure(&event->xconfigure);

			if (Life.canMakeInput)
				return NULL_CMD;

			break;
	}

	return EOF;
}


/*
 * Find our window structure which contains the specified window id.
 * The id can be either for the top level window or the status
 * subwindow or the view subwindow.  Returns 0 if the window id
 * is not known.
 */
static LifeWindow *
FindWindow(Window wid)
{
	const LifeDevice * dev = &x11Device;
	int	index;

	for (index = 0; index < dev->windowCount; index++)
	{
		LifeWindow * window = dev->windowTable[index];
		const X11Data * x11Data = (const X11Data *) window->privateData;

		if (wid == x11Data->mainWid)
			return window;

		if (wid == x11Data->statWid)
			return window;

		if (wid == x11Data->viewWid)
			return window;
	}

	return 0;
}


/*
 * Set the current window to the one specified.
 * This will therefore also set the current object to the
 * one being displayed in the window.
 * Returns TRUE if the window was changed.
 */
static BOOL
SetCurrentWindow(LifeWindow * newWindow)
{
	LifeWindow * oldWindow = Life.currentWindow;
	X11Data * newX11Data = (X11Data *) newWindow->privateData;
	X11Data * oldX11Data = (X11Data *) oldWindow->privateData;

	/*
	 * If the window is not being changed then do nothing.
	 */
	if (newWindow == oldWindow)
		return FALSE;

	/*
	 * The window is being changed.
	 * Indicate to update the old and new status lines
	 * since their background color changes.
	 */
	newWindow->updateFlags |= U_STATUS;
	oldWindow->updateFlags |= U_STATUS;

	/*
	 * If either window has a parent then change the
	 * main backgrounds to match the status background.
	 * This will make their borders change.
	 */
	if (newWindow->parentWindow)
	{
		XSetWindowBackground(display, newX11Data->mainWid, grey);
		XSetWindowBackground(display, newX11Data->statWid, grey);
		DrawBorder(newWindow);
		XClearWindow(display, newX11Data->statWid);
	}

	if (oldWindow->parentWindow)
	{
		XSetWindowBackground(display, oldX11Data->mainWid, darkGrey);
		XSetWindowBackground(display, oldX11Data->statWid, darkGrey);
		DrawBorder(oldWindow);
		XClearWindow(display, oldX11Data->statWid);
	}

	/*
	 * Update the current window and object in the
	 * global data.
	 */
	Life.currentWindow = newWindow;
	Life.currentObject = newWindow->currentObject;

	/*
	 * Clear any characters that are waiting to be
	 * read since they were for the old window and we
	 * don't want them applied to the new one!
	 */
	genStr = genStrBuf;
	genStrBuf[0] = '\0';

	ScanReset();

	/*
	 * Return indicating the window was changed.
	 */
	return TRUE;
}


/*
 * Convert a mouse position in the view sub-window into
 * row and column values in the window's current object
 * and save them in the object for later reference.
 */
static void
SaveMousePosition(const LifeWindow * window, PixelCoord x, PixelCoord y)
{
	LifeObject * obj = window->currentObject;
	int index;

	/*
	 * Calculate the row and column in the view
	 * that the button position is on.
	 */
	NUMBER buttonRow = PixelToScreenCell(window, y);
	NUMBER buttonCol = PixelToScreenCell(window, x);

	/*
	 * Add in the coordinates of the top left cell
	 * of the window.
	 */
	numAddrAdd(&buttonRow, window->minRow);
	numAddrAdd(&buttonCol, window->minCol);

	/*
	 * Save the mouse position in the object.
	 */
	obj->mouseRow = buttonRow;
	obj->mouseCol = buttonCol;

	/*
	 * Scan through the windows to find the ones which
	 * want to keep track of the mouse position and
	 * reposition them.
	 */
	for ( index = 0; index < Life.device->windowCount; index++ )
	{
		LifeWindow * testWindow = Life.device->windowTable[index];

		if (testWindow == window)
			continue;

		if (testWindow->currentObject != obj)
			continue;

		if (GetViewType(&testWindow->settings) != VIEW_TYPE_MOUSE)
			continue;

		PositionView(testWindow);
	}

	UpdateAllWindows();
}


/*
 * Handle a map or unmap event.
 * We trap these so as to avoid doing drawing when the window
 * is not visible.  Nothing else needs doing here, since when the
 * window is mapped exposures will also occur.
 */
static void
DoMapped(const XMapEvent * mp, BOOL isMapped)
{
	LifeWindow * window = FindWindow(mp->window);

	if (window == 0)
		return;

	window->isMapped = isMapped;
}


/*
 * Handle a focus in event.
 * This sets the window to be the current one.
 * As a side effect of this the status sub-window
 * changes color to indicate the focus change.
 */
static void
DoFocusIn(const XFocusChangeEvent * fp)
{
	LifeWindow * window = FindWindow(fp->window);

	if (window)
		SetCurrentWindow(window);
}


/*
 * Handle an exposure event.
 * The appropriate sub-windows are redrawn.
 */
static void
DoExposure(const XExposeEvent * ep)
{
	LifeWindow * window;
	X11Data * x11Data;
	Window	wid;

	//
	// If there are more exposures coming then ignore this one.
	//
	if (ep->count > 0)
		return;

	wid = ep->window;
	window = FindWindow(wid);

	if (window == 0)
		return;

	window->isMapped = TRUE;

	x11Data = (X11Data *) window->privateData;

	if (wid == x11Data->mainWid)
		DrawBorder(window);

	if (wid == x11Data->statWid)
		window->updateFlags |= U_STATUS;

	if (wid == x11Data->viewWid)
		CopyPixmapToWindow(window);
}


/*
 * Handle a button down event.
 * This remembers the current position in the object for use when
 * the button is released, and initialises a drag rectangle at the
 * current position.  This is only done if canMakeInput is true.
 */
static void
DoButtonDown(const XButtonEvent * bp)
{
	Window	wid;
	LifeWindow * window;
	X11Data * x11Data;
	NUMBER	buttonRow;
	NUMBER	buttonCol;

	/*
	 * Find the window which has the button press.
	 */
	wid = bp->window;
	window = FindWindow(wid);

	if (window == 0)
		return;

	lastInputWindow = window;
	lastInputX = bp->x;
	lastInputY = bp->y;

	x11Data = (X11Data *) window->privateData;

	/*
	 * Set the current window to this one.
	 */
	SetCurrentWindow(window);

	/*
	 * If the main window was clicked then raise or
	 * lower the window with respect to its siblings.
	 */
	if (wid == x11Data->mainWid)
	{
		if (bp->button == Button1)
			XRaiseWindow(display, x11Data->mainWid);
		else
			XLowerWindow(display, x11Data->mainWid);
	}

	/*
	 * Clear any motion action.
	 */
	x11Data->isMotionAborted = FALSE;

	ClearMotionAction(window);

	/*
	 * If this is the main window and we have a parent
	 * window then start a window movement or resize action.
	 */
	if (window->parentWindow && (wid == x11Data->mainWid))
	{
		PixelCount borderWidth = GetBorderWidth(&window->settings);

		x11Data->moveOldRootX = bp->x_root;
		x11Data->moveOldRootY = bp->y_root;

		/*
		 * If the mouse was clicked in the lower right corner
		 * then start a resize action otherwise do a move.
		 */
		if ((bp->x >= (window->width - RESIZE_SIZE - borderWidth)) &&
			(bp->y >= (window->height - RESIZE_SIZE - borderWidth)))
		{
			x11Data->motionAction = MOTION_RESIZE;

			x11Data->moveStartWidth = window->width;
			x11Data->moveStartHeight = window->height;
		}
		else
		{
			x11Data->motionAction = MOTION_MOVE;
		}

		return;
	}

	/*
	 * If the button is not in the view sub-window
	 * then ignore it.
	 */
	if (wid != x11Data->viewWid)
		return;

	/*
	 * If we have a parent window and a true modifier key was
	 * used then start a window movement.  This lets the user
	 * move a window even if the border is missing.
	 */
	if (window->parentWindow && (bp->state & ALL_MOD_MASKS))
	{
		x11Data->moveOldRootX = bp->x_root;
		x11Data->moveOldRootY = bp->y_root;
		x11Data->motionAction = MOTION_MOVE;

		return;
	}

	/*
	 * Save the mouse position in the current object.
	 */
	SaveMousePosition(window, bp->x, bp->y);

	/*
	 * If there are any modifier keys then put us
	 * into the mouse position mode which does
	 * nothing except keep the mouse position updated
	 * for other windows to view if they want.
	 */
	if (bp->state & ALL_KEY_MASKS)
	{
		x11Data->motionAction = MOTION_POSITION;

		return;
	}

	/*
	 * There was not a modifier key, so we want to do dragging.
	 * But if input is not possible now then do nothing.
	 */
	if (!Life.canMakeInput)
	{
		return;
	}

	/*
	 * Calculate the row and column in the view that the
	 * button press was on.
	 */
	buttonRow = PixelToScreenCell(window, bp->y);
	buttonCol = PixelToScreenCell(window, bp->x);

	/*
	 * Save the current object and position for later use when
	 * the button is released so we can see if dragging was done.
	 */
	x11Data->dragBeginRow = numAdd(buttonRow, window->minRow);
	x11Data->dragBeginCol = numAdd(buttonCol, window->minCol);

	/*
	 * Remember the initial drag rectangle coordinates
	 * and set the motion action.
	 */
	x11Data->dragBeginX = ScreenCellToPixel(window, buttonCol);
	x11Data->dragBeginY = ScreenCellToPixel(window, buttonRow);
	x11Data->dragEndX = x11Data->dragBeginX;
	x11Data->dragEndY = x11Data->dragBeginY;
	x11Data->motionAction = MOTION_DRAG;

	/*
	 * Draw the initial zero sized rectangle.
	 */
	DrawDragRectangle(window);
	CopyPixmapToWindow(window);
}


/*
 * Handle a button motion event.
 * This is used to update the display of the drag rectangle.
 */
static void
DoButtonMotion(const XButtonEvent * bp)
{
	Window	wid;
	LifeWindow * window;
	X11Data * x11Data;
	NUMBER buttonRow;
	NUMBER buttonCol;
	XEvent latestEvent;

	/*
	 * Find the window which is doing the button motion.
	 */
	wid = bp->window;
	window = FindWindow(wid);

	if (window == 0)
		return;

	lastInputWindow = window;
	lastInputX = bp->x;
	lastInputY = bp->y;

	x11Data = (X11Data *) window->privateData;

	/*
	 * Consume any further button motion events
	 * for the same window since we are really only
	 * interested in the most recent one.
	 */
	while (XCheckWindowEvent(display, wid,
		ButtonMotionMask, &latestEvent))
	{
		bp = &latestEvent.xbutton;
	}

	/*
	 * If there is no motion action set or if it had been
	 * aborted then do nothing.
	 */
	if ((x11Data->motionAction == MOTION_NONE) ||
		x11Data->isMotionAborted)
	{
		return;
	}

	/*
	 * If the motion action is moving of a window then do that.
	 */
	if (x11Data->motionAction == MOTION_MOVE)
	{
		/*
		 * Get the incremental root position offset from the
		 * previous button event and save the new root position.
		 */
		PixelCount offsetX = bp->x_root - x11Data->moveOldRootX;
		PixelCount offsetY = bp->y_root - x11Data->moveOldRootY;

		x11Data->moveOldRootX = bp->x_root;
		x11Data->moveOldRootY = bp->y_root;

		/*
		 * Add the incremental offset to the window position
		 * and then move it there.
		 */
		window->positionX += offsetX;
		window->positionY += offsetY;

		XMoveWindow(display, x11Data->mainWid,
			window->positionX, window->positionY);

		/*
		 * Redraw the window border.
		 * This fixes glitches due to X Server bugs.
		 */
		DrawBorder(window);

		return;
	}

	/*
	 * If the motion action is resizing of a window then do that.
	 */
	if (x11Data->motionAction == MOTION_RESIZE)
	{
		PixelCount offsetX = bp->x_root - x11Data->moveOldRootX;
		PixelCount offsetY = bp->y_root - x11Data->moveOldRootY;

		x11Data->moveOldRootX = bp->x_root;
		x11Data->moveOldRootY = bp->y_root;

		ConfigureWindow(window, window->width + offsetX,
			window->height + offsetY);

		/*
		 * Redraw the border.
		 * This fixes glitches due to X Server bugs.
		 */
		DrawBorder(window);

		return;
	}

	/*
	 * Make sure the window is the view sub-window.
	 */
	if (wid != x11Data->viewWid)
	{
		ClearMotionAction(window);

		return;
	}

	/*
	 * Save the mouse position in the current object.
	 */
	SaveMousePosition(window, bp->x, bp->y);

	/*
	 * If we are in mouse positioning mode then that's
	 * all we want to do.
	 */
	if (x11Data->motionAction == MOTION_POSITION)
		return;

	/*
	 * If we can't safely do dragging stuff then clear
	 * the motion.
	 */
	if (!Life.canMakeInput ||
		(x11Data->motionAction != MOTION_DRAG))
	{
		ClearMotionAction(window);

		return;
	}

	/*
	 * This is the dragging of a rectangle action.
	 * Get the viewed row and column of the mouse.
	 */
	buttonRow = PixelToScreenCell(window, bp->y);
	buttonCol = PixelToScreenCell(window, bp->x);

	/*
	 * Clear the old drag rectangle and update the screen
	 * coordinates of the new rectangle.
	 */
	if (x11Data->isDragShown)
		DrawDragRectangle(window);

	x11Data->dragEndX = ScreenCellToPixel(window, buttonCol);
	x11Data->dragEndY = ScreenCellToPixel(window, buttonRow);

	/*
	 * Draw the new drag rectangle if we aren't currently
	 * running generations.
	 */
	if (window->currentObject->genLeft == 0)
	{
		DrawDragRectangle(window);
		CopyPixmapToWindow(window);
	}
}


/*
 * Handle a button up event.
 * This sets the current row and column of the object to the location
 * of the mouse.  The it executes a macro if one was assigned to the
 * button, or else marks the cells contained within the drag rectangle
 * if any.  This is only done if canMakeInput is true.
 */
static int
DoButtonUp(const XButtonEvent * bp)
{
	Window	wid;
	LifeWindow * window;
	X11Data * x11Data;
	LifeObject * obj;
	char * cp;
	Scale	scale;
	int motionAction;
	int	button;
	int	macro;

	/*
	 * Find the window which had the button up.
	 */
	wid = bp->window;
	window = FindWindow(wid);

	if (window == 0)
		return EOF;

	lastInputWindow = window;
	lastInputX = bp->x;
	lastInputY = bp->y;

	x11Data = (X11Data *) window->privateData;

	/*
	 * Get the motion that that was in progress if any
	 * and then clear it.  The most recent parameters are
	 * still valid though for our use below.
	 */
	motionAction = x11Data->motionAction;

	ClearMotionAction(window);

	/*
	 * If the motion action had been aborted then
	 * do nothing more.
	 */
	if (x11Data->isMotionAborted)
		return EOF;

	/*
	 * If movement was active then move the window to
	 * its final location and clear the movement mode.
	 */
	if (motionAction == MOTION_MOVE)
	{
		/*
		 * Get the incremental root position offset from the
		 * previous button event and save the new root position.
		 */
		PixelCount offsetX = bp->x_root - x11Data->moveOldRootX;
		PixelCount offsetY = bp->y_root - x11Data->moveOldRootY;

		x11Data->moveOldRootX = bp->x_root;
		x11Data->moveOldRootY = bp->y_root;

		/*
		 * Add the incremental offset to the window position
		 * and then move it there.
		 */
		window->positionX += offsetX;
		window->positionY += offsetY;

		XMoveWindow(display, x11Data->mainWid,
			window->positionX, window->positionY);

		/*
		 * Redraw the border.
		 * This fixes glitches due to X Server bugs.
		 */
		DrawBorder(window);

		return EOF;
	}

	/*
	 * If resizing was in progress then set the final size.
	 */
	if (motionAction == MOTION_RESIZE)
	{
		PixelCount offsetX = bp->x_root - x11Data->moveOldRootX;
		PixelCount offsetY = bp->y_root - x11Data->moveOldRootY;

		x11Data->moveOldRootX = bp->x_root;
		x11Data->moveOldRootY = bp->x_root;

		ConfigureWindow(window, window->width + offsetX,
			window->height + offsetY);

		return EOF;
	}

	/*
	 * This is either a single mouse click or a drag.
	 * If the window is not the view sub-window then
	 * ignore it.
	 */
	if (wid != x11Data->viewWid)
		return EOF;

	/*
	 * Save the mouse position in the current object.
	 */
	SaveMousePosition(window, bp->x, bp->y);

	/*
	 * If this was just the mouse position action then
	 * that's all we do.
	 */
	if (motionAction == MOTION_POSITION)
		return EOF;

	/*
	 * We want to move the object position, select a
	 * rectangle, and possibly execute a macro.
	 * If we can't generate any input then do nothing.
	 */
	if (!Life.canMakeInput)
		return EOF;

	/*
	 * Get the position in the viewed object.
	 */
	NUMBER buttonRow = PixelToScreenCell(window, bp->y);
	NUMBER buttonCol = PixelToScreenCell(window, bp->x);

	/*
	 * See what the current position of the pointer is
	 * in the current object, and how much movement there
	 * was from the old curser position of the object.
	 */
	obj = window->currentObject;
	scale = window->scale;

	NUMBER deltaRow = numSub(buttonRow,
		numSub(obj->currentInfo.row, window->minRow));

	NUMBER deltaCol = numSub(buttonCol,
		numSub(obj->currentInfo.col, window->minCol));

	if (scale > 1)
	{
		numAddrDivLong(&deltaRow, scale);
		numAddrDivLong(&deltaCol, scale);
	}

	/*
	 * Calculate the assign index from the mouse button number.
	 * Use any modifier key used with the buttons as three
	 * further possible indexes.  The get the macro which is
	 * assigned to the button, if any.
	 */
	switch (bp->button)
	{
		case Button1:
			button = 0;
			break;

		case Button2:
			button = 1;
			break;

		case Button3:
			button = 2;
			break;

		case Button4:
			button = 3;
			break;

		case Button5:
			button = 4;
			break;

		default:
			button = ASSIGNS;
			break;
	}

/*	if (bp->state & ALL_KEY_MASKS)	*/
/*		button += 5;	*/

	macro = (button < ASSIGNS) ? assigns[button] : '\0';

	/*
	 * Get ready to store commands to be executed.
	 */
	cp = genStrBuf;
	*cp = '\0';

	/*
	 * Generate a relative movement command to move the cursor
	 * position to the cell at the current mouse position.
	 */
	if (numIsNonZero(deltaRow) || numIsNonZero(deltaCol))
	{
		sprintf(cp, "%c%sj%sl", NULL_CMD,
			numToString(deltaRow), numToString(deltaCol));

		cp += strlen(cp);
	}

	/*
	 * If the motion action was dragging then
	 * set the drag area of the object and perhaps
	 * set up a marking of the selected cells.
	 */
	if (motionAction == MOTION_DRAG)
	{
		NUMBER dragEndRow = numAdd(buttonRow, window->minRow);
		NUMBER dragEndCol = numAdd(buttonCol, window->minCol);

		if (numIsLessOrEqual(x11Data->dragBeginRow, dragEndRow))
		{
			obj->dragMinRow = x11Data->dragBeginRow;
			obj->dragMaxRow = dragEndRow;
		}
		else
		{
			obj->dragMinRow = dragEndRow;
			obj->dragMaxRow = x11Data->dragBeginRow;
		}

		if (numIsLessOrEqual(x11Data->dragBeginCol, dragEndCol))
		{
			obj->dragMinCol = x11Data->dragBeginCol;
			obj->dragMaxCol = dragEndCol;
		}
		else
		{
			obj->dragMinCol = dragEndCol;
			obj->dragMaxCol = x11Data->dragBeginCol;
		}

		/*
		 * If a macro is not being executed and no generations
		 * are being run and the drag area is larger than one cell
		 * then set up to do a default action which is to mark the
		 * cells in the drag area.
		 */
		if ((macro == '\0') && (obj->genLeft == 0) &&
			(!numIsEqual(x11Data->dragBeginRow, dragEndRow) ||
			!numIsEqual(x11Data->dragBeginCol, dragEndCol)))
		{
			*cp++ = NULL_CMD;
			*cp++ = 'm';
			*cp++ = 's';
			*cp = '\0';
		}
	}

	/*
	 * See whether there is any macro assigned to the button.
	 * If so, then set up to execute the macro.
	 */
	if (macro != '\0')
	{
		*cp++ = NULL_CMD;
		*cp++ = ESC;
		*cp++ = macro;
		*cp = '\0';
	}

	/*
	 * If no commands were generated then return no character.
	 */
	if (genStrBuf[0] == '\0')
		return EOF;

	/**
	 * Set up to read the commands and return the first character.
	 */
	genStr = genStrBuf;

	return *genStr++ & 0xff;
}


/*
 * Return a key typed in one of our windows, or EOF if no such key was typed.
 * This can change the current window and therefore the current object.
 * In that case, we painfully attempt to save the command to be processed
 * by the command loop in its next iteration by returning a null command.
 */
static int
DoKeyDown(const XKeyEvent * kp)
{
	Window	wid;
	LifeWindow * window;
	X11Data * x11Data;
	BOOL isWindowChanged;
	KeySym	keySym;
	int	len;
	int	ch;
	int	macro;
	char buf[12];

	wid = kp->window;
	window = FindWindow(wid);

	if (window == 0)
		return EOF;

	lastInputWindow = window;
	lastInputX = kp->x;
	lastInputY = kp->y;

	x11Data = (X11Data *) window->privateData;

	/*
	 * If there was any motion action in progress
	 * then abort it and ignore this key.
	 */
	if (x11Data->motionAction != MOTION_NONE)
	{
		x11Data->isMotionAborted = TRUE;

		ClearMotionAction(window);

		return EOF;
	}

	/*
	 * Set the current window to this one and remember
	 * whether or not it was changed.  If the window had
	 * changed then we will have to store the character
	 * and return EOF now, so that the command loop will
	 * restart and use the changed window.
	 */
	isWindowChanged = SetCurrentWindow(window);

	/*
	 * Look up the key symbol.
	 */
	len = XLookupString((XKeyEvent *) kp, buf, sizeof(buf), &keySym, 0);

	/*
	 * Look for special keys which are either cursor movement keys,
	 * or else function keys to execute macros.  Only do this if
	 * canMakeInput is true.
	 */
	macro = '\0';
	ch = '\0';

	if (Life.canMakeInput)
	{
		switch (keySym)
		{
			case XK_Up:
				ch = (kp->state & ALL_KEY_MASKS) ? 'K' : 'k';
				break;

			case XK_Down:
				ch = (kp->state & ALL_KEY_MASKS) ? 'J' : 'j';
				break;

			case XK_Left:
				ch = (kp->state & ALL_KEY_MASKS) ? 'H' : 'h';
				break;

			case XK_Right:
				ch = (kp->state & ALL_KEY_MASKS) ? 'L' : 'l';
				break;

			case XK_F1:	macro = '1'; break;
			case XK_F2:	macro = '2'; break;
			case XK_F3:	macro = '3'; break;
			case XK_F4:	macro = '4'; break;
			case XK_F5:	macro = '5'; break;
			case XK_F6:	macro = '6'; break;
			case XK_F7:	macro = '7'; break;
			case XK_F8:	macro = '8'; break;
			case XK_F9:	macro = '9'; break;
		}
	}

	/*
	 * If it was a function key, then set the associated macro
	 * name as the next input character, and return the macro
	 * invocation character now.
	 */
	if (macro)
	{
		/*
		 * If the window was changed then store the complete
		 * macro command for later and return a null command now.
		 */
		if (isWindowChanged)
		{
			genStrBuf[0] = ESC;
			genStrBuf[1] = macro;
			genStrBuf[2] = '\0';
			genStr = genStrBuf;

			return NULL_CMD;
		}

		genStrBuf[0] = macro;
		genStrBuf[1] = '\0';
		genStr = genStrBuf;

		return ESC;
	}

	/*
	 * If the character wasn't set above then check for a normal key.
	 */
	if (ch == '\0')
	{
		if ((len != 1) || IsModifierKey(keySym))
			return EOF;

		ch = buf[0];
	}

	/*
	 * Translate returns into newlines.
	 */
	if (ch == '\r')
		ch = '\n';

	/*
	 * If the window had changed then store the character
	 * for the next call and return a null command now.
	 */
	if (isWindowChanged)
	{
		genStrBuf[0] = ch;
		genStrBuf[1] = '\0';
		genStr = genStrBuf;

		return NULL_CMD;
	}

	/*
	 * Return the character now.
	 */
	return (ch & 0xff);
}


/*
 * Here for a configuration change to our window.
 * This is typically a resize.
 */
static void
DoConfigure(const XConfigureEvent * cp)
{
	LifeWindow * window;
	X11Data * x11Data;

	/*
	 * Find the window which was reconfigured.
	 */
	window = FindWindow(cp->window);

	if (window == 0)
		return;

	x11Data = (X11Data *) window->privateData;

	/*
	 * If the values aren't being changed then do nothing.
	 */
	if ((cp->height == (x11Data->viewHeight + x11Data->statHeight)) &&
		(cp->width == x11Data->viewWidth))
	{
		return;
	}

	/*
	 * Recalculate values for the new width and height.
	 */
	ConfigureWindow(window, cp->width, cp->height);
}


/*
 * Show the cells around the cursor normally (showing each cell).
 * The absolute value of the scale factor is used as a magnification.
 */
static void
ViewNormal(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	ViewInfo * seenInfo = &x11Data->seenInfo;
	/*const*/ LifeObject * obj = window->currentObject;
	NUMBER minRow = window->minRow;
	NUMBER maxRow = window->maxRow;
	NUMBER minCol = window->minCol;
	NUMBER maxCol = window->maxCol;
	BOOL needClear = FALSE;
	ViewInfo usedInfo;
	ViewInfo currentInfo;

	//
	// Check whether or not the seen data in the window
	// is still valid.  If not then we will need to clear the
	// window and redraw it all.  This call also initialises the
	// seen information for the current view.
	//
	if (!IsSeenInfoValid(window))
		needClear = TRUE;

	if (needClear || (window->updateFlags & U_CLEAR))
	{
		ClearArea(window, 0, 0,
			numToLong(numSub(window->maxRow, window->minRow)) + 2,
			numToLong(numSub(window->maxCol, window->minCol)) + 2);

		x11Data->isCursorShown = FALSE;
		x11Data->isDragShown = FALSE;
		x11Data->isOutlineShown = FALSE;

		FillAuxViewData(seenInfo, CELL_CLEAR);
	}

	//
	// Reset the data offset in the info for another scan.
	//
	ResetView(seenInfo);

	//
	// Get the first non-empty row of the currently seen cells (if any).
	//
	int seenRowIndex = FindNextRowIndex(seenInfo);

	/*
	 * Get the data for the current cells of the object for the
	 * region being viewed and remember how many ON cells are seen.
	 */
	currentInfo = CalculateView(obj, VIEW_ID_OBJECT, 1,
		minRow, maxRow, minCol, maxCol);

	window->seeCount = currentInfo.onCount;

	//
	// Get the view of the overlay object if any.
	//
	ViewInfo overlayInfo = GetOverlayView(window, obj);

	//
	// Get the first row of the overlay object (if any).
	//
	int overlayRowIndex = FindNextRowIndex(&overlayInfo);

	//
	// Get the first nonempty row of the current cells (if any).
	//
	int currentRowIndex = FindNextRowIndex(&currentInfo);

	/*
	 * Get the data for the cells ever used by the object for the
	 * region being viewed, or else an empty info structure if
	 * tracks are currently invisible or not present.
	 * Note: We call the quadcell routine directly since the
	 * used cells always use that algorithm.
	 */
	if (obj->isTracksVisible && !qcIsEmpty(obj->usedQuadCells))
	{
//		usedInfo = qcCalculateView(obj->usedQuadCells,
//			VIEW_ID_TRACKS, FALSE, minRow, maxRow,
//			minCol, maxCol);
		InitEmptyView(&usedInfo);
	}
	else
	{
		InitEmptyView(&usedInfo);
	}

	//
	// Get the first nonempty row of the used cells (if any).
	//
	int usedRowIndex = FindNextRowIndex(&usedInfo);

	/*
	 * Scan the cells of the object row by row.
	 * This scan examines the current cells of the object,
	 * the used cells of the object, the seen cells of the object,
	 * and the cells of an overlay object  all at the same time.
	 */
	while ((currentRowIndex >= 0) || (usedRowIndex >= 0) ||
		(seenRowIndex >= 0) || (overlayRowIndex >= 0))
	{
		/*
		 * Find the minimum row index of the sets of cells.
		 */
		int rowIndex = PickMinimumIndex(currentRowIndex,
			usedRowIndex, overlayRowIndex, seenRowIndex);

		//
		// Examine the cells of the found row to draw them.
		// This will update the seen information structure.
		//
		ViewNormalRow(window, rowIndex, &currentInfo, &usedInfo,
			&overlayInfo, seenInfo);

		//
		// Now advance the current, used, overlay, and seen rows to
		// the next used ones if they were on the row just examined.
		// Note: the checks still work correctly if the appropriate
		// row index was negative when no more rows exist.
		//
		if (rowIndex == currentRowIndex)
			currentRowIndex = FindNextRowIndex(&currentInfo);

		if (rowIndex == usedRowIndex)
			usedRowIndex = FindNextRowIndex(&usedInfo);

		if (rowIndex == overlayRowIndex)
			overlayRowIndex = FindNextRowIndex(&overlayInfo);

		if (rowIndex == seenRowIndex)
			seenRowIndex = FindNextRowIndex(seenInfo);
	}

	//
	// Now overlay the where location data on top.
	//
	DrawAllWhereLocations(window);
}


/*
 * Examine the given information for updating the specified row.
 * The row is completely examined before drawing any of it so that we
 * can minimise the drawing operations.
 */
static void
ViewNormalRow(LifeWindow * window, int rowIndex,
	ViewInfo * currentInfo, ViewInfo * usedInfo,
	ViewInfo * overlayInfo, ViewInfo * seenInfo)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	//
	// Initialise the actions of the row to be ignored until we
	// find the cells which need updating.
	//
	int rowActions = CELL_IGNORE;

	memset(cellActions, CELL_IGNORE, seenInfo->colCount);

	//
	// Find the index of the first ON cell in the row (if any) for the
	// current object, the used object, the overlay object, and for the
	// seen cell if they are on the specified row index.  Note: These calls
	// work correctly even if there were no more rows left to be examined.
	//
	int currentColIndex = -1;
	int usedColIndex = -1;
	int overlayColIndex = -1;
	int seenColIndex = -1;

	if (currentInfo->rowIndex == rowIndex)
		currentColIndex = FindNextColIndex(currentInfo);

	if (usedInfo->rowIndex == rowIndex)
		usedColIndex = FindNextColIndex(usedInfo);

	if (overlayInfo->rowIndex == rowIndex)
		overlayColIndex = FindNextColIndex(overlayInfo);

	if (seenInfo->rowIndex == rowIndex)
		seenColIndex = FindNextColIndex(seenInfo);

	/*
	 * Keep looping while there are columns of interest in any of
	 * the sets of cells.
	 */
	while ((currentColIndex >= 0) || (usedColIndex >= 0) ||
		(overlayColIndex >= 0) || (seenColIndex >= 0))
	{
		int oldAction = CELL_CLEAR;
		int action = CELL_CLEAR;

		/*
		 * Find the minimum column index of the sets of cells.
		 */
		int colIndex = PickMinimumIndex(currentColIndex,
			usedColIndex, overlayColIndex, seenColIndex);

		if (colIndex == seenColIndex)
		{
			oldAction = GetAuxViewData(seenInfo, rowIndex, colIndex);
		}

		/*
		 * Determine what has happened to the cell at this
		 * column and set the action appropriately.
		 * Allow the overlay action to supercede the track action.
		 */
		if (colIndex == overlayColIndex)
			action = CELL_OVERLAY;

		if (colIndex == currentColIndex)
		{
			int state = GetAuxViewData(currentInfo, rowIndex, colIndex);

			if (state & STATE_MARK)
				action = CELL_MARK;
			else if (state & STATE_NEW)
				action = CELL_NORMAL;
			else if (state & STATE_ON)
				action = CELL_OLD;
			else if ((state & STATE_DEAD) && (action != CELL_OVERLAY))
				action = CELL_TRACK;
		}

		/*
		 * If there has been any change, then update it both in
		 * the seen info table for the next update, and also in
		 * the current row actions to be performed this update.
		 */
		if (action != oldAction)
		{
			SetAuxViewData(seenInfo, rowIndex, colIndex, action);

			cellActions[colIndex] = action;
			rowActions |= action;
		}

		//
		// This cell will have to be checked again on the next
		// update if and only if either of the current or the old
		// actions are not CELL_CLEAR.
		//
		SetViewCell(seenInfo, rowIndex, colIndex,
			((action != CELL_CLEAR) || (oldAction != CELL_CLEAR)));

		//
		// Now advance the current, used, overlay, and seen columns to
		// the next used ones on the row.  Note: the checks still work
		// correctly if the appropriate column index was negative
		// when no more columns exist.
		//
		if ((currentInfo->rowIndex == rowIndex) &&
			(colIndex == currentColIndex))
		{
			currentColIndex = FindNextColIndex(currentInfo);
		}

		if ((usedInfo->rowIndex == rowIndex) &&
			(colIndex == usedColIndex))
		{
			usedColIndex = FindNextColIndex(usedInfo);
		}

		if ((overlayInfo->rowIndex == rowIndex) &&
			(colIndex == overlayColIndex))
		{
			overlayColIndex = FindNextColIndex(overlayInfo);
		}

		if ((seenInfo->rowIndex == rowIndex) &&
			(colIndex == seenColIndex))
		{
			seenColIndex = FindNextColIndex(seenInfo);
		}
	}

	/*
	 * Now that the row is scanned, make the indicated changes.
	 * Separate passes are used for supposed X11 efficiency.
	 * Clear cells, then set normal ones, old ones, marked ones,
	 * and overlay ones, and finally tracked ones.  First clear the
	 * cursor if anything changed.
	 */
	if (rowActions != CELL_IGNORE)
		ClearCursor(window);

	if ((rowActions != CELL_IGNORE) && x11Data->isDragShown)
		DrawDragRectangle(window);

	if (rowActions & CELL_CLEAR)
		ClearRow(window, rowIndex);

	if (rowActions & CELL_NORMAL)
		DrawRow(window, rowIndex, CELL_NORMAL);

	if (rowActions & CELL_OLD)
		DrawRow(window, rowIndex, CELL_OLD);

	if (rowActions & CELL_MARK)
		DrawRow(window, rowIndex, CELL_MARK);

	if (rowActions & CELL_OVERLAY)
		DrawRow(window, rowIndex, CELL_OVERLAY);

	if (rowActions & CELL_TRACK)
		DrawRow(window, rowIndex, CELL_TRACK);
}


/*
 * Show the view around the cursor with an arbitrary positive scale factor.
 * Each pixel then represents one or more live cells in an N by N area.
 * If any of those cells are marked, then the pixel will show it.
 * This routine makes no attempt to preserve old data in the window.
 * Instead, the window is just completely rewritten on each call.
 */
static void
ViewScaled(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	const LifeObject * obj = window->currentObject;
	Scale scale = window->scale;
	NUMBER minRow = window->minRow;
	NUMBER maxRow = window->maxRow;
	NUMBER minCol = window->minCol;
	NUMBER maxCol = window->maxCol;
	ViewInfo currentInfo;
	ViewInfo usedInfo;

	//
	// Clear the window data before redrawing it all.
	//
	ClearPixmapArea(window, 0, 0,
		x11Data->viewWidth, x11Data->viewHeight);

	/*
	 * Make sure the scale value is positive.
	 * (This routine isn't called with negative scale values anyway.)
	 */
	if (scale < 1)
		scale = 1;

	//
	// Check whether or not the seen data in the window is still
	// valid, and if not then rebuild the view information.
	// Ignore the return value since we never need to clear the
	// window since it is all completely redrawn anyway.
	//
	IsSeenInfoValid(window);

	/*
	 * Get the data for the current cells of the object for the
	 * region being viewed and remember how many ON cells are seen.
	 */
	currentInfo = CalculateView(obj, VIEW_ID_OBJECT, scale,
		minRow, maxRow, minCol, maxCol);

	window->seeCount = currentInfo.onCount;

	//
	// Get the view of the overlay object if any.
	//
	ViewInfo overlayInfo = GetOverlayView(window, obj);

	int overlayRowIndex = FindNextRowIndex(&overlayInfo);

	//
	// Get the first nonempty row of the current cells (if any).
	//
	int currentRowIndex = FindNextRowIndex(&currentInfo);

	/*
	 * Get the data for the cells ever used by the object for the
	 * region being viewed, or else an empty info structure if
	 * tracks are currently invisible or not present.
	 */
	if (obj->isTracksVisible && !isCellOff(obj->usedQuadCells))
	{
//		usedInfo = qcCalculateScaledView(obj->usedQuadCells,
//			VIEW_ID_TRACKS, FALSE, scale,
//			minRow, maxRow, minCol, maxCol);
		InitEmptyView(&usedInfo);
	}
	else
	{
		InitEmptyView(&usedInfo);
	}

	//
	// Get the first nonempty row of the used cells (if any).
	//
	int usedRowIndex = FindNextRowIndex(&usedInfo);

	/*
	 * Scan the cells of the object row by row.
	 * This scan examines the current cells of the object and the
	 * used cells of the object at the same time.
	 */
	while ((currentRowIndex >= 0) || (usedRowIndex >= 0) ||
		(overlayRowIndex >= 0))
	{
		/*
		 * Find the minimum row index of the sets of cells.
		 */
		int rowIndex = PickMinimumIndex(
			currentRowIndex, usedRowIndex, overlayRowIndex, -1);

		//
		// Now draw the cells of the found row.
		//
		ViewScaledRow(window, rowIndex, &currentInfo,
			&usedInfo, &overlayInfo);

		//
		// Advance to the next row needing drawing if any.
		//
		if (rowIndex == currentRowIndex)
			currentRowIndex = FindNextRowIndex(&currentInfo);

		if (rowIndex == usedRowIndex)
			usedRowIndex = FindNextRowIndex(&usedInfo);

		if (rowIndex == overlayRowIndex)
			overlayRowIndex = FindNextRowIndex(&overlayInfo);
	}

	//
	// Now overlay the where location data on top.
	//
	DrawAllWhereLocations(window);

	//
	// Clear the various show flags since we rewrote the whole window.
	//
	x11Data->isCursorShown = FALSE;
	x11Data->isDragShown = FALSE;
	x11Data->isOutlineShown = FALSE;
}


/*
 * Examine the given information for updating the specified row for a
 * scale factor larger then one.  The row has already been cleared so we
 * only need to draw the non-OFF cells.  This routine makes no use of
 * the seen info since we don't care what is already shown in the window.
 */
static void
ViewScaledRow(LifeWindow * window, int rowIndex,
	ViewInfo * currentInfo, ViewInfo * usedInfo, ViewInfo * overlayInfo)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	//
	// Find the index of the first ON cell in the row (if any) for the
	// current object, the used object, and the overlay object if they
	// are on the specified row index.  Note: These calls work correctly
	// even if there were no more rows left to be examined.
	//
	int currentColIndex = -1;
	int usedColIndex = -1;
	int overlayColIndex = -1;

	if (currentInfo->rowIndex == rowIndex)
		currentColIndex = FindNextColIndex(currentInfo);

	if (usedInfo->rowIndex == rowIndex)
		usedColIndex = FindNextColIndex(usedInfo);

	if (overlayInfo->rowIndex == rowIndex)
		overlayColIndex = FindNextColIndex(overlayInfo);

	/*
	 * Keep looping while there are columns of interest in any of
	 * the sets of cells.
	 */
	while ((currentColIndex >= 0) || (usedColIndex >= 0) ||
		(overlayColIndex >= 0))
	{
		/*
		 * Find the minimum column index of the sets of cells.
		 */
		int colIndex = PickMinimumIndex(currentColIndex,
			usedColIndex, overlayColIndex, -1);

		/*
		 * Determine how to draw this pixel and do it.
		 * The region can have cells of multiple states,
		 * so we rank them in our preferred order.
		 */
		GC gc = viewTracksGC;

		if (colIndex == currentColIndex)
		{
			int state = GetAuxViewData(currentInfo, rowIndex, colIndex);

			if (isStatePartMark(state))
				gc = viewMarkGC;
			else if (isStatePartNew(state))
				gc = viewCellGC;
			else if (isStatePartOn(state))
				gc = viewOldGC;
			else if (isStatePartDead(state))
				gc = viewTracksGC;
		}

		//
		// If this is an overlay cell then draw it if we only
		// have tracks from the main object.
		//
		if ((gc == viewTracksGC) && (colIndex == overlayColIndex))
			gc = overlayGC;

		/*
		 * Note: Do not attempt to collect points in a table and
		 * then do XDrawPoints on them.  That is slower than this!
		 */
		XDrawPoint(display, x11Data->viewPixmap, gc, colIndex, rowIndex);

		//
		// Now advance the current, used, and overlay column to the
		// next used one on the row.
		//
		if ((currentInfo->rowIndex == rowIndex) &&
			(colIndex == currentColIndex))
		{
			currentColIndex = FindNextColIndex(currentInfo);
		}

		if ((usedInfo->rowIndex == rowIndex) &&
			(colIndex == usedColIndex))
		{
			usedColIndex = FindNextColIndex(usedInfo);
		}

		if ((overlayInfo->rowIndex == rowIndex) &&
			(colIndex == overlayColIndex))
		{
			overlayColIndex = FindNextColIndex(overlayInfo);
		}
	}
}


/*
 * Get the overlay view if any for the specified object and scale.
 * An empty view is returned if there is no overlay object.
 */
static ViewInfo
GetOverlayView(const LifeWindow * window, const LifeObject * obj)
{
	//
	// If there is no overlay object then return an empty view.
	//
	if (!obj->isOverlayActive)
	{
		ViewInfo overlayInfo;

		InitEmptyView(&overlayInfo);

		return overlayInfo;
	}

	int scale = window->scale;

	if (scale < 1)
		scale = 1;

	//
	// There is an overlay object.
	// Calculate the view for it using the shifted area according
	// to the relative current positions of the current object and
	// the overlay object.
	//
	LifeObject * overObj = obj->overlayObject;

	NUMBER rowOffset = numSub(
		overObj->currentInfo.row, obj->currentInfo.row);

	NUMBER colOffset = numSub(
		overObj->currentInfo.col, obj->currentInfo.col);

	NUMBER newMinRow = numAdd(window->minRow, rowOffset);
	NUMBER newMaxRow = numAdd(window->maxRow, rowOffset);
	NUMBER newMinCol = numAdd(window->minCol, colOffset);
	NUMBER newMaxCol = numAdd(window->maxCol, colOffset);

	return CalculateView(overObj, VIEW_ID_OVERLAY, scale,
		newMinRow, newMaxRow, newMinCol, newMaxCol);
}


/*
 * Clear the specified area of the pixmap for the window's view.
 * This fills the specified rectangle with the window's background color.
 */
static void
ClearPixmapArea(LifeWindow * window,
	Coord x, Coord y, Count width, Count height)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	//
	// If the view window's background color isn't the same as
	// the clearGC's foreground color, then update the clearGC.
	//
	if (clearColorIndex != x11Data->viewBackgroundIndex)
	{
		clearColorIndex = x11Data->viewBackgroundIndex;

		XSetForeground(display, clearGC, FindColor(clearColorIndex));
	}

	//
	// Now fill the rectangle of the pixmap as specified.
	//
	XFillRectangle(display, x11Data->viewPixmap, clearGC,
		x, y, width, height);
}


/*
 * Determine whether or not the seen information for the window is still
 * valid.  This is true if the window is still showing the same region
 * of the object.  If the information is invalid then the view is initialised
 * for the new region.  The data values in the array are not initialised.
 */
static BOOL
IsSeenInfoValid(LifeWindow * window)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	Scale scale = window->scale;
	ViewInfo * info = &x11Data->seenInfo;

	NUMBER minRow = window->minRow;
	NUMBER maxRow = window->maxRow;
	NUMBER minCol = window->minCol;
	NUMBER maxCol = window->maxCol;

	//
	// Negative scale factors are treated as 1 here.
	// The row and column values depend on them, but that doesn't
	// affect the view info calculations.
	//
	if (scale < 1)
		scale = 1;

	//
	// If the scale and coordinates haven't changed then the
	// window's view info structure is still good for use.
	//
	if (info->isValid && (scale == info->scale) &&
		numIsEqual(minRow, info->minRow) &&
		numIsEqual(maxRow, info->maxRow) &&
		numIsEqual(minCol, info->minCol) &&
		numIsEqual(maxCol, info->maxCol))
	{
		return TRUE;
	}

	//
	// The view info is no longer correct.  Free any previous
	// buffer and allocate a new one of the correct size.
	//
	FreeView(info);

	InitView(info, VIEW_ID_PRIVATE, TRUE, scale, minRow, maxRow, minCol, maxCol);

	return FALSE;
}


/*
 * Clear the cells in a row that require that done.
 * This is only called for negative scale factors.
 */
static void
ClearRow(LifeWindow * window, Coord rowIndex)
{
	const char * cp = cellActions;
	int colCount = numToLong(numSub(window->maxCol, window->minCol)) + 1;
	Coord colIndex = 0;

	while (colIndex < colCount)
	{
		int count;
		cp = memchr(cp, CELL_CLEAR, colCount - colIndex);

		if (cp == 0)
			return;

		colIndex = (cp - cellActions);
		count = 0;

		while ((colIndex < colCount) && (*cp == CELL_CLEAR))
		{
			colIndex++;
			cp++;
			count++;
		}

		ClearArea(window, rowIndex, colIndex - count, 1, count);
	}
}


/*
 * Clear the cells in the specified rectangular region of the screen.
 * This first blanks out the area, and then puts back the grid if
 * the grid is being shown.  This is only called for negative
 * scale factors.
 */
static void
ClearArea(
	LifeWindow *	window,
	Coord		rowIndex,
	Coord		colIndex,
	Count		rowCount,
	Count		colCount
)
{
	Scale	mag = window->scale;

	if (mag < 0)
		mag = -mag;

	if (mag < 1)
		mag = 1;

	/*
	 * First clear the area to the background color.
	 */
	ClearPixmapArea(window, colIndex * mag, rowIndex * mag,
		mag * colCount, mag * rowCount);

	/*
	 * If we don't show a grid then we are done.
	 */
	if ((mag < GRID_MAG) || !GetIsGridEnabled(&window->settings))
		return;

	/*
	 * Draw the appropriate type of grid.
	 */
	switch (GetGridType(&window->settings))
	{
		case GRID_TYPE_DOTS:
			DrawGridDots(window, rowIndex, colIndex,
				rowCount, colCount);

			break;

		case GRID_TYPE_CHECKERBOARD:
			DrawGridCheckerBoard(window, rowIndex, colIndex,
				rowCount, colCount);

			// fall into draw lines to do that too

		case GRID_TYPE_LINES:
			DrawGridLines(window, rowIndex, colIndex,
				rowCount, colCount);

			break;

		default:
			break;
	}
}


/*
 * Draw the grid by putting dots at the corners of the cells.
 */
static void
DrawGridDots(
	LifeWindow *	window,
	Coord		rowIndex,
	Coord		colIndex,
	Count		rowCount,
	Count		colCount
)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	const LifeObject * obj = window->currentObject;
	Drawable drawable = x11Data->viewPixmap;
	Scale	mag = window->scale;
	PixelCoord	xp;
	PixelCoord	yp;
	Coord	col;
	Count	count;
	BOOL	isLarge;
	int	point1Count;
	int	point2Count;
	XPoint	points1[MAX_DRAW];
	XPoint	points2[MAX_DRAW];

	if (mag < 0)
		mag = -mag;

	if (mag < 1)
		mag = 1;

	/*
	 * Show the grid by putting dots at the bottom right corner of
	 * each cell.  Use another color for some rows and columns
	 * depending on the grid size.
	 */
	point1Count = 0;
	point2Count = 0;
	isLarge = FALSE;

	/*
	 * If the magnification is large enough then we want to show
	 * larger grid points.  Subtract one from the coordinate and add
	 * one to the count to draw the larger grid points correctly.
	 */
	if (mag >= GRID_BIG_MAG)
	{
		isLarge = TRUE;
		rowIndex--;
		colIndex--;
		rowCount++;
		colCount++;
	}

	yp = rowIndex * mag + mag - 1;

	while (rowCount-- > 0)
	{
		xp = colIndex * mag + mag - 1;
		count = colCount;
		col = colIndex;

		while (count-- > 0)
		{
			/*
			 * If either of the point tables are full,
			 * then draw both sets of points to empty them.
			 */
			if ((point1Count >= MAX_DRAW - 6) ||
				(point2Count >= MAX_DRAW - 6))
			{
				if (point1Count > 0)
				{
					XDrawPoints(display, drawable,
						gridGC, points1,
						point1Count, CoordModeOrigin);
				}

				if (point2Count > 0)
				{
					XDrawPoints(display, drawable,
						viewMarkGC, points2,
						point2Count, CoordModeOrigin);
				}

				point1Count = 0;
				point2Count = 0;
			}

			/*
			 * Add the point to the appropriate table of points,
			 * depending on the current screen row and column.
			 * Make the differing grid colors depend on the
			 * coordinates relative to the object's origin so
			 * that they don't change when the view is shifted.
			 */
			NUMBER gridRow = numAddLong(
				numSub(window->minRow, obj->originInfo.row),
				rowIndex + 1);

			NUMBER gridCol = numAddLong(
				numSub(window->minCol, obj->originInfo.col),
				col + 1);

			if (GridMod(window, gridRow) && GridMod(window, gridCol))
			{
				point2Count += AddGridPoints(
					&points2[point2Count], xp, yp, isLarge);
			}
			else
			{
				point1Count += AddGridPoints(
					&points1[point1Count], xp, yp, isLarge);
			}

			xp += mag;
			col++;
		}

		yp += mag;
		rowIndex++;
	}

	/*
	 * Finish drawing any leftover points.
	 */
	if (point1Count > 0)
	{
		XDrawPoints(display, drawable, gridGC, points1,
			point1Count, CoordModeOrigin);
	}

	if (point2Count > 0)
	{
		XDrawPoints(display, drawable, viewMarkGC, points2,
			point2Count, CoordModeOrigin);
	}
}


/*
 * Draw the grid by putting lines at the edges of the cells.
 * The lines are drawn thicker for the major grid lines.
 */
static void
DrawGridLines(
	LifeWindow *	window,
	Coord		rowIndex,
	Coord		colIndex,
	Count		rowCount,
	Count		colCount
)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	const LifeObject * obj = window->currentObject;
	Drawable drawable = x11Data->viewPixmap;
	NUMBER gridMod;
	Scale mag = window->scale;
	PixelCoord	minX;
	PixelCoord	maxX;
	PixelCoord	minY;
	PixelCoord	maxY;
	PixelCoord	curX;
	PixelCoord	curY;

	if (mag < 0)
		mag = -mag;

	if (mag < 1)
		mag = 1;

	minX = colIndex * mag;
	minY = rowIndex * mag;

	maxX = minX + mag * colCount - 1;
	maxY = minY + mag * rowCount - 1;

	/*
	 * Draw the horizontal rows.
	 */
	gridMod = numAddLong(
		numSub(window->minRow, obj->originInfo.row), rowIndex);

	curY = minY;

	while (rowCount-- > 0)
	{
		XDrawLine(display, drawable, gridLineGC,
			minX, curY, maxX, curY);

		if (GridMod(window, gridMod) == 0)
		{
			XDrawLine(display, drawable, gridLineGC,
				minX, curY + 1, maxX, curY + 1);
		}

		gridMod = numInc(gridMod);

		curY += mag;
	}

	/*
	 * Draw the vertical rows.
	 */
	gridMod = numAddLong(
		numSub(window->minCol, obj->originInfo.col), colIndex);

	curX = minX;

	while (colCount-- > 0)
	{
		XDrawLine(display, drawable, gridLineGC,
			curX, minY, curX, maxY);

		if (GridMod(window, gridMod) == 0)
		{
			XDrawLine(display, drawable, gridLineGC,
				curX + 1, minY, curX + 1, maxY);
		}

		gridMod = numInc(gridMod);

		curX += mag;
	}
}


/*
 * Draw the grid by drawing a checkerboard.
 * This knows the background is already cleared so we only
 * have to draw half of the squares.
 */
static void
DrawGridCheckerBoard(
	LifeWindow *	window,
	Coord		rowIndex,
	Coord		colIndex,
	Count		rowCount,
	Count		colCount
)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	Drawable drawable = x11Data->viewPixmap;
	Scale	mag = window->scale;
	const LifeObject * obj = window->currentObject;
	Coord	ri;
	Coord	ci;

	if (mag < 0)
		mag = -mag;

	if (mag < 1)
		mag = 1;

	/*
	 * Iterate over the rows and columns.
	 * Use the window's minimum and maximum row numbers
	 * so that the grid is stable as we move.
	 */
	long xx = numToLong(numSub(obj->originInfo.row, window->minRow));
	long yy = numToLong(numSub(obj->originInfo.col, window->minCol));
	long sum = xx + yy + rowIndex + colIndex;

	for (ri = 0; ri < rowCount; ri++)
	{
		for (ci = 0; ci < colCount; ci++)
		{
			long test = sum + ri + ci;

			if ((test & 0x1) == 0)
			{
				XFillRectangle(display, drawable,
					checkerGC, (colIndex + ci) * mag,
					(rowIndex + ri) * mag, mag, mag);
			}
		}
	}
}


/*
 * Add one or more points to the specified array for a grid point.
 * We either add one point or else 5 points in a cross pattern
 * depending on our parameter.  The number of added points is returned.
 * The caller must make sure that the array is large enough for the
 * added points.
 */
static int
AddGridPoints(XPoint * point, PixelCoord xp, PixelCoord yp, BOOL isLarge)
{
	/*
	 * Add the center point of the cross.
	 */
	point->x = xp;
	point->y = yp;

	/*
	 * If the point isn't large then that's all we do.
	 */
	if (!isLarge)
		return 1;

	/*
	 * Add the other points to complete the cross.
	 */
	point++;
	point->x = xp + 1;
	point->y = yp;
	point++;
	point->x = xp;
	point->y = yp + 1;
	point++;
	point->x = xp - 1;
	point->y = yp;
	point++;
	point->x = xp;
	point->y = yp - 1;

	return 5;
}


/*
 * Draw the cells in a row that have the specified action for them.
 * This is used for normal cells, marked cells, previously used cells,
 * or overlaid cells.  This is only called for negative scale values.
 */
static void
DrawRow(LifeWindow * window, Coord rowIndex, int action)
{
	X11Data * x11Data = (X11Data *) window->privateData;

	Drawable drawable = x11Data->viewPixmap;
	Scale mag = window->scale;
	CellShape cellShape = GetCellShape(&window->settings);
	const char * cp = cellActions;
	int colCount = numToLong(numSub(window->maxCol, window->minCol)) + 1;
	Coord colIndex = 0;

	int		offset;
	GC		gc;
	int		count;
	XPoint *	pt;
	XRectangle *	rect;
	XArc *		arc;
	XRectangle	rectangles[MAX_DRAW];
	XArc		arcs[MAX_DRAW];
	XPoint		points[MAX_DRAW];

	if (action == CELL_NORMAL)
		gc = viewCellGC;
	else if (action == CELL_OLD)
		gc = viewOldGC;
	else if (action == CELL_MARK)
		gc = viewMarkGC;
	else if (action == CELL_OVERLAY)
		gc = overlayGC;
	else
		gc = viewTracksGC;

	/*
	 * Make sure the magnification is reasonable.
	 */
	if (mag < 0 )
		mag = -mag;

	if (mag < 1)
		mag = 1;

	//
	// Get the full size of the cell to be drawn.
	//
	int size = ((mag < 3) ? mag : (mag - 1));

	//
	// If this is an overlay then shrink the size.
	//
	if ((action == CELL_OVERLAY) && (size > 5))
		size -= 5;

	/*
	 * Check for a small shape size.
	 */
	if (GetIsCellShapeSmall(&window->settings))
		size = (size * 2) / 3;

	if (size < 1)
		size = 1;

	offset = (mag - size + 1) / 2;

	if (action == CELL_OVERLAY)
		offset++;

	/*
	 * Use the square cell shape if the magnification is too small.
	 */
	if ((cellShape == SHAPE_CIRCLE) && (mag < CIRCLE_MAG))
		cellShape = SHAPE_SQUARE;

	if ((cellShape == SHAPE_DIAMOND) && (mag < DIAMOND_MAG))
		cellShape = SHAPE_SQUARE;

	/*
	 * Loop over the columns of the row looking for those
	 * columns which have the specified action.
	 */
	while (colIndex < colCount)
	{
		/*
		 * Look for the next column having the specified action.
		 */
		cp = memchr(cp, action, colCount - colIndex);

		if (cp == 0)
			return;

		colIndex = (cp - cellActions);
		rect = rectangles;
		arc = arcs;
		pt = points;
		count = 0;

		/*
		 * Check for successive cells having the same action.
		 * We can sometimes store the coordinates for one call.
		 */
		while ((colIndex < colCount) && (*cp == action) &&
			(count < MAX_DRAW))
		{
			//
			// If we are drawing an overlay then clear the cell
			// first so nothing old shows around the edges.
			//
			if (action == CELL_OVERLAY)
				ClearArea(window, rowIndex, colIndex, 1, 1);

			/*
			 * If diamonds are being drawn then do that immediately.
			 * (There isn't a way to draw more than one of them at a time.)
			 */
			if (cellShape == SHAPE_DIAMOND)
			{
				int colBase = colIndex * mag + offset;
				int rowBase = rowIndex * mag + offset;

				points[0].x = colBase + size / 2;
				points[0].y = rowBase;
				points[1].x = colBase + size;
				points[1].y = rowBase + size / 2;
				points[2].x = colBase + size / 2;
				points[2].y = rowBase + size;
				points[3].x = colBase;
				points[3].y = rowBase + size / 2;

				XFillPolygon(display, drawable, gc, points, 4,
					Convex, CoordModeOrigin);

				colIndex++;
				cp++;

				continue;
			}

			/*
			 * If circles are being drawn then add another one to the arc array.
			 */
			if (cellShape == SHAPE_CIRCLE)
			{
				arc->x = colIndex * mag + offset;
				arc->y = rowIndex * mag + offset;
				arc->width = size;
				arc->height = size;
				arc->angle1 = 0;
				arc->angle2 = 360*64;
				arc++;
			}
			else if (mag == 1)
			{
				/*
				 * Add a point to the point array.
				 */
				pt->x = colIndex;
				pt->y = rowIndex;
				pt++;
			}
			else
			{
				/*
				 * Add a rectangle to the rectangle array.
				 */
				rect->x = colIndex * mag + offset;
				rect->y = rowIndex * mag + offset;
				rect->width = size;
				rect->height = size;
				rect++;
			}

			count++;
			colIndex++;
			cp++;
		}

		/*
		 * Now draw the stored items if any.
		 */
		if (cellShape == SHAPE_CIRCLE)
		{
			XFillArcs(display, drawable, gc, arcs, count);
		}

		if ((cellShape == SHAPE_SQUARE) && (mag == 1))
		{
			XDrawPoints(display, drawable, gc, points, count,
				CoordModeOrigin);
		}

		if ((cellShape == SHAPE_SQUARE) && (mag > 1))
		{
			XFillRectangles(display, drawable, gc, rectangles,
				count);
		}
	}
}


/*
 * Draw all the where location letters if they are enabled (including
 * the origin), and also any of the auxillary lines through the locations
 * if they are enabled.
 */
static void
DrawAllWhereLocations(LifeWindow * window)
{
	const LifeObject * obj = window->currentObject;
	int index;

	if (!obj->isLocationsEnabled)
		return;

	for (index = 0; index < 26; index++)
		DrawWhereLocation(window, &obj->whereInfo[index]);

	DrawWhereLocation(window, &obj->originInfo);
	DrawWhereLocation(window, &obj->currentInfo);
}


/*
 * Draw the id and auxillary lines for a where location if enabled.
 * The only lines drawn are horizontal, vertical, and 45 degree diagonals,
 * therefore the clipping for the endpoints is easy.
 */
static void
DrawWhereLocation(LifeWindow * window, const LocationInfo * info)
{
	NUMBER row = info->row;
	NUMBER col = info->col;
	NUMBER minRow = window->minRow;
	NUMBER maxRow = window->maxRow;
	NUMBER minCol = window->minCol;
	NUMBER maxCol = window->maxCol;
	NUMBER left = numAddLong(numSub(col, minCol), 2);
	NUMBER right = numAddLong(numSub(maxCol, col), 2);
	NUMBER up = numAddLong(numSub(row, minRow), 2);
	NUMBER down = numAddLong(numSub(maxRow, row), 2);
	NUMBER leftDown = numMin(left, down);
	NUMBER leftUp = numMin(left, up);
	NUMBER rightDown = numMin(right, down);
	NUMBER rightUp = numMin(right, up);

	if (!info->isEnabled)
		return;

	if (info->isIdShown)
		DrawCharacter(window, info->id, row, col);

	if (info->isHorizShown)
	{
		DrawWhereLine(window, row, row,
			numSubLong(minCol, 2), numAddLong(maxCol, 2));
	}

	if (info->isVertShown)
	{
		DrawWhereLine(window, numSubLong(minRow, 2),
			numAddLong(maxRow, 2), col, col);
	}

	if (info->isForwShown)
	{
		DrawWhereLine(window,
			numAdd(row, leftDown), numSub(row, rightUp),
			numSub(col, leftDown), numAdd(col, rightUp));
	}

	if (info->isBackShown)
	{
		DrawWhereLine(window,
			numSub(row, leftUp), numAdd(row, rightDown),
			numSub(col, leftUp), numAdd(col, rightDown));
	}
}


/*
 * Draw a line through the centers between the specified two cells.
 */
static void
DrawWhereLine(LifeWindow * window, NUMBER minRow, NUMBER maxRow,
	NUMBER minCol, NUMBER maxCol)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	PixelCoord x1;
	PixelCoord x2;
	PixelCoord y1;
	PixelCoord y2;

	//
	// Avoid the drawing if nothing will be seen.
	//
	if (numIsLess(minRow, window->minRow) &&
		numIsLess(maxRow, window->minRow))
	{
		return;
	}

	if (numIsLess(minCol, window->minCol) &&
		numIsLess(maxCol, window->minCol))
	{
		return;
	}

	if (numIsGreater(minRow, window->maxRow) &&
		numIsGreater(maxRow, window->maxRow))
	{
		return;
	}

	if (numIsGreater(minCol, window->maxCol) &&
		numIsGreater(maxCol, window->maxCol))
	{
		return;
	}

	//
	// Convert the coordinates and do the drawing.
	//
	x1 = ScreenCellToPixel(window, numSub(minCol, window->minCol));
	x2 = ScreenCellToPixel(window, numSub(maxCol, window->minCol));
	y1 = ScreenCellToPixel(window, numSub(minRow, window->minRow));
	y2 = ScreenCellToPixel(window, numSub(maxRow, window->minRow));

	XDrawLine(display, x11Data->viewPixmap, whereGC, x1, y1, x2, y2);
}


/*
 * Draw a character in the middle of the specified cell.
 */
static void
DrawCharacter(LifeWindow * window, int ch, NUMBER row, NUMBER col)
{
	X11Data * x11Data = (X11Data *) window->privateData;
	char buf[2];

	//
	// Short cut drawing off of the screen.
	//
	if (numIsLess(row, window->minRow) ||
		numIsGreater(row, numAddLong(window->maxRow, 2)) ||
		numIsLess(col, window->minCol) ||
		numIsGreater(col, numAddLong(window->maxCol, 2)))
	{
		return;
	}

	buf[0] = ch;
	buf[1] = '\0';

	PixelCoord x = ScreenCellToPixel(window, numSub(col, window->minCol));
	PixelCoord y = ScreenCellToPixel(window, numSub(row, window->minRow));

	x -= XTextWidth(font, buf, 1) / 2;
	y += x11Data->charAscent / 2;

	XDrawString(display, x11Data->viewPixmap, pathGC, x, y, buf, 1);
}


/*
 * Try to allocate a color value for the given color name.
 * This associates the name of a color, its pixel value, and an index.
 * Returns the index of the allocated color if it was known with
 * the color pixel value returned.  Returns -1 if not known,
 * with the desired or alternate color pixel value returned.
 */
static ColorIndex
AllocateColor(const char * name, Color * retColor, Color alternate)
{
	XColor	cellColor;
	XColor	nameColor;

	/*
	 * See if the color name is already known.
	 * If so then return the pixel value from our table.
	 */
	ColorIndex index = FindColorIndex(name);

	if (index >= 0)
	{
		*retColor = colorTable[index];

		return index;
	}

	/*
	 * We must try to allocate a new color.
	 * If this fails then return the alternate color.
	 */
	if (!XAllocNamedColor(display, colorMap, name, &cellColor, &nameColor))
	{
		*retColor = alternate;

		return COLOR_INDEX_NONE;
	}

	/*
	 * The color was found.
	 * Return the color pixel value in any case.
	 */
	*retColor = cellColor.pixel;

	/*
	 * Try to add the new color name to the color table.
	 * If this succeeds then save the pixel value in our table.
	 */
	index = AddColorForDevice(name);

	if (index >= 0)
		colorTable[index] = cellColor.pixel;

	/*
 	 * Return the allocated index.
	 */
	return index;
}


/*
 * Get the color value for a color index.
 * Returns black if the index is invalid.
 */
static Color
FindColor(ColorIndex index)
{
	if ((index < 0) || (index >= Life.colorCount))
		return black;

	return colorTable[index];
}


/*
 * Set the view text color to the one given by the specified
 * color index.  If the index is changed then the foreground
 * color in the viewTextGC is changed.
 */
static void
SetTextGC(ColorIndex colorIndex)
{
	if (colorIndex == textColorIndex)
		return;

	XSetForeground(display, viewTextGC, FindColor(colorIndex));

	textColorIndex = colorIndex;
}


/*
 * Set the normal cell color to the one given by the specified
 * color index.  If the index is changed then the foreground
 * color in the viewCellGC is changed.
 */
static void
SetCellGC(ColorIndex colorIndex)
{
	if (colorIndex == cellColorIndex)
		return;

	XSetForeground(display, viewCellGC, FindColor(colorIndex));

	cellColorIndex = colorIndex;
}


/*
 * Set the marked cell color to the one given by the specified
 * color index.  If the index is changed then the foreground
 * color in the viewMarkGC is changed.
 */
static void
SetMarkGC(ColorIndex colorIndex)
{
	if (colorIndex == markColorIndex)
		return;

	XSetForeground(display, viewMarkGC, FindColor(colorIndex));

	markColorIndex = colorIndex;
}


/*
 * Set the tracks color to the one given by the specified
 * color index.  If the index is changed then the foreground
 * color in the viewTracksGC is changed.
 */
static void
SetTracksGC(ColorIndex colorIndex)
{
	if (colorIndex == tracksColorIndex)
		return;

	XSetForeground(display, viewTracksGC, FindColor(colorIndex));

	tracksColorIndex = colorIndex;
}


/*
 * Set the old cells color to the one given by the specified
 * color index.  If the index is changed then the foreground
 * color in the viewOldGC is changed.
 */
static void
SetOldGC(ColorIndex colorIndex)
{
	if (colorIndex == oldColorIndex)
		return;

	XSetForeground(display, viewOldGC, FindColor(colorIndex));

	oldColorIndex = colorIndex;
}


/*
 * Get the status window for displaying text into.
 * This depends on the value of the statusLocation setting.
 * This returns either our own window, our parent window, or null.
 */
static LifeWindow *
GetStatusWindow(LifeWindow * window)
{
	StatusLocation location = GetStatusLocation(&window->settings);

	if (location == STATUS_LOCATION_SELF)
		return window;

	if (location == STATUS_LOCATION_PARENT)
		return window->parentWindow;

	return 0;
}


#endif

/* END CODE */
